//? cpmangle.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_ANGLE_H_
#define CPM_ANGLE_H_
//////////////////////////////////////////////////////////////////////////
// Description: A class describing angles and the 1-dimensional
//   torus
//////////////////////////////////////////////////////////////////////////
#include <cpmc.h>
#include <cpmtypes.h>
#include <cpmv.h>

namespace CpmGeo{

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::Word;
   using CpmRoot::C;
   using CpmArrays::V;

enum AngleUnit { CYCLE, HOUR, DEG, RAD};
   // cycle (full angle), hour, degree, radian
enum AngleRange { POS, NEG};
   // always positive, negative allowed
enum AngleHex {DEC, MIN, SEC};
   // full decimal, minutes, minutes and seconds
// friend functions need to be declared as belonging to the
// present namespace
class Angle;
Word AngleToWord(const Angle& alpha,
   AngleUnit u, AngleRange r, AngleHex h, Z digits);
Angle mean(const CpmArrays::V<Angle>& phi);
Angle mean(const CpmArrays::V<Angle>& angles,
   const CpmArrays::V<R>& weights);
Angle RealToAngle(R x, AngleUnit u, AngleHex h);
Angle realToAngle(R c);

class Angle{ // angles
   // An instance of Angle is an angle, the manifold of all angles is
   // a 1-dimensional torus. There are two interpretations.
   // (i) Angle as a coordinate on a circle (as a curve) and
   // (ii) angle as a descriptor of a rotation arround a given axis.
   // Both are closely related due to the fact, that the rotations act
   // transitively and effectively on a circle. For defining a proper set
   // of methods we have not to make a choice in favor of one of these
   // interpretations.
   // Interpretation (i) endows Angle most directly with the structure of
   // an oriented metric space, and (ii) with the structure of an
   // Abelian group.

   static void normalizeAngle(R& a0);
      // changes a0 into a value in the interval (-Pi,Pi] by adding
      // a multiples of 2*Pi

   R a_; // is read as alpha, is the measure of the angle as arc length
      // on the unit circle (for effective usage of built-in
      // trigonometric functions. The range is by definition (-Pi,Pi].

   void normalize(void){ normalizeAngle(a_);}

public:
   typedef Angle Type;
   typedef R ScalarType;

   CPM_IO
   CPM_ORDER

   CPM_SUM_PLAIN
   CPM_DIFFERENCE_PLAIN
      // Addition and subtraction among angles as borrowed from the
      // mathematical model by residue classes in R 'modulo 2*Pi'.
      // A further mathematical model is given by
      // { z \in C : |z|=1 }. In  this model addition of angles corresponds
      // to multiplication of complex numbers and subtraction of angles
      // corresponds  to division of numbers. If we are simply given a
      // torus (e.g. as a thin ring of polished metal) then we need to mark
      // on it a point as origin and an arrow as direction in order map it
      // to any of the mathematical models considered here.
      // This implies that the algebraic structure of an abelian group
      // exhibited by both mathematical models is not inherent in the
      // geometry of the pure torus but depends on the selection of an
      // origin and a direction (orientation) as indicated above. Obviously
      // the selected origin plays the role of the neutral element of the
      // abelian group. Notice that the pure torus is a metric space
      // where the distance between two points is the length of the
      // curve along the torus (i.e. the shortest arc) which connects the
      // points.

   CPM_SCALAR_M
   // The meaning of multiplication with real numbers derives by
   // continuity from the meaning of multiplication by rational
   // numbers. Let us consider first multiplication by a natural
   // number n. The proper definition follows from the group
   // structure: alpha*n := alpha + ... + alpha (n terms).
   // Multiplication by 1/m for natural number:
   // alpha*(1/m) := smallest angle beta for which beta*m=Angle(),
   // where Angle() is the zero angle which is the unit element of
   // the group. Smallness, obviously, refers to the distance to
   // the unit element. Obviously, one may add (2*Pi)/m to beta
   // (let the result be called beta') and one also has
   // beta'*m=Angle().

   CPM_CONJUGATION
      // An idempotent conjugation operation. If the angle is interpreted
      // as the phase angle of a complex number, this corresponds to
      // the complex conjugation of the number. This is a natural
      // geometric operations for angles, resulting from the distinction
      // of one angle as the zero-angle (origin on torus).

   Word nameOf()const{ return Word("Angle");}

   R operator /(const Angle&)const;
      // alpha/beta is the number r for which beta*r=alpha.
      // Notice the comment to ScalarType on multiplying
      // Angles by numbers. This is not defined in residue
      // classes modulo 2*Pi.
      // Products of angles don't give angles; they are
      // omitted here.

   Angle operator+(R shift)const{ return Angle(a_+shift,false);}
      // Returning the angle which results from *this by adding an arc
      // which may be any length (e.g. much larger than 2*Pi) and sign.
      // This gives Angle the structure of an affine space.

// constructors

   Angle(void):a_(0.){}
      // angle zero

   explicit Angle(R angle, bool deg=true);
      // Constructor for Angles out of real numbers. The first argument
      // is always interpreted as an angle in degrees, unless deg==0
      // Never enable automatic conversion. Would result in ambiguous
      // arithmetics. Input needs not to be normalized.

   Angle(R vAngle, AngleUnit au);
      // creates an angle which is vAngle*au

   explicit Angle(C const& z):a_(z.arg()){}
      // Constructs the phase angle (polar angle) of z

   C expi()const{ return C(1.,a_,"polar");}
      // Returns the complex number exp(i*a). Thus Angle now has a
      // bi-directional efficient interface to C.

// automatic conversion in one direction Angle-->R is OK

   operator R(void)const{return a_;}
   // cast to R for making trigonometric functions defined for angles
   // No longer used in the implementation of Angle.

   R toR()const{ return a_;}
   // Explicit conversion in a standard manner. The implementation of
   // Angle now relies on this.

   R cut( R c)const;
   //: cut
   // Returns a value in the range (c*degree-2*Pi,c*degree] which from
   // a_ is derived by adding or subtracting a multiple of 2*Pi.
   // To have such a facility turned out to be very useful in
   // dealing with angular motion of stepper motors.

   void qun_(Z n);
      //: quantize
      // Changes angle *this into the nearest of n equi-spaced
      // angles which are equi-distributed over the whole range.
      // For n<1, no action.

   R toDeg(Z i=0)const;
      //: to degrees
      // Returns the angle *this in degrees. For the default
      // value 0 of the argument, the range is the standard range
      // (-180,180], and for 1 the range is [0,360).

   Word toWord(AngleUnit u, AngleRange r, AngleHex h, Z digits=1)const;
      // Returns a string which gives the angle in the units
      // given by u, in the range indicated by r, in a hexagesimal
      // subdivision characterized by h and with a number of figures
      // after the decimal point is given by digits. If digits is -1,
      // there is one figure behind the decimal point and this is rounded
      // to 0 or 5 corresponding to the accuracy with which the hour
      // circle of my C5 telescope can be read. The meaning of the
      // parameters follows from the explanation to the
      // enumerations AngleUnit, AngleRange, and AngleHex.

   friend Word AngleToWord(const Angle& alpha,
      AngleUnit u, AngleRange r, AngleHex h, Z digits=1)
      { return alpha.toWord(u,r,h,digits);}

// modifying an angle
   void setArc(R alpha);
      // input is an arc (i.e. something like a, but need not be
      // normalized)

// absolute value
   R abs(void)const{ return (a_>=0. ?  a_ : -a_);}
      // absolute value: positive arc as a number (distance from zero
      // angle)

   R dis(const Angle& phi)const;
      //: distance
      // Lets angles form a metric space. Is symmetric in the arguments.
      // Implementation corrected 2016-03-25 (symmetry was not guaranteed
      // before)

   bool isPos()const{ return a_>=0;}
      // returns true if the angle is in quadrant 1 or 2 (by definition,
      // 0 belongs to quadrant 1, Pi to quadrant 2 ) and false else

   void pos(){ if (a_<0) a_=-a_;}
      // makes the angle positive by conjugation

// mean values taking into account the embedding in the complex numbers
   friend Angle mean(const CpmArrays::V<Angle>& phi);
      // mean of an array of angles.
      // Polar angle of exp(i*phi[1])+...+exp(i*phi[n]).
      // There seems to be no easy way to define this on R/~
      // where a1 ~ a2 <==> a1 = a2 (mod 2Pi)

   friend Angle mean(const CpmArrays::V<Angle>& angles,
      const CpmArrays::V<R>& weights);
      // weighted mean of an array of angles
      // Polar angle of
      // weights[1]*exp(i*phi[1])+...+weights[n]*exp(i*phi[n])

   friend Angle RealToAngle(R x, AngleUnit u, AngleHex h);
      // Returns an angle alpha which is given by a real number
      // according the format indicated by the unit u and the
      // subdivision schema indicated by h.

   friend Angle realToAngle(R c){ return Angle(c);}
      // Converts a real number, interpreted as an arc into a
      // instance of class Angle.

   static V<R> smoothAngleList(const V<Angle>& va);
      // Given a list va of Angles, we return a list res of corresponding
      // R's such that
      //    res[i]==va[i] mod(2*Pi)
      // where adding or subtracting of 2*Pi is done such that the
      // resulting list is as smooth as possible. More precisely we
      // assure that for each i
      // | res[i+i]-res[i]| < |res[i+1]+2*Pi-res[i]|
      // and
      // | res[i+i]-res[i]| < |res[i+1]-2*Pi-res[i]|
      // As a result of this the res[i] may grow to much larger values
      // than 2*Pi as in a list of successive angular positions of a
      // rotating wheel.

    static V<R> smoothList(const V<R>& vr);
      // Similar to previous function. The argument list is not a list
      // of angles but a list of their conversions to numbers (based
      // on the RAD unit).

    static Angle deg1, deg90, deg180, deg270, deg360;
};

} // namespace

#endif
