//? cpmc.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_C_H_
#define CPM_C_H_
/*
   Description: Declaration of a class of complex numbers
       using the system provided type <complex> for implementation was
       tried. But this class has only very limited ammount of
       functions and my component access via [] can't be implemented
       here since the components are private.
       Division by zero will write a message, return 0 and continue.
       Test_c<C> d(10,1,1) gave a result which ~10-7 (including read
       write difference which is OK.
*/

#include <cpmword.h>
#include <complex>
   // connection with std complex numbers is useful when interfacing
   // with the Eigen library.

namespace CpmRoot{

class C;
// These are declarations of functions in scope CpmRoot
// most of them have a friend declaration within class C
// which enables a more efficient implementation.
R arg(C const& z);
C log10(C const& z);
C pow(C const& z, Z const& n);
C pow(C const& z, C const& n);
C sin(C const& z);
C cos(C const& z);
C tan(C const& z);
C cot(C const& z);
C sinh(C const& z);
C cosh(C const& z);
C tanh(C const& z);
C coth(C const& z);
C arcsin(C const& z);
C arccos(C const& z);
C arctan(C const& z);
C arccot(C const& z);
C arcsinh(C const& z);
C arccosh(C const& z);
C arctanh(C const& z);
C arccoth(C const& z);
C operator /(C const& z, R const& s);
bool isVal(C const& z);

class C { // complex numbers

   R re, im;
   typedef C Type;

public:

//// general infrastructure, in part with inline implementation
// This is essentially the general interface of templates Vr<> and Fr<>
// appart from the fact that here no function is virtual since we don't
// need to derive from this 'ultimate' class.

// order and comparison
   CPM_ORDER

// I/O operations
   CPM_IO

// test devices
   C ran(Z j=0)const;
// value of res.re is in the open interval (-|re|,|re|)
// value of res.im is in the open interval (-|im|,|im|)

   R dis(C const& y)const;
   C test(Z)const;
// descriptors
   Word nameOf(void)const{ return Word("C");}

   Word toWord()const;
      //: to word
      // for a nice printable representation

   Z hash(void) const;
      //: hash value

   C net(Z i=0)const{ if (i==1) return C(1,0); else return C(0,0);}
      //: neutrals
      // net(0): neutral element of addition,
      // net(1): neutral element of multiplication

   C neg()const{ return C(-re,-im);}
      //: negative

   C inv(void)const;
      //: inversion
      // Returns C(0,0) upon division by zero, but creates
      // warnings on cpmcerr if this happens. The total
      // number of warnings that can arrise in this way is limited
      // so that this file can't grow too much by this mechanism.

   C operator !(void)const{ return inv();}

// complex conjugation
   C con(void)const{ return C(re,-im);}
   void con_(){im=-im;}

   C operator~(void)const{ return C(re,-im);}
   //: ~
   // instead of z.con() we may write ~z

   C operator|(C const& z)const{ return con()*z;}
   //: (|) scalar product
   // The combination ~z1*z2 can also be written as (z1|z2)

// more specific topics

   static C I;  // 'imaginary unit'
   static C one; // number one = C(1,0)
   static C Ibar; // conjugate of I = - I = C(0,-1)

   C(void):re(0.),im(0.){}

   explicit C(R const& u, R const& v=0. ):re(u),im(v){}
      // constructor from real and imaginary part
      // no automatic conversion from R to C

   explicit C(std::complex<R> const& z):re(z.real()),im(z.imag()){}
      // construction from std::complex

   C(C const& z):re(z.re),im(z.im){}
      // standard construction operator

   C(R r, R phi, Word w):
   re(r*cpmcos(phi)),im(r*cpmsin(phi)){}
      // constructor from polar coordinates
      // In calling the constructor it is helpful to
      // use w="polar"

   R real(void)const{ return re;}

   R imag(void)const{ return im;}

   std::complex<R> std()const{ return std::complex<R>{re,im};}

   C& operator=(C const& c){ re=c.re; im=c.im; return *this;}
      // standard assignment operator

   C& operator=(std::complex<R> const& c)
   { re=c.real(); im=c.imag(); return *this;}

   Z dim(void)const{ return 2;} // dimension as a real linear space
      //: dimension

   R arg(void)const{ return cpmarg(re,im);}
      //: argument
      // The value is in the interval (-Pi,Pi] independent
      // of the convention followed by the system's ::atan2 function

   R abs(bool careful=true)const
   { return  careful ? cpmrho(re,im) : cpmsqrt(re*re+im*im);}
      //: rho
      // common symbol for the polar radius

   R absFast(void)const
   { return cpmsqrt(re*re+im*im);}
      //: rho fast
      // No provision against overflow by sqaring potentially large
      // numbers.

   R absSqr(void)const{ return re*re+im*im;}
      //: absolute (value) squared

   R abs2(void)const{ return re*re+im*im;}
      //: absolute (value) squared

   bool isVal(void)const{ return cpmiva(re)&&cpmiva(im);}
      //: is valid

   C scaleTo(R const& r)const;
      // returns a complex number of absolute value |r| and the direction
      // of *this for positive r and antidirection for negative r.

   C timesI(void)const{ return C(-im,re); }

   C dividedByI(void)const{ return C(im,-re);}

   C dividedBy2I(void)const{ return C(im*0.5,re*(-0.5));}

   void pol_(R r, R phi);
      //: polar
      // changes *this into r*exp(i*phi)
      // 'modern' indication of mutating nature of the function

   static C pol(R r, R phi){C z; z.pol_(r,phi); return z;}
      //: polar
      // construction from polar coordinates as a static
      // function instead of a constructor

   void pol_1_2(R& r, R& phi)const{ r=abs(); phi=arg();}
      //: polar
      // 'modern' indication of reference nature of first
      // and second argument

// heritage versions of functions related to polar coordinates
   void polar(R const& r, R const& phi){ pol_(r,phi);}
      // changes *this into r*exp(i*phi)
      // not conforming to the C+- naming convention

   void toPolar(R& r, R& phi)const;
      // after call r and phi have the values radius() and arg()

// end of heritage versions of functions related to polar coordinates

   R nor_(R r=1)
      //: normalize
   {
      R a=abs();
      if(a!=0){R fac=r/a;re*=fac;im*=fac;}
      else{re=r;}
      return a;
   }
      // turns *this into a C of length r and returns the norm of
      // the original in order not to loose information. If *this was
      // zero, it will be turned into  a 'standard C'
      // of length r.

   C sqrt()const;
   C sqr()const;
   C exp()const;
   C ln()const;

   C pow(C const& p)const;
   C sin()const;
   C cos()const;
   C tan()const;
   C cot()const;
   C sinh()const;
   C cosh()const;
   C tanh()const;
   C coth()const;
   C arcsin()const;
   C arccos()const;
   C arctan()const;
   C arccot()const;
   C arcsinh()const;
   C arccosh()const;
   C arctanh()const;
   C arccoth()const; // these added 2005-08-23

   friend R arg(C const& z);
   friend R rho(C const& z);
   static C expi(R phi){ return C(cpmcos(phi),cpmsin(phi));}
   friend C log10(C const& z);
   friend C pow(C const& z, const Z& n);
   friend C pow(C const& z, C const& n);
   friend C sin(C const& z);
   friend C cos(C const& z);
   friend C tan(C const& z);
   friend C cot(C const& z);
   friend C sinh(C const& z);
   friend C cosh(C const& z);
   friend C tanh(C const& z);
   friend C coth(C const& z);
   friend C arcsin(C const& z);
   friend C arccos(C const& z);
   friend C arctan(C const& z);
   friend C arccot(C const& z);
   friend C arcsinh(C const& z);
   friend C arccosh(C const& z);
   friend C arctanh(C const& z);
   friend C arccoth(C const& z);

// components for uniformity with arrays:  z.re=z[1], z.im=z[2]
// no exceptions !!!

   R& operator[](int i){ return i==2 ? im : re;}

   R const& operator[](int i)const{ return i==2 ? im : re;}

// mutating arithmetics
// There are a few more operations to be defined (such as C+R)
// which are not prepared in the interfaces; so everything
// is done explicitely
   C& operator +=(C const& x){re+=x.re; im+=x.im; return *this;}
   C& operator +=(R const& s){re+=s; return *this;}

   C& operator -=(C const& x){re-=x.re; im-=x.im; return *this;}
   C& operator -=(R const& s){re-=s; return *this;}

   C& operator *=(C const& x)
     {R re0=re; re=re*x.re-im*x.im; im=im*x.re+re0*x.im; return *this;}

   C& operator /=(C const& x){ return operator *=(x.inv());}

   C& operator *=(R const& s){re*=s; im*=s; return *this;}
   C& operator /=(R const& s);

// generating arithmetics
   C operator -(void)const{ return C(-re,-im);}
   C operator +(C const& z)const
      { return C(re+z.re,im+z.im);}
   friend C operator +(R const& s, C const& z2);
   C operator +(R const& s)const
      { return C(re+s,im);}
   C operator -(C const& z)const
      { return C(re-z.re,im-z.im);}
   friend C operator -(R const& s, C const& z2);
   C operator -(R const& s)const
      { return C(re-s,im);}
   C operator *(C const& z)const
      { return C(re*z.re-im*z.im,im*z.re+re*z.im);}
   friend C operator *(R const& s, C const& z2);
   C operator *(R const& s)const
       { return C(s*re,s*im);}
   C operator /(C const& z)const
      { return (*this)*z.inv();}
   friend C operator /(C const& z, R const& s);
};

// These are definitions of functions in scope CpmRoot, declared friend
// in class C.
inline C operator +(R const& s, C const& z2)
      { return C(z2.re+s,z2.im);}
inline C operator -(R const& s, C const& z2)
      { return C(s-z2.re,-z2.im);}
inline C operator *(R const& s, C const& z2)
       { return C(s*z2.re,s*z2.im);}
// The following functions need not to be friends of C.
// their argument in ways that need them to be declared friends of C.
inline bool isVal(C const& x){ return x.isVal();}
inline C ran(C const& x, Z j){ return x.ran(j);}
inline R dis(C const& x1, C const& x2){ return x1.dis(x2);}
inline C test(C const& x, Z cpl){ return x.test(cpl);}
inline Z hash(C const& x){ return x.hash();}
inline R abs(C const& x){ return x.abs();}
inline R rho(C const& x){ return x.abs();}
inline C net(C const& x, Z i=0){ return x.net(i);}
inline C inv(C const& x){ return x.inv();}
inline C con(C const& x){ return x.con();}
inline C sqrt(C const& z){ return z.sqrt();}
inline R Re(C const& z){return z.real();}
inline R Im(C const& z){return z.imag();}
inline C exp(C const& z){return z.exp();}
inline C expI(R phi){ return C::expi(phi);}
      // even for Angle alpha, we can write directly expI(alpha) since
      // CpmGeo::Angle gets automatically converted to an R-valued arc
      // So, instead of saying
      // C z; z.polar(r,phi);
      // we may, much nicer, say:
      // C z=r*expI(phi);
inline C ln(C const& z){ return z.ln();}
inline C log(C const& z){ return z.ln();}
inline C asin(C const& z){ return CpmRoot::arcsin(z);}
inline C acos(C const& z){ return CpmRoot::arccos(z);}
inline C atan(C const& z){ return CpmRoot::arctan(z);}
inline C acot(C const& z){ return CpmRoot::arccot(z);}
inline C asinh(C const& z){ return CpmRoot::arcsinh(z);}
inline C acosh(C const& z){ return CpmRoot::arccosh(z);}
inline C atanh(C const& z){ return CpmRoot::arctanh(z);}
inline C acoth(C const& z){ return CpmRoot::arccoth(z);}

} // namespace
#endif
