//? cpmfr.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_FR_H_
#define CPM_FR_H_
/*
   Purpose: Define classes describing function-like objects with
      arithmetic operations

*/

#include <cpmfa.h>
#include <cpmvo.h>

namespace CpmFunctions{

   using namespace CpmStd;

   using CpmRoot::R;
   using CpmRoot::Word;

template <class X, class Y>

class Fr: public Fa<X,Y> { // version of Fa with rich interface

   static Z imax;
   static Z cpl;
      // means for implementing order and equality
      // by comparison of function values ( actually all
      // values would be needed for being mathematically exact)
      // Another way of looking at the matter is that imax and cpl define
      // an equivalence relation on Fr<X,Y> and order and equality are
      // defined exactly on the set of equivalunce classes. The larger
      // imax and cpl are, the smaller the equivalence classes

public:

   typedef Fr<X,Y> Type;

   Fr(void):Fa<X,Y>(){}
      // default constructor

   Fr(const F<X,Y>& g):Fa<X,Y>(g){}
      // upcast constructor

   Fr(const Fo<X,Y>& g):Fa<X,Y>(g){}
      // upcast constructor

   Fr(const Fa<X,Y>& g):Fa<X,Y>(g){}
      // upcast constructor, note Fa = Fa

   Fr(const Fr<X,Y>& g):Fa<X,Y>(g){}
      // copy constructor

   Fr(FncObj<X,Y>* fop):Fa<X,Y>(fop){}
      // constructor from pointers to FncObj

// Construction from function pointers

   Fr( Y (*f)(const X& )):Fa<X,Y>(f){}
   Fr( Y (*f)(X)):Fa<X,Y>(f){}

// Construtor for constant function

   Fr(const Y& y1):Fa<X,Y>(y1){}
      // takes always the value y1, irrespective of x

   Fr(CpmArrays::V<X> const& xl, CpmArrays::V<Y> const& yl);
      // Construction from a list of x-values and a list of corresponding
      // y-values. Makes a staircase-function since real interpolation
      // assumes R-related operations in Y which we don't want to assme
      // here.

   virtual F<X,Y>* clone()const{ return new Fr(*this);}

   CPM_IO
   CPM_TEST_X
   CPM_DESCRIPTORS
   CPM_CONJUGATION

   Y operator()(X const& x)const;
      // redefinition with better error handling

   R compare(Word top, Fr<X,Y> const& rhs)const;

};

template <class X, class Y>
Y Fr<X,Y>::operator()(const X& x)const
      // notice the * operator from class P<>
{
   using CpmArrays::P;
   try{
      return (P< FncObj<X,Y> >::operator()())(x);
   }
   catch(...){
      CpmRoot::Word wx=CpmRoot::Name<X>()(X());
      Y y=Y();
      CpmRoot::Word wy=CpmRoot::Name<Y>()(y);
      cpmerror("trouble in Y Fa<X,Y>::operator()(const X& x)const\
      for X type "&wx&" and Y type "&wy);
      return y; // to avoid warning
   }
}

// helper classes for implementation

namespace aux{
template <class X, class Y>
class CONJFUO : public FncObj<X,Y>{
   const Fr<X,Y> f;
public:
   CONJFUO(const Fr<X,Y>& f_):f(f_){}
   Y operator()(const X& x)const
   { return CpmRoot::conT<Y>(f(x));}
};

template <class X, class Y>
class RANDFUO : public FncObj<X,Y>{
   const Z i;
   const Z tvs;
   const Z memory;
public:
   RANDFUO(const Z& i_, Z tvs_, Z mem_ ):i(i_),tvs(tvs_),
      memory(mem_){}
   Y operator()(X const& x)const
      // never invoke ran() since then the random functions
      // depend on history
   {
   // would make operator() a mutating operator
   // this would have rather opaque implications
      Z iAct=(i==0 ? memory : i);
      Y ys;
      ys=CpmRoot::testT<Y>(ys,tvs);
      ys=CpmRoot::ranT<Y>(ys,iAct);
      Z j=CpmRoot::hashT<X>(x);
      if (j==0)
         return ys;
      else
         return CpmRoot::ranT<Y>(ys,j);
         // notice that for j==0 ran behaves not as a function
         // (value depends on how often the routine was called)
   }
};

template <class X, class Y>
class IPOLFUO: public FncObj<X,Y>{
   // IPOL stands for interpolation. However, here we don't really
   // interpolate, we use staircase approximation instead.
   // This needs no assumptions concerning Y.
   const CpmArrays::V<X> xl;
   const CpmArrays::V<Y> yl;
public:
   Y operator()(const X& x)const;
   IPOLFUO( const CpmArrays::V<X>& xlIn, const CpmArrays::V<Y>& ylIn ):
      xl(xlIn),yl(ylIn){}
};

template <class X, class Y>
Y IPOLFUO<X,Y>::operator()(X const& x)const
{
   Z i,n=xl.dim();
   Word loc("IPOLFUO<X,Y>::operator()");
   cpmassert(yl.dim()>=n,loc);
   CpmArrays::Vo<R> dis(n);
   for (i=1;i<=n;i++){
      X xd=x-xl[i];
      dis[i]=CpmRoot::absT<X>(xd);
   }
   Z j=dis.indInf();
   return yl[j];
}

} // aux

// end of helper classes

template <class X, class Y>
Z Fr<X,Y>::imax=8;

template <class X, class Y>
Z Fr<X,Y>::cpl=8;

template <class X, class Y>
Fr<X,Y>::Fr(const CpmArrays::V<X>& xl,
   const CpmArrays::V<Y>& yl)
   // a rather crude interpolation based on weighted means
   // of function values from the list
{
   *this=Fr(new aux::IPOLFUO<X,Y>(xl,yl));
}

template <class X, class Y>
Fr<X,Y> Fr<X,Y>::con(void)const
{
   return Fr<X,Y>(new aux::CONJFUO<X,Y>(*this));
}

template <class X, class Y>
Fr<X,Y> Fr<X,Y>::ran(Z i)const
{
   static Z mem=37;
   if (i==0) mem++;
   return Fr<X,Y>(new aux::RANDFUO<X,Y>(i,cpl,mem));
}

template <class X, class Y>
Word Fr<X,Y>::nameOf(void)const
{
   Word wi="Fr<";
   X x; Y y;
   Word wx=CpmRoot::Name<X>()(x);
   Word wy=CpmRoot::Name<Y>()(y);
   return wi&wx&","&wy&">";
}

template <class X, class Y>
Word Fr<X,Y>::toWord(void)const
{
   ostringstream ost;
   prnOn(ost);
   return Word(ost.str());
}

template <class X, class Y>
Fr<X,Y> Fr<X,Y>::test(Z i)const
{
   Y y0;
   Y yt=CpmRoot::testT<Y>(y0,i);
   Fr<X,Y> res(yt);
   imax=i;
   cpl=i;
   return res;
}

template <class X, class Y>
bool Fr<X,Y>::prnOn(ostream& out)const
{
   out<<endl<<"// printing of "<<nameOf()<<endl;
   out<<endl<<"// Number of (x,y)-pairs: "<<endl;
   out<<imax<<endl;
   X x;
   Y y;
   X xs=CpmRoot::testT<X>(x,cpl);
   for (Z i=1;i<=imax;i++){
      x=CpmRoot::ranT<X>(xs,i);
      y=(*this)(x);
      bool b;
      out<<endl<<"// value pair number "<<i<<" of "<<imax;
      out<<endl<<"// x = "<<endl;
      b=CpmRoot::prnT<X>(x,out);
      if (!b) return false;
      out<<endl<<"// y = "<<endl;
      b=CpmRoot::prnT<Y>(y,out);
      if (!b) return false;
   }
   return true;
}

template <class X, class Y>
bool Fr<X,Y>::scanFrom(istream& in)
// reading value pairs and creating an interrpolated function
// from a file that may contain comments
{
   static const Z readLimit=512;
   Z nList;
   if (!CpmRoot::read(nList,in)) return false;
   cpmassert(nList>0,"Fr<X,Y>::scanFrom");
   cpmassert(nList<=readLimit,"Fr<X,Y>::scanFrom");
   CpmArrays::V<X> xl(nList);
   CpmArrays::V<Y> yl(nList);
   for (Z i=1;i<=nList;i++){
      if (!CpmRoot::scanT<X>(xl[i],in)) return false;
      if (!CpmRoot::scanT<Y>(yl[i],in)) return false;
   }
   *this=Fr<X,Y>(xl,yl);
   return true;
}

template <class X, class Y>
Z Fr<X,Y>::hash(void)const
// returns a key value to *this. This should be not
// expensive to calculate. Therfore the number of function
// evaluation is restricted by imax
{
   Z h=0;
   X x,xs;
   Y y;
   xs=CpmRoot::testT<X>(x,cpl);
   for (Z i=1;i<=imax;i++){
      x=CpmRoot::ranT<X>(xs,i);
      y=(*this)(x);
      h^=CpmRoot::hashT<Y>(y);
         // bit operation since += may give overflow
   }
   return h;
}

template <class X, class Y>
R Fr<X,Y>::absSqr(void)const
// returns as an absolute value a mean absolute value of function values
// where the arguments are chosen as a random sequence.
{
try{
   X x,xs;
   Y y;
   xs=CpmRoot::testT<X>(x,cpl);
   R res=0;
   for (Z i=1;i<=imax;i++){
      x=CpmRoot::ranT<X>(xs,i);
      y=(*this)(x);
      R a=CpmRoot::absT<Y>(y);
      res+=a*a;
   }
   cpmassert(imax>0,"Fr<X,Y>::absSqr");
   res/=imax;
   return cpmsqrt(res);
}
catch(...){
   cpmerror("trouble in R Fr<X,Y>::abs(void)const");
   return 0; // to avoid warning
}
}

template <class X, class Y>
R Fr<X,Y>::abs(void)const
{
   return cpmsqrt(absSqr());
}

template <class X, class Y>
R Fr<X,Y>::dis(const Fr<X,Y>& x)const
{
   Fr<X,Y> diffVector=*this;
   diffVector-=x;
   R da=diffVector.abs();
   R xa=x.abs();
   R ya=abs();
   return CpmRoot::disDefFun(xa,ya,da);
}

} // namespace

#endif
