//? cpminterfaces.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_INTERFACES_H_
#define CPM_INTERFACES_H_
/*

   Description: Declaration macros for interfaces for
      CPM classes. No macro argument, assuming that the
      client class will typedef her name as Type; some macros
      assume that the client class also defines ScalarType
      and CloneBase

*/
#include <cpmnumbers.h>
   // includes cpmdefinitions.h
#include <cpmbasicinterfaces.h>
#include <cpmmpi.h>
   // contains valid (though trivial) declarations even for
   // CPM_USE_MPI not defined

// Notice that using types Z and R in macros is only safe as
// CpmRoot::Z and CpmRoot::R since they might be used in a context
// where a corresponding using directive is not available.

// defining member functions send and rec(ieve) for communication between
// processes if the message passing interface library is being used
// (for parallel programming)
//using CpmMPI::Com;
// Definition of send builds on function prnOn, definition of rec builds
// on function scanFrom. The macros CPM_MPI_... will always be used after
// macros which declare these functions prnOn and scanFrom.

#if defined(CPM_USE_MPI)

   #define CPM_MPI_O\
   bool send(Z dest, CpmRoot::Z tag, CpmMPI::Com comm=CpmMPI::Com()\
   )const\
      { \
         std::ostringstream ost;\
         CpmRoot::Z wrtPrcMem=CpmRoot::wrtPrc;\
         CpmRoot::wrtPrc=18;\
         bool b=prnOn(ost);\
         CpmRoot::wrtPrc=wrtPrcMem;\
         return b ? comm.sendStr(ost.str(),dest,tag): false;\
      }

   #define CPM_MPI_I\
      bool rec(CpmRoot::Z source, CpmRoot::Z tag, CpmMPI::Com\
       comm=CpmMPI::Com())\
      {\
         std::string s;\
         bool res=comm.recStr(s,source,tag);\
         std::istringstream ist(s);\
         bool b=scanFrom(ist);\
         return b ? res : false;\
      }

#else // !defined(CPM_USE_MPI)

   #define CPM_MPI_O\
   bool send(CpmRoot::Z dest, CpmRoot::Z tag, CpmMPI::Com comm=\
   CpmMPI::Com())const{ dest; tag; comm; return true;}

   #define CPM_MPI_I\
      bool rec(CpmRoot::Z source, CpmRoot::Z tag, CpmMPI::Com\
       comm=CpmMPI::Com()){ source; tag; comm; return true;}

#endif

// input and output functions; the function read is the common interface
// to both built-in types and Cpm classes to read from commented files
// ( which does not work for R and Z via >>)

// all IMPL-MACROS get completed by their corresponding
// MPI-functionality.
// Then it is guaranteed that all classes define proper inter-process
// communication.
// Notice that for non-class types like Z and R, there is no re-definition
// of operators >> an <<.

#define CPM_IO_IMPL\
   friend std::ostream& operator<<(std::ostream& out, Type const& x)\
      { CpmRoot::IO<Type>().o(x,out); return out;}\
   friend std::istream& operator>>(std::istream& in, Type& x)\
      { CpmRoot::IO<Type>().i(x,in); return in;}\
      CPM_MPI_O\
      CPM_MPI_I

// declaration macros without public qualifications
// with and without virtual qualification.

#define CPM_IO_V\
   virtual bool prnOn(std::ostream&)const;\
   virtual bool scanFrom(std::istream&);\
   CPM_IO_IMPL

#define CPM_IO\
   bool prnOn(std::ostream&)const;\
   bool scanFrom(std::istream&);\
   CPM_IO_IMPL

// sometimes one needs the O-part (= output part) in isolation:

#define CPM_O_IMPL\
   friend std::ostream& operator<<(std::ostream& out, Type const& x)\
      { CpmRoot::IO<Type>().o(x,out); return out;}\
   CPM_MPI_O

#define CPM_O_V\
   virtual bool prnOn(std::ostream&)const;\
   CPM_O_IMPL

#define CPM_O\
   bool prnOn(std::ostream&)const;\
   CPM_O_IMPL

// help to unify defining member functions
//    bool prnOn(ostream& str)const;
//    bool scanFrom(istream& str);
// in classes with many data members that define prnOn and scanFrom.
// Notice that the ; is missing in the definition of cpmp and cpms,
// so that the call of those macros has to provide it.
// These lests these macros behave just as if they were functions.
// E.g.
/*
   bool prnOn(ostream& str)const
   {
      cpmp(x1);
      cpmp(x2);
      cpmp(x3);
      return true;
   }

   bool scanFrom(istream& str)
   {
      cpms(x1);
      cpms(x2);
      cpms(x3);
      return true;
   }
*/
// Defining the nameOf function conveniently, no ';' at the end.
#define CPM_NAM(X) Word nameOf()const{ return Word(#X);}
#define CPM_NAM_V(X) virtual Word nameOf()const{ return Word(#X);}
// Declarations that need a ';' behind. The rule is that any CPM_...
// needs no ; behind it, but cpm... needs it.
#define cpmwt(X) if (!CpmRoot::writeTitle((X),str)) return false
   // wt for 'write title'
#define cpmwat if (!CpmRoot::writeTitle(nameOf().str(),str)) return false
   // wt for 'write automatic title', needs no argument.
#define cpmwbt if (!CpmRoot::writeTitle((nameOf()&" begin").str(),str))\
 return false
   // wbt for 'write begin title'
#define cpmwet if (!CpmRoot::writeTitle((nameOf()&" end").str(),str))\
 return false
   // wet for 'write end title', needs no argument.
#define cpmp(X) if (!CpmRoot::prnT((X),str)) return false
   // p for print
#define cpms(X) if (!CpmRoot::scanT((X),str)) return false
   // s for scan
#define cpmrh(X) rch.read(sec,#X,X)
   // Regularly occurring idiom for reading from a RecordHandler
   // read handler. Stops the program if the quantity cannot be
   // read.
#define cpmrhf(X) rch.read(sec,#X,X,1)
   // Regularly occurring idiom for reading from a RecordHandler
   // read handler, option pedantic = 1, so that only a warning
   // gets issued if the quantity cannot be read.
   // 'f' for 'floppy'
#define cpmr(X) rc.get(sec,#X,X,1)
   // Regularly occurring idiom for reading from a read handler
   // of type Record. Never stops the program if the quantity
   // cannot be read. Issues a warning however.
   // Introduced 2010-08-30.
#define cpmrf(X) rc.get(sec,#X,X,0)
   // Regularly occurring idiom for reading from a read handler
   // of type Record. Never stops the program if the quantity cannot be
   // read. Does not even warn. Introduced 2010-08-30.
#define cpmwh(X) rch.write(sec,#X,X)
   // Regularly occurring idiom for writing to a RecordHandler
   // write handler.
#define cpmrc(X) rc_.get(sec,#X,X,1)
   // regularly occurring idiom for reading from the attribute
   // Recordable::rc_

#define cpmc(X) if (X < obj.X) return 1;\
 if (X > obj.X) return -1
   // c for compare, helps to implement member function com
   // declared in CPM_ORDER (2009-10-02). As in the macros cpmp and
   // cpms the name 'str' for the stream argument was implied, we
   // require here the argument to be named obj'. Works fine and
   // simplifies the implementation of CPM_ORDER considerably.

#define CPM_WORD\
   CpmRoot::Word toWord()const\
   { std::ostringstream ost; prnOn(ost);\
   return CpmRoot::Word(ost.str());}

// order-related stuff now in basicinterfaces

// multiplicative structure with respect to scalars of type ScalarType.
// version where mutating member functions (actually operator *=)  are
// primary
#define CPM_PRODUCT_SCALAR_M\
   Type& operator *=(ScalarType const& s);\
   Type& operator /=(ScalarType const& s)\
   {return operator *=(CpmRoot::invT<ScalarType>(s));}\
   Type operator *(ScalarType const& s)const\
   { Type res=*this; return res*=s;}\
   friend Type operator *(ScalarType const& s, Type const& x)\
   { Type res=x; return res*=s;}\
   Type operator /(ScalarType const& s)const\
   { Type res=*this; return res/=s;}

// version where constant member functions (actually operators *) are
// primary
#define CPM_PRODUCT_SCALAR_C\
   Type operator *(ScalarType const& s)const;\
   Type operator /(ScalarType const& s)const\
   { return *this*CpmRoot::invT<ScalarType>(s);}\
   friend Type operator *(ScalarType const& s, Type const& x)\
    { return x*s;}\
   Type& operator *=(ScalarType const& s){ return *this=*this*s;}\
   Type& operator /=(ScalarType const& s){ return *this=*this/s;}

// linear structure with respect to scalars of type ScalarType.
// version where mutating member functions (actually operator *=, +=) are
// primary
#define CPM_SCALAR_M\
   Type& operator *=(ScalarType const& s);\
   Type& operator +=(ScalarType const& s);\
   Type& operator -=(ScalarType const& s){return operator +=(-s);}\
   Type& operator /=(ScalarType const& s)\
   {return operator *=(CpmRoot::invT<ScalarType>(s));}\
   Type operator *(ScalarType const& s)const\
   { Type res=*this; return res*=s;}\
   friend Type operator *(ScalarType const& s, Type const& x)\
   { Type res=x; return res*=s;}\
   Type operator +(ScalarType const& s)const\
   { Type res=*this; return res+=s;}\
   Type operator -(ScalarType const& s)const\
   { Type res=*this; return res-=s;}\
   Type operator /(ScalarType const& s)const\
   { Type res=*this; return res/=s;}

// version where constant member functions (actually operators *,+) are
// primary
#define CPM_SCALAR_C\
   Type operator *(ScalarType const& s)const;\
   Type operator +(ScalarType const& s)const;\
   Type operator -(ScalarType const& s)const{ return *this+(-s);}\
   Type operator /(ScalarType const& s)const\
   { return *this*CpmRoot::invT<ScalarType>(s);}\
   Type& operator *=(ScalarType const& s){ return *this=*this*s;}\
   Type& operator +=(ScalarType const& s){ return *this=*this+s;}\
   Type& operator -=(ScalarType const& s){ return *this=*this-s;}\
   Type& operator /=(ScalarType const& s){ return *this=*this/s;}

// product, where the mutating member function *= is primary
#define CPM_PRODUCT_M\
   Type& operator *=(Type const& x);\
   Type operator * (Type const& x)const\
   { Type res=*this; return res*=x;}

// product, where the constant member function operator * is primary
#define CPM_PRODUCT_C\
   Type operator * (Type const& x)const;\
   Type& operator *=(Type const& x){ return *this=(*this)*x;}

// inversion
// lean inversion
#define CPM_INVERSION\
   Type operator !(void)const;\
   Type& operator /=(Type const& x){ return operator*=(!x);}\
   Type operator /(Type const& x)const\
   { return *this*!x;}

// more complex version of inversion which also brings the zero into the
// game (which in a group needs not to be a defined concept)
#define CPM_DIVISION\
   Type net(CpmRoot::Z i=0)const;\
   Type inv(void)const;\
   Type operator !(void)const{ return inv();}\
   Type& operator /=(Type const& x){ return operator*=(x.inv());}\
   Type operator /(Type const& x)const\
   { return *this*x.inv();}\
   friend Type net(Type const& x, CpmRoot::Z i=0)\
   { return x.net(i);}\
   friend Type inv(Type const& x){ return x.inv();}

// inversion and netral elements; no multiplication involved.
#define CPM_INV\
   Type net(CpmRoot::Z i=0)const;\
   Type inv(void)const;\
   Type operator !(void)const{ return inv();}\
   friend Type net(Type const& x, CpmRoot::Z i=0)\
   { return x.net(i);}\
   friend Type inv(Type const& x){ return x.inv();}

// minimum infra structure for programming unbiased test for the validity
// of expected mathematical identities

#define CPM_TEST\
   Type ran(CpmRoot::Z j=0)const;\
   Type test(CpmRoot::Z)const;\
   CpmRoot::Z hash(void)const;

#define CPM_TEST_X\
   CPM_TEST\
   CpmRoot::R dis(Type const&)const;\
   CpmRoot::R abs(void)const;\
   CpmRoot::R absSqr(void)const;\
   CpmRoot::R abs2(void)const{ return absSqr();}

// CPM_TEST_ALL, comprises all infra-structure functions
// which are defined in cpmnumbers.h for the bb-types.
#define CPM_TEST_ALL\
   CPM_TEST_X\
   CPM_INV\
   Type con()const;

// CPM_TEST_MOD is a modification of CPM_TEST which turned out
// to be needed in a proper implementation of class Q where it would
// be unnatural to have abs R-valued instead of Q-valued. Also
// implementing abs in terms of absSqure would not be suitable in
// this case.
#define CPM_TEST_MOD\
   CPM_TEST\
   CpmRoot::R dis(Type const&)const;\
   Type abs(void)const;\
   Type absSqr(void)const;

// descriptors
#define CPM_DESCRIPTORS\
   virtual Word nameOf(void) const;\
   virtual Word toWord(void)const;

// arithmetic
#define CPM_DIFFERENCE\
   Type neg(void)const;\
   Type& operator -=(Type const& x){ return operator+=(x.neg());}\
   Type operator -(void)const{ return neg();}\
   Type operator - (Type const& x)const\
   { return *this+x.neg();}

#define CPM_SUM_PLAIN\
   Type& operator +=(Type const& x);\
   Type operator + (Type const& x)const;

#define CPM_DIFFERENCE_PLAIN\
   Type& operator -=(Type const& x);\
   Type operator -(void)const;\
   Type operator - (Type const& x)const;

// sum, where the mutating member function += is primary
#define CPM_SUM_M\
   Type& operator +=(Type const& x);\
   Type operator + (Type const& x)const\
   { Type res=*this; return res+=x;}

// sum, where the constant member function + is primary
#define CPM_SUM_C\
   Type operator + (Type const& x2)const;\
   Type& operator +=(Type const& x){ return *this=(*this)+x;}

// sum, where the constant member function + is primary
#define CPM_DIFF_C\
   Type operator - (Type const& x2)const;\
   Type& operator -=(Type const& x){ return *this=(*this)-x;}

// complex conjugation
#define CPM_CONJUGATION\
   Type con(void)const;\
   friend Type con(Type const& x){ return x.con();}\
   Type operator~(void)const{ return con();}

// lean complex conjugation without friend function
// (friend functions are better avoided since they cause
// problems with todays MS compilers when templates are
// heavily involved)
#define CPM_CONJ\
   Type con(void)const;\
   Type operator~(void)const{ return con();}

// Scalar Product and real valued dependents

#define CPM_DOT_PRODUCT_LEAN\
   ScalarType operator |(Type const&)const;\
   CpmRoot::R absSqr()const\
   { return CpmRoot::absT<ScalarType>(*this|*this);}\
   CpmRoot::R abs()const\
   { return cpmsqrt(absSqr());}

#define CPM_DOT_PRODUCT\
   CPM_DOT_PRODUCT_LEAN\
   CpmRoot::R dis(Type const& x)const\
   {\
      CpmRoot::R a=abs();\
      CpmRoot::R b=x.abs();\
      CpmRoot::R d=(*this - x).abs();\
      return CpmRoot::disDefFun(a,b,d);\
   }

#define CPM_NORMALIZE \
CpmRoot::R nor_()\
{\
   CpmRoot::R nOrig=abs();\
   if (nOrig>0){\
      CpmRoot::R nInv=R(1.)/nOrig;\
      operator*=(ScalarType(nInv));\
   }\
   return nOrig;\
}

// linear structure with respect to scalars of type ScalarType.
// Inversion in ScalarType is not considered; has to be added in a class
// where this is needed (e.g in C)

#define CPM_LINEAR\
   CpmRoot::Z dim()const;\
   Type& operator +=(Type const& a);\
   Type& operator -=(Type const& a);\
   Type& operator *=(ScalarType const& r);\
   Type& neg(void);\
   friend CpmRoot::Z dim(Type const& x){ return x.dim();}\
   friend Type operator +(Type const& x1, Type const& x2)\
      { Type res=x1; return res+=x2;}\
   friend Type operator -(Type const& x1, Type const& x2)\
      { Type res=x1; return res-=x2;}\
   friend Type operator -(Type const& x)\
      { Type res=x; return res.neg();}\
   friend Type operator *(Type const& x, ScalarType const& s)\
      { Type res=x; return res*=s;}\
   friend Type operator *(ScalarType const& s, Type const& x)\
      { Type res=x; return res*=s;}

// a lean and collection of arithmetic operations for linear spaces with
// scalar product. This is now the basic interface of
// VVa<>, VVVa<>, and VVVVa<>.
#define CPM_LIN\
   CPM_SUM_C\
   CPM_DIFF_C\
   Type operator-(void)const;\
   Type operator*(ScalarType const& r)const;\
   Type& operator *=(ScalarType const& r)\
      { return *this=*this*r;}\
   CPM_DOT_PRODUCT\
   CPM_CONJ
#endif
