//? cpmmpi.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_MPI_H_
#define CPM_MPI_H_

/*
   Description: Functions to send and receive objects of C++
   type string via MPI functions. The length of the strings
   is handled automatically. This header file will be included in
   cpminterfaces.h and, therefore, should not include files like
   cpmsystem.h, cpmtypes.h, cpmword.h
*/

#include <cpmdefinitions.h>
   // only to make known CPM_USE_MPI
   // no additional C+- stuff needed here

#include <string>

#if defined(CPM_USE_MPI)
   #include <mpi.h>
      // MPIPro
      // including within the namespace caused trouble!
#endif

namespace CpmMPI{

using std::string;

extern int mpiverbose;
   
#if defined(CPM_USE_MPI)

////////////////////////////// class Com /////////////////////////////////

class Com{ 
// class version of MPI's communicator concept for parallel computing 
   // for which rank varies from 1 to size in order to cooperate with
   // CpmArrays. Note that, due to their template character, the
   // main communication functions can be used to send arround objects
   // of any type which implements the basic C+- marshalling scheme
   // provided by macro CPM_IO.

protected:
   MPI_Comm mc;
   int size;
   int rank; // starts with 1
public:
   static int tagFromToStatic(int f, int t,int size_)
   // returned simply f*t till 2004-07-07. The footnote to p. 254 of GLS
   // (Gropp,Lusk,Skjellum: using MPI) indicates that using the same tag
   // several times in succession might be not so desastrous as one could
   // expect it to be. Nevertheless, one should ensure that the function
   // is injective (appart from tagFromToStatic(f, t, size_)==
   // tagFromToStatic(t, f, size_) )
   {
      int i,j;
      if (f<=t){ i=f-1; j=t;}
      else{ i=t-1; j=f;} 
      return size_*i+j;
   }
   int tagFromTo(int f, int t)const{ return tagFromToStatic(f, t, size);}
   Com(MPI_Comm comm=MPI_COMM_WORLD);
   int getSize()const{ return size;}
      // number of processes between which one can communicate via 
      // sendStr and recStr
   int getRank()const{ return rank;}
   bool sendStr(const string& s, int dest, int tag)const;
      // sending C++-strings without having to input their length
      // return value true means success.
      // I use string as a universal data type singe the C++ class
      // stringstream offers efficient read and write.
      // 1<=dest<=size
   bool bcastStr(string& s)const;
      // broadcasts a string from the process with rank 1 to
      // all other processes
   bool recStr(string& s, int source, int tag)const;
      // receiving strings
      // 1<=source<=size
   int cyc(int i)const;
      // the return value k is equal i modulo size
      // also 1<=k<=size. Useful tool although one could argue that this
      // function belongs to the topic of 'communicator topologies'
      // and not to the communicator itself

// communication patterns:

// The aim of the following functions is to ensure a pairing
// of send and receive activities that avoids deadlock. Although
// it is not difficult to gain an intuitive insight saying that
// the combinations as implemented in these functions are safe against
// deadlock, I have no formal proof for this so far.
// Supporting points are
// 1. empirical: no exception found so far
// 2. philosophical: no simpler expression conceivable except
//    of obviously wrong ones.

// template member functions are very helpful here.
   template <class T, class V>
      // T has to define send() and rec(), V has to have
      // the basic properties of V<T> (see cpmv.h)
   void sendAndRec(const T& t, V& v)const
/*
   Typical application in a SPMD (single program, multiple data )
   situation: (using C+- classes)
   Com com;
   Z size=com.getSize();
   T t(...); // the contribution of com.getRank()
   V<T> v(size); // container for the contributions of others
   com.sendAndRec(t,v); // communication
   R sum=0; // starting to use the communicated stuff
   for (Z i=1;i<=size;i++) sum+=g(v[i]); // R g(const T&)
       // A cummulative effect of all processes is thus available for
       // each process.
*/
   {
      for (int i=1;i<=size;i++){
         if (i==rank){ // OK for size==1, rank==1
            v[i]=t;
         }
         else if (rank<i){
            t.send(i,tagFromTo(rank,i));
            v[i].rec(i,tagFromTo(i,rank));
         }
         else{
            v[i].rec(i,tagFromTo(i,rank));
            t.send(i,tagFromTo(rank,i));
         }
      }
   }

   template <class V>
      // V has to have the basic properties of V<T>, where
      // T defines send() and rec().
   void exchange(V const& s, V& v)const
/*
   Typical application in a SPMD (single program, multiple data )
   situation: (using C+- classes)
   Com com;
   int size=com.getSize();
   int rank=com.getRank();
// the contribution of com.getRank()
   V<T> s(size);
   for (int i=1;i<=size;i++){
      for (i==rank) continue;
         // there is nothing to exchange from rank to rank
      s[i]=....; // this is what rank has to say to i
   }
   V<T> v(size); // container for the messages from others
   com.exchange(s,v); // communication
   R sum=0; // starting to use the communicated stuff
   for (int i=1;i<=size;i++) sum+=g(v[i]); // R g(const T&)
       // Process rank has available the messages from its
       // partners all at once
*/
   {
      for (int i=1;i<=size;i++){
         if (i==rank){
            v[i]=s[i]; // was continue till 2003-06-10
         }
         else if (rank<i){
            s[i].send(i,tagFromTo(rank,i));
            v[i].rec(i,tagFromTo(i,rank));
         }
         else{
            v[i].rec(i,tagFromTo(i,rank));
            s[i].send(i,tagFromTo(rank,i));
         }
      }
   }

//   export // not working with MS cl
   template <class T>
      // see sendAndRec() for the assumed properties of T
   void bcast(T& t)const
      // 'broadcast'
      // Workhorse method for distributing data from the master to
      // the public. See CpmRootX::RecordHandler::ini_ for the typical
      // usage. 
      // Call to bcast has to be preceeded by initialization of t in rank
      // 1. In tis respect rank 1 plays a special role; the call to
      // bcast shoult not be spoilt by conditions on rank!
      // After call for each rank t is assured to have the value which
      // rank 1 had provided. Execution of this function involves each
      // rank with at least one send or receive activity. Since rank 1
      // sends to all processes, and send is not considered done before
      // data are received, rank 1 can only continue after all receives
      // are done in all other processes. Any other process also can
      // only continue after it received. Who received first may hurry
      // away not waiting till others are also ready with receiving. So
      // this is not a strict synchronization mechanism.
      // However, each process that hurries away must have completed its
      // receive and this can only be achieved if all messages sent
      // earlier have been received to. So the function should force
      // completion of all communication tasks that  were pending at the
      // moment of calling bcast. Similar considerations apply to the
      // other collective communication functions.
   
      // A definition of this template providing diagnostics would be
      // desirable. In order to have cpmmessage available one could in
      // principle include  cpmsystems.h into the present file and
      // implement messages much like those of sendStr and recStr. In our
      // special situation this is not working since cpmmpi.h has to be
      // included in cpminterfaces.h where cpmsystem.h should not yet be
      // known. So one had to move the implementation to cpmmpi.cpp which,
      // however, asks for the additional qualification 'export'
      // (see Vandevoorde, Josuttis: C++ Templates, p. 68) which MS cl
      // does not understand.
   {
      if (size==1) return;
      if (rank==1){
         for (int i=2; i<=size;i++) t.send(i,tagFromTo(1,i));
      }
      else  t.rec(1,tagFromTo(1,rank));
   }

   template <class T, class V>
   void gather(const T& t, V& v)const
   // The out-commented messages made me aware of the actual timing of
   // processes in my emulating MPICH runs: Writing of moviefiles for some
   // processes may be in progress long before function step() is done for
   // all tasks. 
   {
      if (size==1){ v[1]=t; return;}
      if (rank!=1){
         t.send(1,tagFromTo(rank,1));
      }
      else{
         v[1]=t;
         for (int i=2;i<=size;i++) v[i].rec(i,tagFromTo(i,1));
      }
   }   
   
   template <class T, class V>
      // see sendAndRec() for the assumed properties of T and V
   void gatherXL(const T& t, V& v, int n)const
      // clear from code
   {
      if (size==1){ v[1]=t; return;}
      if (n<2) {
         gather(t,v);
         return;
      }
      if (rank!=1){
         t.sendXL(1,tagFromTo(rank,1),n);
      }
      else{
         v[1]=t;
         for (int i=2;i<=size;i++) v[i].rec(i,tagFromTo(i,1));
      }
   }
 
   template <class T, class V>
      // see sendAndRec() for the assumed properties of T and V
   void scatter(T& t, const V& v)const
      // clear from code, not yet used or tested
   {
      if (rank==1){ // directly OK for size==1
         t=v[1];
         for (int i=2;i<=size;i++) v[i].send(i,tagFromTo(1,i));
      }
      else{
         t.rec(1,tagFromTo(1,rank));
      }
   }
   
   template <class T, class V>
   void scatterXL(T& t, const V& v, int n)const
      // clear from code, not yet used or tested
   {
      if (rank==1){
         t=v[1];
         for (int i=2;i<=size;i++) v[i].sendXL(i,tagFromTo(1,i),n);
      }
      else{
         t.rec(1,tagFromTo(1,rank));
      }
   }
};

#else // not defined(CPM_USE_MPI)

class Com{ //trivial implementation of Com

public:
   static int tagFromToStatic(int f, int t,int size_)
   {f;t;size_;return 0;}
   static int tagFromTo(int f, int t){ f;t;return 0;}
   Com(){}
   int getSize()const{ return 1;}
   int getRank()const{ return 1;}
   bool sendStr(const string& s, int dest, int tag)const
   {s;dest;tag; return true;}
   bool bcastStr(string& s)const{s; return true;}
   bool recStr(string& s, int source, int tag)const
      {s; source; tag; return true;}
   int cyc(int i)const{ i;return 1;}

   template <class T, class V>
      void sendAndRec(const T& t, V& v)const{t;v;}

   template <class V>
      void exchange(const V& s, V& v)const
      { v[1]=s[1];}
  
   template <class T>
      void bcast(T& t)const{t;}

   template <class T, class V>
      void gather(const T& t, V& v)const{t;v;}
         
   template <class T, class V>
      void gatherXL(const T& t, V& v, int n)const{t;v;n;}

   template <class T, class V>
      void scatter(T& t, const V& v)const{t;v;}
         
   template <class T, class V>
      void scatterXL(T& t, const V& v, int n)const{t;v;n;}
};

#endif // #if defined(CPM_USE_MPI)

extern CpmMPI::Com Cpm_com;
void initialize(int* a, char*** b);
void finalize();

} // namespace

#endif // #if defined(CPM_MPI_H)
