//? cpms.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_S_H_
#define CPM_S_H_
/*

   Description: Defines a template class S<T> of sets
       the elements of which are of type T.
      Intented as a basis for a mathematics style definition of
           associative lists.

*/

#include <cpmvo.h>
#include <cpmtypes.h>

////////////////////////// class S<> ///////////////////////////////////

namespace CpmArrays{

   using CpmFunctions::F;
   using CpmRoot::Word;

template <class T>
class S{ // sets of homotypic elements
   // Represents sets, all the elements of which are of type T.
   // It is assumed that T<T and T>T is defined.

protected:

   void create(Z n)
   // tool for constructors
   { n_=n; Vo<T> xx(n_); Vo<Z> vv(n_); x_=xx; v_=vv;}

   void create(const S& s)
   // tool for constructors
   { n_=s.n_; x_=s.x_; v_=s.v_;}

explicit   S(Z n){ create(n);}
      // Not yet a set since elements not initialized
      // for internal use only !!!!

   void setValidity(Z b);
      // sets a constant value b into the v-list
   void setValidity(T const&  t, Z b);
      // sets the validity of the element which equals t to b

// data all well typed so no copy constructor and assignment operator
// will be defined correctly by the compiler

   Z n_;
      // number of elements (i.e. cardinality)

   Vo<T> x_;   // ordered list of elements

   Vo<Z> v_;
      // auxiliary list indicating validity (valid:1 unvalid: 0 )

   typedef S<T> Type;

public:

   CPM_IO_V
      // CPM_ORDER comes later in this class declaration/definition

// constructors

   S(void){ create(0);}
      // constructor for the void set
   S(Word const& w, T const& t){w; create(1);x_[1]=t;v_[1]=1;}
      // constructor for the set {t}
      // e.g. S<Z> a("",5); seems to be acceptable
      // note that explicit S(T const& t){create(1);x_[1]=t;v[1]=1;}
      // would be ambiguous for T==Z

   S(Type const& s){ create(s);}
      // copy constructor

   void merge(V<Type> const& setList);
      // changes *this into the union of *this and the sets in the list

   explicit S(V<T> const& vec):n_(vec.dim()),x_(vec),v_(n_,1)
      { normalize();}
      // constructor from an array of T-values (V<T> sufficient, not
      // needing Vo<T>; T, however, has to support ordering (which is
      // clear if S<T> is under consideration). Clearly, multiple
      // appearing of components in vec does not prevent their
      // single appearence in the result (which, otherwise, would not be
      // a set)

    S(V<T> const& vec, V<bool> const& vecValid);
      // constructor from an array of T-values, together with a list of
      // validity values. vec[i] is considered valid iff
      // 1<=i<=vec.dim() && 1<=i<=vecValid.dim() && vecValid[i]==true
      // The valid values are put into a set.
      // Clearly, muliply appearing components will appear only
      // once in the result (which, otherwise, would not be a set)

// cardinality access

   Z car()const{ return n_;}
      //: cardinality
   Z dim()const{ return n_;} // for uniformity with other array types
      //: dimension
   Z size()const{ return n_;} // for uniformity with STL code
      //: size

   Z b()const{ return 1;}
      //. begin , as in V<>

   Z e()const{ return n_;}
      //. end , as in V<>

// set formation operations
   bool addAct(T const& t);
      // changes *this into the union of (*this) and {t}
      // and returns true if t was not already there
      // so that it had to be added 'actually'

   void add(T const& t){ addAct(t);}
      // changes *this into the union of (*this) and {t}
      // return value of addAct not used

   void add_(T const& t){ Type s(t); add(s);}
      // experimental

   void add(Type const& s);
      // changes *this into the union of (*this) and s

   void insert(T const&  t){ add(t);}
      // for uniformity with STL code

   void push_back(T const&  t){ add(t);}
      // for uniformity with STL code where the container is vector<>

   bool remove(T const&  t);
      // *this will be changed into *this\{t}
      // If this reduces the cardinality of *this, the return
      // value is 'true' and 'false' else

   Type operator |(Type const&  s)const;
      // forming the union of *this and s (corresponds to the
      // 'or' operation for indicator functions)

   Type operator &(Type const&  s)const;
      // forming the section of *this and s (corresponds to the
      // 'and' operation for indicator functions)

   Type operator -(Type const&  s)const;
      // forming *this\s i.e. the set of all those elements of
      // *this which don't belong to s (relative complement)

   Type operator ^(Type const&  s)const{ return (*this-s)|(s-*this);}
      // forming the symmetric difference of s1 and s2
      // (corresponds to 'exclusive or' for indicator functions)

   T const&  operator[](Z i)const{ return x_[i];}
      // Means of iterating over all elements of the set
      // The specific indexing is gouverned by the < order.
      // If i is 'out of range' i.e. not 1<=i<=card(), this is
      // considered an error which is handled by Vo<T> (finally
      // by V<T>).

   T& ref(Z i){ return x_[i];}
      // dangerous operation since changing x[i] may invalidate the
      // internal representation of *this. Needed for implementing
      // M<X,Y>.
      // Till 02-1-7 I had instead:  T operator[](Z i){ return x[i];}
      // This prevented Y& M<X,Y>::operator[](const X&) from working
      // properly. See also function normalize().

   T const&  fir()const;
      //: first
      // first element of *this according to the ordering on
      // which the construction of Type is based.
      // causes error if *this is void

   T const&  last()const;
      // last element of *this according to the ordering on
      // which the construction of Type is based.
      // causes error if *this is void

   void normalize(void);
      // shortens a list representation with multiple appearence of
      // a component such that all components are mutually different.
      // It also eliminates from the x-list all those elements x[i] for
      // which v[i]==0.
      // It thus achieves the status which for a well formed
      // list representation of a set is mandatory.
      // Not needed to be called normally (although this would not
      // cause problems). However if function ref() was called and
      // used to change elements it might be necessary to call
      // normalize to get an benign set again.

   T operator()(Z i)const{ return x_(i);}
      // Means of iterating over all elements of the set
      // The specific indexing is gouverned by the < order.
      // For i's 'out of range', T() is returned. This is not
      // considered an error.

// boolean operations

   bool isVoid(void)const{ return n_==0;}

   bool hasElm(T const& t)const;
      //: has element
      // returns true iff t is an element of *this

   bool ni(T const&  t)const{ return hasElm(t);}
      // 'in' reversed, from LATEX symbol \ni for the mirror image of
      // the epsilon-like 'is element'-symbol
      // Notice that the 'in' function would have as first argument a T
      // and as second argument a Type. This cannot be made a member
      // function. Introducing a non-member friend function would offend
      // the C+- coding style.

   Z locate(T const&  t)const;
      // returns 0 if t is not an element of *this
      // and its index if it is an element of *this

   bool isSubSetOf(Type const&  s)const;
      // returns true iff *this is a subset of s

   bool isSuperSetOf(Type const&  s)const;
      // returns true iff s is a subset of *this

// selection by functions

   void select(Z (*f)(T const& ));
      // turns *this into the subset consisting of those elements
      // x[i] of *this for which f(x[i])!=0

   void select(F<T,bool> const& f);
      // as previous function, only different type of the selection
      // function

   void select(V<bool> const& sel);
      // eliminates from *this all components x[i] for which sel[i]
      // is defined and statisfies sel[i]==false.
      // (see corresonding (non-mutating) function of V<T>)

   void eliminate(Z i);
      // eliminates from *this just component x[i] for 1<=i<=car()
      // Else, nothing is done. Does the same as remove((*this)[i])
      // but much faster , due to known position of the element to
      // remove.

   Type subSet(F<T,bool> const& f)const;
      // turns *this into the subset consisting of those elements
      // x[i] of *this for which f(x[i])==true

// relations to other classes:

   Vo<T> toV(void)const{ return x_;}
      // getting the representing vector as a new entity. It may be
      // useful to manipulate this vector by, for instance,
      // the V<T>::condense() function and then form a new set via the
      // constructor from vectors

// Infrastructure functions as declaration macros  known from including
// <cpmvo.h>
   CPM_ORDER
      // since T has to allow ordering, we implement the order
      // operations. This allows to form S< Type >.

   virtual S<T>* clone()const{ return new S(*this);}

// static bool effectivelyEqual(T const& t1, const T& t2){ return t1==t2;}
      // If no additional infrastructure in T is available, we can't do
      // better. Could be made non-static and virtual or template argument
      // to allow equivalence classes to be formed during set formation;
      // not persued here.

   Word toWord()const;

   virtual Word nameOf()const{
      Word wi="S< ";
      Word wt=CpmRoot::Name<T>()(T());
      return wi&wt&" >";
   }
};

// Implementation of S<T>

#define CPM_FOR_ALL for (i=1; i<=n_; i++)

template <class T>
void S<T>::setValidity(Z b)
{
   Z i;
   CPM_FOR_ALL v_[i]=b;
}

template <class T>
bool S<T>::addAct(T const&  t)
{
   if (n_==0){
      x_=Vo<T>(1,t);
      v_=Vo<Z>(1,1);
      n_=1;
      return true;
   }
   else if (n_==1){
      if (t==x_[1]){
         return false;
      }
      else{
         x_=x_.enqueue(t);
            // no call to locate() in this case
         v_=v_&true;
         n_++;
         return true;
      }
   }
   else{
      X2<Z,bool> ib=x_.findAsc(t);
      if (ib.c2()) return false; // t found in x_
      Z i=ib.c1()+1;
      x_=x_.ins(i,t);
      v_=v_&true;
      n_++;
      return true;
   }
}

template <class T>
void S<T>::add(const S<T>& s)
{
   Z m=s.car();
   if (m==0) return; // nothing to do
   if (n_==0){
      *this=s;
      return;
   }
   n_+=m;
   x_=x_&s.x_;
   v_=v_&s.v_;
   normalize(); // gives n the right value
}

template <class T>
void S<T>::normalize(void)
{
   Z mL=10;
   cpmmessage(mL,"S<T>::normalize(void) started");
   if (n_==0){
      cpmmessage(mL,"S<T>::normalize(void) done, first n_==0 exit");
      return; // nothing to do
   }

   Z i,j,nOld=n_;

// Eliminating the invalid entries

   n_=0;
   for (i=1;i<=nOld;i++){
      if (v_[i]==1) n_++;
   }
      // now n_ is the number of valid terms

   if (n_==0){
      Vo<T> xx;
      Vo<Z> vv;
      x_=xx;
      v_=vv;
      if (cpmverbose>=4)
         cpmmessage("S<T>::normalize(void) done, second n_==0 exit");
      return;
   }

   Vo<T> arr(n_);
   i=1;j=1;
   while(i<=n_ && j<=nOld ){
      while(j<=nOld && v_[j]==0){ j++; }
      if (j>nOld) break;
      arr[i++]=x_[j++];
   }

   Vo<Z> indx=arr.permutationForIncreasingOrder();

 // Eliminating  repetitions of terms

   Vo<Z> ind(n_);
   T t;
   i=0;j=0;
   while (i<n_ && j<n_){
      i++; j++;
      t=arr[indx[j]];
      // now we have to look whether j has to be increased.
      // while (j<n && effectivelyEqual(t,arr[indx[j+1]]) ){j++;}
      while (j<n_ && t==arr[indx[j+1]] ){j++;}
         // now j is the largest value for which arr[indx[j]]==t
         // since arr[indx[j+1]]!=t by the while condition
      ind[i]=j;
   }

   n_=i;
 //  cpmcerr<<"now n_ equals "<<n_;
   Vo<T> xx(n_);
   Vo<Z> vv(n_);

   x_=xx;
   v_=vv;

   CPM_FOR_ALL{
      x_[i]=arr[indx[ind[i]]];
      v_[i]=true;
   }
   n_=x_.dim();
   cpmmessage(mL,"S<T>::normalize(void) done");
}

template <class T>
S<T> S<T>::operator |(const S<T>& s)const
{
   Z nNew=n_+s.n_, i;
   S res(nNew);
   for (i=1;i<=n_;i++){
      (res.x_)[i]=x_[i];
      (res.v_)[i]=v_[i];
   }
   for (i=1;i<=s.n_;i++){
      (res.x_)[n_+i]=(s.x_)[i];
      (res.v_)[n_+i]=(s.v_)[i];
   }
   res.normalize();
   return res;
}

template <class T>
S<T> S<T>::operator &(const S<T>& s)const
{
   Z i;
   S<T> res;
   if (n_>s.n_){
      res=*this;
      res.setValidity(false);
      for (i=1;i<=s.n_;i++){
         res.setValidity((s.x_)[i],true);
      }
   }
   else{
      res=s;
      res.setValidity(false);
      for (i=1;i<=n_;i++){
         res.setValidity(x_[i],true);
      }
   }
   res.normalize();
   return res;
}

template <class T>
S<T> S<T>::operator -(const S<T>& s)const
{
   Z i;
   S<T> res=*this;
   for (i=1;i<=s.n_;i++){
      res.setValidity((s.x_)[i],false);
   }
   res.normalize();
   return res;
}

template <class T>
bool S<T>::hasElm(T const&  t)const
{
   Z r=x_.find(t);
   return (r!=0);
}

template <class T>
Z S<T>::locate(T const&  t)const
{
   return x_.find(t);
}

template <class T>
bool S<T>::remove(T const&  t)
{
   Z jf=x_.find(t);
   if (jf==0){
      return false;
   }
   else{
      v_[jf]=0;
      normalize();
      return true;
   }
}

template <class T>
void S<T>::setValidity(T const&  t, Z b)
{
   Z r=x_.find(t);
   if (r!=0) v_[r]=b;
}

template <class T>
bool S<T>::isSuperSetOf(const S<T>& s)const
{
   if (n_<s.n_) return false;
   S diff=s-(*this);
   return diff.isVoid();
}

template <class T>
bool S<T>::isSubSetOf(const S<T>& s)const
{
   if (n_>s.n_) return false;
   bool b=s.isSuperSetOf(*this);
   return !b;
}

template <class T>
void S<T>::select(Z (*f)(T const&))
{
   Z i;
   CPM_FOR_ALL v_[i]=f(x_[i]);
   normalize();
}

template <class T>
void S<T>::select(F<T,bool> const& f)
{
   Z i;
   CPM_FOR_ALL v_[i]=f(x_[i]);
   normalize();
}

template <class T >
void S<T>::select(const V<bool>& sel)
{
   Z i;
   Z ns=sel.dim();
   CPM_FOR_ALL v_[i]=(i<=ns ? sel[i] : true);
   normalize();
}

template <class T >
void S<T>::eliminate(Z i)
{
   if (i>1 && i<=n_){
      v_[i]=0;
      normalize();
   }
}

template <class T>
S<T> S<T>::subSet(F<T,bool> const& f)const
{
   S<T> res=*this;
   res.select(f);
   return res;
}

template <class T>
Z S<T>::com(S<T> const& s)const
{
   if (n_<s.n_){
      return 1;
   }
   else if (n_>s.n_){
      return -1;
   }
   else{ // now n==s.n
      for (Z i=1;i<=n_;i++){
         if ( x_[i]<(s.x_)[i]) return 1;
         if (x_[i]>(s.x_)[i] ) return -1;
      }
      return 0; // if all are equal
   }
}

template <class T>
S<T>::S(const V<T>& vec, const V<bool>& vecValid)
{
   n_=vec.dim();
   x_=vec;
   v_=Vo<Z>(n_);
   Z n2=CpmRootX::inf<Z>(n_,vecValid.dim());
   if (n2==0)
      cpmerror("S<T>::S(const V<T>&, const V<bool>&): dimension\
 mismatch");
   for (Z i=1;i<=n2;i++) v_[i]=(vecValid[i] ? 1 : 0);
   normalize();
}

template <class T>
Word S<T>::toWord()const
{
   Z nd=dim();
   Word res="{ ";
   for (Z i=1;i<nd;i++){
      Word wi=CpmRoot::toWord<T>((*this)[i]);
      wi&=", ";
      res&=wi;
   }
   res&=CpmRoot::toWord<T>((*this)[nd]);
   res&=" }";
   return res;
}

template <class T>
void S<T>::merge(const V< Type >& setList)
{
   Z d=setList.dim();
   for (Z i=1;i<=d;i++) add(setList[i]);
}

template <class T>
T const&  S<T>::fir()const
{
   if (n_==0) cpmerror("S<T>::fir(): set is void");
   return x_[1];
}

template <class T>
T const&  S<T>::last()const
{
   if (n_==0) cpmerror("S<T>::last(): set is void");
   return x_[n_];
}

template <class T>
bool S<T>::prnOn(ostream& str)const
{
   Z mL=1;
   static Word loc("S<>::prnOn(ostream&)");
   CPM_MA
   cpmwbt;
   bool bi;
   bi=CpmRoot::IO<Z>().o(n_,str);
   cpmassert(bi==true,loc);
   for (Z i=1;i<=n_;++i){
      if (CpmRoot::wrtTit){
         Word wi("// j=");
         wi&=cpm(i);
         bi=wi.prnOn(str);
            // the method here is to combine Words as Words and
            // write the result to stream (once!)
         cpmassert(bi==true,loc);
      }
      bi=CpmRoot::IO<T>().o(x_[i],str);
      cpmassert(bi==true,loc);
      if (bi==false){
         cpmdebug(x_[i]);
         CPM_MZ
         return false;
      }
   }
   cpmwet;
   CPM_MZ
   return true;
}

template <class T>
bool S<T>::scanFrom(istream& str)
{
   Z mL=2;
   static Word loc("S<>::scanFrom(istream&)");
   CPM_MA
   if (!CpmRoot::IO<Z>().i(n_,str)){
      cpmwarning(loc&": invalid stream after reading cardinality");
      CPM_MZ
      return false;
   }
   else{
      if(n_>dimMax || n_<=0)
         cpmwarning(loc&" n_= "&cpm(n_));
      S<T> res;
      T t;
      Z i;
      CPM_FOR_ALL{
         if (!CpmRoot::IO<T>().i(t,str)){
            cpmwarning(loc&": reading failed for i="&cpm(i)&
               " of "&cpm(n_));
            CPM_MZ
            return false;
         }
         res.add(t);
      }
      *this=res;
      CPM_MZ
      return true;
   }
}

#undef CPM_FOR_ALL

} // CpmArrays

#endif
