//? cpmsr.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_SR_H_
#define CPM_SR_H_
/*
   Description: Defines a class template Sr<T> of sets
       the elements of which are of type T.

   History: Started 95-3-13
      95-3-15 transformed in a class derived from Obj
      97-9-5 transformed as to fit the CTL framework
      99-7-27 transformed as to fit the CPM framework
      02-5-24 implementation changed to STL, Test_set carried out
      2005-09-10 implementation purged from io_clone and r_clone
*/

#include <cpms.h>

namespace CpmArrays{

   using CpmRoot::Word;

template <class T>
class Sr: public S<T>{  // S with a rich interface
   // represents sets, all the elements of which are of type T.
   typedef Sr<T> Type;
   typedef S<T> Base;
public:
// constructors.

   Sr(void):S<T>(){}
         // constructor for the void set (n=0 added 95-11-24)

   Sr(const T& t):S<T>(t){}
         // constructor for the set {t}

   Sr(const S<T>& s):S<T>(s){}
      // down-cast constructor

   Sr<T> condense(void)const;
      // forms a new set obtained by shrinking to a single representative
      // element every equivalence class with respect to function equiv,
      // which itself is controlled by the static element acc

// Infrastructure functions

   virtual S<T>* clone()const{ return new Sr(*this);}

   CPM_IO
      // prnOn, scanFrom
   CPM_TEST_X
      // ran, test, hash, dis, abs, absSqr
   CPM_DESCRIPTORS
      // toWord, nameOf
      // The missing 3 from Root are: net, com, con

   static R acc;

protected:

   static bool equiv(const T& t1, const T& t2)
   { return CpmRoot::disT<T>(t1,t2)<acc;}

};

// test functions based on set theoretical identities

 template <class T>
   void idem1(Sr<T>& lhs, Sr<T>& rhs, const Sr<T>&s)
      { lhs=(s&s); rhs=s;}

   template <class T>
   void idem2(Sr<T>& lhs, Sr<T>& rhs, const Sr<T>& s)
      { lhs=(s|s); rhs=s;}

   template <class T>
   void comm1(Sr<T>& lhs, Sr<T>& rhs, const Sr<T>& s1, const Sr<T>& s2)
      { lhs= (s1&s2); rhs=(s2&s1);}

   template <class T>
   void comm2(Sr<T>& lhs, Sr<T>& rhs, const Sr<T>& s1, const Sr<T>& s2)
      {   lhs= (s1|s2); rhs=(s2|s1);}

   template <class T>
   void assoc1(Sr<T>& lhs, Sr<T>& rhs, const Sr<T>& s1, const Sr<T>& s2,
    const Sr<T>& s3)
      { lhs= ((s1&s2)&s3); rhs=(s1&(s2&s3));}

   template <class T>
   void assoc2(Sr<T>& lhs, Sr<T>& rhs, const Sr<T>& s1, const Sr<T>& s2,
    const Sr<T>& s3)
      { lhs= ((s1|s2)|s3); rhs=(s1|(s2|s3));}

   template <class T>
   void distr1(Sr<T>& lhs, Sr<T>& rhs, const Sr<T>& s1, const Sr<T>& s2,
    const Sr<T>& s3)
      { lhs= (s1|(s2&s3)); rhs=((s1|s2)&(s1|s3)) ;}

   template <class T>
   void distr2(Sr<T>& lhs, Sr<T>& rhs, const Sr<T>& s1, const Sr<T>& s2,
    const Sr<T>& s3)
      { lhs= (s1&(s2|s3)); rhs=((s1&s2)|(s1&s3)) ;}

   template <class T>
   void diffdistr1(Sr<T>& lhs, Sr<T>& rhs, const Sr<T>& s1, const Sr<T>&
    s2, const Sr<T>& s3)
      { lhs= (s1-(s2|s3)); rhs=((s1-s2)&(s1-s3)) ;}

   template <class T>
   void diffdistr2(Sr<T>& lhs, Sr<T>& rhs, const Sr<T>& s1, const Sr<T>&
    s2, const Sr<T>& s3)
      { lhs= (s1-(s2&s3)); rhs=((s1-s2)|(s1-s3)) ;}

// more technical tests

   template <class T>
   int addElements(const Sr<T>& s, const T& t);
      // adding and finding an element

   template <class T>
   R writeRead(const Sr<T>& s);
      // read write test (to and from file)

// Implementation of Sr<T>

#define CPM_FOR_ALL for (i=1;i<=Base::dim();i++)

template <class T>
R Sr<T>::acc=1e-3;

template <class T>
Sr<T> Sr<T>::condense(void)const
{
   Vo<T> xr=Base::toV();
   Vo<T> x1=(static_cast< V<T> >(xr).condense(equiv)).first();
   return S<T>(x1);
}

template <class T>
Word Sr<T>::nameOf(void)const
{
   Word a="Sr<";
   Word b=CpmRoot::Name<T>()(T());
   return a&b&">";
}

template <class T>
R Sr<T>::absSqr(void)const
//
{
   R s=0;
   Z i;
   CPM_FOR_ALL{
      s+=CpmRoot::AbsSqr<T>()((*this)[i]);
   }
   return s;
}

template <class T>
R Sr<T>::abs(void)const
//
{
   return cpmsqrt(absSqr());
}

/*
// looks natural but behaves not well in comparing two sets where the
// second is a noisy copy of the first (e.g. ofter writing to file with
// limited numerical accuracy and reading back. In this case the next
// version is superior. This illustrates the point that the notion of sets
// depends strongly on the notation of equality (and vice versa).

template <class T>
R Sr<T>::dis(const Sr<T>& s)const
{
   Sr<T> sa=(*this).condense();
   Sr<T> sb=s.condense();
   Sr<T> sc=sa^sb;

   R s1=sa.abs();
   R s2=sb.abs();
   R s3=sc.abs();
   return CpmRootX::distFunc(s1,s2,s3);
}
*/

template <class T>
R Sr<T>::dis(const Sr<T>& s)const
{
   if ( Base::n_ != s.car()) return 1;
   Vo<R> d(Base::n_);
   Z i;
   CPM_FOR_ALL d[i]=CpmRoot::Dis<T>()((*this)[i],s[i]);
   d.order();
   return d[Base::n_]; // max of d
}

template <class T>
Sr<T> Sr<T>::ran(Z j)const
{
   Z i;
   Sr<T> res=*this;
   Z j0=1000;
   Z j_=(j!=0 ? CpmRoot::ranT<Z>(j0,j) : 0);
   Z j1=j;
   bool b=true;
   V<T> xr(2*Base::n_);
   V<bool> vr(2*Base::n_);
   CPM_FOR_ALL{
      xr[i]=CpmRoot::ranT<T>(Base::x_[i],j1);
      vr[i]=CpmRoot::ranT<bool>(b,j1);
      j1+=j_;
      // random generator for T and B used
   }
   vr[1]=1; // that not all become false for small n
   return S<T>(xr,vr);
}

template <class T>
Sr<T> Sr<T>::test(Z tsz)const
{
   Z nt,i;
   nt=CpmRoot::testT<Z>(nt,tsz);
   T tTest;
   tTest=CpmRoot::testT<T>(tTest,tsz);
   V<T> xr(nt);
   for (i=1;i<=nt;i++){
      xr[i]=CpmRoot::ranT<T>(tTest,i);
   }
   return S<T>(xr);
}

template <class T>
Z Sr<T>::hash(void)const
{
   Z i,h=0;
   h+=Base::dim();
   CPM_FOR_ALL h+=CpmRoot::hashT<T>((*this)[i]);
   return h;
}

template <class T>
bool Sr<T>::prnOn(ostream& str)const
{
   return Base::prnOn(str);
}

template <class T>
bool Sr<T>::scanFrom(istream& str)
{
   return Base::scanFrom(str);
}

template <class T>
Word Sr<T>::toWord()const
{
   Word res="{ ";
   for (Z i=1;i<Base::n_;i++){
      Word wi=CpmRoot::toWord<T>((*this)[i]);
      wi&=", ";
      res&=wi;
   }
   res&=CpmRoot::toWord<T>((*this)[Base::n_]);
   res&=" }";
   return res;
}

template <class T>
int addElements(const Sr<T>& s, const T& t)
   // Tests whether an element added to a set is retrieved as an element
   {
      Sr<T> s1=s;
      s1.add(t);
      return s1.hasElem(t);
   }

template <class T>
R writeRead(const Sr<T>& s)
   // Tests whether a set can be read again when written to file
   {
      Word filename=s.nameOf()+".dat";
      Word name=s.nameOf();
      ofstream out(-filename);

      if (!out) cpmerror(name,"test: output file cannot be opened");
      out<<s;
      out.close();
      Sr<T> sr;
      R res;
      ifstream in(-filename);
      if (!in) cpmerror(name, "test: input file cannot be opened");
      in>>sr;
      in.close();
      res=s.dis(sr);
      return res;
    }
#undef CPM_FOR_ALL
} // CpmArrays

#endif

