//? cpmsystem.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_SYSTEM_H
#define CPM_SYSTEM_H
/*
   Purpose: Declaring basic facilities for a C+- based program to
   communicate with the user.
*/

#include <cpmword.h>

namespace CpmSystem{ // C+- system basic facilities

   using namespace CpmStd;

 //  typedef double R_; // in CpmSystem we need no arbitrary precision

   using CpmRoot::Word;
   using CpmRoot::Z;
   using CpmRoot::R;

   class Message;
   void updFromIniFile();
   class Error;
   void error(Word const& w);
   void error(std::ostringstream const& ost);
   void error(Word const& w1, Word const& w2);

/////////////////////// function glue //////////////////////////////////

Word glue(Word const& path, Word const& file);
   //: glue
   // Connects a (partial)directory name and a file name
   // to give a full file name. 'glues' path and file together.
   // For extended==false, the result is simply
   // path&Word('/')&file. This is the correct behavior if file is a
   // normal filename which does not contain directory separators.
   // if extended==true we take into account that file may start
   // with ./ or ../. Then these dots will be eliminated.
   // Actually all all dots will be eliminated except the one
   // preceeding the file extension and

/////////////////////////// class Message ////////////////////////////////

class Message{ // provides basic output (unidirectional communication)
      // channels for a program. Two output files (cpmcerr and cpmdata
      // according to their define alias, see comments to data member
      // 'out' for these important entities) and a status bar featuring
      // 4 addressable segments are made available.
      // If we are running a console application, messages also go to the
      // console.
      // There is also limited communication in the other direction:
      // A file cpmconfig.ini, if it can be opened, will be used to
      // modify some static data members of the present class. This
      // excludes the data used to create the files cpmcerr and cpmdata.
      //
      // This class only has static data and methods, thus all instances
      // are identical. The information for initializing the static data
      // comes from the file cpmsystemdependencies.h.
      // A statement
      //    Cpmsystem::updFromIniFile();
      // allows to change most static data by replacing them by values
      // read from the ini-file cpmconfig.ini. Since updFromIniFile is a
      // friend of Message, it can do this replacement of private data of
      // class Message.

   friend void updFromIniFile();
      // this allows to manipulate the private data of Message, by a
      // function which can be implemented only later when class
      // RecordHandler is available. Will be implemented in
      // cpmapplication.cpp. A project which does not call
      // function updFromIniFile() needs not to have cpmapplication.cpp
      // in its file set.

   static bool streamInitialized;

   static Z vpw_, vph_, fsw_, fsh_;
      // viewport width, viewport height,
      // framestore width, framestore height,
   static Word sttmes1;
   static Word sttmes2;
   static Word sttmes3;
   static Word sttmes4;

   static Z sttBarWid1;
      // status bar width 1
      // the status bar has 4 'panes' and the width of each is
      // characterized by a number; sttBarWidi is this number for the
      // i-th pane
   static Z sttBarWid2;
   static Z sttBarWid3;
   static Z sttBarWid4;

   static Word relPos1;
   static Word relPos2;
   static Word relPos3;
   static Word relPos4;
      //: relative position
      // relative directory positions, presently used only to locate
      // basic input file  cpmconfig.ini startng from the location
      // of the executable when starting from the command line) or
      // from the working directory as seen by an IDE one is working
      // with.

   static Word inpDir;
      //: input directory
      // directory root read from cpmconfig.ini that is available for
      // defining interaction of a program with the file system of the
      // machine

   static Word outDir1;
      //: output directory 1
      // directory to which cpmcerr.txt and cpmdata.txt get written.
      // Can't be changed by reading from config.ini.

   static Word outDir2;
      //: output directory 2
      // directory to which output data of a program run will be written
      // can be set to a value different from outDir1 by reading from
      // config.ini

    static Word fontDir;
      //: font directory
      // font location  read from cpmconfig.ini that is available for
      // defining interaction of a program with the file system of the
      // machine. So a C+- user is completely free where to place
      // the C+- font file on his computer.

    static Word runId;
      //: run ID
      // Is created prior to the creation of the log files.

    static bool appRunId;
      //: append run id
      // If this is true, the run ID will be appended to the names of
      // auto-created image files made by class CpmGraphics::Frame
      // and documentation files made by
      // class CpmApplication::IniFileBasedApplication.
      // In the latter case there are provisions in place
      // to prescribe name appendices in ini-files, for instance
/*************************************************************************
         selection
         W runName=090616x
*************************************************************************/
      // This mechanism alone
      // does not make evident what version of the log files describes
      // their generation. So if appRunId is true, one appends runID
      // to this runName so that documentation files and log files which
      // belong togeher always have runID as a common name fragment.
      // For cerrExt==true it is enforced that appRunId==true. Otherwise
      // one may change appRunId from cpmconfig.ini, where there is an
      // entry
/*************************************************************************
         append run id
         B val=false // or true
*************************************************************************/
    // calling void CpmSystem::updFromIniFile() reads cpmconfig.ini and
    // may change appRunId (unless, as stated, cerrExt==true).

   static Z runIdLength;

   static bool initialized(void){ return streamInitialized;}
      //

protected: // for usage by derived class Error

   static Z maxNumMes;
      // number of 'unenforced' messages to be written on out (helps to
      // avoid creating unintentionally huge text files and to slow
      // program run down

   static bool silent;
      // if this is true, the basic function communication runs idle
      // If it is set to true initially no log files will be created
      // and no possibility exists to swich it to false.
      // Thus setting CPM_NOLOG reliably says that there will no log
      // files be created during program run.

   static void communication(Word const&, Word const&, Word const&,
      bool, Z barSeg=1);

public:
// data
   static Z verbose;
      // controls the ammount of messages to be written to file in a way
      // that every message that writes under some value of
      // Message::verbose, will also be written under all higher values.

   static Z debug;
      // controls the behaviour of the cpmassert macro.
      // For debug==1 Message::warning will be executed
      // For debug==2 Message::error instead
      // In all other cases the asserted condition will not be tested

   static Z trigger;
      // Device for debugging, should be used only in code inserted for
      // debugging or temporary analysis. For instance one may
      // add
      //    cpmtrigger=1;
      // just before a dubious function call, and add to the block
      // of this function the statements:
      // if (cpmtrigger>0){
      //    cpmdebug(<argument 1>);
      //    cpmdebug(<argument 2>);
      //       ...
      //    cpmtrigger=0;}
      // This allows to make sure that the function one had in mind was
      // actually called, and it shows the arguments used for the call.
      // Who runs a debugger has probably never to do such tricks.

// functions
   static void ini(void);
      // one has not to call this function, but it can do no harm either,
      // helps experimentation in non-standard situations
   static void fini(void){ out.close(); outData.close();}
      // only after being sure that the communication files have been
      // closed one can copy the for final documentation purposes
   static Z pane1(){ return sttBarWid1;}
      //: pane 1
   static Z pane2(){ return sttBarWid2;}
      //: pane 2
   static Z pane3(){ return sttBarWid3;}
      //: pane 3
   static Z pane4(){ return sttBarWid4;}
      //: pane 4

   static Z w(){ return vpw_;}
      // tendative frame width in pixel
   static Z h(){ return vph_;}
      // tendative frame height in pixel
      // Intended is a call
      // Z nx=Message::w(), ny=Message::h();
      // CpmGraphics::Viewport vp(nx,ny,title);
      // Making a statusbar within this field is the task of Viewport.

   static Z ws(){ return fsw_;}
      // tendative frame store width in pixel
   static Z hs(){ return fsh_;}
      // tendative frame store height in pixel

   static ofstream out;
      // writing via '>>' to Message::out is the primarily provided
      // (unidirectional) communication facility of the system.
      // After program run, one will find the text file cpmcerr.txt in
      // the directory Message::outDir1 that contains all what has been
      // was written to stream Message::out.

      // If CPM_USE_MPI is defined and the number np of processes is > 1
      // the program is aware of its rank and this will be appended
      // to the filenames cpmcerr and cpmdata in a way that
      // alphabetic order is the natural order. E.g. for np=64
      // we have cpmcerr_01.txt,...cpmcerr_09.txt, cpmcerr_10.txt, ...,
      // cpmcerr_64.txt.
      // The next program run will overwrite cpmcerr(_rank).txt. In most
      // cases this is OK in order not to pollute the working directory.
      // Sometimes it is better to retain the cpmcerr, cpmdata files
      // e.g. for comparing them for two related runs. Of course,
      // we may rename the files after creation in an informative way
      // to enable this. There is, however, the function
      // CpmSystem::docRun(bool) which does this using Message::runId
      // as as a name appendix and copies the 'name-mangeled' files to
      // directory Message::outDir2. If cpmapplication.cpp is among the
      // source files of the project, entries in the file cpmconfig.ini
      // determine whether CpmSystem::docRun(bool) will be called
      // automatically and which preexisting directory will be used as
      // Message::outDir2.

   static ofstream outData;
      // writing via '>>' to Message::outData is a second
      // (unidirectional) communication facility of the system.
      // After program run, one will find the text file cpmdata.txt in
      // the execution directory that contains all what was written to
      // stream Message::outData and to the text file
      // text file cpmdata.txt in the directory Message::outDir1. Function
      // CpmSystem::docRun(bool) will also rename and copy this file if the
      // boolean argument is given the value true.

   static void onStatusBar(Word const& w, Z barSegment=1);
      // stores w in the Word memory for 'pane'=barSegment
      // If possible for the system, it displays the Words
      // sttmes1,...sttmes4 on the status bar of the application
      // window. The status bar has several segments, numbered
      // 1,2,... and the second argument determines the field
      // on which the message shall be placed. Presently
      // there are 4 such fields. In CONSOLE applications, only
      // w is displayed on the console.
      // For barSegment<=0 no action is done

   static void setMaxMes(Z mes=1000){maxNumMes=mes;}
      // setting the maximum number of (non-enforced) messages to be
      // written on out.

   // writing to the status bar

   static void progress(Word const& taskName, R frac, Z barSegment=1,
      Z field=3); // field =3 is usual
      // Displays task progress on the status bar.
      // for 0<=frac<=1 we get 000, 001, 002, ... 999, 1000
      // as a sufficient information on the degree of completion, only if
      // a digit changes there is change in the display. This is the
      // situation for field==3; for field=2 we go from
      // 00 to 100; etc. Only field=1,2,3,4,5 work this way; for any other
      // value one gets the same output as for field==4.

// writing to out

   static void report(Word const& text, Z val);
      // writes the number val on out if verbose>0

   static void report(Word const& text, R val);
      // writes the number val on out if verbose>0

// writing to both the status bar and out

   static void value(Word const& text, R val, Z barSegment=1);
      // writes the number val on the status bar and out if verbose>0

   static void values2(Word const& text, R val1, R val2, Z barSegment=1);
      // writes the numbers val1 and val2 on the status bar and out if
      // verbose>0

   static void values3(Word const& text, R val1, R val2, R val3,
      Z barSegment=1);

   static void values4(Word const& text, R val1, R val2, R val3,
      R val4, Z barSegment=1);

   static void value(Word const& text, Z val, Z barSegment=1);
         // writes the number val on the status bar and out if verbose>0

   Message(void);
      // initializes Message::out (if not done earlier)

   explicit Message(Word const& w);
      // initializes out (if not done earlier) and writes w to it

/* gets never called
   ~Message(){
      cout<<"Message destructor called"<<endl;
      out.flush();
      outData.flush();
   }
*/

// Message and warning only differ in the wording of the output.
// Two Word arguments allow to specify the class in the scope of which
// the message is created. For barSeg<=0, no action on the status bar.
// Reason: Sometimes it is important to write some data
// to cpmcerr which would overload and disturb the status display
// on the status bar. If barSeg<0, the message goes to the
// console if we have one, even if CPM_NOGRAPHICS is not defined.

   static void warning(Word const& text, bool fileAlways=false);
      // A simple warning message function writing on out.
      // If fileAlways is false, a limitation of messages to
      // be written to file will take place to avoid slowing down program
      // execution too much.

   static void warning(std::ostringstream const& ost,
      bool fileAlways=false)
   { Message::warning(Word(ost),fileAlways);}

   static void message(Word const& text, Z barSeg=1,
      bool fileAlways=false);

   static void message(std::ostringstream const& ost, Z barSeg=1,
      bool fileAlways=false)
   { Message::message(Word(ost),barSeg,fileAlways);}

   static void message(Word const& className, Word const& text,
      bool fileAlways=false);

   static void warning(Word const& className, Word const& text,
      bool fileAlways=false);

   static void message(Z mL, Word const& text, Z barSeg=1,
      bool fileAlways=false);
      // messaging takes place only if cpmverbose>=mL

   static void urgentMessage(Word const& text)
      // a message that will always be given
   {
      Z verbMem=verbose;
      verbose=1; message(text, 0, false); verbose=verbMem;
   }

   static void setSil(bool sil){ silent=sil;}
      // set silent

   static bool getSil(){ return silent;}

   static Word getRunId(){ return runId;}
      //: get run ID

   static void setRunId(Word const& id){ runId=id;}
      //: set run ID

   static void augRunId(Word const& topic)
      //: augment run ID
   { if (topic.dim()>0) runId=topic&"_"&runId;}
      // allows making the ID more 'speaking'

   static bool getAppRunId(){ return appRunId;}
      //: get append run ID

   static void setAppRunId(bool app){ appRunId=app;}
      //: set append run ID

   static Word getInpDir(){ return inpDir;}
      //: get input directory

   static Word getOutDir1(){ return outDir1;}
      //: get output directory 1

   static Word getOutDir2(){ return outDir2;}
      //: get output directory 2

   static Word getFontDir(){ return fontDir;}
      //: get font directory

   static Word ext(Word const& w){ return glue(inpDir,w);}
      //: extend
      // Returns a file name which can be used to open a file.
      // For instance,
      //    IFileStream ifs(Message::ext("basics.txt"));
      //    Z sel; ifs>>sel;
      // reads an integer sel from file "basics.txt" which will be
      // searched in the directory determined by Message's conception
      // of a input directory. This allows different programs (with
      // different execution directories) access the same input data
      // for reliable performance comparisons. The different
      // programs need to mention in their respective files cpmconfig.ini
      // the same directory location as input directory.
      // It also allows to test a single program with various input
      // data sets (e.g. an experimental one and one already delivered
      // to a customer) by changing the content of the programs
      // cpmconfig.ini. This file may e.g. contain the lines
/*
      d:/cpm/pala
         // stable delivery data
      d:/cpm/palaX
         // experimental data
*/
   // Such entries between which one can switch by commentarization,
   // help to keep track of what one is doing.

   static Word extOut(Word const& w)
   {
      if (w.isVoid()) return Word(); else return glue(outDir2,w);
   }
      //: extend (for) output directory (2)
      // Notice that only outDir2 (not outDir1) can be set via config.ini
      // The void Word as filename as argument in some filing function
      // is always meant as the direction to ignore this function call.
      // So, it should not be transformed in a non-void Word by appending
      // a directory name to it. Changing the old and heavily used function
      // glue(...) to take this into account seems dangerous.

};

void error(Word const& w);
void error(std::ostringstream const& ost);
void error(Word const& w1, Word const& w2);
void lethalError(Word const& w);
   // does not write on cpmcerr.txt

/////////////////////////// class Error /////////////////////////////

class Error: public Message{ // used if C+- classes throw errors
    // No new data to Message, thus also a pure action class.
    // Since the action is determined by the constructor, the
    // functionality is new.
    // Inheritence allows to use the data and methods of Message to use
    // for implementation.

public:

explicit Error(Word const& messageText=Word("unspecified error"));
       // reporting a text in case of error, throws terminating exception

   Error(Word const& className, Word const& messageText);
       // reporting a class name and a text in case of error,
       // throws terminating exception
   friend void error(Word const& w);
      // terminates execution only if Message::debug>0
   friend void error(std::ostringstream const& ost);
   friend void error(Word const& w1, Word const& w2);

};

inline void error(std::ostringstream const& ost)
   { CpmSystem::error(Word(ost));}

struct Exception{ // simple debugging tool
// instead of saying cpmmessage(" reached point 3"); one simply says
// Exception(3);
   Z i_;
   Exception(Z i):i_(i){ Message::message("Exception "&Word::write(i_));}
};

//////////////// class Timer ///////////////////////////////////////////

void wait(R t, Z pane=0);
      // delays program run for t seconds (especially to have time to
      // look to some screen information in absense of getchar() interupt
      // in Win32s programs. When the thread of control reaches a wait
      // statement it is not guaranteed that all graphical effects
      // created earlier on video memory can actually be seen on screen!
      // If the last argument is equal to one of the valid pane indexes
      // 1,2,3,4 and cpmverbose>0 the remaining time in seconds will be
      // shown there.

R time(void);
      // returns the time at evaluation instant in seconds with respect
      // to the first call to time(). Relies on system function
      // gettimeofday.

class Timer{ // implements e.g. getSecondsLeft()
   R tBegin,tEnd;
   bool act;
public:
   Timer():tBegin(0),tEnd(0),act(false){}
   explicit Timer(R sp): tBegin(CpmSystem::time()),tEnd(tBegin+sp),
      act(true){}
      // sets tBegin to the time of function call
      // and tEnd to tBegin+sp. All times in seconds
   R getSecondsLeft()const;
      // returns the remaining span to tEnd expressed in seconds
      // For the default counter this function always returns 137*3600
   R getHoursLeft()const;
      // returns getSecondsLeft()/3600
   bool timeOut()const{ return getSecondsLeft()<=0.;}

};

///////////////////////// class IFileStream ////////////////////////////
//
// typical usage of file stream classes
// Word fIn("c:/d/cpm/myfile.txt");
// Word fOut("log.txt");
// IFileStream infi(fIn);
// OFileStream outfi(fOut);
// V<R> vr;
// vr.scanFrom(infi());
// vr.prnOn(outfi()); // notice '()' in using the streams
// In cpmfile.h there will be defined classes OFile and IFile which
// define assignment and copy.
// See also classes IFile and OFile in cpmfile.h for files which
// satisfy the strict value interface.

const bool binDef=true;
//const bool binDef=false;
   // true is proven, false for experimentation
   // default value for setting the argument 'binary'
   // for opening streams. Introduced 2004-04-29

class IFileStream{ // input file stream

   ifstream* fs_;
   const Word name_;
   typedef IFileStream Type;
   CPM_INVAR(IFileStream)

public:

   explicit IFileStream(Word const& fileName,
      bool safe=true, bool binary=binDef);
      // If the second argument is true one gets an error if the
      // file could not be opened, otherwise only a warning will
      // be issued.

   IFileStream():fs_(0){}

   virtual ~IFileStream(){ fs_->close(); delete fs_;}

   char readChar(){ char res('x'); *fs_>>res; return res;}
      //: read character
      // reading a char from stream

   string readWord(){ string res; *fs_>>res; return res;}
      //: read word
      // reads the next contiguous group of non-whitespace characters

   string readLine(){ string res; std::getline(*fs_,res); return res;}
      //: read line
      // reads the next line with all charcters also whitespace

   bool isVal()const
      //: is valid
      // returns stream status
   { if (!*fs_||fs_->bad()) return false; else return fs_->good(); }

   Word getName()const{ return name_;}
      //: get name

   ifstream& operator()(void){return *fs_;}
      // returns the object for direct access

   bool skipComments();
      // over-reads lines starting with '/', ';', '*', or '#'
};

//////////////////////// class OFileStream ///////////////////////////////

class OFileStream{ // output file stream

   ofstream* fs_;
   const Word name_;
   typedef OFileStream Type;
   CPM_INVAR(OFileStream)
public:

   explicit OFileStream(Word const& fileName,
      bool safe=true, bool binary=binDef);

   OFileStream():fs_(0){}

   virtual ~OFileStream(){ fs_->close();delete fs_;}

    bool isVal()const
       //: is valid
       // returns stream status
    { if (!*fs_||fs_->bad()) return false; else return fs_->good(); }

   Word getName()const{ return name_;}
      //: get name

   ofstream& operator()(void){return *fs_;}
      // returns the object for direct access,
      // this is all one needs to write to the object
};

void copyFile(Word const& fileNameIn, Word const& fileNameOut);
   //: copy file
   // lean file copy function, closely following BS3, p. 638 function main
   // The first argument needs to be the name of a file that can be opened.
   // The output file will be created (and written to) and will have the
   // name given by the second argument. Since as the result of a
   // successful function call the output file exists in the file system
   // of the machine, the file needs not to be returned.
   // Any malfunction exits from the program after having left a message
   // on cout. Since the raison d'etre of this function is to copy
   // the files cpmcerr.txt and cpmdata.txt they are not allowed to
   // write on these files.

void docRun(bool all=false);
   //: document run
   // creates a copy of cpmcerr.txt and, if all==true, also cpmdata.txt
   // with Message::getRunId() appended to the name in the directory
   // Message::outDir2. Notice that Message::outDir1 is the place where
   // the original cpmcerr.txt and cpmdata.txt are placed.

} // namespace

// An assert macro which is fully under the control of the C+- class
// system. The first argument should be a piece of C++ code which
// evaluates to bool. (Thus no C++-function can do the job since there
// are no variables that hold pieces of code)
// Example of usage:
//    cpmassert(x>0,"norm in function eval should be positive");
// notice that the final ';' is not needed

#define cpmassert(X,Y)\
if (CpmSystem::Message::debug==1){\
if (!(X)) CpmSystem::Message::warning(CpmRoot::Word(#X)&\
   " violated: "&Y);}\
else if (CpmSystem::Message::debug==2)\
{ if (!(X)) CpmSystem::error(CpmRoot::Word(#X)&" violated: "&Y);}

///////////////////// short names ////////////////////////////////////////
#define cpmcerr         CpmSystem::Message::out
#define cpmdata         CpmSystem::Message::outData
#define cpmverbose      CpmSystem::Message::verbose
#define cpmtrigger      CpmSystem::Message::trigger
#define cpmsilent       CpmSystem::Message::getSil
#define cpmdbg          CpmSystem::Message::debug
   // the name cpmdebug, which would be natural here, was defined
   // earlier, see cpmtypes.h
#define cpmmessage      CpmSystem::Message::message
#define cpmurgent       CpmSystem::Message::urgentMessage
#define cpmwarning      CpmSystem::Message::warning
#define cpmprogress     CpmSystem::Message::progress
#define cpmreport       CpmSystem::Message::report
#define cpmstatus       CpmSystem::Message::onStatusBar
#define cpmrunid        CpmSystem::Message::getRunId()
#define cpmerror        CpmSystem::error
#define cpmwait         CpmSystem::wait
#define cpmvalue        CpmSystem::Message::value
#define cpmvalues2      CpmSystem::Message::values2
#define cpmvalues3      CpmSystem::Message::values3
#define cpmvalues4      CpmSystem::Message::values4
#define cpmtime         CpmSystem::time
#define cpmexc          CpmSystem::Exception

#endif // guard
