//? cpmtests.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_TESTS_H_
#define CPM_TESTS_H_

/*
Purpose: Uniform test facility for finding out in an
unambiguous way whether a class satisfies a given
interface, such as the strict value interface or the
r-interface.
*/

#include <cpmtypes.h>
#include <cpmc.h>
#include <cpmvr.h>
#include <cpmsr.h>
#include <cpmfr.h>
#include <cpmp.h> // for Vp
#include <iomanip> // for std::setprecision

namespace CpmTests{

   using namespace CpmStd;

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::Word;
   using CpmRoot::Root;
   using CpmArrays::V;
   using CpmArrays::Vo;
   using CpmRootX::B;
   using CpmSystem::IFileStream;
   using CpmSystem::OFileStream;
   using std::setprecision;

// We assume that type T defines the operator  binary bool-valued
// operator ==
template <class T>
class Sym{ // testing symmetry of equality
public:
   Sym(){}
   bool operator()(T const& t1, T const& t2)
      // T == T is symmetric if this function returns true for all
      // input arguments (then we say: 'sym==true')
   {
      if (t1==t2) return (t2==t1); else return true;
   }
};
// Typical usage;
// T t1...;
// T t2...;
// Sym<T> s;
// bool b=s(t1,t2); // b ==true says that symmetry has been confirmed
// for the particular objects t1 and t2.


template <class T>
class Trans{ // testing transitivity of equality
public:
   Trans(){}
   bool operator()(T const& t1, T const& t2, T const& t3)
// T == T is an equivalence relation if sys==true and trans==true
   {
      if (t1==t2 && t2==t3) return (t1==t3); else return true;
   }
};

template <class T>
class DefaultConstructor{ // testing consistency of default constructor
public:
   DefaultConstructor(){}
   bool operator()(T const& t)
   // value of  t not used, carries type information
   {
      t; // to avoid warning concerning not making use of the argument
      T t1; T t2; return (t1==t2);
   }
};

template <class T>
class CopyConstructor{ // testing consistency of copy constructor
public:
   CopyConstructor(){}
   bool operator()(T const& t)
   {
      T t1(t); return t1==t;
   }
};

template <class T>
class Assignment{ // testing consistency of assignment
public:
   Assignment(){}
   bool operator()(T const& t1, T const& t2)
   {
      T tStart(t1);
      tStart=t2;
      return tStart==t2;
   }
};

template <class T>
class StrictAssignment{ // testing strict consistency of assignment
// Here we assume minimal infrastructure (only copy,assignement,equality).
public:
   StrictAssignment(){}

   bool operator()(T const& t1, T const& t2, T const& t3)
      // returns true if OK
   {
      T tStart(t1);
      T t2c(t2);
      tStart=t2c;
      T tMem(tStart);
      t2c=t3; // source for tStart's value re-defined
      return tStart==tMem; // detects change of tStart
         // (which should not happen).
   }

};

// Notice that so far we returned true for passing the test. The following
// two functions return the number of failures instead!

template <class T>
class StrictAssignment2{ // testing strict consistency of assignment
// for values created inside a function body. Needs more infrastructure
// of T than the previous class StrictAssignment<T>. Particularly we
// need a random generator and the distance function.
public:
   StrictAssignment2(){}
   static T gen(T const& t, Z j){ T res = Root<T>(t).ran(j); return res;}
      // studying return values which were primarily created inside a
      // function block
   R operator()(T const& t1)
      // returns an quantitative error varying between 0 (good) and 1
      // (bad)
   {
      T t2=gen(t1,7);
      T t3=gen(t2,6);
      T t4=gen(t3,5);
      T t4M(t4);
      // swapping t2 and t3
      T tTemp=t3;
      t3=t2;
      t2=tTemp;
      return Root<T>(t4).dis(t4M);
   }
};


template <class T>
class ValueBehavior{ // testing the property of a class T
   // to be a 'value class'. Relies on operator == and
   // three instances of T on input.
public:
   ValueBehavior(){}
   R operator()(T const& t1, T const& t2, T const& t3)
      // returns the number of errors
   {
      Z mL=3;
      Word loc("ValueBehavior::operator()(T,T,T)");
      CPM_MA
      DefaultConstructor<T> dc;
      CopyConstructor<T> cc;
      Assignment<T> a;
      StrictAssignment<T> sa;
      Z fc=0; // failure counter
      bool b1=dc(t1);
      if (!b1) {fc++;cpmmessage(loc&" failure in DefaultConstructor");}
      bool b2=cc(t1);
      if (!b2) {fc++;cpmmessage(loc&" failure in CopyConstructor");}
      bool b3=a(t1,t2);
      if (!b3) {fc++;cpmmessage(loc&" failure in Assignment");}
      bool b4=sa(t1,t2,t3);
      if (!b4) {fc++; cpmmessage(loc&" failure in StrictAssignment");}
   //   bool b5=sa(t1);
   //   if (!b5){fc++; cpmmessage(loc&" failure in StictAssignment II");}
      CPM_MZ
      return R(Z(fc));
   }
};

// Let T be a type for which the statement
// strictAssignment<T> sa; compiles then we say that the
// T implements the strict assignment on compilation.
// If we find sa(t1,t2,t3)==true for all T typed argument triplets
// based on the logic of the code we say that T implements the strict
// assignment.
// If we verify sa(t1,t2,t3)==true by running a program with sufficiently
// many arguments we may say that the strict assignment was tested with
// positive result.

// If for a fixed triplet of types T1, T2, C this
// function is true for all arguments, we says that C is a
// polymorphic single-slot-container for T1 and T2.

template <class T1, class T2, class C>
class PolymorphicSingle{ // testing polymorphic containers
public:
   PolymorphicSingle(){}
   bool operator()(const T1& t1, const T2& t2, const C& c )
   {
      C cc(c);
      cc()=t1;
      T1 tc1(cc());
      bool b1=(t1==tc1);
      cc()=t2;
      T2 tc2(cc());
      bool b2=(t2==tc2);
      return b1 && b2;
   }
};

template <class T1, class T2, class C>
class PolymorphicMulti{ // testing polymorphic containers
public:
   PolymorphicMulti(){}
   bool operator()(const T1& t1, const T2& t2, const C& c, Z i1, Z i2 )
   {
      if ( i1>=1 && i1<=c.dim() && i2>=1 && i2<=c.dim() && i1!=i2 ){
         C cc(c);
         cc[i1]=t1;
         cc[i2]=t2;
         T1 tc1(cc[i1]);
         T2 tc2(cc[i2]);
         return t1==tc1 && t2==tc2;
      }
      else return true;
   }
};

// If for a fixed triplet of types T1, T2, C this
// function is true for arguments, we says that c is a
// polymorphic multi-slot-container for T1 and T2.

// We will see that Vp<T> is a polymorphic multi-slot-container for all
// types T1, T2 equal to or derived from T.

template <class T>
class TestBase{ // base class for classes which test that class T
// implements a certain interface
protected:
   R errorSum;

   static R comp(Word top, T const& lhs, T const& rhs);
      // means for comparing results

   static R larger(Word top, R xL, R xS);
      // means for testing xL > xS. Returns 0 if the (expected)
      // inaquality is satisfied and xS-xU else.

   static R equal(Word top, R x1, R x2);
      // means for testing x1 == x2. Returns (x1-x2).abs()

   TestBase():errorSum(0.){}
public:
   static void wrt(ostringstream& ost){
      Word wd=Word(ost.str());
      wd.prnOn(cpmcerr);
      wd.prnOn(cout);
   }
   static void wrt(Word const& wd){
      wd.prnOn(cpmcerr);
      wd.prnOn(cout);
   }
   R val()const{return errorSum;}
};

template <class T>
R TestBase<T>::comp(Word top, T const& lhs, T const& rhs)
{
   using namespace CpmSystem;
   const R tiny=1e-4;

   R al=CpmRoot::absT<T>(lhs);
   R ar=CpmRoot::absT<T>(rhs);
   Word w1="Warning: trivial argument in ";
   Word w2=top;
   ostringstream ost;
   if (al==0. || ar==0.){
      ost<<endl<<(w1&w2);
   }
   R newError=Root<T>(lhs).dis(rhs);
   ost<<endl<<"Test of "
      <<top<<endl<<" error = "<<newError<<endl
      <<"cpmabs(lhs)= "<<al<<endl
      <<"cpmabs(rhs)= "<<ar<<endl
     ;
   if (newError>tiny || !cpmiva(newError)){
     ost<<endl<<" lhs="<<endl;
     CpmRoot::prnT<T>(lhs,ost);
     ost<<endl<<" rhs="<<endl;
     CpmRoot::prnT<T>(rhs,ost);
   }
   wrt(ost);
   return newError;
}

template <class T>
R TestBase<T>::larger(Word top, R xL, R xS)
{
   using namespace CpmSystem;
   Word w1="Warning: trivial argument in ";
   Word w2=top;
   ostringstream ost;
   if (xL==0. || xS==0.){
      ost<<endl<<(w1&w2);
   }
   R newError=xS>xL ? xS-xL : 0.;
   ost<<endl<<"Test of "
      <<top<<endl<<" error = "<<newError<<endl
      <<"xL= "<<xL<<endl
      <<"xS= "<<xS<<endl;
   wrt(ost);
   return newError;
}

template <class T>
R TestBase<T>::equal(Word top, R x1, R x2)
{
   using namespace CpmSystem;
   Word w1="Warning: trivial argument in ";
   Word w2=top;
   ostringstream ost;
   if (x1==0. || x2==0.){
      ost<<endl<<(w1&w2);
   }
   R newError=cpmabs(x1-x2);
   ost<<endl<<"Test of "
      <<top<<endl<<" error = "<<newError<<endl
      <<"x1= "<<x1<<endl
      <<"x2= "<<x2<<endl;
   wrt(ost);
   return newError;
}

////////////// testing the value interface //////////////////////////////

template <class T>
class Test_v: public TestBase<T>{ // testing the value interface

   typedef TestBase<T> Base;

public:
   Test_v(Z tvs=1, Z rep=4);
   // tvs: test value size ~ complexity of test
};

template <class T>
Test_v<T>::Test_v(Z tvs, Z rep):Base()
      // Tests all relations concerning the value interface and returns
      // an error sum
{
   using namespace CpmRoot;
   using namespace CpmRootX;

   const R dSmall=1e-4;
   const R dTiny=1e-12;

   Word top;
   Word loc("Test_v<T>::Test_v(Z tvs, Z rep)");
   R res=0;
   Z counter=0;
   T xd,x;
   T xt=xd.test(tvs);
   cpmdbg=1; // experiment
   while (counter<rep){
      counter++;
      x=xt.ran();
     //  cpmassert(xt.dis(x)>dSmall,loc); // the random generator should
      // generate
      // new values, if it does so only after many iterations it is
      // probably not reliable, so we expect this to be valid for
      // sufficiently small
      // dSmall
      T y=xt.ran();
      cpmassert(y.dis(x)>dSmall,loc);
      T z=xt.ran();
      cpmassert(z.dis(x)>dSmall,loc);
      cpmassert(z.dis(y)>dSmall,loc);
      R r1=y.dis(z);
      R r2=z.dis(y);
      cpmassert(cpmabs(r1-r2)<dTiny,loc);

// now x,y,z are supposed to be sufficiently random to make test
// meaningful
      top="is copy equal to original ?";
      T xl(z);
      res+=Base::comp(top,xl,z);
      top="is assigned value equal to original ?";
      y=x;
      res+=Base::comp(top,y,x);
// This was all to be shown
      xt=z; // to start the new loop were the old one ended
   }
   ostringstream ost;
   ost<<endl<<"  Test of Value Interface; errorsum= "<<res<<endl;
   Base::wrt(ost);
   Base::errorSum=res;
}

///////////// testing the strict value interface ///////////////////////

template <class T>
class Test_sv: public TestBase<T>{ // testing the strict value interface

   typedef TestBase<T> Base;

public:
   Test_sv(Z tvs=1, Z rep=4);
   // tvs: test value size ~ complexity of test
};

template <class T>
Test_sv<T>::Test_sv(Z tvs, Z rep):Base()
      // Tests all relations concerning the strict value interface and
      // returns an error sum
{
   using namespace CpmRoot;
   using namespace CpmRootX;

   const R dSmall=1e-4;
   const R dTiny=1e-12;

   Word top;
   Word loc("Test_sv<T>::Test_sv(Z tvs, Z rep)");
   R res=0;
   Z counter=0;
   T xd,x;
   T xt=xd.test(tvs);
   while (counter++<rep){
      x=xt.ran();
      cpmassert(xt.dis(x)>dSmall,loc); // the random generator should
      // generate new values, if it does so only after many iterations it
      // is probably not reliable, so we expect this to be valid for
      // sufficiently small dSmall
      T y=xt.ran();
      cpmassert(y.dis(x)>dSmall,loc);
      T z=xt.ran();
      cpmassert(z.dis(x)>dSmall,loc);
      cpmassert(z.dis(y)>dSmall,loc);
      R r1=y.dis(z);
      R r2=z.dis(y);
      cpmassert(cpmabs(r1-r2)<dTiny,loc);
      T zr=z;

// now x,y,z are supposed to be sufficiently random to make test
// meaningful
      top="is copy equal to original ?";
      T xl(zr);
      z=z.ran();
      z=z.ran(); // although zr was made from z, this change in z
         // should not affect zr, thus zr and xl have still to be equal
      res+=Base::comp(top,xl,zr);

      top="is assigned value equal to original ?";
      y=x;
      T x1=x;
      T x2=y;
// messing up the originals; x1 and x2 should remain unchanged
      x=x.ran();
      x=x.ran();
      y=y.ran();
      y=y.ran();
      res+=Base::comp(top,x1,x2);
// This was all to be shown
      xt=z; // to start the new loop were the old one ended
   }
   ostringstream ost;
   ost<<endl<<" Test of Strict Value Interface; errorsum="<<res<<endl;
   Base::wrt(ost);
   Base::errorSum=res;
}

/////////// testing the r-interface /////////////////////////////////////

template <class T, class S>
class Test_r: public TestBase<T>{ // testing the r-interface
   // class for testing classes that
   // implement the r-interface
   // Testing e.g. Vr<R> is done e.g. by the following code
   // Test_r<Vr<R>,R> a(10);

   typedef TestBase<T> Base;

public:
   Test_r(Z tvs=1 , Z output=0, Z write=0);
   // tvs: test value size ~ complexity of test
};

template <class T, class S>
Test_r<T,S>::Test_r(Z tvs, Z output, Z write):Base()
      // Tests all relevant mathematical relations and returns
      // an error sum
      // output 0 : only end of test and error sum displayed
      // output 1 : output of every test. The potentially
      //      time consuming output of the test object is suppressed
      // output 2: only the first component of the test object is ??
      // displayed ??
      // output 3: the test object is displayed in full detail ??
      // This form only works for classes (not for the types R and Z. As
      // a compensation it shows the favorable property of not having to
      // know the namespace of T, since all access is intrinsic.??
      // If write is different from 0, a write/read test to and from file
      // will be executed. Only if write > 1,the difference between
      // written and read quantity will enter the error sum. This helps to
      // asses the error behavior for the newly included multiple
      // precision faclity, where otherwise the error would be dominated
      // by the write precision which I set independently from the number
      // precision.
{

   using namespace CpmSystem;
   using namespace CpmFunctions;
   using CpmRoot::nameOf;
   static const Z multOrder=4;
      // order test is made on an array of
      // multOrder*tvs elements
   //using Base::comp;

   Z mL=1;
   Word loc("Test_r(Z,Z,Z)");
   CPM_MA

   Z nOrd=multOrder*tvs;

   T t, v, x, y, z, xl, xr;
   Word top;

   R res=0.;
   ostringstream ost;
   ost<<endl<<" Test of class "<<t.nameOf()<<endl;
   Base::wrt(ost);

   if (output){
      cpmcerr<<endl<<
      " The distinguished objects associated with the class are: "<<endl;
   }

   T xd;
   T xt=xd.test(tvs);
   if (output){
      cpmcerr<<endl<<"test:";
      xt.toWord().prnOn(cpmcerr);
   }
   T xzero=xt.net(0);
   if (output){
      cpmcerr<<endl<<"zero:";
      xzero.toWord().prnOn(cpmcerr);
   }
   T xunit=xt.net(1);
   if (output){
      cpmcerr<<endl<<"unit:";
      xunit.toWord().prnOn(cpmcerr);
   }

   v=xt;
   x=xt.ran(1);
   y=xt.ran(2);
   z=xt.ran(3);
   T xWrite=z;

   if (output){
      cpmcerr<<endl<<"first ran:";  x.toWord().prnOn(cpmcerr);
      cpmcerr<<endl<<"second ran:"; y.toWord().prnOn(cpmcerr);
      cpmcerr<<endl<<"third ran:";  z.toWord().prnOn(cpmcerr);
   }
   T v4=x;
   T v5;
   T v6=y;
   v6=v5=x;
   xl=v6;
   xr=v4;
   top="copy constructor, assignement, and equality";
     res+=Base::comp(top,xl,xr);

   xl=x+y; xr=y+x; top="+ commutativity";
   res+=Base::comp(top,xl,xr);
   xl=(x+y)+z; xr=x+(y+z); top="+ associativity";
   res+=Base::comp(top,xl,xr);
   xl=(x*y)*z; xr=x*(y*z); top="* associativity";
   res+=Base::comp(top,xl,xr);

   T xZero=x.net(0);

   xl=-x; xr=xZero-x; top="sign change";
   res+=Base::comp(top,xl,xr);
   xl=x*(y+z); xr=x*y+x*z; top="left-distributivity";
   res+=Base::comp(top,xl,xr);
   xl=(x+y)*z; xr=x*z+y*z; top="right-distributivity";
   res+=Base::comp(top,xl,xr);
   // not clear why con(x*y) is not understood for T=Fr<Z,R>
   T xTemp=x*y;
   xl=xTemp.con(); xr=y.con()*x.con();
   top="conjugation commutativity";
   res+=Base::comp(top,xl,xr);
   x=z;
   T xu=x;

   xu=xu.net(1);
   if (output){
      cpmcerr<<endl<<"unit=";
      xu.toWord().prnOn(cpmcerr);
   }

   if (output){
      cpmcerr<<endl<<"x=";
      x.toWord().prnOn(cpmcerr);
   }

   T xi=x.inv();

   if (output){
      cpmcerr<<endl<<"xInv=";
      xi.toWord().prnOn(cpmcerr);
   }

   xl=x*xi; xr=xu; top="right inverse";
   res+=Base::comp(top,xl,xr);

   xl=xi*x; xr=xu; top="left inverse";
   res+=Base::comp(top,xl,xr);
// properties with scalars
   S s;
   s=CpmRoot::testT<S>(s,tvs);
   S sZero=CpmRoot::netT<S>(s,0);
   S sUnit=CpmRoot::netT<S>(s,1);
   S sInit=sUnit+sUnit+sUnit;
   sInit+=sInit;
   Z count=0;
   S s1=CpmRoot::ranT<S>(sInit,count);
   const Z countMax=100;
   while (s1==sZero || s1==sUnit){
      s1=CpmRoot::ranT<S>(sInit,++count);
      if (count>countMax){
         cpmwarning(
         "Test_r<T,S>: too many iterations (bad random generator of S)"
         );
         break;
      }
   }
   if (output){
      Word mes="iterations needed to get s1: "&cpmwrite(count);
      cpmcerr<<endl<<mes;
   }
   count=0;
   S s2=CpmRoot::ranT<S>(s1,++count);
   while (s2==sZero || s2==sUnit || s2==s1){
      s2=CpmRoot::ranT<S>(s1,++count);
      if (count>countMax){
         cpmwarning(
         "Test_r<T,S>: too many iterations (bad random generator of S)"
         );
         break;
      }
   }
   if (output){
      Word mes="iterations needed to get s2: "&cpmwrite(count);
      cpmcerr<<endl<<mes;
   }
   xl=s1*(x+y); xr=s1*x+s1*y; top="*s left-distributivity";
   res+=Base::comp(top,xl,xr);
   xl=(x+y)*s2; xr=x*s2+y*s2; top="*s right-distributivity";
   res+=Base::comp(top,xl,xr);
   S s12=s1*s2; // workaround:
      // not clear why xl=(s1*s2)*x; gets not compiled in some cases
   xl=s12*x; xr=s1*(s2*x); top="s*s left-distributivity";
   res+=Base::comp(top,xl,xr);
   xl=x*(s1*s2); xr=(x*s1)*s2; top="s*s right distributivity";
   res+=Base::comp(top,xl,xr);

   xl=x; xl*=sUnit; xl=xl/sUnit; xr=(x*s1); xr/=s1;
   top="/s and *s compatibility";
   res+=Base::comp(top,xl,xr);

// order properties
// rather simple test of ordering consistency

   T xMax=CpmRootX::sup(x,y,z);
   T xMin=CpmRootX::inf(x,y,z);

   if ( xMin.priorTo(xMax) || xMin.equalTo(xMax)  ){
      cpmcerr<<endl<<"comparison test OK"<<endl;
      cout<<endl<<"comparison test OK"<<endl;
   }
   else{
      cpmcerr<<endl<<"comparison test failed"<<endl;
      cout<<endl<<"comparison test failed"<<endl;
      res+=1;
   }

// let us order a list of random values
   CpmArrays::Vo<T> list(nOrd);
   Z i;
   for (i=1;i<=nOrd;i++){
      list[i]=xt.ran(i);
   }
   list.order();
   for (i=2;i<=nOrd;i++){
      if ( list[i].priorTo(list[i-1]) ){
         cpmmessage("comparison failure");
         cout<<endl<<"comparison failure";
         res+=1;
      }
   }

   if (write){
      cpmmessage("Writing path entered");
      Word filename(nameOf(xWrite));
      filename=filename.makeFileName();
      Word filename1="w-"+filename+".txt";
      OFileStream out1(filename1);
      if (!out1()) cpmerror(nameOf(xWrite),
         "test: output file cannot be opened");
      const int precWrite=12;
      out1()<<setprecision(precWrite);
      out1()<<endl<<"// I/O Test:"<<endl;
      out1()<<"// Object to be written: "<<endl;
      xWrite.prnOn(out1());
      cpmmessage("First writing done");
      if (!out1()) cpmerror(nameOf(xWrite),
         "out1 not OK before closing");
      IFileStream in(filename1);
      if (!in()) cpmerror(nameOf(xWrite),
         "test: input file cannot be opened");
      xr.scanFrom(in());
      cpmmessage("First reading done");
      if (!in()) cpmerror(nameOf(xWrite),
         "in not OK after reading and before closing");
      Word filename2="r-"+filename+".txt";
      OFileStream out2(filename2);
      if (!out2()) cpmerror(nameOf(xWrite), "test: out2 file cannot be\
       opened");
      out2()<<"// I/O Test. This is what we got back:"<<endl;
      xr.prnOn(out2());
      cpmmessage("Second writing done");
      if (!out2()) cpmerror(nameOf(xWrite),
         "out2 not OK before closing");
      xl=xWrite; top="file I/O";
      if (write>1) res+=Base::comp(top,xl,xr);
            // otherwise read write reproduction error will not be added
            // to the eror bilance, since it is not purely numerical in
            // nature.
      cpmmessage("Leaving writing path");
   }

   ostringstream ost2;
   ost2<<endl<<"  Test of "<<nameOf(xWrite)<<" errorsum= "<<res<<endl;
   Base::wrt(ost2);
   Base::errorSum=res;
   CPM_MZ
}

///// testing the interface of a commutative ring with unity ///////////
// Essentially Test_r with the test of inversion left out
// and some properties concerning relation of * and abs()
// optionally added.
// Test_rm : r modified

template <class T, class S>
class Test_rm: public TestBase<T>{
   // Testing the modified r-interface.
   // class for testing classes that
   // implement the cr-interface
   // Testing e.g. GaZ<R> is done e.g. by the following code
   // Test_rm<GaZ<R>,R> a(10);

   typedef TestBase<T> Base;

public:
   Test_rm(Z tvs=1 , Z output=0, Z write=0, Z prop=0);
   // tvs: test value size ~ complexity of test
};

template <class T, class S>
Test_rm<T,S>::Test_rm(Z tvs, Z output, Z write, Z prop):Base()
      // Tests all relevant mathematical relations and returns
      // an error sum
      // output 0 : only end of test and error sum displayed
      // output 1 : output of every test. The potentially
      //      time consuming output of the test object is suppressed
      // If write is different from 0, a write/read test to and from file
      // will be executed. Only if write > 1,the difference between
      // written and read quantity will enter the error sum. This helps to
      // asses the error behavior for the newly included multiple
      // precision faclity, where otherwise the error would be dominated
      // by the write precision which I set independently from the number
      // precision.
      // There are three relations of the norm which all are satisfied if
      // T is a C*-algebra. These are ignored for prop=0 and partially
      // or completely tested for larger values of prop.
{

   using namespace CpmSystem;
   using namespace CpmFunctions;
   using CpmRoot::nameOf;
   static const Z multOrder=4;
      // order test is made on an array of
      // multOrder*tvs elements

   Z mL=1;
   Word loc("Test_rm<T>(Z,Z,Z)");
   CPM_MA

   Z nOrd=multOrder*tvs;

   T t, v, x, y, z, xl, xr;
   Word top;

   R res=0.;
   ostringstream ost;
   ost<<endl<<" Test of class "<<t.nameOf()<<endl;
   Base::wrt(ost);

   if (output){
      cpmcerr<<endl<<
      " The distinguished objects associated with the class are: "<<endl;
   }

   T xd;
   T xt=xd.test(tvs);
   if (output){
      cpmcerr<<endl<<"test:";
      xt.toWord().prnOn(cpmcerr);
   }
   T xzero=xt.net(0);
   if (output){
      cpmcerr<<endl<<"zero:";
      xzero.toWord().prnOn(cpmcerr);
   }
   T xunit=xt.net(1);
   if (output){
      cpmcerr<<endl<<"unit:";
      xunit.toWord().prnOn(cpmcerr);
   }

   v=xt;
   x=xt.ran(1);
   y=xt.ran(2);
   z=xt.ran(3);
   T xWrite=z;

   if (output){
      cpmcerr<<endl<<"first ran:";  x.toWord().prnOn(cpmcerr);
      cpmcerr<<endl<<"second ran:"; y.toWord().prnOn(cpmcerr);
      cpmcerr<<endl<<"third ran:";  z.toWord().prnOn(cpmcerr);
   }
   T v4=x;
   T v5;
   T v6=y;
   v6=v5=x;
   xl=v6;
   xr=v4;
   top="copy constructor, assignement, and equality";
     res+=Base::comp(top,xl,xr);

   xl=x+y; xr=y+x; top="+ commutativity";
   res+=Base::comp(top,xl,xr);
   xl=(x+y)+z; xr=x+(y+z); top="+ associativity";
   res+=Base::comp(top,xl,xr);
   xl=(x*y)*z; xr=x*(y*z); top="* associativity";
   res+=Base::comp(top,xl,xr);

   T xZero=x.net(0);

   xl=-x; xr=xZero-x; top="sign change";
   res+=Base::comp(top,xl,xr);
   xl=x*(y+z); xr=x*y+x*z; top="left-distributivity";
   res+=Base::comp(top,xl,xr);
   xl=(x+y)*z; xr=x*z+y*z; top="right-distributivity";
   res+=Base::comp(top,xl,xr);
   T xy=x*y;
   xl=xy.con(); xr=y.con()*x.con();
   top="conjugation commutativity";
   res+=Base::comp(top,xl,xr);
   Z i;
   if (prop>=1){
      // If the iequality is found to hold once,
      // this could be by chance
      Z nTrials=60;
      V<T> trials(nTrials);
      for (i=trials.b();i<=trials.e();++i){
         trials[i]=x.ran();
      }
      top="norm inequality";
      R errTopic=0;
      for (i=trials.b();i<trials.e();++i){
         T xa=trials[i];
         T xb=trials[i+1];
         T xab=xa*xb;
         R xLarger=xa.abs()*xb.abs();
         R xSmaller=xab.abs();
         if (xLarger<xSmaller) errTopic+=(xSmaller-xLarger);
      }
      Base::wrt("Test of "&top&": error="&cpm(errTopic));
      res+=errTopic;
   }
   if (prop>=2){
      top="conjugation norm equality";
      R x1=x.abs();
      R x2=x.con().abs();
      res+=Base::equal(top,x1,x2);
   }
   if (prop>=3){
      top="C* norm equality";
      R x1=x.abs()*x.con().abs();
      R x2=(x.con()*x).abs();
      res+=Base::equal(top,x1,x2);
   }
   x=z;
   T xu=x;
   xu=xu.net(1);
   if (output){
      cpmcerr<<endl<<"unit=";
      xu.toWord().prnOn(cpmcerr);
   }

   if (output){
      cpmcerr<<endl<<"x=";
      x.toWord().prnOn(cpmcerr);
   }
// properties with scalars
   S s;
   s=CpmRoot::testT<S>(s,tvs);
   S sZero=CpmRoot::netT<S>(s,0);
   S sUnit=CpmRoot::netT<S>(s,1);
   S sInit=sUnit+sUnit+sUnit;
   sInit+=sInit;
   Z count=0;
   S s1=CpmRoot::ranT<S>(sInit,count);
   const Z countMax=100;
   while (s1==sZero || s1==sUnit){
      s1=CpmRoot::ranT<S>(sInit,++count);
      if (count>countMax){
         cpmwarning(
         "Test_rm<T,S>: too many iterations (bad random generator of S)"
         );
         break;
      }
   }
   if (output){
      Word mes="iterations needed to get s1: "&cpmwrite(count);
      cpmcerr<<endl<<mes;
   }
   count=0;
   S s2=CpmRoot::ranT<S>(s1,++count);
   while (s2==sZero || s2==sUnit || s2==s1){
      s2=CpmRoot::ranT<S>(s1,++count);
      if (count>countMax){
         cpmwarning(
         "Test_rm<T,S>: too many iterations (bad random generator of S)"
         );
         break;
      }
   }
   if (output){
      Word mes="iterations needed to get s2: "&cpmwrite(count);
      cpmcerr<<endl<<mes;
   }
   xl=s1*(x+y); xr=s1*x+s1*y; top="*s left-distributivity";
   res+=Base::comp(top,xl,xr);
   xl=(x+y)*s2; xr=x*s2+y*s2; top="*s right-distributivity";
   res+=Base::comp(top,xl,xr);
   S s12=s1*s2; // workaround:
      // not clear why xl=(s1*s2)*x; gets not compiled in some cases
   xl=s12*x; xr=s1*(s2*x); top="s*s left-distributivity";
   res+=Base::comp(top,xl,xr);
   xl=x*(s1*s2); xr=(x*s1)*s2; top="s*s right distributivity";
   res+=Base::comp(top,xl,xr);

   xl=x; xl*=sUnit; xl=xl/sUnit; xr=(x*s1); xr/=s1;
   top="/s and *s compatibility";
   res+=Base::comp(top,xl,xr);

// order properties
// rather simple test of ordering consistency

   T xMax=CpmRootX::sup(x,y,z);
   T xMin=CpmRootX::inf(x,y,z);

   if ( xMin.priorTo(xMax) || xMin.equalTo(xMax)  ){
      cpmcerr<<endl<<"comparison test OK"<<endl;
      cout<<endl<<"comparison test OK"<<endl;
   }
   else{
      cpmcerr<<endl<<"comparison test failed"<<endl;
      cout<<endl<<"comparison test failed"<<endl;
      res+=1;
   }
// let us order a list of random values
   CpmArrays::Vo<T> list(nOrd);
   for (i=1;i<=nOrd;i++){
      list[i]=xt.ran(i);
   }
   list.order();
   for (i=2;i<=nOrd;i++){
      if ( list[i].priorTo(list[i-1]) ){
         cpmmessage("comparison failure");
         cout<<endl<<"comparison failure";
         res+=1;
      }
   }
   if (write){
      cpmmessage("Writing path entered");
      Word filename(nameOf(xWrite));
      filename=filename.makeFileName();
      Word filename1="w-"+filename+".txt";
      OFileStream out1(filename1);
      if (!out1()) cpmerror(nameOf(xWrite),
         "test: output file cannot be opened");
      const int precWrite=12;
      out1()<<setprecision(precWrite);
      out1()<<endl<<"// I/O Test:"<<endl;
      out1()<<"// Object to be written: "<<endl;
      xWrite.prnOn(out1());
      cpmmessage("First writing done");
      if (!out1()) cpmerror(nameOf(xWrite),
         "out1 not OK before closing");
      IFileStream in(filename1);
      if (!in()) cpmerror(nameOf(xWrite),
         "test: input file cannot be opened");
      xr.scanFrom(in());
      cpmmessage("First reading done");
      if (!in()) cpmerror(nameOf(xWrite),
         "in not OK after reading and before closing");
      Word filename2="r-"+filename+".txt";
      OFileStream out2(filename2);
      if (!out2()) cpmerror(nameOf(xWrite), "test: out2 file cannot be\
       opened");
      out2()<<"// I/O Test. This is what we got back:"<<endl;
      xr.prnOn(out2());
      cpmmessage("Second writing done");
      if (!out2()) cpmerror(nameOf(xWrite),
         "out2 not OK before closing");
      xl=xWrite; top="file I/O";
      if (write>1) res+=Base::comp(top,xl,xr);
            // otherwise read write reproduction error will not be added
            // to the eror bilance, since it is not purely numerical in
            // nature.
      cpmmessage("Leaving writing path");
   }
   ostringstream ost2;
   ost2<<endl<<"  Test of "<<nameOf(xWrite)<<" errorsum= "<<res<<endl;
   Base::wrt(ost2);
   Base::errorSum=res;
   CPM_MZ
}

/////////// testing the complex-interface ////////////////////////////////
template <class T>
class Test_c: public TestBase<T>{ // testing complex classes
   typedef TestBase<T> Base;
public:
   Test_c(Z tvs=1 , Z output=0, Z write=0);
   // tvs: test value size ~ complexity of test
};


template <class T>
Test_c<T>::Test_c(Z tvs, Z output, Z write):Base()
// The arguments have the same meaning as in Test_r. Look there.
{
   static const Z multOrder=4;
      // order test is made on an array of
      // multOrder*tvs elements

   Z nOrd=multOrder*tvs;

   T  x, y, z, xl, xr, xt;

   ostringstream ost;
   ost<<endl<<" Test of class "<<x.nameOf()<<endl;
   Base::wrt(ost);

   Word top;
   xt=x.test(tvs);
   x=xt.ran(1);
   y=xt.ran(2);
   z=xt.ran(3);
   T xWrite=z;

   if (output){
      cpmcerr<<endl<<"test:"<<xt.toWord();
      cpmcerr<<endl<<"first ran:"<<x.toWord();
      cpmcerr<<endl<<"second ran:"<<y.toWord();
      cpmcerr<<endl<<"third ran:"<<z.toWord();
   }

   T v4=x;
   T v5;
   T v6=y;
   v6=v5=x;
   xl=v6;
   xr=v4;

   R res=0;
   top="copy constructor, assignement, and equality";
     res+=Base::comp(top,xl,xr);
   xl=x+y; xr=y+x; top="+ commutativity";
   res+=Base::comp(top,xl,xr);
   xl=x*y; xr=y*x; top="* commutativity";
   res+=Base::comp(top,xl,xr);
   xl=(x+y)+z; xr=x+(y+z); top="+ associativity";
   res+=Base::comp(top,xl,xr);
   xl=(x*y)*z; xr=x*(y*z); top="* associativity";
   res+=Base::comp(top,xl,xr);

   T xZero=net(x,0);

   xl=-x; xr=xZero-x; top="sign change";
   res+=Base::comp(top,xl,xr);
   xl=x*(y+z); xr=x*y+x*z; top="left-distributivity";
   res+=Base::comp(top,xl,xr);
   xl=(x+y)*z; xr=x*z+y*z; top="right-distributivity";
   res+=Base::comp(top,xl,xr);
   // not clear why con(x*y) is not understood for T=Fr<Z,R>
   T xTemp=x*y;
   xl=con(xTemp); xr=con(y)*con(x); top="conjugation commutativity";
   res+=Base::comp(top,xl,xr);

   x=z;
   T xu=x;

   xu=net(xu,1);
   T xi=inv(x);

   xl=x*xi; xr=xu; top="right inverse";
   res+=Base::comp(top,xl,xr);

   xl=xi*x; xr=xu; top="left inverse";
   res+=Base::comp(top,xl,xr);

// order properties
// rather simple test of ordering consistency

   T xMax=CpmRootX::sup(x,y,z);
   T xMin=CpmRootX::inf(x,y,z);

   if (xMin<=xMax){
      cpmcerr<<endl<<"comparison test OK"<<endl;
   }
   else{
      cpmcerr<<endl<<"comparison test failed"<<endl;
      res+=1;
   }

// let us order a list of random values
   Vo<T> list(nOrd);
   Z i;
   for (i=1;i<=nOrd;i++){
      list[i]=ran(xt,i);
   }
   list.order();
   for (i=2;i<=nOrd;i++){
      if (list[i-1]>list[i]){
         cpmmessage("comparison failure");
         res+=1;
      }
   }
   xl=z;
   R r,phi;
   xl.toPolar(r,phi);
   xr.polar(r,phi);
   top="polar back and forth";
   res+=Base::comp(top,xl,xr);

   xl=x; xr=x.sqrt()*x.sqrt();
   top="sqrt and square";
   res+=Base::comp(top,xl,xr);

   xl=exp(x+y); xr=exp(x)*exp(y);
   top="exp addition law";
   res+=Base::comp(top,xl,xr);

   xl=x; xr=exp(ln(x));
   top="exp(ln(z))=z";
   res+=Base::comp(top,xl,xr);

   xl=sin(x+y); xr=sin(x)*cos(y)+cos(x)*sin(y);
   top="sin addition law";
   res+=Base::comp(top,xl,xr);

   xl=y; xr=sin(arcsin(y));
   top="sin(arcsin(z)=z";
   res+=Base::comp(top,xl,xr);

   xl=z; xr=cot(arccot(z));
   top="cot(arccot(z)=z";
   res+=Base::comp(top,xl,xr);

   Z n=3;
   xl=pow(cosh(x)+sinh(x),n);
   xr=cosh(n*x)+sinh(n*x);
   top="Moivre for cosh and sinh";
   res+=Base::comp(top,xl,xr);
   if (write){
      cpmmessage("Writing path entered");
      Word filename(nameOf(xWrite));
      Word filename1="w-"+filename+".txt";
      OFileStream out1(filename1);
      if (!out1()) cpmerror(nameOf(xWrite),
         "test: output file cannot be opened");
      const int precWrite=12;
      out1()<<setprecision(precWrite);
      out1()<<endl<<"// I/O Test:"<<endl;
      out1()<<"// Object to be written: "<<endl;
      out1()<<xWrite;
      cpmmessage("First writing done");
      if (!out1()) cpmerror(nameOf(xWrite),
         "out1 not OK before closing");
      IFileStream in(filename1);
      if (!in()) cpmerror(nameOf(xWrite),
         "test: input file cannot be opened");
      xr.scanFrom(in());
      cpmmessage("First reading done");
      if (!in()) cpmerror(nameOf(xWrite),
         "in not OK after reading and before closing");
      Word filename2="r-"+filename+".txt";
      OFileStream out2(filename2);
      if (!out2()) cpmerror(nameOf(xWrite),
         "test: out2 file cannot be opened");
      out2()<<"// I/O Test. This is what we got back:"<<endl;
      out2()<<xr;
      cpmmessage("Second writing done");
      if (!out2()) cpmerror(CpmRoot::nameOf(xWrite),
         "out2 not OK before closing");
      xl=xWrite; top="file I/O";
      if (write>1) res+=Base::comp(top,xl,xr);
            // otherwise read write reproduction error will not be added
            // to the eror bilance, since it is not purely numerical in
            // nature.
      cpmmessage("Leaving writing path");
   }
   Base::errorSum=res;
   ostringstream ost2;
   ost2<<endl<<"  Test of "<<nameOf(xWrite)<<" errorsum= "<<res<<endl;
   Base::wrt(ost2);
}

// class for testing consistent polymorphic behavior
template <class T, class Td, class Container> // assumes that Td is
// derived from T
//
class TestOfPolymorphism{ // testing consistent polymorphic behavior
   R eps_;
   public:
   TestOfPolymorphism(T const& t, const Td& td)
   {
       using namespace CpmSystem;
       Word loc("TestOfPolymorphism(T const& t, const Td& td)");
       Z tv=16;
       Container ct;
       T tMem=t;
       Td tdMem=td;
       ct.set(t);
       T t1=ct();
       cpmassert(t1==tMem,loc);
       ct.set(td);
       Td td1=ct();
       eps_=td1.dis(tdMem);
       cpmmessage("TestOfPolymorphism() successfully done");
       cout<<endl<<"TestOfPolymorphism() done with eps="<<eps_<<endl;
    }
    R val()const{ return eps_;}
};

////////////////////////////// Test_Vp /////////////////////////////////
// testing polymorphicity and value interface of Vp<>

template <class T0, class T1, class T2, class T3>
class Test_Vp: public TestBase<T0>{// testing behavior of Vp<>

   typedef TestBase<T0> Base;

public:
   Test_Vp(Z tvs=1 , Z output=1);
};

template <class T0, class T1, class T2, class T3 >
Test_Vp<T0,T1,T2,T3>::Test_Vp(Z tvs , Z output):Base()
      // Tests all relations concerning polymorphic array Vp<> and
      // returns an error sum
{
   using namespace CpmRoot;
   using namespace CpmRootX;
   using namespace CpmArrays;

   Vp<T0> v0;

   ostringstream ost;
   ost<<endl<<" Test of class "<<v0.nameOf()<<endl;
   Base::wrt(ost);

   T3 t3;
   t3=t3.test(tvs);
   T3 t3_1=t3.ran(1);
   T3 t3_2=t3.ran(2);
   T3 t3_3=t3.ran(3);
   T3 t3_4=t3.ran(4);
   T3 t3_5=t3.ran(5);
   T3 t3_6=t3.ran(6);

   T0 t0_1=t3_1.toClnBase();
   T0 t0_2=t3_2.toClnBase();
   T0 t0_3=t3_3.toClnBase();
   T0 t0_4=t3_4.toClnBase();
   T0 t0_5=t3_5.toClnBase();
   T0 t0_6=t3_6.toClnBase();

   T1 t1_1=T1(t0_1);
   T1 t1_2=T1(t0_2);
   T1 t1_3=T1(t0_3);
   T1 t1_4=T1(t0_4);
   T1 t1_5=T1(t0_5);
   T1 t1_6=T1(t0_6);

   T2 t2_1=T2(t0_1);
   T2 t2_2=T2(t0_2);
   T2 t2_3=T2(t0_3);
   T2 t2_4=T2(t0_4);
   T2 t2_5=T2(t0_5);
   T2 t2_6=T2(t0_6);

   Z nvp=12;
   Vp<T0> vp(nvp);
   V<Word> wp(nvp);
   vp[1]=t0_1;
   vp[2]=t1_2;
   vp[3]=t2_3;
   vp[4]=t3_4;
   vp[5]=t0_5;
   vp[6]=t1_6;
   vp[7]=t2_1;
   vp[8]=t3_2;
   vp[9]=t0_3;
   vp[10]=t1_4;
   vp[11]=t2_5;
   vp[12]=t3_6;

   wp[1]=t0_1.nameOf();
   wp[2]=t1_2.nameOf();
   wp[3]=t2_3.nameOf();
   wp[4]=t3_4.nameOf();
   wp[5]=t0_5.nameOf();
   wp[6]=t1_6.nameOf();
   wp[7]=t2_1.nameOf();
   wp[8]=t3_2.nameOf();
   wp[9]=t0_3.nameOf();
   wp[10]=t1_4.nameOf();
   wp[11]=t2_5.nameOf();
   wp[12]=t3_6.nameOf();

// complicated process of creating an array Vp<T0> zp holding the same
// values
// as vp
   Vp<T0> xp; // default element
   xp=vp; // assignment to default element
   Vp<T0> yp(xp); // copy construction
   Vp<T0> zp(5001);
   zp.set(t0_6); // zp something different
   zp=yp; // assignment to something non-trivial

   R errSum=0;
   Z i;
   for (i=xp.b();i<=xp.e();++i){ // testing whether finally zp holds the
   // right objects. Here only the name gets tested
      Word wpi=zp(i).nameOf();
      if (output) cout<<endl<<"wpi="<<wpi<<" , wp[i]="<<wp[i];
      errSum+=wp[i].dis(wpi);
   }

   Base::errorSum=errSum;
   ostringstream ost2;
   ost2<<endl<<" Test of class "<<v0.nameOf()<<" done."
   <<endl<<" Errorsum= "<<errSum;
   Base::wrt(ost2);
}

/////////////////////////////// Test_F ///////////////////////////////////
// Testing the chaining operation & of F<>

template <class T1, class T2, class T3, class T4>
class Test_F: public TestBase<T1>{ // Testing the chaining operation of F
   typedef TestBase<T1> Base;
public:
   Test_F(Z tvs=1 );
   // tvs: test value size ~ complexity of test
};

template <class T1, class T2, class T3, class T4 >
Test_F<T1,T2,T3,T4>::Test_F(Z tvs):Base()
// tvs: test value size ~ complexity of test
{
   using namespace CpmRoot;
   using namespace CpmRootX;
   using namespace CpmFunctions;

   Fr<T1,T2> fa;
   Fr<T2,T3> fb;
   Fr<T3,T4> fc;

   Word mes(" Test of classes ");
   mes&=(fa.nameOf()&", "&fb.nameOf()&", and "&fc.nameOf());
   Base::wrt(mes);

   fa=fa.test(tvs);
   fb=fb.test(tvs);
   fc=fc.test(tvs);

   F<T1,T2> fa_1=fa.ran(1);
   F<T1,T2> fa_2=fa.ran(2);

   F<T2,T3> fb_1=fb.ran(3);
   F<T2,T3> fb_2=fb.ran(4);

   F<T3,T4> fc_1=fc.ran(5);
   F<T3,T4> fc_2=fc.ran(6);

   F<T1,T3> fab=fa_1&fb_2;
   F<T2,T4> fbc=fb_2&fc_1;

   Fr<T1,T4> fab_c=fab&fc_1;
   Fr<T1,T4> fa_bc=fa_1&fbc;

   Fr<T1,T4> fake=fa_2&fbc;

   R errSum=0;
   errSum+=fab_c.dis(fa_bc);
   R dis=fab_c.dis(fake); // should not be small
      // this detects programming bugs that let dis always return 0.
   R tiny=1e-6;
   if (dis<tiny) errSum+=1;

   Base::errorSum=errSum;
   mes&=(" done.\n Errorsum="&cpm(Base::errorSum));
   Base::wrt(mes);
}

/////////////////////////////// Test_set ///////////////////////////////

template <class S, class T> // S is the class of sets of type T;
// typically S=Sr<T>
class Test_set: public TestBase<S>{ // testing behavior of sets

   typedef TestBase<S> Base;

public:
   Test_set(Z tvs=1 , Z output=1, Z write=0);
};

template <class S, class T>
Test_set<S,T>::Test_set(Z tvs , Z output, Z write):Base()
   // Tests all relations concerning the value interface and returns
   // an error sum
{
   using namespace CpmRoot;
   using namespace CpmRootX;
   using namespace CpmArrays;

   const R dSmall=1e-4;
   const R dTiny=1e-12;

   T t;
   S s,s1,s2,s3,s4;
   S lhs, rhs;

   ostringstream ost;
   ost<<endl<<" Test of class "<<s.nameOf()<<endl;
   Base::wrt(ost);

   s=s.test(tvs); // test object, not performing a test.
   t=testT<T>(t,tvs);
   s1=s.ran(1);
   s2=s.ran(2);
   s3=s.ran(3);
   s4=s.ran(4);
   s1=s1|s4;
   s2=s2|s4;
   s3=s3|s4;
      // this achieves that not all sections become trivial
   S xWrite=s3;
   if (output){
      cpmcerr<<endl<<"test:";          s.prnOn(cpmcerr);
      cpmcerr<<endl<<"first ran:";  s1.prnOn(cpmcerr);
      cpmcerr<<endl<<"second ran:"; s2.prnOn(cpmcerr);
      cpmcerr<<endl<<"third ran:";  s3.prnOn(cpmcerr);
   }

   R res=0.;
   Word top;

   idem1(lhs,rhs,s1);
      // test template function defined in cpmsr.h
   top="idem1";
   res+=Base::comp(top,lhs,rhs);

   idem2(lhs,rhs,s2);
   top="idem2";
   res+=Base::comp(top,lhs,rhs);

   comm1(lhs,rhs,s1,s2);
   top="comm1";
   res+=Base::comp(top,lhs,rhs);

   comm2(lhs,rhs,s1,s2);
   top="comm2";
   res+=Base::comp(top,lhs,rhs);

   assoc1(lhs,rhs,s1,s2,s3);
   top="assoc1";
   res+=Base::comp(top,lhs,rhs);

   assoc2(lhs,rhs,s1,s2,s3);
   top="assoc2";
   res+=Base::comp(top,lhs,rhs);

   distr1(lhs,rhs,s1,s2,s3);
   top="distr1";
   res+=Base::comp(top,lhs,rhs);

   distr2(lhs,rhs,s1,s2,s3);
   top="distr2";
   res+=Base::comp(top,lhs,rhs);

   diffdistr1(lhs,rhs,s1,s2,s3);
   top="diffdistr1";
   res+=Base::comp(top,lhs,rhs);

   diffdistr2(lhs,rhs,s1,s2,s3);
   top="diffdistr2";
   res+=Base::comp(top,lhs,rhs);

   S xl,xr;

   if (write){
      cpmmessage("Writing path entered");
      Word filename(nameOf(xWrite));
      filename=filename.makeFileName();
      Word filename1="w-"+filename+".txt";
      OFileStream out1(filename1);
      if (!out1()) cpmerror(nameOf(xWrite),
         "test: output file cannot be opened");
      const int precWrite=12;
      out1()<<std::setprecision(precWrite);
      out1()<<endl<<"// I/O Test:"<<endl;
      out1()<<"// Object to be written: "<<endl;
      xWrite.prnOn(out1());
      cpmmessage("First writing done");
      if (!out1()) cpmerror(nameOf(xWrite),
         "out1 not OK before closing");
      IFileStream in(filename1);
      if (!in()) cpmerror(nameOf(xWrite),
         "test: input file cannot be opened");
      xr.scanFrom(in());
      cpmmessage("First reading done");
      if (!in()) cpmerror(nameOf(xWrite),
         "in not OK after reading and before closing");
      Word filename2="r-"+filename+".txt";
      OFileStream out2(filename2);
      if (!out2()) cpmerror(nameOf(xWrite),
         "test: out2 file cannot be opened");
      out2()<<"// I/O Test. This is what we got back:"<<endl;
      xr.prnOn(out2());
      cpmmessage("Second writing done");
      if (!out2()) cpmerror(nameOf(xWrite),"out2 not OK before closing");
      xl=xWrite; top="file I/O";
      res+=Base::comp(top,xl,xr);
      S xlr=static_cast<S>(xl|xr);
      S xc=xlr.condense();
      cout<<endl<<"condensed union";
      xc.prnOn(cout);
      cpmmessage("Leaving writing path");
   }
   Base::errorSum=res;
   ostringstream ost2;
   ost2<<endl<<" Test of class "<<s.nameOf()<<" done."
   <<endl<<" Errorsum= "<<res;
   Base::wrt(ost2);

}

///////////////////// class Test_logic<> /////////////////////////////////
// Test of classes modeling propositional logic
// So far implemented: cpmlogic1.h, classes PropLog<>, PropLogic

template <class T>
class Test_logic: public TestBase<T>{//test models of propositional logic

   typedef TestBase<T> Base;

public:
   Test_logic(Z tvs=1 , Z output=1, Z write=0);

};

template <class T>
Test_logic<T>::Test_logic(Z n, Z output, Z write):Base()
   // Tests all relevant mathematical relations and returns
   // an error sum
   // output 0 : only end of test and error sum displayed
   // output 1 : output of every test. The potentially
   //      time consuming output of the test object is suppressed
   // output 2: only the first component of the test object is
   // displayed
   // output 3: the test object is displayed in full detail
{
   Z i;
   if (n<3) n=3; // we need as many variables to make indicative examples

   V<T> p=T::list(n);
   T p1=p[1];
   for (i=2;i<=n;i++) p1=p1.et(p[i]);
   p1=p1.non();
   T p2=p[1].non();
   for (i=2;i<=n;i++) p2=p2.vel(p[i].non());
   bool b=p1.iff(p2).taut();
   cpmmessage("b="&B(b).toWord()&"; true is correct");

// example to the commutativity of the diagram shown in
// cpmlogic1.h for class PropLog<X>

   T q1=p[1].non().et(p[2]).vel(p[3]);
   T q2=p[1].vel(p[2]);
   B b1(q1.eval());
   B b2(q2.eval());

   T q12a=q1.et(q2);
   B b12a=b1.et(b2);
   B b12a_(q12a.eval());
   B resa(b12a==b12a_);
   cpmmessage("b12a==b12a_ ?"&resa.toWord()&"; true is correct");

   T q12o=q1.vel(q2);
   B b12o=b1.vel(b2);
   B b12o_(q12o.eval());
   B reso(b12o==b12o_);
   cpmmessage("b12o==b12o_ ?"&reso.toWord()&"; true is correct");

   T q1n=q1.non();
   B b1n=b1.non();
   B b1n_(q1n.eval());
   B resn(b1n==b1n_);
   cpmmessage("b1n==b1n_ ?"&resn.toWord()&"; true is correct");

/************** takes too long so far ***********
   Z nc=3;
   V<T> q=T::list(nc);
   T xorq=T::xorV(q);
   V<T> known(2);
   known[1]=xorq;
   known[2]=q[nc];
   V<T> aux=T::list(nc-1);
   for (i=1;i<nc;i++) aux[i]=q[i].non();
   T hyp=T::andV(aux);
   B resx(hyp.implied(known));
   cpmmessage("resx ? "&resx.toWord()&"; true is correct");
*/
}

} // namespace

#endif
