//? cpmviewport.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_VIEWPORT_H_
#define CPM_VIEWPORT_H_
/*

   class Viewport is the Cpm interface to the graphical
   capabilities of the hardware.

*/
#include <cpmv.h>

namespace CpmGraphics{

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::L;
   using CpmRoot::Word;
   using CpmArrays::V;
   using CpmArrays::IvZ;
   using namespace CpmStd;
      // see cpmnumbers.h, only I/O related stuff from std.
      // I had using namespace std before. This had the bad effect that
      // one of my function names ('ignore' in this case), even if defined
      // in an anonymous namespace, once clashed with std::ignore.
      // So, to avoid clashes with the ever expanding namespace std I will
      // never rely on std as a whole.

#if defined(WIN32)
   const Word Platform("Windows");
#else
   const Word Platform("Cygwin or Linux");
#endif

////////////////////////////////// struct xy ///////////////////////////
// Simple data structure for graphical points. As ever in graphics:
// x-axis horizontal from left to right (thus perfectly normal)
// y-axis downwards. Meaningful coordinates belong to {0,1,2,...}.
   struct xy{ // graphical points
      Z x,y;
      xy(Z a=0, Z b=0):x(a),y(b){}
      xy operator+(xy p)const{ return xy(x+p.x,y+p.y);}
      xy operator-(xy p)const{ return xy(x-p.x,y-p.y);}
      xy operator-()const{ return xy(-x,-y);}
      xy& operator+=(xy p){ x+=p.x;y+=p.y;return *this;}
      xy& operator-=(xy p){ x-=p.x;y-=p.y;return *this;}
      bool operator==(xy p)const{ return x==p.x && y==p.y;}
   };

// simple data structure for graphical rectangles

   struct xyxy{ // graphical rectangles
      Z x1,y1,x2,y2; // (x1,y1) left upper corner,
         // (x2,y2) right lower corner

      xyxy(Z a=0,Z b=0,Z c=0,Z d=0):x1(a),y1(b),x2(c),y2(d){}
         // construction from corner coordinates

      xyxy(xy p1, xy p2):x1(p1.x),y1(p1.y),x2(p2.x),y2(p2.y){}
         // construction from corners

      xyxy(xy p, Z w, Z h):x1(p.x),y1(p.y),x2(x1+w-1),y2(y1+h-1){}
         // construction from left upper corner, width, and height

      Z w()const{ return x2-x1+1;}

      Z h()const{ return y2-y1+1;}

      R aspRat()
      //: aspect ratio
      { Z h_=h(); cpmassert(h_>=1,"xyxy::aspRat()"); return R(w())/h_;}

      bool isVoid()const{ return x1<0;}
         // since the meaningfull graphical rectangles have no
         // negative components, this is an easy way to encode
         // exceeptional behavior

      xyxy operator+(xy p)const
         { return xyxy(x1+p.x,y1+p.y,x2+p.x,y2+p.y);}
      xyxy operator-(xy p)const
         { return xyxy(x1-p.x,y1-p.y,x2-p.x,y2-p.y);}
      xyxy& operator+=(xy p){ return *this=*this+p;}
      xyxy& operator-=(xy p){ return *this=*this-p;}

      xyxy operator|(xyxy r)const;
         // smallest xyxy that contains both *this and r
         // kind of union

      xyxy& operator|=(xyxy r){ return *this=(*this)|r;}
         // | as a mutating operation

      xyxy operator&(xyxy r)const;
         // returns the set section of *this and r.
         // The result res satisfies res.isVoid()==true
         // if this section is the void set.

      xyxy& operator&=(xyxy r){ return *this=(*this)&r;}
         // & as a mutating operation

      xy operator[](Z i)const{ return i<=1 ? xy(x1,y1) : xy(x2,y2);}
         // For an instance r of xyxy r[1] is the upper left corner
         // and r[2] is the lower right corner.

      xy cor(Z i)const
      //: corner
      // Returns the i-th corner. Here corners are numbered counter
      // clock-wise starting with the left upper corner as 1. All
      // non-matching arguments give corner 1. Thus no exceptions!
      {
         if (i==2) return xy(x1,y2);
         if (i==3) return xy(x2,y2);
         if (i==4) return xy(x2,y1);
         return xy(x1,y1);
      }

      bool hasElm(xy const& p)const
      //: has element
      // Makes a xyxy a set of xy's
      {
         return IvZ(x1,x2).hasElm(p.x)&& IvZ(y1,y2).hasElm(p.y);
      }

      xyxy scaleToFit(xyxy r)const;
      //: scale to fit
      // Returns the largest rectangle that
      // 1. has the aspect ratio of r
      // 2. has the same left-upper corner as *this
      // 3. is contained in *this.

      friend ostream& operator<<(ostream& os, xyxy const& r)
      {
         os<<" "<<r.x1<<", "<<r.y1<<", "<<r.x2<<", "<<r.y2<<endl;
         return os;
      }
   };

   struct rgb;
   class Img24;

///////////////////////// class ColRef /////////////////////////////////

const float i255=1./255;

   class ColRef{ //lean 24-bit color-values
      // light 24-bit representation of color-values
      // that easily converts (back and forth) to the low-level color
      // data

      friend struct rgb;
      friend class Img24;
      static Z x(Z z);
      typedef ColRef Type;
   public:
      static Z lowerLimit;
      static Z upperLimit;
      L r,g,b; // all values valid, so public OK
      CPM_IO
      CPM_ORDER
      ColRef():r((L)lowerLimit),g((L)lowerLimit),b((L)lowerLimit){}
      ColRef(Z r_, Z g_, Z b_):r((L)x(r_)),g((L)x(g_)),b((L)x(b_)){}
      ColRef(L r_, L g_, L b_):r(r_),g(g_),b(b_){}
      L getR()const{ return r;}
      L getG()const{ return g;}
      L getB()const{ return b;}
      L operator[](Z i)const
         { if (i==3) return b; else if (i==2) return g; else return r;}
      float glR()const{ return i255*r;}
      float glG()const{ return i255*g;}
      float glB()const{ return i255*b;}
      bool isGray()const{ return r==g && g==b;}
      ColRef add(ColRef const& cr)const
      { return ColRef((Z)r+(Z)cr.r,(Z)g+(Z)cr.g,(Z)b+(Z)cr.b);}
      Word nameOf()const{ return "ColRef";}
   };

////////////////////////// struct rgb //////////////////////////////

struct rgb { // Z-valued red green blue
   // a simple struct to provide a convenient interface for
   // functions draw. This class will be a base class for deriving
   // a more functional class Color in namespace CpmImaging.
   Z r,g,b;

   rgb():r(ColRef::lowerLimit),g(ColRef::lowerLimit),
      b(ColRef::lowerLimit){}

   rgb(Z r_, Z g_, Z b_, bool norm=true):r(r_),g(g_),b(b_)
      {if (norm) normalize();}

   rgb(R r_, R g_, R b_, bool norm=true):
   r(cpmrnd(r_)),g(cpmrnd(g_)),b(cpmrnd(b_))
   {if (norm) normalize();}

   explicit rgb(ColRef cr):r(cr.r),g(cr.g),b(cr.b){}

   virtual Word nameOf()const{ return Word("rgb");}

   operator ColRef()const{ return ColRef(r,g,b);}

   rgb operator~()const
      // returning the complementary color
   {
      Z up=ColRef::upperLimit; return rgb(up-r,up-g,up-b);
   }

   void normalize(){ r=ColRef::x(r); g=ColRef::x(g); b=ColRef::x(b);}
      // ensuring the right value range by simply cutting

   static rgb mix(rgb c1, rgb c2, R w1, R w2, R gamma);
      // mixing colors according to function mixingRule

   R gray()const{return (r+g+b)/3.;}

   bool isGray()const{ return r==g && g==b;}
      //: is gray

   bool isBlack()const{ return r==0 && g==0 && b==0;}
      //:: is black

   bool isWhite()const{ return r==255 && g==255 && b==255;}
      //:: is white

   rgb operator*(R fac)const
   { return rgb(cpmrnd(fac*r),cpmrnd(fac*g),cpmrnd(fac*g));}

   static Z mixingRule(Z z1, Z z2, R w1, R w2, R gamma);
      // defines mixing of colors according to given weights
      // w1+w2=1 is not assumed
      // ( w1=w2=1 corresponds to addition of light)

   void write(ostream& str)const
   { str<<endl<<"rgb: r="<<r<<" g="<<g<<" b="<<b<<endl;}

   friend void operator<<(ostream& str, rgb const& c)
   { str<<c.r; str<<c.g; str<<c.b;}

   friend void operator>>(istream& str, rgb & c)
   { str>>c.r; str>>c.g; str>>c.b;}
};

////////////////////////////// class Font ///////////////////////////////
// Presently only Helvetica 12 needed from this font system

class Font{
// only Helvetica 12 needed from this font system of GLUT
   void* name;
   Z h;
      // height in pixels
   Z mw;
      // maximum width in pixels (estimation based on height)
   Z d;
      // suitable line separation taking decent ('Unterlaenge') into
      // account
public:
   explicit Font(Z type=4);
   explicit Font(R height);
      // creates an instance of a font height that comes closest
      // to the value given by the argument. See the simple
      // implementation
      // code for the details

   Z getHeight()const{ return h;}
   Z getWidth()const{ return mw;}
      // maximum width in pixels (typically width of W in proportional
      // fonts
   Z getLineSep()const{ return d;}
   void* getName()const{ return name;}
};

///////////////////////// class Rec ////////////////////////////////////
// Pixel rectangle rec_ which always fits Viewport::win()
// and linear image buffer mem_ of arbitrary size.
// The display function vis of this class shows mem_
// always whithin rec_. This function is Cpm's means
// to bring pixel-based graphical information to screen.
// Present implementation relies on OpenGL (only if
// CPM_NOGRAPHICS is not set). If CPM_NOGRAPHICS is
// defined we need no graphics libraries from the system
// and can generate all graphics as PPM-images (on an
// pixel rectangle that can be set arbitrarily
// by editing cpmconfig.ini.

class Rec{ // pixel rectangle which always fits Viewport::win()
   xyxy rec_;
   Z n_;
      // dimension of mem_
   L* mem_;
      // should be unsigned char = L
      // Data element of this type needed in function vis()
      // by OpenGL. This is linearized data in two-fold manner:
      // The rectangular image area is un-fold to a linear chain,
      // and the r g b pixel data are represented as a list of
      // characters.

public:
   explicit Rec(xyxy r=xyxy(0,0,0,0));
      // Constructor which fills the image data from the one and only
      // initialized Viewport, more precisely from the data
      // within the set section of r and the Viewport rectangle
      // Viewport::win()
      // Present implementation relies on OpenGL
   explicit Rec(V< V<ColRef> > const& bh);
      // constructs a Rec from the matrix data bh. The corresponding
      // rectangle rec_ is created such that it fits Viewport::win().
   Rec(xyxy r, L c);
      // The pixels within the rectangle r are set equal to c
      // On display this is always a gray (black to white) frame.
      // Notice that colorful colors would be more difficult to
      // handle here.
      // Rectangle is created such that it fits Viewport::win().
   Rec(Rec const& r);
   ~Rec();
   Rec& operator=(Rec const& r);
   Z dim()const{ return n_;}
      // number of bytes = 3*number of pixels
   Z h()const{ return rec_.h();}
   Z w()const{ return rec_.w();}
   Z x1()const{ return rec_.x1;}
   Z y1()const{ return rec_.y1;}
   Z x2()const{ return rec_.x2;}
   Z y2()const{ return rec_.y2;}
   xyxy rec()const{ return rec_;}
   xy operator[](Z i)const{ return rec_[i];}
   xy cor()const{ return rec_[1];}
      //: corner
      // actually left upper corner
   Rec& operator+=(xy shift);
      // result is always within Viewport::win()
   Rec operator+(xy shift)const{ Rec res=*this; res+=shift; return res;}
   Rec& operator-=(xy shift);
      // result is always within Viewport::win()
   Rec operator-(xy shift)const{ Rec res=*this; res-=shift; return res;}
   void vis()const;
      //: visualize
      // Displays the image data in mem_ on the area rec_ (which always
      // is a subset of the viewpoint window).
      // On this function all bitmap imaging in C+- is based.
      // Present implementation relies on OpenGL. Function glDrawPixels
      // needs a L* as the last argument. So it would not be convenient to
      // replace the mem_ data element of the present class by one of type
      // V<L>.
   V< V<ColRef> > fold()const;
      // returns a matrix-like memory version of mem_
};

///////////////////////// class Viewport ////////////////////////////////
/*
class Viewport is a lean interface to the
systems graphical capabilities. It also defines and activates a
statusbar supporting a continuous message stream from a running
program going to a dedicated area of the screen.

In order for the OpenGL version to work (which is the only one that
survived, I had working implementations based on MSFC and on FLTK)
one has to do some initialisation in function main() as is done in
cpmapplication.cpp.

The class as it stands is a rather 'blind viewport'. Apart from filling
rectangles with chosen color and setting text in GLUT's native fonts
it allows  n o t to perform actions which can be seen on screen.
The viewport becomes more active only with class Img24 (see cpmimg24.h)
which provides a data element of type Rec (its name is displayBuffer) and
tools to fill and display it. Each call of an constructor of class
CpmGraphics::Frame activates Img24.
See in particular CpmGraphics::FramesetStaticSize(Z width, Z height)
which is called in int main(int argc, char *argv[]) when this function
is defined in cpmapplication.cpp
*/

class Viewport{ // Lean interface to the system's graphical capabilities.
   // The Viewport class has no non-static data member, so
   // every member function could be declared static without changing
   // their behavior. However, it is more convenient to say
   //
   // Viewport vp;
   // vp.addText(10,10,"hello world");
   //
   // than
   //
   // Viewport::addText(10,10,"hello world");
   //
   // Notice that there is no way to have more than one different
   // instances of Viewport. So we can control a single graphical area,
   // which in my normal applications is one 'application window'.
   // Classes Frame, Graph, and Frames organize as many rectangular
   // subframes of this basic window as we want.
   static bool initialized;
   static Z applWindowWidth;
   static Z applWindowHeight;
   static Z fullWindowHeight;
   static xyxy window;
   static xyxy fullWindow;
   static R gamma;
      // a gamma-value for the systems video screen
      // presently set to 2.2 (usual in Windows systems)
      // in code. Would not be illogical to read this in from
      // cpmsystemdependencies.h.
      // This value of gamma will be used as default for the
      // image class Img24.
   static ColRef colText;
   static xyxy seg1;
   static xyxy seg2;
   static xyxy seg3;
   static xyxy seg4;
   static ColRef cs1;
   static ColRef cs2;
   static ColRef cs3;
   static ColRef cs4;
   static Z heightStatusBar;
   static Z size;
   static Z rank;

public:
   static void setColText(Z r, Z g, Z b){ colText=ColRef(r,g,b);}
   static void setColText(ColRef cr){ colText=cr;}
   static void setColText(rgb crgb){ colText=(ColRef)crgb;}
   static ColRef getColText(){ return colText;}

   static xyxy win(){ return window;}
      // Window available for regular display activities (that do
      // avoid writing to the status bar).
   static xyxy fullWin(){ return fullWindow;}
      // Graphics window including the status bar.

   static Z getHeightStatusBar(){ return heightStatusBar;}
   static void onStatusBar(const Word& w, Z i);
   static void clrPan();
      //: clear panes
   static void fill(xyxy rec, ColRef cr);
      // Fills rectangle rec with solid color cr
   static void placeWord(Z tx, Z ty, Word const& w,
      const Font& font, Z wMax=500);
      // wMax sets the maximum number of pixels for the length of the
      // printed picture of w. If w is longer it gets cut ( e.g. for
      // disciplined writing on status bars.
   static xyxy textLine(Z tx, Z ty, Word const& w,
      ColRef crText, R h);
      // writing w in color crText.
      // Only the pixels making up the characters are written so that
      // previous screen content is only minimally shadowed by the text.
      // The text may not be easy to read, however.
      // h is a proposal for the font hight (the actually chosen
      // hight depends on the available fonts). Introduced
      // 2005-04-14
      // The return value is a enclosing box for the text - a bit larger
      // than the minimum enclosing box.

   Viewport(Z w, Z h, Word const& title);
   Viewport(){}

   static R getGamma(){ return gamma;}
   static void setGamma(R g){ gamma=g;}
   static bool isInitialized(){return initialized!=0;}
      // does not allow to change status !!
   static Z pelX(){ return applWindowWidth;}
   static Z pelY(){ return applWindowHeight;}

   Z getCol()const{return applWindowWidth;}
      // number of columns
      // horizontal extension of the graphically accessible
      // application window in pixels. The range of the first pixel
      // access index (mostely named x or i) is {0,1,...,getCol()-1}
   Z getWidth()const{return applWindowWidth;}
      // legalizing another conventional name

   Z getLin()const{ return applWindowHeight;}
      // number of lines (rows)
      // vertical extension of the graphically accessible
      // application window in pixels. The range of the second pixel
      // access index (mostely named y or j) is {0,1,...,dim2()-1}
   Z getHeight()const{ return applWindowHeight;}
      // legalizing another conventional name

   Z getFullHeight()const{return fullWindowHeight;}

   Z dim1()const{return applWindowWidth;}
   Z dim2()const{return fullWindowHeight;}
      // The screen area under control of Viewport consists of
      // dim1()*dim2() writable and readable pixels. In the case
      // of OpenGL-implementation a part of the lines are used as
      // forming a statusbar. So these should not be used for
      // displaying images and graphics (but they perfectly
      // can from a syntactic point of view).

   Z getRank()const{ return rank;}
   Z getSize()const{ return size;}

   xyxy addText(Z x1, Z y1, Word const& text, R fontHeight=12, Z font=1);
      // (x1,y1) is the upper left corner of the writing box.
      // The return value is a box which encloses the text and is a bit
      // larger than minimal.
      // Thus the last argument remains passive since we select the font
      //  the height of which comes closest to fontHeight.

      // Here we are the names of the available fonts together with a
      // number which does not matter here.
      // font = 1 : GLUT_BITMAP_TIMES_ROMAN_10
      // font = 2 : GLUT_BITMAP_HELVETICA_10
      // font = 3 : GLUT_BITMAP_HELVETICA_12
      // font = 4 : GLUT_BITMAP_8_BY_13
      // font = 5 : GLUT_BITMAP_9_BY_15
      // font = 6 : GLUT_BITMAP_HELVETICA_18
      // font = 7 : GLUT_BITMAP_TIMES_ROMAN_24
};

}// namespace

#endif
