//? cpmc.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmc.h>
#include <cpmtypes.h>

using namespace CpmRoot;
using namespace CpmSystem;

C CpmRoot::C::I=C(0.,1.);
C CpmRoot::C::Ibar=C(0.,-1.);
C CpmRoot::C::one=C(1.,0.);

Word C::toWord()const
      // for a nice printable representation
{
   Word sep_i=" "; // initial
   Word sep_f=" "; // final separator
   Word res;
   if (im==0.) res=cpmwrite(re);
   else if (re==0.){
      if (im>0.) res="i*"&cpmwrite(im);
      else res="-i*"&cpmwrite(-im);
   }
   else if (im>0.){
      res=cpmwrite(re)&"+i*"&cpmwrite(im);
   }
   else{
      res=cpmwrite(re)&"-i*"&cpmwrite(-im);
   }
   return sep_i&res&sep_f;
}

bool C::prnOn(ostream& str )const
{
   //if (!CpmRoot::writeTitle("C",str)) return false;
   cpmwat;
   cpmp(re);
   cpmp(im);
   return true;
}

bool C::scanFrom(istream& str )
{
   cpms(re);
   cpms(im);
   return true;
}

// inversions

C C::inv(void)const
// same logic as inv(R) for error handling
// see Press et al. p. 177 (5.4.5) for the algorithm
{
   static const Z maxMes=100;
   static Z mes=1;
   R c_=(re>=0. ? re : -re);
   R d_=(im>=0. ? im : -im);
   if (c_>=d_){
      if (c_==0) goto DIVBYZERO;
      R y=im/re; R z=R(1.)/(re+im*y); return C(z,-y*z);
   }
   else{
      if (d_==0) goto DIVBYZERO;
      R y=re/im; R z=R(1.)/(re*y+im); return C(y*z,-z);
   }

DIVBYZERO:

      if (mes==maxMes){
         mes++; // needed !
         Message::message(
            "C::inv(): argument is (0,0) ... messages discontinued");
      }
      if (mes<maxMes){
         mes++;
         Message::warning(
            "C::inv(): argument is (0,0);  (0,0) returned");
      }
      return C(0,0);
}

C& C::operator /=(R const& s)
{
   R si=CpmRoot::inv(s);
   re*=si;
   im*=si;
   return *this;
}

C CpmRoot::operator /(C const& z, R const& s)
{
   R si=cpminv(s);
   return z*si;
}

// Functions related to polar representation

R CpmRoot::arg(C const& x)
{
   return x.arg();
}

void C::pol_(R r, R phi)
{
   re=r*cpmcos(phi);
   im=r*cpmsin(phi);
}

void C::toPolar(R& r, R& phi)const
{
   r=abs();
   phi=arg();
}

// Square

C C::sqr(void)const
{
   C res(*this);
   return res*res;
}

// Square root

C C::sqrt(void)const
{
   R r=abs(), phi=arg();
   r=cpmsqrt(r);
   phi*=0.5;
   return C(r*cpmcos(phi),r*cpmsin(phi));
}

// Elementary transcendental functions. We need the following real
// functions to be provided by the compiler:

//      exp, log, sin, cos, sqrt, atan. If complex numbers Ca are to be
//      defined based on a class Ra of numbers of arbitrary precission,
//      we have to implement these functions also for Ra.

C C::exp()const
{
   R r=cpmexp(re);
   return C(r*cpmcos(im),r*cpmsin(im));
}

C C::ln()const
{
   R r=abs(), phi=arg();
   r=cpmlog(r);  // library function
   return C(r,phi);
}

C CpmRoot::log10(C const& z)
// logarithm to basis 10
{
   static const R log10_=1./cpmlog(10.);
   return log10_*ln(z);
}

C CpmRoot::pow(C const& z, const Z& n)
{
   C res(1.,0);
   if (n==0) return res;

   C z_; Z n_;
   if (n>0){
      z_=z;
      n_=n;
   }
   else{
      z_=!z;
      n_=-n;
    }
   for (Z k=1; k<=n_; k++) res*=z_;
   return res;
}

C CpmRoot::pow(C const& z, C const& w)
{
   return exp(w*ln(z));
}

C CpmRoot::sin(C const& z)
{
   return (exp(z.timesI())-exp(z.dividedByI())).dividedBy2I();
}

C CpmRoot::cos(C const& z)
{
   return (exp(z.timesI())+exp(z.dividedByI()))*0.5;
}

C CpmRoot::tan(C const& z){ return sin(z)/cos(z);}

C CpmRoot::cot(C const& z){ return cos(z)/sin(z);}

C CpmRoot::sinh(C const& z){ return 0.5*(exp(z)-exp(-z));}

C CpmRoot::cosh(C const& z){ return 0.5*(exp(z)+exp(-z));}

C CpmRoot::tanh(C const& z){ return sinh(z)/cosh(z);}

C CpmRoot::coth(C const& z){ return cosh(z)/sinh(z);}

C CpmRoot::arcsinh(C const& z)
   // Gradshteyn Ryzhik 1.622 5.
{ return ln(z+(z*z+1).sqrt());}

C CpmRoot::arccosh(C const& z)
   // Gradshteyn Ryzhik 1.622 6.
{ return ln(z+sqrt(z*z-1));}

C CpmRoot::arctanh(C const& z)
   // Gradshteyn Ryzhik 1.622 7.
{ return ln((z+C(1))/(C(1)-z))*0.5;}

C CpmRoot::arccoth(C const& z)
   // Gradshteyn Ryzhik 1.622 8.
{ return ln((z+C(1))/(z-C(1)))*0.5;}

C CpmRoot::arcsin(C const& z)
   // Gradshteyn Ryzhik 1.622 1.
{ return (arcsinh(z.timesI())).dividedByI();}

C CpmRoot::arccos(C const& z)
   // Gradshteyn Ryzhik 1.622 2.
{ return (arccosh(z)).dividedByI();}

C CpmRoot::arctan(C const& z)
   // Gradshteyn Ryzhik 1.622 3.
{ return (arctanh(z.timesI())).dividedByI();}

C CpmRoot::arccot(C const& z)
   // Gradshteyn Ryzhik 1.622 4.
{ return (arccoth(z.timesI())).timesI();}

C C::pow(C const& p)const{ return CpmRoot::pow(*this,p);}
C C::sin()const{ return CpmRoot::sin(*this);}
C C::cos()const{ return CpmRoot::cos(*this);}
C C::tan()const{ return CpmRoot::tan(*this);}
C C::cot()const{ return CpmRoot::cot(*this);}
C C::sinh()const{ return CpmRoot::sinh(*this);}
C C::cosh()const{ return CpmRoot::cosh(*this);}
C C::tanh()const{ return CpmRoot::tanh(*this);}
C C::coth()const{ return CpmRoot::coth(*this);}
C C::arcsin()const{ return CpmRoot::arcsin(*this);}
C C::arccos()const{ return CpmRoot::arccos(*this);}
C C::arctan()const{ return CpmRoot::arctan(*this);}
C C::arccot()const{ return CpmRoot::arccot(*this);}
C C::arcsinh()const{ return CpmRoot::arcsinh(*this);}
C C::arccosh()const{ return CpmRoot::arccosh(*this);}
C C::arctanh()const{ return CpmRoot::arctanh(*this);}
C C::arccoth()const{ return CpmRoot::arccoth(*this);}
   // these added 2005-08-23

// Indicators

/*************** utility functions ******************************/

C C::test(Z cpl)const
{
   const R red=0.5;
      // complex functions like exp() should be included in tests and
      // don't behave uncritical for large arguments
   R a=0.;
   R ar=CpmRoot::test(a,cpl);
   R ai=ar*red;
   return C(ar,ai);
}

C C::ran(Z j)const
{
   R a,b;
   if (j==0){
      a=CpmRoot::ran(re,0);
      b=CpmRoot::ran(im,0);
   }
   else{
      Z j2=j*2;
      a=CpmRoot::ran(re,j2);
      b=CpmRoot::ran(im,j2+1);
   }
   return C(a,b);
}

R C::dis(C const& z)const
{
   return CpmRoot::disDefFun(abs(),z.abs(),(*this-z).abs());
}

Z C::com(C const& s)const
{
   if (re<s.re) return 1;
   if (re>s.re) return -1;
   if (im<s.im) return 1;
   if (im>s.im) return -1;
   return 0;
}

Z C::hash()const
{
   Z ix=CpmRoot::hash(re);
   Z iy=CpmRoot::hash(im);

   return ix^iy;
}

C C::scaleTo(R const& r)const
      // returns a complex number of absolute value |r| and the direction
      // of *this for positive r and antidirection for negative r.
{
   R rOrig=abs();
   if (rOrig==0.) return C(r,0);
   C res=*this;
   return res*(r/rOrig);
}
