//? cpmclinalg.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_CLINALG_H_
#define CPM_CLINALG_H_
/*
   Description: classes C1, C2, C3, C4, CC2, CC3, CC4
   Basis for effective geometry in 1,2,3,4-dimensional spaces.

*/
#include <cpmlinalg.h>
#include <cpmperm.h>

namespace CpmLinAlg{

   using namespace CpmStd;

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::C;
   using CpmRoot::Word;
   using CpmArrays::V;
   using CpmArrays::T2;
   using CpmArrays::X2;
   using CpmArrays::Perm;

///////////////////////////// class C2 /////////////////////////////////
class C4;
class CC2;

class C2{  // tuples of two complex elements
   // In physics this is also the state space of the spin degree of
   // freedom for spin 1/2. Some of the functions to be defined here deal
   // with this situation.
public:
   C x1;
   C x2;
   void ini_(R theta, R phi){
      R ph=phi*0.5, th=theta*0.5;
      x1=C::expi(-ph)*cpmcos(th);
      x2=C::expi(ph)*cpmsin(th);
   }
public:
   friend class C4;
   friend class CC2;
   typedef C2 Type;

// constructors
   C2():x1(0.),x2(0.){}
   explicit C2(C const& z):x1(z),x2(z){}
   C2(C const& a, C const& b):x1(a),x2(b){}
   explicit C2(V<C> const&);
      // Construction from dimension-independent lists.
      // No restiction for the argument.
   explicit C2(R theta, R phi=0){ini_(theta,phi);}
      // construction from the usual polar coordinates on the Bloch sphere
      // Gives always a unit vector. phi does not change the proportion
      // |x1| : |x2|. theta==0 => x2=0, ('north pole')
      // theta==pi => x1=0 ('south pole')
      // see function blcAng for getting the angles from a C2 object.

   C2(R theta, R phi, R chi){
      ini_(theta,phi);
      C omega=C::expi(chi);
      x1*=omega;
      x2*=omega;
   }

// access
   C& operator[](Z i){ if (i>=2) return x2; else return x1;}
   C const& operator[](Z i)const{ if (i>=2) return x2; else return x1;}
   void set(C t){ x1=t; x2=t;}
// mutating arithmetics (linear algebra)
   C2& operator +=(C2 const& a)
   { x1+=a.x1; x2+=a.x2; return *this;}
   C2& operator -=(C2 const& a)
   { x1-=a.x1; x2-=a.x2; return *this;}
   C2& operator *=(C const& r)
   { x1*=r; x2*=r; return *this;}
   C2& operator *=(R const& r)
   { x1*=r; x2*=r; return *this;}
   C2& neg(void)
   { x1=-x1; x2=-x2; return *this;  }
   C operator |(C2 const& a)const{ return ~x1*a.x1+~x2*a.x2;}
   C trc(C2 const& a)const{ return x1*a.x1+x2*a.x2;}
   R abs(bool careful=true)const;
 //R abs(bool careful=false)const;
   R absSqr()const{ return x1.absSqr()+x2.absSqr();}
   R abs2()const{ return x1.absSqr()+x2.absSqr();}
   R dis(C2 const&)const;
   bool isVal()const{ return x1.isVal()&&x2.isVal();}
   C2 ran(Z j=0)const{
      if (j==0) return C2(x1.ran(),x2.ran());
      else{
         Z j2=j*2; return C2(x1.ran(j2),x2.ran(j2+1));
      }
   }
   C2 con()const{ return C2(x1.con(),x2.con());}
   C2 swap()const{ return C2(x2,x1);}
      //: swap
   C2& swap_(){ C temp=x1; x1=x2; x2=temp; return *this;}
      //: swap
      // obviously we have for the CC2 matrix sig3 and any C2 instance x:
      // sig3*x == R(r)*x ==> sig*x.swap() == R(-r)*x.swap() i.e. swap()
      // transforms a spin up state into a spin down state.
   C2 operator-()const{ return C2(-x1,-x2);}
   X2<R,C2 > polDec()const;
      //: polar decomposition
      // First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // a unit vector n such that *this = n*r. If *this is the zero
      // vector, n is arbitrarily chosen as (1,0)
   R normalize()
      { X2<R,C2> p=polDec(); *this=p.c2();  return p.c1();}
      // turns *this into a unit vector and returns the norm of the
      // original in order not to loose information
   R nor_(){ return normalize();}
      //: normalize

   R nor1_(){ R a=x1.arg(); C w=C::expi(-a); operator*=(w); return a;}
      //: normalize 1 (i.e. component 1)
      // normalizing means making real in this case

   V<C> toVC()const{ return V<C>{x1,x2};}

   R2 blcAng()const;
      //: bloch angles R2(theta,phi)

   C blc()const{ R2 ba=blcAng();
      return C::expi(ba[2])*(cpmpi-ba[1])/cpmpi;}
      // Bloch angles coded as a single C.
      // Useful for conversion to display color
      // Then theta modulates luminance: theta=0 (north pole) bright
      // theta=Pi (south pole) dark. Phi determines hue along
      // the color circle.

   R3 toR3()const
      //: to R3
      // returns (y1,y2,y3) with 0<=yi<=1;
      // Let C2n := { c \in C2 | c.abs()==1}
      // and R3n := { r \in R3 | r[1],r[2],r[3] \in Iv(0,1)}
      // then c |--> c.toR3() defines a surjective and continuous
      // mapping A: C2n --> R3n. As a surjective mapping it has a
      // right inverse.
      // The main reason for considering this function (and the next)
      // is that R3 can be naturally converted to display color.
      // See class CpmImaging::Color.
   {
      Word loc("C2::toR3()");
      R r1,r2,phi1,phi2;
      x1.pol_1_2(r1,phi1);
      x2.pol_1_2(r2,phi2);
      R phi3=C(r1,r2).arg();
      R y1=(cpmsin(phi1)+1.)*0.5;
      R y2=(cpmsin(phi2)+1.)*0.5;
      R y3=cpmsin(phi3);
      return R3(y1,y2,y3);
   }

   C2(R3 const& y)
      // constructor from y \in R3n (see previous function).
      // This function B: R3n --> C2n, y |--> C2n(y) is the
      // right inverse of A (as defined above) i.e.
      // A(B(y)==y
   {
      Word loc("C2(R3)");
      R r2=y[3];
      R r1=cpmsqrt(-r2*r2+1);
      R phi1=cpmasin(y[1]*2-1);
      R phi2=cpmasin(y[2]*2-1);
      x1=C::pol(r1,phi1);
      x2=C::pol(r2,phi2);
   }

   C2 operator*(R r)const{ return C2(x1*r,x2*r);}

   C sel(bool second)const{ return second ? x2 : x1;}

   static R blcTest(Z n=100);
      //: Bloch test
      // tests with random input whether
      // for any R2 r (sic!), we have z ~ z' for
      // z:=C2(r), r':= z.blcAng, z':=C2(r')
      // '~' means equality up to a common phase.
      // The accumulated error (which should be numerical
      // noise) is returned. Worked with result ~ 1e-13
      // for n=1000 (2007-12-05)

   static R codTest(Z n=1000);
      //: coding test
      // tests whether going from R3 to C2 and back to R3 gives
      // the identity (see C2(R3 const& y) for the context).

   CPM_IO
   CPM_WORD
   CPM_ORDER
   CPM_NAM(C2)
   Z dim()const{ return 2;}
      //: dimension
};

inline C2 operator +(C2 const& x1, C2 const& x2)
{ C2 res=x1; return res+=x2;}

inline C2 operator -(C2 const& x1, C2 const& x2)
{ C2 res=x1; return res-=x2;}

inline C2 operator *(C2 const& x1, C const& s)
{ C2 res=x1; return res*=s;}

inline C2 operator *(C const& s, C2 const& x1)
{ C2 res=x1; return res*=s;}

///////////////////////////// class C3 /////////////////////////////////

class C3{ // tuples of 3 complex elements
public:
   typedef C3 Type;
   C x1;
   C x2;
   C x3;
public:
// constructors
   C3():x1(0.),x2(0.),x3(0.){}
   explicit C3(C const& z):x1(z),x2(z),x3(z){}
   C3(C const& a, C const& b, C const& c):x1(a),x2(b),x3(c){}
   C3(const V<C>&); // construction from dimension-independent lists
      // no restiction to the argument
   C3(R3 const& re, R3 const& im):x1(C(re.x1,im.x1)), x2(C(re.x2,im.x2)),
      x3(C(re.x3,im.x3)){}
      // construction from real part and imaginary part
// access
   C& operator[](Z i)
   { if (i>=3) return x3; else if (i==2) return x2; else return x1;}
   C const& operator[](Z i)const
   { if (i>=3) return x3; else if (i==2) return x2; else return x1;}
   void set(C t){ x1=t; x2=t; x3=t;}
// mutating arithmetics (linear algebra)
   C3& operator +=(C3 const& a)
   { x1+=a.x1; x2+=a.x2; x3+=a.x3; return *this;}
   C3& operator -=(C3 const& a)
   { x1-=a.x1; x2-=a.x2; x3-=a.x3; return *this;}
   C3& operator *=(C const& r)
   { x1*=r; x2*=r; x3*=r; return *this;}
   C3& operator *=(const R& r)
   { x1*=r; x2*=r; x3*=r; return *this;}
   C3& neg(void)
   { x1=-x1; x2=-x2; x3=-x3; return *this;  }
   C3 operator-()const{ return C3(-x1,-x2,-x3);}
   C3 con()const{ return C3(~x1,~x2,~x3);}
   C trc(C3 const& a)const{ return x1*a.x1+x2*a.x2+x3*a.x3;}
   C operator |(C3 const& a)const
   { return ~x1*a.x1+~x2*a.x2+~x3*a.x3;}
   R abs(bool careful=true)const;
   R dis(C3 const&)const;
   X2<R,C3 > polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // a unit vector n such that *this = n*r. If *this is the zero
      // vector, n is arbitrarily chosen as (1,0,0)
   R normalize()
   { X2<R,C3> p=polDec(); *this=p.c2();  return p.c1(); }
      // turns *this into a unit vector and returns the norm of the
      // original in order not to loose information
   R nor_(){ return normalize();}
   C3 ran(Z j=0)const{
      if (j==0) return C3(x1.ran(),x2.ran(),x3.ran());
      else{
         Z j3=j*3;
         return C3(x1.ran(j3),x2.ran(j3+1),x3.ran(j3+2));
      }
   }
   V<C> toVC()const{ return V<C>{x1,x2,x3};}
   CPM_IO
   CPM_WORD
   CPM_ORDER
   CPM_NAM(C3)
   Z dim()const{ return 3;}
};

inline C3 operator +(C3 const& x1, C3 const& x2)
{ C3 res=x1; return res+=x2;}

inline C3 operator -(C3 const& x1, C3 const& x2)
{ C3 res=x1; return res-=x2;}

inline C3 operator *(C3 const& x1, C const& s)
{ C3 res=x1; return res*=s;}

inline C3 operator *(C const& s, C3 const& x1)
{ C3 res=x1; return res*=s;}

///////////////////////////// class C4 /////////////////////////////////

class C4{  //  tuples of 4 complex elements
public:
   typedef C4 Type;
   C x1;
   C x2;
   C x3;
   C x4;
public:
// constructors
   C4():x1(0.),x2(0.),x3(0.),x4(0.){}
   explicit C4(C const& z):x1(z),x2(z),x3(z),x4(z){}
   C4(C const& a, C const& b, C const& c, C const& d):x1(a),x2(b),
      x3(c),x4(d){}
   explicit C4(const V<C>&); // construction from dimension-independent
   // lists
      // no restiction to the argument
   C4(C2 const& a, C2 const& b):x1(a.x1*b.x1),x2(a.x1*b.x2),x3(a.x2*b.x1),
      x4(a.x2*b.x2){}
      // Here we consider C4 as the tensor product C2 times C2 (not the
      // direct sum.
   C4(Word const& idle, C2 const& a, C2 const& b):x1(a.x1),x2(a.x2),
      x3(b.x1),x4(b.x2){}
      // Here we consider C4 as the direct sum of  C2 and C2
      // The dummy argument is to discern this (less important)
      // construct from the more canonical tensor product
// access
   C& operator[](Z i)
   { if (i>=4) return x4; else if (i==3) return x3;
      else if (i==2) return x2; else return x1;}
   C const& operator[](Z i)const
   { if (i>=4) return x4; else if (i==3) return x3;
      else if (i==2) return x2; else return x1;}
   void set(C t){ x1=t; x2=t; x3=t; x4=t;}
// mutating arithmetics (linear algebra)
   C4& operator +=(C4 const& a)
   { x1+=a.x1; x2+=a.x2; x3+=a.x3; x4+=a.x4; return *this;}
   C4& operator -=(C4 const& a)
   { x1-=a.x1; x2-=a.x2; x3-=a.x3; x4-=a.x4; return *this;}
   C4& operator *=(C const& r)
   { x1*=r; x2*=r; x3*=r; x4*=r; return *this;}
   C4& operator *=(const R& r)
   { x1*=r; x2*=r; x3*=r; x4*=r; return *this;}
   C4& neg(void)
   { x1=-x1; x2=-x2; x3=-x3; x4=-x4; return *this;  }
   C operator |(C4 const& a)const
   { return (~x1)*a.x1 + (~x2)*a.x2 + (~x3)*a.x3 + (~x4)*a.x4;}
   C trc(C4 const& a)const{ return x1*a.x1+x2*a.x2+x3*a.x3+x4*a.x4;}
   R abs(bool careful=true)const;
   R dis(C4 const&)const;
   X2<R,C4 > polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // a unit vector n such that *this = n*r. If *this is the zero
      // vector, n is arbitrarily chosen as (1,0,0,0)
   R normalize(){ X2<R,C4> p=polDec(); *this=p.c2(); return p.c1(); }
      // turns *this into a unit vector and returns the norm of the
      // original in order not to loose information
   R nor_(){ return normalize();}
   R norm()const{ return abs(); }
   C4 con()const{ return C4(~x1,~x2,~x3,~x4);}
   C4 rev()const{ return C4(x4,x3,x2,x1);}
      // multiplication with alpha_1 (topic: Dirac equation)
   C4 beta()const{ return C4(x1,x2,-x3,-x4);}
      // multiplication with  beta (topic: Dirac equation)
   C4 operator-()const{ return C4(-x1,-x2,-x3,-x4);}
   C4 operator*(R r)const{return C4(x1*r,x2*r,x3*r,x4*r);}
   C4 ran(Z j=0)const{
      if (j==0) return C4(x1.ran(),x2.ran(),x3.ran(),x4.ran());
      else{
         Z j4=j*4;
         return C4(x1.ran(j4),x2.ran(j4+1),x3.ran(j4+2),x4.ran(j4+3));
      }
   }
   V<C> toVC()const{ return V<C>{x1,x2,x3,x4};}
   CPM_IO
   CPM_WORD
   CPM_ORDER
   CPM_NAM(C4)

   Z dim()const{ return 4;}
   Z b()const{ return 1;}
   Z e()const{ return 4;}
   bool isVal()const{
      return x1.isVal()&&x2.isVal()&&x3.isVal()&&x4.isVal();
   }
};

inline C4 operator +(C4 const& x1, C4 const& x2)
{ C4 res=x1; return res+=x2;}

inline C4 operator -(C4 const& x1, C4 const& x2)
{ C4 res=x1; return res-=x2;}

inline C4 operator *(C4 const& x1, C const& s)
{ C4 res=x1; return res*=s;}

inline C4 operator *(C const& s, C4 const& x1)
{ C4 res=x1; return res*=s;}


//////////// lean square matrices (2*2) - (4*4) /////////////////////////

// comment to the last function where matrices are acting on vectors
// from the right side. This kind of action on a vector space gives a
// representation of the multiplicative monoid of matrices if the product
// law for endomorphisms of linear spaces is defined as contravariant.
// In our specific notation here:
// x*(a*b)=(x*a)*b
// Proof: x*(a*b)=~(a*b)*x=((~b)*(~a))*x=(~b)*((~a)*x)=((~a)*x)*b=(x*a)*b

/// notice: operator ~ not yet analysed and tested for the complex case
// !!!!!!!!!

///////////////////////////// class CC2 /////////////////////////////////
class CCC2;
class CCCC2;
class CC2{  //  complex 2*2 matrices

public:
   friend class CCC2;
   friend class CCCC2;
   typedef CC2 Type;
   C2 x1;
   C2 x2;
      // The usual matrix element xij is written in terms of the
      // data elements x1, x2 as xi.xj which implies that x1 is the first
      // row of the matrix and x2 the second row. This agrees with
      // Mathematica's way of representing matrices as lists of lists.
   static CC2 zero;
   static CC2 unit;
   static CC2 sig0,sig1,sig2,sig3;
      // Pauli sigma matrices, also named sigma_x, sigma_y, sigma_z;
      // sig0 is the unit matrix.
   CPM_IO
   CPM_TEST_ALL
      // this interface was added 2016-05-22 to facilitate testing.
      // Would be natural to have it implemented for all classes defined
      // clinalg.
   CPM_MAT_ALG(C2,CC2)  // see cpmlinalg.h
// constructors
   CC2():x1(),x2(){}
   CC2( C2 const& a1, C2 const& a2):x1(a1),x2(a2){}
   CC2(C z11, C z12, C z21, C z22):x1(z11,z12),x2(z21,z22){}
   explicit CC2(C4 const& c4):x1(c4[1],c4[2]),x2(c4[3],c4[4]){}
   explicit CC2(V<C> const& c4):x1(c4[1],c4[2]),x2(c4[3],c4[4]){}
   CC2(R4 const& x):x1(C(x.x1+x.x4),C(x.x2,-x.x3)),x2(C(x.x2,x.x3),
      C(x.x1-x.x4)){}
      // hermitean matrix generated from a four-vector, which is the basis
      // of spinor calculus
// access
   C2& operator[](Z i){ if (i>=2) return x2; return x1;}
   C2 const& operator[](Z i)const{ if (i>=2) return x2; return x1;}
   C det()const{ return x1.x1*x2.x2-x1.x2*x2.x1;}
//   CC2 operator !()const;
      // inverse matrix can be expressed easily in this case
   CC2 her()const;
      //: hermitean (conjugate)
      // returns the hermitean conjugate
   CC2 trn()const;
      //: transpose
 //  C operator *(C z)const{ return (x1.x1*z+x1.x2)/(x2.x1*z+x2.x2);}
      // CC2 acts on C as a 'broken linear transformation'

   C act(C z)const{ return (x1.x1*z+x1.x2)/(x2.x1*z+x2.x2);}
      // CC2 acts on C as a 'broken linear transformation'

   CC2 operator*(C z)const{ return CC2(x1*z,x2*z);}

   C4 toC4()const{ return C4(x1.x1,x1.x2,x2.x1,x2.x2);}
      // one has: CC2 x => x.toC4().toCC2()==x
      // and:     C4 y => y.toCC2().toC4()==y

   V<C> toVC()const{ return V<C>{x1.x1,x1.x2,x2.x1,x2.x2};}
      //: to V<C>

   CC2 per(Perm const& p)const;
      //: permute
      // assuming that p.dim()==4 the permutation p mixes the 4 components
      // and takes these as the components of a new CC2-matrix which
      // gets returned.

   R4 toR4(bool careful=true)const
      // recovering the four-vector from the hermitean matrix.
      // If *this is known to be hermitean one may call the function
      // with careful==false.
   {
      R y0,y1,y2,y3;
      // *this is the matrix
      //  y0 + y3  y1 -iy2
      //  y1 +iy2  y0 - y3
      // and the simple task is to extract the yi.
      if (careful){
         CC2 h=(*this+her())*0.5;
         y0=(h.x1.x1+h.x2.x2)[1]*0.5;
         y3=(h.x1.x1-h.x2.x2)[1]*0.5;
         y1=(h.x2.x1+h.x1.x2)[1]*0.5;
         y2=(h.x2.x1-h.x1.x2)[2]*0.5;
      }
      else {
         y0=(x1.x1+x2.x2)[1]*0.5;
         y3=(x1.x1-x2.x2)[1]*0.5;
         y1=(x2.x1+x1.x2)[1]*0.5;
         y2=(x2.x1-x1.x2)[2]*0.5;
      }
      return R4(y0,y1,y2,y3);
   }

   C2 trc2(C2 const& a)const{
      C u1=a.x1.con();
      C u2=a.x2.con();
      return C2(u1*x1.x1+u2*x1.x2,u1*x2.x1+u2*x2.x2);
   }

   C2 trc1(C2 const& a)const{
      C u1=a.x1.con();
      C u2=a.x2.con();
      return C2(u1*x1.x1+u2*x2.x1,u1*x1.x2+u2*x2.x2);
   }

   C trc()const{ return x1.x1+x2.x2;}
      //: trace

   C app(C2 const& a, bool second)const
      //: apply
      // returns the first or the second component of (*this)*a
      // depending of the speakingly named second argument.
      // A boolean argument is chosen here since it is available in this
      // form most directly in function Qreg::app_
   {
      return
         second ? x2.x1*a.x1+x2.x2*a.x2 : x1.x1*a.x1+x1.x2*a.x2;
   }

   C biLinForm(C2 const& a, C2 const& b)const
      //: bi-linear form
   {
      return x1.x1*a.x1*b.x1 +
             x1.x2*a.x1*b.x2 +
             x2.x1*a.x2*b.x1 +
             x2.x2*a.x2*b.x2 ;
   }

   C operator()(bool i, bool j)const{
      if (i && j)  return x2.x2;
      if (i && !j) return x2.x1;
      if (!i && j) return x1.x2;
      if (!i && !j)return x1.x1;
      return C();
   }

   V< V<C> > toVVC()const{ return V< V<C> >{x1.toVC(),x2.toVC()};}

   CC2 operator*(R r)const
   {
      return CC2(x1*r,x2*r);
   }

   CC2 operator-()const{ return CC2(-x1,-x2);}

   CC2 cc()const{ return CC2(x1.con(),x2.con());}
   //. complex conjugate

   static CC2 dydPrd(C2 const& a, C2 const& b)
      //: dyadic product
      // building a matrix as a dyadic product of vectors
      // (tensor product)
   {
      CC2 r;
      C2 c=b.con();
      r.x1.x1=a.x1*c.x1;
      r.x1.x2=a.x1*c.x2;
      r.x2.x1=a.x2*c.x1;
      r.x2.x2=a.x2*c.x2;
      return r;
   }
};

C2 operator*(const CC2&, const C2&);
CC2 operator*(const CC2& , const CC2&);
CC2 operator+(const CC2& , const CC2&);
CC2 operator~(const CC2&);
C2 operator*(const C2&, const CC2&);
CC2 operator-(const CC2&, const CC2&);

///////////////////////////// class CC3 /////////////////////////////////

class CC3{  //  complex 3*3 matrices
public:
   typedef CC3 Type;
   C3 x1;
   C3 x2;
   C3 x3;
public:
   static CC3 zero;
   static CC3 unit;
// constructors
   CC3(){}
   CC3(C3 const& a1, C3 const& a2, C3 const& a3):x1(a1),x2(a2),x3(a3){}
// access
   C3& operator[](Z i)
   { if (i>=3) return x3; if (i==2) return x2; return x1;}

   C3 const& operator[](Z i)const
   { if (i>=3) return x3; if (i==2) return x2; return x1;}

   CC3 operator-()const{ return CC3(-x1,-x2,-x3);}

   CC3 cc()const{ return CC3(x1.con(),x2.con(),x3.con());}
   //. complex conjugate

   V< V<C> > toVVC()const{ return
      V< V<C> >{x1.toVC(),x2.toVC(),x3.toVC()};}
   R abs()const{ return cpmsqrt(((x1|x1)+(x2|x2)+(x3|x3))[1]);}
   CPM_IO
   CPM_MAT_ALG(C3,CC3)
};

C3 operator*(const CC3&, const C3&);
CC3 operator*(const CC3& , const CC3&);
CC3 operator+(const CC3& , const CC3&);
CC3 operator~(const CC3&);
C3 operator*(const C3&, const CC3&);
CC3 operator-(const CC3&, const CC3&);

///////////////////////////// class CC4 /////////////////////////////////

class CC4{  // complex 4*4 matrices
public:
   typedef CC4 Type;
   C4 x1;
   C4 x2;
   C4 x3;
   C4 x4;

 //  static CC4 zero;
//   static CC4 unit;
 //  static CC4 gam0,gam1,gam2,gam3;
      // gamma matrices
      // Definition as in Itzykson and Zuber (IZ), p. 693 gamma matrices
      // in Dirac representation, here we give the version which in IZ
      // carries upper indexes.
 //  static CC4 beta,alpha1,alpha2,alpha3;
      // alpha, beta matrices for representing the Dirac Hamiltonian.

   // The corresponding quantities with appended W are the Weyl version.
//   static CC4 gam0W,gam1W,gam2W,gam3W;
//   static CC4 betaW,alpha1W,alpha2W,alpha3W;

// The corresponding quantities with appended M are the Majorana version.
 //  static CC4 gam0M,gam1M,gam2M,gam3M;
 //  static CC4 betaM,alpha1M,alpha2M,alpha3M;


// constructors

   CC4(){}
   CC4(C4 const& a1, C4 const& a2, C4 const& a3, C4 const& a4):
      x1(a1),x2(a2),x3(a3),x4(a4){}

   CC4(C4 const& a, C4 const& b):x1(a.x1*b),x2(a.x2*b),
      x3(a.x3*b),x4(a.x4*b){}
         // terse but correct coding of c_ij = a_i * b_j,
         // dyadic product |a><b*|

   CC4(CC2 const& a, CC2 const& b)
   {
      *this=CC4(a.toC4(),b.toC4());
   }

   CC4(CC2 const& a11, CC2 const& a12, CC2 const& a21, CC2 const& a22)
   {
      x1.x1=a11.x1.x1;
      x1.x2=a11.x1.x2;
      x2.x1=a11.x2.x1;
      x2.x2=a11.x2.x2;

      x3.x1=a12.x1.x1;
      x3.x2=a12.x1.x2;
      x4.x1=a12.x2.x1;
      x4.x2=a12.x2.x2;

      x1.x3=a21.x1.x1;
      x1.x4=a21.x1.x2;
      x2.x3=a21.x2.x1;
      x2.x4=a21.x2.x2;

      x3.x3=a22.x1.x1;
      x3.x4=a22.x1.x2;
      x4.x3=a22.x2.x1;
      x4.x4=a22.x2.x2;
   }


// access
   C4& operator[](Z i)
      { if (i>=4) return x4; if (i==3) return x3; if (i==2) return x2;
       return x1;}

   C4 const& operator[](Z i)const
       { if (i>=4) return x4; if (i==3) return x3; if (i==2) return x2;
        return x1;}

   CC4 operator-()const{ return CC4(-x1,-x2,-x3,-x4);}

   CC4 operator*(R r)const{ return CC4(x1*r,x2*r,x3*r,x4*r);}

   CC4 operator*(C z)const{ return CC4(x1*z,x2*z,x3*z,x4*z);}

   R abs()const{ return cpmsqrt(((x1|x1)+(x2|x2)+(x3|x3)+(x4|x4))[1]);}

   CC4 hc()const{ return ~CC4(x1.con(),x2.con(),x3.con(),x4.con());}
      //. hermitian conjugate

   CC4 cc()const{ return CC4(x1.con(),x2.con(),x3.con(),x4.con());}
   //. complex conjugate

   V< V<C> > toVVC()const{ return
      V< V<C> >{x1.toVC(),x2.toVC(),x3.toVC(),x4.toVC()};}

   CPM_IO
   CPM_MAT_ALG(C4,CC4) // ~ is transposition
   Word nameOf()const{ return "CC4";}

};

C4 operator*(const CC4&, const C4&);
CC4 operator*(const CC4& , const CC4&);
CC4 operator+(const CC4& , const CC4&);
CC4 operator~(const CC4&);
C4 operator*(const C4&, const CC4&);
CC4 operator-(const CC4&, const CC4&);

///////////////////////////// class Gamma //////////////////////////////
// Presenting means for defining Dirax 4x4 matrices: alpha, beta, gamma
// by simple initializers
class Gamma{
   CC4 g_;
   Gamma(CC4 const& g):g_(g){}
public:
   Gamma(Z i)
   {
      if (i==0)
         g_=CC4(CC2::sig0, CC2::zero, CC2::zero, -CC2::sig0);
      else if (i==1)
         g_=CC4(CC2::zero, CC2::sig1, -CC2::sig1, CC2::zero);
      else if (i==2)
            g_=CC4(CC2::zero, CC2::sig2, -CC2::sig2, CC2::zero);
      else if (i==3)
            g_=CC4(CC2::zero, CC2::sig3, -CC2::sig3, CC2::zero);
      else if (i==5)
            g_=CC4(CC2::zero, CC2::sig0, CC2::sig0, CC2::zero);
      else
            cpmerror("unvalid i = "&cpm(i)&" in Gamma(Z i)");
   }

   static CC4 beta(){ return Gamma(0).rep(); }
   static CC4 alpha(Z i){ return Gamma(0).rep()*Gamma(i).rep();}

   Gamma(Word const& w)
   {
      if (w=="unit")
             g_=CC4(CC2::sig0,CC2::zero,CC2::zero,CC2::sig0);
      else if (w=="zero")
             g_=CC4(CC2::zero,CC2::zero,CC2::zero,CC2::zero);
      else if (w=="U_Weyl")
            g_=CC4(CC2::sig0,-CC2::sig0,CC2::sig0,CC2::sig0)*
               cpminv(cpmsqrt(R(2.)));
      else if (w=="U_Majorana")
            g_=CC4(CC2::sig0,CC2::sig2,CC2::sig2,-CC2::sig0)*
               cpminv(cpmsqrt(R(2.)));
            // from Itzykson Zuber: Quantum Field Theory p. 694
      else cpmerror("unvalid w in Gamma(Word w)");
   }

   Gamma(Z i, Word const& w)
   {
      if (w=="Dirac") g_=Gamma(i).rep();
      else if (w=="Weyl"){
         CC4 U = Gamma("U_Weyl").rep();
         g_ = U*Gamma(i).rep()*U.hc();
      }
      else if (w=="Majorana"){ // For i=0,1,2,3,5 the matrix
         // Gamma(i,"Majorana") is purely imaginary. Tested!
         CC4 U = Gamma("U_Majorana").rep();
         g_ = U*Gamma(i).rep()*U.hc();
      }
      else cpmerror("unvalid w in Gamma(Z i, Word w)");
   }

   static CC4 beta(Word const& w)
      { return Gamma(0,w).rep(); }

   static CC4 alpha(Z i,Word const& w)
      { return Gamma(0,w).rep()*Gamma(i,w).rep();}

   CC4 rep()const{ return g_;}
};

/// special topic: tensor products of C2: CCC2, CCCC2
/// for my spin interaction application I do not longer need CC4
/// and use CCCC2 instead. CC4 is needed for Dirac equation in
/// cpmstatedirac.h.

class CCC2{ // pairs of CC2
   CC2 x1,x2;
   CCC2(CC2 const& y1,CC2 const& y2):x1(y1),x2(y2){}
public:
   friend class CCCC2;
   CCC2():x1(),x2(){}
   CCC2 operator + (CCC2 const& a)const{ return CCC2(x1+a.x1,x2+a.x2);}
   CC2 sel(bool second)const{ return second ? x2 : x1;}
      //: select
   CCC2 operator* (R r)const
   {
      return CCC2(x1*r,x2*r);
   }
};

class CCCC2{ // pairs of CCC2, for implementation
   // of tensor products of sigma matrices
   CCC2 x1,x2;
   CCCC2(CCC2 const& y1,CCC2 const& y2):x1(y1),x2(y2){}
public:
   CCCC2():x1(),x2(){}

// constructor for the tensor product of a with b
   CCCC2(CC2 const& a, CC2 const& b):x1(),x2()
   {
      x1.x1.x1.x1 = a.x1.x1 * b.x1.x1;
      x1.x1.x1.x2 = a.x1.x1 * b.x1.x2;
      x1.x1.x2.x1 = a.x1.x1 * b.x2.x1;
      x1.x1.x2.x2 = a.x1.x1 * b.x2.x2;

      x1.x2.x1.x1 = a.x1.x2 * b.x1.x1;
      x1.x2.x1.x2 = a.x1.x2 * b.x1.x2;
      x1.x2.x2.x1 = a.x1.x2 * b.x2.x1;
      x1.x2.x2.x2 = a.x1.x2 * b.x2.x2;

      x2.x1.x1.x1 = a.x2.x1 * b.x1.x1;
      x2.x1.x1.x2 = a.x2.x1 * b.x1.x2;
      x2.x1.x2.x1 = a.x2.x1 * b.x2.x1;
      x2.x1.x2.x2 = a.x2.x1 * b.x2.x2;

      x2.x2.x1.x1 = a.x2.x2 * b.x1.x1;
      x2.x2.x1.x2 = a.x2.x2 * b.x1.x2;
      x2.x2.x2.x1 = a.x2.x2 * b.x2.x1;
      x2.x2.x2.x2 = a.x2.x2 * b.x2.x2;
   }

// self-evident
   CCCC2 operator + (CCCC2 const& a)const{ return CCCC2(x1+a.x1,x2+a.x2);}

// the isotropic interaction term between two spin 1/2
// magnetic moments (or Heisenberg exchange interaction
// between spins)
   static CCCC2 spinSpin(R q, bool full=true){
      CCCC2 resPrel=CCCC2(CC2::sig1*q,CC2::sig1)
              + CCCC2(CC2::sig2*q,CC2::sig2);
      if (full) return resPrel+CCCC2(CC2::sig3*q,CC2::sig3);
      else return resPrel;
   }

// more flexible form for the interaction term between two spin 1/2
// magnetic moments
   static CCCC2 spinSpinXYZ(R qXY, R qZ){
      CCCC2 res;
      if (qXY != 0.) res = res + CCCC2(CC2::sig1*qXY,CC2::sig1)
              + CCCC2(CC2::sig2*qXY,CC2::sig2);
      if (qZ != 0.) res = res + CCCC2(CC2::sig3*qZ,CC2::sig3);
      return res;
   }

// Returns *this(a,b) where *this is considered as a bi-linear form.
   CC2 operator()(C2 const& a, C2 const& b)const
   {
      CC2 res;
      res.x1.x1= x1.x1.x1.x1 * a.x1 * b.x1
               + x1.x1.x1.x2 * a.x1 * b.x2
               + x1.x1.x2.x1 * a.x2 * b.x1
               + x1.x1.x2.x2 * a.x2 * b.x2 ;

      res.x1.x2= x1.x2.x1.x1 * a.x1 * b.x1
               + x1.x2.x1.x2 * a.x1 * b.x2
               + x1.x2.x2.x1 * a.x2 * b.x1
               + x1.x2.x2.x2 * a.x2 * b.x2 ;

      res.x2.x1= x2.x1.x1.x1 * a.x1 * b.x1
               + x2.x1.x1.x2 * a.x1 * b.x2
               + x2.x1.x2.x1 * a.x2 * b.x1
               + x2.x1.x2.x2 * a.x2 * b.x2 ;

      res.x2.x2= x2.x2.x1.x1 * a.x1 * b.x1
               + x2.x2.x1.x2 * a.x1 * b.x2
               + x2.x2.x2.x1 * a.x2 * b.x1
               + x2.x2.x2.x2 * a.x2 * b.x2 ;

      return res;
   }

   CCC2 sel(bool second)const{ return second ? x2 : x1;}
      //: select

   CC2 operator * (CC2 const& a)const;
      // strangling the second and the fourth index of *this by the
      // 2*2 matrix a and thus getting a 2*2 matrix, which gets returned.

   C app(CC2 const& a, bool b1, bool b2)const;
      // returns the matrix element (i,j) of (*this)*a
      // where
      // i=1 for b1=false, i=2 for b1=true
      // j=1 for b2=false, j=2 for b2=true
      // Only those products and sums are computed which
      // make up the requested matrix element. Thus if not
      // all matrix elements of (*this)*a are needed, we do less
      // work than computing the whole matrix (*this)*a.
};

#undef CPM_MAT_ALG

} // namespace

#endif
