//? cpmconstphys.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_CONSTPHYS_H_
#define CPM_CONSTPHYS_H_

// Values of physical constants in SI units. Comments indicate the SI
// unit
// Source: Huette: Grundlagen ... 29 Auflage p. B2, source there is
// CODATA Bull 63 Nov 1986
// All quantities end in _PHYS, names of greek letters are abreviated
// to three letters.
// 2007-02-24 comment added to Boltzmann's constant

#include <cpmsystem.h>

namespace CpmPhysics{

   using CpmRoot::R; // should not be multiple precision

const R c_PHYS=299792458.;
   //  m/s  speed of light in vacuum

const R my0_PHYS=cpmpi*4e-7;
//const R my0_PHYS=3.14159*4e-7;
   // (V*s)/(A*m) permeability of vaccum

const R eps0_PHYS=R(1)/(c_PHYS*c_PHYS*my0_PHYS);
   // (A*s)/(V*m) dielectric constant of vaccum
   // Notice my0 * eps0 * c * c = 1

const R influence_PHYS=R(1)/(cpmpi*eps0_PHYS*4);
   // (V*m)/(A*s) 'influence constant'

const R mag_influence_PHYS=R(1)/(cpmpi*my0_PHYS*4);
   // (A*m)/(V*s) 'magnetic influence constant'

const R e_PHYS=1.60217733e-19;
   // A*s electrical charge of proton

const R h_PHYS=6.6260755e-34;
   // J*s = A*V*s Planck's constant

const R h__PHYS=h_PHYS/(cpmpi*2);
   // J*s = A*V*s 'h quer'

const R NA_PHYS=6.0221367e23;
   // 1 Avogadro's constant

const R k_PHYS=1.380658e-23;
   // J/K Boltzmann's constant
   // Laut Physik Journal Dezember 2006 p. 37
   // ist geplant den Wert 1.38065e-23 als exakt
   // festzulegen und zur Neudefinition der Temperatur zu
   // verwenden.

const R R_PHYS=k_PHYS*NA_PHYS;
   // J/K universal gas constant

const R sig_PHYS=(cpmpi*cpmpi/60.)*(k_PHYS*k_PHYS/(h__PHYS*c_PHYS))*
(k_PHYS*k_PHYS/(h__PHYS*c_PHYS))/h__PHYS;
   // W/(m^2 * K^4) Stefan-Boltzmann constant

const R me_PHYS=9.1093897e-31;
   // kg mass of electron

const R mp_PHYS=1.6726231e-27;
   // kg mass of proton

const R mu_PHYS=1.6605402e-27;
   // kg mass unit mu = mass(C12)/12

const R muBohr_PHYS=e_PHYS*h__PHYS/(me_PHYS*c_PHYS*2);
   // A*m^2 Bohr magneton. The magnetic moment (Ampere's magnetic Moment)
   // of the electron is 1.00115966 times muBohr_PHYS.
   // Coulomb's magnetic moment is mu0 times this and has dimension T*m^3
   // T = kg*A^-1*s^-2 = J*A^-1*m^-2

const R G_PHYS=6.67259e-11;
   // N*m^2/kg^2 gravitational constant G , sometimes gamma

// more astronomical additions (from Kuchling, not particulary accurate)

const R mEarth_PHYS=5.97e24;
   // kg mass of Earth

const R mSun_PHYS=mEarth_PHYS*3.334e5;
   // kg mass of Sun

const R AU_PHYS=1.49598e11;
   // m Astronomical Unit (of length)

const R gEarth_PHYS=9.81;
   // acceleration of free fall as a conventional constant

extern R gEarth;
   // acceleration of free fall at the place under consideration
   // can meanigfully get changed, so should not be treated as constant

} // namespace

#endif

