//? cpmcvector.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_CVECTOR_H_
#define CPM_CVECTOR_H_

/*
   Purpose: see class C_Vector
*/
#include <cpmvr.h>
#include <cpmperm.h>
#include <cpmclinalg.h>
#include <cpminterval.h>
#include <cpmrvector.h>
#include <cpmsystemdependencies.h>

#include <vector>

namespace CpmArrays{

   using std::vector;
   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::C;
   using CpmArrays::Perm;
   using CpmLinAlg::R3;
   using CpmLinAlg::C2;
   using CpmLinAlg::C3;
   using CpmLinAlg::C4;
   using CpmGeo::Iv;
   using CpmGeo::Iv2;

#if defined(CPM_USE_EIGEN)
   using namespace Eigen;
#endif

   typedef Vr<Z> I_Vector;

   class C_Vector{
// vectors with C-valued components, rich interface including arithmetic
      // arbitrary length, dynamic allocation

   Vr<C> rep;
      // we choose composition instead of inheritance, so the
      // interface of Vr<C> needs to be known only to the implementer
      // of this class, the user finds the interface of C_Vector
      // in the present file together with the macro definitions
      // in cpminterfaces.h

   C_Vector(const Vr<C>& x):rep(x){}

   static C addFunc(const C& r1, const C& r2, const C& t)
      { return r1+r2*t;}

public:
   typedef C_Vector Type;
   typedef C ScalarType;
   CPM_IO
   bool scanFromFreeFor(istream&);
      // FreeFor stands for free format
      // reads numbers from a stream as long as this can successfully
      // be done and changes *this into the C_Vector having all these
      // numbers as components. So, one doesn't need to know the
      // number of these numbers in advance
   CPM_ORDER
   CPM_TEST
   CPM_DESCRIPTORS
   CPM_CONJUGATION
   CPM_DOT_PRODUCT
   CPM_NORMALIZE
   CPM_SUM_M
   CPM_PRODUCT_M
   CPM_SCALAR_M
   CPM_DIFFERENCE
   CPM_DIVISION
   Z dim()const{ return rep.dim();}
   Z size()const{ return rep.dim();}
   Z b()const{ return rep.b();}
   Z e()const{ return rep.e();}
   C_Vector& add(const C_Vector& v, C t)
      // adds v*t to *this and returns *this.
   { rep.fg_(addFunc,v.rep,t); return *this;}

   static R etaInv;
// constructors
   explicit C_Vector(Z n=0):rep(n,C()){}
   explicit C_Vector(const V<C>& v):rep(v){}
   explicit C_Vector(vector<C> const& v);
   explicit C_Vector(const C2& v):rep(V<C>{v[1],v[2]}){}
   explicit C_Vector(const C3& v):rep(V<C>{v[1],v[2],v[3]}){}
   explicit C_Vector(const C4& v):rep(V<C>{v[1],v[2],v[3],v[4]}){}
   explicit C_Vector(IvZ ivz):rep(ivz){}

   C_Vector(std::initializer_list<C> il):rep(V<C>(il)){}
      // requires C++11

   C_Vector(R_Vector const& re, R_Vector const& im);
      // constructor from two R_Vectors which represent the real part
      // and the imaginary part respectively.

   C_Vector(Z n, C x):rep(n,x){}
      // constructor for a vector with n comonents equal to x

   C_Vector(Z n, Iv2 iv, Z j=0);
      // Random constructor:
      // Creates a vector of index range 1,...,n and components
      // for which the real part ranges randomly over the interval ivR
      // and the imaginary part ranges randomly over the interval ivI.
      // The parameter j controls the seed of the random generator.

   C_Vector(Iv2 iv, Z n);
      // n equidistant components filling the interval iv (boundaries
      // included, (*this)[1]=iv.inf(), (*this)[n]=iv.sup() )
      // For n=1 (*this)[1]=(iv.inf()+iv.sup())/2
      // n=0 calls cpmassert


   C_Vector(C2 p, Z n);
      // List of n equidistant components starting with p[1] and ending
      // exactly with p[2]
      // For n=1 (*this)[1]=(p[1]+p[2])/2
      // n=0 calls cpmassert

#if defined(CPM_USE_EIGEN)
   explicit C_Vector( Matrix<std::complex<R>,Dynamic,1> const& );

   Matrix<std::complex<R>,Dynamic,1> toEigen()const;

#endif

  // end of constructors
   operator V<C>()const{ return rep;}
      // the opposite conversion is not automatic !

   vector<C> std()const{ return rep.std();}
      // returns a std::vector with the same components
      // Let res be the return value, then clearly
      // res[0]=rep[1]

   Vr<C> const& c()const{ return rep;}
   Vr<C> & c(){ return rep;}

   T2<R_Vector> prv()const;
      //. pair (of) real vectors
      // returns real parts and imaginary parts as R_Vectors

   R_Vector real()const;
      // returns the real part as a R_Vectors

   R_Vector imag()const;
      // returns the imaginary part as a R_Vectors

   R_Vector toRealForm()const;
      // returns a real 2*d-vector

   static C_Vector fromRealForm(const R_Vector& r);
      // returns a complex vector which is built in a natural way from r

   const C& cyc(Z i)const{ return rep.cyc(i);}
      //: cyclic
   C& cyc(Z i){ return rep.cyc(i);}
      //: cyclic

   C& operator[](Z i){ return rep[i];}
   const C& operator[](Z i)const{ return rep[i];}

   C& operator()(Z i){ return rep[i];}
   const C& operator()(Z i)const{ return rep[i];}

   C& cui(Z i){ return rep.cui(i);}
      //. component (with) unchecked index
   const C& cui(Z i)const{ return rep.cui(i);}
      //. component (with) unchecked index

   C& fir(){ return rep.fir();}
      //: first

   C const& fir()const{ return rep.fir(); }
      //: first

   C& last(){ return rep.last(); }
      //: last

   C const& last()const{ return rep.last(); }
      //: last

   C const& read(Z i, Outside mode)const{ return rep.read(i,mode);}
      // reading components in an efficient (as &'s), safe,
      // and flexible manner
      // mode 0: values outside assumed to be given as T()
      // mode 1: values outside assumed to be given by cyclic
      // continuation
      // else: values outside assumed to be given by constant
      // continuation

   C in_(C const& t, bool reverse=false){ return rep.in_(t,reverse);}
      // This operations treats *this as a shift register:
      // All components get shifted by one position 'to the right' and
      // the total length (dim) of the vector remains the same. So what
      // was the last component prior to the operation has to be removed
      // from the vector, in order to not wasting information, this
      // removed component will shine up as the return value of the
      // function. After the operation, the first component of the vector
      // will be t.
      // If reverse==true, t gets inputted at the end, and all shift
      // operations go 'to the left'.

   C_Vector eliFirst(Z nEli=1)const{ return C_Vector(rep.eliFirst(nEli));}

   C_Vector eliLast()const{ return C_Vector(rep.eliLast());}

   C_Vector eli(Z i, Z nEli=1)const{ return C_Vector(rep.eli(i,nEli));}

   C_Vector eli(IvZ const& iv)const{ return C_Vector(rep.eli(iv));}
      //:  eliminate
      // Eliminate the components the indexes of which belong to iv.

   C_Vector cut(Z i)const{ return C_Vector(rep.cut(i));}
      //: cut
      // returned is a C_Vector which results from *this by removing i
      // components from the end

   V<Z> permutationForIncreasingOrder()const
      {return rep.permutationForIncreasingOrder();}

   C_Vector permute(const V<Z>& p)const
      { return C_Vector(rep.permute(p));}

   C_Vector compose(const V<Z>& p)const
      { return C_Vector(rep.compose(p));}

   C_Vector operator*(const Perm& p)const
      { return C_Vector(rep.permute(p.getRep()));}

   C_Vector increasingCopy()const
      { return C_Vector(rep.increasingCopy());}

   C_Vector decreasingCopy()const
      { return C_Vector(rep.decreasingCopy());}

   C_Vector select(const V<bool>& sel)const
      { return C_Vector(rep.select(sel));}

   void order(){ rep.order();}

   void sort_(){ rep.order();}

   void set_(C const& t){ rep.set_(t);}
      // changes all components to the value t

   C_Vector& b_(Z i){ rep.b_(i); return *this;}
      // setting the beginning of the valid index range

   C_Vector& e_(Z i){ rep.e_(i); return *this;}
      // setting the end of the valid index range

   bool isVal()const;
      // returns false iff for any of the components
      // x[i] we have CpmRoot::isVal(x[i])==false

   C inf(Z i=0)const;
      // returns the 'ith-smallest' component. No ordering
      // for i=0. For i>0 it is the component indexed 1+i
      // of the increasingly ordered version of rep

   C sup(Z i=0)const;
      // returns the 'ith-largest' largest component.
      // For i=0 fast operation, no ordering
      // For i>0 it is the component indexed rep.dim()-i
      // of the increasingly ordered version of rep

   R supAbs()const;
      // returns sup{ |rep[i]|}

   Z indSup()const{ return rep.indSup();}
      // returns the smallest i for which (*this)[i]=(*this).sup()

   Z indInf()const{ return rep.indInf();}
      // returns the smallest i for which (*this)[i]=(*this).inf()

   C sum()const{ return rep.sum();}
      // returns the sum of all components

   C mean()const{Z n=rep.dim(); return n==0 ? C() : rep.sum()/n;}
      // arithmetic mean value of all components, fast

   C med()const // median value of all components
      { return infMedSup()[2];}

   Iv infSup()const{ return real().infSup();}
   // returns the smallest and the largest component. Fast operation,
   // no ordering

   C3 infMeanSup()const;
   // returns the smallest component, the arithmetic mean value,
   // and the largest component of *this. Fast operation, no ordering

   C3 infMedSup()const;
   // returns the smallest component, the median value,
   // and the largest component of *this. Achieved by ordering the
   // components. If one should also need the mean value, don't call
   // the previous function in addition, but simply the cheap function
   // mean().

   C_Vector pseudoInverse(R acc)const;
// Gives the pseudoinverse of the vector *this.
// This means that all components get inverted except those that are
// close to zero in comparison to the component with the largest
// absolute value.
// For small components c ( |c| ~ acc*sup |(*this)[i]| ) a continuous
// transition between the inverse and zero becomes operational.

    C_Vector scaleTo(const R& r)const;
// returns a vector of absolute value |r| and the direction of
// *this for positive r and antidirection for negative r.
// same function is available for complex n

   C_Vector resize(Z newDim)const
   { return C_Vector(rep.resize(newDim));}

   C_Vector subMean()const{ C s=-mean(); return C_Vector(rep+s);}
      //: subtract mean

   C subMean_(){ C m=mean(); rep-=m; return m;}
      //: changes *this by subtractng the mean from all components
      // and returns the mean

   C_Vector posCopy()const;
      // positive copy
      // returns a C_Vector obtained from *this by replacing all
      // components by other components in which the real part
      // is replaced by its absolute value and so is the imaginary part.

   C_Vector log10()const;
      // log10
      // returns a C_Vector obtained from *this by replacing all
      // components
      // by the log10 of their absolute value. The safe version of log10
      // which is a member function of R1 is employed for this.

   C_Vector operator()(F<C,C> const& f)const{ return C_Vector(rep(f));}

   C_Vector cutCopy(Iv iv)const;
      // Returns a vector with selected components, where the
      // selection is determined by iv.
      // Let ivr the section of iv and [0,1]. The interval [0,1] is
      // mapped in a natural way to the full index set {1,2,...,dim()}
      // and ivr then determines a contiguous subset of the full index
      // set. Returned is a vector which has just the components the
      // original indexes of which belong to this subset. In effect
      // the returned vector differs from *this by deleting some
      // (0 included) components at the beginning and some at the end.

   C3 variation()const;
      // returns (a,b,c) such that a=Min{ x[i+1]-x[i] : i=1,...d-1}
      // c=Max{ x[i+1]-x[i] : i=1,...d-1}
      // b=Mean{ |x[i+1]-x[i]| : i=1,...d-1}.
      // Thus a<=c, b<=Max(|a|,|c|). If *this is monotone ascending,
      // a is never negative. If *this consists of empirical data
      // which in the absence of errors and uncertainties would
      // be monotone ascending, a slightly negative a may be OK
      // if |a|<<b. Implemented as a single pass operation
      // without ordering or copying.
      // Notice that the meanvalue of x[i+1]-x[i] (thus without the abs)
      // is simply (x[n]-x[1])/(n-1) and thus has not to be determined by
      // a loop over all components.
      // Notice also that b*(d-1) is the sum of all |x[i+1]-x[i]|
      // which comes close to a polygon length if the spacing in
      // i-direction is neglegiable compared to the swing in x-direction.

   R var()const;
      // variant of variation which returns the length of the polgon
      // determined by the components.

   void adjPhase_(R eps);
      // adjust phase such that the component of largest absolute
      // value is real and positive. Actually the component of
      // largest absolute value is not always uniquely determined
      // independent of numerical noise. Therefore we give a
      // control parameter to determine a set of components that
      // will be under consideration together with the one of
      // largest absolute value: This set consists of those components
      // c of *this such that  dis(|c|,|cMax|)<eps ,
      // cMax: largest component.
      // From all these components under consideration, we make
      // the one with the smallest index positive.

   C_Vector der(R h=1, bool periodic=false) const;
// The vector a.der() results from the vector a by numerical
// differentiation of a function, the equi-distantly sampled version
// (sampling step = h) of which the vector a is considered to be.
// Formulas: Bronstein Semendiajew, Taschenbuch der Mathematik S. 768
// Tabelle 7.13 for generic points and Abramowitz Stegun p. 914
// Table 25.2 for the boundary points.
// If the second argument is true the function is periodic so that
// for i outside the allowed range we use this->cyc(i) as the component

   C_Vector rot(Z i,Outside mode=CYCLIC) const
//: rotate
// Name as the list transformation function Rotate of Mathematica.
// v.rot(s)[i] == v(i-s,mode)
// This means that we shift the component i of the original vector
// into the new position i+s. See V<>::rot(Z,Outside) in cpmv.h.
   { return C_Vector(rep.rot(i,mode));}

   C_Vector disLap(Outside mode=CYCLIC) const;
//: discrete Laplacean
// Here mode controls how to interprete indexes out of range. No reference
// to the lattice spacing is made.

   C sumPrd(C_Vector const& x )const;
      //: sum (of) products
      // Kind of scalar product without conjugation

   C_Vector& operator &=(const C_Vector& x1)
   { rep&=x1.rep; return *this;}
   // mutating form of concatenation

   C_Vector& operator &=(C r)
   { rep&=r; return *this;}

   C_Vector operator & (const C_Vector& x1)const
   { return C_Vector(rep&x1.rep);}
   // concatenation

   void operator<<(C const& r){ rep<<r;}
      // changes *this by appending a component, Ruby style

   void operator<<(C_Vector const& x1){ rep<<x1.rep;}
      // changes *this by appending a C_Vector, Ruby style

   C_Vector operator & (C const& r)const
   { return C_Vector(rep&r);}

   C_Vector prepend(C const& r)const
   { return C_Vector(rep.prepend(r));}

   void printLineOn(ostream& out , const Word& separator="  " )const
      // prints the components of *this, separated by the character
      // string separator. In default mode all numbers are in a line
      // which is usefull for creating data that later have to be given
      // to a plot program such as gnuplot.
   { for (Z i=1;i<=rep.dim();i++) out<<separator<<rep[i];}

   static V<C_Vector> E_Schmidt(V<C_Vector>& x, R eps=1e-6);
      // returns an orthonormal family which spans the same space
      // as family x. Made by E. Schmidt's orthogonalization;
      // eps controles the termination of the process and helps to avoid
      // returning vectors which are made 'out of numerical noise'.
      // More specifically we assume that x.dim() is an even number 2*n
      // and that there are just n complex-linearly independent members
      // in the family x. (all 2*n are real-orthonormal in our
      // application). Thus we try to return just n vectors. As is well
      // known, E. Schmidt's procedure tends to strongly accumulate
      // round-off errors for families of more than, say, 20 members.
      // So, we do a semi-understood reordering for larger families which
      // seems to cure all deseases

 };

// inline reduction of most functions to those of rep in Vr<R>
// in order not to loose speed

inline bool C_Vector::prnOn(ostream& out)const
{ return rep.prnOn(out);}

inline bool C_Vector::scanFrom(istream& in)
{ return rep.scanFrom(in);}

inline Z C_Vector::com(const C_Vector& x)const
{ return rep.com(x.rep);}

inline C_Vector& C_Vector::operator +=(const C_Vector& x)
{ rep+=x.rep; return *this;}

inline C_Vector& C_Vector::operator *=(const C& r)
{ rep*=r; return *this;}

inline C_Vector& C_Vector::operator *=(const C_Vector& x)
{ rep*=x.rep; return *this;}

inline C_Vector C_Vector::neg()const
{ return C_Vector(-rep);}

inline C_Vector C_Vector::ran(Z j)const
{ return C_Vector(rep.ran(j));}

inline Z C_Vector::hash()const
{ return rep.hash();}

inline C_Vector C_Vector::test(Z j)const
{ return C_Vector(rep.test(j));}

inline Word C_Vector::nameOf()const
{ return Word("C_Vector");}

inline Word C_Vector::toWord()const
{ return rep.toWord();}

inline C_Vector C_Vector::net(Z i)const
{ return C_Vector(rep.net(i));}

inline C_Vector C_Vector::inv()const
{ return pseudoInverse(etaInv);}

inline C_Vector C_Vector::con()const
{ return C_Vector(rep.con());}

inline C C_Vector::operator |(const C_Vector& x )const
{ return rep|x.rep;}

inline C C_Vector::sumPrd(const C_Vector& x )const
{ return rep.sumPrd(x.rep);}

} // namespace

#endif
