//? cpmfft.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_FFT_H
#define CPM_FFT_H
/*

   Description: Fourier transform of new effective template
      matrices and tensors. Intended for fourier transformation of
      images. E.g. determination of frequency response of cameras
      by dividing the fourier transform of a camera image by
      the fourier transform of the corresponding areal scene (e.g.
      test pattern). More functionality is provided by
      cpmfftmore.(h,cpp)

*/

#include <cpmc.h>
#include <cpmv.h>

// Implementation based on a modification of
// function fourn in Press et al.,
// Numerical Recipes in C, second edition, p. 523

/*** Definitions concerning the Discrete Fourier Transform *************
We consider complex-valued arrays with indexes running from 1 to some
even number N=2*n. Since we use i and j as array indexes, we have to
use for the imaginary unit something different, namely I.
We define a mapping f : {1,...,N} --> Q
   ( Q denotes the set of rationals, f stands for 'frequencies')
   f(i) := i<=n ? (i-1)/N : (i-1)/N - 1
and note the following properties:
   f(1)   = 0
   f(2)   = 1/N
   ...      ...
   f(n)   = (n-1)/N = 1/2 - f(2)
   f(n+1) =  n/N -1 = -1/2
   f(n+2) = (n+1)/N-1 = (n+1-N)/N = (-n+1)/N = -f(n)
   ...       ...
   f(N)   =  -1/N = -f(2)

Thus all values f(i) lie in the intervall [-1/2, 1/2 - 1/N ].
Roughly f(i) is positive in the first half of the 'interval'
[1,...,N] and negative in the last one; the graph of |f(i)|
grows linearly towards the center and then falls linearly
to (nearly) zero.

Now we define for a C-valued array x of dimension N=2*n the
C-valued array y=F(x) of the same dimension by

   y[i] : =
      N^(-1/2) times sum from j=1 to j=N of x[j]*exp[2*pi*I*(j-1)*f(i)]

   =: (Fy)[i]

for all i in {1,...,N}. Of course, this defines the discrete
Fourier transform. As the definition stands, it involves
repetitious computations that can be very effectively
streamlined if N is a power of two and thus gives rise to
the fast fourier transform (which, as just explained, is
the discrete fourier transformation, but implemented efficiently).
What are the properties of F(x) if x is real (i.e. x[i]=x[i]~,
where the complex conjugate of z \in C is written as z~)?
We compute for y := F(x) for general x (not necessarily real)
y[i]~ =
   N^(-1/2) times sum from j=1 to j=N of x[j]~*exp[2*pi*I*(j-1)*-f(i)]

 = ...

In order to recognize the right-hand side as F(x*) at some
other index position we define the mapping
 ~ : {1,...,N} --> {1,...,N}, i |--> i~ by

      i~ := i==1 ? 1 : N+2-i

which achieves

   f(i~) = -f(i) for all i in {1,...,N}

and satisfies
   i~~ = i

This allows to define two idempotent operations C and K
on complex arrays:

   (Cx)[i]  := x[i]~  (anti-linear operation)

   (Kx)[i]  := x[i~]  (linear operation)

Obviously C K = K C.

With these definitions we resume our calculation ...

... = N^(-1/2) times sum from j=1 to j=N of x[j]~*exp[2*pi*I*(j-1)*f(i~)]

      = K(F(Cx)).

Thus the whole computation shows
      C(F(x)) = K(F(C(x)))

or C F = K F C or

   C' F = F C , where C' := K C = C K   (*)

Notice  (C'x)[i] = x[i~]~ and that C' is an anti-linear idempotent
operation. We have
   x==Cx <==> Fx==C'Fx

Due to the pre-factor N^(-1/2) in the definition, the transformation
F is unitary ( <Fx|Fy> == <x|y> ). This is not the case with the
definition in Press et al.

*/

namespace CpmFourier{

   using CpmRoot::Z;
   using CpmRoot::C;
   using CpmArrays::V;
   using CpmArrays::VV;
   using CpmArrays::VVV;

// general fast Fourier Transform of complex 1-dimensional arrays
// See the 2-dimensional analog for explanations

V<C> fft(const V<C>& x, Z exponent=1);
// For real x and y=fft(x) we have
// y[i]~ = y[i~] (*)
// Assume y comes from a real and positive x (e.g. an linear image), and
// was modified by some image processing algorithm and than
// should be converted back to an image. If the algorithm did
// not some magic to ensure that relation (*) is still valid
// after the manipulation, the back-transformed image will no longer
// be real and ---a fortiori--- no longer positive.
// For multiplicative manipulations T (i.e.  (Ty)[i]:=a[i]*y[i] )
// the persistance of (*) can easily be ensured: one has to
// have a[i]~=a[i~], for real a thus simply a[i]=a[i~].
// If this is satisfied, the back transform is real and to
// get a posive function one could cut the negative parts to zero
// or take the absolute value. Both ways are not very natural and
// so it is interesting that there is a general natural prescription
// for unlimited image manipulations in 'Fourier space':
// Write the positive image x[i] as x[i]=|a[i]|^2 (e.g. a[i]=x[i]^(-1/2),
// a[i] could be chosen as complex in principle). Form
// y=F(a) manipulate y as to get a result y'. Transform back to
// a':=F'y' (here F' is the inverse of F, for which one has the
// same formula as for F but with f replaced by -f so that
// (F'x)[i]=(Fx)[i~].

// General fast Fourier Transform of complex 2-dimensional arrays

VV<C> fft(const VV<C>& x, Z exponent=1);
   // exponent 1 for fft, -1 for inverse transformation
   // The dimension of x is arbitrary. It will however be filled
   // with zeros up to the next power of two (not necessarily the same
   // in i and j direction)
   // The connection between matrix indexes in and frequencies is
   // in the previous function.
// For real x and y=fft(x) we have
// y[i][j]~ = y[i~][j~]

// general fast Fourier Transform of complex 3-dimensional arrays
// See the 2-dimensional analog for explanations

VVV<C> fft(const VVV<C>& x, Z exponent=1);
// For real x and y=fft(x) we have
// y[i][j][k]~ = y[i~][j~][k~]

} // namespace

#endif
