//? cpmimage.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_IMAGE_H_
#define CPM_IMAGE_H_
/*
    Description: Defines template class Image<T>
      Notice: Reading images from file is done in MultiImage, not here!
*/
#include <cpmgraph.h>
#include <cpmc.h>
#include <cpmhistogr.h>
#include <cpmfftmore.h>
#include <cpmfa.h>
#include <cpmimagingtools.h>

namespace CpmImaging{

   using namespace CpmStd;

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::C;
   using CpmRoot::L;
   using CpmRoot::toDouble;
   using CpmRoot::Conv;
   using CpmRoot::randomR;

   using CpmRootX::B;

   using CpmArrays::IvZ;
   using CpmArrays::VV;
   using CpmArrays::X2;
   using CpmArrays::R_Vector;
   using CpmArrays::R_Matrix;

   using CpmFunctions::F;
   using CpmFunctions::Fa;

   using CpmLinAlg::R2;
   using CpmLinAlg::Z2;

   using CpmFourier::fft;

   using CpmGraphics::Frame;
   using CpmGraphics::Graph;
   using CpmGraphics::Img24;

   using CpmGeo::Iv;

template<class T>
class Image : public ImagingTools, public VV<T> {
// real-valued image, gray image with high dynamic range
   // In namespace CpmGraphics (e.g. class CpmGraphics::Img24, class
   // CpmGraphics::Frame) pixel arrays were indexed according to
   // the conventional graphics indexing scheme. Here, in namepace
   // CpmImaging we switch to the mathematically much more natural
   // matrix indexing scheme: increasing the first index means moving
   // downwards, increasing the second index means moving to the right.
   // Valid pixel indexes start with 1.

   // Also continuous image coordinates (x,y) have different meaning here
   // and for CpmGraphics::Img24. Here (i.e. in the present class Image)
   // the image area is considered as the unit square with respect to a
   // coordinate system that has its origin at the left-lower corner of
   // the image and standard axes (x-axis horizontally to the right,
   // y-axis vertically upwards). So, the right-upper corner has
   // coordinates x=1, y=1. The idea is that in namespace CpmImaging
   // no graphics specific low level concepts should enter and that
   // therefore a rectangular area should be described in terms of
   // coordinates as in usual in analytic geometry.

   // In order for most of the following functions be meaningful it is
   // assumed that T is a type such that for t of type T statements
   //
   // R x=t;
   // R y=...; t=y;
   // make sense

   // From a data structure point of view T could also be a
   // multi-component type. Since earlier implementations of the basic
   // matrix template (now VV<T> did not allow this, I created the type
   // MultiImage which allows several image planes.

   R1 aspRat;    // aspect ratio = image width / image hight
      // (e.g. 1.5 for 35 mm film format) is n/m for square pixels
      // aspRat>1 for landscape and <1 for portrait
   typedef VV<T> Base;
   typedef Image Type;
public:

// constructors

   Image(void): VV<T>(1,1){aspRat=(R)Base::dim2()/Base::dim1();}

   Image(Img24 const&);
      // constructs a B/W image from an instance of the most
      // basic one of my image classes. Notice that instabces of class
      // Img24 can be read from PPM files.

   Image(Z mI, Z nI): VV<T>(mI,nI){aspRat=(R)Base::dim2()/Base::dim1();}
      // constructs a black image mI pixels high and nI pixels wide

   Image(Z mI, Z nI, T const& t): VV<T>(mI,nI,t)
      {aspRat=(R)Base::dim2()/Base::dim1();}

   Image(const VV<T>& y): VV<T>(y){aspRat=(R)Base::dim2()/Base::dim1();}

   Image(Z mI, Z nI, R xL, R yL, R dx, R dy,
      const F<C,C>& f, const F<C,R>& g):VV<T>(mI,nI)
      // see sampleComplexFunction(R,R,R,R,R,F<C,C>,F<C,R>) for details
      {   sampleComplexFunction(xL,yL,dx,dy,f,g);   }

   Image(Z mI, Z nI, R xL, R yL, R dx, R dy,
      const F<C,R>& f):VV<T>(mI,nI)
      // see sampleComplexFunction(R,R,R,R,R,F<C,C>,F<C,R>) for details
      {   sampleComplexFunction(xL,yL,dx,dy,f);   }

   explicit Image(VV<C> const& sp);
      // Construction of an image back from its fourier
      // components. So this is not a way to visualize the
      // sp itself. Notice that function
      // Graph::show(V<V<C> > const&,Iv,Z,bool,bool,R,bool)
      // provides convenient and fast visualization of the
      // objects of type VV<C>.


// access functions

   Z getLin()const{return Base::dim1();}
      // number of lines = height of image
   Z getCol()const{return Base::dim2();}
      // number of columns = width of image
   Z h()const{ return Base::dim1();}
      // height of image
   Z w()const{ return Base::dim2();}
      //  width of image

   T getPix(Z i, Z j)const;
     // getting pixel values, if pixels outside the image are
     // assessed, the image appears 'continued as constant'

   T getLocalMean(Z, Z, const Mask& )const;
      // get a weighted mean of pixel value (i,j) and its neighbors
      // where the contributing neighbors and their weights are
      // determined by the Mask argument

   R locVar(Z i, Z j)const;
      //: local variation
      // ad hoc method that is intended to give large values
      // at a position where an electronic image is in focus and
      // small values at positions out of focus.

   Image<R> locVar(bool useLocVar=false)const;
      //: local variation
      // Should discern between focussed parts and out-of-focus
      // parts of electronic images. I tested many ad hoc methods.
      // he present one, based on the Laplacian is simpler and
      // not worse then these.

   void putPix( Z i, Z j, T const& t);
     // setting pixel values, if pixels outside the image are
     // assessed nothing is done

   R getAspRat(void)const{ return aspRat;}
      //: get aspect ratio
      // returns the aspect ratio
      // aspRat=(R)dim2()/dim1();

   void setAspRat(R const& r){ aspRat=r;}
      //: set aspect ratio
      // sets the aspect ratio

 // others

   Image<T> interpolate(Z m, Z n, Z method)const;
      // Returns a m*n Image created from *this by a interpolation
      // method selected by the parameter method.
      // method =1 means bilinear interpolation and method=0 means
      // pixelation. The latter is useful if the number of pixels is to
      // be compared to expectations.
      // Also intended for m << height and n << width.

   Image<T> intPolAux(Z m, Z n, Z method)const;
      // Auxiliar interpolation method intended to be used
      // only for m > height/2, n > width/2; thus not for the case
      // m << height and n << width included above.

   Image<T> scl(R magFac, Z method=1)const
      //: scale
      // change number of columns and rows by a common factor
   {
      Z m=cpmrnd(magFac*h());
      Z n=cpmrnd(magFac*w());
      return interpolate(m,n,method);
   }

   T operator()(R x, R y, Z method=1)const;
      // returns the value at a point with respect to a
      // normal coordinate system (origin lower-left, x-axis
      // to the right, y-axis vertically ascending. The upper-right
      // image corner has coordinates x=1, y=1. This system
      // will be referred to as base system. Pixelation, or
      // bilinear interpolation according to 'method' as in previous
      // function interpolate(). For points outside the basic unit-square
      // the image is replicated periodically. This assures that an
      // arbitrary transformation of the plane will always define a image
      // with image content in the basic unit square. This function is
      // implemented such that all calls to that function for one image
      // take approximately the same time. So no preparatory calculations
      // for the first call are allowed. This is essential for forming a
      // Cpm-function out of an image (see class R2_Func). Details for
      // interpolation: The unit square is divided into (m*n) rectangular
      // pixels of identical size. The image values (pixel values) are
      // considered to be valid function values (of a real function of
      // two real variables) at the center locations of the pixels.
      // Function values at intermediate positions are defined by
      // interpolation (dependent on 'method') from these values. Notice
      // that it would be more natural to consider the pixel value as a
      // integral of some unknown distribution over a pixel area and not
      // as a sampled function value. Probably such an understanding also
      // would lead to bilinear interpolation.

   T evaluate( R x, R y)const;
      // returns the value at a point with respect to a
      // normal coordinate system (origin lower-left, x-axis
      // to the right, y-axis vertically ascending. The upper-right
      // image corner has coordinates x=1, y=1. This system
      // will be referred to as base system. Pixelation, no
      // interpolation.

   R2 pixelCenter(Z i, Z j)const;
      // Let i,j be matrix indexes. Then the return value
      // is the pixel center in the base system (see previous
      // function for definition).

   R2 minMax(R xL=0, R xU=1, R yL=0, R yU=1)const;
      // Let res be the return value. Then res.x is the minimum
      // and res.x2 the maximum of the pixel values of *this
      // This refers to the part of the image defined by xL,xU,yL,yU.
      // The pre-assigned values correspond to the whole image.

   R_Vector statistics(R xL=0, R xU=1, R yL=0, R yU=1)const;
      // Let res be the return value. Then res[1] is the minimum
      // and res[2] is the mean value and res[3] is the the maximum
      // of the pixel values of *this.
      // This refers to the part of the image defined by
      // xL,xU,yL,yU. The pre-assigned values correspond to the whole
      // image.
      // Future extensions could also give standard deviation and median.

   Histogram histogram(Z nBins=16,R valMin=0, R valMax=255)const;
      // makes a histogramm of an Image
      // nBins:   number of Sectors
      // valMin:   minimal value (normaly 0)
      // valMax:   maximal value (normaly 255)

   //void gamma(R val1, R val2, R valMin, R valMax, R gammaI);
      // Changes *this in a manner that is usually called
      // a tone-scale transformation and, more specifically,
      // a gamma conversion.
      // Notice that the tone scale control in functions mark
      // renders gamma transformations superfluous in all cases where
      // only the appearence on screen matters.
      // Let the pairs (Val1,Val2), and (ValMin,ValMax) result from
      // ordering the pairs (val1,val2), and (valMin, valMax)
      // (to be sure: Val1 < Val2), valMin <= valMax).
      // Uses a suitably shifted gamma curve f(.) to achieve
      //    f(Val1)=ValMin
      //    f(Val2)=ValMax.
      // Gamma has the usual meaning: gamma=1 for linear
      // curve, gamma <1 for enhancing shadow contrast, gamma > 1 for
      // enhancing contrast in full light.

   //void gamma(R valMin, R valMax, R gammaI=1.);
      // see gamma(R val1, R val2, R valMin, R valMax, R gammaI);
      // Uses a suitably shifted gamma curve for achieving a range of
      // gray-values from valMin to valMax. Gamma has the usual meaning:
      // gamma=1 for linear curve, gamma <1 for enhancing shadow
      // contrast,
      // gamma > 1 for enhancing contrast in full light.

   //void gamma(R gammaI);
      // see gamma(R val1, R val2, R valMin, R valMax, R gammaI);
      // Uses a gamma curve such that the range of gray-values remains
      // unchanged. Gamma has the usual meaning:
      // gamma=1 for linear curve, gamma <1 for enhancing shadow
      // contrast,
      // gamma > 1 for enhancing contrast in full light.

   Image<T>& setRan_(R valMin=0, R valMax=255, R gamma=Color::getGamma());
      //: set range
      // Changes *this in a manner that is usually called
      // a tone-scale transformation and, more specifically,
      // a gamma transformation.
      // The first two arguments fix the minimal pixelvalue and the
      // maximal pixel value of the transformed image. The value of
      // gamma may also be negative. In this case we have a logarithmic
      // curve. See CpmImagingTools::response for details.

   void sampleComplexFunction(const Graph& gr, C (*f)(const C&),
      R (*g)(const C&, const R_Vector&), const R_Vector& p );
      // not as flexible as the next function.

   void sampleComplexFunction(R xL, R yL, R dx, R dy,
      const F<C,C>& f, const F<C,R>& g);
      // Changes *this into an image with the same number of pixels
      // and values determined by f and transformed by g.
      // We consider the area of the complex plane C determined by
      //
      //      xL <= Re(z) <= xU,   xU = xL + dx
      //      yL <= Im(z) <= yU,   yU = yL + dy
      //
      // This area is subdivided in m rows and n colums of rectangular
      // pixels. The pixel values of the image are g(f(zPixCen)) where
      // zPixCen is the center of the pixel under consideration.
      // The correspondence between position in C and in the
      // resulting Image is matrix oriented: Image(1,1) corresponds
      // to the upper left corner of the rectangle in C.

   void sampleComplexFunction(R xL, R yL, R dx, R dy, const F<C,R>& f);
      // same as last function but f here corresponds to the composition
      // f&g there

   void insertPattern
      (R xL, R yL, R dx, R dy, const F<C,T>& f, const F<C,B>& g);
      // as previous function, but pixels are set only if g(z)=TRUE
      // otherwise pixel value is not changed

   void frame(const T& t);
      // a 1-pixel boundary line with value t is drawn in the picture

   void mark(Frame& fr, Response const& rsp, Z method=0,
      Z faithful=1)const;
      // Writes the image *this on the Frame fr.
      // This does not display the image. fr.vis() needs to
      // be called for showing the image on screen.
      // The object rsp transforms the pixel values
      // into values of type Color. The various member functions
      // of CpmImaging::Response give a wide range of options
      // for this transformation. E.g. rsp.setMethod(6) enables
      // pseudo-color representation of gray values.
      // The starting point so far is always the C++-provided
      // conversion from type T to type R so that Image<R> is the
      // most natural instantiation of the present template class.
      // The frame fr also determines the number of pixels to
      // be displayed. Since this may not coincide with the number of
      // pixels of the image, a suitable transformation has to be done.
      // This is controlled via parameters method and faithful.
      // The parameter 'method' controls pixel rendition:
      //    0 normal pixelation
      //    1 bilinear interpolation.
      // If the parameter 'faithful' is 1, the image will be placed in
      // such a manner to the upper left position that its aspect ratio
      // is correctly reproduced on the screen. Otherwise it is made
      // filling the whole frame.

   void mark(Frame& fr, R2 limits, Z method=0, Z faithful=1)const
      // Grayscale representation such lowest pixel value
      // (after C++-provided conversion to R)
      // corresponds to limits[1] and the highest color value to
      // limits[2]. Gamma is taken as Color::gamma.
      // If the image is created as a visualization tool for
      // non-image data one should consider using the previous
      // function for display that allows also pseudo-color
      // representation of real data.
   {
      Response rsp(limits[1],limits[2]);
      rsp.setGamma(Color::getGamma());
      mark(fr,rsp,method,faithful);
   }

   R2 mark(Frame& fr, Z method=0, Z faithful=1)const
      // As above, but limits is adjusted according to the value
      // range of the image. This value range is returned.
      {
         R2 limits=minMax();
         Response rsp(limits[1],limits[2]);
         rsp.setGamma(Color::getGamma());
         mark(fr,rsp,method,faithful);
         return limits;
      }

// trivial ordering not used but needed to allow Image as an template
// argument of V<> in its normal form
   Z operator < (const Image<T>& a){ return 0;}

   Image<T> cut(R xL, R xU, R yL, R yU)const;
      // reduces the size of the image by cutting the borders. All
      // arguments have to be in the interval [0,1].
      // The part of the image which is left after cut has normal
      // coordinates (x,y) with respect to the original which satisfy
      //   xL <= x <= xU
      //   yL <= y <= yU
      // Normal coordinate system: origin lower-left, x-axis
      // to the right, y-axis vertically ascending. The upper-right
      // image corner has coordinates x=1, y=1.

   Image<T> crop(Iv const& ivx, Iv const& ivy)const
      //: crop
      // creates a rectangular cut-out of the image by giving the
      // subinterval of the x-axis and the subinterval of the
      // y-axis to be used. Taking the whole image corresponds
      // to ivx=Iv(0,1), ivy=Iv(0,1). The origin (x,y)=(0,0) is
      // the upper-left corner of the image. Notice that for images with
      // an aspect ratio different from 1, the aspect ratio of the croped
      // image differs from the quotient ivx.abs()/ivy.abs().
      // same style as Img24::crop()
   { return cut(ivx.b(),ivx.e(),R(1)-ivy.e(),R(1)-ivy.b());}

   Image<T> cut(Z2 cLL, Z2 cUR)const;
      // similar to the previous function. Here the lower left corner
      // and the upper right corner are given as pixel positions in
      // graphics format. Most recent version allows to use any pair of
      // diagonally opposed corners (without that, the function is very
      // prone to error)

   Image<T> cut(Z i1, Z i2, Z j1, Z j2)const;
      // returning the subimage with pixels characterized by their
      // boundaries in matrix indexing scheme:
      // getPix(i,j) i1<=i<=i2, j1<=j<=j2

   Image<T> clip(R clipL, R clipU)const;
      // x.clip(a,b) takes values between a*x.min() and b*x.max()

   Image<T> cut(IvZ const& iRange, IvZ const& jRange)const
   { return cut(iRange.b(),iRange.e(),jRange.b(),jRange.e());}

   template <class Y>
   Image<Y> transformToneScale(F<T,Y> const& f)const;
      // returns a new image that takes pixel value f(val)
      // where the original image had pixel value val.

   Image<T> conv(R_Matrix const& a, Z si=0, Z sj=0)const;
      // R_Matrix a is taken as an convolution kernel with a(ci+si,cj+sj)
      // considered the center of the kernel.
      // Here
      // ci := 1+dim1(a)/2
      // cj := 1+dim2(a)/2
      // which is the actual center if the dimensions of a both are odd
      // which is the normal case.
      // No normalization will be done on a, since this can be done more
      // flexibly in advance.

   Image<T> conv(R_Vector const& a, R_Vector const& b)const;
      // does the same as conv(R_Matrix(a,b)) but much more
      // efficiently. Makes use of the separability of the
      // kernel.

   Image<T> smooth(Z n=1)const;
      // convolution with simple binomial kernel
      // operation is done n times

   Image<R> lap(bool makePositive=true, Z p=1)const;
      //: Laplace
      // Since the Laplacian kernel {{0,1,0},{1,-4,1},{0,1,0}} may
      // give negative pixel values, these can be made positive by
      // giving the argument the value true (which also is the default).

   Image<T> conv(const Mask& a)const;
      // returns the image res for which
      // res[i][j]=getLocalMean(i,j,a) for all i,j
      // 'modern' form of convolution (?)
      // conv(R_Vector,R_Vector) is by far most efficient;
      // but not all kernels are seperable.

   Image<T>& operator *= (R r);
      // changing the image into one pixelwhise multilpied by r

   Image<T>& operator += (R r);
      // changing the image into one where pixelwhise r is added

   VV<C> fourierTransform()const;
      // returns the fourier-transformed image as it is used in
      // function fourierFilter. The function
      // Graph::show(V<V<C> > const&,Iv,Z,bool,bool,R,bool)
      // provides convenient and fast visualization of the
      // result. Also the constructor
      // MultiImage( Word w, VV<C> const& vv) allows
      // to generate a color image with all options of tonescale
      // transformation, display, and filing (via Img24).
      // The upper-left corner of the VV<C>-typed result corresponds
      // to areal frequency 0. The center of the matrix to Nyquist
      // frequency in both directions, and the parts to the right or/and
      // below the center to frequencies which are negative at least
      // for one of the two directions.

   Image<T> fourierFilter(F<R2,R> const& resp)const;
      // multiplies the fourier transform by function rep which
      // takes meaningful arguments in [0,1]x[0,1], where 0 corresponds
      // to frequency 0 and 1 to Nyquist frequency along the
      // respective axis.
      // The negative frequencies (which are not independent) are treated
      // automatically in a consistent manner.
      // T is assumed to allow meaningful two-way conversion to R.
      // The resulting image has lines and columns equal to powers of
      // 2. The input image is not required to have this property.
      // For calculating its fourier transform we fill the missing pixels
      // in by 0.

   Image<T> fftFilter(const F<C,C>& resp)const;
      // similiar as above but using the more natural
      // rectangle [-0.5,0.5]x[-0.5,0.5] and operating on
      // the amplitude instead of the intensity. Assumes T=R.

   Image<T> cut(const Rect& rect)const;
      // creates a rectangular cut-ut of the image
      // See class Rect for details on the description of the
      // rectangular area.

   R norm1(void)const;
       // mean value of absolute pixel values

   VV<T> toVV()const{ return (Base)(*this);}

   Img24 toImg24()const;

   Img24 halfTone(R const& gamma=1.)const;
      //: half tone
      // Returns a monochrome image obtained by a primitive
      // but natural rasterless half tone algorithm.

   Image<R> diff(Image<T> const& im)const;
      // returns the positive version of the difference

   Image<R> operator*(Image<T> const& im)const;
      // returns the pixelwise product of *this and im.

   Image<R> operator+(Image<T> const& im)const;
      // returns the pixelwise sum of *this and im.

   Image<R> operator-(Image<T> const& im)const;
      // returns the pixelwise difference of *this and im.

   void horLine(Z x1,Z x2, Z y, T c);
      // fast way to draw a horizontal line

   static T biLin(R f11, R f12, R f21, R f22,
      R w11, R w12, R w21, R w22, Z method);
   // law for bi-linear interpolation (for method==1, else pixelation)

   bool isVal()const;
      //: is valid

};

typedef Image<R> GrayImgR;
   // in analogy to typedef MultiImage<R> ImgR

template <class T>
Image<T>::Image(const Img24& img)
{
   cpmmessage("Image(Img24) started");
   Z m=img.getLin();
   Z n=img.getCol();
   Image<T> res(m,n);
   Z i,j;
   R prfrac=1./m;
   for (i=1;i<=m;i++){
      cpmprogress(" putPix()-loop in constructor",R(i)*prfrac);
      for (j=1;j<=n;j++){
         rgb pixel=img.getPel(j-1,i-1);
         res.putPix(i,j,pixel.gray());
      }
   }
   *this=res;
   cpmmessage("Image(Img24) done");
}

template<class T>
Image<T>::Image(VV<C> const& sp)
{
   Z mL=2;
   static Word loc("Image<>::Image(VV<C>)");
   CPM_MA
   Z m=sp.dim1(),n=sp.dim2(),i,j;
   cpmassert(m%2==0 && n%2==0,loc);
   VV<C> res=fft(sp,-1);
   *this=Image<T>(m,n);
   for (i=1;i<=m;++i) for (j=1;j<=n;++j){
      (*this)[i][j]=(res[i][j])[1];
   }
   CPM_MZ
}

template<class T>
bool Image<T>::isVal()const
{
   Z i,j,m=Base::dim1(),n=Base::dim2();
   for (i=1;i<=m;++i){
      for (j=1;j<=n;++j){
         bool b=cpmiva((*this)[i][j]);
         if (!b) return false;
      }
   }
   return true;
}


template<class T>
T Image<T>::getPix(Z i, Z j)const
{
   Z m=Base::dim1(),n=Base::dim2();
   if (i<1) i=1;
   if (j<1) j=1;
   if (i>m) i=m;
   if (j>n) j=n;
   return (*this)[i][j];
}

template<class T>
void Image<T>::putPix( Z i, Z j, T const& t)
{
   if (i<1) return;
   if (j<1) return;
   Z m=Base::dim1(),n=Base::dim2();
   if (i>m) return;
   if (j>n) return;
   (*this)[i][j]=t;
}

template<class T>
T Image<T>::evaluate(R xI, R yI)const
   // returns the value at a point with respect to a
   // normal coordinate system (origin lower-left, x-axis
   // to the right, y-axis vertically ascending. The upper-right
   // image corner has coordinates x=1, y=1.
    // no interpolation involved
{
   Z i,j;
   const R tiny=1e-7;
   if (xI<=0) xI=tiny;
   if (xI>=1) xI=R(1)-tiny;
   if (yI<=0) yI=tiny;
   if (yI>=1) yI=R(1)-tiny;
   Z m=Base::dim1(), n=Base::dim2();
   i=(Z)(yI*m);  // i=0,...m-1
   j=(Z)(xI*n);  // j=0,...n-1
      // i,j discretized y,x
   j++; // j=1,...n
   i=m-i; // i=m,...1
      // now i,j matrix indexes
   return (*this)[i][j];
}

template<class T>
R2 Image<T>::pixelCenter(Z i, Z j)const
   // Let i,j be matrix indexes. Then the return value
   // is the pixel center in the base system
{
   Z m=Base::dim1(),n=Base::dim2(); // a silly compiler asked for
      // scope resolution
   R dx=1./n, dy=1./m;
   Z iy=m-i; //  iy= 0,...m-1
   Z jx=j-1; //  ix=0,...n-1
   R2 res(dx*(jx+0.5),dy*(iy+0.5));
   return res;
}

template <class T>
inline T r2t(R r, T t){ t; return static_cast<T>(toDouble(r));}

template<class T>
T Image<T>::biLin(R f11, R f12, R f21, R f22,
   R w11, R w12, R w21, R w22, Z method)
{
   T t=T();
   if (method==1){
      R p11=w11*w21;
      R p12=w11*w22;
      R p21=w12*w21;
      R p22=w12*w22;
      R fp = f11*p11 + f12*p12 + f21*p21 + f22*p22;
      return r2t(fp,t);
   }
   else{
      Z i= w11<=w12 ? 2 : 1;
      Z j= w21<=w22 ? 2 : 1;
      if (i==1 && j==1) return r2t(f11,t);
      else if (i==1 && j==2) return r2t(f12,t);
      else if (i==2 && j==1) return r2t(f21,t);
      else return r2t(f22,t);
   }
}

template<class T>
Image<T> Image<T>::intPolAux(Z mI, Z nI, Z method)const
{
   Z mL=3;
   cpmmessage(mL,"Image<T>::interpolate(...) started. m="&
      CpmRoot::toWord(mI)&" n="&CpmRoot::toWord(nI));
   Image<T> res(mI,nI);
   Z i,j;
   R f11,f12,f21,f22;
   Z m=h();
   Z n=w();
   if (m==mI && n==nI) return *this;
   Iv iv(0,1);
   V<R> ci=iv.centers(m);
   V<R> cj=iv.centers(n);
   V<R> cIi=iv.centers(mI);
   V<R> cIj=iv.centers(nI);
   for (i=1;i<=mI;i++){
      R xiI=cIi[i];
      X2<R2,Z2> qi=iv.intPolInfo(xiI,m);
      R2 wi=qi.c1();
      Z2 zi=qi.c2();
      R wi1=wi.x1;
      R wi2=wi.x2;
      Z i1=zi.x1;
      Z i2=zi.x2;
      for (j=1;j<=nI;j++){
         R xjI=cIj[j];
         X2<R2,Z2> qj=iv.intPolInfo(xjI,n);
         R2 wj=qj.c1();
         Z2 zj=qj.c2();
         R wj1=wj.x1;
         R wj2=wj.x2;
         Z j1=zj.x1;
         Z j2=zj.x2;
         f11=(R)(*this)[i1][j1];
         f12=(R)(*this)[i1][j2];
         f21=(R)(*this)[i2][j1];
         f22=(R)(*this)[i2][j2];
         res[i][j]=(T)biLin(f11,f12,f21,f22,wi1,wi2,wj1,wj2,method);
      }
   }
   return res;
}

template<class T>
Image<T> Image<T>::interpolate(Z mI, Z nI, Z method)const
{
   Z mL=3;
   cpmmessage(mL,"Image<T>::interpolate(...) started. m="&
      CpmRoot::toWord(mI)&" n="&CpmRoot::toWord(nI));
   Z m=getLin();
   Z n=getCol();
   if (mI>m/2 && nI>n/2){ // traditional case, interpolation to
      // similar or larger resolution
      return Image<T>::intPolAux(mI,nI,method);
   }
   else{ // new case: interpolating to lower resolution
         // based on averaging; works well for characters
         // in fonts
      R mIinv=1./mI, nIinv=1./nI;
      Z mf=cpmrnd(R(m)*mIinv), nf=cpmrnd(R(n)*nIinv);
      Z mA=mI*mf, nA=nI*nf;
      Image<T> resA=Image<T>::intPolAux(mA,nA,method);
         // image of similiar resolution than the original
         // but allowing schematic averaging
      Image<T> res(mI,nI);
      Z i,j,iR,jR,k,l;
      R sum,nmInv=1./(mf*nf);
      for (i=1;i<=mI;++i){
         iR=1+(i-1)*mf;
         for (j=1;j<=nI;++j){
            sum=0;
            jR=1+(j-1)*nf;
            for (k=0;k<mf;++k){
               for (l=0;l<nf;++l){
                  sum+=resA[iR+k][jR+l];
               }
            }
            res[i][j]=toDouble(sum*nmInv);
         }
      }
      return res;
   }
}

namespace{
// tools for next function
inline void dec(Z& i){ if (i>1) i--;}
inline void inc(Z& i, Z m){ if (i<m) i++;}
   // for fast access, this checks correctness of indexes only where
   // leaving the range may happen, a inline version of getPix() would
   // check always.
}
// main part
template<class T>
T Image<T>::operator()(R x, R y, Z method)const
// should be as efficient as possible, thus let us accept the ugly code
// structure. In a typical application I observes 8% speed loss by
// replacing
// (*this)[i][j] by getPix(i,j). Since this code is at the heart of
// considering
// images as functions it might be worthwhile to be as efficient as
// possible.
{
   Z m=getLin();
   Z n=getCol();
   x-=cpmfloor(x);
   y-=cpmfloor(y);
      // now 0<=x<1, 0<=y<1, makes the periodicity with unit period
   x*=n;
   y*=m;
      // 0<=x<n, 0<=y<m,
   Z xd=cpmfloor(x);
      // varies from 0 to n-1
   Z yd=cpmfloor(y);
      // varies from 0 to m-1
   Z i=m-yd;   // varies from m to 1
   Z j=xd+1;   // varies from 1 to n
   if (method==0){
       return (*this)[i][j];
    //  return getPix(i,j);
   }
   else if (method==1){
      // unfortunately every case seems to need special consideration:
      // see my pages UM/CTL 98-3-6 in the Folder UM 98, CTL, KAGUM
        x-=xd;
      y-=yd;
      R p1,p2,f00,f01,f10,f11;
      Z iNext=i;
      Z jNext=j;
         // these will be changed by reference by dec and inc
         // this will happen only once
      if (x==0.5){
         if (y==0.5) return (*this)[i][j] ;

         if (y>0.5) {
            f00=(*this)[i][j];
            dec(iNext);
            f01=(*this)[iNext][j];
            p2=y-0.5;
         }
         else{ // y<0.5
            inc(iNext,m);
            f00=(*this)[iNext][j];;
            f01=(*this)[i][j];
            p2=y+0.5;
         }
         return f01*p2-f00*(p2-1);
      }

      if (x>0.5){
         if (y>0.5){  // OK
            dec(iNext);
            inc(jNext,n);
            f00=(*this)[i][j];
            f10=(*this)[i][jNext];
            f01=(*this)[iNext][j];
            f11=(*this)[iNext][jNext];
            p1=x-0.5;
            p2=y-0.5;
         }
         else if (y<0.5){   // 0K
            inc(iNext,m);
            inc(jNext,n);
            f00=(*this)[iNext][j];
            f10=(*this)[iNext][jNext];
            f01=(*this)[i][j];
            f11=(*this)[i][jNext];
            p1=x-0.5;
            p2=y+0.5;
         }
         else if (y==0){
            inc(jNext,n);
            f00=(*this)[i][j];
            f10=(*this)[i][jNext];
            p1=x-0.5;
            return f10*p1-f00*(p1-1);
         }
      }
      else{ // then x<0.5, since == was carried to return earlier
         if (y>0.5){  // OK
            dec(iNext);
            dec(jNext);
            f00=(*this)[i][jNext];
            f10=(*this)[i][j];
            f01=(*this)[iNext][jNext];
            f11=(*this)[iNext][j];
            p1=x+0.5;
            p2=y-0.5;
         }
         else if (y<0.5){
            inc(iNext,m);
            dec(jNext);
            f00=(*this)[iNext][jNext];
            f10=(*this)[iNext][j];
            f01=(*this)[i][jNext];
            f11=(*this)[i][j];
            p1=x+0.5;
            p2=y+0.5;
         }
         else if (y==0){
            inc(jNext,n);
            f00=(*this)[i][j];
            f10=(*this)[i][jNext];
            p1=x-0.5;
            return f10*p1-f00*(p1-1);
         }
      }
      R b=f10-f00;
      R c=f01-f00;
      R d=f11-f00-b-c;
      return f00+b*p1+(c+d*p1)*p2;
         // f(x,y):=f00+b*x+c*y+d*x*y has to obey
         // f(0,0)=f00, f(0,1)=f01, f(1,0)=f10, f(1,1)=f11
         // indexes and arguments refer to the (x,y)-system for which
         // lower left corner is (0,0) and right upper corner is (1,1)
   }
   else{
      cpmerror("Image<T>::operator(): unvalid parameter method");
      return 0;
   }
}

template <class T>
Image<T>& Image<T>::setRan_(R valMin, R valMax, R gamma)
{
   Z mL=2;
   Word loc("Image<T>::setRan_(R,R,R)");
   CPM_MA
   R2 mm=minMax();
   R vmin=mm.x1;
   R vmax=mm.x2;
   Z m=getLin();
   Z n=getCol();
   Z i,j;
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         R pix=getPix(i,j);
         R pixNew= ImagingTools::response(vmin,vmax,valMin,valMax,
            gamma,pix);
         putPix(i,j,Conv<R,T>(pixNew)());
      }
   }
   CPM_MZ
   return *this;
}


template<class T>
void Image<T>::sampleComplexFunction(const Graph& gr, C (*f)(const C&),
    R (*g)(const C&, const R_Vector&), const R_Vector& p )
   // The rectangle in object space which belongs to *this is considered
   // a m*n pixel matrix independently of the pixel structure which also
   // is determined by *this. We take samples of the function f (defined
   // on object space) at the positions of pixel centers.
   // The correspondence between position in object space and in the
   // resulting GrayImage is matrix oriented: GrayImage(1,1) corresponds
   // to the upper left corner of object space.
   // The complex values of f at the sampling positions are transformed
   // into real values via the transformation g under control of the
   // parameter vector p ( allows maximum flexibility due to variable
   // number of components.
{
   C z,w;
   R aspRatGraph=cpmabs(gr.getyUpper()-gr.getyLower());
   cpmassert(aspRatGraph!=0.,"");
   aspRatGraph=cpmabs(gr.getxRight()-gr.getxLeft())/aspRatGraph;
   setAspRat(aspRatGraph);
      // aspect ration of image should be that of the object space
      // area of interest. Then an orthogonal line system in the
      // complex plane will be shown as an orthogonal line grid if in
      // Frame::show(GrayImage a, R_Vector par, Z method, Z faithful )
        // the last parameter is set to 1.
   Z i,j,i_;
   R u,v;
   T r;
   Z m=Base::dim1(), n=Base::dim2();
   V<R> iLattice=ImagingTools::pixelCenters(gr.getyLower(),
      gr.getyUpper(),m);
      // matrix index i varies in y direction
   V<R> jLattice=ImagingTools::pixelCenters(gr.getxLeft(),
      gr.getxRight(),n);
      // matrix index j varies in x direction
   for (i=1;i<=m;i++){
      i_=m-i+1;
      v=iLattice[i_];
      for (j=1;j<=n;j++){
         u=jLattice[j];
         z=C(u,v);
         w=f(z);
         r=g(w,p);  // transformation C to R
         (*this)[i][j]=r;
      }
   }
}

template<class T>
void Image<T>::sampleComplexFunction(R xL, R yL, R dx, R dy,
      const F<C,C>& f, const F<C,R>& g)
{
   const Z nTacit=4;
      // controls frequency of messages in verbose mode

   cpmassert(dx>=0.,"");
   cpmassert(dy>0.,"");

   R xU=xL+dx;
   R yU=yL+dy;

   setAspRat(dx/dy);
      // aspect ration of image should be that of the object space
      // area of interest. Then an orthogonal line system in the
      // complex plane will be shown as an orthogonal line grid if in
      // Frame::show(GrayImage a, R_Vector par, Z method, Z faithful )
      // the last parameter is set to 1.
   Z m=Base::dim1(), n=Base::dim2();

   Z i,j,i_;
   R u,v;
   V<R> iLattice=ImagingTools::pixelCenters(yL,yU,m);
      // matrix index i varies in y direction
   V<R> jLattice=ImagingTools::pixelCenters(xL,xU,n);
      // matrix index j varies in x direction

   C z,w;
   T r;
   Z lineCounter=0;
   R mInv=1./m;

   for (i=1;i<=m;i++){
      i_=m-i+1;
      v=iLattice[i_];
        if (verbose==1)
      {   lineCounter++;
         if (lineCounter==nTacit){
            lineCounter=0;
            R fr=R(i)*mInv*100;
            cout<<endl<<"Image "<<fr<<" % completed";
         }
        }

      for (j=1;j<=n;j++){
         u=jLattice[j];
         z=C(u,v);
         w=f(z);
         r=g(w);  // transformation C to R
         (*this)[i][j]=r;
      }
   }
}

template<class T>
void Image<T>::
sampleComplexFunction(R xL, R yL, R dx, R dy, const F<C,R>& f)
{
   const Z nTacit=4;
      // controls frequency of messages in verbose mode

   cpmassert(dx>=0.,"");
   cpmassert(dy>0.,"");

   R xU=xL+dx;
   R yU=yL+dy;

   setAspRat(dx/dy);
      // aspect ration of image should be that of the object space
      // area of interest. Then an orthogonal line system in the
      // complex plane will be shown as an orthogonal line grid if in
      // Frame::show(GrayImage a, R_Vector par, Z method, Z faithful )
      // the last parameter is set to 1.
   Z m=Base::dim1(), n=Base::dim2();

   Z i,j,i_;
   R u,v;
   V<R> iLattice=ImagingTools::pixelCenters(yL,yU,m);
      // matrix index i varies in y direction
   V<R> jLattice=ImagingTools::pixelCenters(xL,xU,n);
      // matrix index j varies in x direction

   C z,w;
   T r;
   Z lineCounter=0;
   R mInv=R(1)/m;

   for (i=1;i<=m;i++){
      i_=m-i+1;
      v=iLattice[i_];
        if (verbose==1)
      {   lineCounter++;
         if (lineCounter==nTacit){
            lineCounter=0;
            R fr=R(i)*mInv*100;
            cout<<endl<<"Image "<<fr<<" % completed";
         }
      }

      for (j=1;j<=n;j++){
         u=jLattice[j];
         z=C(u,v);
         r=f(z); // transformation C to R
         (*this)[i][j]=r;
      }
   }
}

template<class T>
void Image<T>::
insertPattern(R xL, R yL, R dx, R dy, const F<C,T>& f, const F<C,B>& g)
{
   const Z nTacit=4;
      // controls frequency of messages in verbose mode

   cpmassert(dx>=0.,"");
   cpmassert(dy>0.,"");

   R xU=xL+dx;
   R yU=yL+dy;

   setAspRat(dx/dy);
      // aspect ration of image should be that of the object space
      // area of interest. Then an orthogonal line system in the
      // complex plane will be shown as an orthogonal line grid if in
      // Frame::show(GrayImage a, R_Vector par, Z method, Z faithful )
      // the last parameter is set to 1.
   Z m=Base::dim1(), n=Base::dim2();

   Z i,j,i_;
   R u,v;
   V<R> iLattice=ImagingTools::pixelCenters(yL,yU,m);
      // matrix index i varies in y direction
   V<R> jLattice=ImagingTools::pixelCenters(xL,xU,n);
      // matrix index j varies in x direction

   C z,w;
   T r;
   Z lineCounter=0;
   R mInv=1./m;
   B valid;

   for (i=1;i<=m;i++){
      i_=m-i+1;
      v=iLattice[i_];
      if (verbose==1)
      {
         lineCounter++;
         if (lineCounter==nTacit){
            lineCounter=0;
            R fr=R(i)*mInv*100;
            cout<<endl<<"Image "<<fr<<" % completed";
         }
      }

      for (j=1;j<=n;j++){
         u=jLattice[j];
         z=C(u,v);
         valid=g(z);
         if (valid){   // thus only those parts are
            r=f(z); // transformation C to R
            (*this)[i][j]=r;
         }
      }
   }
}

template<class T>
void Image<T>::mark(Frame& fr, const Response& rsp, Z method,
   Z faithful)const
{
   Z mL=2;
   Word loc("Image<>::show(Frame,R2,Z,Z,Z)");
   CPM_MA
   Z i,j,m=Base::dim1(),n=Base::dim2();
   R pI, qI, width=1., hight=1.;
   if (faithful){ // we assume square pixels on screen
      R alpha_=getAspRat(); // aspect ratio of image
      R alpha=fr.getAspRat(); // aspect ratio of frame
      if (alpha_ > alpha){ // image has to be placed full width
         width=1.;
         hight=alpha/alpha_;
      }
      else{ // image has to be placed full hight
         width=alpha_/alpha;
         hight=1.;
      }
   }

   V<R> lattice1=ImagingTools::pixelCenters(0.,width,n);
   V<R> lattice2=ImagingTools::pixelCenters(0.,hight,m);

   Z2 index;
   R2 p;
   Z left,right,lower,upper;
   for (i=1;i<=fr.getLin();i++){
      index.x1=i;
      for (j=1;j<=fr.getCol();j++){
         index.x2=j;
         p=fr.getxFyF(index); // point in frame system
         if (p.x2>hight || p.x1>width) continue;
         ImagingTools::find(lattice1, p.x1, pI, left, right);
         ImagingTools::find(lattice2, p.x2, qI, lower, upper);
            // left ... upper refer to discretized xF, yF
         lower=1+m-lower;
         upper=1+m-upper;
            // now left ... upper refer to the matrix index grid
   // pixel values are converted to R for (optional) interploation
   // and input to rsp
         R a11=getPix(upper,left);
         R a12=getPix(upper,right);
         R a21=getPix(lower,left);
         R a22=getPix(lower,right);
         R val=ImagingTools::int2(a11,a12,a21,a22,pI,qI,method);
         Color c=rsp.color(val);
         fr.put(index,c);
      }
   }
   CPM_MZ
}

namespace{

Z multPix(Z n, R x)
{
   Z res=cpmrnd(x*n);
   if (res==0) return 1;
   if (res>n) return n;
   return res;
}

}

template<class T>
R2 Image<T>::minMax(R xL, R xU, R yL, R yU)const
   // Let res be the return value. Then res.x is the minimum
   // and res.x2 the maximum of the pixel values of *this
{
   const R veryLarge=1e30;
   R max=-veryLarge, min=veryLarge, a;
   cpmassert(xL>=0 && xU<=1,"");
   cpmassert(yL>=0 && yU<=1,"");
   cpmassert(xL<xU,"");
   cpmassert(yL<yU,"");
   Z m=getLin();
   Z n=getCol();
   Z j1=multPix(n,xL);
   Z j2=multPix(n,xU);
   Z i1=multPix(m,R(1)-yU);
   Z i2=multPix(m,R(1)-yL);
   Z i,j;
   for (i=i1;i<=i2;i++) for (j=j1;j<=j2;j++){
      a=getPix(i,j);
      if (a>max) max=a;
      if (a<min) min=a;
   }
   R2 res(min,max);
   return res;
}

template<class T>
R_Vector Image<T>::statistics(R xL, R xU, R yL, R yU)const
{
   const R veryLarge=1e30;
   R max=-veryLarge, min=veryLarge, a;
   cpmassert(xL>=0 && xU<=1,"");
   cpmassert(yL>=0 && yU<=1,"");
   cpmassert(xL<xU,"");
   cpmassert(yL<yU,"");
   Z m=getLin();
   Z n=getCol();
   Z j1=multPix(n,xL);
   Z j2=multPix(n,xU);
   Z i1=multPix(m,R(1)-yU);
   Z i2=multPix(m,R(1)-yL);
   Z i,j,count=0;
   R sum=0;
   for (i=i1;i<=i2;i++) for (j=j1;j<=j2;j++){
      a=getPix(i,j);
      sum+=(R)a;
      count++;
      if (a>max) max=a;
      if (a<min) min=a;
   }
   cpmassert(count>0,"");
   R mean=sum/count;
   R_Vector res(3);
   res[1]=min;
   res[2]=mean;
   res[3]=max;
   return res;
}

template<class T>
Image<T> Image<T>::cut(R xL, R xU, R yL, R yU)const
{
   if (verbose) cpmmessage("Image<>::cut(R,R,R,R) started");
   cpmassert(xL>=0 && xU<=1,"");
   cpmassert(yL>=0 && yU<=1,"");
   cpmassert(xL<xU,"");
   cpmassert(yL<yU,"");
   Z m=getLin();
   Z n=getCol();
   Z j1=multPix(n,xL);
   Z j2=multPix(n,xU);
   Z i1=multPix(m,R(1)-yU);
   Z i2=multPix(m,R(1)-yL);
   Z mc=1+i2-i1;
   Z nc=1+j2-j1;
   Z si=i1-1;
   Z sj=j1-1;

   Image<T> res(mc,nc);
   Z i,j;
   for (i=1;i<=mc;i++){
      for (j=1;j<=nc;j++){
         res.putPix(i,j,getPix(i+si,j+sj));
      }
   }
   if (verbose) cpmmessage("Image<>::cut(R,R,R,R) done");
   return res;
}

template<class T>
Image<T> Image<T>::cut(Z2 cLL, Z2 cUR)const
{
   Z2 cLLVVStyle(cLL[1],cLL[0]);
   Z2 cURVVStyle(cUR[1],cUR[0]);

                           //               cUR
   Z i1=cURVVStyle[0];     //   (i1,j1)      (i1,j2)
                           //                          matrix indexing
                           //   (i2,j1)      (i2,j2)
   Z i2=cLLVVStyle[0];     //  cLL

   Z j1=cLLVVStyle[1];
   Z j2=cURVVStyle[1];

   CpmRootX::order(i1,i2);
   CpmRootX::order(j1,j2);
   cpmassert(i1<i2 && j1<j2,""); // for utmost security
   Z m=getLin();
   Z n=getCol();
   cpmassert(i2<=m && j2<=n,"");
      // to catch inconsitent input of cLL, cUR

   Z mc=1+i2-i1;
   Z nc=1+j2-j1;
   Z si=i1-1;
   Z sj=j1-1;
   if (verbose){
      Word mess="Image<>::cut(Z2, Z2) creating";
      Word mess1=" m="+CpmRoot::toWord(mc)+" n="+CpmRoot::toWord(nc);
      mess=mess+mess1;
      cpmmessage(mess);
   }
   Image<T> res(mc,nc);
   Z i,j;
   for (i=1;i<=mc;i++){
      for (j=1;j<=nc;j++){
         res.putPix(i,j,getPix(i+si,j+sj));
            // notice mc+si=i2, nc+sj=j2, thus i2<=m and j2<=n
            // is a guarantee that the index is never out of range
      }
   }
   return res;
}

template<class T>
Image<T> Image<T>::cut(Z i1, Z i2, Z j1, Z j2)const
{
   CpmRootX::order(i1,i2);
   CpmRootX::order(j1,j2);
   Z m=i2-i1+1;
   Z n=j2-j1+1;
   Z i,j;
   Image<T> res(m,n);
   for (i=1;i<m;i++){
      for (j=1;j<=n;j++){
         T val=getPix(i+i1-1,j+j1-1);
         res.putPix(i,j,val);
      }
   }
   return res;
}

template<class T>
Image<T> Image<T>::clip(R clipL, R clipU )const
{
   R vMax=minMax().get2();
   T mi=clipL*vMax;
   T ma=clipU*vMax;
   if (verbose) cout<<endl<<"lower clip level = "<<mi;
   if (verbose) cout<<endl<<"upper clip level = "<<ma;
   Image<T> res=*this;
   Z m=getLin();
   Z n=getCol();
   Z i,j;
   T val;
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         val=getPix(i,j);
         if (val<mi) res.putPix(i,j,mi);
         if (val>ma) res.putPix(i,j,ma);
      }
   }
   return res;
}

template<class T>
template<class Y>
Image<Y> Image<T>::transformToneScale(F<T,Y> const& f)const
      // returns a new image that takes pixel value f(val)
      // where the original image had pixel value val.
{
   Z m=getLin();
   Z n=getCol();
   Z i,j;
   T val;
   Image<Y> res(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         val=(*this)[i][j];
         res[i][j]=f(val);
      }
   }
   return res;
}

template<class T>
Image<R> Image<T>::diff(Image<T> const& im)const
{
   Word loc("Image<T>::diff(Image<T>)");
   Z m=getLin();
   Z n=getCol();
   cpmassert(m==im.getLin(),loc);
   cpmassert(n==im.getCol(),loc);
   Z i,j;
   Image<R> res(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         R tij=(R)(*this)[i][j], imij=(R)im[i][j];
         res[i][j]=cpmabs(tij-imij);
      }
   }
   return res;
}

template<class T>
Image<R> Image<T>::operator*(Image<T> const& im)const
{
   Word loc("Image<T>::operator*(Image<T>)");
   Z m=getLin();
   Z n=getCol();
   cpmassert(m==im.getLin(),loc);
   cpmassert(n==im.getCol(),loc);
   Z i,j;
   Image<R> res(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         R tij=(R)(*this)[i][j], imij=(R)im[i][j];
         res[i][j]=tij*imij;
      }
   }
   return res;
}

template<class T>
Image<R> Image<T>::operator+(Image<T> const& im)const
{
   Word loc("Image<T>::operator+(Image<T>)");
   Z m=getLin();
   Z n=getCol();
   cpmassert(m==im.getLin(),loc);
   cpmassert(n==im.getCol(),loc);
   Z i,j;
   Image<R> res(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         R tij=(R)(*this)[i][j], imij=(R)im[i][j];
         res[i][j]=tij+imij;
      }
   }
   return res;
}

template<class T>
Image<R> Image<T>::operator-(Image<T> const& im)const
{
   Word loc("Image<T>::operator-(Image<T>)");
   Z m=getLin();
   Z n=getCol();
   cpmassert(m==im.getLin(),loc);
   cpmassert(n==im.getCol(),loc);
   Z i,j;
   Image<R> res(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         R tij=(R)(*this)[i][j], imij=(R)im[i][j];
         res[i][j]=tij-imij;
      }
   }
   return res;
}

template<class T>
T Image<T>::getLocalMean(Z i, Z j, const Mask& a)const
{
   T res=T();
   Z p,d=a.dim();
   for (p=1;p<=d;p++) res+=(getPix(i+a.di(p),j+a.dj(p))*a.w(p));
      // no out of range problem due to getPix's robustness
      // assumes  T+=T and T*R, satisfied e.g. for T=C
   return res;
}

template<class T>
R Image<T>::locVar(Z i, Z j)const
{
   R e2=R(getPix(i,j))*2;
   R a,b,c,d,f,g,h,k;
   a=(R)getPix(i-1,j-1);
   b=(R)getPix(i-1,j);
   c=(R)getPix(i-1,j+1);
   d=(R)getPix(i,j);
   f=(R)getPix(i,j+1);
   g=(R)getPix(i+1,j-1);
   h=(R)getPix(i+1,j);
   k=(R)getPix(i+1,j+1);
   R_Vector res(4);
   res[1]=e2-a-k;
   res[2]=e2-b-h;
   res[3]=e2-c-g;
   res[4]=e2-d-f;
   return res.supAbs();
}

template<class T>
Image<R> Image<T>::locVar(bool useLocVar)const
{
   if (useLocVar){
      Z m=getLin(), n=getCol();
      Image<R> res(m,n);
      R prfrac=1./m;
      for (Z i=1;i<=m;i++){
         cpmprogress("Image<T>::locVar(bool)-loop",R(i)*prfrac);
         for (Z j=1;j<=n;j++){
            res[i][j]=locVar(i,j);
         }
      }
      return res;
   }
   else return lap(true,2); // new
}

template<class T>
Image<T> Image<T>::smooth(Z n)const
{
   R_Vector rv{0.25,0.5,0.25};// linear binomial kernel
   if (n==1)
      return conv(rv,rv);
   else{
      Image<T> res=*this;
      for (Z i=1;i<=n;++i){
         res=res.conv(rv,rv);
      }
      return res;
   }
}

template <class T>
R fPos(T const& t){ return cpmabs((R)t);}

template <class T>
R fPlain(T const& t){ return (R)t;}

template<class T>
Image<R> Image<T>::lap(bool makePositive, Z p)const
{
   R_Vector rv;
   if (p==1){
      rv=R_Vector{1,-2,1};
      rv.b_(-1);
   }
   else {
      rv=R_Vector{2,2,-2,-4,-2,2,2};
      rv.b_(-3);
   }
    // linear Laplace kernel
   Image<T> res=conv(rv,rv);
   if (makePositive){
      return res.transformToneScale(F<T,R>(fPos));
   }
   else{
      return res.transformToneScale(F<T,R>(fPlain));
   }
}

template<class T>
Image<T> Image<T>::conv(const Mask& a)const
{
   cpmmessage( "Image<>::conv(const Mask& ) started");
   if (a.isVoid()) {
      cpmmessage("Image<>::conv(const Mask& ) trivially done");
      return *this;
   }
   Z m=getLin(), n=getCol();
   Z i, j;
   Image<T> res(m,n);
   R prfrac=1./m;
   for (i=1;i<=m;i++){
      cpmprogress("Image<T>::conv()-loop",R(i)*prfrac);
      for (j=1;j<=n;j++){
         res[i][j]=getLocalMean(i,j,a);
      }
   }
   cpmmessage( "Image<>::conv(const Mask& ) done");
   return res;
}


template<class T>
Image<T> Image<T>::conv(const R_Matrix& a, Z si, Z sj)const
{
   cpmmessage( "Image<>::conv(const R_Matrix& , Z , Z) started");

   Z ma=a.dim1(),na=a.dim2(), ia, ja;
   if (ma==0 && na==0){
      cpmmessage( "Image<>::conv(const R_Matrix&,Z,Z) trivially done");
      return *this;
   }
   Z m=getLin(), n=getCol(), i, j;
   Z ci=1+ma/2, cj=1+na/2;
   Z i0=ci+si, j0=cj+sj;
   T val=T();
   Image<T> res(m,n);
     R prfrac=1./m;
   for (i=1;i<=m;i++){
      cpmprogress("Image<T>::conv()-loop",R(i)*prfrac);
      for (j=1;j<=n;j++){
         val=0.;
         for (ia=1;ia<=ma;ia++){
            for (ja=1;ja<=na;ja++){
               val+=getPix(i+ia-i0,j+ja-j0)*a[ia][ja];
                  // if ma and na are odd and si=sj=0, the range should be
                  // i-ma/2, i+ma/2
                  // j-na/2, j+ma/2
                  // Thus i+1+ti=i-ma/2, thus ti=-ma/2-1
                  // Correspondingly tj=-na/2-1. For even ma or na this
                  // is taken over as definition. Formulae 1 and 2 are
                  // definitions of si,sj
            }
         }
         res.putPix(i,j,val);
      }
   }
   cpmmessage( "Image<>::conv(const R_Matrix& , Z , Z) done");
   return res;
}

template<class T>
Image<T> Image<T>::conv(R_Vector const& a, R_Vector const& b)const
{
   Z mL=1;
   static Word loc("Image<>::conv(R_Vector,R_Vector)");
   CPM_MA
   Z m=getLin(),n=getCol(),i,j,p;
   T val=T();
   Image<T> res1(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         val=0.;
         for (p=a.b();p<=a.e();++p){
            val+=getPix(i+p,j)*a[p];
         }
         res1.putPix(i,j,val);
      }
   }
   Image<T> res2(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         val=0.;
         for (p=b.b();p<=b.e();++p){
            val+=res1.getPix(i,j+p)*b[p];
         }
         res2.putPix(i,j,val);
      }
   }
   CPM_MZ
   return res2;
}

template<class T>
Image<T>& Image<T>::operator *= (R r)
{
   Z m=getLin(), n=getCol(), i, j;
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         putPix(i,j,getPix(i,j)*r);
      }
   }
   return *this;
}

template<class T>
Image<T>& Image<T>::operator += (R r)
{
   Z m=getLin(), n=getCol(), i, j;
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         putPix(i,j,getPix(i,j)+r);
      }
   }
   return *this;
}

template<class T>
void Image<T>::frame(const T& t)
{
   Z m=getLin(), n=getCol(), i, j;
   i=1;
   for (j=1;j<=n;j++){
      putPix(i,j,t);
   }
   i=m;
   for (j=1;j<=n;j++){
      putPix(i,j,t);
   }
   j=1;
   for (i=1;i<=m;i++){
      putPix(i,j,t);
   }
   j=n;
   for (i=1;i<=m;i++){
      putPix(i,j,t);
   }
}

template<class T>
Image<T> Image<T>::fftFilter(const F<C,C>& resp)const
{
   using namespace CpmFourier;
   Z i,j,m=Base::dim1(),n=Base::dim2();
   VV<C> mat(m,n);
   for (i=1;i<=m;++i) for (j=1;j<=n;++j){
      R vij=(*this)[i][j];
      vij=cpmabs(vij);
         // image should be positive so cpmabs is only
         // a defensive provision
      mat[i][j]=C(cpmsqrt(vij));
         // amplitude is the square-root of the intensity
   }
   SD2 x(mat,true);
      // considering this as distribution in space domain
   Z mx=x.dim1();
   Z nx=x.dim2();
   R mI=1./(mx-1);
   R nI=1./(nx-1);
   FD2 y=x.fft();
      // going to frequency domain
   for (i=1;i<=mx;++i){
      R yc=R(i-1)*mI-0.5;
         // yc from -1/2 to 1/2 for i 1 to mx
      for (j=1;j<=nx;++j){
         R xc=R(j-1)*nI-0.5;
            // xc from -1/2 to 1/2 for i 1 to nx
         C arg(xc,yc);
         C fac=resp(arg);
            // multiplicative modification
         y(i,j)*=fac;
      }
   }
   SD2 xt=y.fft();
      // back to space domain amplitude
   Image<T> res(mx,nx);
   for (i=1;i<=mx;++i) for (j=1;j<=nx;++j){
      R rij=xt(i,j).absSqr();
      res[i][j]=rij;
         // back to intensity as square of abs
         // of the amplitude
   }
   return res;
}

template<class T>
Image<T> Image<T>::fourierFilter(F<R2,R> const& resp)const
{
   Z mL=2;
   static Word loc("Image<>::fourierFilter(F<R2,R>)");
   CPM_MA
   VV<C> a=fourierTransform();
// filtering the Fourier image
   Z m2=a.dim1(),n2=a.dim2();
   R fac,xi,xj;
   Z i,j,ic,jc;
   Z mNy=m2/2+1;
   Z nNy=n2/2+1;
   R ci=R(2)/m2, cj=R(2)/n2;
   for (i=1;i<=m2;i++){
      ic=(i<=mNy ? i : 2+m2-i);
         // this ensures that the components belonging to frequency
         // f will be multiplied by the same factor as those of
         // frequency -f
      xi=R(ic-1)*ci;
         // xi varies in [0,1]
      for (j=1;j<=n2;j++){
         jc=(j<=nNy ? j : 2+n2-j);  // as for i
         xj=R(jc-1)*cj;
         fac=resp(R2(xi,xj));
         a(i,j)*=fac;
      }
   }
   CPM_MZ
   return Image<T>(a); // implies Fourier transformation
}

template<class T>
VV<C> Image<T>::fourierTransform()const
{
   Z mL=1;
   static Word loc("Image<>::fourierTransform()");
   CPM_MA
   Z m=getLin(), n=getCol(), i, j;
   cpmassert(m>=2 && n>=2,loc);
   Z m2=2; while (m2<m) m2*=2;
   Z n2=2; while (n2<n) n2*=2;
   C zero;
   VV<C> ori(m2,n2,zero); // sufficiently large for fourier purposes
      // and initialized by zero
   for (i=1;i<=m;i++) for (j=1;j<=n;j++){
      R vij=getPix(i,j);
      ori(i,j)=C(vij);
   }
   CPM_MZ
   return fft(ori,1);
}

template<class T>
Image<T> Image<T>::cut(const Rect& rect)const
{
   R xL=rect.getxL();
   R xU=rect.getxU();
   R yL=rect.getyL();
   R yU=rect.getyU();
   Z m=getLin();
   Z n=getCol();
   Roi r(m,n,xL,xU,yL,yU);
   Image<T> res(r.m(),r.n());
   Z i,j;
   for (i=1;i<=r.m();i++){
      for (j=1;j<=r.n();j++){
         res[i][j]=getPix(r.i(i),r.j(j));
      }
   }
   return res;
}

template<class T>
R Image<T>::norm1(void)const
   // sum over the absolute values over all pixel values, divided by
   // number of pixels
{
   Z m=getLin();
   Z n=getCol();
   Z i,j;
   R sum=0.;   // even if T is a integer type !
   R val;
   if (verbose){ cout<<endl<<"MultImage<>::norm1 started";}
   for (i=1;i<=m;i++){
      if (verbose>9){ cout<<endl<<"treating line "<<i;}
      for (j=1;j<=n;j++){
         val=getPix(i,j);
         R valr=CpmRoot::absT<T>(val);
         sum+=valr;
      }
   }
   R nPix=1.*m*n;
   cpmassert(nPix>0,"");
   sum/=nPix;
   return sum;
}

template<class T>
Histogram Image<T>::histogram(Z nBins, R valMin, R valMax )const

{
   Z m=getLin();
   Z n=getCol();
   Z i,j;
   R val;
   Histogram b(1,nBins,valMin,valMax );
   for(i=1;i<=m;i++){
      for(j=1;j<=n;j++){
         val=getPix(i,j);
         b.put(1,val);
      }
   }
   return b;
}

template <class T>
Img24 Image<T>::toImg24()const
{
   Z m=getLin();
   Z n=getCol();
   Z i,j;
   Img24 img(n,m);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         R val=getPix(i,j);
         rgb c(val,val,val);
         img.putPel(j-1,i-1,c);
      }
   }
   return img;
}

template <class T>
Img24 Image<T>::halfTone(R const& gamma)const
{
   rgb c0(0,0,0);
   rgb c1(255,255,255);
   R2 mM=minMax();
   R m0=mM[1];
   R m1=mM[2];
   R mInv=cpminv(m1-m0);
   Z m=getLin();
   Z n=getCol();
   Z i,j;
   Z k=137;
   Img24 img(n,m);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         R p=randomR(k++);
         R val=getPix(i,j);
         R vNorm=cpmpow((val-m0)*mInv,gamma);
         rgb c= p>vNorm ? c0 : c1;
            // If vNorm is large, the probability for p>vNorm is small,
            // so the probability for the dark value c0 is small. So the
            // image tends to be bright here.
         img.putPel(j-1,i-1,c);
      }
   }
   return img;
}

} // namespace

#endif
