//? cpmimagingtools.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_IMAGINGTOOLS_H_
#define CPM_IMAGINGTOOLS_H_
/*
   Description: Defines some functions to be used in image classes
*/
#include <cpmrvector.h>
#include <cpmangle.h>
#include <cpmviewport.h>
#include <cpmalgorithms.h>

namespace CpmImaging{

   using CpmRoot::Z;
   using CpmRoot::Word;
   using CpmRoot::R;
   using CpmRoot::C;

   using CpmRootX::Z1;
   using CpmRootX::R1;
   using CpmRootX::B;

   using CpmLinAlg::R2;
   using CpmLinAlg::R3;

   using CpmArrays::V;
   using CpmArrays::Va;
   using CpmArrays::R_Vector;

   using CpmGraphics::rgb;
   using CpmGraphics::ColRef;

   using CpmGeo::Iv;
   using CpmGeo::Angle;

   using CpmAlgorithms::CyclicAlarm;
   using CpmAlgorithms::NonCyclicAlarm;

   extern Z stereoLeftR;
   extern Z stereoLeftG;
   extern Z stereoLeftB;

   extern Z stereoRightR;
   extern Z stereoRightG;
   extern Z stereoRightB;

//////////////////// enum Palette ////////////////////////////////////////

// There is a (non-explicit) constructor Color(Palette) so Palette-typed
// quantities may -via user-defined conversion- go automatically into
// Color objects. This is very convenient in definitions such as that of
// 'rainBow' below. It is dangerous however to use Palette-typed
// quantities as arguments of function. For instance the class Camera
// has many member functions to 'mark' objects on video memory. Two for
// marking Spc-objects are:
//    void Camera::mark(Spc,Color);
//    void Camera::mark(Spc,R);
// A call ca.mark(p,BLUE);
// interprets BLUE as a double (C++ preferred conversion from integer to
// double) instead of considering the user-defined conversion to Color.
// The call ca.mark(p,Color(BLUE)); does what was intented. Even clearer
// is:
// Spc p(...); Color cb(BLUE); ca.mark(p,cb);
   enum Palette {
    BLACK,            // dark colors
    BLUE,
    GREEN,
    CYAN,
    RED,
    MAGENTA,
    BROWN,
    FAINTGRAY,
    FAINTYELLOW,
    LIGHTGRAY,
    DARKGRAY,           // light colors
    LIGHTBLUE,
    LIGHTGREEN,
    LIGHTCYAN,
    LIGHTRED,
    LIGHTMAGENTA,
    YELLOW,
    ORANGE,
    WHITE,
    DARKBLUE,    // dark colors
    DARKGREEN,
    DARKCYAN,
    DARKRED,
    DARKMAGENTA,
    INVISIBLE,   // special colors
      // If the color of a body is set 'invisible' this does not mean
      // that its screen image can not be seen. Most of my graphics
      // functions Camera::mark(...) then make an anaglyphic stereo
      // representation in 3D and the corresponding flat line
      // representation in 2D. In a sense, then the color of the body is
      // in fact invisible; the body is 'mere form'
    STEREOBLACK, // background for stereo (anaglyph) images.
      // So far no need has been seen to use anything different from pure
      // black.
    STEREOLEFT, // pure red, brightness set such that seen through the red
    //
      // (left-eye) window of our anaglyph viewing device it is equally
      // bright as STEREOCYAN seen through the blue or green (right-eye)
      // window. All my various anaglyph viewers are red-green 'glasses'
      // When seen through the red filter, the green color channel of
      // my LCD-screen is sufficiently bright to interfere disturbingly
      // with the red image. So it is better to use as color for the right
      // eye the blue channel of the LCD-screen.
      // Since this blue light is nevertheless viewed throuph the green
      // filter of the glasses, it is seen not as bright as red light
      // through the red filter. So this system works best if
      // STEREOLEFT = (190,0,0) and STEREORIGHT = (0,0,255)(thus pure
      // blue).
      // This shows that it may be useful to be able to set the colors
      // STEREOLEFT and STERORIGHT from files
      // cpmsystemdependencies.h and cpmconfig.ini

    STEREORIGHT, // see STEREOLEFT
    TRANSPARENT_ ,// without _ it clashes with a MS-name
    STEREO,
      // same as INVISIBLE
    WRITECOLOR,
    GRIDCOLOR
};

///////////////////// class Color ///////////////////////////////////////

   class Color: public rgb{//data of CpmGraphics::rgb + gamma, arithmetics
      static R gamma;
         // comes into action only in converting C to Color

      void make(Palette c);
         // sets r,g,b according to a Palette entry
         // (Palette is an enum !)
         // actually this function defines what is the operational
         // meaning of a Palette entry

      static R valMax_;
   public:
      typedef Color Type;

      typedef R ScalarType;
         // for arithmetics defined via CPM_PRODUCT_SCALAR_C

      virtual Color* clone()const{ return new Color(*this);}
         // allows Color to be inserted as X in the polymorphic array
         // template Vp<X>
      CPM_IO_V
      CPM_ORDER
      CPM_PRODUCT_SCALAR_C
      CPM_SUM_C
      Word nameOf()const{ return Word("Color");}

      explicit Color(rgb const& c):rgb(c){}
      Color():rgb(){}
      Color(Z r_, Z g_, Z b_):rgb(r_,g_,b_){}
      explicit Color(V<Z> const& v):rgb(v(1),v(2),v(3)){}
         // This is usefull since V<Z> is among the data types that can be
         // exchanged with class Record on which my ini-files are based.
         // The components of v get cut down to the range [0,255]

      Color(R r_, R g_, R b_):rgb(cpmrnd(r_),cpmrnd(g_),cpmrnd(b_)){}
         // Here the usefull values of r_,g_ b_ range from -0.5 to 255.5,
         // not from 0 to 1 as in Color(R3 const& y).

      Color(Palette c){ make(c);}
         // So a Palette-value will directly be converted into a Color;
         // make(c) does not call a normalizing constructor of rgb,
         // so it may set negative values as done for c=INVISIBLE.
         // Notice that C++ prefers converting Palette (as an integer
         // type) into R over converting to Color in function argument
         // matching. So instead of giving an Color-expecting argument
         // slot the input c, it is safer to employ Color(c).

      Color(Word w, R br):rgb(cpmrnd(br),cpmrnd(br),cpmrnd(br)){}
         // first argument 'dummy' second gives the brightness
         // of gray

      explicit Color(R3 const& y);
         // construction from real triples
         // Usefull values satisfy 0<=y[i]<=1, i=1,2,3.
         // Other values get squeezed to 0 or 255.
         // This constructor is usefull in order to construct color
         // representations of arithmetical constructs.

      R3 toR3()const;
         //: to R3
         // For each Color c, we have Color(c.toR3())==c.
         // The converse property can not hold: For arbitrary
         // y in R3 such that 0<=y[i]<=1, i=1,2,3 we are not sure
         // that Color(y).toR3() == y since the transition from
         // y to Color(y) waists bits.

      explicit Color(Z colNumber);
         // so colors can be initialized also by numbers,
         // numbers <=0 create Color(INVISIBLE). Large numbers repeat
         // periodically the basic Palette (the colors from BLACK to
         // WHITE). Thus:
         // Color(0)=INVISIBLE,
         // Color(1)=BLACK, Color(2)=BLUE, Color(3)=GREEN,
         // Color(4)=CYAN, Color(5)=RED
         // Color(16)=ORANGE,
         // Color(17)=WHITE, Color(18)=BLACK,...
         // No automatic conversion.
         // Notice that there is another way of generating colors from
         // integers:
         // i|--> CpmImaging::rainBow.cyc(i);
         // For these repeated 8 colors the intended order is evident,
         // so it especially useful for displaying families of curves

      static Z toZ(Palette p);
         // Defined such that for Palette p we have
         // Color(p)==Color(Color::toZ(p)) (for all p
         // for which there is Z i such that Color(i)==Color(p)
         // Thus INVISIBLE, BLACK, BLUE, ... WHITE

      static Palette toPal(Z i);
         // Returns the Palette instance that belongs to Color(i)

      explicit Color(const C& z, Iv minMax=Iv(0,1));
         // coding complex numbers as colors, the absolute value controls
         // luminance and the phase angle controls the hue. We use
         // maximally saturated colors only and the angle is 0 for red,
         // 120 for green, 240 for blue. Thus the order is in the
         // direction of increasing polar angle (0->360):
         // red,yellow,green,cyan,blue,purple,red
         // Notice: red   == real positive
         //         cyan  == real negative
         //         green == imaginary positive
         //         blue  == imaginary negative
         // other algorithms that I tested made negative real and
         // negative imaginary appear rather similar.
         // The absolute value of z gets transformed into brightness
         // via ImagingTools::response according to:
         // response(minMax.inf(), minMax.sup(), 0, 255, gamma,
         // z.abs());
         // For gamma=0 one shows only the phase of z coded as color
         // and the absolute value of z does not matter. This is
         // sometimes usefull.

      explicit Color(R r, Iv minMax=Iv(0,1));
         // coding real numbers as a color such that two pure colors get
         // connected over the white point.
         // Presently: (0,1/2,1)|--> red,white,blue
         // This just fits my EP-application

      explicit Color(ColRef cr):rgb(cr){}

      Color operator~()const{ return Color(~rgb(r,g,b)); }
         // returns the color complementary to *this
      static Color hue(R const& h, bool finalMagenta=true);
         // h ranges from 0 to 1 and gives colors red orange yellow green
         // cyan blue magenta. If the last argument is set to false we
         // close the color circle to red. So the ful rainbow results
         // for finalMagenta=true.
      Color scale(R const& s)const{ return Color(s*r,s*g,s*b);}
         // result gets cut to available range
      Color addWhite(R const& w)const
         // w should vary in [0,1]. 1 means that maximum ammount of
         // white will be added.
      {
         Z gap=ColRef::upperLimit-max();
         Z add=cpmtoz(w*gap);
         return Color(r+add,g+add,b+add);
      }

      Color brighten(R const& b=1.)const;
         // multiplies all channels by the same factor in a way that at
         // least one channel reaches ColRef::upperLimit if b=1,
         // and proportionaly by a smaller factor for 0<=b<1.

      Z Red()const{return r<0 ? 0 : r;}
      Z Green()const{return g<0 ? 0 : g;}
      Z Blue()const{return b<0 ? 0 : b;}
      Z max()const{ return CpmRootX::sup<Z>(r,g,b);}

      R red()const{return r<0 ? R(0) : R(r);}
      R green()const{return g<0 ? R(0) : R(g);}
      R blue()const{return b<0 ? R(0) : R(b);}

      bool visible()const{ return r>=0;}
         // notice that for Color(INVISIBLE) we have r=g=b=-1
      bool transparent()const{ return r==-137;}
      bool stereo()const{return r<0;}
      ColRef toColRef()const{ return ColRef(r,g,b);}
      static void setGamma(R gamma_){ gamma=gamma_;}
      static R getGamma(){ return gamma;}
      rgb toRGB()const{ return rgb(r,g,b);}
   };

   const CpmArrays::V<Color> rainBow{BLACK,LIGHTMAGENTA,BLUE,CYAN,
      GREEN,YELLOW,ORANGE,LIGHTRED,RED};

   const CpmArrays::V<Color> colors9{MAGENTA,LIGHTMAGENTA,
      BLUE,LIGHTBLUE,GREEN,LIGHTGREEN,RED,LIGHTRED,BLACK};
      // an ad hoc collection of 9 colors
      // Useful for visualization of C4-valued functions such as
      // Dirac electron wave functions.

// some dark colors which look good as backgrounds for graphics made of
// white or yellow dots or lines
   const CpmImaging::Color darkSky(Z(0),Z(0),Z(50));
      //: dark sky
   const CpmImaging::Color twiSky(Z(30),Z(0),Z(50));
      //: twilight sky
   const CpmImaging::Color morSky(Z(50),Z(60),Z(100));
      //: morning sky

/////////////////// class ImagingTools /////////////////////////////////

class ImagingTools{ // collection of helper functions

public:

// functions

static R fMono(R_Vector const& a);
   // Function for combining R G B to Gray

static R fMax(R_Vector const& a);
   // Function for combining all channels to a maximum value

static R fMin(R_Vector const& a);
   // Function for combining all channels to a maximum value

static R int2(R f11, R f12, R f21, R f22, R p, R q, Z method);
// int2 stands for interpolation in 2 dimensions.
// The returned value is a weighted meen of the fij. Notice
// p+p_=1, q+q_=1, which leads to p*q_+p_*q_+p*q+p_*q=1.
// For interpretation: p and q follow x and y axis.
// The corners of the rectangle in which the interpolation point is
// placed and to which the reduced coordinates p and q refer are as
// follows:
//
//      f11      f12
//
//      f21      f22
//
// Method =1 means bilinear interpolation and method=0 means
// pixelation. The latter is useful if the number of pixels is to be
// compared to expectations. The bilinear formula is
// f11*p_*q+f12*p*q+f21*p_*q_+f22*p*q_ which implies for
// p=q=0 obviously f21 ( namely the coefficient of p_*q_ ). Thus
// the indexes of f are matrix indexes and p and q refer to normally
// oriented x,y-axes (y from bottom to top )

static CpmArrays::V<R> pixelCenters(R a, R b, Z n);
// The interval [a,b] is divided into n subintervalls of equal length
// and the n centers of these subintervals are returned as a R_Vector
// in ascending order

static void find(const CpmArrays::V<R>& lattice, R x, R& p, Z& i, Z& j);
// After function call, i is the largest integer for which lattice[i]
// is defined and for which lattice[i]<=x.
// p=(x-lattice[i])/(lattice[i+1]-lattice[i]).
// It is assumed that lattice[j+1]-lattice[j]>0 and independent of j,
// whenever none of the indexes is out of range.
// Extension 95-3-3: for values outside the lattice points, linear extra-
// polation may be useful.

static R response(R x1, R x2, R y1, R y2, R gamma, R x);
// Auxiliar function for e.g. Image<T>::gamma(...).
// The function x |---> f(x):=response(...,x) has the following
// properties:
// f(x)=y1 for x<=x1
// f(x)=y2 for x>=x2
// For gamma>=0: f(x)= y1+(y2-y1)*((x-x1)/(x2-x1))^gamma
// For gamma<0: f(x)=alpha+beta*log(x) with alpha and beta
// determined such that f(x1)=y1, f(x2)=y2
// Also x1 is forced to be positive and x2 is reduced if too large
// for gamma<0.

static R random(Z i);
// returns randomly distributed numbers in the interval [0,1]
// as a function of i=1,2,3,....

static Z random(Z i, Z k);
// returns a randomly distributed numbers in {1,2,...k}
// as a function of i=1,2,3,....

// data
static Z verbose;
// verbose=0, no information outputted during calculations

};

////////////////////// class Response ////////////////////////////////////
/*
   considerably extended (and slightly modified) 2013-04-18
*/

class Response{ // versatile conversion from R to screen color
   // The intention is as follows: Given a R-valued function f on C or on
   // R2, by changing the function value f(z) into rp.color(f(z)) for a
   // suitably constructed instance rp of class Response one should obtain
   // a Color-image which shows the features of f clearly upon suitably
   // setting the many parameters which govern the behavior of Response.
   Iv rx_,ry_; // input range and output range
      // Assume Response r, then r(rx_)=ry_
   R1 gamma_; // exponent in a transformation from numerical values to
      // screen color (in a wide sense).
   Z1 mode_;
      // parameter that controls the strategy for transforming numerical
      // values in screen color (this may be black & white )
   B periodic_;
      // This is a switch that determines the working of the first stage
      // function f1_. New
   B bw_; //. black and white
      // This is a switch in the function color.
      // Notice that we use for R, Z, bool the class typed substitudes, so
      // that hey all are initialized even if not mentioned in any
      // constructor.
   static Z1 levels;
      // |levels| is the number of intervals into which the whole
      // color-spectrum in function color() is divided
      // by 'Fraunhofer lines'. Could also be a member, static for back-
      // compatibility. If levels is negative each such interval
      // contributes the half of a line at each of its ends (one full line
      // pro toto), if levels > 0 each interval has a single full line at
      // its center. If Response is used to visualize a real field in the
      // plane, the 'Fraunhofer lines' mark lines of equal field value and
      // are a very ehfficient analysis tool. They also make often
      // esthetically appealing images. Nwe is that now the sme mechanism
      // is implemented for true Colors and for gray images.
   R f1_(R const& x)const
      // This 'first stage function', for more uniform processing, sends
      // all input values x into the interval [0,1].
   {
      if (periodic_) return rx_.periodic(x); // periodic (saw tooth)
      else return rx_.ramp(x); // linear on rx_, constant outside
   }
   R levFac_( R const& p)const
      //: level factor
      // see levels for comments
   {
      R y=p*levels.x1*cpmpi;
      R q=levels<0 ? cpmcos(y):cpmsin(y);
      // 0<=q<=1
      q*=q;
      q*=q; // Points with q==1 remain so, all others move more to 0
         // so that the bright areas shrink.
      return R(1)-q; // Also in [0,1] but now the dark areas are narrow.
   }

   R trn_(R const& p)const{ return cpmpow(p,gamma_.x1);}
   //: transformation
   // transforms [0,1] into itself, via exponential function with exponent
   // gamma_.

   R res_(R const& x)const;
      // building block for the newly added mode_==7 path of function
      // Color color(R)
   Color col_(R const& x)const;
      // building block for the newly added mode_==7 path of function
      // Color color(R)
   typedef Response Type;
public:
// infrastructure for enabling I/O
// and polymorphic arrays of derived types
   CPM_IO_V
   CPM_NAM(Response)
   virtual Response* clone()const{ return new Response(*this);}
// constructors
   Response(){}
   Response(Iv rx, Iv ry=Iv(0,255), R gamma=1, Z mode=2):
      rx_(rx),ry_(ry),gamma_(gamma),mode_(mode){}
   Response(R xL, R xU):
      rx_(Iv(xL,xU)),ry_(Iv(0,255)),gamma_(1),mode_(Z(0)){}
// raison d'etre
   R operator()(const R& x)const;
      // Assume Response r, and  R x, then r(x) is the response of r
      // to x.
   Color color(R x)const;
      // for mode<=5 this returs Color(v,v,v) if v is the return value
      // of operator(), thus it returns a B/W color (unbunte Farbe).
      // For mode==6 it gives a 'spectral color' under the control of
      // rx and gamma. Colors correspond to x just as x would be a
      // photon energy lowest values correspond to red (also red numbers
      // in economy correspond to lowest, here negative, values of
      // earnings) highest values to violet. The whole range rx gets
      // subdivided into 10 subintervals by darker lines like the
      // Fraunhofer lines in the spectrum of solar light. Notice that
      // the spectral lines have sharp, pixeled, edge towards the
      // boundaries of rx, wheras the interior lines have diffuse edges.
      // mode_==7 is new. For this mode the new attribute bw_ allows us to
      // switch between gray images and color images.

// modifying functions
   void setRangeX(R xL, R xU){ rx_=Iv(xL,xU);}
   void setRangeY(R xL, R xU){ ry_=Iv(xL,xU);}
   void setGamma(R gamma, bool lean=false)
   { gamma_=gamma; if (!lean) mode_=0;}

   R getGamma()const{ return gamma_; }
   bool normalizeGamma()
      {bool b=gamma_>R1(0); if (!b) gamma_*=-1; return b;}

   void setMode(Z mode){ mode_=mode;}
   Z getMode()const{ return mode_;}
      // mode_ 0: response contolled by a traditional gamma
      // mode_ 1: response curve is arctan, gamma has no influence
      // mode_ 2: periodic response, gamma has no influence
      // mode_ 3: response curve is sin, gamma has no influence
      // mode_ 4: band pass response: for values outside rx we return 0
      // mode_ 5: asymmetric sine response, in between of 2 and 3
      // less aliasing than 2, but visibility of growth and decline as
      // for 2
      // mode_ 6: colormode; here the the new member-function color
      //    gives meaningful values, see function color() for the meaning
      //    of the colors. catchword: photon energy
      // mode_ 7: unites essentially all functionality of modes 1-6 in a
      //    single mode

   void setGammaResponse(){mode_=0;}
   void setArcTanResponse(){mode_=1;}
   void setPeriodicResponse(){ mode_=2;}
   void setSineResponse(){ mode_=3;}
   void setBandResponse(){ mode_=4;}
   void setAsySineResponse(){ mode_=5;}
   void setColorResponse(){ mode_=6;}
   void setLevels(Z lev){ levels=lev;}
   void setPeriodic(bool per){ periodic_=B(per);}
   void setBW(bool bw){ bw_=B(bw);}

   bool getColorResponse(){ return mode_>=6;}
   Iv getRangeX()const{ return rx_;}
   Iv getRangeY()const{ return ry_;}
};

///////////////////////////// class Roi //////////////////////////////////
// 97-5-18(?)

class Roi{  // region of interest

   // The following is the logic to copy the Roi-part of image *this into
   // a new image res

/*
      Z m=getLin();
      Z n=getCol();
      Roi r(m,n,xL,xU,yL,yU);
      Image<T> res(r.m(),r.n());
      Z i,j;
      for (i=1;i<=r.m();i++){
         for (j=1;j<=r.n();j++){
            res.putPix(i,j,getPix(r.i(i),r.j(j)));
         }
      }
*/
protected:
   Z mc,nc,si,sj;
public:
   Roi(Z m, Z n, R xL, R xU, R yL, R yU);
      // The part of the image (m=image.getCol(), n=image.getLin() )
      // which is described by the Roi has normal
      // coordinates (x,y) with respect to the original which satisfy
      //   xL <= x <= xU
      //   yL <= y <= yU
      // Normal coordinate system: origin lower-left, x-axis
      // to the right, y-axis vertically ascending. The upper-right
      // image corner has coordinates x=1, y=1.
   Roi(Z m, Z n, Iv ivx, Iv ivy);
      // same as previous constructor but with more convenient
      // arguments (ivx[1]=xL, ivx[2]=xU, ivy[1]=yL, ivy[2]=yU)
   Z m(void)const{ return mc;}
   Z n(void)const{ return nc;}
   Z i(Z i0)const{ return i0+si;}
   Z j(Z j0)const{ return j0+sj;}
};

////////////////////////////// class Rect ////////////////////////////////

class Rect{  // rectangular sub-area of the unit-square

protected:
   R xL,xU,yL,yU;
      // The points (x,y) of the rectangle satisfy
      // xL<=x<=xU, yL<=y<=yU
   typedef Rect Type;
public:
   CPM_IO
   CPM_NAM(Rect)
   Rect(R a=0, R b=1, R c=0, R d=1);
      // defines a sub-rectangle of the unit-square [0,1]x[0,1].
      // If the arguments are as intended, we have
      // 0<=a<b<=1
      // 0<=c<d<=1.
      // Then the sub-rectangle is simply [a,b]x[c,d]
      // If they are out of range or out of order, they will be brought
      // back in an obvious manner

   Rect(Iv iv1, Iv iv2);
      // 'modern form' defines a rectangle in terms of
      // intervals on the x-axis and the y-axis.
      // the section of iv1 with [0,1] defines [xL,xU]
      // and the section of iv2 with [0,1] defines [yL,yU]

   Rect(Z c1x, Z c1y, Z c2x, Z c2y, Z cx, Z cy);
      // defines the sub-rectangle of the unit-square [0,1]x[0,1]
      // which corresponds to graphics pixel locations for subarea
      // corners as can be read from PhotoStyler screen. Marking the
      // lower right corner of the whole image  with the pipette tool of
      // PhotoStyler (same with similar programs) we read the pixel
      // coordinates (cx,cy). Marking two corners c1 and c2 which span
      // the subarea of interest (i.e. to be described by the Rect under
      // construction) we get the pixel readings
      // (c1x, c1y)
      // (c2x, c2y)

   R getxL(void)const{ return xL;}
   R getxU(void)const{ return xU;}
   R getyL(void)const{ return yL;}
   R getyU(void)const{ return yU;}

   Rect sub(const Rect& rec)const;
      // returns the rectangle which is formed by taking rec of *this
      // in a way that *this is now considerd a unit square
};

////// class Mask for defining local mean values of pixels ///////////////
//97-9-12
// 2008-01-30 Convolutions of images may also be defined in a way that
// the process is given by a R_Matrix which is more direct than with
// Mask

class Mask{ // tool for defining convolution
   // tool for defining linear combinations of pixelvalues in the
   // neighborhood of a given pixel.
   // 'The data are the interface' in this case.
   // The constructor assures reasonable initialisation.
   // Since members are constant, no inconsitencies can be created.
   // Default constructor gives a mask a0 such that
   // ImageUM<R>::getLocalMean(i,j,a0) =  ImageUM<R>::getPix(i,j)
   // 97-9-15: now members not constant, since with constant members
   // no operator = is defined automatically. Returned to the
   // orthodox standpoint where access to data should be define by proper
   // memberfunctions and not by public members. This actually improves
   // the readability of formulas for ImageUM<>::getLocalMean(). This
   // functions completely explains the quantities entering the
   // definition of the present class.
public:
   CPM_NAM(Mask)
// constructors
   Mask(V<Z> const& diI, V<Z> const& djI, V<R> const& wI):
      di_(diI),dj_(djI),w_(wI)
   { cpmassert(diI.dim()==djI.dim()&&djI.dim()==wI.dim(),"Mask(...)");}

   Mask():di_(1,0),dj_(1,0),w_(1,1){}

   explicit Mask(Z n);
      // a uniform (constant) n by n mask. n is internally chosen as odd
      // if n was even (next higher odd value).

   Mask(R rMax, const CpmFunctions::F<R,R>& f);
      // a radially symmetric mask of total weight 1, where the
      // individual weights are proportion to f(r), where r is
      // the distance from the center. No contributions for r > rMax,
      // Distance from between pixel centers of horizontally or
      // vertically adjacent pixels is 1. (and sqrt(2) for diagonally
      // adjacent ones)
      // implementation did not work till 03-03-25

static Mask parabolic(R rMax);
   // static function as a 'named constructor'. returns the mask
   // Mask(rMax,f) where f(r):=1-(r/rMax)^2. For rMax=2 the
   // functionality looks very good

static Mask bin3();
   //: binomial 3 (by 3)
   // The smallest from the family of binomial masks which represented
   // as a matrix is
   // ((1,2,1),(2,4,2),(1,2,1))/16

// access functions
   Z dim(void)const{return di_.dim();}
   Z di(Z i)const{ return di_[i];}
   Z dj(Z i)const{ return dj_[i];}
   R w(Z i)const{ return w_[i];}
   Mask invert(void)const;
      // returns a mask which is 'reflected at the origin'
   R rms(void)const;
      // returns the root of the mean (with respect to w) of r square.
      // We do not assume that the components of w_ are positive and/or
      // normalized in some way. We take  always the absolute value of
      // these components and normalize with respect to these positive
      // quantities
   bool isVoid()const{ return (di_.dim()==0 && dj_.dim()==0);}

protected:
   CpmArrays::Va<Z> di_;
   CpmArrays::Va<Z> dj_;
   CpmArrays::Va<R> w_;
};

//////////////////////// struct ShowControl //////////////////////////////

struct ShowControl{ // collection of display control parameters
   // combining the 'historically' grown screen
   // display control parameters into a single data structure
   mutable CyclicAlarm cyalClear;
   mutable CyclicAlarm cyalView;
   mutable CyclicAlarm cyalImg;
   mutable NonCyclicAlarm nocyalImg;
   mutable Z1 countImg;
   Z1 offSet;
   Z1 maxNumberImages;
   R1 tWait;
   R1 autoFrameFactor;
   B boxSel;
   Iv ivx,ivy,ivz;
   V<Z> listColorize;
   V<Z> invisibleParticles;
      // particles wit an id in this list will not be marked on screen
      // (introduced 2006-01-02 for creating graphics for printing)
   Word modality;
      // More flexible control than using several B's.
   Word imgFile;
   ShowControl(){}
   Word nameOf()const{ return Word("ShowControl");}
};

} // namespace

#endif
