//? cpmimg24.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_IMG24_H_
#define CPM_IMG24_H_

/*
   Description: Declaration and partial definition of class
   Img24 which serves as memory for graphics windows and also
   allows to use all graphical functions for console applications
   to represent screen images as image files of *.ppm format.

*/

#include <cpmviewport.h>
#include <cpminterval.h>

namespace CpmGraphics {

   using CpmRoot::Z;
   using CpmRoot::L;
   using CpmRoot::R;
   using CpmRoot::Word;
   using CpmRootX::B;
   using CpmRootX::R1;
   using CpmLinAlg::R2;
   using CpmLinAlg::R3;
   using CpmGeo::Iv;
   using CpmArrays::VV;
   using CpmArrays::VVV;
   using CpmArrays::IvZ;

/////////////////////// enum ColSty //////////////////////////////////////

enum ColSty {SET_COLOR,ADD_COLOR,MIX_COLOR,STEREO};
   // color style
    // 2010-06-29 NEW_LINES lets any call to lineThin finally be
    // re-directed to the new function lineThinNew. Not yet mature.

//////////////////// class TriAng ////////////////////////////////////////

class TriAng{
// triangles in the plane, geometry only as needed for graphics
   // the plane is considered as R2
   // No restrictions concerning the points.
   // Lean class dealing only with aspects relevant to graphics

//independent quantities
   R2 p1_,p2_,p3_;
public:
   Word nameOf()const{ return Word("TriAng");}
      //: name of
   TriAng(R2 const& p1, R2 const& p2, R2 const& p3):
      p1_(p1),p2_(p2),p3_(p3){}
      // constructor from three points

   TriAng shf(R2 const& s)const{ return TriAng(p1_+s,p2_+s,p3_+s);}
      //: shift

   IvZ i1()const;
      //: i1
      // returns the smallest IvZ which contains the set
      // {p1_.x1, p2_.x1, p3_.x1}

   IvZ i2()const;
      //: i2
      // returns the smallest IvZ which contains the set
      // {p1_.x2, p2_.x2, p3_.x2}

   R area()const;
      //: area
      // this area is positive iff the rotation sense
      // p1 -> p2 -> p3 is positive (i.e. counter clock wise)

   bool out(R2 const& p)const;
      //: out
      // returns true iff p is outside the triangle *this
      // Source: see function barCor

   R3 barCor(R2 const& p)const;
      //: barycentric coordinates
      // returns the barycentric coordinates of p with respect
      // to the triangle *this.
      // Source: G. Aumann, K. Spitzmueller:
      // Computerorientierte Geometrie, BI 1993,  p.43
};

///////////////////////////////////// class Img24 ////////////////////////

class Img24{ // class of 24-bit (R,G,B)-images
// Notice that in the CpmGraphics system, the generation and display of
// screen graphics always make uses of Img24.
/*
Convention concerning the geometry of a digital image:
Considering images as data structures implies to abstract from their
pictorial content almost completely. What has to be left over from
content is the notion of orientation which says us what the upper, lower,
left, right parts of the image are.
A convenient formalization of an oriented pixel arrangement is given
by an instance of struct xyxy as introduced in cpmviewport.h.
Here the indexes of a pixel can be considered as integer-valued
coordinates with respect to a coordinate system which has its
origin in the left-upper corner of the image, and an x-axis in horizontal
direction to the right, and a y-axis in vertical direction downwards.
By using R-valued coordinates (x,y) one gets positions in between of pixel
positions that can be useful for interpolation.
Whenever R-valued coordinates x and y are used in this class
declaration/definition this coordinate system is selected.
Notice that mathematical matrices use a different relation between
argument (here: index) position and directions (left-right, up-down):
Here the first index grows down and the second index grows to the right.
The internal object bh_ uses just this matrix order of things.
Also the two or three dimensional arrays that appear as arguments
of constructors are placed as mathematical matrices into the image
area.
*/
// Data (all private)

   Z width;
   // Width of the bitmap

   Z height;
   // height of the bitmap

   V< V<ColRef> > bh_;
      // Image data.
      // Indexing corresponds to considering the image rectangle as
      // a mathematical matrix; indexing is however starting at 0.
      // Especially
      //    height==bh_.dim()
      // and
      //    widtht==bh_[i].dim()
      // for all i \in {0,...,height-1}

   mutable Rec displayBuffer;
      // see cpmviewport.h for class Rec. As the name says, this data
      // member holds (in his linear image memory) the pixel values
      // of *this which can be displayed via our viewport.
      // In addition to this linear memory it holds the descriptors for
      // a rectangular subarea of the viewport window. See cpmviewport.h.
      // The size of this viewport window and also the size of  *this
      // may be much larger than any physical display.

   mutable B readyForDisplay;
      // true says that the data from bh_ have been mapped consistently
      // to displayBuffer so that calling displayBuffer.vis() shows
      // the part of *this which fits the viewport.  The function which
      // achieves this consistent mapping is adjDisBuf().

   void adjDisBuf()const;
      //: adjust display buffer
      // Declared constant since the data which it changes are
      // declared mutable.

   ColRef c0;
      // Base color of the image with which the bitmap gets filled
      // on initialization in all cases that don't provide
      // pixel data.

   ColSty sty;
      // sets the color mixing mode of graphical actions
      // The meaning of color adding in case of sty==ADD_COLOR
      // is determined by the static public variable
      // rgb::addingMethod for which presenty values 0,1,2
      // have a meaning

   B antiAliasLines;
      // influences only function line(...) in a way that is clear
      // from the name

   B thickLines;
      // influences only function line(...) in a way that is clear
      // from the name.
      // There is only a single method of thickening which makes
      // horizontal and vertical lines 3 pixels wide and thus effectively
      // covers pixel irregularities in fine-grained polygons.
      // Graphics generated with thick lines can be converted (by some
      // image processor) into PNG and then included in LATEX-documents.
      // Without thickening graphics which contain horizontal or vertical
      // lines look very poor in pdflatex-generated documents since some
      // of the lines are visible, others not.

   B newLines;
      // If true, the lines are drawn according to a new method (developed
      // during a stay in Munich, June 2010) which takes care of the non-
      // integer coordinates of initial and final point of a line.

   R1 gamma;
      // controls how to add colors

   R1 wMix;
      // Controls how to mix colors (i.e. only if sty==MIX_COLOR,
      // only then it is not excluded from reading and writing).
      // It is the weight of the actual color relative to the
      // background color.

// private functions

   V< V<ColRef> >& rep_(){ return bh_;}

   xy disXY(R x, R y)const // discretize
      // returning the discrete representative of the
      // continuous location (x,y). Meaningful only for
      // (x,y) \in W:=[0,width]x[0,height].
      // Thus x-axis horizontal, to the right
      // y-axis vertical downwards. This means that here
      // (unlike in Frame) x,y are just the continuous versions
      // of graphics indexes. Notice that the indexes of bh_ are
      // matrix indexes (starting with 0).
   {
      return xy(x==width ? width-1 : cpmtoz(x),
         y==height ? height-1 : cpmtoz(y));
   }

   Z disX(R x)const
      // For all R x, R y the term
      // bh_[disY(y)][disX(x)] is guaranteed to be
      // well-defined.
   {
      if (x>=width) return width-1;
      else if (x<=0) return 0;
      else return cpmtoz(x); // x always positiv, has to be fast
   }

   Z disY(R y)const
      // see Z disX(R x)const
   {
      if (y>=height) return height-1;
      else if (y<=0) return 0;
      else return cpmtoz(y);
   }

   L outcode(R x, R y)const
      // Sutherland's outcode. Setting bits intelligently
   {
      L res=0;
      if (x<0) res+=8;        // 8=2^3
      if (x>width) res+=4;    // 4=2^2
      if (y<0) res+=2;        // 2=2^1
      if (y>height) res+=1;   // 1=2^0
      return res;
   }

   bool clip(R& x1, R& y1, R& x2, R& y2)const;
      // The points (x1,y1) and (x2,y2) determine a closed line
      // segment S. The image *this determines the closed
      // rectangle W:=[0,width]x[0,height] in R^2 (sic, no -1 !).
      // We consider D := section of sets S and W.
      // There are two cases:
      // (1) D == void set
      // (2) else
      // The function returns false in the first case and true else.
      // In the second case the function potentially modifies the
      // arguments (which is possible since these are references) in such
      // a way that D is the closed line segment determined by (x1,y1)
      // and (x2,y2).

    Img24 intPolAux(Z mI, Z nI, Z method)const;
      // tool for implementing function interpolate

    Rec toRec();
      //: to rectangle

    Img24(Rec const& r);
   // construcor from graphical rectangles with linear byte memory

    typedef Img24 Type;

public:
   CPM_IO
      // only the non-mutable data are written and read
   CPM_ORDER
      // Only to not unnecessarily deviate from the standard
      // infrastructure of value classes.

   Img24();
   // trivial constructor which does not allocate memory

   explicit Img24(ColRef c, R fracX=1, R fracY=1);
      // Image of uniform color c which fills the viewport
      // according to the fractions given for x and y.
      // sty=SET_COLOR, gamma=Viewport::getGamma()

   Img24(Z w, Z h, rgb c=rgb(), ColSty cs=SET_COLOR);
      // constructor which allocates memory for x*y pixels and initialize
      // id with Pix(r,g,b)
      // Z w: Width of the bitmap
      // Z h: Height of the bitmap
      // We use SET_COLOR as the default so that classes Img24
      // and Viewport, behave as exchangable as possible

explicit Img24(Word const& Filename);
   // constructor from a *.ppm (binary portable pixmap,
   // classified P6 or P3).
   // P6 has all data beyond 'depth' in binary format; i.e.
   // each character counts, no one is overread as whitespace, new-line
   // or something like that. Therefore P6 is faster to read and to write
   // and creates smaller files. Example of usage:
   //    Img24 im2("pisces.ppm");
   // 2014-07-31:
   // If the underlying system allows commandline inputs "gm convert ...."
   // also "pisces.jpg" and "pisces.JPG" works. The implementation then
   // calls gm convert internally and creates a ppm image from the jpg
   // image. Here gm is the free GraphicsMagick program that on most
   // Unix/Linux systems is installed and otherwise is easily obtained
   // from the net.

explicit Img24(VVV<L> const& vv);
   // generation from 'standard imaging data type'
   // Pixel locations in 'matrix style'.
   // The first index gives the color plane.

explicit Img24(VV<L> const& vv);
   // generation of a gray image from 'standard imaging data type'
   // Pixel locations in 'matrix style'.

explicit Img24(VV<ColRef> const& vv);
   // generation from a 'standard imaging data type'
   // Pixel locations in 'matrix style'.

explicit Img24(VVV<R> const& vv);
   // Pixel locations in 'matrix style'.
   // The first index gives the color plane.
   // Generation from floating point pixeldata. Such data can be
   // retreived from instances of my CpmImaging::MultiImage<T> for T=R
   // (see cpmmultiimage.h) which is capable of many image processing
   // operations such as Fourier filtering.
   // For vv.dim3()<3 a gray image will be created
   // (using the first image plane only). And for vv.dim3()>=3
   // a color image will be created (using the lower 3 image
   // planes only). If the values don't fit the range [0,255]
   // they will scaled by a linear transformation (same for all
   // image planes) to attain minimum 0 and maximum 255.
   // Small negative values (>-0.5) as may result from image
   // processing in floating point numbers are dealt with by a slight
   // shift and no multiplicative scaling. So natural color images
   // don't get automaticly boosted in a way that the value 255
   // appears in one channel at least.

VVV<L> toVVV()const;
   // creation of an equivalent 'standard imaging data type'
   // Let res be the return value. Then res[1] is the red
   // image matrix (i.e. res[1][1][1] left upper corner,
   // res[1][res.dim2()][res.dim3()] right lower corner).
   // res[2] is the green one, and res[3] the blue one.

VV<ColRef> toVVColRef()const;
   // creation of an equivalent 'standard imaging data type'
   // Let res be the return value. Then res[1][1] is the
   // color of the left upper corner,

VV<R> toVVR(R wR=1, R wG=1, R wB=1, Iv val=Iv(0,1))const ;
   // Returns a real image matrix as a gray level image
   // (obtained by forming an weighted average of
   // the r,g,b-plane with weights (wR,wG,wB)/(wR+wG+wB).
   // The values of the image are such that (r,g,b)=(0,0,0)
   // corresponds to val.inf() and (r,g,b)=(255,255,255)
   // to val.sup().

V<V<ColRef> > get_bh()const{ return bh_;}
   //: get bh_
   // Getting the pixel matrix in the original form for
   // experimentation.

Img24& operator+=(xy const& s);//{ displayBuffer+=s; return *this;}
   //: operator +=
   // No utilization of this, so far.

Img24& operator-=(xy const& s){ return (*this)+=(-s);}
   //: operator -=
   // No utilization of this, so far.

xy cor()const{ return displayBuffer.cor();}
   //: corner
   // Position of upper left corner of *this within the
   // rectangle given by the viewport window (see explanation
   // to member displayBuffer).

Img24 trn()const;
   //: transpose
   // Returns the image which has the transposed image matrix.

Img24 ref(Z i)const;
   //: reflect
   // Returns the image which results from *this by exchanging
   // left and right          for i==1
   // floor and ceiling       for i==2
   // both                    for i==12

V<ColRef> const& operator[](Z i)const{ return bh_[i];}
   // returns the i-th line of the image (i=0,...height-1)
   // in a form that allows to pass it over to image
   // processing algorithms in a fast way (notice that
   // we don't allow to change it.

// Handling color style
   void setCol(){ sty=SET_COLOR;}
   void addCol(){ sty=ADD_COLOR;}
   void mixCol(R w=0.5){ sty=MIX_COLOR; wMix=R1(w);}
      // setting color mix mode
   void stereo(){ sty=STEREO;}
   ColSty getSty()const{ return sty;}
   void setSty(ColSty sty_){ sty=sty_;}

// Handling the new implementation of lines
   bool getNewLines()const{ return newLines;}
   void setNewLines(bool aal=true){ newLines=B(aal);}
// Handling anti-aliasing, thick lines, and assumed display gamma
   bool getAntiAliasLines()const{ return antiAliasLines;}
   void setAntiAliasLines(bool aal=true){ antiAliasLines=B(aal);}

   void setThickLines(bool b=true){ thickLines=B(b);}
   bool getThickLines()const{ return thickLines;}

   R getGamma()const{ return gamma;}
   void setGamma(R g){ gamma=R1(g);}

// Handling base color
   void setColor(rgb c){ c0=ColRef(c);}
      // notice that filling the image with this color is the
      // task of function clear.
   rgb getColor()const{ return rgb(c0);}

// Initializing change of image size.
   void resize(Z widthNew, Z heightNew);
   // We re-set the size of the image data matrix and
   // initializes it with with c0. Input <0 will be handled as 0.
   // Thus, no logical exceptions. For large input, storage allocation
   // may fail however.
   // This action takes place even if width==widthNew and
   // height==heightNew.
// Descriptors for image size

   xyxy to_xyxy()const{ return xyxy(xy(),width,height);}
      //: to xyxy
      // Gives the set of pixels as a rectangular lattice

   bool isVoid()const{ return width<=0 || height<=0;}

   Z getLin()const{ return height;}
      // number of pixel lines

   Z getCol()const{ return width;}
      // number of pixel columns

   Z dim1()const{ return height;}
   Z dim2()const{ return width;}
      // for uniformity with matrices. Considers the image
      // as a pixel matrix in the natural manner
      // For uniformity with Image<>

   Z w()const{ return width;}
   Z h()const{ return height;}

   R getAspRat()const{ return R(width)/height;}
      //: get aspect ratio
      // Aspect ratio > 1 for landscape and <1 for portrait.

   IvZ i1()const{ return IvZ(0,width-1);}
      // index-range in 1-direction

   IvZ i2()const{ return IvZ(0,height-1);}
      // index-range in 2-direction

   Iv r1()const{ return Iv(0.,width);}
      // real range in 1-direction, the interval is to be considered
      // as [0,width)

   Iv r2()const{ return Iv(0.,height);}
      // real range in 2-direction, the interval is to be considered
      // as [0,height)

   R xr(R xi)const{ return xi*width;}
      // x from a 'relative coordinate' xi which varies in [0,1]

   R yr(R eta)const{ return eta*height;}
      // y from a 'relative coordinate' eta which varies in [0,1]

    bool out(R x, R y)const
      // Returns true if (x,y) is outside the image area
      // A=[0,width]x[0,height]
   {
      if (x<0) return true;
      if (y<0) return true;
      if (x>width) return true;
      if (y>height) return true;
      return false;
   }

   bool out(Z i, Z j)const
      // Returns true if (i,j) is outside the image area
      // [0,width-1]x[0,height-1]
   {
      if (i<0) return true;
      if (j<0) return true;
      if (i>=width) return true;
      if (j>=height) return true;
      return false;
   }

// Showing image data

   void vis()const;
      //: visualize
      // Displays *this on Viewport in a way that the left upper corner
      // of the image rectangele (on screen) is the screenpoint cor().
      // 2014-07-31: actually the left lower corner of the image is placed
      // at the left lower point on screen (empiricly!).

// Reading image data

    rgb getPel(Z i, Z j)const
   // Save reading access for discrete location. If (i,j)
   // is outside the image area, the background color of the
   // image gets returned.
   {
      if (out(i,j)) return rgb(c0);
      return rgb(bh_[j][i]);
   }

   rgb getPel(R x, R y)const
   // Save reading access for continuous location. If (x,y)
   // is outside the image area, the background color of the
   // image gets returned.
   {
      if (out(x,y)) return rgb(c0);
      xy z=disXY(x,y);
      return rgb(bh_[z.y][z.x]);
   }

// Changing image data

   void clear(void);
   // Fills the Bitmap with base color c0

   Img24 crop(Iv const& ivx, Iv const& ivy)const;
   //: crop
   // creates a rectangular cut-out of the image by giving the
   // subinterval of the x-axis and the subinterval of the
   // y-axis to be used. Taking the whole image corresponds
   // to ivx=Iv(0,1), ivy=Iv(0,1). The origin (x,y)=(0,0) is
   // the upper-left corner of the image. Notice that for images with
   // an aspect ratio different from 1, the aspect ratio of the croped
   // image differs from the quotient ivx.abs()/ivy.abs().

   void paint(rgb c){ setColor(c); clear();}
   // Fills the Bitmap with base color c and changes c0 to c

   void frame(rgb c=rgb(R(255.),R(0.),R(0.)));
   // sets the rim pixels to color c

   void putPel(Z x, Z y, rgb c)
   // Puts a pixel in a 24 bit bitmap at the position x, y to
   // the color specified in c.
   // Meaningful is 0<=x<width, 0<=y<height. If (x,y) violate this
   // there is no problem: the function simply does nothing in this case.
   // rgb is a struct with the Z-typed elements r, g, b.
   // colorStyle: SET_COLOR the new color is painted over the old one,
   //             ADD_COLOR the new color is added to the old one
   //                adding identical colors increases brightness
   //             STEREO adds colors in a simple manner which makes
   //                sense for colors being either (255,0,0) or
   //                (0,255,255) as is the case in anaglyphic
   //                represetation
   //             MIX_COLOR the new color is mixed with the old one
   //                mixing identical colors gives no change
   // It is only in the last case that the weight parameter w gets used.
   // w near 1 indicates dominance of the new color and w near 0
   // indicates dominance of the old color (e.g. background color).
   // Since access is over operator[] of class V<>, indexes out of range
   // will stop the program with an error depending on
   // defined(CPM_RANGE_CHECK).
   // (x,y)=(0,0) is the upper left corner and
   // (x,y)=(width-1,height-1) is the lower right one (graphics pixel
   // coordinates).
   // Thus the x-axis is as usual and the y-axis goes downwards
   // inlining for utmost efficiency
   {
      if (out(x,y)) return;
      if (sty==SET_COLOR){
         bh_[y][x]=c;
      }
      else{ ColRef cB=bh_[y][x]; bh_[y][x]=mix(cB,c);}
      readyForDisplay=B();
   }

   void putPel(R x, R y, rgb pix)
      // Defined for all (x,y). But does nothing for (x,y) outside
      // the image area
   {
      if (out(x,y)) return;
      xy z=disXY(x,y);
      putPel(z.x,z.y,pix);
   }

   void lineThin(R x1, R y1, R x2, R y2, rgb c);
   // Draws the clipped part of the line from x1, y1 to x2, y2.
   // Clipping is performed by the Cohen-Sutherland algorithm.
   // If the line is horizontal, a fast implementation based
   // on non-public function horLine is chosen.
   // The color of the line, specified in r, g, b, is combined with the
   // already existing color of each pixel according to the value of sty.
   // Depending on the attribute antiAliasLines the representation
   // reduces aliasing significantly if gamma has a value corresponding to
   // the system (e.g. 2.2).
   // The R-typed arguments actually enter ony through their integer
   // parts disX(x1),disY(y1),disX(x2),disY(y2).
   // This gives good-looking lines, but at the edges of polygons ugly
   // patterns like
   //              x
   //            x
   //          x x
   //        x
   // may appear. Notice that the C+- graphical system allows to create
   // arbitrarily large Img24 objects and write them to file as
   // ppm-images. If such large images get processed by a, say, 1.5 pixel
   // wide unsharp mask, one easily obtains print quality polygons.
   // Also employing function line(...) instead of lineThin(...)
   // together with thickLines==true  conceals irregularities of polygons.
   // 2010-6-22: Attempt to improve this by making better use of the real
   // input values (Munich idea, not yet ready, see next function).

   void lineThinNew(R x1, R y1, R x2, R y2, rgb c);
   // Now this function is used instead of lineThin(R,R,R,R,rgb)
   // if antialiaslines==false and sty==SET_COLOR. The R-typed
   // arguments act as they are and not only through their integer
   // part. This lets the ugly patterns like
   //              x
   //            x
   //          x x
   //        x
   // no longer appear.
   // The new method holds a strong potential for anti-aliasing, but
   // this potential cannot be realized with a line function in isolation.
   // Since typically the lines are arranged to form polygons, the last
   // pixel of one line gets a color the weight of wich should be known
   // to the first pixel of the next line to get the color right.
   // Now it looks to me that these reservations are immaterial and that
   // a solution based on a line-function (even the present one) should
   // work.
   // Not mature, tests not yet ready.

   void line(R x1, R y1, R x2, R y2, rgb c)
      // as lineThin, but thickness controlled by data member thickLines
   {
      lineThin(x1,y1,x2,y2,c);
      if (thickLines){
         lineThin(x1+1,y1,x2+1,y2,c);
         lineThin(x1-1,y1,x2-1,y2,c);
         lineThin(x1,y1+1,x2,y2+1,c);
         lineThin(x1,y1-1,x2,y2-1,c);
      }
   }

   void fillRectangle(R x1, R y1, R x2, R y2, rgb c);
   // Fills a rect with the RGB color specified in c. x1, y2 is the
   // upper left corner (beginning with 0, 0) and x2, y2 is the lower
   // right corner of the recangle to fill.
   // Now no longer implemented by putPel but by much faster horLine.

   void fill(TriAng const& tri, rgb c);
   // Fills a triangle tri with the color specified in c.

   void fillEllipse(R x1, R y1, R x2, R y2, rgb c);
   // The function paints a solid ellipse into the rectangle defined
   // by x1, y1 (left, top) and x2, y2 (right, bottom).
   // Actually only the part of the ellipse wich lies in the
   // image area gets painted.

   bool horLine(Z x1, Z x2, Z y, rgb c);
   // Draws a line from (x1,y) to (x2,y). The color of the line,
   // specified in r, g, b, is combined with the already existing color
   // of each pixel according to the value of sty.
   // Depending on the attribute antiAliasLines the representation
   // reduces aliasing significantly if gamma has a value corresponding to
   // the system (e.g. 2.2)
   // Returns true if at least one pixel was affected, else false.

   void horIns(xy p, V<ColRef> const& line,
      ColRef cr=ColRef());
      // Horizontal insert, no restrictions to the length of the line.
      // Only the part that fits the frame will be used. Insertion
      // is always in color setting mode (the old image is locally
      // replaced by the color data from line. p is the first point in
      // which the replacement becomes active, and the replacement goes
      // of course from left to right. If cr has not the default value,
      // it causes an admixture of this color in a way that was intro-
      // duced for implementing writing letters. Special usage, not of
      // general relevance.

   void insert(xy p, Img24 const& img , ColRef cr=ColRef());
      // the image img replaces the part of *this which is determined
      // by the size of img and the prescription that p is the
      // left upper corner of the replacement area. If due to
      // the size of img and/or the location of p *this can't
      // incorporate the whole of img, an appropriate outcut
      // of img will be used for this purpose.
      // Since 2005-09-01 implementation is based on the fast
      // function horIns and no longer on looping over individual
      // pixels.
      // The ColRef argument is needed in some of my font generation
      // functions. Not needed in general applications.

// storing image data on file

//   Word write(const Word& Filename="img24_write", bool binary=true);
   Word write(const Word& Filename="img24_write", bool binary=false);

   // The function writes a bitmap to Filename+"extension".
   // Return value is the file name with proper extension added by the
   // function.
   // "extension"=".ppm"
   // The binary=true gives compact files of portable pixmap P6 format
   // that can be used to create a new  image via the constructor
   // Img24(Word const&). binary=false gives portable pixmap P3 format
   // in a form that can be easily interpreted by a text editor. A
   // complete legende gives date of generation and also the value of the
   // data member text. Explanatory text is preceded by a '#' so that a
   // ppm-reader overreads them and can display the image
   // directly. My constructor reads both variants correctly.
   // Note 2008-03-06 at least working under MS compiler, binary sometimes
   // gets reconstructed badly and (missing R-plane ?) and the
   // PhotoImpact program fails to read some of the frames. The binary=
   // false setting works correctly. So P3 should be preferred over P6 in
   // my setting. The Mingw-executable works (at least in this case)
   // perfectly also for binary=true.

// Auxiliary functions

   static ColRef miX(ColRef const& cB, ColRef const& c,
      ColSty const& s, R const& g, R const& w);
      // this is the primary definition of making
      // display color from existing background color cB
      // and the color c that 'is aimed for' in the
      // graphical action calling this function.
      // g is a gamma.

   ColRef mix(ColRef const& cB, ColRef const& c)const
      // making this a member function; notice that making
      // F-typed function objects as needed in the fast
      // definition of horLine, one needs miX and not mix
      // Here Img24::gamma comes in.
   {
      if (sty==STEREO) return cB.add(c);
      else return miX(cB,c,sty,gamma,wMix);
   }

   Img24 interpolate(Z wNew, Z hNew, Z method=1)const;
      // Returns an Image of with=wNew and height hNew which is created
      // from *this by a interpolation method selected by the parameter
      // method; method =1 means bilinear interpolation and method=0 means
      // pixelation. The latter is useful if the number of pixels is to
      // be compared to expectations.
      // This works well also for wNew << width and hNew << height in
      // which case a combination of interpolation and averaging is done.

   Img24 pair(Img24 const& img, Z gap=2)const;
      //: pair
      // returns a paired image made from *this and img.
      // Assumes that these two images have the same height.
      // The width of the result is the sum of the widths of the
      // components. No interpolation. Argument gap is the width of
      // a separating gap between the images.

   Word nameOf()const{ return "Img24";}

   static ColRef biLin(ColRef f11, ColRef f12, ColRef f21, ColRef f22,
      R w11, R w12, R w21, R w22, Z method);
   // law for bi-linear interpolation (for method==1, else pixelation)

};

} // namespace

#endif
