//? cpminterval.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_INTERVAL_H_
#define CPM_INTERVAL_H_
/*
    Description:
      open intervals on the real axis endowed with the lattice structure
      meet and join (here operator | and operator &)
*/

#include <cpmlinalg.h>
   // it is considered natural to treat geometry
   // on the basis of linear algebra. Presently
   // CpmLinAlg only needed for being able to use R2 and C2
#include <cpmclinalg.h>
#include <cpmx.h>
#include <cpmzinterval.h>

namespace CpmGeo{

   using namespace CpmStd;

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::Word;
   using CpmArrays::IvZ;
   using CpmLinAlg::Z2;
   using CpmLinAlg::R2;
   using CpmLinAlg::C2;
   using CpmArrays::V;
   using CpmArrays::X2;
   using CpmFunctions::F;

////////////////////// class Iv ////////////////////////////////////////

class Iv{ // open intervals in R
   // could be extended to interval arithmetics, therefore by the
   // clone macro polymorphic containers are supported

//independent quantities
   R left;
   R right;
// dependent quantities
   R center;
   R halfWidth;
   R absInv; // 1/abs() or 0
// giving the dependent quantities the right values
   void update();
   void updateLight(); // does not order left,right
public:
   typedef Iv Type;
   typedef Iv CloneBase;
   virtual Iv* clone()const{ return new Iv(*this);}
   CPM_IO
   CPM_ORDER
   CPM_NAM(Iv)
   CPM_WORD
   R dis(Iv const& iv)const;
   Iv(R a, R b, bool ordered=false):left(a),right(b)
      { if (!ordered) update(); else updateLight();}
   Iv():left(0),right(0),center(0),halfWidth(0),absInv(0){}

   explicit Iv(R a):left(-cpmabs(a)),right(cpmabs(a)){updateLight();}
      // Interval [-|a|,|a|]

   explicit Iv(R2 const& r2):left(r2[1]),right(r2[2]){update();}
      // making an interval out of a pair of numbers

   Iv(const V<R>& vr);
      // generates the interval determined by the smallest
      // and the largest of the components vr

   bool isVoid()const{ return left==right;}
      // This shows that our intervals here are not fully
      // characterized when considered point sets. There
      // is only one void point set in R but many Iv's
      // with left==right

   R abs()const{ return halfWidth*2;}
      // length of interval = diameter of the set
   R radius()const{ return halfWidth;}
      // in one dimension an interval also is a sphere
   R rad()const{ return halfWidth;}
      //: radius

   R inf()const{ return left;}
      //: infimum
   R b()const{ return left;}
      //. begin

   R sup()const{ return right;}
      //: supremum
   R e()const{ return right;}
      //. end

   R operator[](Z i)const{ if (i<=1) return left; else return right;}
      // since Iv and R2 are often to be used interchangingly
      // it is convenient to get to the data by the same syntax

   R mean()const{ return center;}
      // center

   R cen()const{ return center;}
      //: center

   R mod()const{ return left==right ? R(0.) :
      R(4.)*halfWidth/(cpmabs(left)+cpmabs(right));}
      //: modulation

   R x(R p)const{ return left+p*abs();}
      // returns the left end for p=0, the center for p=0.5,
      // and the right end for p=1.

   R operator()(R x)const{return cpmabs(x-center)-halfWidth;}
      // geometrical indicator function
      // zero on the boundary, positive outside
      // negative inside; gradient of the function gives the
      // outside-directed normal vector on the boundary

   Iv operator|(const Iv& iv)const;
      // minimum open interval that contains the union
      // join or l.u.b. (lowest upper bound) in lattice terminology

   Iv operator&(const Iv& iv)const;
      // section, intersection
      // meet or g.l.b (greates lower bound) in lattice terminology

   bool contains(R x, bool leftOpen=true, bool rightOpen=true)const
      // answers the question whether point x is an element of the
      // interval *this. The second and the third argument allow to
      // set the interpretation of *this as an open, closed , or
      // semi-closed interval
   {
      bool leftOK=(leftOpen ?  x>left : x>=left);
      bool rightOK=(rightOpen ?  x<right : x<=right);
      return leftOK&&rightOK;
   }

   bool contains(const Iv& iv)const
      // says whether iv is a subset (true subset 'or equal') of *this
      // or not.
   {  return left<=iv.left && iv.right<=right;}

   R containsFuzzy(R x, R frac)const;
      // Returns 1 if x actually belongs to *this, and returns 0 if
      // x does not belong to *this*frac. For this to make sense, we
      // assume frac>=1. Betwee the two cases the returnvalue decreases
      // linearly from 1 to 0.

   bool isSubsetOf(const Iv& iv)const
      // says whether *this is a subset (true subset 'or equal') of iv
      // or not.
   {  return iv.left<=left && right<=iv.right;}

   const Iv& stretch(R k, Z s=0);
      // multiplies the size of the interval by |k| and leaves
      // a point p invariant, where this (fixed) point is
      // (i)   the left end if s=-1
      // (ii)  the center if s=0
      // (iii) the right end if s=1
      // Error for other values of s

   const Iv& shift(R s){left+=s;right+=s; center+=s; return *this;}
      // shifts both ends of the interval by s

// modern form of moving both ends of the interval by the same
// step
   Iv& shf_(R s){left+=s; right+=s; center+=s; return *this;}
      //: shift
      // no change of the other dependent data
   Iv& operator+=(R s){left+=s;right+=s; center+=s; return *this;}
   Iv& operator-=(R s){left-=s;right-=s; center-=s; return *this;}
   Iv& operator*=(R s)
      // assumes s>0, multiplies the size by s, leaving the center
      // invariant
   {
      halfWidth*=s; absInv*=cpminv(s);
      right=center+halfWidth; left=center-halfWidth;
      return *this;
   }
   Iv operator*(R s)const{ Iv res=*this; return res*=s;}
      // for convenience
   const Iv& shiftLeftEnd(R s){ left+=s; update(); return *this;}

   const Iv& shiftRightEnd(R s){ right+=s; update(); return *this;}

   V<R> cen(Z n)const;
      //: centers
      // The interval is divided into n subintervals of equal length
      // and the n centers of these subintervals are returned as a
      // real vector in ascending order. n>=1 needed.

   R cen(Z i, Z n)const;
      // returns cen(n)[i] without creating the array

   V<R> centers(Z n)const{ return cen(n);}
      //:: returns a conventional array

   V<R> div(Z n)const;
      //: division
      // Equidistant division of an interval giving n values such that the
      // first and the last value coincide with the ends of the innterval.
      // Regular case: n>=2
      // We return res such that res.dim()==n,
      // res[1]=inf(), res[n]=sup().
      // and res[i+1]-res[i]=const
      // Special cases: negative n: n acts as |n|
      // n=0 V<R>(0) returned
      // n=1 V<R>(1,mean()) returned

   V<R> divGeoProg(R dt,Z n)const;
      // division according to geometric progression.
      // We return res such that res.dim()==n,
      // res[1]=inf(), res[n]=sup().
      // res[2]-res[1]=dt
      // and res[i+1]-res[i]=q*(res[i]-res[i-1])
      // for all i=2,...n-1 with some fixed q
      // Assumption thus n>2, 0<dt<abs(). These
      // are checked via cpmassert

   V<R> divSmooth(R dti,R dtf,Z n)const;
      // We return res such that res.dim()==n,
      // res[1]=inf(), res[n]=sup().
      // res[2]-res[1]=dti
      // res[n]-res[n-1]=dtf
      // and res[i] 'as slowly varying with i as possible'.
      // Assumption thus n>3, 0<dti+dtf<abs(). These
      // are checked via cpmassert.
      // First intention was to minimize sum of (res[i]-res[i-1])^2
      // which leads (Lagrange multipliers needed for constraints
      // of the problem) to a matrix inversion problem which
      // I used numerically and found steps si=res[i]-res[i-1]
      // smoothly going to negative values and then smoothly
      // going back to positive ones. This is no solution to the
      // present problem. The present robust method is to place the
      // n-4 not yet determined points equidistantly and then
      // smoothing many times by a simple linear filter of mask size 3.
      // See implementation code for details.

   V<Iv> divide(Z n)const;
      // The interval is divided into n subintervals of equal length
      // which are returned as a vector in ascending order. n>=1 needed.

   R periodic(R x)const;
      // evaluates a function which is periodic with respect to
      // a periodic repetition of the interval and takes the value 0
      // at the left end of the interval and the value 1 at the right
      // one. In between it is linear. (thus it is a saw-tooth function)
      // This function creates an error if abs()==0.

   R toTorus(R x)const
   {
      if (contains(x,false,false)) return x; // if x lies in
      // the closed interval *this x is what we search for
      return linear(periodic(x),0,1);
   }
      // Returns a y that lies in *this just
      // as x lies in a periodic copy of *this. So, if we bend the
      // interval to a circle and then roll the circle along the
      // real axis (in the direction to point x) eventually x will be
      // in contact with the circle and can be considered to make a mark
      // on it. This mark on the circle corresponds (upon unwinding our
      // circle to *this) to the point y we asked for.

   R ramp(R x)const;
      //: ramp
      // is a function which takes the value 0
      // for all points left of the interval and the value 1 everywhere
      // right of it. In between it is linear. (thus it is a linear ramp)
      // If the interval has length 0, we return 0.5 for x==left==right

   R win(R x, R eps=0.05)const;
      //: window
      // is a smooth function which is 0 outside the open interval *this
      // and is given as a 1-valued plateau inside the interval.
      // The transition zone from 0 to 1 has the width abs()*eps/2. Thus
      // eps is the probability to find a win(x,eps)<1 for x uniformly
      // distributed in *this.

   Z qun(R x, Z n, bool restrict=false)const;
      //: quantize
      // simple quantization into n levels
      // assumes and checks n>0.
      // We divide the interval [left,right) into a disjoint sum of
      // intervals of type [ ) and of equal length. For x in [left,right)
      // we return the number of the subinterval (thus a number in
      // {1,2,...n} to which x belongs.
      // For x==right, we also return n. For x outside of
      // [left,right] the return value is determined by the formula
      // 1+(Z)(n*(x-left)/(right-left)) for restrict==false
      // and 1 for x<=left , n for x>=right for restrict==true

   Z quant(R x, Z n, bool restrict=false)const{ return qun(x,n,restrict);}
      // old name, for compatibility

   R qunInv(Z n, IvZ const& iv)const;
      //: quantize inverse
      // returns y which is related to *this just as n is
      // related to iv.

   X2<R2,Z2> intPolInfo(R x, Z n)const;
      //: inter-polation information
      // For the following explanation we assume the return value
      // being written as (R2(w1,w2),Z2(i1,i2)) giving us two
      // weights w1, w2 (wi>=0, w1+w2=1) and two indexes i1,i2
      // (1 <= i1 <= i2 <= n, i2-i1 \in {0,1}).
      // Consider a real-valued positive function f defined on
      // the interval *this. We consider the natural partition of
      // *this into n subintervals of equal length and we index them
      // in increasing order which gives us a well-defined list
      // I_1, ..., I_n of intervals. With interval I_i we associate
      // the real number f_i, by definition the mean value of function f
      // in I_i (i.e. the integral of f over I_i, divided by the length
      // of I_i). An important concrete situation is that the f_i are
      // pixel exposures and f an illuminance.
      // Then w1,w2,i1,i2 are set such that
      //       w1*f_i1 + w2*f_i2
      // is the most natural hypothesis for f(x). If x is outside
      // the interval, there is no problem: Then simply i1=i2, w1=w2=0.5.
      // More specificly: for x<inf() we have i1=i2=1
      // and for x>sup we have i1=i2=n.

   X2<R2,Z2> intPolInfo2(R x, Z n)const
      { R y=per(x); return intPolInfo(y,n); }
      //: inter-polation information 2
      // Clear from code.

   R linear(R x, R yL=-1, R yR=1 )const{ return
      (left==right) ? (yL+yR)*0.5 : (yL*(right-x)+yR*(x-left))*absInv;}
      // Is a linear function which takes the value yL at the left
      // end of the interval and the value yR at the right one.
      // If the interval has length 0, we return 0.5*(yL+yR).

   Z stair(R x, Z n)const{ return 1+cpmtoz(ramp(x)*n);}
      // Is a function which takes the value 0
      // at the left end of the interval and the value n+1 at the right
      // one. In between it is a monotone stair function taking the
      // values 1,2,...,n in succession, on intervals of equal width.

   V<bool> bisection(R x, Z n)const;
      // returns n most significant bits (starting with the
      // utmost significant one) of the number ramp(x).
      // This is the explanation, the implementation is
      // different and aims at efficiency

   R bisectionInv(const V<bool>& vb)const;
      // inversion; as accurate as the number of bits allows it
      // 02-10-02 tested with a few examples. Works !!!

   R2 toR2()const{ return R2(left,right);}
      // returns *this as a pair of real numbers

   R ran(Z j=0)const;
      // returns a random number that lies in the interval *this.
      // Here the argument j plays the same role as in
      //    R CpmRoot::randomR(Z j=0);
      // (see cpmtypes.h)

   IvZ toIvZ()const{ return IvZ(left,right);}
      // returns an approximate representation as a contiguous set of
      // integers

   R per(R x)const;
      //: periodic
      // returns y such that y lies in *this and y-x is an integer multiple
      // of abs().

// Auxiliar functions for relativistic dynamics of wave functions:
   Iv& lorentz_(R chi);
      //.. Lorentz transformation
      // \Lambda(chi)(0,x) = (t',x')
      // Here the Lorentz transformation acts on (t,x) and leaves x=0
      // invariant.
      // Let x vary over iv then x' varies over iv.lorentz(chi).
      // chi is here the rapitidy: v/c = tanh(chi).

   Iv exp(R t)const
      //: expand
      // Interpreting *this as a region of causal influence it expands in
      // the course of time to the interval returned by this function
   {
      R a = cpmabs(t);
      return Iv(left-a,right+a);
   }

   Iv boost(R chi)const
      // boost
      // Interpreting the interval iv as the carrier of a Dirac wave
      // function psi in 1D position space. Then iv.boost(chi) is the
      // carrier of exp(i chi N) psi where N is the boost generator.
      // N = (X H + H X)/2 when expressed in terms of Hamiltonian and
      // position operator.
   {
      R c=cpmcosh(chi), s=cpmsinh(chi);
      return Iv(left*c-cpmabs(s*left),right*c+cpmabs(s*right));
   }

   Iv& step_(R ds)
      //: step
      // In cases where the interval plays the role of a relativistic
      // domain of influence it evolves over time by expanding with
      // constant speed (here set to 1).
   { left-=ds; right+=ds; update(); return *this; }
};

/////////////////////// class DiIv /////////////////////////////////////

class DiIv{ // directed interval in R
   // tool for letting two arbitrary (but not coinciding) points of
   // R play the role of the origin (x0) and of the unity (x1)
// independent data
   R x0,x1;
// dependent data
   R range,rangeInv;
// setting the dependent data
   void update();
public:
   typedef DiIv Type;
   CPM_IO
   explicit DiIv(Iv iv=Iv(0,1,true));
   explicit DiIv(R2 const& r2);
   DiIv(R x0_, R x1_);
   R p(R x)const{ return (x-x0)*rangeInv;}
      // function p returns something like the 'propoportional part' p
      // used in table interpolation; p=0 for x=x0 and p=1 for x=x1
   R x(R p_)const{ return x0+p_*range;}
      // returns a value x for which the 'proportional part' is p_
   Iv toIv()const{ return Iv(x0,x1,false);}
   R2 toR2()const{ return R2(x0,x1);}
   R operator[](Z i)const{ if (i<=1) return x0; else return x1;}
      // since DiIv,Iv and R2 are often to be used interchangingly
      // it is convenient to get to the data by the same syntax
};

////////////////////////// proportions /////////////////////////////////

// Functions related to pairs of points, which are not assumed to
// be ordered according to size as in an interval

R proportion(R x1, R x2, R xd);
   // Look at x1,x2,xd as three points on the a staight line.
   // The position of xd with respect to the ordered pair (x1,x2)
   // of points may be described by a proportion ('Teilverhaeltnis'
   // in German). This is given by (xd-x1)/(x2-xd) so that has the value
   // takes the following special values (|huge|>>1>>|tiny|>0)
   // xd=x1          proportion=0
   // xd=(x1+x2)/2   proportion=1
   // xd=x2+tiny     proportion=huge
   // xd=huge        proportion=-1+tiny
   // The relative sign of huge and tiny there depends on the sign of
   // x2-x1.

R pointFromProportion(R x1, R x2, R prp);
   // Reconstructing a point from its proportion relative to x1 and x2:
   // For
   //       prp=proportion(x1,x2,xd)
   // we get
   //       xd=pointFromProportion(x1,x2,prp)
   // Conversely, for
   //       xd=pointFromProportion(x1,x2,prp), prp!=-1
   // we get
   //       prp=proportion(x1,x2,xd).

////////////////////////////////// class Iv2 /////////////////////////////

namespace{
   R f1(R2 const& x){ return x[1];}
   R f2(R2 const& x){ return x[2];}
} // namespace

class Iv2{ // Interval in two dimensions: closed rectangle
   Iv iv1_, iv2_;
   typedef Iv2 Type;
public:
   CPM_IO
   CPM_ORDER
   Word nameOf()const{ return Word("Iv2");}
   Iv2():iv1_(),iv2_(){}
   Iv2(Iv const& iv1, Iv const& iv2):iv1_(iv1),iv2_(iv2){}

   Iv2(V<R2> const& vr);
      // generates the minimal rectangle which contains
      // all components of vr

   Iv2 operator|(Iv2 const& iv)const
   { return Iv2(iv1_|iv.iv1_,iv2_|iv.iv2_);}
      // minimum rectangle that contains the union
      // join or l.u.b. (lowest upper bound) in lattice terminology

   Iv2 operator&(Iv2 const& iv)const
   { return Iv2(iv1_&iv.iv1_,iv2_&iv.iv2_);}
      // section, intersection
      // meet or g.l.b (greates lower bound) in lattice terminology

   bool contains(R2 const& p)const;
      // returns true iff p[1] belongs to iv1_ and
      // p[2] belongs to iv2_. The boundary-question is
      // controlled by the default behavior of Iv::contains(R,bool,bool).

   bool isVoid()const{ return iv1_.isVoid() && iv2_.isVoid();}

   R2 cen()const{ return R2(iv1_.cen(),iv2_.cen());}
      //: center

   Iv const& c1()const{ return iv1_;}
   Iv const& c2()const{ return iv2_;}

   Iv& c1(){ return iv1_;}
   Iv& c2(){ return iv2_;}
};

} // namespace CpmGeo

#endif
