//? cpmrecord.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_RECORD_H_
#define CPM_RECORD_H_
/*
Record is a data structure capable of holding data of types bool, Z, R,
Word and accessing those data via a section and an item.
The main constructor for this class is one which reads all
the data from file. This file has to follow formatting
rules very similar to those for Windows ini-files.
The main difference is that we need an explicit type
indicator at the beginning of each value defining line.
Also, a section is a single word (that may contain spaces)
on a line. No '[' and ']' is needed.
*/
#include <cpmm.h>
   // map template. Powerful tool to achieve most of the functionality.

namespace CpmRootX{

   using namespace CpmStd;

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::Word;

   using CpmArrays::V;
   using CpmArrays::T2;
   using CpmArrays::M;
   using CpmArrays::S;

///////////////////////// class SecCol ////////////////////////////
// 'Section Collector'
// an auxiliary class
// Collects lines from various files and places them consistently
// in sections which are defined by headlines. (There are no
// headlines of different hierarchical level). See comment to
// SecCol::str_ for details.
// Tool for reading ('parsing') the most recent version of ini-Files
// which allow as data entries the names of other ini-files
// which will be used as data sources just as the one appearing as an
// argument of the main constructor.
// assume a file we want to use is named ".\\iniFiles\\aux.ini"
// then the corresponding entry in the basic ini file is
//
/*
F ./iniFiles/aux.ini

  or

include ./iniFiles/aux.ini
*/
// where F tries to suggest the usage of File as a data type
// and the word 'include' is familiar from C programming.
// File names that include whitespace cannot be used.

// The result of parsing is the well organized single string
// str_, which can be accessed via function str()
// and finally by Record::readFormatted(...).

// The class SecCol is used to implement
// bool Record::readFromFile(const Word& fileName, bool binary);
// which in turn implements
// RecordHandler::RecordHandler(const Word& file=Word(), bool
// verbose=true);
// and
// IniFileBasedApplication::IniFileBasedApplication(const Word& prog).

class Record; // forward declaration

namespace aux{

   class SecCol{// section collector,

      B binary;
         // file open mode, true and false seem to work equally well
         // on my present Windows system

      string el;
         // line terminator ('end of line'), may need to depend on the
         // parameter binary

      S<Word> fns;
         // file names, initialized as void set

      Z1 nFiles_;
         // total number of files which were used to create str_

      string str_;
         // the content of all contributing files as a single string which,
         // however, is subdivided into lines by 'end of line characters'
         // (see 'el' above).
         // Each data line is (somewhere) preceeded by its section
         // head-line. Sections are ordered (according to the default order
         // of class string) by their section head-lines.
         // Each line which does not begin with the comment indicator ("c")
         // or a type indicator (such as "B", "Z", "Zs" ...)  is (by
         // definition) a section head-line.
         // The detailed structure of the data entries in ini-files
         // does not matter for the present class.

      bool ini();
         // main work to implement the main constructor

      B valid;
         // initialized as false !

   public:

      Word nameOf()const{ return Word("SecCol");}

      explicit SecCol(const Word& fileName, bool binaryI=true);
         // raison d'etre (sorry for missing accent, 7bit ASCII
         // limitation)

      SecCol(){}
         // useful ???

      const string& str()const{ return str_;}
         // raison d'etre

      Z nFiles()const{ return nFiles_;}
         // for information; may help to see whether all intended files
         // were found

      bool isVal()const{ return valid;}
   };

//////////////////////// class Key ///////////////////////////////////
// another auxiliary class

class Key: public T2<Word>{
      // now IO and ORDER from T2
      void normalize(){ c1().skipSurSpc_(); c2().skipSurSpc_();}
   public:
      typedef T2<Word> Base;
      typedef Key Type;
      Key():Base(){}
      Key(const Word& w1, const Word& w2):Base(w1,w2){normalize();}
         // assures that space and whitespace (such as carriage return
         // and line feed do not shine up in the keys even if
         // written to and read from file without too much care.
         // This is important since identity of keys is required for
         // functional identity.
   };

///////////////////////// class TextH ////////////////////////////////
// again an auxiliary class

   class TextH{
      // Text with special header of each line; too special
      // to justify the general name Text. Should grow into a general
      // class Text after some experience.
      V<string> vs;
      typedef TextH Type;
   public:
      CPM_IO
      Word nameOf()const{ return Word("TextH");}
      friend class Record;
      TextH():vs(0){}
         // Here it is important to get a truely void text.
      explicit TextH(const string& s):vs(1,s){}
      explicit TextH(Word const& w):vs(1,w.str()){}
      void operator &=(const string& s){ vs&=s;}
      void operator &=(const TextH& t){ vs&=t.vs;}
      void operator &=(Word const& line){ vs&=line.str();}
   };

} // aux

class Text{// texts as lists of Words
      V<Word> vw_;
      explicit Text(V<Word> const& vw):vw_(vw){}
      typedef Text Type;
   public:
      CPM_IO
      Word nameOf()const{ return Word("Text");}
      Text():vw_(0){}
      explicit Text(Word const& w):vw_(1,w){}
      explicit Text(string const& s):vw_(1,Word(s)){}
      static Text getLine(istream& in);
      V<Word> const& rep()const{ return vw_;}
   };

/////////////////////// class Record /////////////////////////////////

   class Record{ // data structure for communication with ini-files
      // and other input mechanisms to complex programs and functions. The
      // idea is, that all input data for defining the conditions for
      // a program run should be collectable in a Record.
      // We do not follow the self-suggesting idea to hold in Record
      // only a string representation of all data, together with a type
      // indicator. Instead we like to have all data in their
      // 'native' binary format.
      // A sufficienly general set of types for this usage turned out to
      // be:
      // bool, Z, R, Word, V<bool>, V<Z>, V<R>, V<Word>
      // Reading and writing data of these types is supported by
      // functions set() and get()
      // Usage and motivation for the V<>-types is explained in the
      // comment to
      //    explicit Record(const Word& fn);

   typedef M<aux::Key,B>            m0;
   typedef M<aux::Key,V<B> >        v0;

   typedef M<aux::Key,Z1>           m1;
   typedef M<aux::Key,V<Z1> >       v1;

   typedef M<aux::Key,R1>           m2;
   typedef M<aux::Key,V<R1> >       v2;

   typedef M<aux::Key,Word>         m3;
   typedef M<aux::Key,V<Word> >     v3;

   typedef M<aux::Key,aux::TextH>   m4;
   typedef M<Word,aux::TextH>       m5;
   typedef aux::TextH               m6;

   m0 rep0;
      // 'list' of names, sections, and values of B-valued variables
   v0 vec0;
      // 'list' of names, sections, and values of B-valued lists

   m1 rep1;
      // 'list' of names, sections, and values of Z1-valued variables
   v1 vec1;
      // 'list' of names, sections, and values of Z1-valued lists

   m2 rep2;
      // 'list' of names, sections, and values of R1-valued variables
   v2 vec2;
      // 'list' of names, sections, and values of R1-valued lists

   m3 rep3;
      // 'list' of names, sections, and values of Word-valued variables
   v3 vec3;
      // 'list' of names, sections, and values of Word-valued lists

// Handling comments (do work well after a change in implementation
// that eliminats all CR symbols from comments (2004-09)

   m4 rep4;
      // holds the comments to section-item pairs
   m5 rep5;
      // holds the comments to the sections
   m6 rep6;
      // comment for the whole Record *this

   B valid;
      // initialized as false !

   public:
   typedef Record Type;

   static bool startsEntry(Word const& w)
      // returns true if w is one of the type indicating words,
      // which, by definition, preceed a value entry.
      // Since 2004-10-12 all non-comment lines for which
      // the first Word w1 says startsEntry(w1)==false is
      // interpreted as a section title. Thus such a title may
      // be a chain of whitespace separated contiguous Words.
      // This allows using unmodified plain language
      // in complicated section titles e.g.
      // 'kinetic energy' instead of 'kinetic_energy' or 'eKin'.
      // Notice that the list types Bs, Zs, Rs, Ws are followed by
      // value lists which are white-space separated,  n o t  comma
      // separated. E.g.
      // Rs times= 0.1 0.2 0.4 1.
   {
      return
      w=="F"  ||
      w=="c"  ||
      w=="B"  ||
      w=="Bs" ||
      w=="Z"  ||
      w=="Zs" ||
      w=="R"  ||
      w=="Rs" ||
      w=="W"  ||
      w=="Ws"  ;
   }

   CPM_IO
   CPM_ORDER
  // This cannot be implemented in an efficient
  // manner with the given data members. An efficient implementation
  // will be given in class Recordable.

  Word nameOf()const{ return Word("Record");}

   bool writeFormatted(ostream&, bool writeComments=false)const;
      // writes the content of *this (which, of cource, is being
      // held in the attributes rep0, ..., rep6) in such a
      // way to a stream, that the following read function can
      // recover it from there. Whether 'c'-comments (that tend to
      // become lengthy in my applications) will be written
      // or not is controlled by the second argument.
   bool readFormatted(istream&);
      // reading from a file that follows the syntax for my ini-files
      // Also writes to cpmcerr what it found
      // This is the function which transforms strings into arithmetic
      // types. Here I had to make additions for allowing the syntax
      // R a= 1 3
      // for giving item a of type R the value 0.33333333....
   bool readFromFile(const Word& fileName, bool binary);

   Record(){};
      // this creates a void record. Data are inserted via set. e.g.
      // Record rc;
      // R v=1000, J=1.8, el=30000, hVal=200000;
      // Z verb=2;
      // bool nog=false;
      // Word conf("config1");
      // V<R> waitTimes("",1.1,2.2,5,5);
      // rc.set("Electricity","V",v);
      // rc.set("Magnetism","J",J);
      // rc.set("Electricity","E",el);
      // rc.set("Magnetism","H",hVal);
      // rc.set("General","data",conf);
      // rc.set("General","verbose",verb);
      // rc.set("General","noGraphics",nog);
      // rc.set("General","waitTimes",waitTimes);

      // These set() statement create the following entries in the
      // file by rc.writeFormatted(out), where the stream out is given by
      // cpmofstream(out,filename)
/*
Electricity
R E=30000.
R V=1000.
General
W data=config1
B noGraphics=false
Z verbose=2
Rs waitTimes=1.1 2.2 5. 5.
Magnetism
R J=1.8
R H=20000
*/
   // A even more convenient way of defining a Record is by reading it
   // from a textfile:
   //
   // Record rc("input.ini");
   //
   // The previously displayed
   // data (between /********... and ...*****/) have the matching format
   // for this reading function. In addition such a file may contain
   // ignored lines starting with '/' or ';' or '*'.
   //
   // Lines starting with 'c' (optionally preceeded by some white space)
   // are constituting elements of the Record and are, depending on their
   // position in the file, comments to the whole file, or to a section,
   // or to an item within a section.

   explicit Record(const Word& fn);

// bool, Z, R, Word are chosen as the basic data types for controlling
// programs. Since we are dealing with the interface of a program with
// the outside world, program specific user defined types and
// mathematical abstractions such as complex numbers, should not be
// considered here. However, it turned out that lists of the types
// mentioned above occur often, also under circumstances where
// the length of the list has to be changed often so that an independent
// entry for the number of list members is not convenient. This new
// extension now allows entries
//
/*
example
   Rs xl=137 3.14 1.4142
   R  x=1.7321
      // single R
   R a=1 3
      // single R with value 1./3. To allow a division sign '/' would need
      // major design changes. R b=1.54 0.42 results in b=1.54/0.42 so this
      // format is not restricted to fractions of integers.
   Rs y=3.14
      // V<R> with only one component
   Bs bl=true false false true
   Ws a=eins zwei drei
   Zs b=1 2 3 137
*/
// The correct reading of these quantities is
/*
   Word sec="example";
   V<R> xl,y;
   R x;
   V<Z> b;
   V<Word> a;
   V<bool> bl;
   rc.get(sec,"xl",xl);
   rc.get(sec,"y",y);
   rc.get(sec,"x",x);
   rc.get(sec,"bl",bl);
   rc.get(sec,"a",a);
   rc.get(sec,"b",b);
*/
// Reading the fundamental basic types from a position determined by
// section and key. A failure in doing so never stops the
// program. For pedantic!=0 diagnostic messages will be left on file
// cpmcerr.txt.
// It is only for return value==true that argument value may get changed.

   bool get(const Word& section, const Word& item, bool& value,
      Z pedantic=1)const;
   bool get(const Word& section, const Word& item, V<bool>& value,
      Z pedantic=1)const;

   bool get(const Word& section, const Word& item, Z& value,
      Z pedantic=1)const;
   bool get(const Word& section, const Word& item, V<Z>& value,
      Z pedantic=1)const;

   bool get(const Word& section, const Word& item, R& value,
      Z pedantic=1)const;
   bool get(const Word& section, const Word& item, V<R>& value,
      Z pedantic=1)const;

   bool get(const Word& section, const Word& item, Word& value,
      Z pedantic=1)const;
   bool get(const Word& section, const Word& item, V<Word>& value,
      Z pedantic=1)const;

// new fast lean access to data, return the return classes default value
// if there is no matching entry. For data that need a conversion, the
// implementation is not inline, see cpmrecord.cpp

// constant get functions
   bool getB(Word const& section, Word const& item)const
   { return rep0[aux::Key(section,item)];}
   V<bool> getBs(Word const& section, Word const& item)const;
   Z getZ(Word const& section, Word const& item)const
   { return rep1[aux::Key(section,item)];}
   V<Z> getZs(Word const& section, Word const& item)const;
   R getR(Word const& section, Word const& item)const
   { return rep2[aux::Key(section,item)];}
   V<R> getRs(Word const& section, Word const& item)const;
   Word getW(Word const& section, Word const& item)const
   { return rep3[aux::Key(section,item)];}
   V<Word> getWs(Word const& section, Word const& item)const
   { return vec3[aux::Key(section,item)];}
// mutating set functions
   void setB(Word const& section, Word const& item, bool x)
   { rep0[aux::Key(section,item)]=B(x);}

   void setBs(Word const& section, Word const& item, V<bool> const& x);

   void setZ(Word const& section, Word const& item, Z x)
   { rep1[aux::Key(section,item)]=Z1(x);}

   void setZs(Word const& section, Word const& item, V<Z> const& x);

   void setR(Word const& section, Word const& item, R x)
   { rep2[aux::Key(section,item)]=R1(x);}

   void setRs(Word const& section, Word const& item, V<R> const& x);

   void setW(Word const& section, Word const& item, Word const& x)
   { rep3[aux::Key(section,item)]=x;}

   void setWs(Word const& section, Word const& item, V<Word> const& x)
   { vec3[aux::Key(section,item)]=x;}
// end of new lean access

// Handling comments
   bool get(aux::TextH& comment)const;
      // getting a text ment as a comment to the whole Record
   bool get(const Word& section, aux::TextH& comment)const;
      // getting a text ment as a comment to a section
   bool get(const Word& section, const Word& item,
      aux::TextH& comment)const;
      // getting a text ment as a comment to a section-item pair

// setting the value at a position determined by section and item.
// If section or item was not yet created earlier they become installed
// now. The return value is true for success and false for failure.
// In the present implementation success seems to be guaranteed
// and we return true in any case.

   bool set(const Word& section, const Word& item, const bool& value);
   bool set(const Word& section, const Word& item, const V<bool>& value);
   bool set(const Word& section, const Word& item, const V<B>& value);

   bool set(const Word& section, const Word& item, const Z& value);
   bool set(const Word& section, const Word& item, const V<Z>& value);
   bool set(const Word& section, const Word& item, const V<Z1>& value);

   bool set(const Word& section, const Word& item, const R& value);
   bool set(const Word& section, const Word& item, const V<R>& value);
  // #if !defined(CPM_ULM)
   bool set(const Word& section, const Word& item, const V<R1>& value);
  // #endif

   bool set(const Word& section, const Word& item, const Word& value);
   bool set(const Word& section, const Word& item, const V<Word>& value);

// Handling comments
   bool set(const aux::TextH& comment);
      // introducing a text as a comment to the whole Record
   bool set(const Word& section, const aux::TextH& comment);
      // introducing a text as a comment to a section
   bool set(const Word& section, const Word& item,
      const aux::TextH& comment);
      // introducing a text as a comment to a section-item pair

   bool isVoid()const
      { return (rep0.isVoid() && rep1.isVoid()
         && rep2.isVoid() && rep3.isVoid()&&vec3.isVoid());}

   bool isVal()const{ return valid;}

   Record operator()(Word const& sec)const;
      // returns all entries of the section sec of *this
      // as a single-section record which is endowed with a
      // suitable overall comment.

   Record operator()(Word const& sec1, Word const& sec2)const;
      // returns all entries of the section sec1 of *this
      // as a single-section record the single section of which
      // is named sec2.

   M<aux::Key,V<B> > getBs()const{ return vec0;}
   M<aux::Key,V<Z1> > getZs()const{ return vec1;}
   M<aux::Key,V<R1> > getRs()const{ return vec2;}
   M<aux::Key,V<Word> > getWs()const{ return vec3;}
};

///////////////////// class Recordable ///////////////////////////////
/*
This class arrose as an abstraction from my first implementation
of class R_FuncRec, which now is implemented as being derived
from Recordable. One could even carry the abstraction one step
further and make class Record a template argument.

This class uses Record to hold the complete state information
and implements the CPM_IO and CPM_ORDER interfaces based on
this.

One could consider to implement Recordable as a class derived from
Record. I prefer not to do this in order to be free in the
implementation of CPM_IO and CPM_ORDER which should here be based
on Record::readFormatted and Record::writeFormatted
instead of Record::scanFrom and Record::prnOn.
*/

class Recordable{// deriving from this class provides IO for complex data

   mutable bool valid_;
   mutable Word rep_;
      // this is a text representation of the content on which
      // the order operations are based. This is a dependent
      // quantity that will be set with function adjust().
      // Notice that stream related operations work with rc_.
protected:

   Record rc_;
      // this is -by definition- the only independent
      // information that makes up the state of an instance
      // of the class.

   virtual bool vadjust()const{ return true;}
      // setting all additional (dependent)
      // data members if rc_ is given
      // returns true if it was successful. For enabling
      // the definition of those functions as const which
      // are expected to be constant, all dependent data
      // members to be added in derived classes have to
      // be declared mutable (just as the attribute 'valid' above).
      // See R_FuncRec::vadjust() for a typical redefinition
      // of this function.

   bool adjust()const;
      // sets valid_, rep_, and (by calling vadjust) all
      // additional dependent data members
      // so that they are in accordance with the data in
      // rc_. If valid_==true at calling the function,
      // the call has no effect and is not expensive.
      // The function is const in the sense that it does not change
      // the real content. It does, however change the representation
      // of this content, which is possible due to mutable data
      // members.

   void renew_(){ valid_=false;}
      // setting valid_ to false makes the next call of
      // adjust busy. Any function in a derived class which
      // changes rc_ (which is possible since we chose to define
      // rc_ as protected) has to call renew_ in order to
      // let adjust do the necessary adjustments.

   typedef Recordable Type;

public:

   CPM_IO
   CPM_ORDER
   virtual Word nameOf()const{ return Word("Recordable");}

   explicit Recordable(Record const& rc=Record()):rc_(rc),valid_(false){}
      // if the constructor would be allowed to call a virtual function
      // one could form a valid instance already here. Things being as
      // they are enforce that adjust can be called only in calls
      // of member functions. See class R_FuncRec (cpmrfuncrec.h)
      // as an example for this mechanism

   Recordable(Record const& rc, Word const& sec):rc_(rc(sec)){}
      // generating a Recordable from one section of a Record

   Record rc()const{ return rc_;}

// access to Boolean content

   void setB(Word const& section, Word const& name, bool x)
      { rc_.setB(section,name,x); renew_();}

   bool getB(Word const& section, Word const& name)const
      { return rc_.getB(section,name);}

   void setBs(Word const& section, Word const& name, V<bool> const& x)
      { rc_.setBs(section,name,x); renew_();}

   V<bool> getBs(Word const& section, Word const& name)const
      { return rc_.getBs(section,name);}

// access to integer content

   void setZ(Word const& section, Word const& name, Z x)
      { rc_.setZ(section,name,x); renew_();}

   Z getZ(Word const& section, Word const& name)const
      { return rc_.getZ(section,name);}

   void setZs(Word const& section, Word const& name, V<Z> const& x)
      { rc_.setZs(section,name,x); renew_();}

   V<Z> getZs(Word const& section, Word const& name)const
      { return rc_.getZs(section,name);}

// access to real content

   void setR(Word const& section, Word const& name, R x)
      { rc_.setR(section,name,x); renew_();}

   R getR(Word const& section, Word const& name)const
      { return rc_.getR(section,name);}

   void setRs(Word const& section, Word const& name, V<R> const& x)
      { rc_.setRs(section,name,x); renew_();}

   V<R> getRs(Word const& section, Word const& name)const
      { return rc_.getRs(section,name);}

// access to Word content

   void setW(Word const& section, Word const& name, Word const& x)
      { rc_.setW(section,name,x); renew_();}

   Word getW(Word const& section, Word const& name)const
      { return rc_.getW(section,name);}

   void setWs(Word const& section, Word const& name, V<Word> const& x)
      { rc_.setWs(section,name,x); renew_();}

   V<Word> getWs(Word const& section, Word const& name)const
      { return rc_.getWs(section,name);}

// access to rep_
   Word rep()const // needs to update rep_
   {
      if( adjust())return rep_;
      else {
         cpmerror("Recordable::rep() failed");
         return Word();
      }
   }
};

} // namespace

#endif
