//? cpmrecordhandler.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_RECORDHANDLER_H_
#define CPM_RECORDHANDLER_H_

/*
A tool for creating and initializing instances of class
   CpmRootX::Record from 'ini-files'.
   Suitable for data distribution in parallel computing based on MPI.
   See class RecordHandler.
*/

#include <cpmrecord.h>

namespace CpmRootX{

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::Word;
   using CpmRootX::Z1;
   using CpmRootX::R1;
   using CpmRootX::Record;

   V<Word> combine(V<Word> const& argsConsole, V<Word> const& argsFile);
      // combining commandline info and command line substitute info
      // into a single command line substitute, where console input
      // is given priority. Needed exclusively for implementing the
      //  following function.

   V<Word> extComLine(int argc, char *argv[]);
      //: extended command line
   // This is a building block in the implementation of
   //    int main(int argc, char *argv[])
   // in file cpmapplication.cpp.

   // The argument is actual console input to the command line
   // for starting a console application or the command line
   // surrogate which is made availabe upon (double)
   // clicking on its representation in something like Windows Explorer).
   // This surrogate is equivalent to true console input argc=1
   // and argv[0]=full name(icluding path) of the program's
   // executable. The information to be used for extending the
   // information content of argc and argv is to be taken from an
   // ini-file xxx.ini in the form RecordHandler(xxx.ini).
   // The base name of the ini-file is given by the program's name and
   // the search positions primarily determined by entries in
   // cpmsystemdependencies.h and if a file cpmconfig.ini
   // was found (at a position also determined in cpmsystemdependencies.h)
   // also by a directory name given there.
   // What is to be taken as the program's name is subject to
   // interpretation by the present function: if the actual name contains
   // '_' or uppercase letters then a part of the name may be interpreted
   // as a user created name appendix that is better ignored in searching
   // for the proper ini-file. I had a program for which the IDE created
   // the exeutable name graphicsTest.exe and this program unexpectedly
   // tried to get input from a file graphics.ini. After forcing the IDE
   // to give the executable the name graphicstest.exe it properly asked
   // for input from graphicstest.ini. Notice that it is not the data
   // element 'title_' that determines the name of the program specific
   // ini-file.

   class RecordHandler{ // tool for reading and writing a Record
   // RecordHandler is a tool reading a Record from file putting it
   // into memory and manipulating it there (by the write functions).
   // The read functions access the data in memory. In a SPMD
   // (single program, multiple data ) context, the functionality
   // is such that file opening and reading (by the constructor) is done
   // for rank==1 and the results are broadcasted to the memory of
   // all other ranks.
   // The assumed syntax of the file is described in class Record.
   // Important: Never use
   //    RecordHandler rch;
   // in a context in which
   //    CpmSystem::Message::ini_()
   // has not yet been called. Causes segmentation fault just at the
   // beginning of the program run.
   // Under Visual C++ this remains undetected and causes dark
   // behavior afterwards. Under cygwin the segmentation fault is reported
   // immediately and the program stops. Debugging (compilation option -g,
   // calling gdb run) gave the essential hint.

   void ini_(V<Word> const& files, Word const& path,
      Word const& callMessage );
   // construction tool

public:
   explicit RecordHandler(V<Word> const& files=V<Word>(0),
      Word const& path="", Word const& callMessage="")
   { ini_(files,path,callMessage);}
   // Constructor from a list of file names.
   // If rank==1, the files in the list will be opened in their order
   // until function readFromFile was able to read a valid
   // Record. If this is impossible for all files in the list,
   // the constructed instance res has the property
   // res.isVal()==false. This constructor distributes
   // the data found for rank==1 to all other processes such that
   // following read and write functions work in each process just as in
   // a single process program.
   // The last argument causes
   //       cpmmessage(callMessage);
   // near to the beginning of the defining function block. This helps
   // to stay informed if many ini-files are involved in a program.
   // The argument 'path' has a special meaning, if it equals "fromCon"
   // (meaning 'from configuration'): Then we use Message::getDirBase()
   // for path. This quantity can be set on the last line of the file
   // cpmconfig.ini and thus allows to switch between data sources without
   // recompilation.

   explicit RecordHandler(Word const& file, Word const& path="fromCon",
      Word const& callMessage="")
   { ini_(V<Word>{file},path,callMessage);}
   // As previous function, but for convenience, a single
   // file name has not to be made a vector before it can be used as an
   // argument. Also for 'path' the default argument is given that
   // is probably most useful in most cases.

   explicit RecordHandler(Record const& rc):
   valid(true),rank(1),size(1),rec(rc){}
      // added 2005-04-05. Since much older code uses class RecordHandler
      // in situations in which Record would be sufficient, here is a
      // conversion function:

// writing the fundamental basic types at a position determined by
// section and key.  The return value false indicates failure.

   bool write(const Word& section, const Word& key,
      const Z& value);

   bool write(const Word& section, const Word& key,
      const R& value);

   bool write(const Word& section, const Word& key,
      const Word& value);

   bool write(const Word& section, const Word& key,
      const Z1& value);

   bool write(const Word& section, const Word& key,
      const R1& value);

   bool write(const Word& section, const Word& key,
      const bool& value);

   bool write(const Word& section, const Word& key,
      const B& value);

// vector types (only those with CpmRoot-typed components)

   bool write(const Word& section, const Word& key,
      const V<bool>& value);

   bool write(const Word& section, const Word& key,
      const V<Z>& value);

   bool write(const Word& section, const Word& key,
      const V<R>& value);

   bool write(const Word& section, const Word& key,
      const V<Word>& value);

   bool writeFormatted(ostream& , bool writeComments)const;
      // readable representation of content according to the same grammar
      // as reading from file in the constructor. But different
      // order (alphabetic) than in input

// reading the fundamental basic types from a position determined by
// section and key. A failure in doing so stops the
// program for pedantic==2 and diagnostic messages will be left on file
// cpmcerr.txt. If pedantic==1 a message will be left, but the
// program will not stop. For pedantic==0 there is even no message.
// It is only for return value==true that argument value may get changed.

   bool read(const Word& section, const Word& key,
      Z& value, Z pedantic=2)const;

   bool read(const Word& section, const Word& key,
      R& value, Z pedantic=2)const;

   bool read(const Word& section, const Word& key,
      Z1& value, Z pedantic=2)const;

   bool read(const Word& section, const Word& key,
      R1& value, Z pedantic=2)const;

   bool read(const Word& section, const Word& key,
      Word& value, Z pedantic=2)const;

   bool read(const Word& section, const Word& key,
      bool& value, Z pedantic=2)const;

   bool read(const Word& section, const Word& key,
      B& value, Z pedantic=2)const;

// vector types (only those with CpmRoot-typed components)
   bool read(const Word& section, const Word& key,
      V<bool>& value, Z pedantic=2)const;

   bool read(const Word& section, const Word& key,
      V<Z>& value, Z pedantic=2)const;

   bool read(const Word& section, const Word& key,
      V<R>& value, Z pedantic=2)const;

   bool read(const Word& section, const Word& key,
      V<Word>& value, Z pedantic=2)const;

   Word getFileName()const{ return filename;}
      //: get file name

   bool prnOn(ostream& str)const{ return rec.prnOn(str);}

   Record getRecord()const{ return rec;}

   Record toRec()const{ return rec;}
      //: to record

   Record operator()(void)const{ return rec;}
      // my idiom for accessing the basics

   bool isVal()const{ return valid;}
      //: is valid

   V<Word> comLine()const;
      // read a command line substitute and return it.
      // If nothing was found, we return a V of length 0.

private:
   B valid;
   Z rank;
   Z size;
   Word filename;
      // Filename of the inifile
   CpmRootX::Record rec;
      // replacement for the functionality of Windows ini-files
};

} // namespace
#endif
