//? cpmrmateig.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_RMATEIG_H_
#define CPM_RMATEIG_H_
/*
   cpmrmateig.h   stands for R_Matrix Eigenvectors

   Description:
      Eigenvalues and eigenvectors of a symmetric matrix.

*/
// credit: main algorithm modified from Press. et al.

#include <cpmrmatrix.h>

namespace CpmArrays{

R_Vector eigen(const R_Matrix& a, R_Matrix& eig, bool test=false,
   Z iterMax=30);
/*
Let a be a symmetric (n,n)-matrix and let eig be a defined but not
necessarily meaningful matrix. After a statement like
   R_Vector d=eigenvalues(a,eig)
eig will be a (n,n)-matrix such that
   a*eig[i]=d[i]*eig[i]
i.e. the return vector contains the eigenvalues and the matrix eig is
formed by the eigenvectors. The enumeration of eigenvalues is in
descending order (this is usefull when applying the method for
orthonormalization of function systems with respect to a given scalar
product) and the open +1/-1 phase factor of eigenvectors (if all
eigenvalues are different, which is the natural case, there are no more
degrees of freedom for the eigenvectors) is chosen such that the largest
component (in absolut value) of every eigenvector is positive. Let D :=
diag(d[1],...d[n])
then
   a=~eig*D*eig    (position of ~ is important !)
i.e. the matrix eig also performs the principal axis transformation.
We have: (~eig)*eig=eig*(~eig)=(n,n)-unit matrix. Thus this
is a special case of a singular value decomposition. However the SVD of
a will not always coincide with this decomposition ??? not yet clarified
UM 99-6-8: This is 'the' singular value decomposition if the diagonal
elements of D are non-negative. If some elements are negative, we change
the sign of the corresponding line in in the right factor eig and obtain
a=~eig*|D|*eigmod. Both, eig and eigmod are orthogonal and this, then,
is a SVD.
UM 99-12-18: If the test agument is set to true, an error-value,
indicating the accuracy of the eigenvectors and eigen values will be
written on file cpmcerr.
UM 00-1-5: the test is modified such that the completeness of the
eigenvectors is tested too. This is done by testing
(~eig)*eig = eig*(~eig) = (n,n)-unit matrix.
*/

R scalar_product(const R_Vector& x, const R_Vector& y,
   const R_Vector& d);
/*
Scalar product overloaded such that a differential operator is applied to
the arguments as in Sobolev spaces.
*/

void adjust_scalar_product(
// input
   const R_Vector& x, // vector used for adjusting
   R a, // norm that the vector x shall get
   R_Vector w,     // gives the relative weight of the contributions
      // to <x|x>
// output (defined but meaningless as input)
   R_Vector& d // vector defining the adjusted scalar product
);
/*
With the d obtained from this routine, we will have
   scalar_product(x,x,d) = a*a
and
   contrib[1]:contrib[2]:...:contrib[n]=w[1]:w[2]:...:w[n],
where the contributions are defined by
   scalar_product(x,x,d) = d[1]*scalar_product(x,x)
   +d[2]*scalar_product(x.der(1),x.der(1)
   +d[3]*scalar_product((x.der(1)).der(1),(x.der(1)).der(1))
      ...
   +d[n]*scalar_product(x der(1) applied n-1 times,...)
   =: contrib[1]+...+contrib[n].
*/

R_Vector projection(const R_Vector& x, const R_Matrix& f,
   const R_Vector& d, Z i);
/*
The matrix f describes in the standard way an orthogonal function system
f[1], ...f[m]. The return vector is the orthogonal projection of x on
span(f[1],...,f[m]), with respect to the scalar product
<x|y> := scalar_product(x,y,d).
This is the case for all values of i except i=-1. Then the projection on
the orthogonal complement of span(f[1],...,f[m]) is returned.
Although it is easy to write down the result of such a projection (see
code), it is harder to write down an expression for the projection
operator as a matrix. This will not be carried out here since it seems
not to be needed.
*/

void ortho_system(const R_Matrix& fs, const R_Vector& d, R_Matrix& os);
/*
The matrix fs describes in the standard way a function system fs[1],
 ...fs[m]. We adopt the scalar product <x|y> := scalar_product(x,y,d).
By forming suitable linear combinations we obtain the orthonormal
function system os (which is defined but meaningless on input and having
the stated meaning on output) by a principal axis transformation of the
matrix
   g[i][j]=<fs[i]|fs[j]>/sqrt(<fs[i]|fs[i]><fs[j]|fs[j]>).
*/

R testR_MatEig(const R_Matrix& a);
   // Performs a diagonalisation of (a+~a)/2 and outputs
   // an error sum which should be <<1.

class SpcObjR{
// Eigen values and eigenvectors of a symmetric real (n,n)-matrix
// Seems to give a few powers of 10 larger errors than the complex
// version. Not yet clarified.
   Z n_;
      // 'dimension of the problem'
      // val_.dim() == vec_.dim1() == vec_.dim2() == mat_.dim1()
      // == mat_.dim2();
   R_Matrix mat_;
      // Matrix for which eigenvalues and eigenvectors are requested
   R_Matrix vec_;
      // The rows of this matrix are the eigenvectors
   R_Vector val_;
      // The components of this vector are the eigenvalues

public:
   explicit SpcObjR(const R_Matrix& a, bool enfSym=false);
      // creates a instance of SpcObjR which provides all quantities which
      // result from a spectral representation of a. If such an
      // instance is created once, several 'services' like finding f(a)
      // for arbitrary f:R->R can be provided without much computational
      // burden.
   R_Matrix func(const F<R,R>& f)const;
      // returns f(mat_)
   R_Matrix operator()(const F<R,R>& f)const{ return func(f);}

// simple access functions
   R_Matrix mat()const{ return mat_;}
   R_Matrix vec()const{ return vec_;}
   R_Vector val()const{ return val_;}

   R operator()(Z i)const{ return val_[i];}
   R_Vector operator[](Z i)const{ return vec_[i];}

   Z dim()const{ return n_;}

   R getError()const;
      // defined such that the result should deviate from 0 only by
      // numerical inaccuracies. Any deviation from the defining
      // properties of a spectral representation will be detected.
};

} // namespace

#endif
