//? cpmrnfunc.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_RNFUNC_H_
#define CPM_RNFUNC_H_
/*

Description: Working out in C++ the topic from 
   page 6 (03-05-14) of my yellow Konzeptheft 

*/

#include <cpmfa.h>
#include <cpmrvector.h>
#include <cpmalgorithms.h>

namespace CpmFunctions{

   using CpmRoot::R;
   using CpmRootX::R1;
   using CpmLinAlg::R2;
   using CpmRoot::Z;
   using CpmFunctions::F;
   using CpmArrays::R_Vector;
   using CpmArrays::V;
   using CpmAlgorithms::bitsFromZ;

/*********************** class Rn<n>  ********************************/
   // class of real n-tupels. Unlike in R_Vector, the dimension
   // of the space is fixed (by a template parameter)
   // class will form the domain of the following function class
   // Rn_Func<n>
   // Presently this is exploratory. If it should be seriously used,
   // one had to specialize Rn<n> for n=1,2,3,4,5,6
   // to CpmLinAlg::R1,CpmLinAlg::R2,CpmLinAlg::R3,...CpmLinAlg::R6 

template <Z n>
class Rn: public R_Vector{ // R_Vector with specified dimension 

   typedef Rn<n> Type;
   typedef R_Vector Base;
   static Z ok(Z i) // constant continuation outside the 
         // primary index range
   { if (i<1) return 1; else if (i>n) return n; else return i;}

   static Type fFunc(const Z& i, const Type& a, const Type& b, 
      const bool& random);
      // basic function for implementing f

public:

   Rn():Base(n){}

   Rn(const Base& x):Base(n)
   { Z i=1; while( i<=n && i<=x.dim() ){ (*this)[i]=x[i]; i++;} }

   R vol(const Type& y);
      // a.vol(b) is the product over i of
      // b[i]-a[i]. 
      // It is, up possibly to a sign equal to the volume of
      // the n-rectangle with corners a and b

   static R1 c(const Type& x, const Z& i){ return R1(x[ok(i)]);}
      // basic function Type x Z --> R1
      // for forming the component functions in the next class

   static R1 volFunc(const Type& x);
      // returns the product of the components of x

// means for discrete exhaustion of Type.
// The basic feature of Type is that a very general useful subset
// of type can ge characterized by only two instances of Type 
// by a natural geleralization of the concept of an interval in R.
// The arguments a and b in the following two functions just play 
// this role.

   static V< Type > v(Z m, const Type& a, const Type& b,
      bool random=true );
      // Let
      // Z m=...; Rn<4> a=...; Rn<4> b=...;
      // Let V< Rn<4> > x=Rn<4>::v(m,a,b);
      // then x.dim()==m and 
      //  inf<R>(a[j],b[j])<=x[i][j]<=sup<R>(a[j],b[j]) for all
      // i \in {1,...,m} and j \in {1,...,n}.
      // The sequence forms an equi-spaced lattice for random=false
      // (not yet implemented) and a random cloud for random=true
      // (implemented).
      
   static  F<Z, Type > f(const Type& a, const Type& b, bool random=true);
      // same as above with x[i] repalced by f(i), so the length has not 
      // to be given. Actually this is the basic algorithm and the
      // previous
      // function is implemented through the present one.
};

template <Z n>
R Rn<n>::vol(const Type& y)
{
   R ai,bi,res=1;
   for (Z i=1;i<=n;i++){
      ai=(*this)[i];
      bi=y[i];
      res*=(bi-ai);
   }
   return res;
}

template <Z n>
R1 Rn<n>::volFunc(const Type& x)
{
   R res=1;
   for (Z i=1;i<=n;i++) res*=x[i];
   return R1(res);
}

template <Z n>
Rn<n> Rn<n>::fFunc(const Z& i, const Type& a, const Type& b,
 const bool& random)
{
   Type res;  
   if (random){
      Z shift = n==137 ? 127:137;
      const Z iMax=137731127;
      Z i0=i;
      for (Z j=1;j<=n;j++){
         res[j]=a[j]+CpmRoot::randomR(i0)*(b[j]-a[j]);
         i0+=shift;
         if (i0>iMax) i0=0;
      }
   }
   else{
      cpmerror("Rn<n>::fFunc(...random=false) not implemented");
   }
   return res;
}

template <Z n>
F<Z, Rn<n> > Rn<n>::f(const Type& a, const Type& b, bool random)
      // same as above with x[i] repalced by f(i), so th elength has not 
      // to be given
{
   return F3<Z,Type,Type,bool,Type >(a,b,random)(Type::fFunc);
}

template <Z n>
V< Rn<n> > Rn<n>::v(Z m, const Type& a, const Type& b, bool random)
{
   F<Z, Type> ff=Type::f(a,b,random);
   V< Type > res(m);
   for (Z i=1;i<=m;i++) res[i]=ff(i);
   return res;
}
/////////////////////////// class Rn_Func<n> /////////////////////////////

template <Z n>
class Rn_Func: public Fa<Rn<n>,R1>{ // R1-valued functions on Rn<n>
   // R1-valued functions on Rn<n>  with arithmetics and trivially
   // defined order properties, and construction from discrete data via
   // interpolation.

   typedef Rn<n> Dom;
      // domain, where the function is defined
   typedef R1 Ran;
      // 'range' (co-domain), type to which the values of the function
      // belongs
   typedef Rn_Func<n> Type;
   typedef Fa<Dom,Ran> Base;
      // 'a' stands for arithmetic, defines addition and multiplication
      // of functions

   static V<Type> iniX();
      // initializes the array x of component functions 
   static Ran parDer(const Dom& x, const Type& f, const Z& i, 
      const R& h, const bool& lean);

   static Ran fullDer(const Dom& x, const Type& f, const R& h);

   static Ran linFunc(const Dom& x, const Dom& y)
   {
      R res=0;
      for (Z i=1;i<=n;i++) res+=x[i]*y[i];
      return Ran(res);
   }

   static Ran shiftFunc(const Dom& x, const Type& f, const Dom& s)
   {  return f(x-s);}

public:

// carrying over the constructors from Fa<R,R>

   Rn_Func (const F<Dom,Ran>& g):Base(g){}
      // most universal constructor. Cooperates with e.g.
      // F2<Dom,X1,X2,Ran>

   Rn_Func (const Base& g):Base(g){}
      // upcast constructor

   Rn_Func (const Type & g):Base(g){}
      // copy constructor

   Rn_Func (void):Base(){}
      // default constructor

   explicit   Rn_Func ( Ran (*f)(const Dom &)):Base(f){};
      // constructor from native function pointer

   explicit   Rn_Func ( Ran (*f)(Dom)):Base(f){};
      // constructor from native function pointer

   explicit   Rn_Func (R r):Base(Ran(r)){};
     // constructor for constant function

   explicit   Rn_Func (const Dom& a):Base(F1<Dom,Dom,Ran>(a)(linFunc)){}
     // constructor for linear form determined by a

   Type shift(const Dom& s)const
      { return F2<Dom,Type,Dom,Ran>(*this,s)(shiftFunc);}
      // If f has a peak at xP, f.shift(s) has a peak at xP+s !
      // Thus we need a '-' (sic!) in the definition of shiftFunc.

   Type der(Z i, R h=1e-3)const;
      // returns the partial derivative of function *this
      // with respect to component i

   Type der(R h=1e-3)const;
      // f.der() is the complete mixed derivative 
      // f.der(1).der(2).... .der(n)

   Type derX(R h=1e-3)const;
      // returnes the complete mixed derivative
      // according to a second implementation based on 
      // evl2. This consists of 2^n terms with 'alternating' signs
      // so that for n>4 one may get no reliable results for any h.
      // Before this bad situation occurs, one may find varying h
      // at least a platau-dependence on h over, say, a order of
      // magnitude; the stable plateau value may than be taken
      // as the value of the function. This remark also applies to
      // function der, but there the dependence on h seems to be 
      // less critical.

   Ran evl2(const Dom& x, const Dom& y )const;
      // the two argument evaluation 
      // f.eval2(a,b) generalizes f(b)-f(a)
      // basic construct for my formulation of the 'fundamental
      // theorem of calculus' in n dimensions

   static R test1(Z i=1000);
      // should give very small values for all i
      // Volume of a rectangle can be expressed by eval2()
      // acting on the volume-generating function

   static R2 test2(R h=1e-3);
      // should give a pair of values which are close together
      // presently based only on one function as which 'waird'
      // is used since the random generator is not yet implemented
      // in a useful manner

   static V<Type> x;
      // list of component functions
      // x[i].der(j)=kronecker(i,j)

   static Type waird(Z i=1000);
      // For each i this is a
      // product of n random generated affine-linear functions.
      // Thus a rather general n-order polynomial in the 
      // n component functions Type::x[1],...Type::x[n].
};


template <Z n>
R1 Rn_Func<n>::evl2(const Dom& x, const Dom& y )const
{ 
   Z i, j, p=Z1(2).pow(n);
   R si,sum=0;
   for (i=0;i<p;i++){
      V<bool> b=bitsFromZ(i,n);
      Z par=1;
      Dom z;
      for (j=1;j<=n;j++){
         bool bj=b[j];
         if (bj){
            par*=-1;
            z[j]=x[j];
         }
         else{
            z[j]=y[j];
         }
      }
      si=par*evl(z);
      sum+=si;
   }
   return R1(sum);
}

template <Z n>
R Rn_Func<n>::test1(Z i)
{
   R size=2;
   R_Vector x(n,size);
   R_Vector x1=x.ran(i);
   R_Vector x2=x.ran(i+1);
   Dom y1(x1);
   Dom y2(x2);
   Type vol(Dom::volFunc);
   R1 r1=vol.evl2(y1,y2);
   R1 r2=R1(y1.vol(y2));
   return r1.dis(r2);
}

template <Z n>
R2 Rn_Func<n>::test2(R h)
{
   R size=2;
   R_Vector x(n,size);
   R_Vector x1=x.ran();
   Dom y1(x1);
   Type w=waird();
   Type f1=w.der(h);
   Type f2=w.derX(h);

   R r1=f1(y1);
   R r2=f2(y1);
   return R2(r1,r2);
}

template <Z n>
R1 Rn_Func<n>::parDer(const Dom& x, const Type& f, const Z& i, 
   const R& h, const bool& lean)
{
   static Word loc("Rn_Func<n>::parDerparDer(...)");
   cpmassert(i>=1 && i<=n,loc);
   cpmassert(h!=0,loc);
   R h2=2*h;
   Dom x_=x;
   R y;
   if (lean){
      x_[i]+=h;
      y=f(x_); 
      x_[i]-=h2;
      y-=f(x_); 
      y/=h2;
   }
   else{
      x_[i]-=h2;
      y=f(x_); x_[i]+=h;
      y-=8*f(x_); x_[i]+=h2;
      y+=8*f(x_); x_[i]+=h;
      y-=f(x_);
      y/=(12*h);
   }
   return R1(y);
}

template <Z n>
R1 Rn_Func<n>::fullDer(const Dom& x, const Type& f, const R& h)
{
   static Word loc("Rn_Func<n>::fullDerparDer(...)");
   cpmassert(h!=0,loc);
   Dom y=x;
   R z=1;
   for (Z i=1;i<=n;i++){
      y[i]+=h;
      z*=h;
   }
   return f.evl2(x,y)/z;
}


template <Z n>
Rn_Func<n> Rn_Func<n>::waird(Z j)
{
   R size=2;
   Z i,ic=j;
   V<R_Vector> tv(n,R_Vector(n,size));
   for (i=1;i<=n;i++){
      tv[i]=tv[i].ran(ic); 
      ic++;
   }
   V<R_Vector> sv(n,R_Vector(n,size));
   for (i=1;i<=n;i++){
      sv[i]=sv[i].ran(ic);
      ic++;
   }
   Type res(1.); // function constant 1
   for (i=1;i<=n;i++){
      Type fi=Type(tv[i]).shift(sv[i]);
         // shifted linear form
      res*=fi; // product from Fa<>, works !
   }
   return res;
}

template <Z n>
V<Rn_Func<n> > Rn_Func<n>::iniX()
{
   V<Type > res(n);
   for (Z i=1;i<=n;i++){
      res[i]=F1<Dom,Z,R1>(i)(Dom::c);
   }
   return res;
}

template <Z n>
V<Rn_Func<n> > Rn_Func<n>::x=Rn_Func<n>::iniX();

template <Z n>
Rn_Func<n> Rn_Func<n>::der(Z i, R h)const
{
   F<Dom,R1> f=F4<Dom,Type,Z,R,bool,R1>(*this,i,h,true)(Type::parDer);
   return Type(f);
}

template <Z n>
Rn_Func<n> Rn_Func<n>::derX(R h)const
{
   F<Dom,R1> f=F2<Dom,Type,R,R1>(*this,h)(Type::fullDer);
   return Type(f);
}

template <Z n>
Rn_Func<n> Rn_Func<n>::der(R h)const
{
   Type f=*this;
   for (Z i=1;i<=n;i++){
      f=f.der(i,h);
   }
   return f;
}

} // namespace

#endif
