//? cpmroot.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_ROOT_H_
#define CPM_ROOT_H_

/*

Purpose: Root finding for a real-valued function of a real
   argument in 'modern' form, where the function is
   represented as a instance of F<R,R> which is a data member of
   class FindRoot.
   Notice that this includes functions which depend on an additional
   parameter which considerably widens the scope of the method.

*/

#include <cpmtypes.h>
#include <cpmf.h>

namespace CpmAlgorithms{

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmFunctions::F;

/************* Typical usage of class FindRoot ***************************

FindRoot r(F<R,R>(cos),0,1);
   // sets up the problem to find a solution of cos(x)=0 for 0<=x<=1.

R myPi=2*r.getRoot();
   // should be PI

*/

/////////////////////////// class FindRoot ///////////////////////////////

class FindRoot{  // finds a root of a function

public:

   FindRoot(const F<R,R>& f, R x1, R x2, R eps=1e-9);
      // setting the function under consideration and the interval
      // in which the root is searched first (will be extended if
      // no root can be found there) . Also the accuracy of the
      // root can be set

   FindRoot(R x1=0, R x2=1, R eps=1e-9);
      // also default constructor, does not specify a function, the root
      // of which should be found (the previous constructor does). This
      // can be done by function setF()

   void setLimits(R x1, R x2);
      // sets the interval on which root searching is based
      // correct order is assured internally.

   void setAcc(R eps);
      // sets the relative accuracy of the root

   void setF(const F<R,R>& f);
      // setting the 'function' ff for which a root is to to be found.

   R getRoot();
      // finds and returns the root x0. Provisions against calling this
      // function for undefined ff are made.

   R operator()(void){ return getRoot();}

   R getRootCrit(Z& success);
      // computes and returns the root x0
      // sets success to 0 if no root was found .
      // Provisions against calling this function for undefined ff are
      // made.

   R getValue()const;
      // returns ff(x0);
      // Can meaningfully be called only after one of the
      // previous functions was called, so that a root finding process
      // was already performed.

private:

   R x0, y0, xL, xU, tol;
   F<R,R> ff;
      // 'function' the root of which has to be found
   void solve(void);
      // activates root finding
   void solveCrit(Z& success);
      // activates root finding
      // sets success to 0 if no root was found
   Z done;
   Z ffDefined;
   Z brac(R* x1, R* x2);
   R brent(R x1, R x2);
};

} // namespace

#endif
