//? cpmunits.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_UNITS_H_
#define CPM_UNITS_H_

// Values of non-SI physical units expressed as multiples of the
// corresponding SI units

// Source: (1) For 'emu' matters: John Crangle, Mike Gibbs: Units and
// unity in magnetism: a call for consistency, Physics World November
// 1994 p. 31-32
// (2) Abramowitz, Stegun: Handbook of Mathematical Functions, Chapter
// 2.2 (p. 6)
// In the explanation of an entry, we give the SI unit first, and then an
// description of the physical quantity measured in this unit

#include <cpmnumbers.h>

namespace CpmUnits{

   using CpmRoot::R; // real numbers

const R gauss=1e-4;
   // T=Tesla=V*s/m^2 (magnetic polarization J, and magnetic flux density
   // B, B = B0 + J)

const R oersted=1000.;   // divide this number by 4*pi if 'oersted'
// refers to the
   // 'rationalized CGS-system' instead of the 'unrationalized' one.
   // A/m        (magnetic field strength H )

const R emu_per_cm=1000.;
   // J * T^-1 * m^-3 = A/m ( M = Magnetization  =
   // = (Ampere's)magnetic moment 'per unit volume',  B = B0 + mu0*M )
   // notice [M]=[H],  mu0 = 4*pi*1e-7 V*s/(A*m).
   // For convenience, we mention here the sibling of mu0 too:
   // eps0=1/(c^2 * mu0)  = permittivity of vacuum  = dielectric constant
   // of vacuum, c = speed of light in vacuo := 299792458 m/s
   // [eps0] = (A*s)/(V*m)
   // Remember eps0 * mu0 * c^2 =1. Since c is large in SI units , both
   // eps0 and mu0 are very small in SI units:
   // mu0 = 1.256637e-6 H/m (this is to infer H=Henry=V*s/A)
   // eps0 = 8.854188e-12 F/m (F=Farad=A*s/V)
   // Also remember Wb=Weber=T*m^2

const R emu_per_g = 1.;
   // J/(T*kg) = A * m^2 / kg  ( sigma = M/rho = Ampere's magnetic moment
   // 'per unit mass' ). Example: Electrophotographic carrier particles
   // are reported to have a remanent magnetization of 30 emu/g and a
   // density of 3.5 g/cm^3. What is the polarization J (Coulomb's
   // magnetic moment 'per unit volume') ?
   // sigma=30 emu/g = 30 A*m^2/kg, M = sigma * rho = 30 * 3500 A/m
   // J = M*mu0 = 30 * 3500 * 4pi*1e-7 = 0.132 V*s/m^2 = 0.132 T

const R inch=2.54e-2;
   // m

const R foot=inch*12;
   // m

const R mil=inch*0.001;
   // m

const R gallon=3.785e-3;
   // m^-3
   // abbreviation gal, this is the US gallon, there is a UK gallon too

const R pound=0.4536;
   // kg

const R ounce=pound/16;
   // kg

const R RPM=cpmpi/R(30);
   // rad/s

//const R degree=CpmRoot::Pi/180;
   // rad, now CpmGeo::degree defined in cpmangle.h

const R cm=1e-2;
   // m

const R mm=1e-3;
   // m

const R micrometer=1e-6;
   // m

const R nm=1e-9;
   // m

} // namespace

#endif
