//? cpmvectormore.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_VECTOR_MORE_H_
#define CPM_VECTOR_MORE_H_
/*
       additional topics for R_Vector (and also to R_Matrix)
*/

#include <cpmrmatrix.h>
#include <cpmalgorithms.h>

namespace CpmArrays{
   using CpmLinAlg::Z2;
   using CpmLinAlg::R2;

R2 lineFit(R_Vector const& x, R_Vector const& y,
   R_Vector const& w);
   //: line fit
   // Fits a straight line to weighted data.
   // Returns R2(a,b) such that
   // sum from i=1 to i=Min(x.dim(),y.dim(),w.dim()) of
   //    |w[i]|*( a + b*x[i] - y[i] )^2
   // is minimal. Modfied from function fit in Press et al.
   // Added 2010-09-05.

R evalPoly(R const& x, R_Vector const&  c);
   // evaluates the real polynomial function determined by data c
   // at position x. This polynomial function is:
   //    p(x)=c[3]+z*c[4]+z*z*c[5]+z*z*z*c[6]+ ...,
   //    where
   //       z:=(x-x0)*rInv, x0:=c[1], rInv:=c[2].
   // This is the same format in which the next fit function
   // returns its result. Notice that writing the polynomial function
   // simply as a polynomial in x leads to coefficients varying
   // over many order of magnitude with the index position in cases where
   // the value of x is large in the whole range under consideration.
   // Especially the determination of these coefficients by any fit
   // becomes numerically unstable.
   // Actually, we deal here with a collective description of a type of
   // functions, which calls for becoming transformed into a class!
   // Later ??

R_Vector polyFit(R_Vector const& x, R_Vector const& y,
   Z deg, R x0=0., R r=1., bool autoAdjust=true );
   // Least mean square fit of a polynomial function of degree deg
   // to data (x[i],y[i]).
   // Let res be the return value, then
   // t |--> evalPoly(t,res) is the function that fits the data.
   // The first 2 components res[1], res[2] of res are either equal
   // to x0 and 1/r (for autoAjust==false), or they are
   // (x[1]+x[n])/2 and abs(2/(x[n]-x[1])) (for autoAjust==true).
   // Let res be the return value. Then res.dim()=deg+3 and
   // the polynomial function x |--> res[3]+z*res[4]+z*z+res[5]+...,
   //    z=(x-x0)/r
   // is the least mean square fit to the list (x[i],y[i]).
   // Elegant application of the pseudo-inverse. Let us explain it in
   // the general case of linear fit with respect to an arbitrary system
   // of functions (not just powers of x as in the present case). Let
   // f[i], i=1,...,d be the system of functions. We define the matrix
   // a[i][j]:=f[i](x[j]). We ask for coefficients c[i],i=1,...,d such
   // that y[j]-Sum(i=1,d) c[i]*f[i](x[j])=y[j]-(~a*c)[j]
   // becomes as small as possible in the mean over j. Thus we search the
   // vector c of coefficients such that y '=' ~a * c. The best solution
   // to this is c = !~a * y , where !a is the pseudo-inverse of matrix
   // a.
   // For the case of f[i](x):=((x-x0)/r)^(i-1) we can obtain the matrix
   // a by the elegant algorithm (see implementation code).

R_Vector linFit(const V<F<R,R> >& f, R_Vector const& x,
   R_Vector const& y);
   //: linear fit
   // Least mean square fit by a linear combination
   // c[1]*f[1]+...+c[n]*f[n], n=f.dim()
   // of a given family f of functions to data (x[i],y[i]).
   // The array c is the return value.

R_Vector nonLinFit(const F<X2<R,R_Vector>,R>& f,
   R_Vector const& x, R_Vector const& y, R_Vector const&  pStart,
   R fracTol=1e-3, bool verbose=false);
   //: non-linear fit
   // Least mean square fit by a function x|--> f(X2(x,p)) to data
   // (x[i],y[i]). Here, R_Vector p is determined from minimizing misfit.
   // p is the return value.
   // The minimizing method needs a starting value for p and a tolerance
   // limit on input, these things are provided with the last two
   // arguments. Notice that the dimension of p equals that of pStart
   // by definition.
   // If verbose=true, some diagnostic information is written on cpmcerr.

R polyInte_4(R_Vector const& x, R_Vector const& y, R x0, R& error);
   // Polynomial interpolation.
   // Name appendix _4 says that the fourth argument is a non-constant
   // reference.
   // Modified from polint() of Press et al.
   // x is a list of x-values, all are assumed to be different (i.e. list
   // is injective). y is a list of at least equal length of
   // corresponding y-values. Returned is the value of the interpolating
   // polynom at/ x-position x0. After function call, error is an error
   // estimate for the function value as indicated in Press et al.

R polyInte(R_Vector const& x, R_Vector const& y, R x0);
   // Version of the last function which has more natural argument list
   // by not indicating an error estimate (which is not too attractive an
   // concept here anyway).

R ratiInte_4(R_Vector const& x, R_Vector const& y, R x0, R& error);
   // rational interpolation
   // Name appendix _4 says that the fourth argument is a non-constant
   // reference.
   // 'Diagonal rational interpolation' according to Bulirsch and Stoer.
   // Modified from function ratint() of Press et al.
   // Same meaning of arguments as in polyInte(). Creates a warning and
   // returns a large value if x0 is located at a pole of the
   // interpolating function.

R ratiInte(R_Vector const& x, R_Vector const& y, R x0);
   // As previous function but no error indication.

Z  locate(R_Vector const& x, R x_);
// Let x be ascending:  x[1] < x[2] <...< x[n]. For a given
// real number x_, we find the largest integer j such that x[j] <= x_.
// This j is the return value of the function. If all x[...] are larger
// than x_, we return 0. Algorithm is most effective for equidistant or
// nearly equidistant lists.

R locatePeak(R_Vector const& x, R p=2.);
// The values |x[i]|^p are considered defining a probability
// distribution over [1,....n] (n=x.dim()) and the expectation value
// of the 'position i' is calculated. If all x[i]=0, no distribution
// is defined and we return 0 (from definition all regular return
// values are >= 1 and <= n)

R2 peakAndSigma(R_Vector const& x, R p=2.);
// Just as previous function but returns the standard deviation
// (half-width of the peak) as the second component of the result

// The following functions smoothen() ... interpolateInverse() interprete
// R_Vector& x as a list of function values belonging to a argument
// values 1,....x.dim().

R_Vector smoothen(R_Vector const& x, bool periodic=false);
// fast convolution with binomial kernel (1 2 1)/4
// end points treated as if data would be extended as constant
// unless argument periodic equals true

R_Vector smoothen(R_Vector const& x, Z p, bool periodic=false);
// applying previous function p times, p>=1
// Returns x for p<=0

X2< V<Z>, V<Z> >  locExt(R_Vector const& x);
//: local extrema
// Let res be the return value of the function. Then
// res.c1() is a list of local minima, and res.c2() a list
// of local maxima. Take, for instance, the first:
// V<Z> a=res.c1();
// Then there are a.dim() local minima of x at indexes
// i=a[1], a[2], ...
// and the x values there are, of course
// x[a[1]], x[a[2]], ...
// In order to obtain only robust extremal values, not created
// by numerical noise, it may be advisable to apply function
// smoothen(...) to x before handing x over to the present function.
// Extrema here are strict: i rightfully appears in the minimum
// list res.c1() iff
//    i>=2 and i<x.dim() and x[i-1]>x[i] and x[i+1]>x[i]
// and in the maximum list res.c2() iff
//    i>=2 and i<x.dim() and x[i-1]<x[i] and x[i+1]<x[i]
// Thus the lists res.c1() and res.c2() are void for all monotone
// lists x, not only for strictly monotone ones.

R_Vector derive(R_Vector const& x);
// foreward derivation, function assumed to be constant outside
// defined area

T2<R_Vector> derive(R_Vector const& x, R_Vector const& y);
//

R_Vector integrate(R x1, R_Vector const& x);
// Integrate a function from left to right. Start value for result
// at left end given as x1.

R_Vector accumulate(R_Vector const& x, bool average=false);
// res[i]=x[1]+x[2]+...+x[i] for i>2 and mutandis mutatis for
// smaller i and smaller dimension
// Needed e.g. for cumnulative histograms.
// notice derive(accumulate(x))=x
// If the newly introduced second argument is 'true', the
// defining formula is replaced by
// res[i]=(x[1]+x[2]+...+x[i])/i so that a limit may
// exist x's which represents a mildly varying function.

R extMean(R_Vector const& x);
//: extrapolated mean
// Estimating 'the DC component of x' without making use of FFT.
// Estimates a limit of the sequence
// x[1], (x[1]+x[2])/2, (x[1]+x[2]+x[3])/3, ...
// by determining the local maxima and the local minima of this sequence,
// and then returning the mean of the last minimum and the last maximum.
// To make local extrema independent of numerical noise, it may be
// advisable to apply this function to a smoothened version of the
// original data x.

void makePositive(R_Vector& x, R tiny=1.e-3);
// set to zero where function is negative
// more precisely set equal to tiny > 0 where the function is <tiny.
// This is done to obtain stricly monotone (even only slightly
// increasing) functions in monotonize

R_Vector reverse(R_Vector const& x);
// 'interchange left and right'

R_Vector monotonize(R_Vector const& x, Z pSmooth=3);
// Change empiric data which should represent a monotone increasing
// function into a slightly modified data set which is strictly
// increasing.

R interpolate(R ic, R_Vector const& x);
// ic 'continuous index' for i in Z we should get back x[i]
// linear interpolation and constant extrapolation

R interpolate(R xp, R_Vector const& x, R_Vector const& y);
// Interpolation from non-equidistant lists:
// x is a list of x-values and y a list of y-values, xp is the value
// for which an interpolated value is needed.
// linear interpolation and constant extrapolation

R interpolateInverse(R xc, R_Vector const& x);
// we try to find a R ic such that xc=interpolate(ic,x).
// We assume that the list x is strict monotonely ascending.
// This is achieved by getting x from function monotonize.

// The following functions moments() and statistics() interpret
// R_Vector& x as a list of empirical data resulting from a stochastic
// experiment(or process).

R_Vector moments(R_Vector const&  data);
// modified from Press et al. moment().
/*
Given a list of R data (forming a R_Vector) we get a result list res
which gives the most important statistical descriptors for this data set.
These are the  descriptors:

     res[1] average of data (mean value)
     res[2] average deviation (estimating the uncertainty of the mean)
     res[3] standard deviation
     res[4] variance (square of standard deviation)
     res[5] skewness
     res[6] kurtosis
   See function statistics() for definition of skewness and kurtosis.
   Notice that in the present function the conventional (not normalized)
   values of skewness and curtosis  are given just as defined in Press et
   al. as Skew() and Kurt().
*/

R_Vector statistics(R_Vector const&  dat);
/*
Given a list of R data (forming a R_Vector) we get a result list res
which gives the most important statistical descriptors for this data set.
These are the  descriptors:

     res[1] average of data (mean value)
     res[2] average deviation (estimating the uncertainty of the mean)
     res[3] standard deviation
     res[4] variance (square of standard deviation)
     res[5] skewness divided by sqrt(6/n)
     res[6] kurtosis divided by sqrt(24/n)

     res[7] confidence level for the hypothesis that data follow a
      normal distribution (with mean and standard deviation from above).
      This is found from the Kolmogorov-Smirnov test which is considered
      reliable for n>=20.

     res[8] minimum of data
     res[9] median of data
     res[10] maximum of data

Skewness and kurtosis are zero for normal distributions. The
normalization denominators in res[5] and res[6] are the standard
deviation of the skewness and kurtosis of a normal distribution, so that
values greater, say, 2 indicate significant deviation from normality.
Negative skewness indicates that the distribution is broader for small
values than for large ones. Negative kurtosis indicates a 'Tafelberg' and
positive kurtosis a 'Matterhorn'. See Press, Flannery et al. Chapter 13.1
p. 474. for moments and p.490-94 for Kolmogorov-Smirnov test.
*/

R_Vector bimodalAnalysis(R_Vector const&  dat);
// This is still an ad hoc device working well for a set of data
// which strongly cluster arround two different values. The actual
// application case helps to understand the situation. We sample the
// electrical potential along a surface which consists of uniformely
// charged areas with two different values of the areal charge density.
// Due to the presence of field-creating particles the potentials
// are not stricly constant on the sub-areas with constant charge
// density. We want to get the mean values over these subareas without
// knowing to what a position and hence to what a charge density
// the samled potential values (=dat) belong.
// What is actually done is to divide dat into "upper values"
// being closer to dat.sup() and lower values being closer to dat.inf()
// and returning res=R_Vector(2), res[1]=lower_values.mean(),
// res[2]=upper_values.mean().
// Presently no test is implemented whether dat is actually bimodal
// or not.

R_Vector gradient( const R_Matrix& x, R_Vector const& y);
// Gradient vector from sampled values of a scalar function.
// Consider a function f such that y[i]=f(x[i]) i=1,...n
// Then we determine the gradient vector gr such that
// f(x[i])=f(x[1])+(gr|x[i]-x[1]).  i=2,....n, (.|.) scalar product
// Let d the dimension of the x[i], then gr has d components. In order to
// have enough equations one has to have n>=d+1. For the method of
// computation to be used here, more equations than unknowns poses no
// problem. gr is the gradient vector of the function f at point x[1].

R_Matrix gradient( const R_Matrix& x, const R_Matrix& y);
// Gradient matrix from sampled values of a vector-valued function.
// Consider an the situation as above but f vector-valued
// f(x)=(f1,....fp)
// Then we determine the gradient matrix gr such that
// f(x[i])[k]=f(x[1])[k]+ (x[i]-x[1]|gr[k]).  i=2,....n, k=1...p
// Then the same solution as above. It is worthwhile to have this
// serialized version of the function , since each application of the
// previous function calculates the pseudo-inverse of a matrix which is
// determined by the first argument alone, and which has to be calculated
// only once in the present version

V< X2<Z,R> > levelCrossing(R_Vector const& y, R frac,
   bool autoLevel=true, R levelSpezifier=0.99);
// We consider y as defining a function f :[1,n]-->R the graph of
// which is the polygon (1,y[1]),(2,y[2]), ... (n,y[n]), n=dy.dim().
// Consider
// V<X2<Z,R> > res=levelCrossing(y,frac,autoLevel,levelSpezifier);
// Z m=res.dim();
// Then for each j \in {1,...,m} we have f(res[j].c2())==threshold
// and signum(f'(res[j].c2())) = res[j].c1().
// Here threshold=levelSpezifier*frac for autoLevel==false
// and  threshold=yVraiMax*frac for autoLevel==true
// where yVraiMax is the real number below which the fraction
// fr of all y-components lie. The fraction fr is given by the
// argument levelSpezifier (which thus has quite a different meaning
// for the two cases autoLevel==true and autoLevel==false

Iv pairedLevelCrossing(R_Vector const& y, R frac,
   bool autoLevel=true, R levelSpezifier=0.99);
// returns the position first and the last of a matching pair
// level crossings (ascending crossing matches descending crossing and
// vice versa). Thus, if y resemples a profile through the alps
// from north to the south, we get the positions of the first
// point in the north where height reaches (say) 2000 m and
// the most southern point where we leave the niveau 2000 m.

/////////////////////////////// geometry /////////////////////////////////

void rotation(R_Vector& x, R_Vector const&  a, R_Vector const&  b);
// If neither a nor b is a zero vector, there is a uniquely defined
// minimal rotation which carries a into b. In calling rotation,
// this rotation is applied to vector x, which after the function call
// holds the value of x after rotation.
// See UM/Elmo 94-10-10/3-5 for a derivation of the algorithm.
// To be used as a tool in implementing function simplex

void simplex(R_Matrix& s, Z d);
// The statements
//
//   R_Matrix s;
//   simplex(s,3);
//
// will turn s into a 4*3 matrix such that s[1],s[2],s[3],s[4] are the
// corners of a regular tetrahedron. The center of gravity of the
// tetrahedron is the zero vector (origin) and the distance of every
// corner from the origin is 1. The position of the corners with respect
// to the axes is not the one which one would normally choose in
// drawing a regular tetrahedron. The present choice comes from a
// construction which is valid in every dimension (the analogue to
// regular tetrahedron then is called a regular simplex) and tries to
// make all corners as close to coordinate axes as possible.
// See UM/Elmo 94-10-10/1-2

R area(R_Vector const& x, R_Vector const& y);
// Assumed is x.dim() == y.dim() =: n. The points (x[i],y[i]) define a
// polygon, where the last point (x[n],y[n]) is assumed to be a neighbor
// of (x[1],y[1]). We consider the (closed, of course) polygon
// (x[1],y[1]) --> (x[2],y[2]) --> ... --> (x[n],y[n]) --> (x[1],y[1])
// and return its area. In case of crossing lines of the polygon some
// enclosed areas count positive, others negative, according to the
// rotational sense of their enclosing curve.

////////////////////////////////// miscellaneous /////////////////////////

Z2 linearToMatrix(Z k, Z m, Z n);
// Consider m*n positions that are spatially layed out like the matrix
// elements of a (m,n)-Matrix. Let k be in {1,2,...m*n}. Then the
// return value is the position (i,j) in matrix indexing notation ([i][j]
// in my C++ matrices, indexing starting with 1) of a position that by
// the normal linearization of a matrix (appending the rows (not
// columns!) starting with the first one. Notice that this is the normal
// order in reading letters from a page. Beyond the primary domain for k,
// the function is extended periodically so that no restrictions for k
// have to be obeyed. m>0 and n>0 are required.

Z matrixToLinear(Z i, Z j, Z m, Z n);
// finding the linear position (see previous function) of position (i,j)
// of a (m,n) matrix. Again, i,j are shifted into the proper range and
// m>0 and n>0 are required.

R_Vector binomialCoefficients(Z n);
// returns (n over 0), (n over 1), .... (n over n) as an R_vector
// of dimension n+1. Notice that computing these coefficiens all at
// once maskes much sense since they normally are used together and
// much synergy is available then.

///////////////////////// function tenToMat ////////////////////////////

template <class T>
V< V<T> > tenToMat(V<V<V<T> > > const& x)
//: tensor to matrix
// transforms a 3-tensor into a matrix. The type of tensor elements
// is handled as a template parameter. Useful for graphical
// representation.
{
   Word loc("CpmArrays::tenToMat(V<V<V<T>>>)");
   Z d1=x.dim();
   cpmassert(d1>0,loc);
   Z d2=x[0].dim();
   cpmassert(d2>0,loc);
   Z d3=x[0][0].dim();
   cpmassert(d3>0,loc);
   Z2 f=CpmAlgorithms::genSqrRoot(d1);
   Z f1=f[1], f2=f[2];
   Z m=d2*f1, n=d3*f2;
   V< V<C> > res(m,V<C>(n,LEAN),0);
   for (Z h=0;h<d1;++h){
      Z2 h2=linearToMatrix(h+1,f1,f2);
      Z si=h2[1]-1,sj=h2[2]-1;
         // the +1 and -1 operations take into account that
         // here indexing starts with 0
      si*=d2; sj*=d3;
      for (Z i=0;i<d2;++i){
         for (Z j=0;j<d3;++j){
            res[i+si][j+sj]=x[h][i][j];
         }
      }
   }
   return res;
}

////////// functions for handling data series ////////////////////////////
// Modelling computations turned out to generate data in many
// cases most naturally as instances of V< V<R2> > i.e.
// as lists of data series, where a data series is a list of
// R2's which typically are pairs (t,y) where t is a point in model
// time to which the computed value y belongs. Notice that
// CpmGraphics::Graph(V< V<R2> > const&, ....) is a flexible function to
// show such data on screen.

V< V<R2> > homDataRan(V< V<R2> > const& x);
//: homogenize data range
// x.dim() has to be even = 2*dh
// x[i] and x[i+dh] contain data of comparable origin.
// Typically they may represent data of various quantities computed
// by two different methods. In the R2-typed data, the first component
// is interpreted as time to which the value of the second component
// refers.
// Since the values of these quantities may vary by
// a large factor, we multiply x[i] and x[i+dh] by a common
// (i-dependent) positive factor which forces these values
// into the interval [-1,1] and let them reach one of the
// boundaries at least once.
// The data series x[i] and x[i+dh] are treated symmetrically, so that
// absolute values > 1 can't occur in the processed data.
// Corresponding data need not to fit in length or in the points
// in time to which they refer.

V< V<R2> > difOfCorData(V< V<R2> > const& xa, bool homFirst=true);
//: difference of corresponding data
// Same data structure as for the previous function assumed.
// After adjusting the data ranges via the previous function
// (remember: corresponding data are multiplied by a common
// positive factor) we form the difference between corresponding data.
// Since this makes sense only if corresponding data belong to the same
// point in time, appropriate tests are built in.
// If the last argument is set to false, the prepararoty
// call to homDataRan is avoided.

bool toGnuPlot(Word const& filename, V< V<R2> > const& x);
//: to gnu-plot
// Creates an outputfile with name 'fileName' and writes into it
// the elements of x in a manner that the file is suitable
// as a data input file for making gnuplot graphics with them.
// There is a similar function bool R_Matrix::toGnuPlot(Word const&)const
// for filing data held by a R_Matrix.

bool toGnuPlot(ostream& str, V< V<R2> > const& x);
//: to gnu-plot
// Similar to the previous function, but writes on an ostream, that may
// have been created before by opening a file.

} // namespace

#endif
