//? cpmclinalg.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmclinalg.h>
#include <cpmvr.h>

using namespace CpmRoot;
using namespace CpmLinAlg;

/////////////////////////// class C2 ///////////////////////////////////

C2::C2(const V<C>& v)
{
   Z d=v.dim();
   if (1<=d) x1=v[1];
   if (2<=d) x2=v[2];
}

bool C2::prnOn(ostream& out)const
{
   if (!CpmRoot::writeTitle("C2",out)) return false;
   if (!x1.prnOn(out)) return false;
   if (!x2.prnOn(out)) return false;
   return true;
}

bool C2::scanFrom(istream& in)
{
   if (!x1.scanFrom(in)) return false;
   if (!x2.scanFrom(in)) return false;
   return true;
}

Z C2::com(const C2& a)const
{
   if (x1<a.x1) return 1;
   if (x1>a.x1) return -1;
   if (x2<a.x2) return 1;
   if (x2>a.x2) return -1;
   return 0;
}

R C2::abs(bool careful)const
{
   CpmLinAlg::R2 r(x1.abs(careful),x2.abs(careful));
   return r.abs(careful);
}

R C2::dis(const C2& x)const
{
   R a=abs();
   R b=x.abs();
   R c=(*this-x).abs();
   return disDefFun(a,b,c);
}

X2<R,C2 > C2::polDec()const
{
   R r=abs(true);
   C2 pol=*this;
   R z(0.);
   if (r!=z){
     R rInv=cpminv(r); pol*=rInv;
     return X2<R,C2>(r,pol);
   }
   else{
      return X2<R,C2>(z,C2(C(1),C()));
   }
}

R2 C2::blcAng()const
{
   R r1,phi1,r2,phi2;
   x1.pol_1_2(r1,phi1);
   x2.pol_1_2(r2,phi2);
   R thetah=C(r1,r2).arg();
   R phi=phi2-phi1;
   if (phi<0) phi+=cpm2pi;
   return R2(thetah*2,phi);
}

/*
// instructive but inefficient implementation
R2 C2::blcAng()const
{
   CC2 m=CC2::dydPrd(*this,*this);
   R u1=(m*CC2::sig1).trc()[1];
   R u2=(m*CC2::sig2).trc()[1];
   R u3=(m*CC2::sig3).trc()[1];
   R3 u(u1,u2,u3);
   R3 res=u.carToPol();
   return R2(res[2],Pi-res[3]);
}
*/

R C2::blcTest(Z n)
{
   const R a=100;
   R2 r0(-a,a);
   R err=0;
   for (Z i=1;i<=n;++i){
      R2 r1=r0.ran(i); // random pair of arguments
      C2 z1(r1[1],r1[2]);
      z1.nor1_(); // rephase to make first component real
      R2 r2=z1.blcAng();
      C2 z2(r2[1],r2[2]);
      z2.nor1_(); // rephase to make first component real
         // now z1 and z2 should differ only by numerical noise
      err+=z1.dis(z2);
   }
   return err;
}

/**********
 R C2::codTest(Z n)
 {
    C z1(1);
    C z2(1,-1);
    C2 a(z1,z2);
    R err=0;
    for (Z i=1;i<=n;++i){
       C2 b=a.ran(i);
       b.nor_();
       R3 r3=b.toR3();
       C2 c(r3);
       R d=b.dis(c);
       err+=d;
       cpmdebug(b);
       cpmdebug(c);
       cpmdebug(d);
    }
    return err;
 }
 ************/

R C2::codTest(Z n)
 {
    R err=0;
    for (Z i=1;i<=n;++i){
       R r1=CpmRoot::randomR();
       R r2=CpmRoot::randomR();
       R r3=CpmRoot::randomR();
       C2 a(R3(r1,r2,r3));
       R3 b=a.toR3();
       R d=a.dis(b);
       err+=d;
       cpmdebug(a);
       cpmdebug(b);
       cpmdebug(d);
    }
    return err;
 }

/////////////////////////// class C3 ///////////////////////////////////

C3::C3(const V<C>& v)
{
   Z d=v.dim();
   if (1<=d) x1=v[1];
   if (2<=d) x2=v[2];
   if (3<=d) x2=v[3];
}

bool C3::prnOn(ostream& out)const
{
   if (!CpmRoot::writeTitle("C3",out)) return false;
   if (!x1.prnOn(out)) return false;
   if (!x2.prnOn(out)) return false;
   if (!x3.prnOn(out)) return false;
   return true;
}

bool C3::scanFrom(istream& in)
{
   if (!x1.scanFrom(in)) return false;
   if (!x2.scanFrom(in)) return false;
   if (!x3.scanFrom(in)) return false;
   return true;
}


Z C3::com(const C3& a)const
{
   if (x1<a.x1) return 1;
   if (x1>a.x1) return -1;
   if (x2<a.x2) return 1;
   if (x2>a.x2) return -1;
   if (x3<a.x3) return 1;
   if (x3>a.x3) return -1;
   return 0;
}

R C3::abs(bool careful)const
{
   CpmLinAlg::R3 r;
   r.x1=x1.abs(careful);
   r.x2=x2.abs(careful);
   r.x3=x3.abs(careful);
   return r.abs(careful);
}

R C3::dis(const C3& x)const
{
   R a=abs();
   R b=x.abs();
   R c=(*this-x).abs();
   return disDefFun(a,b,c);
}

X2<R,C3 > C3::polDec()const
{
   R r=abs(true);
   C3 pol=*this;
   R z(0.);
   if (r!=z){
     R rInv=cpminv(r); pol*=rInv;
     return X2<R,C3>(r,pol);
   }
   else{
     return X2<R,C3>(z,C3(C(1),C(),C()));
   }
}

/////////////////////////// class C4 ///////////////////////////////////

C4::C4(const V<C>& v)
{
   Z d=v.dim();
   if (1<=d) x1=v[1];
   if (2<=d) x2=v[2];
   if (3<=d) x2=v[3];
   if (4<=d) x2=v[4];
}

bool C4::prnOn(ostream& out)const
{
   if (!CpmRoot::writeTitle("C4",out)) return false;
   if (!x1.prnOn(out)) return false;
   if (!x2.prnOn(out)) return false;
   if (!x3.prnOn(out)) return false;
   if (!x4.prnOn(out)) return false;
   return true;
}

bool C4::scanFrom(istream& in)
{
   if (!x1.scanFrom(in)) return false;
   if (!x2.scanFrom(in)) return false;
   if (!x3.scanFrom(in)) return false;
   if (!x4.scanFrom(in)) return false;
   return true;
}

Z C4::com(const C4& a)const
{
   if (x1<a.x1) return 1;
   if (x1>a.x1) return -1;
   if (x2<a.x2) return 1;
   if (x2>a.x2) return -1;
   if (x3<a.x3) return 1;
   if (x3>a.x3) return -1;
   if (x4<a.x4) return 1;
   if (x4>a.x4) return -1;
   return 0;
}

R C4::abs(bool careful)const
{
   CpmLinAlg::R4 r;
   r.x1=x1.abs(careful);
   r.x2=x2.abs(careful);
   r.x3=x3.abs(careful);
   r.x4=x4.abs(careful);
   return r.abs(careful);
}

R C4::dis(const C4& x)const
{
   R a=abs();
   R b=x.abs();
   R c=(*this-x).abs();
   return disDefFun(a,b,c);
}

X2<R,C4> C4::polDec()const
{
   R r=abs(true);
   C4 pol=*this;
   if (r!=0){
     R rInv=R(1)/r; pol*=rInv;
     return X2<R,C4>(r,pol);
   }
   else{
      return X2<R,C4>(0.,C4(C(1),C(),C(),C()));
   }
}

//////////////////////////// class CC2 ///////////////////////////////////

C2 CpmLinAlg::operator*(const CC2& a, const C2& x)
// avoid trc for utmost efficiency
{
   return C2(a.x1.x1 * x.x1 + a.x1.x2 * x.x2,
             a.x2.x1 * x.x1 + a.x2.x2 * x.x2);
}

R CC2::abs()const{ return cpmsqrt(((x1|x1)+(x2|x2))[1]);}

R CC2::absSqr()const{ return ((x1|x1)+(x2|x2))[1];}


//CC2 CC2::operator !()const
CC2 CC2::inv()const
{
   C detInv=det().inv();
   CC2 res;
   res.x1.x1=x2.x2;
   res.x2.x2=x1.x1;
   res.x1.x2=-x1.x2;
   res.x2.x1=-x2.x1;
   return res;
}

CC2 CpmLinAlg::operator*(const CC2& a, const CC2& b)
   {
      CC2 res;
      res.x1.x1=a.x1.x1 * b.x1.x1 + a.x1.x2 * b.x2.x1;
      res.x1.x2=a.x1.x1 * b.x1.x2 + a.x1.x2 * b.x2.x2;
      res.x2.x1=a.x2.x1 * b.x1.x1 + a.x2.x2 * b.x2.x1;
      res.x2.x2=a.x2.x1 * b.x1.x2 + a.x2.x2 * b.x2.x2;
      return res;
   }

CC2 CpmLinAlg::operator+(const CC2& a, const CC2& b)
   {
      CC2 res=a;
      res.x1+=b.x1;
      res.x2+=b.x2;
      return res;
   }

CC2 CpmLinAlg::operator~(const CC2& a)
// transpose
   {
      CC2 res=a;
      res.x1.x2=a.x2.x1;
      res.x2.x1=a.x1.x2;
      return res;
   }

CC2 CC2::trn()const
// transpose
{
   return CC2(x1.x1,x2.x1,x1.x2,x2.x2);
}

CC2 CC2::her()const
// hermitean conjugate
{
   return CC2(x1.x1.con(),x2.x1.con(),x1.x2.con(),x2.x2.con());
}

CC2 CC2::con()const
{
   return her();
}

CC2 CC2::per(Perm const& p)const
{
   cpmassert(p.dim()==4,"CC2::per(Perm)");
   V<C> c4 = toVC();
   c4 = c4.compose(p.getRep());
   return CC2(c4);
}

CC2 CC2::zero=CC2();
CC2 CC2::unit=CC2(C(1.),C(),C(),C(1.));
CC2 CC2::sig0=CC2(C(1.),C(),C(),C(1.));
CC2 CC2::sig1=CC2(C(),C(1.),C(1.),C());
CC2 CC2::sig2=CC2(C(),C(0.,-1.),C(0.,1.),C());
CC2 CC2::sig3=CC2(C(1.),C(),C(),C(-1.));

bool CC2::prnOn(ostream& out)const
{
     if (!x1.prnOn(out)) return false;
     if (!x2.prnOn(out)) return false;
     return true;
}

bool CC2::scanFrom(istream& in)
{
   if (!x1.scanFrom(in)) return false;
   if (!x2.scanFrom(in)) return false;
   return true;
}

CC2 CC2::ran(Z i)const
{
   CpmArrays::Vr<C> temp(toVC());
   temp=temp.ran(i);
   return CC2(temp);
}

CC2 CC2::test(Z cpl)const
{
   const C red(0.9,-0.95);
   C z;
   C z1=CpmRoot::Root<C>(z).test(cpl);
   C z2=z1*red;
   C z3=z2*red;
   C z4=z3*red;
   return CC2(C4(z1,z2,z3,z4));
}

CC2 CC2::net(Z i)const
{
   return i==1 ? CC2::unit : CC2::zero;
}

Z CC2::hash()const
{
   C4 c4=toC4();
   Z i1=CpmRoot::hash(c4[1]);
   Z i2=CpmRoot::hash(c4[2]);
   Z i3=CpmRoot::hash(c4[3]);
   Z i4=CpmRoot::hash(c4[4]);
   return i1^i2^i3^i4;
}

R CC2::dis(CC2 const& a)const
{
   R r1=abs();
   R r2=a.abs();
   R r3=(*this-a).abs();
   return CpmRoot::disDefFun(r1,r2,r3);
}

//////////////////////////// class CC3 ///////////////////////////////////

C3 CpmLinAlg::operator*(const CC3& a, const C3& x)
   {
      C3 res;
      res.x1=a.x1.trc(x);
      res.x2=a.x2.trc(x);
      res.x3=a.x3.trc(x);
      return res;
   }

CC3 CpmLinAlg::operator*(const CC3& a, const CC3& b)
   {
      CC3 res;
      res.x1.x1=a.x1.x1*b.x1.x1 + a.x1.x2*b.x2.x1 + a.x1.x3*b.x3.x1;
      res.x1.x2=a.x1.x1*b.x1.x2 + a.x1.x2*b.x2.x2 + a.x1.x3*b.x3.x2;
      res.x1.x3=a.x1.x1*b.x1.x3 + a.x1.x2*b.x2.x3 + a.x1.x3*b.x3.x3;

      res.x2.x1=a.x2.x1*b.x1.x1 + a.x2.x2*b.x2.x1 + a.x2.x3*b.x3.x1;
      res.x2.x2=a.x2.x1*b.x1.x2 + a.x2.x2*b.x2.x2 + a.x2.x3*b.x3.x2;
      res.x2.x3=a.x2.x1*b.x1.x3 + a.x2.x2*b.x2.x3 + a.x2.x3*b.x3.x3;

      res.x3.x1=a.x3.x1*b.x1.x1 + a.x3.x2*b.x2.x1 + a.x3.x3*b.x3.x1;
      res.x3.x2=a.x3.x1*b.x1.x2 + a.x3.x2*b.x2.x2 + a.x3.x3*b.x3.x2;
      res.x3.x3=a.x3.x1*b.x1.x3 + a.x3.x2*b.x2.x3 + a.x3.x3*b.x3.x3;

      return res;
   }

CC3 CpmLinAlg::operator+(const CC3& a, const CC3& b)
   {
      CC3 res=a;
      res.x1+=b.x1;
      res.x2+=b.x2;
      res.x3+=b.x3;
      return res;
   }

CC3 CpmLinAlg::operator~(const CC3& a)
   {
      CC3 res(a);

      res.x1.x2=a.x2.x1;
      res.x1.x3=a.x3.x1;

      res.x2.x3=a.x3.x2;
      res.x2.x1=a.x1.x2;

      res.x3.x1=a.x1.x3;
      res.x3.x2=a.x2.x3;

      return res;
   }

CC3 CC3::zero=CC3();
CC3 CC3::unit=CC3(C3(C(1.),C(),C()),C3(C(),C(1.),C()),C3(C(),C(),C(1.)));


bool CC3::prnOn(ostream& out)const
{
     if (!x1.prnOn(out)) return false;
     if (!x2.prnOn(out)) return false;
     if (!x3.prnOn(out)) return false;
     return true;
}

bool CC3::scanFrom(istream& in)
{
   if (!x1.scanFrom(in)) return false;
   if (!x2.scanFrom(in)) return false;
   if (!x3.scanFrom(in)) return false;
   return true;
}

//////////////////////////// class CC4 ///////////////////////////////////

C4 CpmLinAlg::operator*(const CC4& a, const C4& x)
   {
      C4 res;
      res.x1=a.x1.trc(x);
      res.x2=a.x2.trc(x);
      res.x3=a.x3.trc(x);
      res.x4=a.x4.trc(x);
      return res;
   }

CC4 CpmLinAlg::operator*(const CC4& a, const CC4& b)
   {
      CC4 res;
      res.x1.x1=a.x1.x1*b.x1.x1 + a.x1.x2*b.x2.x1 + a.x1.x3*b.x3.x1 +\
       a.x1.x4*b.x4.x1;
      res.x1.x2=a.x1.x1*b.x1.x2 + a.x1.x2*b.x2.x2 + a.x1.x3*b.x3.x2 +\
       a.x1.x4*b.x4.x2;
      res.x1.x3=a.x1.x1*b.x1.x3 + a.x1.x2*b.x2.x3 + a.x1.x3*b.x3.x3 +\
       a.x1.x4*b.x4.x3;
      res.x1.x4=a.x1.x1*b.x1.x4 + a.x1.x2*b.x2.x4 + a.x1.x3*b.x3.x4 +\
       a.x1.x4*b.x4.x4;

      res.x2.x1=a.x2.x1*b.x1.x1 + a.x2.x2*b.x2.x1 + a.x2.x3*b.x3.x1 +\
       a.x2.x4*b.x4.x1;
      res.x2.x2=a.x2.x1*b.x1.x2 + a.x2.x2*b.x2.x2 + a.x2.x3*b.x3.x2 +\
       a.x2.x4*b.x4.x2;
      res.x2.x3=a.x2.x1*b.x1.x3 + a.x2.x2*b.x2.x3 + a.x2.x3*b.x3.x3 +\
       a.x2.x4*b.x4.x3;
      res.x2.x4=a.x2.x1*b.x1.x4 + a.x2.x2*b.x2.x4 + a.x2.x3*b.x3.x4 +\
       a.x2.x4*b.x4.x4;

      res.x3.x1=a.x3.x1*b.x1.x1 + a.x3.x2*b.x2.x1 + a.x3.x3*b.x3.x1 +\
       a.x3.x4*b.x4.x1;
      res.x3.x2=a.x3.x1*b.x1.x2 + a.x3.x2*b.x2.x2 + a.x3.x3*b.x3.x2 +\
       a.x3.x4*b.x4.x2;
      res.x3.x3=a.x3.x1*b.x1.x3 + a.x3.x2*b.x2.x3 + a.x3.x3*b.x3.x3 +\
       a.x3.x4*b.x4.x3;
      res.x3.x4=a.x3.x1*b.x1.x4 + a.x3.x2*b.x2.x4 + a.x3.x3*b.x3.x4 +\
       a.x3.x4*b.x4.x4;

      res.x4.x1=a.x4.x1*b.x1.x1 + a.x4.x2*b.x2.x1 + a.x4.x3*b.x3.x1 +\
       a.x4.x4*b.x4.x1;
      res.x4.x2=a.x4.x1*b.x1.x2 + a.x4.x2*b.x2.x2 + a.x4.x3*b.x3.x2 +\
       a.x4.x4*b.x4.x2;
      res.x4.x3=a.x4.x1*b.x1.x3 + a.x4.x2*b.x2.x3 + a.x4.x3*b.x3.x3 +\
       a.x4.x4*b.x4.x3;
      res.x4.x4=a.x4.x1*b.x1.x4 + a.x4.x2*b.x2.x4 + a.x4.x3*b.x3.x4 +\
       a.x4.x4*b.x4.x4;
      return res;
   }

CC4 CpmLinAlg::operator+(const CC4& a, const CC4& b)
   {
      CC4 res=a;
      res.x1+=b.x1;
      res.x2+=b.x2;
      res.x3+=b.x3;
      res.x4+=b.x4;
      return res;
   }

CC4 CpmLinAlg::operator~(const CC4& a)
   {
      CC4 res(a);

      res.x1.x2=a.x2.x1;
      res.x1.x3=a.x3.x1;
      res.x1.x4=a.x4.x1;

      res.x2.x1=a.x1.x2;
      res.x2.x3=a.x3.x2;
      res.x2.x4=a.x4.x2;

      res.x3.x1=a.x1.x3;
      res.x3.x2=a.x2.x3;
      res.x3.x4=a.x4.x3;

      res.x4.x1=a.x1.x4;
      res.x4.x2=a.x2.x4;
      res.x4.x3=a.x3.x4;

      return res;
   }

// When using multiple precision (relying on pmreal) it is impossible to
// define static quantities building on other static quantities (probably
// depending on compiler options). The presently implemented workarround
// makes CC4::gam0 ... CC4::alpha3W well defined if before there was a call
// to some CC4 constructor oder to function CC4::iniGam().

/**********************************
CC4 CC4::zero=Gamma("zero").rep();
CC4 CC4::unit=Gamma("unit").rep();

CC4 CC4::gam0=Gamma(0).rep();
CC4 CC4::gam1=Gamma(1).rep();
CC4 CC4::gam2=Gamma(2).rep();
CC4 CC4::gam3=Gamma(3).rep();

CC4 CC4::beta=Gamma::beta();
CC4 CC4::alpha1=Gamma::alpha(1);
CC4 CC4::alpha2=Gamma::alpha(2);
CC4 CC4::alpha3=Gamma::alpha(3);

CC4 CC4::gam0W=Gamma(0,"Weyl").rep();
CC4 CC4::gam1W=Gamma(1,"Weyl").rep();
CC4 CC4::gam2W=Gamma(2,"Weyl").rep();
CC4 CC4::gam3W=Gamma(3,"Weyl").rep();

CC4 CC4::betaW=Gamma::beta("Weyl");
CC4 CC4::alpha1W=Gamma::alpha(1,"Weyl");
CC4 CC4::alpha2W=Gamma::alpha(2,"Weyl");
CC4 CC4::alpha3W=Gamma::alpha(3,"Weyl");

CC4 CC4::gam0M=Gamma(0,"Majorana").rep();
CC4 CC4::gam1M=Gamma(1,"Majorana").rep();
CC4 CC4::gam2M=Gamma(2,"Majorana").rep();
CC4 CC4::gam3M=Gamma(3,"Majorana").rep();

CC4 CC4::betaM=Gamma::beta("Majorana");
CC4 CC4::alpha1M=Gamma::alpha(1,"Majorana");
CC4 CC4::alpha2M=Gamma::alpha(2,"Majorana");
CC4 CC4::alpha3M=Gamma::alpha(3,"Majorana");
****************************************************/


bool CC4::prnOn(ostream& out)const
{
     if (!x1.prnOn(out)) return false;
     if (!x2.prnOn(out)) return false;
     if (!x3.prnOn(out)) return false;
     if (!x4.prnOn(out)) return false;
     return true;
}

bool CC4::scanFrom(istream& in)
{
   if (!x1.scanFrom(in)) return false;
   if (!x2.scanFrom(in)) return false;
   if (!x3.scanFrom(in)) return false;
   if (!x4.scanFrom(in)) return false;
   return true;
}

//////////////////////////// class CCCC2 /////////////////////////////////

CC2 CCCC2::operator * (CC2 const& a)const
{
   C r11=x1.x1.x1.x1*a.x1.x1 + x1.x1.x1.x2*a.x1.x2
      + x1.x2.x1.x1*a.x2.x1 + x1.x2.x1.x2*a.x2.x2;
   C r12=x1.x1.x2.x1*a.x1.x1 + x1.x1.x2.x2*a.x1.x2
      + x1.x2.x2.x1*a.x2.x1 + x1.x2.x2.x2*a.x2.x2;
   C r21=x2.x1.x1.x1*a.x1.x1 + x2.x1.x1.x2*a.x1.x2
      + x2.x2.x1.x1*a.x2.x1 + x2.x2.x1.x2*a.x2.x2;
   C r22=x2.x1.x2.x1*a.x1.x1 + x2.x1.x2.x2*a.x1.x2
      + x2.x2.x2.x1*a.x2.x1 + x2.x2.x2.x2*a.x2.x2;
   return CC2(r11,r12,r21,r22);
}

C CCCC2::app(CC2 const& a, bool b1, bool b2)const
{
   if (!b1 && !b2)
      return x1.x1.x1.x1*a.x1.x1 + x1.x1.x1.x2*a.x1.x2
      + x1.x2.x1.x1*a.x2.x1 + x1.x2.x1.x2*a.x2.x2;
   if (!b1 && b2)
      return x1.x1.x2.x1*a.x1.x1 + x1.x1.x2.x2*a.x1.x2
      + x1.x2.x2.x1*a.x2.x1 + x1.x2.x2.x2*a.x2.x2;
   if ( b1 && !b2)
      return x2.x1.x1.x1*a.x1.x1 + x2.x1.x1.x2*a.x1.x2
      + x2.x2.x1.x1*a.x2.x1 + x2.x2.x1.x2*a.x2.x2;
   if (b1 && b2)
      return x2.x1.x2.x1*a.x1.x1 + x2.x1.x2.x2*a.x1.x2
      + x2.x2.x2.x1*a.x2.x1 + x2.x2.x2.x2*a.x2.x2;
   return C(); // never active
}
