//? cpmcmatrix.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmcmatrix.h>
#include <cpmrmateig.h>
#include <cpmc.h>
#include <cpmrfunc.h>
#include <cpmfile.h>
//#include <cpmsystemdependencies.h>

using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmSystem;

using CpmArrays::C_Matrix;
using CpmArrays::R_Matrix;
using CpmArrays::C_Vector;
using CpmArrays::R_Vector;
using CpmArrays::SVDC;
using CpmArrays::SpcObj;
using CpmArrays::V;
using CpmArrays::VV;
using CpmArrays::X2;
using CpmFunctions::F;
using CpmFunctions::F2;
using CpmFunctions::F3;
using CpmGeo::Iv;
using CpmLinAlg::R2;

R C_Matrix::etaInv=1e-12;

#if defined(CPM_USE_EIGEN)
using namespace Eigen;
using namespace std;
C_Matrix::C_Matrix(
      Matrix<complex<R>,Dynamic,Dynamic> const& mat)
   {
      Z m=mat.rows();
      Z n=mat.cols();
      VV<C> res(m,n);
      for (Z i=1;i<=m;++i){
         for (Z j=1;j<=n;++j){
            res[i][j] = C(mat(i,j));
         }
      }
      *this=C_Matrix(res);
   }

   Matrix<complex<R>,Dynamic,Dynamic> C_Matrix::toEigen()const
   {
      Z m=dim1(),n=dim2(),i,j;
      Matrix<complex<R>,Dynamic,Dynamic> res(m,n);
      for (i=1;i<=m;++i) for (j=1;j<=n;++j)
         res(i,j)=rep.cui(i).cui(j).std();
      return res;
   }

#endif

C_Matrix::C_Matrix(const VV<C>& vv):rep(vv.dim())
{
   Z m=rep.dim();
   for (Z i=1;i<=m;i++) rep[i]=C_Vector(vv[i]);
}

C_Matrix::C_Matrix(const V< V<C> >& vv):rep(vv.dim())
{
   Z m=rep.dim();
   for (Z i=1;i<=m;i++) rep[i]=C_Vector(vv[i]);
}

C_Matrix::C_Matrix(C_Vector const& a, C_Vector const& b):
rep(a.dim(),C_Vector(b.dim()))
{
   for (Z i=a.b();i<=a.e();++i){
      C ai = a[i];
      for (Z j=b.b();j<=b.e();++j){
         rep[i][j]=ai*~b[j];
      }
   }
}

C_Matrix::C_Matrix(C_Vector const& col, Z n):
rep(col.dim(),C_Vector(n))
{
   for (Z i=1;i<=col.dim();++i){
      for (Z j=1;j<=n;++j){
         rep[i][j]=col[i];
      }
   }
}

C_Matrix C_Matrix::permuteColumns(const V<Z>& per)
{
   Z n=dim();
   C_Matrix res(n);
   for (Z i=1;i<=n;i++) res[i]=(*this)[i].permute(per);
   return res;
}

C_Matrix C_Matrix::trn(void)const
{
   Z m=dim1();
   Z n=dim2();
   C_Matrix res(n,m); // sic, (m,n) are the data for *this
   for (Z i=1;i<=m;i++) for (Z j=1;j<=n;j++) res[j][i]=rep[i][j];
   return res;
}

C_Matrix C_Matrix::con(void)const
{
   Z m=dim1();
   Z n=dim2();
   C_Matrix res(n,m); // sic, (m,n) are the data for *this
   for (Z i=1;i<=m;i++) for (Z j=1;j<=n;j++) res[j][i]=rep[i][j].con();
   return res;
}

C_Matrix C_Matrix::cc(void)const
{
   Z m=dim1();
   Z n=dim2();
   C_Matrix res(m,n);
   for (Z i=1;i<=m;i++) for (Z j=1;j<=n;j++) res[i][j]=rep[i][j].con();
   return res;
}

R_Matrix C_Matrix::real()const
{
   Z m=dim1();
   Z n=dim2();
   R_Matrix res(n,m); // sic, (m,n) are the data for *this
   for (Z i=1;i<=m;i++) for (Z j=1;j<=n;j++) res[i][j]=rep[i][j].real();
   return res;
}

R_Matrix C_Matrix::toRealForm(bool test)const
{
   Z i,j,i2,j2,d=dim();
   Z d2=d*2;
   R_Matrix res(d2,d2);
   C val;
   R valr,vali;
   for (i=1;i<=d;i++){
      i2=i+d;
      for (j=1;j<=d;j++){
         j2=j+d;
         val=rep.cui(i).cui(j);
         valr=val[1];
         vali=val[2];
         res[i][j]=valr;
         res[i2][j2]=valr;
         res[i2][j]=vali;
         res[i][j2]=-vali;
      }
   }
   if (test){
      C_Matrix rc = C_Matrix::fromRealForm(res);
      R err = dis(rc);
      cout<<"C_Matrix::toRealForm(bool): err = "<<err<<endl;
      cpmassert(err<1e-6,"C_Matrix::toRealForm(bool)");
   }
   return res;
}

C_Matrix C_Matrix::fromRealForm(R_Matrix const& r)
// static function
{
   Z i,j,i2,d2=r.dim1();
   cpmassert(r.dim2()==d2,"");
   Z d=d2/2;
   cpmassert(2*d==d2,"");
   C_Matrix res(d,d);
   C val;
   R valr,vali;
   for (i=1;i<=d;i++){
      i2=i+d;
      for (j=1;j<=d;j++){
         valr=r[i][j];
         vali=r[i2][j];
         val=C(valr,vali);
         res[i][j]=val;
      }
   }
   return res;
}

C_Vector C_Matrix::operator *(const C_Vector& x)const
// always defined
{
   Z m=dim1();
   Z n=dim2();
   Z nx=x.dim();
   C_Vector res(m);
   if (n==nx){
      for (Z i=1;i<=m;i++){
         res[i]=rep[i].sumPrd(x); // fast normal case
      }
   }
   else{ // slow irregular case
      Z nMin=(nx<=n ? nx : n);
      for (Z i=1;i<=m;i++){
         C sum;
         for (Z j=1;j<=nMin;j++) sum+=rep[i][j]*x[j];
         res[i]=sum;
      }
   }
   return res;
}

# if defined(CPM_USE_EIGEN)

C_Matrix C_Matrix::operator *(const C_Matrix& a2)const
// Notice: the product of arbitrarily framed matrices is well defined.
// If the ranges fit the the requirement for regular matrix multiplication
// the speed is 18 times larger.
{
   using namespace Eigen;
   using std::complex;

   Z m1=dim1();
   Z n1=dim2();
   Z m2=a2.dim1();
   Z n2=a2.dim2();
   cpmassert(n1==m2,"C_Matrix * C_Matrix");
   Z i,j,k,l;
   C_Matrix res(m1,n2);
   Matrix<complex<R>,Dynamic,Dynamic> x1(m1,n1);
   for (i=1,k=0; i<=m1; i++,k++) for (j=1,l=0;j<=n1; j++,l++)
      x1(k,l) = rep[i][j].std();

   Matrix<complex<R>,Dynamic,Dynamic> x2(m2,n2);
   for (i=1,k=0; i<=m2; i++,k++) for (j=1,l=0;j<=n2; j++,l++)
      x2(k,l)=a2[i][j].std();

   Matrix<complex<R>,Dynamic,Dynamic> x3 = x1*x2;
   for (i=1,k=0; i<=m1; i++,k++) for (j=1,l=0;j<=n2; j++,l++)
   res[i][j] = x3(k,l);
   return res;
}

# else

C_Matrix C_Matrix::operator *(const C_Matrix& a2)const
// Notice: the product of arbitrarily framed matrices is well defined.
// If the ranges fit the the requirement for regular matrix multiplication
// the speed is 18 times larger.
{
   Z m1=dim1();
   Z n1=dim2();
   Z m2=a2.dim1();
   Z n2=a2.dim2();
   C_Matrix res(m1,n2);
   if (n1==m2){ // regular case
      C_Matrix b=a2.trn(); // O(n^2) negligible compared to the  O(n^3)
      // main
         // task
      for (Z i=1;i<=m1;i++){ // O(n^3)
         for (Z j=1;j<=n2;j++){
            res[i][j]=rep[i].sumPrd(b.rep[j]);
               // making the inner loop fast
         }
      }
   }
   else{ // irregular case
      Z nk=(n1<=m2 ? n1 : m2);
      for (Z i=1;i<=m1;i++){
         for (Z j=1;j<=n2;j++){
            C sum;
            for (Z k=1;k<=nk;k++){
               sum+=rep[i][k]*a2[k][j]; // slow inner loop
            }
            res[i][j]=sum;
         }
      }
   }
   return res;
}
# endif

R C_Matrix::supAbs()const
{
   R res=0;
   for (Z i=1;i<=dim();++i){
      R xi=rep[i].supAbs();
      if (xi>res) res=xi;
   }
   return res;
}

C_Matrix C_Matrix::append2(const C_Matrix& x1)const
{
   Z m=CpmRootX::inf<Z>(dim1(),x1.dim1());
   Z n1=dim2();
   Z n2=x1.dim2();
   Z i,j,k,n=n1+n2;
   C_Matrix res(m,n);
   for (i=1;i<=m;++i){
      for (j=1;j<=n1;++j){
         res[i][j]=rep[i][j];
      }
      for (j=1,k=n1+1; j<=n2; ++j,++k){
         res[i][k]=x1[i][j];
      }
   }
   return res;
}

C_Vector C_Matrix::getCol(Z j)const
{
   Word loc("C_Matrix::getCol(Z)");
   Z m=dim1(),n=dim2();
   cpmassert(j>=1 && j<=n,loc);
   C_Vector res(m);
   for (Z i=1;i<=m;i++) res[i]=rep[i][j];
   return res;
}

void C_Matrix::setCol(Z j, const C_Vector& v)
{
   Word loc("C_Matrix::setCol(Z,C_Vector)");
   Z m=dim1(),n=dim2();
   cpmassert(j>=1 && j<=n,loc);
   Z mm=CpmRootX::inf<Z>(m,v.dim());
   for (Z i=1;i<=mm;i++) rep[i][j]=v[i];
}

C_Matrix& C_Matrix::operator +=(const C_Matrix& a)
{
   Z n=dim();
   Z na=a.dim();
   Z nnaMin=(n<=na ? n : na);
   for (Z i=1;i<=nnaMin;i++) rep[i]+=(a.rep)[i];
   return *this;
}

C_Matrix& C_Matrix::operator -=(const C_Matrix& a)
{
   Z n=dim();
   Z na=a.dim();
   Z nnaMin=(n<=na ? n : na);
   for (Z i=1;i<=nnaMin;i++) rep[i]-=(a.rep)[i];
   return *this;
}

C_Matrix C_Matrix::operator -(const C_Matrix& a)const
 {  C_Matrix res=*this; return res-=a;}

R2 C_Matrix::centerOfMass()const
{
   Z mL=3;
   static Word loc("C_Matrix::centerOfMass()");
   cpmmessage(mL,loc&" started");
   Z m=dim();
   Z n=dim2();
   R si=0;
   R sj=0;
   R s=0;
   Z i,j;
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         R val=((*this)[i][j]).abs2();
         si+=R(i)*val;
         sj+=R(j)*val;
         s+=val;
      }
   }
   cpmassert(s!=0,loc);
   R sInv=cpminv(s);
   cpmmessage(mL,loc&" done");
   return R2(si*sInv,sj*sInv);
}

C_Vector C_Matrix::linForm()const
// In principle rep[1]&rep[2]&...&rep[n], n=rep.dim()
// But due to the multiple allocations this is much too slow.
{
   Z m=dim1();
   if (m<1) return C_Vector(0);
   Z n=dim2();
   if (n<1) return C_Vector(0);
   Z mn=m*n;
   C_Vector res(mn);
   Z k=0;
   for (Z i=1;i<=m;++i){
      for (Z j=1;j<=n;++j){
         res[++k]=rep[i][j];
      }
   }
   return res;
}

C_Vector C_Matrix::dig()const
{
   Z dd=CpmRootX::inf<Z>(dim1(),dim2());
   C_Vector res(dd);
   for (Z i=1;i<=dd;++i) res[i]=rep[i][i];
   return res;
}

namespace{
   C_Vector ffAux(C_Vector const& x, F<C,C> const& f){ return x(f);}
   //C_Vector logAux(C_Vector const& x){ return x.log10();}
}

C_Matrix C_Matrix::operator()(F<C,C> const& g)const
{
   return C_Matrix(rep(F1<C_Vector,F<C,C>,C_Vector>(g)(ffAux)));
}

VV<C> C_Matrix::toVV()const
{
   Z m=dim1();
   Z n=dim2();
   Z i,j;
   VV<C> res(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         res[i][j]=rep[i][j];
      }
   }
   return res;
}

C toFfunc(R2 const& xy, C_Matrix const& mat, Iv const& ivx, Iv const& ivy)
{
   Z m=mat.dim1(),n=mat.dim2();
   Z i=ivx.qun(xy.x1,m,true);
   Z j=ivy.qun(xy.x2,n,true);
   return mat[i][j];
}

F<R2,C> C_Matrix::toF(Iv const& iR,Iv const& jR)const
{
   return F3<R2,C_Matrix,Iv,Iv,C>(*this,iR,jR)(toFfunc);
}

bool C_Matrix::wrtFor(ostream& str)const
{
   const Word sep(" "); // as a Word, sep ends with a \n
   Z m=dim1();
   Z n=dim2();
   Z i,j;
   str<<m<<sep<<n<<endl;
   for (i=1;i<=m;++i){
      for (j=1;j<=n;++j){
         str<<rep[i][j]<<sep;
      }
      str<<endl;
   }
   return (str) ? true : false;
}

bool C_Matrix::readFor_(istream& str)
{
   Z m=-1,n=-1;
   str>>m;
   if (m<0 || !str) return false;
   str>>n;
   if (n<0 || !str) return false;
   rep=V<C_Vector>(m,C_Vector(n));
   Z i,j;
   for (i=1;i<=m;++i){
      for (j=1;j<=n;++j){
         str>>rep[i][j];
      }
   }
   return (str) ? true : false;
}

bool C_Matrix::wrtForTrn(ostream& str)const
{
   const Word sep(" ");
   Z m=dim1();
   Z n=dim2();
   Z i,j;
   str<<m<<sep<<n<<endl;
   for (i=1;i<=n;++i){
      for (j=1;j<=m;++j){
         str<<rep[j][i]<<sep;
      }
      str<<endl;
   }
   return (str) ? true : false;
}

bool C_Matrix::toGnuPlot(Word const& fileName)const
{
   Z m=dim1(), n=dim2();
   CpmSystem::OFile ofi(fileName);
   for (Z i=1;i<=n;++i){
      for (Z j=1;j<=m;++j){
         ofi()<<rep[j][i]<<" ";
      }
      ofi()<<endl;
   }
   bool res=ofi.isVal();
   return res;
}

bool C_Matrix::toGnuPlot(ostream& str)const
{
   Z m=dim1(), n=dim2();
   for (Z i=1;i<=n;++i){
      for (Z j=1;j<=m;++j){
         str<<rep[j][i]<<" ";
      }
      str<<endl;
   }
   return (str) ? true : false;
}

bool C_Matrix::readForTrn_(istream& str)
{
   Z m=-1,n=-1;
   str>>m;
   if (m<0 || !str) return false;
   str>>n;
   if (n<0 || !str) return false;
   rep=V<C_Vector>(m,C_Vector(n));
   Z i,j;
   for (i=1;i<=n;++i){
      for (j=1;j<=m;++j){
         str>>rep[j][i];
      }
   }
   return (str) ? true : false;
}

// implementation 2016 using the Eigen library. Here the functionality is
// also available for complex matrices and probably also for user-defined
// scaler types
# if defined(CPM_USE_EIGEN)
void C_Matrix::svdcmp(C_Matrix& a, R_Vector& w, C_Matrix& v)
{
   using namespace Eigen;
   using std::complex;
   Z mL=3;
   Word loc("svdcmp(C_Matrix,C_Vector,C_Matrix)");
   CPM_MA

   Z m=a.dim1(), n=a.dim2();

   if (!(m>=n))
      cpmerror("svdcmp(): bad frame of a");
   if (n!=w.dim())
      cpmerror("svdcmp(): bad frame of w");
   if ((n!=v.dim1()) || (n!=v.dim2()))
      cpmerror("svdcmp(): bad frame of v");
   Z i,j,k,l;
   Z d=CpmRootX::inf(m,n);
   Matrix<complex<R>,Dynamic,Dynamic> ea(m,n);
   for (i=1,k=0; i<=m; i++,k++) for (j=1,l=0;j<=n; j++,l++)
      ea(k,l)=a[i][j].std();

   JacobiSVD<Matrix<complex<R>,Dynamic,Dynamic> > svd(ea,
      ComputeFullU | ComputeFullV);
   Matrix<R,Dynamic,1> ew = svd.singularValues();
   Matrix<complex<R>,Dynamic,Dynamic> eu=svd.matrixU();
   Matrix<complex<R>,Dynamic,Dynamic> ev=svd.matrixV();
   cpmassert(d==ew.size(),loc);

   for (i=1,k=0; i<=d; i++,k++)
   w[i] = ew(k);

   Z u1 = eu.rows();
   Z u2 = eu.cols();
   for (i=1,k=0; i<=u1; i++,k++) for (j=1,l=0;j<=u2; j++,l++)
   a[i][j] = eu(k,l);

   Z v1 = ev.rows();
   Z v2 = ev.cols();
   for (i=1,k=0; i<=v1; i++,k++) for (j=1,l=0;j<=v2; j++,l++)
   v[i][j] = ev(k,l);

   CPM_MZ
   return;
}
#else
// singular value decomp for complex matrices not given in Press et al ////
void C_Matrix::svdcmp(C_Matrix& a, R_Vector& w, C_Matrix& v)
{
   Z mL=3;
   Word loc("svdcmp(C_Matrix,C_Vector,C_Matrix)");
   CPM_MA
   cout<<endl<<"svdcmp(C_Matrix,C_Vector,C_Matrix): not implemented"<<endl;
   cpmwarning("svdcmp(C_Matrix, C_Vector, C_Matrix): not implemented ");
   CPM_MZ
   return;
}
#endif
//////////// end of singular value decomposition algorithm /////////////

C_Matrix C_Matrix::inv()const
{
   const Z testLimit=5;
   C sum;
   Z i, j, k, m_=dim1(), n=dim2();
   bool rightframe=(m_>=n);
   C_Matrix a=*this;
   if (!rightframe) a=~a;
   m_=a.dim1(); n=a.dim2();
   C_Matrix v(n,n);
   R_Vector w(n);
   C_Matrix vr(n,m_); //  sic
   svdcmp(a,w,v);
   if (cpmverbose>=testLimit){ // test of svdcmp,
      // svdcmp got corrupted once, so good to test
      C_Matrix u=v.net(1);
      R error1=(~a*a-u).abs();
      cpmmessage("C_Matrix::inv(): error1= "&cpm(error1));
      R error2=(~v*v-u).abs();
      cpmmessage("C_Matrix::inv(): error2= "&cpm(error2));
      R error3=(v*~v-u).abs();
      cpmmessage("C_Matrix::inv(): error3= "&cpm(error3));
      R error4=(*this-a*C_Matrix(w)*~v).abs();
      cpmmessage("C_Matrix::inv(): error4= "&cpm(error4));
      R err=error1+error2+error3+error4;
      cpmmessage(" total inaccuracy in svdcmp = "&cpmwrite(err));
   }
   w=w.pseudoInverse(etaInv);  // w does not come out orderd
      // now w is the pseudo-inverse of the original w
   for (i=1;i<=n;i++){
      for (j=1;j<=m_;j++){
         sum=C();
         for (k=1;k<=n;k++){
            sum+=v[i][k]*w[k]*a[j][k];
         }
         vr[i][j]=sum;
      }
   }
   if (!rightframe) vr=~vr;
   return vr;
}

namespace{
// without the order-related operations the quality of the SVD is much
// considerably less accurate.
R_Vector svd(C_Matrix& u, C_Matrix& v)
{
   C_Matrix u_=u;
   Z m=u_.dim1(), n=u_.dim2(), i;
   if (m>=n){
      C_Matrix vv(n,n);
      v=vv;
      R_Vector w(n);
      C_Matrix::svdcmp(u_,w,v);
      R_Vector w_(n);
      for (i=1;i<=n;i++) w_[i]=-cpmabs(w[i]);
      V<Z> p=w_.permutationForIncreasingOrder();
      u_=u_.permuteColumns(p);
      v=v.permuteColumns(p);
      w=w.permute(p);
      u=u_;
      { // making the dominant component of ~v positive
         C_Matrix vt=~v;
         C_Matrix ut=~u;
         Z nv=vt.dim1(),nu=ut.dim1();
         if (nv!=nu) cpmerror("svd: malfunction in +/- part");
         for (Z i=1;i<=nv;i++){
            Iv ivt=vt[i].infSup();
            R xmin=ivt.inf(), xmax=ivt.sup();
            if (cpmabs(xmin)>cpmabs(xmax)){
               vt[i]=vt[i]*C(-1.);  // makes dominant part positiv
               ut[i]=ut[i]*C(-1.); // necessary to stay with a svd
            }
         }
         u=~ut; v=~vt;
      }
      return w;
   }
   else{
      n=u_.dim1();   // now m>=n  and u is a (n,m)-matrix
      C_Matrix vv(n,n);
      v=vv;
      R_Vector w(n);
      u_=~u_; //  now u_ is a (m,n)-matrix
      C_Matrix::svdcmp(u_,w,v);
      R_Vector w_(n);
      for (i=1;i<=n;i++) w_[i]=-cpmabs(w[i]);
      V<Z> p=w_.permutationForIncreasingOrder();
      u_=u_.permuteColumns(p);
      v=v.permuteColumns(p);
      w=w.permute(p);
      C_Matrix umem=u_; u_=v; v=umem;
      u=u_;
      { // making the dominant component of ~v positive
         C_Matrix vt=~v;
         C_Matrix ut=~u;
         Z nv=vt.dim1(), nu=ut.dim1();
         if (nv!=nu) cpmerror("svd: malfunction in +/- part");
         for (Z i=1;i<=nv;i++){
            Iv ivt=vt[i].infSup();
            R xmin=ivt.inf(),xmax=ivt.sup();
            if (cpmabs(xmin)>cpmabs(xmax)){
               vt[i]=vt[i]*C(-1.);  // makes dominant part positiv
               ut[i]=ut[i]*C(-1.); // necessary to stay with a svd
            }
         }
         u=~ut; v=~vt;
      }
      return w;
   }
}
}

C_Matrix C_Matrix::net(Z i)const
{
   Z m=dim1();
   Z n=dim2();
   C_Matrix res(m,n); // all zero by definition
   if (i!=1) return res;
// now i==1, thus unit matrix required
   if (m!=n) cpmwarning(
      "C_Matrix::net(Z): non-square unit matrix formed");
   Z mnMin=(m<=n ? m : n);
   for (Z j=1; j<=mnMin; j++) res[j][j]=C(1.);
   return res;
}

//////////////////////////// class SVDC ///////////////////////////////

SVDC::SVDC(const C_Matrix& a):m(a.dim1()),n(a.dim2()),
A(a),U(a),V()
{
   d=CpmRootX::inf(m,n);
   w=svd(U,V);
   Vt=~V;
}
//#endif

R SVDC::getError()const
{
   C_Matrix u1=~U*U;
   R error1=(u1-u1.net(1)).abs();
      // deviation from isometry
//   cout<<"deviation 1 from isometry ="<<error1<<endl;
   C_Matrix v1=~V*V;
   R error2=(v1-v1.net(1)).abs();
//   cout<<"deviation 2 from isometry ="<<error2<<endl;
      // deviation from isometry
   //R error3=(A-U*C_Matrix(w)*Vt).abs();
   R error3=A.dis(U*C_Matrix(w)*Vt);
      // deviation from representing the matrix we started from
//   cout<<"deviation from reliability ="<<error3<<endl;
   bool b1=(U.dim1()==m);
   bool b2=(U.dim2()==d);
   bool b3=(w.dim()==d);
   bool b4=(Vt.dim1()==d);
   bool b5=(Vt.dim2()==n);
   R error4=(b1 && b2 && b3 && b4 && b5) ? 0 : 1;
      // deviation from the expected frame-type of
      // the matrices
//   cout<<"deviation from dimensionality ="<<error4<<endl;
   return error1+error2+error3+error4;
}

C_Matrix SVDC::operator()(const F<R,R>& f)const
{
   R_Vector fw(d);
   for (Z i=1;i<=d;i++) fw[i]=f(w[i]);
   return U*C_Matrix(fw)*Vt;
}

//////////////////////////// class SpcObj /////////////////////////////////

SpcObj::SpcObj(const C_Matrix& a, bool enfSym):n_(a.dim1()),mat_(a),
   val_(n_), vec_(n_,n_)
{
   const Z mL=2;
   Word loc("SpcObj(C_Matrix,bool)");
   CPM_MA
   cpmassert(n_==mat_.dim2(),loc);
   if (enfSym) mat_ = (mat_ + ~mat_)*C(0.5);
#if defined(CPM_USE_EIGEN)
   using namespace Eigen;
   using std::complex;
   Z i,j,k,l;
   Matrix<complex<R>,Dynamic,Dynamic> ea(n_,n_);
   for (i=1,k=0;i<=n_;++i,++k) for (j=1,l=0;j<=n_;++j,++l)
      ea(k,l) = mat_[i][j].std();
   SelfAdjointEigenSolver<Matrix<complex<R>,Dynamic,Dynamic> > ses(ea);
   Matrix<R,Dynamic,1> val=ses.eigenvalues();
   Matrix<complex<R>,Dynamic,Dynamic> vec=ses.eigenvectors();
   for (i=1,k=0;i<=n_;++i,++k) for (j=1,l=0;j<=n_;++j,++l)
      vec_[i][j]=vec(l,k);
   for (i=1,k=0;i<=n_;++i,++k) val_[i] = val(k);
#else
   Z i,j,d=n_;
   R am = mat_.supAbs();
   if (am==R(0.)){
      vec_=a.net(1);
   }
   else{
      R_Matrix matR = mat_.toRealForm();
      // is a 2n_ times 2n_ matrix
      SpcObjR sor(matR);
      R_Vector realSpec = sor.val();
      R_Matrix realEig = sor.vec();
      Z dR=realSpec.dim();
      cpmassert(dR==2*d,loc);
      for (i=1,j=1;i<=d;i++,j+=2){
         val_[i]=realSpec[j];
         C_Vector cfj=C_Vector::fromRealForm(realEig[j]);
         R r=cfj.nor_(); // important since the real eigenvectors are
         // orthonormal with respect to a real-valued scalar product
         cpmassert(r>0,loc);
         vec_[i]=cfj;
      }
   }
#endif
   CPM_MZ
}

C_Matrix SpcObj::func(F<R,R> const& f)const
{
   // return vec_.trn()*C_Matrix(val_(f))*vec_.cc(); // OK but more cryptic
   C_Matrix res(n_,n_); // dyadic representation, presumably faster
   for (Z i=1;i<=n_;++i) res+=C_Matrix(vec_[i],vec_[i])*C(f(val_[i]));
   return res;
}

C_Matrix SpcObj::func(F<R,C> const& f)const
{
   // return vec_.trn()*C_Matrix(val_(f))*vec_.cc(); // OK but more cryptic
   C_Matrix res(n_,n_); // dyadic representation, presumably faster
   for (Z i=1;i<=n_;++i) res+=C_Matrix(vec_[i],vec_[i])*f(val_[i]);
   return res;
}

namespace{
   R squareFunc(const R& x){return x*x;}
}

R SpcObj::getError()const
{
   R sum1=0.;
   Z i,j;
   R nInv = 1./n_;
   for (i=1;i<=n_;++i) sum1+=(mat_*vec_[i] - vec_[i]*C(val_[i])).abs();
   sum1*=nInv;
   R sum2=0.;
   for (i=1;i<=n_;i++){
         for (j=1;j<=n_;j++){
            sum2+=((vec_[i]|vec_[j]) - C(krn(i,j))).abs();
               //krn = Kronecker's delta
         }
   }
   sum2*=(nInv*nInv);
   R err3=(mat_*mat_-func(F<R,R>(squareFunc))).abs();
   R res = sum1 + sum2 + err3;
   return res;
}
