//? cpmdistribution.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmdistribution.h>
#include <cpmgraph.h>
#include <cpmalgorithms.h>
#include <cpmmini.h>

#include <algorithm>
#include <functional>

using CpmArrays::DisR;
using CpmArrays::R_Vector;
using CpmArrays::R_Matrix;
using CpmArrays::V;
using CpmRoot::R;
using CpmRoot::Z;
using CpmRoot::Word;
using CpmLinAlg::R2;
using CpmGeo::Iv;
using CpmFunctions::F;
using CpmFunctions::F2;
using CpmGraphics::Color;
using CpmGraphics::RED;
using CpmGraphics::Graph;
using CpmGraphics::rainBow;

//////////////////////// class DisR ////////////////////////////////////

Z DisR::verbose=0;

void DisR::adjust()
{
   f=CpmAlgorithms::Quantizer(range,n);
}

DisR::DisR(const std::vector<R>& v, Z n, R fac1, R fac2, bool all):
   Base(n)
{
   Z i,vs=(Z)v.size();
   if (vs<1)
      cpmerror("DisR::DisR(const std::vector<R>& v, Z n_): v.size()==0");
   R vL=v[0];
   R vU=vL;
   for (i=0;i<vs;i++){
      R val=v[i];
      if (val>vU) vU=val;
      if (val<vL) vL=val;
   }
   range=Iv(vL*fac1,vU*fac2);
   adjust();
   put(v,all);
}

DisR::DisR(const std::vector<Z>& v, Z n, R fac1, R fac2, bool all):
   Base(n)
{
   Z i,vs=(Z)v.size();
   if (vs<1)
      cpmerror("DisR::DisR(const std::vector<Z>& v, Z n_): v.size()==0");
   vector<R> vr(vs);
   for (i=0;i<vs;i++) vr[i]=v[i];
   R vL=vr[0];
   R vU=vL;
   for (i=0;i<vs;i++){
      R val=vr[i];
      if (val>vU) vU=val;
      if (val<vL) vL=val;
   }
   range=Iv(vL*fac1,vU*fac2);
   adjust();
   put(vr,all);
}

DisR::DisR(const V<R>& v, Z n, R fac1, R fac2, bool all):Base(n)
{
   Z i,vs=v.dim();
   if (vs<1)
      cpmerror("DisR::DisR(const V<R>& v, Z n_): v.dim()==0");
   vector<R> vr(vs);
   for (i=0;i<vs;i++) vr[i]=v[i+1];
   R vL=vr[0];
   R vU=vL;
   for (i=0;i<vs;i++){
      R val=vr[i];
      if (val>vU) vU=val;
      if (val<vL) vL=val;
   }
   range=Iv(vL*fac1,vU*fac2);
   adjust();
   put(vr,all);
}

bool DisR::prnOn(std::ostream& str)const
{
   cpmp(n);
   cpmp(rep);
   cpmp(range);
   return true;
}

bool DisR::scanFrom(std::istream& str)
{
   cpms(n);
   cpms(rep);
   cpms(range);
   adjust();
   return true;
}

void DisR::show(CpmGraphics::Graph& g, bool autoX)const
{
   Z mL=2;
   Word loc("DisR::show(CpmGraphics::Graph& g, bool autoX)");
   CPM_MA
   if (autoX){
      g.setX(range);
      cpmcerr<<endl<<"range="<<range<<endl;
   }
   R_Vector x(g.getIvX().centers(n));
   R_Vector y(n);
   for (Z i=1;i<=n;i++) y[i]=rep[i];
   g.show(x,y,Color(RED),1);
   CPM_MZ
}

namespace{
   R func(const R_Vector& a, const V<R>& x, const V<Z>& y)
      // misfit indicator with suitably normalized return value
   {
      R alpha=a[1];
      R s1=0;
      R s2=0;
      Z n=x.dim(),i;
      for (i=1;i<=n;i++){
         R val=-alpha*x[i];
         s1+=cpmexp(val*2);
         s2+=cpmexp(val)*y[i];
      }
      R c=s2/s1;
      R res1=0;
      R res2=0;
      for (i=1;i<=n;i++){
         R ri1=y[i];;
         R ri2=ri1-c*cpmexp(-alpha*x[i]);
         res1+=ri1*ri1;
         res2+=ri2*ri2;
      }
      return cpmsqrt(res2/res1);
   }
}

R2 DisR::tBoltzmann(R tShow, Word title)const
{
   const R tol=1e-3; // was 1e-3
      // relative tolerance for localizing a minimum
   const R biasWeight=3; // was 3
      // the larger this value, the closer to 0 one starts with
      // looking for the proper 'thermal energy'

   if (singular()>=0){  // then the
      // distribution is either void or concentrated on a
      // single point. The fit is trivial but perfect.
      // Normally no noticible computational burden.
      return R2(0,0);
   }

   R_Vector x(range.centers(n));
   F<R_Vector,R> f=F2<R_Vector,V<R>,V<Z>,R>(x,rep)(func);
   CpmAlgorithms::Mini mini(f,tol);
 //  mini.verbose=2;
   R_Vector start(1);
   start[1]=biasWeight/range.mean();
   R_Vector res=mini.getPoint(start);
   R tInv=res[1];
   R fit=mini.getValue();
   bool succ=mini.getSuc();
   if (!succ) fit=1;
   if (tShow>0){
      R_Matrix view(2);
      view[1]=R_Vector(n);
      view[2]=R_Vector(n);
      Z i;
      for (i=1;i<=n;i++)view[1][i]=rep[i];
      R s1=0;
      R s2=0;
      for (i=1;i<=n;i++){
         R val=-tInv*x[i];
         s1+=cpmexp(val*2);
         s2+=cpmexp(val)*rep[i];
      }
      R c=s2/s1;
      for (i=1;i<=n;i++){
         view[2][i]=c*cpmexp(-tInv*x[i]);
      }
      Graph g;
      g.show(x,view,rainBow,title,0.2);
      cpmwait(tShow);
   }
   return R2(R(1)/tInv,fit);
}

R2 CpmArrays::disBoltzmann(std::vector<R>& v, R frac, Z nBins)
{
   const R largeMisFit=2;
   const R uselessValue=0;
   using namespace std;
   Z n=cpmtoz(frac*static_cast<int>(v.size()));
   if (n<nBins) return R2(uselessValue,largeMisFit);
      // no useful result possible
   sort(v.begin(),v.end(),less<R>());
   vector<R> vCut(n);
   copy(v.begin(),v.begin()+n,vCut.begin());
   DisR dist(vCut,nBins);
   return dist.tBoltzmann();
}


