//? cpmlinalg.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmlinalg.h>
#include <cpmrmatrix.h>

using namespace CpmRoot;
using namespace CpmLinAlg;

using CpmArrays::X2;
using CpmArrays::R_Vector;
using CpmFunctions::F1;

/////////////////////////// class R2 ///////////////////////////////////

namespace{

R pytCom(R x1, R x2, bool careful)
{
   if (careful==false) return cpmsqrt(x1*x1+x2*x2);
   else return cpmrho(x1,x2);
}

}

R2::R2(V<R> const& v)
{
   Z d=v.dim();
   R zero=0;
   x1= (1<=d ? v[1] : zero);
   x2= (2<=d ? v[2] : zero);
}

bool R2::prnOn(ostream& out)const
{
   if (!CpmRoot::writeTitle("R2",out)) return false;
   if (!CpmRoot::write(x1,out)) return false;
   if (!CpmRoot::write(x2,out)) return false;
   return true;
}

bool R2::scanFrom(istream& in)
{
   if (!CpmRoot::read(x1,in)) return false;
   if (!CpmRoot::read(x2,in)) return false;
   return true;
}

Z R2::com(const R2& a)const
{
   if (x1<a.x1) return 1;
   if (x1>a.x1) return -1;
   if (x2<a.x2) return 1;
   if (x2>a.x2) return -1;
   return 0;
}

R R2::abs(bool careful)const
{
   return pytCom(x1,x2,careful);
}

R R2::dis(const R2& x)const
{
   R a=abs();
   R b=x.abs();
   R c=(*this-x).abs();
   return disDefFun(a,b,c);
}

X2<R,R2 > R2::polDec()const
{
   R r=abs(true);
   R2 pol=*this;
   R z(0.);
   if (r!=z){
     R rInv=cpminv(r); pol*=rInv;
     return X2<R,R2 >(r,pol);
   }
   else{
      return X2<R,R2 >(z,R2(R(1.),z));
   }
}

R R2::normalize_(R r)
{
   X2<R,R2> p=polDec(); R r0=p.first();
   *this=p.second()*r;  return r0;
}

R2 R2::carToPol()const
{
   return R2(abs(),cpmatan2(x2,x1));
}

R2 R2::polToCar()const
{
   return R2(x1*cpmcos(x2),x1*cpmsin(x2));
}

R R2::angle(R2 const& a)const
{
    R c = (*this|a); // scalar product
    R s=cpmabs(crsPrd(a)); // vector product
    return R2(c,s).carToPol()[2]; // polar angle
}

#define CPM_SC\
   Type res;\
   R ri;\
   Z d=res.dim(),i;\
   if (j==0){\
      for (i=1; i<=d; i++){\
         ri=(*this)[i];\
         res[i]=CpmRoot::ran(ri,j);\
      }\
   }\
   else{\
      Z j0=1000;\
      Z jIncr=CpmRoot::ranT<Z>(j0,j);\
      Z jc=j;\
      for (i=1; i<=d; i++){\
         ri=(*this)[i];\
         jc+=jIncr;\
         res[i]=CpmRoot::ran(ri,jc);\
      }\
   }\
   return res


R2 R2::ran(Z j)const{ CPM_SC;}


/////////////////////////// class Z2 ///////////////////////////////////

Z2::Z2(const V<Z>& v)
{
   Z d=v.dim();
   x1= (1<=d ? v[1] : 0);
   x2= (2<=d ? v[2] : 0);
}

bool Z2::prnOn(ostream& out)const
{
   if (!CpmRoot::writeTitle("Z2",out)) return false;
   if (!CpmRoot::write(x1,out)) return false;
   if (!CpmRoot::write(x2,out)) return false;
   return true;
}

bool Z2::scanFrom(istream& in)
{
   if (!CpmRoot::read(x1,in)) return false;
   if (!CpmRoot::read(x2,in)) return false;
   return true;
}

Z Z2::com(const Z2& a)const
{
   if (x1<a.x1) return 1;
   if (x1>a.x1) return -1;
   if (x2<a.x2) return 1;
   if (x2>a.x2) return -1;
   return 0;
}

R Z2::abs(bool careful)const
{
   return R2(R(x1),R(x2)).abs(careful);
}

R Z2::dis(const Z2& x)const
{
   R a=abs();
   R b=x.abs();
   R c=(*this-x).abs();
   return disDefFun(a,b,c);
}

X2<R,R2 > Z2::polDec()const
{
   R r=abs(true);
   R2 pol(x1,x2);
   if (r!=0){
     R rInv=R(1)/r; pol*=rInv;
     return X2<R,R2 >(r,pol);
   }
   else{
      return X2<R,R2 >(0.,R2(1.,0.));
   }
}

/////////////////////////// class Z3 ///////////////////////////////////

Z3::Z3(const V<Z>& v)
{
   Z d=v.dim();
   x1= (1<=d ? v[1] : 0);
   x2= (2<=d ? v[2] : 0);
   x3= (3<=d ? v[3] : 0);
}

bool Z3::prnOn(ostream& out)const
{
   if (!CpmRoot::writeTitle("Z3",out)) return false;
   if (!CpmRoot::write(x1,out)) return false;
   if (!CpmRoot::write(x2,out)) return false;
   if (!CpmRoot::write(x3,out)) return false;
   return true;
}

bool Z3::scanFrom(istream& in)
{
   if (!CpmRoot::read(x1,in)) return false;
   if (!CpmRoot::read(x2,in)) return false;
   if (!CpmRoot::read(x3,in)) return false;
   return true;
}

Z Z3::com(const Z3& a)const
{
   if (x1<a.x1) return 1;
   if (x1>a.x1) return -1;
   if (x2<a.x2) return 1;
   if (x2>a.x2) return -1;
   if (x3<a.x3) return 1;
   if (x3>a.x3) return -1;
   return 0;
}

R Z3::abs(bool careful)const
{
   return R3(R(x1),R(x2),R(x3)).abs(careful);
}

R Z3::dis(const Z3& x)const
{
   R a=abs();
   R b=x.abs();
   R c=(*this-x).abs();
   return disDefFun(a,b,c);
}

X2<R,R3> Z3::polDec()const
{
   R r=abs(true);
   R3 pol(x1,x2,x3);
   if (r!=0){
     R rInv=R(1)/r; pol*=rInv;
     return X2<R,R3>(r,pol);
   }
   else{
      return X2<R,R3>(0.,R3(1.,0.,0.));
   }
}

/////////////////////////// class R3 ///////////////////////////////////

R3::R3(const V<R>& v)
{
   Z d=v.dim();
   R zero=0;
   x1= (1<=d ? v[1] : zero);
   x2= (2<=d ? v[2] : zero);
   x3= (3<=d ? v[3] : zero);
}

bool R3::prnOn(ostream& out)const
{
   if (!CpmRoot::writeTitle("R3",out)) return false;
   if (!CpmRoot::write(x1,out)) return false;
   if (!CpmRoot::write(x2,out)) return false;
   if (!CpmRoot::write(x3,out)) return false;
   return true;
}

bool R3::scanFrom(istream& in)
{
   if (!CpmRoot::read(x1,in)) return false;
   if (!CpmRoot::read(x2,in)) return false;
   if (!CpmRoot::read(x3,in)) return false;
   return true;
}

Z R3::com(const R3& a)const
{
   if (x1<a.x1) return 1;
   if (x1>a.x1) return -1;
   if (x2<a.x2) return 1;
   if (x2>a.x2) return -1;
   if (x3<a.x3) return 1;
   if (x3>a.x3) return -1;
   return 0;
}

R R3::abs(bool careful)const
{
   if (careful==false) return cpmsqrt(x1*x1+x2*x2+x3*x3);
   R z(0.);
   R y1=(x1>=z ? x1 : -x1);
   R y2=(x2>=z ? x2 : -x2);
   R y3=(x3>=z ? x3 : -x3);
   R y=y1+y2+y3;
   if (y==z) return z; // to avoid zero-division warnings
   R yInv=cpminv(y);
   y1*=yInv;
   y2*=yInv;
   y3*=yInv;
   return y*cpmsqrt(y1*y1+y2*y2+y3*y3);
}

R R3::dis(const R3& x)const
{
   R a=abs();
   R b=x.abs();
   R c=(*this-x).abs();
   return disDefFun(a,b,c);
}

X2<R,R3 > R3::polDec()const
{
   R r=abs(true);
   R3 pol=*this;
   R z(0.);
   if (r!=z){
     R rInv=cpminv(r); pol*=rInv;
     return X2<R,R3 >(r,pol);
   }
   else{
     return X2<R,R3 >(z,R3(R(1.),z,z));
   }
}

R R3::normalize_(R r)
{
   X2<R,R3> p=polDec(); R r0=p.first();
   *this=p.second()*r;  return r0;
}

R3 R3::carToPol()const
{
   R r=pytCom(x1,x2,true);
   R phi=cpmatan2(x2,x1);
   R rho=pytCom(r,x3,true);
   R decl=cpmatan2(x3,r);
   return R3(rho,phi,decl);
}

R3 R3::polToCar()const
{
   R z=x1*cpmsin(x3); // x1=r, x2=phi, x3=declination
   R r=x1*cpmcos(x3);
   R x=r*cpmcos(x2);
   R y=r*cpmsin(x2);
   return R3(x,y,z);
}

R3 R3::ast(R3 const& x)const
{
   R rr=x1*x1+x2*x2+x3*x3;
   R f1=R(1)-rr;
   R f2=(x1*x.x1+x2*x.x2+x3*x.x3)*2;
   R f3=R(1)/(rr+1); // nominator >= 1
   return R3((f1*x.x1+f2*x1+(x2*x.x3-x3*x.x2)*2)*f3,
             (f1*x.x2+f2*x2+(x3*x.x1-x1*x.x3)*2)*f3,
             (f1*x.x3+f2*x3+(x1*x.x2-x2*x.x1)*2)*f3 );
}

R3 R3::circ(R3 const& x)const
{
   const R eps=1e-12;
   R sp=x1*x.x1+x2*x.x2+x3*x.x3;
   R f1=R(1)-sp;
   if (cpmabs(f1)<eps){
      if (f1>=0.) f1 = eps;
      else f1 = -eps;
   }
   R f2=R(1)/f1;
   return R3((x1+x.x1+x2*x.x3-x3*x.x2)*f2,
             (x2+x.x2+x3*x.x1-x1*x.x3)*f2,
             (x3+x.x3+x1*x.x2-x2*x.x1)*f2 );
}

R3 R3::toEulerAng()const
// Here we interpret *this as a rotation vector
// (list of Euler-Rodrigues parameters)
// and return the Euler  angles phi, theta, psi
// Source: my notes UM/Astro 87-11-4/6
{
   R2 a=R2(x1,x2).carToPol();
   R2 b=R2(1,x3).carToPol();
   R phi=a.x2+b.x2;
   R psi=-a.x2+b.x2;
   R theta=cpmatan(a.x1/b.x1)*2.0;
   return R3(phi,theta,psi);
}

R R3::angle(R3 const& a)const
{
    R c = (*this|a); // scalar product
    R s=crsPrd(a).abs(); // vector product
    return R2(c,s).carToPol()[2]; // polar angle
}

R3 R3::ran(Z j)const{ CPM_SC;}

R3 R3::ran(R r)const
{
   R x=randomR()*2.-1.;
   R y=randomR()*2.-1.;
   R z=randomR()*2.-1.;
   R r2=x*x+y*y+z*z;
   while (r2 > R(1.)){
      x=randomR()*2.-1.;
      y=randomR()*2.-1.;
      z=randomR()*2.-1.;
      r2=x*x+y*y+z*z;
   }
   return *this + R3(x,y,z)*r;
}

/////////////////////////// class Z4 ///////////////////////////////////

Z4::Z4(const V<Z>& v)
{
   Z d=v.dim();
   x1= (1<=d ? v[1] : 0);
   x2= (2<=d ? v[2] : 0);
   x3= (3<=d ? v[3] : 0);
   x4= (4<=d ? v[4] : 0);
}

bool Z4::prnOn(ostream& out)const
{
   if (!CpmRoot::writeTitle("Z4",out)) return false;
   if (!CpmRoot::write(x1,out)) return false;
   if (!CpmRoot::write(x2,out)) return false;
   if (!CpmRoot::write(x3,out)) return false;
   if (!CpmRoot::write(x4,out)) return false;
   return true;
}

bool Z4::scanFrom(istream& in)
{
   if (!CpmRoot::read(x1,in)) return false;
   if (!CpmRoot::read(x2,in)) return false;
   if (!CpmRoot::read(x3,in)) return false;
   if (!CpmRoot::read(x4,in)) return false;
   return true;
}

Z Z4::com(const Z4& a)const
{
   if (x1<a.x1) return 1;
   if (x1>a.x1) return -1;
   if (x2<a.x2) return 1;
   if (x2>a.x2) return -1;
   if (x3<a.x3) return 1;
   if (x3>a.x3) return -1;
   if (x4<a.x4) return 1;
   if (x4>a.x4) return -1;
   return 0;
}

R Z4::abs(bool careful)const
{
   return R4(R(x1),R(x2),R(x3),R(x4)).abs(careful);
}

R Z4::dis(const Z4& x)const
{
   R a=abs();
   R b=x.abs();
   R c=(*this-x).abs();
   return disDefFun(a,b,c);
}

X2<R,R4> Z4::polDec()const
{
   R r=abs(true);
   R4 pol(x1,x2,x3,x4);
   if (r!=0){
     R rInv=R(1)/r; pol*=rInv;
     return X2<R,R4>(r,pol);
   }
   else{
      return X2<R,R4>(0.,R4(1.,0.,0.,0.));
   }
}

/////////////////////////// class R4 ///////////////////////////////////

R4::R4(const V<R>& v)
{
   Z d=v.dim();
   R zero=0;
   x1= (1<=d ? v[1] : zero);
   x2= (2<=d ? v[2] : zero);
   x3= (3<=d ? v[3] : zero);
   x4= (4<=d ? v[4] : zero);
}

bool R4::prnOn(ostream& out)const
{
   if (!CpmRoot::writeTitle("R4",out)) return false;
   if (!CpmRoot::write(x1,out)) return false;
   if (!CpmRoot::write(x2,out)) return false;
   if (!CpmRoot::write(x3,out)) return false;
   if (!CpmRoot::write(x4,out)) return false;
   return true;
}

bool R4::scanFrom(istream& in)
{
   if (!CpmRoot::read(x1,in)) return false;
   if (!CpmRoot::read(x2,in)) return false;
   if (!CpmRoot::read(x3,in)) return false;
   if (!CpmRoot::read(x4,in)) return false;
   return true;
}

Z R4::com(const R4& a)const
{
   if (x1<a.x1) return 1;
   if (x1>a.x1) return -1;
   if (x2<a.x2) return 1;
   if (x2>a.x2) return -1;
   if (x3<a.x3) return 1;
   if (x3>a.x3) return -1;
   if (x4<a.x4) return 1;
   if (x4>a.x4) return -1;
   return 0;
}

R R4::abs(bool careful)const
{
   if (careful==false) return cpmsqrt(x1*x1+x2*x2+x3*x3+x4*x4);
   R z(0.);
   R y1=(x1>=z ? x1 : -x1);
   R y2=(x2>=z ? x2 : -x2);
   R y3=(x3>=z ? x3 : -x3);
   R y4=(x4>=z ? x4 : -x4);
   R y=y1+y2+y3+y4;
   if (y==z) return z; // to avoid zero-division warnings
   R yInv=cpminv(y);
   y1*=yInv;
   y2*=yInv;
   y3*=yInv;
   y4*=yInv;
   return y*cpmsqrt(y1*y1+y2*y2+y3*y3+y4*y4);
}

R R4::dis(const R4& x)const
{
   R a=abs();
   R b=x.abs();
   R c=(*this-x).abs();
   return disDefFun(a,b,c);
}

X2<R,R4 > R4::polDec()const
{
   R r=abs(true);
   R4 pol=*this;
   R z(0.);
   if (r!=z){
     R rInv=cpminv(r); pol*=rInv;
     return X2<R,R4 >(r,pol);
   }
   else{
      return X2<R,R4 >(z,R4(R(1.),z,z,z));
   }
}

R R4::normalize_(R r)
{
   X2<R,R4> p=polDec(); R r0=p.first();
   *this=p.second()*r;  return r0;
}

R4 R4::ran(Z j)const{ CPM_SC;}

#undef CPM_SC
///////////////////////// RR2 //////////////////////////////////////////

R2 CpmLinAlg::operator*(const RR2& a, const R2& x)
   {
      R2 res;
      res.x1=(a.x1|x);
      res.x2=(a.x2|x);
      return res;
   }

RR2 operator*(const RR2& a, const RR2& b)
   {
      RR2 res;
      res.x1.x1=a.x1.x1*b.x1.x1+a.x1.x2*b.x2.x1;
      res.x1.x2=a.x1.x1*b.x1.x2+a.x1.x2*b.x2.x2;
      res.x2.x1=a.x2.x1*b.x1.x1+a.x2.x2*b.x2.x1;
      res.x2.x2=a.x2.x1*b.x1.x2+a.x2.x2*b.x2.x2;
      return res;
   }

RR2 operator+(const RR2& a, const RR2& b)
   {
      RR2 res=a;
      res.x1+=b.x1;
      res.x2+=b.x2;
      return res;
   }

RR2 CpmLinAlg::operator~(const RR2& a)
   {
      RR2 res=a;
      res.x1.x2=a.x2.x1;
      res.x2.x1=a.x1.x2;
      return res;
   }

RR2 RR2::operator !()const
{
   R detInv=CpmRoot::inv(det());
   RR2 res;
   res.x1.x1=x2.x2*detInv;
   res.x2.x2=x1.x1*detInv;
   res.x1.x2=-x1.x2*detInv;
   res.x2.x1=-x2.x1*detInv;
   return res;
}

RR2 RR2::zero;

RR2 RR2::unit=RR2(R2(1,0),R2(0,1));

bool RR2::prnOn(ostream& out)const
{
     if (!x1.prnOn(out)) return false;
     if (!x2.prnOn(out)) return false;
     return true;
}

bool RR2::scanFrom(istream& in)
{
   if (!x1.scanFrom(in)) return false;
   if (!x2.scanFrom(in)) return false;
   return true;
}

RR2::RR2(R2 const& x, F<R2,R2> const& f, R h)
{
   cpmassert(h!=0,"RR2(R2,F<R2,R2>,R)");

   R h2=h*2;
   R hi=cpminv(h*12);

   R2 x1p=x, x1m=x, x2p=x, x2m=x;
   R2 y1p=x, y1m=x, y2p=x, y2m=x;

   x1p.x1+=h;
   x1m.x1-=h;

   x2p.x2+=h;
   x2m.x2-=h;

   y1p.x1+=h2;
   y1m.x1-=h2;

   y2p.x2+=h2;
   y2m.x2-=h2;

   RR2 res;
   // avoid making more than one subtraction
   R2 fp=f(x1p)*8+f(y1m);
   R2 fm=f(x1m)*8+f(y1p);
   res.x1=(fp-fm)*hi;

   fp=f(x2p)*8+f(y2m);
   fm=f(x2m)*8+f(y2p);
   res.x2=(fp-fm)*hi;

   *this=~res;
}

namespace{
   R2 toF_aux2(R2 const& x, RR2 const& mat){ return mat*x;}
}

F<R2,R2> RR2::toF()const
{
   return F1<R2,RR2,R2>(*this)(toF_aux2);
}

///////////////////////////////// class RR3 //////////////////////////////

R3 CpmLinAlg::operator*(const RR3& a, const R3& x)
   {
      R3 res;
      res.x1=(a.x1|x);
      res.x2=(a.x2|x);
      res.x3=(a.x3|x);
      return res;
   }

RR3 operator*(const RR3& a, const RR3& b)
   {
      RR3 res;
#ifndef CPM_NOT_KILL_OPTIMIZER
      res.x1.x1=a.x1.x1*b.x1.x1 + a.x1.x2*b.x2.x1 + a.x1.x3*b.x3.x1;
      res.x1.x2=a.x1.x1*b.x1.x2 + a.x1.x2*b.x2.x2 + a.x1.x3*b.x3.x2;
      res.x1.x3=a.x1.x1*b.x1.x3 + a.x1.x2*b.x2.x3 + a.x1.x3*b.x3.x3;

      res.x2.x1=a.x2.x1*b.x1.x1 + a.x2.x2*b.x2.x1 + a.x2.x3*b.x3.x1;
      res.x2.x2=a.x2.x1*b.x1.x2 + a.x2.x2*b.x2.x2 + a.x2.x3*b.x3.x2;
      res.x2.x3=a.x2.x1*b.x1.x3 + a.x2.x2*b.x2.x3 + a.x2.x3*b.x3.x3;

      res.x3.x1=a.x3.x1*b.x1.x1 + a.x3.x2*b.x2.x1 + a.x3.x3*b.x3.x1;
      res.x3.x2=a.x3.x1*b.x1.x2 + a.x3.x2*b.x2.x2 + a.x3.x3*b.x3.x2;
      res.x3.x3=a.x3.x1*b.x1.x3 + a.x3.x2*b.x2.x3 + a.x3.x3*b.x3.x3;
#else
      cpmerror(
         "definition of RR3*RR3 disabled by a macro in cpmlinalg.cpp");
#endif
      return res;
   }

RR3 operator+(const RR3& a, const RR3& b)
   {
      RR3 res=a;
      res.x1+=b.x1;
      res.x2+=b.x2;
      res.x3+=b.x3;
      return res;
   }

RR3 CpmLinAlg::operator~(const RR3& a)
   {
      RR3 res(a);

      res.x1.x2=a.x2.x1;
      res.x1.x3=a.x3.x1;

      res.x2.x3=a.x3.x2;
      res.x2.x1=a.x1.x2;

      res.x3.x1=a.x1.x3;
      res.x3.x2=a.x2.x3;

      return res;
   }

RR3 RR3::zero;

RR3 RR3::unit=RR3(R3(1,0,0),R3(0,1,0),R3(0,0,1));

bool RR3::prnOn(ostream& out)const
{
     if (!x1.prnOn(out)) return false;
     if (!x2.prnOn(out)) return false;
     if (!x3.prnOn(out)) return false;
     return true;
}

bool RR3::scanFrom(istream& in)
{
   if (!x1.scanFrom(in)) return false;
   if (!x2.scanFrom(in)) return false;
   if (!x3.scanFrom(in)) return false;
   return true;
}

RR3::RR3(R3 const& r):x1(),x2(),x3()
// nearly verbatim carried over from function AxVec::toMat
{
   R3 d=-r; // overall adjustment
   R d11=d.x1*d.x1;
   R d22=d.x2*d.x2;
   R d33=d.x3*d.x3;
   R d2=d11+d22+d33;
   R ha=(R(1)-d2)*0.5;
   R hb=R(2)/(d2+1); // OK without test for 0
   x1.x1=hb*(ha+d11);
   x2.x2=hb*(ha+d22);
   x3.x3=hb*(ha+d33);
   R dxy=d.x1*d.x2;
   x1.x2=hb*(dxy+d.x3);
   x2.x1=hb*(dxy-d.x3);
   dxy=d.x1*d.x3;
   x1.x3=hb*(dxy-d.x2);
   x3.x1=hb*(dxy+d.x2);
   dxy=d.x2*d.x3;
   x2.x3=hb*(dxy+d.x1);
   x3.x2=hb*(dxy-d.x1);
}

RR3::RR3(R3 const& x, F<R3,R3> const& f, R h)
{
   cpmassert(h!=0,"RR3(R3,F<R3,R3>,R)");

   R h2=h*2;
   R hi=cpminv(h*12);

   R3 x1p=x, x1m=x, x2p=x, x2m=x, x3p=x, x3m=x;
   R3 y1p=x, y1m=x, y2p=x, y2m=x, y3p=x, y3m=x;

   x1p.x1+=h;
   x1m.x1-=h;

   x2p.x2+=h;
   x2m.x2-=h;

   x3p.x3+=h;
   x3m.x3-=h;

   y1p.x1+=h2;
   y1m.x1-=h2;

   y2p.x2+=h2;
   y2m.x2-=h2;

   y3p.x3+=h2;
   y3m.x3-=h2;

   RR3 res;
   // avoid making more than one subtraction
   R3 fp=f(x1p)*8+f(y1m);
   R3 fm=f(x1m)*8+f(y1p);
   res.x1=(fp-fm)*hi;

   fp=f(x2p)*8+f(y2m);
   fm=f(x2m)*8+f(y2p);
   res.x2=(fp-fm)*hi;

   fp=f(x3p)*8+f(y3m);
   fm=f(x3m)*8+f(y3p);
   res.x3=(fp-fm)*hi;

   *this=~res;
}

X2<R,R3> RR3::toAngAndAxis()const
// Source: UM/T3 86-1-17/3 (Folder Task 3)
{
   R trace=x1.x1+x2.x2+x3.x3;
   R d1=x3.x2-x2.x3; // the second term determines the cyclic
   // supplement: 2 3 1, thus d1
   R d2=x1.x3-x3.x1; // 3 1 2
   R d3=x2.x1-x1.x2; // 1 2 3
   R3 d(d1,d2,d3);
   R da=d.nor_();
   R phi=R2(trace-1,da).arg();
   return X2<R,R3>(phi,d);
}

R RR3::det()const
{
   return
       x1.x1*(x2.x2*x3.x3-x2.x3*x3.x2)
      -x1.x2*(x2.x1*x3.x3-x2.x3*x3.x1)
      +x1.x3*(x2.x1*x3.x2-x2.x2*x3.x1);
}

inline R inl(R a, R b, R c, R d){ return  a*d-b*c;}

RR3 RR3::conGrd()const
{
   R di=cpminv(det());
   RR3 res;
   res.x1.x1= inl(x2.x2,x2.x3,x3.x2,x3.x3)*di;
   res.x1.x2=-inl(x2.x1,x2.x3,x3.x1,x3.x3)*di;
   res.x1.x3= inl(x2.x1,x2.x2,x3.x1,x3.x2)*di;

   res.x2.x1=-inl(x1.x2,x1.x3,x3.x2,x3.x3)*di;
   res.x2.x2= inl(x1.x1,x1.x3,x3.x1,x3.x3)*di;
   res.x2.x3=-inl(x1.x1,x1.x2,x3.x1,x3.x2)*di;

   res.x3.x1= inl(x1.x2,x1.x3,x2.x2,x2.x3)*di;
   res.x3.x2=-inl(x1.x1,x1.x3,x2.x1,x2.x3)*di;
   res.x3.x3= inl(x1.x1,x1.x2,x2.x1,x2.x2)*di;

   return res;
}

RR3 RR3::operator !()const
// built as previous code, only transpose the result
// directly (eficiency)
{
   R di=cpminv(det());
   RR3 res;
   res.x1.x1= inl(x2.x2,x2.x3,x3.x2,x3.x3)*di;
   res.x2.x1=-inl(x2.x1,x2.x3,x3.x1,x3.x3)*di;
   res.x3.x1= inl(x2.x1,x2.x2,x3.x1,x3.x2)*di;

   res.x1.x2=-inl(x1.x2,x1.x3,x3.x2,x3.x3)*di;
   res.x2.x2= inl(x1.x1,x1.x3,x3.x1,x3.x3)*di;
   res.x3.x2=-inl(x1.x1,x1.x2,x3.x1,x3.x2)*di;

   res.x1.x3= inl(x1.x2,x1.x3,x2.x2,x2.x3)*di;
   res.x2.x3=-inl(x1.x1,x1.x3,x2.x1,x2.x3)*di;
   res.x3.x3= inl(x1.x1,x1.x2,x2.x1,x2.x2)*di;

   return res;
}

namespace{
   R3 toF_aux(R3 const& x, RR3 const& mat){ return mat*x;}
}

F<R3,R3> RR3::toF()const
{
   return F1<R3,RR3,R3>(*this)(toF_aux);
}

//////////////////////////////////// class RR4 ///////////////////////////

R4 operator*(const RR4& a, const R4& x)
{
   R4 res;
   res.x1=(a.x1|x);
   res.x2=(a.x2|x);
   res.x3=(a.x3|x);
   res.x4=(a.x4|x);
   return res;
}

RR4 operator*(const RR4& a, const RR4& b)
   {
      RR4 res;

      res.x1.x1=a.x1.x1*b.x1.x1 + a.x1.x2*b.x2.x1 + a.x1.x3*b.x3.x1
         + a.x1.x4*b.x4.x1;
      res.x1.x2=a.x1.x1*b.x1.x2 + a.x1.x2*b.x2.x2 + a.x1.x3*b.x3.x2
         + a.x1.x4*b.x4.x2;
      res.x1.x3=a.x1.x1*b.x1.x3 + a.x1.x2*b.x2.x3 + a.x1.x3*b.x3.x3
         + a.x1.x4*b.x4.x3;
      res.x1.x4=a.x1.x1*b.x1.x4 + a.x1.x2*b.x2.x4 + a.x1.x3*b.x3.x4
         + a.x1.x4*b.x4.x4;

      res.x2.x1=a.x2.x1*b.x1.x1 + a.x2.x2*b.x2.x1 + a.x2.x3*b.x3.x1
         + a.x2.x4*b.x4.x1;
      res.x2.x2=a.x2.x1*b.x1.x2 + a.x2.x2*b.x2.x2 + a.x2.x3*b.x3.x2
         + a.x2.x4*b.x4.x2;
      res.x2.x3=a.x2.x1*b.x1.x3 + a.x2.x2*b.x2.x3 + a.x2.x3*b.x3.x3
         + a.x2.x4*b.x4.x3;
      res.x2.x4=a.x2.x1*b.x1.x4 + a.x2.x2*b.x2.x4 + a.x2.x3*b.x3.x4
         + a.x2.x4*b.x4.x4;

      res.x3.x1=a.x3.x1*b.x1.x1 + a.x3.x2*b.x2.x1 + a.x3.x3*b.x3.x1
         + a.x3.x4*b.x4.x1;
      res.x3.x2=a.x3.x1*b.x1.x2 + a.x3.x2*b.x2.x2 + a.x3.x3*b.x3.x2
         + a.x3.x4*b.x4.x2;
      res.x3.x3=a.x3.x1*b.x1.x3 + a.x3.x2*b.x2.x3 + a.x3.x3*b.x3.x3
         + a.x3.x4*b.x4.x3;
      res.x3.x4=a.x3.x1*b.x1.x4 + a.x3.x2*b.x2.x4 + a.x3.x3*b.x3.x4
         + a.x3.x4*b.x4.x4;

      res.x4.x1=a.x4.x1*b.x1.x1 + a.x4.x2*b.x2.x1 + a.x4.x3*b.x3.x1
         + a.x4.x4*b.x4.x1;
      res.x4.x2=a.x4.x1*b.x1.x2 + a.x4.x2*b.x2.x2 + a.x4.x3*b.x3.x2
         + a.x4.x4*b.x4.x2;
      res.x4.x3=a.x4.x1*b.x1.x3 + a.x4.x2*b.x2.x3 + a.x4.x3*b.x3.x3
         + a.x4.x4*b.x4.x3;
      res.x4.x4=a.x4.x1*b.x1.x4 + a.x4.x2*b.x2.x4 + a.x4.x3*b.x3.x4
         + a.x4.x4*b.x4.x4;
      return res;
   }

RR4 operator+(const RR4& a, const RR4& b)
   {
      RR4 res=a;
      res.x1+=b.x1;
      res.x2+=b.x2;
      res.x3+=b.x3;
      res.x4+=b.x4;
      return res;
   }

RR4 operator~(const RR4& a)
   {
      RR4 res(a);

      res.x1.x2=a.x2.x1;
      res.x1.x3=a.x3.x1;
      res.x1.x4=a.x4.x1;

      res.x2.x1=a.x1.x2;
      res.x2.x3=a.x3.x2;
      res.x2.x4=a.x4.x2;

      res.x3.x1=a.x1.x3;
      res.x3.x2=a.x2.x3;
      res.x3.x4=a.x4.x3;

      res.x4.x1=a.x1.x4;
      res.x4.x2=a.x2.x4;
      res.x4.x3=a.x3.x4;

      return res;
   }

RR4 RR4::zero;

RR4 RR4::unit=
RR4(R4(1,0,0,0),R4(0,1,0,0),R4(0,0,1,0),R4(0,0,0,1));

RR4 RR4::g(R4(1,0,0,0),R4(0,-1,0,0),R4(0,0,-1,0),R4(0,0,0,-1));


bool RR4::prnOn(ostream& out)const
{
     if (!x1.prnOn(out)) return false;
     if (!x2.prnOn(out)) return false;
     if (!x3.prnOn(out)) return false;
     if (!x4.prnOn(out)) return false;
     return true;
}

bool RR4::scanFrom(istream& in)
{
   if (!x1.scanFrom(in)) return false;
   if (!x2.scanFrom(in)) return false;
   if (!x3.scanFrom(in)) return false;
   if (!x4.scanFrom(in)) return false;
   return true;
}
