//? cpmperm.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmperm.h>
#include <cpmrvector.h>
#include <regex>

using namespace CpmArrays;
using namespace CpmRoot;

Perm::Perm(const R_Vector& v):rep(v.permutationForIncreasingOrder()){}

Perm::Perm(Z n):rep(n)
{
   for (Z i=1;i<=n;i++) rep[i]=i;
}

Perm::Perm( Z n, Z s):rep(n)
{
   Z i,sr=s;
   while (sr<0) sr+=n; // does not change the meaning of sr
   while (sr>=n) sr-=n; // now 0 <= sr <= n-1
   for (i=1;i<=n-sr;i++) rep[i]=i+sr;
   for (i=n-sr+1;i<=n;i++) rep[i]=i+sr-n;
}

Perm::Perm( Z n, Z i, Z j):rep(n)
   // constructor for a transposition of i and j.
{
   if (i<1 || i>n) cpmerror("Perm", "transposing index out of range");
   if (j<1 || j>n) cpmerror("Perm", "transposing index out of range");
   for (Z k=1;k<=n;k++) rep[k]=k;
   rep[i]=j;
   rep[j]=i;
}

Perm Perm::operator * (const Perm& p)const
{
   Z n=dim();
   cpmassert(n==p.dim(),"Perm::operator * (const Perm& p)const");
   Perm res(n);
   for (Z i=1;i<=n;i++) res.rep[i]=rep[p[i]];
   return res;
}

Perm Perm::operator ! (void)const
   // Returns a Perm which is the inverse of *this.
{
     Vo<Z> repRes=rep.permutationForIncreasingOrder();
   return Perm(repRes,true);
}

V<Z> Perm::cycDec()const
 // The programm is tested with many examples.
 // Presently the unorthodox control structure looks well understandable
 // to the author.
{
   Z n=rep.dim(), i, i_start=0, i_current, m=n, i_cy=1;
   V<Z> present(n);
      // An integer-valued array which monitors whether an element is
      // still present in the list and waits for being grouped into a
      // cycle.
   V<Z> cyclelength(n);
   for (i=1;i<=n;i++) present[i]=1;
Newcycle:     // Begin with creating a new cycle
   for (i=1;i<=n;i++){
      if (present[i]){     // find smallest available starting point
         i_start=i;
         break;
      }
   }
   cyclelength[i_cy]++;
   i_current=i_start;
Newelement:
   present[i_current]=0; // this value of i_current should never be
      // chosen as a starting point
   m--;
   i_current=rep[i_current];
   if (i_current==i_start){
      if (m==0) goto End; else {i_cy++; goto Newcycle;}
   }
   else{
      cyclelength[i_cy]++;
      goto Newelement;
   }
End:
   n=i_cy;
   V<Z> vr(n);
   for (i=1;i<=n;i++) vr[i]=cyclelength[i];
   return vr;
}

namespace{
Z numberParity(Z i)
   // Parity of integers
{
   if (2*(i/2)==i) return 1; else return -1;
}
}

Z Perm::par()const
   // Parity of permutations based on cycle decomposition.
   // See S. Mac Lane, G. Birkhoff: Algebra, MacMillan Company, 1967
   // p.95-96, Theorem 17 and Corollary 2.
{
   V<Z> vr=cycDec();
   Z n=vr.dim();
   Z x=-n;
   for (Z i=1;i<=n;i++) x+=vr[i];
   return numberParity(x);
}

Z Perm::com(Perm const& obj)const
{
   return rep.com(obj.rep);
}

bool Perm::prnOn(ostream& out)const
{
   Z i,n=dim();
   CpmRoot::writeTitle("Perm",out);
   CpmRoot::write(n,out);
   for (i=1;i<=n;i++) CpmRoot::write(rep[i],out);
   return true;
}

bool Perm::scanFrom(istream& in)
{
   Z i,n=dim();
   CpmRoot::read(n,in);
   for (i=1;i<=n;i++) CpmRoot::read(rep[i],in);
   return true;
}

Perm Perm::ran(Z j)const
{
   Z n=dim();
   R_Vector r(n,1);
   r=r.ran(j);
   return Perm(r);
}

V<Perm> Perm::all(Z n)
{
   cpmassert(n>0,"Perm::all(Z)");
   std::vector<Z> vz(n);
   Z i;
   for (i=0;i<n;++i) vz[i]=i+1;
   Z nf=CpmRootX::fac(n);
   V<Perm> res(nf);
   for (i=1;i<=nf;++i){
      res[i]=Perm(V<Z>(vz,1),true);
      std::next_permutation(vz.begin(),vz.end());
   }
   return res;
}
