//? cpmrecord.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <vector>

#include <cpmrecord.h>
#include <cpmsystem.h>
#include <cpmgreg.h>

using std::vector;
using std::istringstream;
using std::ostringstream;

using CpmRoot::Word;
using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::toStr;
using CpmRootX::B;
using CpmRootX::Z1;
using CpmRootX::R1;
using CpmRootX::Record;
using CpmRootX::Recordable;
using CpmRootX::Text;
using CpmRootX::aux::Key;
using CpmRootX::aux::TextH;
using CpmRootX::aux::SecCol;
using CpmArrays::V;
using CpmArrays::S;
using CpmArrays::M;
using CpmArrays::X2;
using CpmFunctions::F;
using CpmFunctions::F1;
using CpmSystem::Message;
using CpmSystem::IFileStream;
using CpmSystem::OFileStream;
using namespace CpmStd;

////////////////// class SecCol //////////////////////////

CpmRootX::aux::SecCol::SecCol(const Word& fileName, bool binaryI):
binary(binaryI),nFiles_(1)
{
   el=binary ? "\r\n" : "\n";
      // works, but with binary replaced by !binary
      // it works too at least in some contexts
   fns.add(fileName);
   valid=B(ini());
}

namespace {

   bool skipHead(string& str)
      // returns true if a leading "c " (named head) was found.
      // after call, str does no longer contain the head.
      // This will be applied only for a str which already is
      // freed from leading white space
   {
      Z n=(Z)str.size();
      if (n<3){
         return false;
      }
      char c1=str[0];
      char c2=str[1];
      if (c1=='c' && c2==' '){
         str.erase(0,2);
         return true;
      }
      else{
         return false;
      }
   }

   bool replaceFillCharacters_(string& str)
      // defining replacements that are OK for non-comment lines
      // commas should not be replaced since they are important at least in
      // values of type Ws
   {
      Z nl=(Z)str.size();
      bool res=false;
      char rep=' '; // rep=replacement
      for (Z i=0;i<nl;++i){
         char c=str[i];
         if (c=='='||/* c==',' || */c==';'){
            str[i]=rep;
            if (/*c==',' ||*/ c==';') res=true;
         }
      }
      return res;
   }

   bool isInclude(const string& line, Word& fnFound)
      // returns true if line imports a file name
   {
      Word w1("include");
      Word w2("F");
      Word l(line);
      Word fw=l.firstWord_();
      if (fw==w1 || fw==w2){
         fnFound=l.firstWord_();
         fnFound=Message::ext(fnFound); // 2008-04-21
         return true;
      }
      else{
         return false;
      }
   }

   bool ignore(const string& line)
      // returns true if line is to be ignored
   {
      Word l(line);
      Word w1=l.firstWord_();
      return w1.length()==0 || CpmRoot::startsComment(w1[1]);
   }

   bool ignoreX(const string& line)
      // returns true if line is to be ignored
   {
      Word l(line);
      Word w1=l.firstWord_();
      return w1.length()==0 || CpmRoot::startsComment(w1[1]);
   }

   bool isSectionHeader(const string& line, Word& headerFound )
      // returns true if line imports a section heading
      // notice that lines starting with '/' or ';' or ....
      // will never shine up as 'line'
   {
      Word l(line);
      Word w1=l.firstWord();
      bool res=!Record::startsEntry(w1);
      if (res==true){
         l.skipSurSpc_();
         headerFound=l;
      }
      return res;
   }
}

bool CpmRootX::aux::SecCol::ini()
{
   Z mL=2;
   static Word loc("CpmRootX::SecCol::ini()");
   CPM_MA
   M<string,string> sections;
   bool safe=false;
   while (!fns.isVoid()){ // work through all files to which the
      // primary file refers and that may get referred to in such
      // files
      Word fn=fns.fir();
      IFileStream fs(fn,safe,binary);
      if (!fs.isVal()){
         cpmwarning(loc&" file stream unvalid");
         CPM_MZ
         return false;
      }
      Word activeHeader;
      Word nextFile;
      string line;

      while ( std::getline(fs(),line) ){ // reading fs line by line
         if (ignoreX(line)){
            ; // do nothing
         }
         else{
            replaceFillCharacters_(line);
               // This means also eliminating ';' which
               // could turn comment-lines into non-comment lines
               // if applied to comment lines. In the present
               // place we are sure not to make this mistake
            if (isInclude(line,nextFile)){
               fns.add(nextFile);
               nFiles_+=1;
               // a new action item for the outer while loop
            }
            else{
               string lineToAdd=line+el;
               bool b=isSectionHeader(line,activeHeader);
               // find the correct section to which line should
               // be added
               if (!b){
                  string key=activeHeader.str();
                  sections[key]+=lineToAdd;
               }
            }
         }
      }
      fns.remove(fn);
         // fn is worked through
   }
   Z ns=sections.car();
   for (Z i=1;i<=ns;i++){
      X2<string,string> xyi=sections.xy(i);
      string xi=xyi.c1();
      string yi=xyi.c2();
      str_+=el; // adding end of line characters let str_ be scanned line
         // by line
      str_+=xi;
         // section header
      str_+=el;
      str_+=el;
      str_+=yi;
         // section content
   }
   CPM_MZ
   return true;
}

////////////////////// class TextH //////////////////////////////////////

namespace{
   string skipCR(const string& w)
   {
      string::size_type nCR=w.find(0x0D);
      string res=w;
      return nCR==string::npos ? res : res.replace(nCR,1,"");
   }
}

bool TextH::prnOn(ostream& str)const
{
   const string head("  c ");
   if (!str) return false;
   Z n=vs.dim();
   if (n>=1){
      for (Z i=1;i<=n;i++){
         str<<endl<<head<<skipCR(vs[i]);
         if (!str) return false;
      }
      str<<endl;
   }
   return true;
}

bool TextH::scanFrom(istream& str)
{
   if (!str) return false;
   using CpmRoot::readLine;
   string line;
   while (readLine(line,str)){
      *this&=line;
   }
   return true;
}

///////////////////// class Record /////////////////////////////////////

Record::Record(const Word& fn)
{
   Record rc;
   IFileStream in(fn);
   bool bv=rc.readFormatted(in());
   *this=rc;
   valid=bv;
}

bool Record::set(const Word& section, const Word& item,
   const bool& value)
{
   B val(value);
   Key k(section,item);
   rep0.set(k,val);
   return true;
      // no non-mystical error condition can happen
}

namespace{
B f_bool(bool x){ return B(x);}
Z1 f_Z(Z x){ return Z1(x);}
//#if !defined(CPM_ULM)
R1 f_R(R x){ return R1(x);}
//#endif

bool g_bool(B x){ return x.x1;}
Z g_Z(Z1 x){ return x.x1;}
//#if !defined(CPM_ULM)
R g_R(R1 x){ return x.x1;}
//#endif
}

bool Record::set(const Word& section, const Word& item,
   const V<bool>& value)
{
   Key k(section,item);
   vec0.set(k,value(F<bool,B>(f_bool)));
      // see V<>::operator()(const F<T,X>& g)const
   return true;
}

bool Record::set(const Word& section, const Word& item,
   const V<B>& value)
{
   Key k(section,item);
   vec0.set(k,value);
   return true;
}

bool Record::set(const Word& section, const Word& item, const Z& value)
{
   Z1 val(value);
   Key k(section,item);
   rep1.set(k,val);
   return true;
}

bool Record::set(const Word& section, const Word& item,
   const V<Z>& value)
{
   Key k(section,item);
   vec1.set(k,value(F<Z,Z1>(f_Z)));
   return true;
}

bool Record::set(const Word& section, const Word& item,
   const V<Z1>& value)
{
   Key k(section,item);
   vec1.set(k,value);
   return true;
}

bool Record::set(const Word& section, const Word& item, const R& value)
{
   R1 val(value);
   Key k(section,item);
   rep2.set(k,val);
   return true;
}

bool Record::set(const Word& section, const Word& item,
   const V<R>& value)
{
   Key k(section,item);
   vec2.set(k,value(F<R,R1>(f_R)));
   return true;
}

bool Record::set(const Word& section, const Word& item,
   const V<R1>& value)
{
   Key k(section,item);
   vec2.set(k,value);
   return true;
}

bool Record::set(const Word& section, const Word& item,
   const Word& value)
{
   Key k(section,item);
   rep3.set(k,value);
   return true;
}

bool Record::set(const Word& section, const Word& item,
   const V<Word>& value)
{
   Key k(section,item);
   vec3.set(k,value);
   return true;
}

bool Record::set(const Word& section, const Word& item,
   const TextH& value)
{
   Key k(section,item);
   rep4.set(k,value);
   return true;
}

bool Record::set(const Word& section, const TextH& value)
{
   Word sectionAct=section;
   sectionAct.skipSurSpc_();
      // notice that using this function with 'section' read from file
      // after another function has it written to file may result in
      // having the argument cluttered with non-printable characters.
      // So a Word may loose its power as a key
   rep5.set(sectionAct,value);
   return true;
}

bool Record::set(const TextH& value)
{
   rep6=value;
   return true;
}

void Record::setBs(Word const& section, Word const& item,
   V<bool> const& x)
 { vec0[Key(section,item)]=x(F<bool,B>(f_bool));}

void Record::setZs(Word const& section, Word const& item, V<Z> const& x)
{ vec1[Key(section,item)]=x(F<Z,Z1>(f_Z));}

void Record::setRs(Word const& section, Word const& item, V<R> const& x)
{ vec2[Key(section,item)]=x(F<R,R1>(f_R));}

bool Record::get(const Word& section, const Word& item, bool& value,
   Z verb)const
{
   Key k(section,item);
   B val(value);
   bool b=rep0.get(k,val);
   if (b) value=val;
   else if (verb>0){
      cpmwarning(
         "section="&section&", item="&item&", not able to read bool");
   }
   return b;
}

bool Record::get(const Word& section, const Word& item, Z& value,
   Z verb)const
{
   Key k(section,item);
   Z1 val(value);
   bool b=rep1.get(k,val);
   if (b) value=val;
   else if (verb>0){
      cpmwarning(
         "section="&section&", item="&item&", not able to read Z");
   }
   return b;
}

bool Record::get(const Word& section, const Word& item, R& value,
   Z verb )const
{
   Key k(section,item);
   R1 val(value);
   bool b=rep2.get(k,val);
   if (b) value=R(val);
   else if (verb>0){
      cpmwarning(
         "section="&section&", item="&item&", not able to read R");
   }
   return b;
}

bool Record::get(const Word& section, const Word& item,
   Word& value, Z verb)const
{
   Key k(section,item);
   Word val(value);
   bool b=rep3.get(k,val);
   if (b) value=val;
   else if (verb>0){
      cpmwarning(
         "section="&section&", item="&item&", not able to read Word");
   }
   return b;
}

bool Record::get(const Word& section, const Word& item,
   V<bool>& value, Z verb)const
{
   Key k(section,item);
   V<B> val;
   bool b=vec0.get(k,val);
   if (b) value=val(F<B,bool>(g_bool));
   else if (verb>0){
      cpmwarning(
         "section="&section&", item="&item&", not able to read Bs");
   }
   return b;
}

   V<bool> Record::getBs(Word const& section, Word const& item)const
   { return vec0[Key(section,item)](F<B,bool>(g_bool));}

   V<Z> Record::getZs(Word const& section, Word const& item)const
   { return vec1[Key(section,item)](F<Z1,Z>(g_Z));}

   V<R> Record::getRs(Word const& section, Word const& item)const
   { return vec2[Key(section,item)](F<R1,R>(g_R));}

bool Record::get(const Word& section, const Word& item,
   V<Z>& value, Z verb)const
{
   Key k(section,item);
   V<Z1> val;
   bool b=vec1.get(k,val);
   if (b) value=val(F<Z1,Z>(g_Z));
   else if (verb>0){
      cpmwarning(
         "section="&section&", item="&item&", not able to read Zs");
   }
   return b;
}

bool Record::get(const Word& section, const Word& item,
   V<R>& value, Z verb)const
{
   Key k(section,item);
   V<R1> val;
   bool b=vec2.get(k,val);
   if (b) value=val(F<R1,R>(g_R));
   else if (verb>0){
      cpmwarning(
         "section="&section&", item="&item&", not able to read Rs");
   }
   return b;
}

bool Record::get(const Word& section, const Word& item,
   V<Word>& value, Z verb)const
{
   Key k(section,item);
   V<Word> val;
   bool b=vec3.get(k,val);
   if (b) value=val;
   else if (verb>0){
      cpmwarning(
         "section="&section&", item="&item&", not able to read Ws");
   }
   return b;
}

bool Record::get(TextH& comment)const{
   comment=rep6;
   return true;
}

bool Record::get(const Word& section, TextH& comment)const
{
   Word sectionAct=section;
   sectionAct.skipSurSpc_();
      // see comment to
      // bool Record::set(const Word& section, const TextH& value)
   bool b=rep5.get(sectionAct,comment);
   // no warning for !b since comments need'nt to be present
   return b;
}

bool Record::get(const Word& section, const Word& item, TextH& com)const
{
   Key k(section,item);
   bool b=rep4.get(k,com);
   // no warning for !b since comments need'nt to be present
   return b;
}

// I/0 interface

bool Record::prnOn(ostream& str)const
{
   cpmwt("Record");
   cpmp(valid);
   cpmp(rep0);
   cpmp(vec0);
   cpmp(rep1);
   cpmp(vec1);
   cpmp(rep2);
   cpmp(vec2);
   cpmp(rep3);
   cpmp(vec3);
   cpmp(rep4);
   cpmp(rep5);
   cpmp(rep6);
   return true;
}

bool Record::scanFrom(istream& str)
{
   cpms(valid);
   cpms(rep0);
   cpms(vec0);
   cpms(rep1);
   cpms(vec1);
   cpms(rep2);
   cpms(vec2);
   cpms(rep3);
   cpms(vec3);
   cpms(rep4);
   cpms(rep5);
   cpms(rep6);
   return true;
}

namespace{ // local tools

   string skipWhitespace(const string& w)
   {
      Z i,n=(Z)w.size();
      if (n==0) return w;
      V<char> v(n);
      V<bool> val(n);
      for (i=1;i<=n;++i){
         char c=w[i-1];
         v[i]=c;
         val[i]=(isspace(c) ? false : true);
            // isspace(ws)==true for whitespace ws
      }
      V<char> prel=v.select(val);
      n=prel.dim();
      string res(n,'x');
      for (i=1;i<=n;++i) res[i-1]=prel[i];
      return res;
   }

   vector<string> parseLine(const string& str)
      // parsing a line into whitespace-separated words
   {
      vector<string> res;
      istringstream ist(str);
      string st;
      while (ist>>st){
         res.push_back(st);
      }
      return res;
   }

   vector<string> parseStream(istream& in)
      // parsing a text stream into non-blank, non-comment lines
      // The lines do not contain leading whitespace.
      // Since input from file can get seriously disturbed by violation
      // of syntax, this basic input function writes to cpmcerr what it
      // is dealing with
   {
      Z mL=3;
      static Word loc("parseStream(istream&)");
      CPM_MA
      if (cpmverbose>mL){ // intentionally '>'
         cpmcerr<<"these are all data found in the istream"<<endl<<
         "sections are in alphabetic order,"<<endl<<
         "entries within sections in the order of reading"<<endl;
      }
      vector<string> res;
      while(in){
         string st;
         //std::getline(in,st);
         CpmRoot::readLine(st,in);
            // this eliminates carriage returns in addition to newlines
            // added 2004-09-15
         if (cpmverbose>mL) cpmcerr<<st<<endl;
         st=CpmRoot::skipLeadingWhitespace(st);
         Z ds=(Z)st.size();
         if (ds>0){
            char c1=st[0];
            if (c1!='*'&&c1!='/'&&c1!=';'){
                  // lines starting (maybe after whitespace) with
                  // '*' or '/' or ';' are ignored
               replaceFillCharacters_(st);
               res.push_back(st);
            }
         }
      }
      if (cpmverbose>mL)
         cpmcerr<<"end of data found in the istream"<<endl;
      CPM_MZ
      return res;
   }

   bool getText(vector<string> vs, Z& k, TextH& text)
   {
      Z n=(Z)vs.size();
      if (k>=n){
         return false;
      }
      bool res=false;
      while(skipHead(vs[k])){
         text&=vs[k];
         k++;
         res=true;
         if (k>=n) break;
         res=true;
      }
      return res;
   }

   bool getSection(vector<string> vs, Z& k, Word& section)
   {
      Z mL=4;
      Word loc("bool anon::getSection(...)");
      CPM_MA
      if (k>=(Z)vs.size()) return false;
      string line=vs[k];

      bool res=isSectionHeader(line,section);
      if (res){
         k++;
         if (cpmverbose>=mL) cpmcerr<<"section="<<section<<endl;
      }
      CPM_MZ
      return res;
   }

   bool getItem(vector<string> vs, Z& k, const Word& section,
      Word& type, Word& item, V<Word>& val)
   {
      Z mL=3;
      Word loc("anon::getItem(...)");
      CPM_MA
      if (k>=(Z)vs.size()) return false;
      string line=vs[k];
      vector<string> wordsOfLine=parseLine(line);
      Z i,wols=(Z)wordsOfLine.size();
      Word sc;
      if (isSectionHeader(line,sc)){ // then we don't get a
         CPM_MZ
         return false;
      }
      k++;
      type=Word(wordsOfLine[0]);
      item=Word(wordsOfLine[1]);
      Z nList=wols-2; // nList>=0 since wols>=2
      bool insertDefault=false;
      if (nList==0){
         if (type=="F" || type=="include"){
            // for F and include it is natural to have wols==2
            // but no value has to be read
            CPM_MZ
            return false;
         }
         else{
            ostringstream ost;
            ost<<loc<<" value missing at section "<<section<<
               " type"<<type
            <<" item "<<item<<" default value chosen";
            cpmwarning(ost);
            insertDefault=true;
            nList++;
         }
      }
      val=V<Word>(nList); // now nList>=1
      if (insertDefault){
         if (type=="B" || type=="Bs") val[1]="false";
         else if (type=="Z" || type=="Zs" || type=="R" || type=="Rs")
            val[1]=0;
         else if (type=="W" || type=="Ws") val[1]="";
         else cpmerror(loc&": unvalid type");
         CPM_MZ
         return true;
      }
      for (i=1;i<=nList;++i) val[i]=Word(wordsOfLine[i+1]);
         // wordsOfLine[2], wordsOfLine[3], ....
         // transformed into a list of Words
      if (cpmverbose>=mL){
         cpmcerr<<"section="<<section<<", type="<<type<<", item="<<item;
         for (i=1;i<=nList;++i) cpmcerr<<", val["<<i<<"]="<<val[i];
         cpmcerr<<endl;
      }
      CPM_MZ
      return true;
   }

} // namespace

Z Record::com(Record const& s)const
{
   if (rep0<s.rep0) return 1;
   if (rep0>s.rep0) return -1;
   if (vec0<s.vec0) return 1;
   if (vec0>s.vec0) return -1;

   if (rep1<s.rep1) return 1;
   if (rep1>s.rep1) return -1;
   if (vec1<s.vec1) return 1;
   if (vec1>s.vec1) return -1;

   if (rep2<s.rep2) return 1;
   if (rep2>s.rep2) return -1;
   if (vec2<s.vec2) return 1;
   if (vec2>s.vec2) return -1;

   if (rep3<s.rep3) return 1;
   if (rep3>s.rep3) return -1;
   if (vec3<s.vec3) return 1;
   if (vec3>s.vec3) return -1;

   return 0;
}

bool Record::writeFormatted(ostream& str, bool writeComments)const
{
   const Z mL=1;
   Word loc("Record::writeFormatted(ostream&,bool)");
   CPM_MA
   if (!str){
      cpmwarning(loc&": called for unvalid ostream");
      CPM_MZ
      return false;
   }
   if (!isVal()){
      cpmwarning(loc&": denied to write unvalid Record");
      CPM_MZ
      return false;
   }
   S<Word> sections, krep0, kvec0, krep1, kvec1, krep2, kvec2,
      krep3, kvec3;
   Z i,n;

   n=rep0.dim();
   for (i=1;i<=n;i++){
      Key k=rep0.x(i);
      sections.add(k.c1());
      krep0.add(k.c2());
   }

   n=vec0.dim();
   for (i=1;i<=n;i++){
      Key k=vec0.x(i);
      sections.add(k.c1());
      kvec0.add(k.c2());
   }

   n=rep1.dim();
   for (i=1;i<=n;i++){
      Key k=rep1.x(i);
      sections.add(k.c1());
      krep1.add(k.c2());
   }

   n=vec1.dim();
   for (i=1;i<=n;i++){
      Key k=vec1.x(i);
      sections.add(k.c1());
      kvec1.add(k.c2());
   }

   n=rep2.dim();
   for (i=1;i<=n;i++){
      Key k=rep2.x(i);
      sections.add(k.c1());
      krep2.add(k.c2());
   }

   n=vec2.dim();
   for (i=1;i<=n;i++){
      Key k=vec2.x(i);
      sections.add(k.c1());
      kvec2.add(k.c2());
   }

   n=rep3.dim();
   for (i=1;i<=n;i++){
      Key k=rep3.x(i);
      sections.add(k.c1());
      krep3.add(k.c2());
   }

   n=vec3.dim();
   for (i=1;i<=n;i++){
      Key k=vec3.x(i);
      sections.add(k.c1());
      kvec3.add(k.c2());
   }

   n=rep4.dim();
   for (i=1;i<=n;i++){
      Key k=rep4.x(i);
      sections.add(k.c1());
   }

   bool wc=writeComments;

   ostringstream res;
   rep6.prnOn(res);

   Z ns=sections.car();
   cpmmessage(loc&": "&cpm(ns)&" sessions found");
   Z nk,j,i_;
   Word secSep(
      "//--------------------------------\
----------------------------------------");

   for (i=1;i<=ns;++i){
      Word seci=sections[i];
      res<<endl<<seci.str();
      res<<endl<<secSep;
      TextH valsc;
      bool bsc=get(seci,valsc);
         // This is the place where rep5 is used
         // careful version of rep5[seci] that accepts
         // whitespace-cluttered versions of the key
      if (bsc&&wc) valsc.prnOn(res);
      nk=krep0.car();
      for (j=1;j<=nk;j++){
         Word itemj=krep0[j];
         Key k(seci,itemj);
         B value;
         bool b=rep0.get(k,value);
         if (b){
             res<<endl<<"B "<<itemj.str()<<"="<<toStr(value);
             TextH valc;
             bool bc=rep4.get(k,valc);
             if (bc&&wc) valc.prnOn(res);
         }
      }
      nk=kvec0.car();
      for (j=1;j<=nk;j++){
         Word itemj=kvec0[j];
         Key k(seci,itemj);
         V<B> value;
         bool b=vec0.get(k,value);
         if (b){
            n=value.dim();
            res<<endl<<"Bs "<<itemj.str()<<"=";
            for (i_=1;i_<=n;++i_) res<<toStr(value[i_])<<" ";
            TextH valc;
            bool bc=rep4.get(k,valc);
            if (bc&&wc) valc.prnOn(res);
         }
      }
      nk=krep1.car();
      for (j=1;j<=nk;j++){
         Word itemj=krep1[j];
         Key k(seci,itemj);
         Z1 value;
         bool b=rep1.get(k,value);
         if (b){
             res<<endl<<"Z "<<itemj.str()<<"="<<value.x1;
             TextH valc;
             bool bc=rep4.get(k,valc);
             if (bc&&wc) valc.prnOn(res);
         }
      }
      nk=kvec1.car();
      for (j=1;j<=nk;j++){
         Word itemj=kvec1[j];
         Key k(seci,itemj);
         V<Z1> value;
         bool b=vec1.get(k,value);
         if (b){
            n=value.dim();
            res<<endl<<"Zs "<<itemj.str()<<"=";
            for (i_=1;i_<=n;++i_) res<<toStr(value[i_].x1)<<" ";
            TextH valc;
            bool bc=rep4.get(k,valc);
            if (bc&&wc) valc.prnOn(res);
         }
      }
      nk=krep2.car();
      for (j=1;j<=nk;j++){
         Word itemj=krep2[j];
         Key k(seci,itemj);
         R1 value;
         bool b=rep2.get(k,value);
         if (b){
             res<<endl<<"R "<<itemj.str()<<"="<<value.x1;
             TextH valc;
             bool bc=rep4.get(k,valc);
             if (bc&&wc) valc.prnOn(res);
         }
      }
      nk=kvec2.car();
      for (j=1;j<=nk;j++){
         Word itemj=kvec2[j];
         Key k(seci,itemj);
         V<R1> value;
         bool b=vec2.get(k,value);
         if (b){
            n=value.dim();
            res<<endl<<"Rs "<<itemj.str()<<"=";
            for (i_=1;i_<=n;++i_) res<<toStr(value[i_].x1)<<" ";
            TextH valc;
            bool bc=rep4.get(k,valc);
            if (bc&&wc) valc.prnOn(res);
         }
      }
      nk=krep3.car();
      for (j=1;j<=nk;j++){
         Word itemj=krep3[j];
         Key k(seci,itemj);
         Word value;
         bool b=rep3.get(k,value);
         if (b){
             res<<endl<<"W "<<itemj.str()<<"="<<value.str();
             TextH valc;
             bool bc=rep4.get(k,valc);
             if (bc&&wc) valc.prnOn(res);
         }
      }
      nk=kvec3.car();
      for (j=1;j<=nk;j++){
         Word itemj=kvec3[j];
         Key k(seci,itemj);
         V<Word> value;
         bool b=vec3.get(k,value);
         if (b){
            n=value.dim();
            res<<endl<<"Ws "<<itemj.str()<<"=";
            for (i_=1;i_<=n;++i_) res<<-value[i_]<<" ";
            TextH valc;
            bool bc=rep4.get(k,valc);
            if (bc&&wc) valc.prnOn(res);
         }
      }
      res<<endl;
   }
   str<<res.str();
   CPM_MZ
   //return (str!=0);
   return !str ? false : true;
}

bool Record::readFormatted(istream& str)
// When this function is called from Record::readFromFile(), the argument
// is a stringstream that passed a lot of processing already. So it is
// not necessarily of the same line order as the ini-file under
// consideration (reason is the new file include mechanism).
// This is the function which transforms strings into arithmetic types.
{
   const Z mL=2;
   Word loc("Record::readFormatted(istream&)");
   CPM_MA

   if (!str){
      cpmwarning("Record::readFormatted() called for unvalid stream");
      return false;
   }

   rep0=m0();
   vec0=v0();

   rep1=m1();
   vec1=v1();

   rep2=m2();
   vec2=v2();

   rep3=m3();
   vec3=v3();

   rep4=m4();
  // rep5=m5();
   rep6=m6();

   vector<string> pf=parseStream(str);
      // list of all non-void, non-comment lines, not starting with
      // white-space. All intermediate ";" and "=" are eliminated.
      // Till 2011-07-17 also "," was eliminated which turned out not to
      // be adequate for data input of type Ws.
      // At the end of the function str is no longer valid, so return
      // (str!=0) is not good!
      // This is the function that writes to cpmcerr.
   Z pfs=(Z)pf.size();
   if (pfs==0){
      cpmwarning(loc&": false returned due to pfs==0");
      CPM_MZ
      return false;
   }
   Z n,i,k=0;
// reading the explanation for the whole Record
   getText(pf,k,rep6);
   Word section;
   while (getSection(pf,k,section)){
      TextH sectionText;
      while(getText(pf,k,sectionText)){
         set(section,sectionText);
      }
      Word item,type;
      V<Word> val;
      while(getItem(pf,k,section,type,item,val)){
         Key key(section,item);
         if (type=="B"){
            B b(val[1].toBool());
            rep0.set(key,b);
         }
         else if (type=="Bs"){
            n=val.dim();
            V<B> vb(n);
            for (i=1;i<=n;++i) vb[i]=B(val[i].toBool());
            vec0.set(key,vb);
         }
         else if (type=="Z"){
            Z1 z1(val[1].toZ());
            rep1.set(key,z1);
         }
         else if (type=="Zs"){
            n=val.dim();
            V<Z1> vz1(n);
            for (i=1;i<=n;++i) vz1[i]=Z1(val[i].toZ());
            vec1.set(key,vz1);
         }
         else if (type=="R"){
            n=val.dim();
            if (n>2) cpmdebug(val);
            cpmassert(n<=2,"Record::readFormatted(istream) n = "&cpm(n));
            if (n==1){
               R1 r1(val[1].toR()); // aha! Here a Word becomes converted
               // to
               // a R by Word::toR()
               rep2.set(key,r1);
            }
            else{ // added 2016-05-26. Now for type=R we may input also
               // two numbers r1 r2 to the effect that the true value is
               // r1/r2. This allows to input 2/3 as 2 3 instead of
               // 0.66666666666667
               R r1=val[1].toR();
               R r2=val[2].toR();
               rep2.set(key,R1(r1/r2));
            }
         }
         else if (type=="Rs"){
            n=val.dim();
            V<R1> vr1(n);
            for (i=1;i<=n;++i) vr1[i]=R1(val[i].toR());
            vec2.set(key,vr1);
         }
         else if (type=="W"){
            rep3.set(key,val[1]);
         }
         else if (type=="Ws"){
            vec3.set(key,val);
         }
         else {
            cpmerror("Record::readFormatted: undefined data type "&type);
         }
         TextH itemText;
         while(getText(pf,k,itemText)){
            rep4.set(key,itemText);
         }
      }
   }
   CPM_MZ
   return true;
}

bool Record::readFromFile(const Word& fn, bool binary)
{
   Z mL=2;
   Word loc("Record::readFromFile(const Word& fn, bool) fn="&fn);
   CPM_MA
   SecCol indat(fn,binary);
   if (!indat.isVal()){
      cpmwarning(loc&": indat not valid");
      CPM_MZ
      return false;
   }
   Z n=indat.nFiles();
   string data=indat.str();
   istringstream ist(data);
   if (cpmverbose>mL) cpmdata<<"Data collected from "<<n<<
      " ini-files:"<<data<<endl;
   // note that the constructor of IniFileBasedApplication creates a nice
   // record of the read data on file cpmdata.txt, so that the present
   // output may be useful for critical diagnostics only
   bool res=readFormatted(ist);
   valid=B(res); // added 2006-02-07
   if (!valid) cpmwarning(loc&": return value is false");
   CPM_MZ
   return res;
}

namespace{
   bool selFunc(Key const& k, Word const& sec){ return k.c1()==sec;}
}

Record Record::operator()(Word const& section)const
{
   F<Key,bool> sel=F1<Key,Word,bool>(section)(selFunc);
   Record res;
   res.rep0=rep0.select(sel);
   res.vec0=vec0.select(sel);
   res.rep1=rep1.select(sel);
   res.vec1=vec1.select(sel);
   res.rep2=rep2.select(sel);
   res.vec2=vec2.select(sel);
   res.rep3=rep3.select(sel);
   res.vec3=vec3.select(sel);
   Word comment1(
      "This single-section record was made by extracting section");
   comment1&=section;
   Word comment2("from a record with the following overall comment:");
   TextH overall;
   get(overall);
   TextH comm(comment1);
   comm&=comment2;
   comm&=overall;
   res.set(comm);
   res.valid=B(true);
   return res;
}

Record Record::operator()(Word const& sec1, Word const& sec2)const
{
   F<Key,bool> sel=F1<Key,Word,bool>(sec1)(selFunc);
   Z i;
   Record res;
   m0 r0s=rep0.select(sel);
   for (i=1;i<=r0s.dim();++i){
      Word item=r0s.x(i).c2();
      B value=r0s.y(i);
      res.set(sec2,item,value);
   }

   v0 v0s=vec0.select(sel);
   for (i=1;i<=v0s.dim();++i){
      Word item=v0s.x(i).c2();
      V<B> value=v0s.y(i);
      res.set(sec2,item,value);
   }

   m1 r1s=rep1.select(sel);
   for (i=1;i<=r1s.dim();++i){
      Word item=r1s.x(i).c2();
      Z1 value=r1s.y(i);
      res.set(sec2,item,value);
   }

   v1 v1s=vec1.select(sel);
   for (i=1;i<=v1s.dim();++i){
      Word item=v1s.x(i).c2();
      V<Z1> value=v1s.y(i);
      res.set(sec2,item,value);
   }

   m2 r2s=rep2.select(sel);
   for (i=1;i<=r2s.dim();++i){
      Word item=r2s.x(i).c2();
      R1 value=r2s.y(i);
      res.set(sec2,item,value);
   }

   v2 v2s=vec2.select(sel);
   for (i=1;i<=v2s.dim();++i){
      Word item=v2s.x(i).c2();
      V<R1> value=v2s.y(i);
      res.set(sec2,item,value);
   }

   m3 r3s=rep3.select(sel);
   for (i=1;i<=r3s.dim();++i){
      Word item=r3s.x(i).c2();
      Word value=r3s.y(i);
      res.set(sec2,item,value);
   }

   v3 v3s=vec3.select(sel);
   for (i=1;i<=v3s.dim();++i){
      Word item=v3s.x(i).c2();
      V<Word> value=v3s.y(i);
      res.set(sec2,item,value);
   }

   Word comment1(
      "This single-section record was made by extracting section");
   comment1&=sec1;
   Word comment2("from a record with the following overall comment:");
   TextH overall;
   get(overall);
   TextH comm(comment1);
   comm&=comment2;
   comm&=overall;
   res.set(comm);
   res.valid=B(true);
   return res;
}

///////////////// class Recordable ///////////////////////////////////

bool Recordable::adjust()const
{
   if (valid_) return true;
   else{
      valid_=true;
      ostringstream ost;
      rc_.writeFormatted(ost,false);
      rep_=Word(ost.str());
      return vadjust();
   }
}

bool Recordable::prnOn(ostream& str)const
{
   cpmwt(-nameOf());
   cpmp(rc_);
   return true;
}

bool Recordable::scanFrom(istream& str)
{
   cpms(rc_);
   renew_();
   return adjust();
}

Z Recordable::com(Recordable const& r)const
{
   return rep().com(r.rep());
}

/////////////////////// class Text /////////////////////////////////////
Text Text::getLine(istream& str)
{
   string line;
   CpmRoot::readLine(line ,str);
      // reads over comments an eliminates CRs (not newlines)
   F<string,Word> trans;
   vector<string> resPrel=parseLine(line);
   Z n=(Z)resPrel.size();
   V<Word> res(n);
   for (Z i=1;i<=n;++i) res[i]=Word(resPrel[i-1]);
   return Text(res);
}

