//? cpmrmatrix.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmrmatrix.h>
#include <cpmc.h>
#include <cpmrfunc.h>
#include <cpmfile.h>
#include <cpmsystemdependencies.h>

using namespace CpmRoot;
using namespace CpmSystem;

using CpmArrays::R_Matrix;
using CpmArrays::R_Vector;
using CpmArrays::SVD;
using CpmArrays::V;
using CpmArrays::VV;
using CpmArrays::X2;
using CpmFunctions::F;
using CpmFunctions::F2;
using CpmFunctions::F3;
using CpmGeo::Iv;
using CpmLinAlg::R2;

R R_Matrix::etaInv=1e-12;

R_Matrix::R_Matrix(const VV<R>& vv):rep(vv.dim())
{
   Z m=rep.dim();
   for (Z i=1;i<=m;i++) rep[i]=R_Vector(vv[i]);
}

R_Matrix::R_Matrix(R_Vector const& a, R_Vector const& b):
rep(a.dim(),R_Vector(b.dim()))
{
   for (Z i=a.b();i<=a.e();++i){
      for (Z j=b.b();j<=b.e();++j){
         rep[i][j]=a[i]*b[j];
      }
   }
}

R_Matrix::R_Matrix(R_Vector const& col, Z n):
rep(col.dim(),R_Vector(n))
{
   for (Z i=1;i<=col.dim();++i){
      for (Z j=1;j<=n;++j){
         rep[i][j]=col[i];
      }
   }
}

R_Matrix R_Matrix::permuteColumns(const V<Z>& per)
{
   Z n=dim();
   R_Matrix res(n);
   for (Z i=1;i<=n;i++) res[i]=(*this)[i].permute(per);
   return res;
}

R_Matrix R_Matrix::con(void)const
{
   Z m=dim1();
   Z n=dim2();
   R_Matrix res(n,m); // sic, (m,n) are the data for *this
   for (Z i=1;i<=m;i++) for (Z j=1;j<=n;j++) res[j][i]=rep[i][j];
   return res;
}

R_Vector R_Matrix::operator *(const R_Vector& x)const
// always defined
{
   Z m=dim1();
   Z n=dim2();
   Z nx=x.dim();
   R_Vector res(m);
   if (n==nx){
      for (Z i=1;i<=m;i++){
         res[i]=(rep[i]|x); // fast normal case
      }
   }
   else{ // slow irregular case
      Z nMin=(nx<=n ? nx : n);
      for (Z i=1;i<=m;i++){
         R sum=0;
         for (Z j=1;j<=nMin;j++) sum+=rep[i][j]*x[j];
         res[i]=sum;
      }
   }
   return res;
}

# if defined(CPM_USE_EIGEN)

R_Matrix R_Matrix::operator *(const R_Matrix& a2)const
// Notice: the product of arbitrarily framed matrices is well defined.
// If the ranges fit the the requirement for regular matrix multiplication
// the speed is 18 times larger.
{
   using namespace Eigen;

   Z m1=dim1();
   Z n1=dim2();
   Z m2=a2.dim1();
   Z n2=a2.dim2();
   cpmassert(n1==m2,"R_Matrix * R_Matrix");
   Z i,j,k,l;
   R_Matrix res(m1,n2);
   Matrix<R,Dynamic,Dynamic> x1(m1,n1);
   for (i=1,k=0; i<=m1; i++,k++) for (j=1,l=0;j<=n1; j++,l++)
      x1(k,l) = rep[i][j];

   Matrix<R,Dynamic,Dynamic> x2(m2,n2);
   for (i=1,k=0; i<=m2; i++,k++) for (j=1,l=0;j<=n2; j++,l++)
      x2(k,l)=a2[i][j];

   Matrix<R,Dynamic,Dynamic> x3 = x1*x2;
   for (i=1,k=0; i<=m1; i++,k++) for (j=1,l=0;j<=n2; j++,l++)
   res[i][j] = x3(k,l);
   return res;
}

# else

R_Matrix R_Matrix::operator *(const R_Matrix& a2)const
// Notice: the product of arbitrarily framed matrices is well defined.
// If the ranges fit the the requirement for regular matrix multiplication
// the speed is 18 times larger.
{
   Z m1=dim1();
   Z n1=dim2();
   Z m2=a2.dim1();
   Z n2=a2.dim2();
   R_Matrix res(m1,n2);
   if (n1==m2){ // regular case
      R_Matrix b=~a2; // O(n^2) negligible compared to the  O(n^3) main
         // task
      for (Z i=1;i<=m1;i++){ // O(n^3)
         for (Z j=1;j<=n2;j++){
            res[i][j]=(rep[i]|b.rep[j]); // making the inner loop fast
         }
      }
   }
   else{ // irregular case
      Z nk=(n1<=m2 ? n1 : m2);
      for (Z i=1;i<=m1;i++){
         for (Z j=1;j<=n2;j++){
            R sum=0;
            for (Z k=1;k<=nk;k++){
               sum+=rep[i][k]*a2[k][j]; // slow inner loop
            }
            res[i][j]=sum;
         }
      }
   }
   return res;
}
# endif

R R_Matrix::supAbs()const
{
   R res=0.;
   for (Z i=1;i<=dim();++i){
      R xi=rep[i].supAbs();
      if (xi>res) res=xi;
   }
   return res;
}

R_Matrix R_Matrix::append2(const R_Matrix& x1)const
{
   Z m=CpmRootX::inf<Z>(dim1(),x1.dim1());
   Z n1=dim2();
   Z n2=x1.dim2();
   Z i,j,k,n=n1+n2;
   R_Matrix res(m,n);
   for (i=1;i<=m;++i){
      for (j=1;j<=n1;++j){
         res[i][j]=rep[i][j];
      }
      for (j=1,k=n1+1; j<=n2; ++j,++k){
         res[i][k]=x1[i][j];
      }
   }
   return res;
}

R_Vector R_Matrix::getCol(Z j)const
{
   Word loc("R_Matrix::getCol(Z)");
   Z m=dim1(),n=dim2();
   cpmassert(j>=1 && j<=n,loc);
   R_Vector res(m);
   for (Z i=1;i<=m;i++) res[i]=rep[i][j];
   return res;
}

void R_Matrix::setCol(Z j, const R_Vector& v)
{
   Word loc("R_Matrix::setCol(Z,R_Vector)");
   Z m=dim1(),n=dim2();
   cpmassert(j>=1 && j<=n,loc);
   Z mm=CpmRootX::inf<Z>(m,v.dim());
   for (Z i=1;i<=mm;i++) rep[i][j]=v[i];
}

R_Matrix& R_Matrix::operator +=(const R_Matrix& a)
{
   Z n=dim();
   Z na=a.dim();
   Z nnaMin=(n<=na ? n : na);
   for (Z i=1;i<=nnaMin;i++) rep[i]+=(a.rep)[i];
   return *this;
}

R_Matrix& R_Matrix::operator -=(const R_Matrix& a)
{
   Z n=dim();
   Z na=a.dim();
   Z nnaMin=(n<=na ? n : na);
   for (Z i=1;i<=nnaMin;i++) rep[i]-=(a.rep)[i];
   return *this;
}

R_Matrix R_Matrix::operator -(const R_Matrix& a)const
 {  R_Matrix res=*this; return res-=a;}

R2 R_Matrix::centerOfMass()const
{
   Z mL=3;
   static Word loc("R_Matrix::centerOfMass()");
   cpmmessage(mL,loc&" started");
   Z m=dim();
   Z n=dim2();
   R si=0;
   R sj=0;
   R s=0;
   Z i,j;
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         R val=(*this)[i][j];
         si+=R(i)*val;
         sj+=R(j)*val;
         s+=val;
      }
   }
   cpmassert(s!=0,loc);
   R sInv=cpminv(s);
   cpmmessage(mL,loc&" done");
   return R2(si*sInv,sj*sInv);
}

R_Vector R_Matrix::linForm()const
// In principle rep[1]&rep[2]&...&rep[n], n=rep.dim()
// But due to the multiple allocations this is much too slow.
{
   Z m=dim1();
   if (m<1) return R_Vector(0);
   Z n=dim2();
   if (n<1) return R_Vector(0);
   Z mn=m*n;
   R_Vector res(mn);
   Z k=0;
   for (Z i=1;i<=m;++i){
      for (Z j=1;j<=n;++j){
         res[++k]=rep[i][j];
      }
   }
   return res;
}

R_Vector R_Matrix::dig()const
{
   Z dd=CpmRootX::inf<Z>(dim1(),dim2());
   R_Vector res(dd);
   for (Z i=1;i<=dd;++i) res[i]=rep[i][i];
   return res;
}

namespace{
   R_Vector ffAux(R_Vector const& x, F<R,R> const& f){ return x(f);}
   //R_Vector logAux(R_Vector const& x){ return x.log10();}
}

R_Matrix R_Matrix::operator()(F<R,R> const& g)const
{
   return R_Matrix(rep(F1<R_Vector,F<R,R>,R_Vector>(g)(ffAux)));
}

R_Matrix R_Matrix::log10()const
{
   return this->operator()(CpmFunctions::R_Func::log10);
}

VV<R> R_Matrix::toVV()const
{
   Z m=dim1();
   Z n=dim2();
   Z i,j;
   VV<R> res(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         res[i][j]=rep[i][j];
      }
   }
   return res;
}

R toFfunc(R2 const& xy, R_Matrix const& mat, Iv const& ivx, Iv const& ivy)
{
   Z m=mat.dim1(),n=mat.dim2();
   Z i=ivx.qun(xy.x1,m,true);
   Z j=ivy.qun(xy.x2,n,true);
   return mat[i][j];
}

F<R2,R> R_Matrix::toF(Iv const& iR,Iv const& jR)const
{
   return F3<R2,R_Matrix,Iv,Iv,R>(*this,iR,jR)(toFfunc);
}

bool R_Matrix::wrtFor(ostream& str)const
{
   const Word sep(" "); // as a Word, sep ends with a \n
   Z m=dim1();
   Z n=dim2();
   Z i,j;
   str<<m<<sep<<n<<endl;
   for (i=1;i<=m;++i){
      for (j=1;j<=n;++j){
         str<<rep[i][j]<<sep;
      }
      str<<endl;
   }
   return (str) ? true : false;
}

bool R_Matrix::readFor_(istream& str)
{
   Z m=-1,n=-1;
   str>>m;
   if (m<0 || !str) return false;
   str>>n;
   if (n<0 || !str) return false;
   rep=V<R_Vector>(m,R_Vector(n));
   Z i,j;
   for (i=1;i<=m;++i){
      for (j=1;j<=n;++j){
         str>>rep[i][j];
      }
   }
   return (str) ? true : false;
}

bool R_Matrix::wrtForTrn(ostream& str)const
{
   const Word sep(" ");
   Z m=dim1();
   Z n=dim2();
   Z i,j;
   str<<m<<sep<<n<<endl;
   for (i=1;i<=n;++i){
      for (j=1;j<=m;++j){
         str<<rep[j][i]<<sep;
      }
      str<<endl;
   }
   return (str) ? true : false;
}

bool R_Matrix::toGnuPlot(Word const& fileName)const
{
   Z m=dim1(), n=dim2();
   CpmSystem::OFile ofi(fileName);
   for (Z i=1;i<=n;++i){
      for (Z j=1;j<=m;++j){
         ofi()<<rep[j][i]<<" ";
      }
      ofi()<<endl;
   }
   bool res=ofi.isVal();
   return res;
}

bool R_Matrix::toGnuPlot(ostream& str)const
{
   Z m=dim1(), n=dim2();
   for (Z i=1;i<=n;++i){
      for (Z j=1;j<=m;++j){
         str<<rep[j][i]<<" ";
      }
      str<<endl;
   }
   return (str) ? true : false;
}

bool R_Matrix::readForTrn_(istream& str)
{
   Z m=-1,n=-1;
   str>>m;
   if (m<0 || !str) return false;
   str>>n;
   if (n<0 || !str) return false;
   rep=V<R_Vector>(m,R_Vector(n));
   Z i,j;
   for (i=1;i<=n;++i){
      for (j=1;j<=m;++j){
         str>>rep[j][i];
      }
   }
   return (str) ? true : false;
}

namespace{

   R fAux(const R& x, const Iv& iv1, const Iv& iv2)
   {
      return iv1.linear(x,iv2.inf(),iv2.sup());
   }

}

X2<R_Matrix,Iv> R_Matrix::adjustRange(const Iv& ivRange)const
{
   Iv ivVal=infSup();
   F<R,R> g=F2<R,Iv,Iv,R>(ivVal,ivRange)(fAux);
   return X2<R_Matrix,Iv>(operator()(g),ivVal);
}

// implementation 2016 using the Eigen library. Here the functionality is
// also available for complex matrices and probably also for user-defined
// scaler types
# if defined(CPM_USE_EIGEN)
void R_Matrix::svdcmp(R_Matrix& a, R_Vector& w, R_Matrix& v)
{
   using namespace Eigen;
   Z mL=3;
   Word loc("svdcmp(R_Matrix,R_Vector,R_Matrix)");
   CPM_MA

   Z m=a.dim1(), n=a.dim2();

   if (!(m>=n))
      cpmerror("svdcmp(): bad frame of a");
   if (n!=w.dim())
      cpmerror("svdcmp(): bad frame of w");
   if ((n!=v.dim1()) || (n!=v.dim2()))
      cpmerror("svdcmp(): bad frame of v");
   Z i,j,k,l;
   Z d=CpmRootX::inf(m,n);
   Matrix<R,Dynamic,Dynamic> ea(m,n);
   for (i=1,k=0; i<=m; i++,k++) for (j=1,l=0;j<=n; j++,l++)
      ea(k,l)=a[i][j];

   JacobiSVD<Matrix<R,Dynamic,Dynamic> > svd(ea,
      ComputeFullU | ComputeFullV);
   Matrix<R,Dynamic,1> ew = svd.singularValues();
   Matrix<R,Dynamic,Dynamic> eu=svd.matrixU();
   Matrix<R,Dynamic,Dynamic> ev=svd.matrixV();
   cpmassert(d==ew.size(),loc);

   for (i=1,k=0; i<=d; i++,k++)
   w[i] = ew(k);

   Z u1 = eu.rows();
   Z u2 = eu.cols();
   for (i=1,k=0; i<=u1; i++,k++) for (j=1,l=0;j<=u2; j++,l++)
   a[i][j] = eu(k,l);

   Z v1 = ev.rows();
   Z v2 = ev.cols();
   for (i=1,k=0; i<=v1; i++,k++) for (j=1,l=0;j<=v2; j++,l++)
   v[i][j] = ev(k,l);

   CPM_MZ
   return;
}
#else

////////////// singular value decomp, modified from Press et al. //////////
// written such that no conversion to NR data structures is needed
// modified from Press et al.
// an more recently introduced error was corrected 2000-10-31.
// (and one more in 2003-07-14)

namespace{

R sign2(R r1, R r2)
{
   return ( r2 >= 0. ? cpmabs(r1) : -cpmabs(r1) );
}

R pythag(R r1, R r2)
{
   using CpmRoot::C;
   return C(r1,r2).abs();
}
} // anon

void R_Matrix::svdcmp(R_Matrix& a, R_Vector& w, R_Matrix& v)
{
   const Z iterMax=40;

 //  using CpmRoot::abs;
   using CpmRootX::sup;
   using CpmRootX::inf;
   Z mL=3;
   Word loc("svdcmp(R_Matrix,R_Vector,R_Matrix)");
   CPM_MA

   Z m=a.dim1(), n=a.dim2();

   if (!(m>=n))
      cpmerror("svdcmp(): bad frame of a");
   if (n!=w.dim())
      cpmerror("svdcmp(): bad frame of w");
   if ((n!=v.dim1()) || (n!=v.dim2()))
      cpmerror("svdcmp(): bad frame of v");

   Z flag,i,its,j,jj,k;
   Z l=0,nm=0;
   R anorm,c,f,g,h,s,scale,x,y,z;

// almost verbally Press et al. with some names replaced
   R_Vector rv1(n);
   g=scale=anorm=0.0;
   for (i=1;i<=n;i++) {
      l=i+1;
      rv1[i]=scale*g;
      g=s=scale=0.0;
      if (i <= m) {
         for (k=i;k<=m;k++) scale += cpmabs(a[k][i]);
         if (scale!=0.) {
            for (k=i;k<=m;k++) {
               a[k][i] /= scale;
               s += a[k][i]*a[k][i];
            }
            f=a[i][i];
            g = -sign2(cpmsqrt(s),f);
            h=f*g-s;
            a[i][i]=f-g;
            for (j=l;j<=n;j++) {
               for (s=0.0,k=i;k<=m;k++) s += a[k][i]*a[k][j];
               f=s/h;
               for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
            }
            for (k=i;k<=m;k++) a[k][i] *= scale;
         }
      }
      w[i]=scale *g;
      g=s=scale=0.0;
    //  if (i <= m && i+1 != n) { // i != n before, reported as bug
      // on the official NR bug report, corrected 2007-08-03, and
      // withdrawn 2008-01-20, since only the old form works properly
      if (i <= m && i != n) { //old and good form
         for (k=l;k<=n;k++) scale += cpmabs(a[i][k]);
         if (scale!=0.) {
            for (k=l;k<=n;k++) {
               a[i][k] /= scale;
               s += a[i][k]*a[i][k];
            }
            f=a[i][l];
            g = -sign2(cpmsqrt(s),f);
            h=f*g-s;
            a[i][l]=f-g;
            for (k=l;k<=n;k++) rv1[k]=a[i][k]/h;
            for (j=l;j<=m;j++) {
               for (s=0.0,k=l;k<=n;k++) s += a[j][k]*a[i][k];
               for (k=l;k<=n;k++) a[j][k] += s*rv1[k];
            }
            for (k=l;k<=n;k++) a[i][k] *= scale;
         }
      }
      anorm=sup(anorm,(cpmabs(w[i])+cpmabs(rv1[i])));
   }
   for (i=n;i>=1;i--) {
      if (i < n) {
         if (g!=0.) {
            for (j=l;j<=n;j++) v[j][i]=(a[i][j]/a[i][l])/g;
            for (j=l;j<=n;j++) {
               for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
               for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
            }
         }
         for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
      }
      v[i][i]=1.0;
      g=rv1[i];
      l=i;
   }
   for (i=inf(m,n);i>=1;i--) {
      l=i+1;
      g=w[i];
      for (j=l;j<=n;j++) a[i][j]=0.0;
      if (g!=0.) {
         g=R(1)/g;
         for (j=l;j<=n;j++) {
            for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
            f=(s/a[i][i])*g;
            for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
         }
         for (j=i;j<=m;j++) a[j][i] *= g;
      } else for (j=i;j<=m;j++) a[j][i]=0.0;
      ++a[i][i];
   }
   for (k=n;k>=1;k--) {
      for (its=1;its<=iterMax;its++) {
         flag=1;
         for (l=k;l>=1;l--) {
            nm=l-1;
            if ((cpmabs(rv1[l])+anorm) == anorm) {
                  // if one casts to float on both sides of ==, one gets
                  // indexes out of range!. Very subtle point that fooled
                  // me once. Erronous version detected and corrected
                  // 2000-10-30.
                  // Not clear when the error was introduced. Long ago,
                  // everything was OK since tested several times
               flag=0;
               break;
            }
            if ((cpmabs(w[nm])+anorm) == anorm) break;
               // see comment above
         }
         if (flag) {
            c=0.0;
            s=1.0;
            for (i=l;i<=k;i++) {
               f=s*rv1[i];
               rv1[i]=c*rv1[i];
               if ((cpmabs(f)+anorm) == anorm) break;
                   // see comment above
               g=w[i];
               h=pythag(f,g);
               w[i]=h;
               h=R(1)/h;
               c=g*h;
               s = -f*h;
               for (j=1;j<=m;j++) {
                  y=a[j][nm];
                  z=a[j][i];
                  a[j][nm]=y*c+z*s;
                  a[j][i]=z*c-y*s;
               }
            }
         }
         z=w[k];
         if (l == k) {
            if (z < 0.0) {
               w[k] = -z;
               for (j=1;j<=n;j++) v[j][k] = -v[j][k];
            }
            break;
         }
         if (its == iterMax)
            cpmerror("svdcmp: no convergence in "&cpmwrite(iterMax)&
            " iterations");
         x=w[l];
         nm=k-1;
         y=w[nm];
         g=rv1[nm];
         h=rv1[k];
         f=((y-z)*(y+z)+(g-h)*(g+h))/(h*y*2);
         g=pythag(f,1.0);
         f=((x-z)*(x+z)+h*((y/(f+sign2(g,f)))-h))/x;
         c=s=1.0;
         for (j=l;j<=nm;j++) {
            i=j+1;
            g=rv1[i];
            y=w[i];
            h=s*g;
            g=c*g;
            z=pythag(f,h);
            rv1[j]=z;
            c=f/z;
            s=h/z;
            f=x*c+g*s;
            g = g*c-x*s;
            h=y*s;
            y *= c;
            for (jj=1;jj<=n;jj++) {
               x=v[jj][j];
               z=v[jj][i];
               v[jj][j]=x*c+z*s;
               v[jj][i]=z*c-x*s;
            }
            z=pythag(f,h);
            w[j]=z;
            if (z!=0.) {
               z=R(1)/z;
               c=f*z;
               s=h*z;
            }
            f=c*g+s*y;
            x=c*y-s*g;
            for (jj=1;jj<=m;jj++) {
               y=a[jj][j];
               z=a[jj][i];
               a[jj][j]=y*c+z*s;
               a[jj][i]=z*c-y*s;
            }
         }
         rv1[l]=0.0;
         rv1[k]=f;
         w[k]=x;
      }
   }
   CPM_MZ
   return;
}
#endif

//////////// end of singular value decomposition algorithm /////////////

R_Matrix R_Matrix::inv()const
{
   const Z testLimit=5;
   R sum;
   Z i, j, k, m_=dim1(), n=dim2();
   bool rightframe=(m_>=n);
   R_Matrix a=*this;
   if (!rightframe) a=~a;
   m_=a.dim1(); n=a.dim2();
   R_Matrix v(n,n);
   R_Vector w(n);
   R_Matrix vr(n,m_); //  sic
   svdcmp(a,w,v);
   if (cpmverbose>=testLimit){ // test of svdcmp,
      // svdcmp got corrupted once, so good to test
      R_Matrix u=v.net(1);
      R error1=(~a*a-u).abs();
      cpmmessage("R_Matrix::inv(): error1= "&cpm(error1));
      R error2=(~v*v-u).abs();
      cpmmessage("R_Matrix::inv(): error2= "&cpm(error2));
      R error3=(v*~v-u).abs();
      cpmmessage("R_Matrix::inv(): error3= "&cpm(error3));
      R error4=(*this-a*R_Matrix(w)*~v).abs();
      cpmmessage("R_Matrix::inv(): error4= "&cpm(error4));
      R err=error1+error2+error3+error4;
      cpmmessage(" total inaccuracy in svdcmp = "&cpmwrite(err));
   }
   w=w.pseudoInverse(etaInv);  // w does not come out orderd
      // now w is the pseudo-inverse of the original w
   for (i=1;i<=n;i++){
      for (j=1;j<=m_;j++){
         sum=0;
         for (k=1;k<=n;k++){
            sum+=v[i][k]*w[k]*a[j][k];
         }
         vr[i][j]=sum;
      }
   }
   if (!rightframe) vr=~vr;
   return vr;
}

namespace{
// without the order-related operations the quality of the SVD is much
// considerably less accurate.
R_Vector svd(R_Matrix& u, R_Matrix& v)
{
   R_Matrix u_=u;
   Z m=u_.dim1(), n=u_.dim2(), i;
   if (m>=n){
      R_Matrix vv(n,n);
      v=vv;
      R_Vector w(n);
      R_Matrix::svdcmp(u_,w,v);
      R_Vector w_(n);
      for (i=1;i<=n;i++) w_[i]=-cpmabs(w[i]);
      V<Z> p=w_.permutationForIncreasingOrder();
      u_=u_.permuteColumns(p);
      v=v.permuteColumns(p);
      w=w.permute(p);
      u=u_;
      { // making the dominant component of ~v positive
         R_Matrix vt=~v;
         R_Matrix ut=~u;
         Z nv=vt.dim1(),nu=ut.dim1();
         if (nv!=nu) cpmerror("svd: malfunction in +/- part");
         for (Z i=1;i<=nv;i++){
            Iv ivt=vt[i].infSup();
            R xmin=ivt.inf(), xmax=ivt.sup();
            if (cpmabs(xmin)>cpmabs(xmax)){
               vt[i]=(-1.)*(vt[i]);  // makes dominant part positiv
               ut[i]=(-1.)*(ut[i]); // necessary to stay with a svd
            }
         }
         u=~ut; v=~vt;
      }
      return w;
   }
   else{
      n=u_.dim1();   // now m>=n  and u is a (n,m)-matrix
      R_Matrix vv(n,n);
      v=vv;
      R_Vector w(n);
      u_=~u_; //  now u_ is a (m,n)-matrix
      R_Matrix::svdcmp(u_,w,v);
      R_Vector w_(n);
      for (i=1;i<=n;i++) w_[i]=-cpmabs(w[i]);
      V<Z> p=w_.permutationForIncreasingOrder();
      u_=u_.permuteColumns(p);
      v=v.permuteColumns(p);
      w=w.permute(p);
      R_Matrix umem=u_; u_=v; v=umem;
      u=u_;
      { // making the dominant component of ~v positive
         R_Matrix vt=~v;
         R_Matrix ut=~u;
         Z nv=vt.dim1(), nu=ut.dim1();
         if (nv!=nu) cpmerror("svd: malfunction in +/- part");
         for (Z i=1;i<=nv;i++){
            Iv ivt=vt[i].infSup();
            R xmin=ivt.inf(),xmax=ivt.sup();
            if (cpmabs(xmin)>cpmabs(xmax)){
               vt[i]=(-1.)*(vt[i]);  // makes dominant part positiv
               ut[i]=(-1.)*(ut[i]); // necessary to stay with a svd
            }
         }
         u=~ut; v=~vt;
      }
      return w;
   }
}
}

R_Matrix R_Matrix::net(Z i)const
{
   Z m=dim1();
   Z n=dim2();
   R_Matrix res(m,n); // all zero by definition
   if (i!=1) return res;
// now i==1, thus unit matrix required
   if (m!=n) cpmwarning(
      "R_Matrix::net(Z): non-square unit matrix formed");
   Z mnMin=(m<=n ? m : n);
   for (Z j=1; j<=mnMin; j++) res[j][j]=R(1.);
   return res;
}

bool R_Matrix::scanFromFreeFor(istream& ifs)
{
   Word loc("R_Matrix::scanFromFreeFor(istream& ifs)");
   Z mL=2;
   CPM_MA
   CpmRootX::R1 x;
   Z m=dim();
   if (m==0){
      cpmwarning("trivial exit m==0");
      CPM_MZ
      return false;
   }
   R_Vector y(m);
   std::vector<R_Vector> vec;
   Z i,j;
   bool work=true;
   while (work){
      for (i=1;i<=m;++i){
         bool succ=x.scanFrom(ifs);
         if (succ)y[i]=R(x);
         else{
            work=false;
            cpmmessage(loc&": last valid x was "&cpm(x));
            break;
         }
      }
      if (work) vec.push_back(y);
   }
   Z n=(Z)vec.size();
   if (n==0){
      cpmwarning("trivial exit n==0");
      CPM_MZ
      return false;
   }
   R_Matrix res(m,n);
   for (i=1;i<=m;++i) for (j=1;j<=n;++j) res[i][j]=vec[j-1][i];
   *this=res;
   CPM_MZ
   return true;
}

bool R_Matrix::scanLineWise(istream& ifs)
{
   Word loc("R_Matrix::scanLineWise(istream& ifs)");
   Z mL=2;
   CPM_MA
   bool work=true;
   vector<vector<R> > vvr;
   bool first=true;
   Z dimFirst=0;
   while(work){
      string line;
      work = readLine(line,ifs);
      if (work){
         istringstream ist(line);
         vector<R> vr;
         R r;
         while (ist>>r){
            vr.push_back(r);
         }
         if (first){
            dimFirst=vr.size();
            first=false;
            vvr.push_back(vr);
         }
         else{
            work = (vr.size()==dimFirst);
            if (work) vvr.push_back(vr);
         }
      }
   }
   Z m=vvr.size();
   if (m==0){
      cpmwarning("trivial exit m==0");
      CPM_MZ
      return false;
   }
   Z n=vvr[0].size();
   Z i,j,err=0;
   for (i=1;i<m;++i){
      if (vvr[i].size()!=n){
         err++;
      }
   }
   cpmassert(err==0,loc);
   if (n==0){
      cpmwarning("trivial exit n==0");
      CPM_MZ
      return false;
   }
   R_Matrix res(m,n);
   for (i=1;i<=m;++i) for (j=1;j<=n;++j) res[i][j]=vvr[i-1][j-1];
   *this=res;
   CPM_MZ
   return true;
}

//////////////////////////// class SVD ///////////////////////////////

SVD::SVD(const R_Matrix& a):m(a.dim1()),n(a.dim2()),
A(a),U(a),V()
{
   d=CpmRootX::inf(m,n);
   w=svd(U,V);
   Vt=~V;
}
//#endif

R SVD::getError()const
{
   R_Matrix u1=~U*U;
   R error1=(u1-u1.net(1)).abs();
      // deviation from isometry
//   cout<<"deviation 1 from isometry ="<<error1<<endl;
   R_Matrix v1=~V*V;
   R error2=(v1-v1.net(1)).abs();
//   cout<<"deviation 2 from isometry ="<<error2<<endl;
      // deviation from isometry
   //R error3=(A-U*R_Matrix(w)*Vt).abs();
   R error3=A.dis(U*R_Matrix(w)*Vt);
      // deviation from representing the matrix we started from
//   cout<<"deviation from reliability ="<<error3<<endl;
   bool b1=(U.dim1()==m);
   bool b2=(U.dim2()==d);
   bool b3=(w.dim()==d);
   bool b4=(Vt.dim1()==d);
   bool b5=(Vt.dim2()==n);
   R error4=(b1 && b2 && b3 && b4 && b5) ? 0 : 1;
      // deviation from the expected frame-type of
      // the matrices
//   cout<<"deviation from dimensionality ="<<error4<<endl;
   return error1+error2+error3+error4;
}

R_Matrix SVD::operator()(const F<R,R>& f)const
{
   R_Vector fw(d);
   for (Z i=1;i<=d;i++) fw[i]=f(w[i]);
   return U*R_Matrix(fw)*Vt;
}
