//? cpmrvector.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmrvector.h>
#include <cpmrfunc.h>

using namespace CpmRoot;
using namespace CpmSystem;
using namespace CpmArrays;

R R_Vector::etaInv=1e-12;

bool R_Vector::isVal()const
{
   Z n=dim();
   for (Z i=1;i<=n;i++){
      if (!CpmRoot::isVal(rep[i])) return false;
   }
   return true;
}

R_Vector::R_Vector(Iv iv, Z n):rep(n)
{
   Word loc("R_Vector(Iv iv, Z n)");
   cpmassert(n>0,loc);
   R x1=iv.inf();
   R xn=iv.sup();
   if (n==1){
      rep[1]=(x1+xn)*0.5;
   }
   else if (n==2){
      rep[1]=x1;
      rep[2]=xn;
   }
   else if (n>2){
      rep[1]=x1;
      rep[n]=xn;
      R h=(xn-x1)/(n-1.);
      for (Z i=2;i<n;i++) rep[i]=rep[i-1]+h;
   }
}

R_Vector::R_Vector(Z n, Iv iv, Z j):rep(n)
{
   Word loc("R_Vector(Z,Iv)");
   cpmassert(n>0,loc);
   Z jj=j;
   for (Z i=1;i<=n;++i) rep[i]= (jj==0 ? iv.ran() : iv.ran(jj++));
}

R_Vector::R_Vector(R xStart, R step, Z n):rep(n)
{
   Word loc("R_Vector(R,R,Z)");
   cpmassert(n>=0,loc);
   R xi=xStart;
   if (n>=1){
      rep.cui(1)=xi;
   }
   if (n>=2){
      for (Z i=2;i<=n;++i){
         xi+=step;
         rep.cui(i)=xi;
      }
   }
}

R_Vector::R_Vector(R2 p, Z n):rep(n)
{
   Word loc("R_Vector(R2 p, Z n)");
   cpmassert(n>0,loc);
   R x1=p[1];
   R xn=p[2];
   if (n==1){
      rep[1]=(x1+xn)*0.5;
   }
   else if (n==2){
      rep[1]=x1;
      rep[2]=xn;
   }
   else if (n>2){
      rep[1]=x1;
      rep[n]=xn;
      R h=(xn-x1)/(n-1.);
      for (Z i=2;i<n;i++) rep[i]=rep[i-1]+h;
   }
}

R3 R_Vector::infMeanSup()const
{
   Z n=rep.dim();
   if (n==0) return R3();
   else if (n==1){
      R r=rep[1];
      return R3(r,r,r);
   }
   else{
      R r=rep[1];
      R vMin=r,vMax=r,sum=r;
      for (Z i=2;i<=n;i++){
         r=rep[i];
         sum+=r;
         if (r>vMax) vMax=r;
         if (r<vMin) vMin=r;
      }
      R vMean=sum/n;
      return R3(vMin,vMean,vMax);
   }
}

R3 R_Vector::variation()const
{
   Z n=rep.dim();
   if (n==1){
      return R3();
   }
   else if (n==2){
      R diff=rep[2]-rep[1];
      return R3(diff,cpmabs(diff),diff);
   }
   else{
      Z i=2;
      R xim=rep[1];
      R xi=rep[i];
      R di=xi-xim;
      R a=di;
      R b=cpmabs(di);
      R c=di;
      for (i=3;i<=n;i++){
         xim=xi;
         xi=rep[i]; // only one call to [] needed!
         di=xi-xim;
         if (di<a) a=di;
         if (di>c) c=di;
         b+=cpmabs(di);
      }
      b/=(n-1); // there were n-1 terms in the sum
      return R3(a,b,c);
   }
}

R_Vector::R_Vector(R xStart, R step, R xEnd,
   bool multiplicative, bool precautious )
{
   Word loc("R_Vector::R_Vector(R,R,R,bool)");
   if (xEnd == xStart){
      *this = R_Vector(1,xStart);
      return;
   }
   R dx = step;
   bool reg = (xEnd > xStart);
   R tiny = cpmabs(xEnd-xStart)*1.e-9; // to avoid dependence on numerical
      // noise

   if (multiplicative){
      cpmassert(step != 1.,loc);
      if (reg){
         dx = step > 1. ? step : cpminv(cpmabs(step));
      }
      else{
         dx = step > 1. ? cpminv(step) : cpmabs(step);
      }
   }
   else{
      cpmassert(step != 0.,loc);
      dx = reg ? cpmabs(step) : - cpmabs(step);
   }
   Z n = 0;
   R x = xStart;
   Iv iv = reg ? Iv(xStart-tiny,xEnd+tiny) : Iv(xStart+tiny,xEnd-tiny);
   vector<R> res;
   while (iv.contains(x)){
      res.push_back(x);
      if (multiplicative) x *= dx; else x += dx;
   }
   if (precautious){ // computationally cheap
      bool b1 = iv.contains(res.back());
      bool b2 = iv.contains(x);
      cpmassert(b1 == true,loc);
      cpmassert(b2 == false,loc);
   }
   *this=R_Vector(res);
}

R_Vector R_Vector::pseudoInverse(R acc) const
{
   R_Vector d=*this;
   Z i, n=dim();
   acc=cpmabs(acc);
   if (acc==0.){ // most common case, thus first and without ordering
   // overhead
      for (i=1;i<=n;i++) d[i]=cpminv((*this)[i]);
      return d;
   }
   R_Vector absthis(n);
   for (i=1;i<=n;i++) absthis[i]=cpmabs((*this)[i]);
   R value, maxValue=-1;
   for (i=1;i<=n;i++){ // finding the maximum of the absolute values of
   // the
         // components as fast as possible, thus without ordering
      value=absthis[i];
      if (value>maxValue) maxValue=value;
   }
   R eta=maxValue*acc;
   eta*=0.5;
   eta*=eta;
   R nom,xx;
   for (i=1;i<=n;i++){
      xx=(*this)[i];
      nom=xx*xx+eta;
      cpmassert(nom>0,"R_Vector::pseudoInverse(R acc)");
      d[i]=xx/nom;
   }
   return d;
}

R_Vector R_Vector::scaleTo(const R& r)const
      // returns a vector of absolute value |r| and the direction of
      // *this for positive r and antidirection for negative r.
      // same function is available for complex numbers
{
   R rOrig=abs();
   R_Vector res=*this;
   if (rOrig!=0.){
      R fac=(r/rOrig);
      res*=fac;
   }
   return res;
}

R_Vector R_Vector::posCopy()const
{
   return this->operator()(F<R,R>(CpmFunctions::R_Func::abs));
}

R_Vector R_Vector::log10()const
{
   return this->operator()(F<R,R>(CpmFunctions::R_Func::log10));
}

R_Vector R_Vector::cutCopy(Iv iv)const
{
   R x1=iv[1];
   R x2=iv[2];
   if (x1<0) x1=0;
   if (x2>1) x2=1;
   Z i,n=dim();
   Z span=n-1;
   Z i1=cpmrnd(x1*span+1);
   Z i2=cpmrnd(x2*span+1);
   Z nRes=1+i2-i1;
   Z i1_=i1-1;
   R_Vector res(nRes);
   for (i=1;i<=nRes;++i){
      res.rep[i]=rep[i1_+i];
   }
   return res;
}

R3 R_Vector::infMedSup()const
{
   Z n=dim();
   R r=rep[1];
   if (n==1) return R3(r,r,r);
   else{
      V<Z> indNew=rep.permutationForIncreasingOrder();
      R v_min=rep[indNew[1]];
      R v_max=rep[indNew[n]];
      R v_median;
      Z n_half=n/2;
      if (n_half*2==n){ // then n is even
         v_median=(rep[indNew[n_half]]+rep[indNew[1+n_half]])*0.5;
      }
      else{
         v_median=rep[indNew[1+n_half]];
      }
      return R3(v_min,v_median,v_max);
   }
}

R R_Vector::sup(Z i)const
{
   if (i==0) return rep.sup();
   static Word loc("R_Vector::sup(Z i)");
   cpmassert(i>0,loc);
   Z n=dim();
   cpmassert(n>=1,loc);
   if (n==1) return rep[1];
   else{
      Z iRes=n-i;
      cpmassert(iRes>=1,loc);
      V<Z> indNew=rep.permutationForIncreasingOrder();
      return rep[indNew[iRes]];
   }
}

R R_Vector::inf(Z i)const
{
   if (i==0) return rep.inf();
   static Word loc("R_Vector::inf(Z i)");
   cpmassert(i>0,loc);
   Z n=dim();
   cpmassert(n>=1,loc);
   if (n==1) return rep[1];
   else{
      Z iRes=1+i;
      cpmassert(iRes<=n,loc);
      V<Z> indNew=rep.permutationForIncreasingOrder();
      return rep[indNew[iRes]];
   }
}

R R_Vector::supAbs()const
{
   R res=0.;
   for (Z i=1;i<=dim();++i){
      R xi=cpmabs(rep[i]);
      if (xi>res) res=xi;
   }
   return res;
}

R_Vector R_Vector::der(R h, bool periodic) const
// The vector a.der() results from the vector a by numerical
// differentiation of a function, the equi-distantly sampled image
// (sampling step = h) of which the vector a is considered to be.
// Formulas: Bronstein Semendiajew, Taschenbuch der Mathematik S. 768
// Tabelle 7.13 for generic points and Abramowitz Stegun p. 914 Table
// 25.2 for the boundary points.
// Notice: D2x:=x[i+1]-x[i-1]/2
// D4x:=x[i+2]-x[i-2]/4, then the formula to be used can be written as
// D := (4/3)D2-(1/3)D4 which is OK since for D2==D4 we get D=D2.
{
   Z mL=4;
   static Word loc("R_Vector::der(R,bool)");
   cpmmessage(mL,loc&" started");
   Z i,n=dim();
   R_Vector res(n);
   if (n<=1) return res;
   cpmassert(h!=0,loc);
   R s;
   if (!periodic){
      if (n==2){
         s=R(1)/h;
         res[1]=s*((*this)[2]-(*this)[1]);
         res[2]=res[1];
      }
      else{  // n>=3
         s=R(1)/(h*12);
         R s1=R(1)/(h*2);
         res[1]=s1*((*this)[1]*(-3)+(*this)[2]*4-(*this)[3]);
         res[2]=s1*(-(*this)[1]+(*this)[3]);
         res[n-1]=s1*(-(*this)[n-2]+(*this)[n]);
         res[n]=s1*((*this)[n-2]-(*this)[n-1]*4+(*this)[n]*3);
         for (i=3;i<=n-2;i++){
            res[i]=s*((*this)[i-2]-(*this)[i-1]*8+
               (*this)[i+1]*8-(*this)[i+2]);
         }
      }
   }
   else{
      s=R(1)/(h*12);
      for (i=1;i<=n;i++){
         R a1=this->cyc(i-2);
         R a2=this->cyc(i-1);
         R a3=this->cyc(i+1);
         R a4=this->cyc(i+2);
         res[i]=s*(a1-a2*8+a3*8-a4);
      }
   }
   return res;
}

R_Vector R_Vector::disLap(Outside mode)const
{
   return (*this).rot(1,mode)+(*this).rot(-1,mode)+(*this)*(-2.0);
}

R_Vector R_Vector::crsPrd(const R_Vector& x)const
{
   Z d1=dim();
   if (d1==2){
      Z d2=x.dim();
      if (d2<d1){
         cpmerror(
            "R_Vector::crsPrd(const R_Vector&): dimension mismatch"
         );
         return R_Vector(); // never reached; to avoid warning
      }
      else{
         R_Vector res(1);
         res[1]=rep[1]*(x.rep)[2]-rep[2]*(x.rep)[1];
         return res;
      }
   }
   if (d1==3){
      Z d2=x.dim();
      if (d2<d1){
         cpmerror(
            "R_Vector::crsPrd(const R_Vector&): dimension mismatch"
         );
         return R_Vector(); // never reached; to avoid warning
      }
      else{
         R_Vector res(3);
         res[1]=rep[2]*(x.rep)[3]-rep[3]*(x.rep)[2];
         res[2]=rep[3]*(x.rep)[1]-rep[1]*(x.rep)[3];
         res[3]=rep[1]*(x.rep)[2]-rep[2]*(x.rep)[1];
         return res;
      }
   }
   else{
      cpmerror(
         "R_Vector::crsPrd(const R_Vector&): dim differs from 2 and 3"
      );
      return R_Vector(); // never happens; to avoid warning
   }
}

bool R_Vector::scanFromFreeFor(istream& ifs)
{
   CpmRootX::R1 x;
   std::vector<R> res;
   while (x.scanFrom(ifs)){
      res.push_back((R)x);
   }
   *this=R_Vector(res);
   return true;
}

Z R_Vector::indOfNerCom(R const& r)const
{

   if (r<=rep.fir()) return rep.b();
   if (r>=rep.last()) return rep.e();
   Z i=rep.locAsc(r);
   Z ip=i+1;
   R mi=(rep[i]+rep[ip])*0.5;
   if (r<=mi) return i;
   else return ip;
}

