//? cpmcamera3.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
     Description: class Camera achieves visualization of 3-dimensional
      objects via anaglyph-technique
*/

namespace CpmCamera3{

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::C;
   using CpmRoot::Word;

   using CpmGeo::Angle;
   using CpmGeo::Iv; // Interval

   using CpmRootX::B;
   using CpmRootX::R1;
   using CpmRootX::Z1;

   using CpmLinAlg::R2;
   using CpmLinAlg::R3;
   using CpmLinAlg::RR2;
   using CpmLinAlg::RR3;

   using CpmArrays::T2;
   using CpmArrays::X2;

   using CpmFunctions::F;
   using CpmFunctions::F1;

   using CpmArrays::V;
   using CpmGraphics::VecGraph;
   using CpmGraphics::VecGraphs;

   using CpmImaging::Response;

using namespace CpmColCam;
using namespace CpmDim3;

//////////////////////////// class Camera ///////////////////////////////

   class Camera{ // maps the 3D object space onto a 2D image rectangle 
      // Tool for visualizing the motion of clusters
      // in space using the 'red green glasses method' (anaglyph method)

// independent data

      Color color1;
         // color for image corresponding to image 1 (left eye)

      Color color2;
         // color for image corresponding to image 2 (right eye)

      Color color3;
         // color for painting the background. It may be better to have
         // this not completely black in order to reduce visibility of
         // reflexes.

      R s;
         // distance from each of the viewpoints ('eyes') to the point of
         // convergence pc (i.e the point where the optical axes of the
         // eyes intersect, we only consider stereo cameras for which
         // such an intersection point exist)

      R frac;
         // s*frac is the distance from the viewpoint to an imaginary
         // transparent coordinate-projection screen on which the
         // intersection of a ray from object point to viewpoint
         // generates the (x,y) record of the point.
         // (1.-frac)*s is the distance from the point of convergence
         // to the point of interest (poi).
         // For frac<1, the topological order is: eye,poi,pc.
         // For frac>1 we have eye, pc, poi
         // frac==1 is a natural situation.
         // frac>0 is assumed.

      Vec e1;
         // unit vector from viewpoint 1 to the point of convergence

      Vec e2;
         // unit vector from viewpoint 2 to the point of convergence

      Vec v;
         // unit vector in direction of e1.cross(e2).
         // Vertical direction in the image planes. If e1 and e2 are
         // close together or identical, then they don't determine v,
         // So we have to treat this quantity as independent

      Spc poi;
         // 'point of interest'= center of field of view in object space
         // all construction methods of Camera make this point the origin
         // of Spc(). Geometrical transformations enable this to change
         // after construction.

      mutable Graph gr;  
         // viewing window in the film plane. Defines the position and
         // size of the 'film gate'.
         // Note that the film gate has a well defined origin.

      mutable Response rsp;
         // a quantity which controls the conversion from R to 
         // pixel brightness control values

 // dependent data

      Spc vp1; // view point 1, left eye

      Spc vp2; // view point 2, right eye

      Spc vp;  // mean of vp1 and vp2

      Spc pc; // point of convergence

      Vec h1;
         // e1.cross(v), horizontal direction in image plane 1

      Vec h2;
         // e2.cross(v), horizontal direction in image plane 2

      Vec h;
         // unit vector in direction of (h1+h2)/2

      Vec e;
         // unit vector in direction of (e1+e2)/2

      void complete(void);
         // assumes that the indpendent data have their valid values
         // and initializes the dependent data accordingly

      B stereo_;
         // initialized as false. If false we use the camera as a 
         // mono-camera. This is useful for capturing 3D line art 
         // for creating scientific illustrations
         // Till 2009-11-30 I had 'B mono;' instead. Now the logic is
         // clearer since 2D is treated correctly automatically.

   public:
      typedef Camera Type;
      CPM_IO
      Word nameOf()const{ return Word("CpmCamera3::Camera");}
      Camera(R imageHeight=1, R aspectRatio=1.4,
         Angle const& viewAngle=Angle(10), // was 50
         R frac=1., Angle const& conv=Angle(2.5), // was 5
         Frame const& fr=Frame());
   // Takes frac from the argument and sets s from imageHeight,
   // aspectRatio, and viewAngle.   poi is the origin. 
   // The viewpoints are located in the (1,3)-plane
   // and the mean vision direction is the -3 direction.
   // This is natural in a coordinate system in which the screen is the
   // (1,2) plane with 1,2 corresponding to conventional x,y ( x-axis to
   // the right, y-direction upwards). In order to follow conventional
   // orientation, the 3 direction points towards the observer of the
   // screen plane, that direction of vision is -3 direction.
   // The angle conv is the deviation of the individual vision directions
   // against the mean vision direction. Thus the angle between the two
   // vision directions is 2*conv. The colors for marking on screen are
   // presently set such that the left eye is assumed to look through a
   // red filter and the right eye through a green filter. Camera
   // position in space different from the one adopted here, is achieved
   // by operator *=(Group const&).

   // To precisely characterize the geometry of the objects, we consider
   // an idealized derived Camera, which results from the actual one by
   // shrinking the convergence angle to zero while the point of
   // convergence remains fixed. This associated mono camera has a unique
   // center of field. The overall location of the generated Camera is
   // such that this 'image center' coincides with the origin of Spc. 
   // The Graph data  member gr has the object space data corresponding
   // to the film piece of the associated mono camera. It is a rectangle
   // of height 'imageHeight' and a width 'imageHeight'*'aspectRatio'.
   // The viewAngle gives the angular size of the film diagonal as
   // seen from the eye position (distance frac*s from the film center).

      Camera(const Spc& cenFilm, const Spc& cenScn, const Vec& ver,
         R aspectRatio=1.4,
         const Angle& viewAngle=Angle(50),
         const Angle& conv=Angle(5),
         const Frame& fr=Frame());
         // center of film, center of scene (=point of concergence),
         // vertical vector, aspect ratio, viewing angle (corresponds
         // to film diagonal, angle of convergence, display frame on
         // screen

      T2< X2<C,R> > capture(const Spc& p, const R& r)const;
         // the values for radius and position of a sphere at point p
         // with radius r as seen on the two image planes of the stereo
         // camera are returned. Notice that simply marking corresponding
         // circles in red and green on screen is not sufficient to get
         // an anaglyph picture due to the 'solid' colors. In the overlap
         // area of red and green one needs yellow and not the solid
         // color of the circle to be drawn last. The first return data
         // component corresponds to the left eye and the second
         // component to the right eye.

      T2< C > capture(const Spc& p)const;
         // Same a previous function, but ignoring the radius and thus
         // returning two graphical positions only

      X2<C,R> captureMono(const Spc& p, const R& r)const;
         // same as the previous function, but activating only 'one
         // half' of the stereo camera and using a vision direction
         // which is the mean of the two stereo vision directions
         // (namely e).

      C captureMono(const Spc& p)const;
         // Same a previous function, but ignoring the radius and thus
         // returning one graphical position only

      Spc inverse(const C& x, const R& mag )const;
         // returns the position p in space of a point which on the first
         // stereo-image gets the 'film-plane-coordinates' x and for
         // which an small extended lateral object near p of linear size
         // s gets imaged as one linear size mag*s on the film.
         // Abstract definition:
         // captureMono(inverse(x,mag),r)=X2<R2,R>(x,r*mag)
         // for all R2 x and R r, mag

      AxVec getAxis()const{ return AxVec(e());}
         // returns a unit-vector in viewing direction as a data 
         // type which makes sense in 2D and 3D

      bool is3D()const{ return true;}
      bool is2D()const{ return false;}
      void setBW_(){ color1=color2=BLACK; color3=WHITE;}

      void fillCircle(const R2& center, R radius,
         const Color& color )const
         {gr.fillCircle(center, radius, color);}
         // shows a circular disk

      Camera& operator *=(const Group& g);
         // moving the whole camera

      void shift(const R& sx, const R& sy, const R& sz=0.);
         // translation of the whole camera. sx is the shift in direction
         // of h (horizontal image direction)
         // and sy is the shift in direction v (vertical image direction)
         // sz is the shift in direction of -e , the z-direction in a
         // local system of reference in which h=x-direction,
         // v=y-direction.

// in the following 3 operations the 'point of interest' is held fixed
// The teminology for the next three functions follows the book Isaac
// Victor Kerlow: The Art of 3-D Computer Animation and Imaging,
// Van Nostrand Reinhold, 1996
      void tilt(const Angle& a);
         // positive a lets the camera look 'more to the sky'
         // (positive rotation arround h)
      void pan(const Angle& a);
         // positive a lets the camera look 'more to the left'
         // (positive rotation arround v)
      void roll(const Angle& a);
         // positive a lets the 'horizon' (horizontal line before roll)
         // appear ascending from the left to the right
         // (positive rotation arround e)

      void adjustFieldOfView( R enlarge, R sx=0, R sy=0, R sz=0);
      // enlarge is a factor by which the viewing window gets enlarged
      // compared to the one which is presently valid. (sx,sy) describes
      // a shift of the viewing window such that (1,1) shifts the lower
      // left corner of the enlarged window to the position previously
      // held by the upper-right corner. sz describes a shift in
      // direction of the mean viewing direction in units of the image
      // diagonal. The latter parameter, as enlarge,
      // is capable of changing the visual size of objects, using sz
      // changes the perspective too.

      void shiftGate(R sx=0, R sy=0){gr.shift(sx,sy);}
         // shifts the 'filmgate' in a manner that is clear from th code

      Color getColor1()const{ return color1;}
      Color getColor2()const{ return color2;}
      Color getColor3()const{ return color3;}

      void mark(const RefSys& rs, Z i, R len=1.)const;
         // marks the positive coordinate i-axis by a chain of points
         // If i>3 all axes are shown, then only a coordinate range
         // not yet ready

      #include <cpmcamerax.h>
         // declarations which do not depend on dimension
   };

} // namespace
