//? cpmcamera3cpp.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*

Description: specific implementation for 3D

*/

namespace CpmCamera3{

using namespace CpmStd;
using namespace CpmRoot;
using namespace CpmColCam;

using CpmDim3::AxVec;
using CpmDim3::Vec;
using CpmDim3::Spc;
using CpmDim3::RefSys;
using CpmDim3::Group;
using CpmDim3::SubsetOfSpace;
using CpmDim3::Skeleton;
using CpmDim3::WireFrame;
using CpmDim3::Texture;
using CpmGraphics::BLUE;

/////////////////// class Camera ///////////////////////////////////

void Camera::complete(void)
// s,frac,e1,e2,poi,v given
{
   e=(e1+e2).polDec().c2();
   R s1f=s*(R(1)-frac);
   pc=poi+e*s1f;
   vp1=pc-e1*s;
   vp2=pc-e2*s;
   vp=pc-e*s;
   h1=e1.crossExc(v);
   h2=e2.crossExc(v);
   h=(h1+h2).polDec().c2();
   gr.stereo();
}

V<Vec> Camera::getVectors()const
{
   V<Vec> res(3);
   res[1]=h;
   res[2]=v;
   res[3]=-e;
   return res;
}

Camera::Camera(R imageHeight, R aspectRatio, const Angle& viewAngle,
   R frac_, const Angle& conv, const Frame& fr):
   color1(STEREOLEFT), color2(STEREORIGHT),color3(STEREOBLACK),
   frac(frac_),e1(0,0,-1),e2(0,0,-1),v(0,1,0),poi(),gr(fr)
{
   R width=imageHeight*aspectRatio;
   gr.setXY(-R(0.5)*width,-R(0.5)*imageHeight,width,imageHeight);
   C z(imageHeight,width);
   R diag=z.abs();
   R alpha=viewAngle; // alpha now in rad
   R t=cpmtan(alpha*R(0.5)); // frac*s*tan(alpha/2) = diag/2  defines s
   t*=(frac*2);
   if (t<=0) cpmerror("Camera(...): bad viewing angle");
   s=diag/t;
   AxVec av(conv,v);
   Group g2(av);
   Group g1=!g2;
   e1*=g1; // 1 corresponds to the left eye
   e2*=g2;  // 2 corresponds to the right eye
   complete();
}

Camera::Camera(const Spc& cenFilm, const Spc& cenScn, const Vec& ver,
   R aspectRatio, const Angle& viewAngle, const Angle& conv,
   const Frame& fr):
   color1(STEREOLEFT), color2(STEREORIGHT),color3(STEREOBLACK),
   frac(1),poi(cenFilm),gr(fr)
{
   Z mL=2;
   cpmmessage(mL,"Camera(Spc,Spc,...) started");
   Word loc("Camera(Spc,Spc,Vec,R,Angle,R,Angle,Frame)");
   Vec cc=cenScn-poi;
   s=cc.abs();
   cpmassert(s>0,loc);
   R sInv=R(1)/s;
   R diag=s*cpmtan(viewAngle.toR()*R(0.5))*2;
   R imageHeight=diag/cpmsqrt(aspectRatio*aspectRatio+1);
   R width=imageHeight*aspectRatio;
   gr.setXY(-R(0.5)*width,-R(0.5)*imageHeight,width,imageHeight);
   e=cc*sInv;
   e1=e;
   e2=e;
   Vec ver1=ver-e*(e|ver);
   R ver1av=ver1.abs();
   cpmassert(ver1av>0.,loc);
   R ver1avInv=R(1)/ver1av;
   v=ver1*ver1avInv;
   AxVec av(conv,v);
   Group g2(av);
   Group g1=!g2;
   e1*=g1; // 1 corresponds to the left eye
   e2*=g2;  // 2 corresponds to the right eye
   complete();
   cpmmessage(mL,"Camera(Spc,Spc,...) done");
}

bool Camera::prnOn(ostream& out)const
{
   using CpmRoot::write;
   if (!CpmRoot::writeTitle("Camera",out)) return false;
   if (!color1.prnOn(out)) return false;
   if (!color2.prnOn(out)) return false;
   if (!color3.prnOn(out)) return false;
   if (!write(s,out)) return false;
   if (!write(frac,out)) return false;
   if (!poi.prnOn(out)) return false;
   if (!v.prnOn(out)) return false;
   if (!e1.prnOn(out)) return false;
   if (!e2.prnOn(out)) return false;
  // if (!gr.prnOn(out)) return false; //experiment
   return true;
}

bool Camera::scanFrom(istream& in)
{
   using CpmRoot::read;
   if (!color1.scanFrom(in)) return false;
   if (!color2.scanFrom(in)) return false;
   if (!color3.scanFrom(in)) return false;
   if (!read(s,in)) return false;
   if (!read(frac,in)) return false;
   if (!poi.scanFrom(in)) return false;
   if (!v.scanFrom(in)) return false;
   if (!e1.scanFrom(in)) return false;
   if (!e2.scanFrom(in)) return false;
   if (!gr.scanFrom(in)) return false;
   complete();
   return true;
}

Camera& Camera::operator *=(const Group& g)
{
   poi*=g; e1*=g; e2*=g; v*=g; complete();
      // using complete here assures constancy of geometry even in the
      // presence of accumulating roundoff errors
   return *this;
}

void Camera::shift(const R& sx, const R& sy, const R& sz)
{
   Vec tr=h*sx+v*sy-e*sz;
      // notice that h,v,-e is a camera-fixed system of reference
   Group g(tr);
   operator *=(g);
}

void Camera::tilt(const Angle& a)
{
   AxVec av(a,h);
   Group g(poi,av);
   operator *=(g);
}

void Camera::pan(const Angle& a)
{
   AxVec av(a,v);
   Group g(poi,av);
   operator *=(g);
}

void Camera::roll(const Angle& a)
{
   AxVec av(a,e);
   Group g(poi,av);
   operator *=(g);
}

void Camera::adjustFieldOfView( R enlarge, R sx, R sy, R sz)
{
   gr.scale(enlarge);
   R sxa=gr.getdX();
   R sya=gr.getdY();
   R sza=( sz==0.  ?  R(0)  :  sz*cpmsqrt(sxa*sxa+sya*sya) );
   sxa*=sx;
   sya*=sy;
   shift(sxa,sya,sza);
}

void Camera::scale(R enlarge)
{
   gr.scale(enlarge);
   s*=enlarge;
   complete();
}

T2<C> Camera::capture(const Spc& p)const
{
   R f=s*frac;

   Vec d=p-vp;
   R sp=(d|e);
   cpmassert(sp!=0,"Camera::capture(Spc)");
   R demag=f/sp;
      // if one has a different demag for each eye, one gets a curved
      // visual representation of planes which is not yet understood

   Vec s1=(p-vp1)*demag;
   R x1=(s1|h1);
   R y1=(s1|v);
   C res1(x1,y1);

   Vec s2=(p-vp2)*demag;
   R x2=(s2|h2);
   R y2=(s2|v);
   C res2(x2,y2);

   return T2<C>(res1,res2);
}

T2< X2<C,R> > Camera::capture(const Spc& p,
         const R& r)const
{
   R f=s*frac;

   Vec d=p-vp;
   R sp=(d|e);
   cpmassert(sp!=0,"Camera::capture(Spc,R)");
   R demag=f/sp;
      // if one has a different demag for each eye, one gets a curved
      // visual representation of planes which is not yet understood

   Vec s1=(p-vp1)*demag;
   R x1=(s1|h1);
   R y1=(s1|v);
   C res1(x1,y1);

   Vec s2=(p-vp2)*demag;
   R x2=(s2|h2);
   R y2=(s2|v);
   C res2(x2,y2);

   R rd=r*demag;

   typedef X2<C,R> prr2;
   prr2 p1(res1,rd);
   prr2 p2(res2,rd);
   return T2<prr2>(p1,p2);
}

C Camera::captureMono(const Spc& p)const
{
   Vec d=p-vp;
   R f=s*frac;
   R sp=(d|e);
   cpmassert(sp!=0,"Camera::captureMono(Spc)");
   R demag=f/sp;
   d*=demag;
   R x=(d|h);
   R y=(d|v);
   return C(x,y);
}

X2<C,R> Camera::captureMono(const Spc& p, const R& r)const
{
   Vec d=p-vp;
   R f=s*frac;
   R sp=(d|e);
   cpmassert(sp!=0,"Camera::captureMono(Spc,R)");
   R demag=f/sp;
   d*=demag;
   R x=(d|h);
   R y=(d|v);
   C pr(x,y);
   return X2<C,R>(pr,r*demag);
}

void Camera::mark(const Spc& x, R r, const Color& c)const
{
   if (c.transparent()) return;
   else if (c.stereo()){
      T2< X2<C,R> > ppp=capture(x,r);
      R r1=ppp.c1().c2();
      R2 x1=R2(ppp.c1().c1());
      R r2=ppp.c2().c2();
      R2 x2=R2(ppp.c2().c1());
      fillCircle(x1,r1,getColor1());
      fillCircle(x2,r2,getColor2());
   }
   else{
      X2<C,R> prr=captureMono(x,r);
      fillCircle(R2(prr.c1()),prr.c2(),c);
   }
}

void Camera::mark(const Spc& x, R r)const
{
   mark(x,r,Color(INVISIBLE));
}

bool Camera::mark(const Spc& x, const Color& c)const
   // shows the point x in color c (as seen by camera *this in
   // 1-direction)
{
   const Z size=4;
   C px;
   bool stereo=!c.visible()&&is3D()&&stereo_;
   if (stereo){
      T2< C > ppp=capture(x);
      C p1=ppp.c1();
      C p2=ppp.c2();
      px=(p1+p2)*0.5;
      gr.mark(p1,size,color1);
      gr.mark(p2,size,color2);
   }
   else{
      px=captureMono(x);
      if (!c.transparent()) gr.mark(px,size,c);
   }
   return gr.indFunc(R2(px));
}

void Camera::mark(const Spc& a, const Spc& b, R rFoot)const
{
   T2<C> pa=capture(a);
   C a1=pa.c1();
   C a2=pa.c2();
   T2<C> pb=capture(b);
   C b1=pb.c1();
   C b2=pb.c2();
   gr.draw(R2(a1),R2(b1),color1);
   gr.draw(R2(a2),R2(b2),color2);
   if (rFoot>0){
      fillCircle(R2(a1),rFoot,color1);
      fillCircle(R2(a2),rFoot,color2);
   }
}

void Camera::mark(const Spc& x1, const Spc& x2,
                  const Color& c, R rFoot)const
   // shows a line from x1 to x2 in color c (as seen by camera *this
   // in 1-direction)
{
   bool stereo=!c.visible()&&is3D()&&stereo_;
   if ( stereo ){
      mark(x1,x2,rFoot);
   }
   else{ // mono
      C y1=captureMono(x1);
      C y2=captureMono(x2);
      gr.draw(R2(y1),R2(y2),c);
         // was draw_simple
      if (rFoot>0) fillCircle(R2(y1),rFoot,c);
   }
}

Spc Camera::inverse(const C& x, const R& mag)const
{
   R f=s*frac;
   Vec vFilm=e*f+h*x[1]+v*x[2];
   if (mag<=0)
      cpmerror("Spc Camera::inverse(...)const: unexpected mag<=0");
   R magInv=R(1)/mag;
   vFilm*=magInv;
   return vp+vFilm;
}

void Camera::mark(const RefSys& rs, Z i, R len)const
{
   len; // not used
   if (i<=3){
      const R step=2e-6;
      Spc p=rs.o();
      Vec sh;
      if (i==1) sh=rs.e1();
      else if (i==2) sh=rs.e2();
      else if (i==3) sh=rs.e3();
      else if (i==-1) sh=-rs.e1();
      else if (i==-2) sh=-rs.e2();
      else if (i==-3) sh=-rs.e3();
      else cpmerror("Camera::mark(RefSys,Z i), i out of range");
      sh*=step;
      Color c(Z(-1));
      while (isVisible(p)){p+=sh; mark(p,c);}
   }
   else{

   }
}

void Camera::dirTo_(Spc const& x, R const& r, R fitFactor,
   Angle tilt, Angle pan, RefSys const& rs)
{
   Z mL=3;
   Word loc("Camera::dirTo_(Spc,R,...)");
   CPM_MA
   RefSys rsOrig=getRefSys();
   Group g=rs/rsOrig;
   operator *=(g); // now the camera has standard orientation with
      //respect to system rs
   Vec sh=x-poi;
   operator *=(Group(sh));
   operator *=(Group(x,AxVec(tilt,rs.e1())));
   operator *=(Group(x,AxVec(pan,rs.e2())));
// characterizing the present camera:
   R hh=getdY()*0.5;
   if (hh<=0) cpmerror(loc&": unexpected hh<=0");
   R wh=getdX()*0.5;
   R aspectRatio=wh/hh;
   R dis=s*frac;
   if (dis<=0) cpmerror("Camera::dirTo_(...): unexpected s*frac<=0");
   R alpha_hh=cpmatan(hh/dis);
       // angle under which the half hight of the object is seen
   R nom=cpmsin(alpha_hh)*frac;
   if (nom<=0) cpmerror("Camera::dirTo_(...): unexpected nominator<=0");
// adjusting the camera such that it is capable of framing the
// bounding sphere of sos. Viewing angle and aspect ratio of frame
// remain unchanged.
   R rAct= r<=0 ? R(1) : r; // r==0 occurs e.g. in visualizing a single
      // point and this should not disappear, added 2007-10-13
   R s= rAct/nom;
   R hhNew=s*frac*cpmtan(alpha_hh)*fitFactor;
   R imageHeight=hhNew*2;
   R width=imageHeight*aspectRatio;
   gr.setXY(-R(0.5)*width,-R(0.5)*imageHeight,width,imageHeight);
   complete();
   CPM_MZ
}

void Camera::show(const SubsetOfSpace& sos, const Texture& tx,
   bool bodyFix)const
   // areal visualisation of extended bodies under control of
   // method parameters dr, ds.
{
    using CpmArrays::VV;
    using CpmLinAlg::Z2;
    using namespace CpmGraphics;

   Z mL=3;
   Word loc("Camera::show(SubsetOfSpace,Texture,bool)");
   CPM_MA

    const R fineFac=0.25;
    const R coarseFac=5;
      // factors for setting the method parameters of SubsetOfSpace::hit
      // in terms of the basic visual resolution as given by the
      // pixel size.

    const R tWait=2;
      // time for display of a important message

    if (sos(vp1)<=0){
       cpmwarning("Camera::show(...): left eye inside the body");
       cpmwait(tWait);
    }

    if (sos(vp2)<=0){
       cpmwarning("Camera::show(...): right eye inside the body");
       cpmwait(tWait);
    }

    Z i0=gr.a0();
    Z j0=gr.b0();

    Z nx=dim1();
    Z ny=dim2();
    R dx=getdX()/nx;
    R dy=getdY()/ny;
    R resolution=CpmRootX::inf(dx,dy);

    R sFine=fineFac*resolution;
    R sCoarse=coarseFac*resolution;

    R xStart=getxLeft()+dx*0.5;
    R yStart=getyUpper()-dy*0.5;
         // (x,y) pixel center of the first (top-left) film pixel
         // in object space
    R x=xStart;
    R y;
    R h0=s*frac*0.1;
    R nxInv=R(1)/nx;
    RefSys sosrs=sos.getRefSys();
    for (Z i=0;i<nx; i++){
       cpmprogress("capture",R(i)*nxInv,3);
       y=yStart;
       for (Z j=0; j<ny; j++){
          Spc sp=poi+h*x+v*y;
          Vec dir1=sp-vp1;
          Vec dir2=sp-vp2;
          dir1.nor_();
          dir2.nor_();

          X2<Spc,bool> prz1=sos.hit(vp1,dir1,sFine,sCoarse);
          X2<Spc,bool> prz2=sos.hit(vp2,dir2,sFine,sCoarse);
          R br1,br2,zero=0;
          if (!bodyFix){
            br1= prz1.c2() ?  tx(prz1.c1()) : zero;
            br2= prz2.c2() ?  tx(prz2.c1()) : zero;
          }
          else{
            br1= prz1.c2() ?  tx.eval(sosrs,prz1.c1()) : zero;
            br2= prz2.c2() ?  tx.eval(sosrs,prz2.c1()) : zero;
          }

          Color c;
          if (br1==0 && br2==0) c=color3; else c=br1*color1+br2*color2;
          gr.setPel(i+i0,j+j0,c);
          y-=dy;
       }
       x+=dx;
    }
    gr.vis();
    CPM_MZ
}

void Camera::mark(SubsetOfSpace const& sos, Texture const& tx,
   bool bodyFix)const
   // areal visualisation of extended bodies under control of
   // method parameters dr, ds.
{
    using CpmArrays::VV;
    using CpmLinAlg::Z2;
    using namespace CpmGraphics;

   Z mL=3;
   Word loc("Camera::mark(SubsetOfSpace,Texture,bool)");
   CPM_MA

    const R fineFac=0.25;
    const R coarseFac=5;
      // factors for setting the method parameters of SubsetOfSpace::hit
      // in terms of the basic visual resolution as given by the
      // pixel size.

    const R tWait=2;
      // time for display of a important message

    if (sos(vp1)<=0){
       cpmwarning("Camera::show(...): left eye inside the body");
       cpmwait(tWait);
    }

    if (sos(vp2)<=0){
       cpmwarning("Camera::show(...): right eye inside the body");
       cpmwait(tWait);
    }

    Z i0=gr.a0();
    Z j0=gr.b0();

    Z nx=dim1();
    Z ny=dim2();
    R dx=getdX()/nx;
    R dy=getdY()/ny;
    R resolution=CpmRootX::inf(dx,dy);

    R sFine=fineFac*resolution;
    R sCoarse=coarseFac*resolution;

    R xStart=getxLeft()+dx*0.5;
    R yStart=getyUpper()-dy*0.5;
         // (x,y) pixel center of the first (top-left) film pixel
         // in object space
    R x=xStart;
    R y;
    R h0=s*frac*0.1;
    R nxInv=1./nx;
    RefSys sosrs=sos.getRefSys();
    for (Z i=0;i<nx; i++){
       cpmprogress("capture",R(i)*nxInv,3);
       y=yStart;
       for (Z j=0; j<ny; j++){
          Spc sp=poi+h*x+v*y;
          Vec dir1=sp-vp1;
          Vec dir2=sp-vp2;
          dir1.nor_();
          dir2.nor_();

          X2<Spc,bool> prz1=sos.hit(vp1,dir1,sFine,sCoarse);
          X2<Spc,bool> prz2=sos.hit(vp2,dir2,sFine,sCoarse);
          R br1,br2,zero=0;
          if (!bodyFix){
            br1= prz1.c2() ?  tx(prz1.c1()) : zero;
            br2= prz2.c2() ?  tx(prz2.c1()) : zero;
          }
          else{
            br1= prz1.c2() ?  tx.eval(sosrs,prz1.c1()) : zero;
            br2= prz2.c2() ?  tx.eval(sosrs,prz2.c1()) : zero;
          }

          Color c;
          if (br1==0 && br2==0) c=color3; else c=br1*color1+br2*color2;
          gr.setPel(i+i0,j+j0,c);
          y-=dy;
       }
       x+=dx;
    }
    CPM_MZ
}

#include <cpmcameraxcpp.h>
   // common part of the implementation

} // namespace
