//? cpmdim3.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*

    Description:
      declares and partly defines classes describing geometrical objects
      which depend on the space dimension in a non-trivial manner.
      The idea is to select the actual version for use by a namespace
      using declaration:
      using namespace CpmDim3;
      will give the classes Vec, AxVec, InTens the form of classes
      describing polar vectors, axial vectors, tensors of inertia in
      3-dimensional space. Wheras
      using namespace CpmDim2;
      selects the proper implementation for 2-dimensional geometry.

      Contains some very efficient concepts of computational geometry
         which seem not to be widely (if ever) known.
      2009-12-04: classes AxVec and Vec are no longer derived classes
      (from R2 or R3) but use numerical tupel spaces as components.
      This lets the arithmetic operations create objects of the
      exactly right type so that one has no longer to rely on automatic
      conversions from R2,R3 on one hand and Vec,AxVec on the other hand.
*/

   class Group;
   class RefSys;
   class Vec;
   class Plane;

//doc namespace CpmDim3{

//////////////////// class AxVec ////////////////////////////////////////

   class AxVec{
// axial vectors, rotation descriptors
      // space of 'axial vectors' i.e. rotations in
      // space and their infinitesimal generators. Interestingly, this is
      // an autonomous class (i.e. needs no support from other classes).
      // The vector space structure over R is inherited from base class.
      // If we interprete an axial vector e.g. as an angular velocity,
      // the the absolute value of the AxVec is dphi/dt (?? dphi/2dt).
      //
      // The natural combination of angular velocites corresponds to
      // relative rotations of frames: consider a rigid body rb_1
      // rotating with respect to the master system of reference with
      // angular velocity omega_1. Then consider a second rigid body rb_2
      // rotating with respect to rb_1 with an angular velocity omega_2.
      // Then rb_2 rotates with respect to the master system with an
      // angular velocity omega_1 + omega_2.
      //
      // If it is a finite rotation, the length is tan(phi/2).
      // This description of finite rotations is closely related to the
      // description by four-component Euler vectors (quaternions). I
      // never found this representation in the literature. Especially
      // the usefulness of this representation (as opposed to the
      // standard representation by orthogonal matrices) in computational
      // geometry seems to be not generally recognized so far.
      // Notice that AxVec defines all operations of a field (in the
      // sense of algebra) without being one: a*(x+y)!=a*x+a*b (i.e.
      // multiplication and addition are not distributive). Both
      // structures come from the Lie group structure of SO(3) endowed
      // with a canonical system of coordinates which approximately
      // canonical in the neighborhood of the unit element. As a Lie
      // group, SO(3) has a meaningful linear space structure near unity
      // and this is (artificially as far as the group is concerned)
      // extended over the whole space. Other axial vectors (e.g. angular
      // momentum, angular velocity) directly correspond to infinitesimal
      // rotations and for those, the linear structure is the primary
      // one.

      R3 x_;

      typedef AxVec Type;

      friend class Group;
      void make(R phi, const R3& n);
         // building block for two constructors of AxVec's corresponding
         // to finite rotations
      R3 toR3()const{ return x_;}
         //: to R3
   public:
      CPM_IO
      CPM_ORDER
      static const Z dimension;
         // in our case 3
      Word nameOf()const{ return Word("CpmDim3::AxVec");}

      AxVec():x_(){}
         // AxVec with all components equal to zero

      AxVec(const R3& a, const RefSys& rs);
         // AxVec with given components with respect to the master system
         // of reference

      explicit AxVec(R3 const& a):x_(a){}
         // constructor from components respect to the master system of
         // reference

      AxVec(R r1, R r2, R r3):x_(r1,r2,r3){}
         // constructor from components respect to the master system of
         // reference

      explicit AxVec(V<R> const& v):x_(v){}
         // For uniformity with Vec

      explicit AxVec(R r3):x_(0,0,r3){}
         // uniformity with 2-dimensional case

      AxVec(Angle const& phi, R3 const& n);
         // constructor from rotation angle and rotation axis
         // n is a descriptor of the rotation axis
         // It has not to be a unit vector, but the corresonding unit
         // vector will be taken for describing the axis. The components
         // of n refer to the master system of reference

      AxVec(Angle const& phi, Vec const& axis);
         // constructor from rotation angle and rotation axis.
         // The Vec axis is a descriptor of the rotation axis
         // It has not to be a unit vector, but the corresonding unit
         // vector will be taken for describing the axis.

      AxVec(Angle const& phi, AxVec const& axis);
         // same as previous function but axis given more reasonably
         // as an axial vector

      explicit AxVec(Angle const& phi, Z i=3);
         // i=1,2,3 rotation arround i-axis (master system !) with angle
         // phi.
         // The skrew sense of the rotation is the standart one:
         // Looking against the axis vector, a positive phi indicates
         // counter-clock-wise rotation. The default is set such that
         // AxVec(phi) in 2-dimennsional geometry and in
         // three-dimensional geometry are consistent

      AxVec( V<Vec> const& b1, V<Vec> const& b2, bool careful=true);
         // constructs the finite rotation which transforms the
         // orthonormal basis b1 into the orthonormal basis b2.
         // If the third parameter has value true, one actually works
         // with potentially modified versions of b1 and b2 which are
         // guaranteed to be orthonormal triplets. Notice that this is
         // (appart from the default) the first constructor which does
         // not refer to a reference system.

// These are the functions which I once got from Rdim, when AxVec was
// derived fro Rdim.
      AxVec& operator *=(R const& r){ x_*=r; return *this;}
      AxVec& operator +=(AxVec const& v){ x_+=v.x_; return *this;}
      AxVec& operator -=(AxVec const& v){ x_-=v.x_; return *this;}
      AxVec operator *(R const& r)const{ return AxVec(x_*r);}
      AxVec operator +(AxVec const& v)const{ return AxVec(x_+v.x_);}
      AxVec operator -(AxVec const& v)const{ return AxVec(x_-v.x_);}
      AxVec operator -(void)const{ return AxVec(-x_);}
      friend AxVec operator *(R const& r, AxVec const& v){ return v*r;}
      R operator |(AxVec const& v)const{ return (x_|v.x_);}
      R nor_(R r=1){ return x_.nor_(r);}
      R abs()const{ return x_.abs();}
      R absSqr()const{ return x_.absSqr();}
      Rdim operator()(void)const{ return x_;}
      R dis(AxVec const& v)const;
      bool isVal()const{ return x_.isVal();}
      R& operator[](Z i){ return x_[i];}
      const R& operator[](Z i)const{ return x_[i]; }
      Z dim()const{ return x_.dim();}
      void neg(){ x_.neg();}
      X2<R,AxVec> polDec()const{
         X2<R,Rdim> y=x_.polDec();
         return X2<R,AxVec>(y.c1(),AxVec(y.c2()));
      }

// other topics:

// group structure, composition of rotations.
      AxVec operator *(const AxVec& a)const;
         // 'product of ternions'
      AxVec operator !()const{ return AxVec(-x_);}
         // !a is the inverse of a with respect to the multiplication
         // given above.
         // The default AxVec is the unit element of the group structure
         // of AxVec.
         // Notice that the inversion for the additive structure and the
         // multiplicative structures are the same and, consequently, the
         // neutral elements are the same too.
      AxVec& operator *=(const AxVec& a){ return *this=(*this)*a;}
         // group multiplication as a mutating operation. This hides all
         // R3::operator*=(...) irrespective of argument type ...
         // (which is const R& in our case).
         // Thus we have to define multiplication by R also in the scope
         // of AxVec if we don't want the user of this class to have to
         // be aware of its relation to R3 and make the suitable casts
         // for accessing this operation.

      AxVec& operator *=(const Group&);
         // here (*this) is considered a pure rotation and, hence, a
         // element of a subgroup of Group. The '*' referes to this
         // group.

      AxVec& ref_(Vec const& n);
         //: reflection
         // Changes *this into the axial vector which is obtained from
         // *this by a reflection from a plane which has the normal vector
         // n, which is assumed to be a unit vector. This action differs
         // from the corresponding action on polar vectors by a minus
         // sign.

      AxVec& ref_(Plane const& p);
         //: reflection
         // returns ref_(p.n()) with the previously defined function
         // ref_(Vec).

// operations for infinitesimal rotations and angular velocities

      AxVec exp(const R& t, bool accurate=true)const;
         // the standard exponential function from the Lie algebra
         // (to which *this belongs here) and the Lie group (return
         // value).
         // If omega is a constant angular velocity, then omega.exp(t) is
         // the finite rotation which results from having this angular
         // velocity be effective over the time span t.
         // If the last argument is false, we return the approximation
         // for small angles which is a.exp(t)=a*t*0.5

      AxVec log()const;
         // inverse function to exp()

      AxVec pow(R p)const;
         // AxVec x=..; x.pow(3)=x*x*x etc. with the normal
         // generalization to rational and, final, real exponents.

      AxVec adjointAction(const AxVec& d)const;
         // In mathematical terminology this is the natural action of the
         // Lie group (to which argument d belongs) on its Lie algebra
         // (to which *this belongs). This natural action sometimes is
         // called the adjoint representation of the Lie group.
         // In Physics the most elementary appearence of this operation
         // is the natural transformation of angular velocities under
         // rotations.

// vector product
      AxVec cross(const AxVec& a)const
         { return AxVec(x_.crsPrd(a.x_));}
         // cross product or vector product of axial vectors. This is
         // related to the commutator of infinitesimal rotations:
         // Generally we have
         // d1*d2-d2*d1 = 2*(d1.cross(d2))/(1- (d1|d2))
         // And for infinitesimal transformations the nominator can be
         // neglected.

      Vec cross(Vec const& v)const;

      Angle phi()const;
         //: phi
         // returns the rotation angle belonging to *this: Range is [0,
         // Angle(piAppr)] piAppr=2*atan(hugeVal), where hugeVal is a
         // singularity management constant defined in cpmdim3.cpp,
         // presently set to 1e12

      AxVec& phi_(Angle const& phi);
         //: phi
         // Changes *this into a rotation vector which has as rotation
         // axis the unit vector given by *this and phi as the rotation
         // angle.
         // This is a useful operation in situations in which the axis
         // of rotation can be defined as a vector product of two vectors,
         // and the rotation angle has to be set from requirements which
         // can't be expressed in terms of these vectors.

      X2<Angle,Vec> toAngAxs()const;
         // returns a pair made out of rotation angle and rotation
         // axis (as unit vector)

      AxVec ran(Z j=0)const;
         // returns a random value such that the rotation value of
         // the result is in [-|a|,|a|], where a is the rotation
         // angle of *this. The rotation axis of the result is
         // random and independent on the axis of *this.
         // (It employes Vec::ranUnitVec.)
      static V<AxVec> canBas()
      { return V<AxVec>{AxVec(1,0,0),AxVec(0,1,0),AxVec(0,0,1)};}
   };

   RR3 toMat(const AxVec& d);
      // rotation matrix belonging to the axial vector. I.e.
      // x*toMat(d)=x*d, where on the left-hand side we have R3*RR3
      // (see cpmlinalg.h). Although the matrix is less efficient in
      // acting on vectors compared to the right-hand side x*d, the
      // matrix is the only thing to describe easily the transformation
      // properties of tensors of rank 2 and higher.

////////////////////// class Vec ////////////////////////////////////////

   class Vec{
// vectors, space of point differences and velocities
      friend class Group;
      friend class AxVec;
      Rdim x_;
      typedef Vec Type;
   public:
      CPM_IO
      CPM_ORDER
      static const Z dimension;
      // CPM_IO and CPM_ORDER from R3 !!!
      // also function R normalize()
      Word nameOf()const{ return Word("CpmDim3::Vec");}
      Vec():x_(){}

      explicit Vec(Rdim const& a):x_(a){}

      explicit Vec(AxVec const& ax):x_(ax()){}

      Vec(const R3& a, const RefSys& rs);

      Vec(R r1, R r2, R r3=0):x_(r1,r2,r3){}

      explicit Vec(C const& a):x_(a[1],a[2],0){}
      explicit Vec(V<R> const& v):x_(v){}
         // dimension independent input (now very useful since
         // V<R> is now a class that can be directly inputted
         // from class RecordHandler (added 2004-05-03)

// These are the functions which I once got from Rdim, when Vec was
// derived from Rdim.
      Vec& operator *=(R const& r){ x_*=r; return *this;}
      Vec& operator +=(Vec const& v){ x_+=v.x_; return *this;}
      Vec& operator -=(Vec const& v){ x_-=v.x_; return *this;}
      Vec operator *(R const& r)const{ return Vec(x_*r);}
      Vec operator +(Vec const& v)const{ return Vec(x_+v.x_);}
      Vec operator -(Vec const& v)const{ return Vec(x_-v.x_);}
      Vec operator -(void)const{ return Vec(-x_);}
      friend Vec operator *(R const& r, Vec const& v){ return v*r;}
      R operator |(Vec const& v)const{ return (x_|v.x_);}
      R nor_(R r=1){ return x_.nor_(r);}
    //  R normalize(){ return x_.nor_();} // old form
      R abs()const{ return x_.abs();}
      R absSqr()const{ return x_.absSqr();}
      Rdim operator()(void)const{ return x_;}
      R dis(Vec const& v)const{ return x_.dis(v.x_);}
      bool isVal()const{ return x_.isVal();}
      R& operator[](Z i){ return x_[i];}
      const R& operator[](Z i)const{ return x_[i]; }
      Z dim()const{ return x_.dim();}
      void neg(){ x_.neg();}
      X2<R,Vec> polDec()const{
         X2<R,Rdim> y=x_.polDec();
         return X2<R,Vec>(y.c1(),Vec(y.c2()));
      }

      void set(R t){ x_.set(t);}

// other topics:

      C toC()const{ return C(x_.x1,x_.x2);}

      R3 toRdim()const{ return x_;}

      friend Vec operator *(const Vec& x, const AxVec& d);
         // x*d is the image of x under the active transformation
         // represented by d. This is the 'master representation' which
         // defines the geometrical meaning of d. The group
         // multiplication in AxVec is defined just such that
         // x*(d1*d2) = (x*d1)*d2.
         // Notice that for mappings R(d): Vec --> Vec, x |--> x*d
         // this implies
         // R(d1*d2)=R(d2) \circ R(d1), if \circ denotes the conventional
         // composition of mappings:
         // (f \circ g)(x)=f(g(x))
         // For the 'contravariant composition' f&g, which is the natural
         // one for mappings written as acting 'from the right'
         // x(f&g)=(xf)g
         // we have
         // R(d1*d2)=R(d1)&R(d2)
         // Notice that we have x*(d1*d2) = (x*d1)*d2 only 'up to
         // roundoff errors'.
         // So x*(d1*d2*....dn)=(...((x*d1)*d2)*....*dn) with an
         // error which increases with n. The vector
         // yn:=(...((x*d1)*d2)*....*dn) will finally change its length
         // considerably with increasing n. On the other hand
         // xn:=x*(d1*d2*....dn) will always result from a single AxVec
         // on x and will thus have the same length as x up to roundoff
         // errors of a single function evaluation. Here we clearly see
         // the benefit from representing rotations by a set of
         // independent data. If we would use matrices (9 real
         // parameters, so 6 more than needed), these matrices have to be
         // orthogonal and each matrix multiplication would spoil the
         // orthogonality of the result by roundoff errors.
         // So we would have to implement a mechanism that restitutes
         // orthogonaliy if deviations have grown to some size.
         // Since enforcing orthogonality is computationally expensive,
         // the only staightforward strategy, namely to do the adjustment
         // after each multiplication, is hardly affordable.

      Vec& operator *=(const AxVec& d){ return *this=(*this)*d;}
         // x*=d resuls from x by applying rotation d

      AxVec cross(const Vec& a)const{ return AxVec(x_.crsPrd(a()));}
         // vector product of vectors is an axial vector
         //(uses R3 CpmLinAlg::crsPrd(const R3&, const R3&))

     Vec cross(const AxVec& a)const{ return Vec(x_.crsPrd(a()));}
        // vector product of a vector with an axial vector is a vector

     Vec crossExc(Vec const& v)const{ return Vec(x_.crsPrd(v.x_));}
         //: cross exceptional

      R spat(const Vec& a, const Vec& b)const
         { return operator|(a.crossExc(b));}

      Vec& operator *=(const Group&);

      Vec rot90()const{ return Vec(x_.x2,-x_.x1,x_.x3);}
         // rotation 90 (degrees)
         // Returns a vector which results from *this by a rotation
         // of 90 degrees arround the 3 axis in positive direction.
         // I came to consider this an important operation by studying
         // lecture notes of Professor Engelbert Schlipf
         // (Fachhochschule Ulm) on technical mechanics.

      X2<Vec,Vec> tanNormDec(Vec const& b)const;
         // the first component of the return pair is the component of b
         // in the direction of *this, and the second component is the
         // normal component

      Vec ref(Vec const& n)const{Vec d=n*((*this|n)*2); return *this-d;}
         //: reflection
         // Returns a vector which is obtained from *this by a reflection
         // from a plane which has the normal vector n, which is assumed
         // to be a unit vector.

      Vec& ref_(Vec const& n)
      {Vec d=n*((*this|n)*2); *this-=d; return *this;}
         //: reflection
         // Changes *this into the vector which is obtained from *this
         // by a reflection from a plane which has the normal vector n,
         // which is assumed to be a unit vector.

      Vec& ref_(Plane const& p);
         //: reflection
         // Reflectio with respect to plane p.

      V<Vec> base(const Vec& b=Vec(0,1,0) )const;
         // returns a positively oriented triplet of orthogonal unit
         // vectors e[1],e[2],e[3].  If *this is not zero, e[1] is
         // parallel to *this else e[1]=Vec(1,0,0).
         // If e[1] and b determine a plane, then e[1],b and e[1],e[2]
         // determine the same oriented plane. Otherwise, e[2] is the
         // first vector from the list Vec(1,0,0), Vec(0,1,0), Vec(0,0,1)
         // which is not proportional to e[1].

      void setDir_(const Vec& v){*this=v.polDec().second()*abs();}
         //: set direction
         // gives *this the direction of v, without changing length

      AxVec rotTo(const Vec& v)const;
         //: rotate to
         // if the return value is interpreted as rotation, it is the
         // most direct one carrying *this into v.

      Angle angTo(const Vec& v)const{ return rotTo(v).phi();}
         //: angle to
         // angle 'from *this to v'

      AxVec rotParOrAntiPar(const Vec& v);
         //: rotate parallel or anti-parallel
         // The smallest rotation (i.e. the one with the absolute value
         // of the rotation angle) which makes *this parallel to the
         // the lines determined by {v,-v}.

      Vec randomize(R dv, Angle const& a, Z j)const;
         // Returns a vector which is a stochastic
         // modification of a copy of *this. The length of the
         // result is uniformly distributed (as a function of j)
         // in the interval [max(0,|*this|-|dv|),|*this|+|dv|], and
         // the direction is obtained from the direction of *this
         // by rotating arround a random axis by an angle which is
         // uniformely distributed between -a and a.

      static Vec ranUnitVec(Z j=0);
         //: random unit vector

      static V<Vec> canBas()
      { return V<Vec>{Vec(1,0,0),Vec(0,1,0),Vec(0,0,1)};}
         //: canonical basis
         // 'Base' seems not to be idiomatic here so I discontinue
         // writing base.

      static V<Vec> homUnitVec(Z n, Z nSteps=10, Z stoppingPeriod=10,
         R dtFac=0.07, R eps=1e-2);
         //: homogeneous unit vectors
         // generates a system of n unit vectors which are
         // first selected at random and then distributed more
         // homogeneously by applying nSteps steps of a
         // repulsive dynamical process. Implementation is identical
         // for 2D and 3D and is in cpmhomunitvecxcpp.h.

     };

//////////////////// class Group ////////////////////////////////////////

   class Spc;
   class InTens;
   class LieAlg;

class Group{
// group of active Euclidean transformations (translations and rotations)
      // Group of active transformations which defines
      // the 'physical geometry' under consideration.
      // All classes X that cooperate with this geometry
      // have to define an action of G on X.
      // In physics, the action of groups on system states x is normally
      // written as x |--> D(g)x for which a natural C++ formulation
      // would be x |--> g*x which would be implemented as
      // X operator *(const Group& g, const X& x).
      //
      // For classes from which one may wish to derive new classes
      // with aditional geometry-related data (e.g. particles with spin
      // derived from a class of scalar particles) one would like to
      // realize this as a virtual function so that 'incremental
      // programming' for the additional degrees of freedom becomes
      // possible and is supported by the programing language.
      // So, only a member function is the canonical implementation
      // of the group action. This means that Group has to act from
      // the right on X. I choose
      // X& X::operator *=(const Group& g);
      // (or also
      //    void X::operator *=(const Group& g); )
      // for the implementation of the group action. In order to let
      // such an operator access the private members of Group
      // all geometrical classes should be friends of Group. So the
      // following friend list has to be extended on occasion.
      // Although we consider the action of reflections on vectors,
      // axial vectors and space points, these reflections are n o t
      // integrated into Group, which thus remains the 'resticted'
      // Euclidean group as opposed to the 'full' Euclidean group.
      friend class AxVec;
      friend class Vec;
      friend class InTens;
      friend class Spc;
      friend Spc operator * (const Spc& x, const Group& g);
// private data: same structure also for 2 dimensions. Discrete symmetries
// (inversion) not present in Group (could be added in a derived group)

      AxVec A;
      Vec a;
         // we write (A,a) for a Group instance with these data
   public:
      typedef Group Type;
      CPM_INVERSION
      CPM_PRODUCT_M
      CPM_IO
      CPM_ORDER
      Word nameOf()const{ return Word("CpmDim3::Group");}

      Group(){}
         // unit element

      Group(const Spc& x0, const AxVec& A1);
         // constructor from fixed point and axial vector for
         // rotation arround that point.
         // inline definition in the Spc group, since Spc has to be
         // defined for implementation of this function.
         // x|--> x0 + (x-x0)*A1, thus x|--> (A1,x0-x0*A1)x
         // Group(x0,A1)=(0,-x0)*(A1,0)*(0,x0)
         // Here we wrote x0 also for the position vector of x0 in the
         // master frame.

      Group( Spc const& x0, AxVec const& A1, Vec const& a1);
         // constructor from fixed point and axial vector for
         // rotation arround that point and a polar vector for
         // a final translation.

      Group(const Spc& x0, const Angle& phi);
         // =Group(x0,AxVec(phi))

      Group(const AxVec& A1, const Vec& a1):A(A1),a(a1){}
         // constructor from axial vector and vector

      explicit Group(const AxVec& A1):A(A1),a(){}
         // constructor from axial vector

      explicit Group(Angle const& alpha):A(AxVec(alpha)),a(){}
         // constructor from angle; this is usefull in context where
         // Group is available and AxVector not (as in my ParDynSys
         // template)

      explicit Group(const Vec& a1):A(),a(a1){}
         // constructor from vector

      Vec shift(void)const{ return a;}
         // returns the shift (translation) part

      AxVec rot(void)const{ return A;}
         // returns the rotation part (the homogeneous transformation)

      friend Vec operator * (const Vec& x, const Group& g)
         // transforming a vector via the fundamental transformation
         // group
      { return x*g.A;}

      static Rdim act(const Rdim& x, const Group& g)
         // transforming a point via the fundamental transformation
         // group. Here the point is represented by its coordinate-
         // tuple with respect to the master system of reference
      { return (Vec(x)*g.A+g.a)();}

      F<Rdim,Rdim> actRdim()const
      { return F1<Rdim,Group,Rdim>(*this)(act);}

      R dis(const Group& g)const
         { return CpmRootX::sup(A.dis(g.A),a.dis(g.a));}

      Group ran(Z j=0)const{ return Group(A.ran(j));}
         // random rotations arround ori with rotation angle up to that of
         // *this. Useful in context where AxVec (which defines the
         // corresponding random sequence) is not available.

      Group ran(Spc const& x0, Z j=0)const
      { return Group(x0, A.ran(j));}
         // random rotations arround x0 with rotation angle up to that
         // of *this. Useful in context where AxVec (which defines the
         // corresponding random sequence) is not available.

      LieAlg log()const;
   };

   inline Group& Group::operator *=(const Group& g)
   {
      A*=g.A; (a*=g.A)+=g.a; return *this;
   }

   inline Group Group::operator !()const
   {
      return Group(!A,(-a)*(!A));
   }

//////////////////// class LieAlg ////////////////////////////////////////

class LieAlg{ // Lie algebra of class Group.

      AxVec A_;
      Vec a_;
         // we write (A_,a_) for a LieAlg instance with these data
      typedef LieAlg Type;
   public:

      CPM_IO
      CPM_ORDER
      Word nameOf()const{ return Word("CpmDim3::LieAlg");}

      LieAlg(){}
         // zero element

      LieAlg(AxVec const& A, Vec const& a):A_(A),a_(a){}
         // constructor from axial vector and vector

      explicit LieAlg(AxVec const& A):A_(A),a_(){}
         // constructor from axial vector

      explicit LieAlg(Vec const& a):A_(),a_(a){}
         // constructor from vector

      Vec shift(void)const{ return a_;}
         // returns the shift (translation) part

      AxVec rot(void)const{ return A_;}
         // returns the rotation part (the homogeneous transformation)

      LieAlg operator +(LieAlg const& x)const
         { return LieAlg(A_+x.A_,a_+x.a_);}

      LieAlg& operator +=(LieAlg const& x)
         { A_+=x.A_; a_+=x.a_; return *this;}

      LieAlg operator -(LieAlg const& x)const
         { return LieAlg(A_-x.A_,a_-x.a_);}

      LieAlg& operator -=(LieAlg const& x)
         { A_-=x.A_; a_-=x.a_; return *this;}

      LieAlg operator *(R r)const{ return LieAlg(A_*r,a_*r);}

      LieAlg& operator *=(R r)
         { A_*=r; a_*=r; return *this;}

      LieAlg operator *(LieAlg const& x)const
         // commutator as algebra product
         { return LieAlg(A_.cross(x.A_),A_.cross(x.a_)-x.A_.cross(a_));}

      Group exp(R t)const;
         // the standard exponential function from the Lie algebra
         // (to which *this belongs here) and the Lie group (return
         // value).
         // If omega is a constant angular velocity, then omega.exp(t) is
         // the finite rotation which results from having this angular
         // velocity be effective over the time span t.
         // If the last argument is false, we return the approximation
         // for small angles which is a.exp(t)=a*t*0.5

/***********

      static Rdim act(const Rdim& x, const LieAlg& g)
         // transforming a point via the fundamental transformation
         // group. Here the point is represented by its coordinate-
         // tuple with respect to the master system of reference
      { return (Vec(x)*g.A+g.a)();}

      F<Rdim,Rdim> actRdim()const
      { return F1<Rdim,LieAlg,Rdim>(*this)(act);}

      R dis(const LieAlg& g)const
         { return CpmRootX::sup(A.dis(g.A),a.dis(g.a));}

      LieAlg ran(Z j=0)const{ return LieAlg(A.ran(j));}
         // random rotations arround ori with rotation angle up to that of
         // *this. Useful in context where AxVec (which defines the
         // corresponding random sequence) is not available.

      LieAlg ran(Spc const& x0, Z j=0)const
      { return LieAlg(x0, A.ran(j));}
         // random rotations arround x0 with rotation angle up to that
         // of *this. Useful in context where AxVec (which defines the
         // corresponding random sequence) is not available.

*************/
   };
/*******
   inline LieAlg& LieAlg::operator *=(const LieAlg& g)
   {
      A*=g.A; (a*=g.A)+=g.a; return *this;
   }

   inline LieAlg LieAlg::operator !()const
   {
      return LieAlg(!A,(-a)*(!A));
   }
***********/

////////////////////// class Spc ////////////////////////////////////////

   class Spc{
// affine Euclidian space, instances are points and not vectors
      friend class Group;
      friend class RefSys;
      Vec p;
      explicit Spc(const Vec& p1):p(p1){}
      static R refLen;
         // reference length, a length to which the function dis
         // referes in the sense that p1.dis(p2) is |p1-p2|/refLen
         // for points the Euclidian distance of which is much smaller
         // than refLen.
   public:
      typedef Spc Type;
      CPM_IO
      CPM_ORDER

      Word nameOf()const{ return Word("CpmDim3::Spc");}

      Spc():p(){}

      Spc(R x1, R x2, R x3=0):p(x1,x2,x3){}
         // construction from components with respect
         // to the master system of reference, giving two components is
         // sufficient so there is uniformity with 2-dimensional case

      explicit Spc(Rdim const& x):p(x[1],x[2],x[3]){}
         // construction from components with respect
         // to the master system of reference

      explicit Spc(V<R> const& x):p(x[1],x[2],x.dim()<3 ? R(0) : x[3]){}

      Spc(const R3& x, const RefSys& rs);
         // construction from components with respect
         // to the reference system rs

      explicit Spc(const V<R>& x, const RefSys& rs);
         // construction from components with respect
         // to the reference system rs, where the components
         // are represented by a dimension-independent type

      explicit Spc(const C& a):p(a[1],a[2],0){}

      C toC()const{ return p.toC();}

      Rdim toRdim()const{ return p.toRdim();}

      bool isVal()const{ return p.isVal();}

      static Rdim toRdimStatic(const Spc& x)
      { return x.p.toRdim();}

      static F<Spc,Rdim> FSR;

      R& operator[](Z i){return p[i];}
      const R& operator[](Z i)const{return p[i];}
         // access to the components with respect to the master system
         // of reference

      Z dim()const{ return p.dim();}

      void setRefLen(R rl)
      { cpmassert(rl>0,"Spc::setRefLen(R)");refLen=rl;}
      R getRefLen(){ return refLen;}

      friend Spc operator * (const Spc& x, const Group& g)
         // transforming a point via the fundamental transformation
         // group
      { return Spc(x.p*g.A+g.a);}

      Spc& operator *=(const Group& g)
         // action of Group as a mutating operation
      { p=p*g.A+g.a; return *this;}

      Spc& operator /=(const Group& g)
         // inverse action of Group as a mutating operation
         // for 'utmost efficiency' not written as *=!g
       { p-=g.a; p*=!g.A; return *this;}

      Spc& rot_(Spc const& x0,AxVec const& a)
         //: rotation
         // rotation by a, which leaves the point x0 fixed
      { Vec pr=p-x0.p; p=x0.p+pr*a; return *this;}

      Spc& rot(const Spc& x0, const AxVec& a){ return rot_(x0,a);}
         // old style

      friend Spc operator + (const Spc& x, const Vec& a)
         // shifting position by adding a vector
       { return Spc(x.p+a);}

      friend Spc operator - (const Spc& x, const Vec& a)
         // shifting position by adding the vector -a
      { return Spc(x.p-a);}

      Spc& operator +=(const Vec& a)
         // adding a vector as a mutating operation
      { p+=a; return *this;}

      Spc& operator -=(const Vec& a)
         // subtracting a vector as a mutating operation
      { p-=a; return *this;}

      friend Vec operator - (const Spc& x, const Spc& y)
         // getting a vector from subtracting two points
      { return Vec(x.p-y.p);}

      Spc ref(Vec const& n, Spc const& s0)const
         //: reflection
         // Returns the image of *this under a reflection from
         // a plane which goes through point s0 and has the
         // normal unit vector n.
      { return s0+(*this-s0).ref(n);}

      Spc& ref_(Vec const& n, Spc const& s0)
         //: reflection
         // Changes *this into its image under a reflection from
         // a plane which goes through point s0 and has the
         // normal unit vector n.
      { return *this = ref(n,s0);}

      Spc& ref_(Plane const& p);
         //: reflection
         // Changes *this into its image under a reflection from
         // the plane p.

      Vec vecTo(const Spc& goal)const{ return Vec(goal.p-p);}
         // Vector which shifts *this into goal

      Spc cen(Spc const& x)const{ return Spc((p+x.p)*0.5);}
         //: center
         // returns the center (midpoint) between *this and x

      Spc scl(Spc const& fix, R fac)const{ return fix+(*this-fix)*fac;}
         //: scale
         // Returns the point resulting from *this by
         // multiplying the position vector relative to
         // the point 'fix' by factor fac.

      Spc& scl_(Spc const& fix, R fac){ return *this=scl(fix,fac);}

      static Spc convex(const V<Spc>& x, V<R>& w, bool careful=true);
         // convex linear combination of the x.
         // Most important application: center-of-mass of a particle
         // system.
         // w is a family of as many positive weights (that add up to 1)
         // as there are members of x.
         // If the boolean parameter careful is set to true no
         // assumptions on w are made and w will be changed if positivity
         // and normalization are not satisfied on input.
         // The mode of enforcement is natural: Changing the sign of
         // negative w[i]'s and finally dividing by the sum of the
         // w[i]'s.
         // If careful is set to false, the w's will not be changed and
         // used as they are. The resulting point then may not belong to
         // the convex hull of the x[i]'s

      static X2<Spc,R> envSph(V<Spc> const& vs);
         //: enveloping sphere
         // center and radius of an envelopping sphere of an array
         // of points

      void convexAsMember(const V<Spc>& x, V<R>& w, bool careful=true)
      { *this=convex(x,w,careful);}
         // changes *this into the convex linear combination of the x

      Spc versus(const Spc& a, R frac)const
         { return Spc(p+(a.p-p)*frac);}
         // returns a point which lies on the line from *this to a.
         // frac==0 gives *this, frac==1 gives a;  for values
         // outside [0,1] the meaning is an obvious generalization of
         // this

      R dis(const Spc& a)const;
         // p1.dis(p2)=(2/cpmpi)arctan((cpmpi/2)*|p1-p2|/refLen)
         // translation invariant function.

      R distance(const Spc& a)const // p1.distance(p2):=|p1-p2|
      {  return (p-a.p).abs();}

      R distSquare(const Spc& a)const // p1.distSquare(p2):=|p1-p2|^2
      {  return (p-a.p).absSqr();}

      static R vol(R r){ return (R(4)/R(3))*cpmpi*r*r*r;}
         // volume of the sphere (radius=r) in the space under
         // consideration.

      static R sur(R r){ return cpmpi*r*r*4;}
         // surface area of the sphere (radius=r) in the space under
         // consideration.

      static R radFrmVol(R vol)
         { return vol<=0 ? R(0) : cpmpow((vol*3)/(cpmpi*4),R(1)/R(3));}
         // radius from volume (of a sphere)

      static R radFrmSur(R ar)
         { return ar<=0 ? R(0) : cpmpow(ar/(cpmpi*4),R(0.5));}
         // radius from surface (of a sphere)

      static Z dimension(void){ return 3;}
         // dimension of the space. Avoids to have to say:
         // Spc x; Z d=x.dim()

      static V<Spc> tet();
         //: tetraeder

      static X4<Spc,R,R,R> meet(Spc const& p1, Vec const& v1,
         Spc const& p2, Vec const& v2);
         //: meet
         // We consider the lines r1 := { p1 + v1*a | a \in R },
         // r2 := { p2 + v2*a | a \in R }. Let res be the return
         // value of the function. Then for pc=res.c1();
         // rho := res.c2(); the sphere Shere(pc,rho) around
         // pc with radius rho is the smallest sphere which containst the
         // two points of closest approach of r1 and r2.
         // With a1:=res.c3(), a2:=res.c4(), these points are
         // q1=p1+v1*a1, q2=p2+v2*a2; thus pc=q1.cen(q2),
         // rho=|q1-q2|/2.
         // The natural situation to which this function applies is that
         // these two lines intersect and we are interested in the
         // crossing point. The second component of the result then allows
         // to assess the precision with which the lines hit each other.
         // In such cases it is natural that only solutions with
         // a1>0, a2>0 are intended ones.
   };

   inline Group::Group(const Spc& x0, const AxVec& A1):A(A1)
   { Vec x0p=x0.p; a=x0p; a-=(x0p*=A);}

   inline Group::Group(Spc const& x0, AxVec const& A1, Vec const& a1):
   A(A1){ Vec x0p=x0.p; a=x0p; a-=(x0p*=A); a+=a1;}

   inline Group::Group(const Spc& x0, const Angle& phi):A(AxVec(phi))
   { Vec x0p=x0.p; a=x0p; a-=(x0p*=A);}

//////////////////////// class RefSys ///////////////////////////////////
// orthonormal, positively oriented coordinate system in Euclidean affine
// 3-space

class RefSys{
//orthonormal, positively oriented reference system in space

      friend class AxVec;
      friend class Vec;
      friend class Spc;

// independent data
      Group g0;
         // describes the transformation which generates RefSys *this
         // from the master system of reference
// dependent data for efficiency
      Group g0Inv;
      Vec v1,v2,v3;
         // triplet of axis unit vectors
      Spc x0;
         // actual origin

      void update();

      static Vec u1,u2,u3;
         // axis triplet of Master system
      static Spc orig;
         // origin of master system

   public:
      typedef RefSys Type;
      CPM_IO
      Word nameOf()const{ return Word("CpmDim3::RefSys");}
      RefSys();
         // creates the master system of reference. All other systema can
         // be created via group actions

      explicit RefSys(const Group& g0_):g0(g0_){update();}

      RefSys(const Spc& x0_, const AxVec& av);
         // construct the RefSys which has x0_ as origin and the axes of
         // which result from the master axes by applying the rotation
         // described by the axial vector av

      explicit RefSys(const V<Spc>& vs);
         // assumption: vs.dim()>=Rdim
         // only the first Rdim components of vs get used.
         // vs[1] is the origin
         // e1:=(vs[2]-vs[1]), normalized is the 1-axis
         // e2:=(component of vs[3]-vs[1] which is orthogonal to e1,
         // normalized is the 2-axis
         // e3:=e1.cross(e2) is the 3-axis
         // If vs is too special for this to work, cpmassert will signal
         // the situation. In case that cpmassert is set not to stop but
         // only to warn (so that the function does not terminate) the
         // default RefSys gets constructed.
         // Implemented in cpmdimxcpp.h.

      RefSys& operator *=(const Group g)
         { g0*=g; update(); return *this;}
         // translation and rotation combined

      RefSys& operator +=(const Vec& a)
         { g0*=Group(a); update(); return *this;}

      RefSys& operator *=(const AxVec& A)
         { g0*=Group(A); update(); return *this;}

      RefSys& rotByEulAng(const Angle& phi, const Angle& theta,
         const Angle& psi);
         // rotate by Euler angles
         // transforms the system by Euler angle rotations referring
         // to system *this, leaving the origin of *this unchanged.

      RefSys& rot(const Angle& phi, Z i);
         //: rotate
         // transforms the system by phi-rotation arround axis
         // i of *this, leaving the origin of *this unchanged.

   //   RefSys& ref_(Plane const& p);
         //: reflect
         // Changes *this in a way that the origin and the three
         // axis directions get reflected by plane p.
         // bad concept: since each RefSys results from the master
         // system by an orientation-preserving transformation,
         // the ref_ transformation has to leave the class RefSys
         // as defined here.

  //    RefSys& ref_(Vec const& n);
         //: reflect , bad concept, see previous function.

      RefSys per(const Word& p)const;
         // permutation
         // p may be one of the following words:
         // "123","231","312","1-32","32-1","-213" which correspond
         // directly to the 6 permutations of the set {1,2,3} if the
         // minus signs are ignored.
         // The return value is a reference system rs  which has the same
         // origin and the same set of axes as *this (up to a reversal of
         // at most one axis, this is indicated by the - sign), but
         // differs in the indexing of these axes:
         // Let any of the allowed p's be written as 'abc'
         // (e.g. a=1, b=-3, c=2 for the fourth entry in the list), then
         // the 1-axis of *this is the a-axis of rs, the 2 axis of *this
         // is the b-axis of rs, and the 3-axis of *this is the c-axis of
         // rs. If abc, as a permutation of {1,2,3}, is a transposition
         // (of negative parity in a more general context) one axis (or 3
         // axes) have to change sign in order for the permutation of
         // axes to be achievable by a rotation in space (notice that we
         // consider only reference systems which originate from the
         // master system by rotations and translations; reflections are
         // excluded).
         // There are much more possible permutations of axes if other
         // positions of the - sign and also three minus signs are taken
         // into account. Here a minimal set is selected that does not
         // need rotations of 180 degrees (which depend on a regularization
         // of a singularity in our parameter space)

// getting the components of objects with respect to the reference system
// *this
      R3 operator()(const Spc&)const;
      R3 operator()(const Vec&)const;
      R3 operator()(const AxVec&)const;
// getting the objects from their components
      Spc operator()(Rdim x)const{ return x0+x[1]*v1+x[2]*v2+x[3]*v3;}
      Vec operator[](Rdim x)const{ return x[1]*v1+x[2]*v2+x[3]*v3;}

      Group frmMasSys()const{ return g0;}  // from master system
         // returns the transformation which creates *this from the
         // master system of reference (which is created by the default
         // constructor) In self-explanatory pseudo-code (pseudo-code
         // elements '...' and '==')
         // RefSys rs=...; Group g=rs.frmMasSys(); RefSys master;
         // master*=g;
         // master==rs;

      AxVec rot()const{ return g0.rot();}
         // evident

      Vec shift()const{ return g0.shift();}
         // evident

      Group toMasSys()const{ return g0Inv;} // to master system
         // returns the transformation which creates the master system of
         // reference from system *this
         // RefSys rs=...; Group g=rs.toMasSys(); RefSys master; rs*=g;
         // master==rs;

      Group operator /(const RefSys&)const;
         // Let r1 and r2 two RefSys's. Then there is a uniquely
         // determined Group instance g such that r2 is obtained from r1
         // by group action:
         // r2 equals r1*=g. We write r2/r1 for this g.
         // ( plausible since then (r1*g)/r1==g )
         // This establishes between RefSys and Group a relation which is
         // analogous to the relation between Spc and Vec which is
         // characteritic for affine geometry. Group g=r2/r1 corresponds
         // to Vec v=x2-x1 for x2+=v for Spc x1,x2. (Vec is a group,
         // wheras Spc is'nt)

      Group toSys(const RefSys& rs)const{ return rs/(*this);}
         // More suggestive (at least at a first glance) notation
         // for the / operator.
         // RefSys rs1=...,rs2=...;
         // Group g=rs1.toSys(rs2);
         // bool b=(rs2==(rs1*=g)); // b==true always

// getting axes and origin associated with the RefSys
      Vec e1()const{ return v1;}
         // e1 and e2 are orthogonal and lie in the RefSys; they
         // determine the orientation of the RefSys

      Vec e2()const{ return v2;}

      Vec e3()const{ return v3;}

      Vec n()const{ return v3;}
         // it is useful to have a dimension-independent name for the
         // 'last vector'

      Spc o()const{ return x0;}

      V<Vec> getVectors()const
         { return V<Vec>{v1,v2,v3};}

      F<Spc,Rdim> toFSR()const;
         // creates a CPM-function which transforms from Spc to
         // coordinates with respect to system *this

      F<Rdim,Spc> toFRS()const;
         // creates a CPM-function which transforms from coordinates
         // (with respect to system *this) to Spc

      Plane pln(Z i)const;
         //: plane
         // rs.pln(i) is the plane Plane(rs.ei(),rs.o()) if i=1,2,3
         // and Plane() else.
   };

//////////////////////// class Plane ///////////////////////////////////

   class Plane{ // oriented affine plane
      RefSys rs;
         // the reference system in which *this is the 12-plane
   public:
      typedef Plane Type;
      CPM_IO
      Word nameOf()const{ return Word("CpmDim3::Plane");}

      Plane();
         // creates the 12-plane of the master system of reference

      explicit Plane(const RefSys& rs_):rs(rs_){}
         // creates a plane, which is the 1-2-plane of RefSys rs_

      Plane(const Vec& n, const Spc& x0);
         // creates a plane containing point x0 and having a normal unit
         // vector parallel to n (irrespective of normalization of n)

      R dist(const Spc& x)const{ return rs(x)[3];}
         // returns the signed distance of x from the plane. Positive
         // distance means positive 3-component with respect to a
         // positively oriented 'Dreibein' the 1-direction and
         // 2-direction of which are in the plane and follow after each
         // other in the orientation of the plane.

      Spc reflect(const Spc& x)const
         { Spc res = x; return res.ref_(*this);}
         // returns the mirror point of x with respect to the plane *this

      Vec reflect(const Vec& x)const
         { Vec res = x; return res.ref_(*this);}
         // returns the mirror image of vector x with respect to the
         // plane *this

      AxVec reflect(const AxVec& x)const
         { AxVec res = x; return res.ref_(*this);}
         // returns the mirror image of axial vector x with respect to
         // the plane *this

      Spc project(const Spc& x)const
         { R3 y=rs(x); y[3]=0; return Spc(y,rs);}
         // returns the projected image of x on the plane *this

      Vec project(const Vec& x)const
         { R3 y=rs(x); y[3]=0; return Vec(y,rs);}
         // returns the projected image of x on the plane *this

      Plane& operator *=(const Group g){ rs*=g; return *this;}
         // translation and rotation combined

      Plane& operator +=(const Vec& a){ rs*=Group(a); return *this;}

      Plane& operator *=(const AxVec& A){ rs*=Group(A); return *this;}

      R3 operator()(const Spc& x){ return rs(x);}

      R3 operator()(const Vec& v){ return rs(v);}

      R3 operator()(const AxVec& av){ return rs(av);}

// getting directions associated with the plane
      Vec e1()const{ return rs.e1();}
         // e1 and e2 are orthogonal and lie in the plane; they determine
         // the orientation of the plane

      Vec e2()const{ return rs.e2();}

      Vec n()const{ return rs.n();}
         // normal direction

      Spc o()const{ return rs.o();}

      RefSys getRefSys()const{ return rs;}

      Word des()const;
         // description as a Word for diagnosis
   };

   class SubsetOfPlane{
// subset of a Plane defined via an indicator function
         // stream I/O makes no sense for this class due to f being not
         // serializable
      RefSys rs;
         // the plane under consideration is the 12-plane of rs
      F<R2,bool> f;
         // f(x1,x2)==true says that a point x with rs-coordinates
         // (x1,x2,0) belongs to the subset to be defined here.

   public:
      Word nameOf()const{ return Word("CpmDim3::SubsetOfPlane");}
      SubsetOfPlane(const RefSys& rs_,const F<R2,bool>& f_):
         rs(rs_),f(f_){}
         // the constructor for a subset the 12-plane of system rs_ such
         // that the rs-coordinates (x1,x2,x3) of its points satisfy
         // just f(x1,x2)==true

      explicit SubsetOfPlane(const RefSys& rs_):rs(rs_),f(true){}
         // this constructor thus creates the whole 12-plane of the
         // system rs_

      SubsetOfPlane():rs(),f(true){}
         // the default constructor thus creates the whole 12-plane of
         // the master system of reference

      X2<R,bool> ind(const Spc& x)const  // indicator
      { R3 y=rs(x); return X2<R,bool>(y[3],f(R2(y[1],y[2])));}
         // returns the distance of x from the plane and the truth value
         // of the 'belonging to set'-relation for the projection of x
         // onto the plane

      bool indLean(const Spc& x)const  // lean indicator
      { R3 y=rs(x); return f(R2(y[1],y[2]));}
         // returns the truth value of the 'belonging to set'-relation
         // for the projection of x onto the plane

      SubsetOfPlane& operator *=(const Group g){ rs*=g; return *this;}
         // translation and rotation combined. Notice that f has not to
         // be transformed since it works on the components in the
         // 'body-fixed' system of reference

      SubsetOfPlane& operator +=(const Vec& a)
         { rs*=Group(a); return *this;}

      SubsetOfPlane& operator *=(const AxVec& A)
         { rs*=Group(A); return *this;}
   };

//////////////////// class InTens////////////////////////////////////////
// inertial tensor, tensor of inertia

   class InTens{ // tensors of inertia
      // Space of symmetric tensors. Data refer to a spectral
      // representation. Here used as Tensors of Inertia.

      R3 eigVal;  // list of eigenvalues
      // |eigVal[1]|>=|eigVal[2]|>=|eigVal[3]|

      RR3 eigVec; // again trying to improve speed by having a
      // pointer-less data type here. Unfortunately the data
      // have to satisfy arithmetic conditions (three orthogonal unit-
      // vectors). When one computes these in standard precision and
      // then uses these values as input of a multiple precision
      // one introduces an error which may be disturbing. Using a
      // RefSys as a data element looked good but finally caused
      // consistency problems since RefSys, since defined as positively
      // oriented cannot be reflected (unless we change either RefSys
      // or -- moresystematically -- Group, by allowing for inversions.
      // These also make technical problems. The full rotation group may
      // be represented as a direct product (sic!) of the group {1,-1}
      // and the proper rotation group. This shows that a reflection
      // is of the form r = (-1,g) with a g satisfying g*g = 1
      // (since 1 = r*r = (-1,g)*(-1,g) = (-1)*(-1)*g*g = g*g )
      // and thus needs the regularization precaution in the definition
      // of the product in the proper rotation group.

   public:
      typedef InTens Type;
      CPM_IO
      Word nameOf()const{ return Word("CpmDim3::InTens");}

      InTens():eigVec()
         // standard orthonormal basis, eigenvalues all 1
      {  eigVec[1][1]=1; eigVec[2][2]=1;
         eigVec[3][3]=1; eigVal[1]=1; eigVal[2]=1; eigVal[3]=1;
      }

      explicit InTens(R3 v):eigVec()
         // standard orthonormal basis, eigenvalues given by v
         {
            eigVec[1][1]=1; eigVec[2][2]=1; eigVec[3][3]=1;
            eigVal[1]=v[1]; eigVal[2]=v[2]; eigVal[3]=v[3];
         }

      InTens(RR3 const& );
         // construction from a matrix representation, the input matrix
         // will not be taken as it is, but a symmetrized version of it
         // will be created. So the most obvious consistency condition
         // for a inertia matrix will be enforced. Other conditions of
         // elements of a tensor of inertia (e.g. Huette(Springer,1991),
         // E31, between (9) and (10), will not be considered.

      AxVec operator * (const AxVec& a )const;
         // inertia tensors transform axial vectors (eg. applied to
         // angular velocity to give angular momentum). Since we know the
         // eigen vectors and -values this can be expressed as a sum of
         // dyades

      void balance();
         // changes the inertia tensor *this into one which has equal
         // eigenvalues and the Euclidean length of eigVal remains the
         // same. This is useful for plausibility checks

      InTens& operator *=(const AxVec& a );
         // The definition is such that (I*a)*(x*a)=(I*x)*a

      InTens adjointAction(AxVec const& a)const
         // non-mutating version of operator *=(const AxVec& a )
      { InTens res=*this; return res*=a;}

      InTens& operator *=(const Group& g )
         // g acts only through the rotation component
      { return operator*=(g.rot());}

      void scale(R fac){eigVal*=fac;}
         // multiplying the eigenvalues by a common factor fac. No
         // dimension reflecting power of fac is taken. If, for instance,
         // the material system described by *this is geometrically
         // enlarged by a (linear) factor y and the density remains
         // constant, then fac=pow(y,5)

      R3 volume(R m)const;
         // Let res be the return value. Then res[1], res[2], res[3] are
         // the sites of a homogeneous Quader of total mass m
         // which has the inertial tensor *this (res[i] is along axis
         // eigVec[i] i=1,2,3)

      InTens operator !()const;
         // inverse matrix, here implemented by inverting the eigen
         // values

      InTens& ref_(Vec const& n);
         //: reflection
         // Changes *this into the reflection-transformed tensor.
         // This is the tensor of inertia of the reflected body.
         // (where the plane of reflection contains the center-of-mass
         // of the body)

      InTens& ref_(Plane const& p){ return ref_(p.n());}
         //: reflection

      InTens ref(Vec const& n)const
         //: reflection
      { InTens res(*this); return res.ref_(n);}

      Vec e1()const{ return Vec(eigVec.x1);}
      Vec e2()const{ return Vec(eigVec.x2);}
      Vec e3()const{ return Vec(eigVec.x3);}

      R i1()const{ return eigVal.x1;}
      R i2()const{ return eigVal.x2;}
      R i3()const{ return eigVal.x3;}
   };

   InTens rotate(const InTens&, const AxVec& );
