//? cpmdimx.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*

    Description:
      declares and partly defines those geometricl objects which,
      though depending on dimension (as Spc, Vec ... do) but can
      be coded in a way not depending essentially on dimension

*/
using CpmFunctions::FncObj;
using CpmFunctions::F1;
using CpmFunctions::F2;

using CpmArrays::X2;
using CpmArrays::X3;
using CpmArrays::VV;

using CpmGraphics::VecGraph;

//doc namespace CpmDimx{

//////////////////////// function grad ///////////////////////////

   Vec grad(const Spc& x, const F<Spc,R>& f, const R& h);
   // gradient of a scalar function as a vector-valued function of
   // position. h is the distance used for finite differencing

///////////////////////// class BndSph //////////////////////////

   namespace{ // technicalities
      const Z nFindDef=500;
         // default for the number of points to find in a
         // SubsetOfSpace for basing an estimation of volume or diameter
         // on
      const Z nMaxDef=10000;
         // default for the maximum number of points to try
         // to find points in a SubsetOfSpace
   }

   class BndSph{ // BndSph = Bounding Sphere
   // an abstraction from a specific tool that helped to implement
   // effectively the function hit in class SubsetOfSpace

   Spc x;
      // center of a sphere
   R r;
      // radius of a sphere

public:

   Spc get_x()const{ return x;}
   R get_r()const{ return r;}

   BndSph(Spc x_,R r_):x(x_),r(r_){}
   BndSph():r(0){}
   BndSph operator|(const BndSph& bs)const;
      // minimum sphere which contains the set union of
      // *this and bs
   BndSph operator&(const BndSph& bs)const;
      // minimum sphere which contains the set intersection
      // *this and bs
   Iv hit(const Spc& x, const Vec& v)const;
      // {x+a*v| a in Iv} is the intersection of *this with the
      // ray { x+a*v| a>=0}
      // v has to be a unit vector

   X2<V<Spc>,Z> ran(Z n, const F<Spc,bool>& cond,
      Z tryMax=nMaxDef)const;

   X2<V<Spc>,Z> ran(Z n, Z tryMax=nMaxDef)const
   {
      F<Spc,bool> cond(true);
     // return ran(n,cond,nMaxDef);
      return ran(n,cond,tryMax);
         // changed 2006-06-22 old form seems to be nonsense
   }
      // returns a random sequence of n uniformely distributed
      // points in the interior of the bounding sphere.
      // The second argument specifies a condition that is guaranteed to
      // evaluate to true for the points of the returned list.
      // Further we return the number of points which were tried.
      // we never try more than tryMax points. If after trying tryMax
      // points we did not find n valid points for the return list, we
      // return the (shorter) list of the valid points found so far.
      // Unfortunately the gnu compiler does not understand the natural
      // version with default arguments
      //  X2<V<Spc>,Z> ran(Z n, const F<Spc,bool>& cond=
      //    F<Spc,bool>(true), Z tryMax=nMaxDef)const;

   X2<Z,Z> ranLean(Z n, const F<Spc,bool>& cond,
      Z tryMax=nMaxDef)const;

   X2<Z,Z> ranLean(Z n, Z tryMax=nMaxDef)const
   {
      F<Spc,bool> cond(true);
     // return ranLean(n,cond,nMaxDef);
      return ranLean(n,cond,tryMax);
          // changed 2006-06-22 old form seems to be nonsense
   }
      // similar to function ran. Let (v,n) be the return value
      // of ran, then (v.dim(),n) is the return value of the present
      // function.

   R vol()const{ return Spc::vol(r);}
};

//////////////// struct DisPar ////////////////////////////////////////

   struct DisPar{  // discretization parameters
      // the discretization parameters need not be set
      // when Camera::show(const SubsetOfSpace&,...) is
      // used for visualization. Quantities corresponding to
      // ds and dr are set from the pixel size of the camera
      R1 dr;
         // small length compared to the whole object, e.g.
         // used as step length in numerical differentiation
      R1 ds;
         // characteristing length for sampling tasks
      Z1 nFind;
      Z1 nMax;
      void ini(){nFind=nFindDef; nMax=nMaxDef;}
      typedef DisPar Type;
      CPM_IO
      DisPar(R dr_, R ds_):dr(dr_),ds(ds_){ini();}
      DisPar(){ini();}
   };

///////////////////////// class SubsetOfSpace //////////////////////////

   class SubsetOfSpace{ // subsets of space with boolean operations
   // (union, section, ... of sets)
   // Visualisation by Camera::show(const SubsetOfSpace&,...)
   // and via
   // Scalarfield f(aSubsetOfSpace);
   // f.mark(aCamera)
   // (this is much faster and allows to locate even objects which
   // may not be found by Camera::show(const SubsetOfSpace&,...)).
   // See also cpmshapesx.h for derived classes that have to be
   // visalized by just the same means.

   F<Spc,R> f;
      // f(x)<0 says that a point x belongs to the interior of the subset
      // to be defined here. f(x)==0 characterizes the surface
      // and f(x)>0 the exterior. In some algorithms it helps if
      // |f(x)| is close to the distance from the surface
      // at least for points close to the surface.
      // If this is the case we say that f is 'locally metric'.
      // In the present implementation f is a dependent quantity
      // which is set by function update() from the independent data
      // fr and rs. In a previous version, fr and rs were not present
      // and the following outdated comment gave the motivation for
      // this: It is reproduced here in order to prevent me from falling
      // into this trap again:
      // 'Notice that here it makes little sense to replace this f by a
      // RefSys and a F<R3,R> since the boolean combinations of two
      // SubsetOfSpace's having different RefSys's would be unnatural to
      // define'.
      // Now a natural definition has been found (see implementation
      // of the boolean combinations). Introducing rs on this
      // level has the benefit that operator *=(const Group&) can be
      // implemented in a way that consecutive applications of *= don't
      // grow in execution time.
      // Although it is standard to describe the surface of a smooth
      // point set by a equation f(x)=0, the extension of this
      // description -as given here- seems not to be in use in
      // computational geometry.
      // Among the strong points of this description is the elegant
      // implementation of boolean ('set theoretical') combinations
      // of point sets (see implementation of operators &,|,-, and ^).

   F<Rdim,R> fr;
      // f,fr, and rs are interrelated by f(x)=fr(rs(x)) for
      // all x of type Spc. (Notice that rs(x) is of type Rdim for all
      // x of type Spc). Thus fr describes the subset of space as a
      // subset of coordinate tuples referring to the actual
      // system of reference rs

   RefSys rs;
      // system of reference to which fr refers. On this attribute an
      // efficient implementation of transformations by Group
      // (see operator *=(const Group&)) is based.

   Spc x;
      // x and r belong together: Ideally BndSph(x,r) is a bounding
      // sphere for *this. The value of x and r has an influence
      // only on the result of the following functions:
      //    R improveBoundingSphere(...)
      //    R vol()const
      //    V<Spc> ran()const
      //    X2<Spc,Z> hitFast(...)const
      // In my present applications by far the most important effect
      // is made by the last function on which all my visualization
      // methods for SubseOfSpace's is based. This implies that
      // only the part of the point set which is in BndSph(x,r)
      // is visible for these methods. Since it might be useful
      // to have only a part of a point set in 'visual scope'
      // an instance of SubsetOfSpace is not considered corrupted
      // if BndSph(x,r) is not a bounding sphere for *this.
      // Therefore we allow a constructor to be public which does not
      // care about the relation between f and BndSph(x,r).
      // Also, we would like to define unbounded sets like half-spaces
      // or infinite plates for which BndSph only makes sense as
      // a visual scope.

   R1 r;
     // see x

   DisPar dp;
      // DisPar = discretizing parameters
      // data helping algorithms that need to discretize
      // *this in some way

   void update();
      // adjusts f to rs and fr

public:
   typedef SubsetOfSpace Type;
   virtual SubsetOfSpace* clone()const
      { return new SubsetOfSpace(*this);}

   CPM_O_V
      // no stream input unless we have a constructor from streamable
      // date as in the derived classes defines in cpmshapesx.h

   virtual Word nameOf()const{ return Word("SubsetOfSpace");}

// constructors

   SubsetOfSpace():f(F<Spc,R>(1)),fr(F<Rdim,R>(1)){}
         // constructor of the void subset of Spc

explicit SubsetOfSpace(const F<Rdim,R>& fr_, const RefSys& rs_=RefSys(),
      const Spc& x_=Spc(), R r_=0, const DisPar& dp_=DisPar()):
         fr(fr_),rs(rs_),x(x_),r(r_),dp(dp_)
   { update();}
      // Defining the object by prescribing fr.
      // Does not make sure that x_,r_ constitutes
      // a bounding sphere of the set described by f_
      // see explanation of x for reasons that may legitimate
      // to make such a disorganized function public.

explicit SubsetOfSpace(const F<Spc,R>& f_, const RefSys& rs_=RefSys(),
      const Spc& x_=Spc(), R r_=0, const DisPar& dp_=DisPar());
      // Defining the object by prescribing f.
      // The further comment of the previous function applies.

// access functions

   void setShape(const F<Rdim,R>& fr_, const RefSys& rs_=RefSys())
   // tool for defining f,fr,rs consistently.
   // Notice that x and r may not match the shape generated
   // by these settings. See comment to previous function
   {
      fr=fr_;
      rs=rs_;
      update();
   }

   void set_x(const Spc& x_){ x=x_;}
   void set_x(const Rdim& xr){ x=rs(xr);}
   void set_r(R r_){ r=r_;}
      // see comment to attribute x for reasons that allow this to be
      // public

   void set_dr(R dr){ dp.dr=dr;}
   void set_ds(R ds){ dp.ds=ds;}
   void set_nFind(Z nF){dp.nFind=nF;}
   void set_nMax(Z nM){dp.nMax=nM;}
   void set_dp(const DisPar& dp_){ dp=dp_;}

   Spc get_x()const{ return x;}
   R get_r()const{ return r;}
   RefSys getRefSys()const{ return rs;}

   Group frmMasSys()const{ return rs.frmMasSys();}
   Group toMasSys()const{ return rs.toMasSys();}

   R get_dr()const{ return dp.dr;}
   R get_ds()const{ return dp.ds;}
      // only usage so far in
      // void Camera::mark(const SubsetOfSpace& sos,
      // SitRec k, const Color& c)const
   Z get_nFind()const{ return dp.nFind;}
   Z get_nMax()const{ return dp.nMax;}
   DisPar get_dp()const{ return dp;}

   bool ind(const Spc& x)const{ return f(x)<=0;}  // indicator
   R operator()(const Spc& x)const{ return f(x);}

   Rdim refSys(const Spc& x)const{ return rs(x);}
   Spc refSys(const Rdim& x)const{ return rs(x);}
      // notice that there is no automatic conversion Rdim --> Spc
      // since the corresponding constructor is explicit

   R improveBoundingSphere(Z n=nFindDef, Z tryMax=nMaxDef);
      // A series of n randomly selected space points of *this will be
      // used to get an improved estimate for the bounding sphere.
      // As an illustration for the need of such an operation consider
      // two parallel long cylinders which are close together but
      // don't intersect. The bounding sphere computed for the
      // intersection will have a diameter similar to the length of the
      // cylinders, wheras the intersection is void, so that the radius
      // of the minimal bounding sphere is zero. The present function
      // will give radius zero.
      // It will, however, give radius zero also in cases where the set
      // fills only a very small fraction of the un-improved bounding
      // sphere.
      // The return value is the fraction of tried points that was found
      // to belong to *this. This is helpful as an estimate for the
      // volume of the set *this.
      // In order to avoid infinit work in case of void sets (which
      // easily result, for instance, from forming intersections of sets)
      // we will try not more than tryMax points

   virtual R vol()const;
      // estimates the volume of the set by testing nFind random points
      // in the bounding sphere for the property to belong to the subset
      // or not. If we don't find nFind such points after having tried
      // nMax points, we base the estimate on what we have got so far.

   virtual V<Spc> ran()const;
      // random sequence of nFind points in the subset; if
      // we don't find nFind such points after having tried
      // nMax points, we return what we have got so far.

   Plane tangentialPlane(const Spc& x)const;
      // returns the tangential plane to the boundary near x

   X2<R,Vec> distAndDirection(const Spc& x)const;
      // The first component of the return pair is the distance of x from
      // the surface (positive for points outside the point set described
      // by *this; needs not to be correct for points far away from the
      // set).
      // The second component is the outwards-directed normal unit
      // vector.
      // Thus, denoting the returned pair as (d,n), the surface point
      // nearest to x is approximately x-n*d. This does not assume that
      // f is locally metric. The step size for numerically computing
      // grad f is get_dr().

   void shiftExternalPointToSurface4(Spc& x, R& dOut,
      R& fxBefore, R& fxAfter)const;
      // Remark to a new notational device:
      // if a function name ends in a number n this means that the first
      // n arguments are non-constant references and that the intended
      // usage is to change the values of those (using them effectively
      // as 'return'-values).
      // If the function is mutating itself, I use a _ directly after the
      // letter part of the name. So 'X::some_3' would be the
      // name of a mutating member function of class with n>=3 arguments
      // the first 3 of which are mutable.
      //
      // Consider f(x). If this is <=0 we have an non-external point
      // so that x will not be changed by function call and
      // it will be set fxBefore=fxAfter=f(x) and dOut=0
      // (dOut=distance by which the point is 'outside').
      // If f(x)>0, after function call x is very close to the surface of
      // *this if it was not too far away before.
      // The second and third argument give f(x) for x before and after
      // function call.
      // dOut is the (xBefore-xAfter).abs() and thus a good
      // approximation to the distance of the outside point x to the
      // surface of *this.

   BndSph toBndSph()const{ return BndSph(x,r);}
      // returning x and r as a BndSph

// boolean operations

   SubsetOfSpace operator &(const SubsetOfSpace& sos)const;
      // forming the intersection of *this and sos (corresponds to the
      // 'and' operation for indicator functions)
      // Notice (sos1&sos2)(x)=max(sos1(x),sos2(x))

   SubsetOfSpace operator |(const SubsetOfSpace& sos)const;
      // forming the union of *this and s (corresponds to the
      // 'or' operation for indicator functions)
      // Notice (sos1|sos2)(x)=min(sos1(x),sos2(x))

   SubsetOfSpace operator -(const SubsetOfSpace& sos)const;
      // forming *this\sos i.e. the set of all those elements of
      // *this which don't belong to sos (relative complement)

   SubsetOfSpace operator ^(const SubsetOfSpace& sos)const;
      // (*this-sos)|(sos-*this) i.e.
      // the symmetric difference of *this and sos
      // (corresponds to 'exclusive or' for indicator functions)
      // a direct definition as for the previous operations is
      // slightly more effective so that we do not
      // simply say:
      // { return (*this-sos)|(sos-*this);}

   friend  SubsetOfSpace unionOfArray(const V<SubsetOfSpace>& vs);
      // returns the union of all vs[i]

   friend  SubsetOfSpace sectionOfArray(const V<SubsetOfSpace>& vs);
      // returns the intersection of all vs[i]

    SubsetOfSpace combine(const V<SubsetOfSpace>& vs)const;
      // here we consider *this as a region of interest and return a
      // body which is formed out of all intersections vs[i]&*this
      // Those terms for which this intersection is void, are left out of
      // the result for efficiency.

// transformations
// Notice that derived classes will have additional
// members to define the shape by serializable data. These will have to
// become included into the transformation. Since the MS compiler does
// not allow redefinition of virtual functions in a way that the return
// value is a reference to a derived class, we have to deviate from
// established customs to define transformations as returning a reference
// to the transformed object. This is a minor inconvenience if any.

   virtual void operator *=(const Group& g);
      // translation and rotation combined

   void operator *=(const F<Spc,Spc>& f);
      // If y is the generic point belonging to
      // *this before the action, then f^(-1)(y) is the generic point
      // of *this after the function call.
      // No change of rs and of x,r.
      // For x the natural change would be to f^(-1)(x).
      // change of r would be more complicated.
      // Note that x,r define only the visual scope and not
      // the set itself. So this is no problem.
      // So a perfect implementation is only possible if
      // f and f^(-1) are given on input.
      // Notice the strange situation: we need f for the implementation
      // and f^(-1) for a description of the meaning (semantics)
      // Notice also that action of Group hides this problem since
      // for g also g^(-1) is available

   void operator +=(const Vec& a){ *this*=Group(a);}
      // translation

   void rot(const AxVec& A){ *this*=Group(x,A);}
      // not defined as *= since overloading of virtual functions creates
      // unexpected (though correct) name hiding phenomena in derived
      // classes rotation arround the central point x

   void setCenter(const Spc& xNew){ operator+=(xNew-x);}
      // bringing the center into the new position by translation
      // Should set_x become replaced by this function ???
      // no !!!! Making this replacement caused severe malfunctions in
      // the padyan program (detected 2002-2-12, the bad change was made
      // not earlier than 2 weeks ago).

   void setRefSys(const RefSys& rsNew)
 //     { operator*=(rs.toSys(rsNew));}
   {
      rs=rsNew;
      x=rs.o();
      update();
   }
      // moving *this such that rs equals the input RefSys rsNew.

   Group toSys(const RefSys& rsNew)const{ return rs.toSys(rsNew);}

   F<Spc,R> getF()const{ return f;}

   void rotByEulAng(const Angle& phi,const Angle& theta,const Angle& psi)
   { rs.rotByEulAng(phi,theta,psi); update();}
      // rotation by Euler angles (see Goldstein: Klassische Mechanik,
      // Chapter 4-4, p. 119)

   X2<Spc,bool> hit(Spc const& xi, Vec const& v,
      R sFine, R sCoarse)const;
      // Let the return value be written as (r1,r2). Then:
      // If r2==true (regular case), the point r1 has a distance from
      // the surface which is not larger than sFine.
      // If r2==false, no indication of a body surface was found.
      // No assumption is made on whether xi is outside the body or
      // inside.
      // Improved method, based on a two-level analysis of the function
      // t |--> f(xi+t*v) , t>=0.
      // The first step scans this function at step size sCoarse. So
      // detail much finer than this value will not be detected.
      // v needs to be a unit vector !!!
      // Only the part of the body within the assigned bounding
      // sphere is considered for hitting.

// Static generating function (more convenient than a constructor.
// Notice that cpmshapes.h defines also special instance by
// defining derived classes with constructors. So the names of
// these derived classes provide the intended meaning that here
// the function name indicates. I'm not sure which (if at all) of the
// two mechanisms is superior.

   static SubsetOfSpace band(Z i, const Iv& iv,
      const RefSys& rf=RefSys());
      // returns the 'set of points' x \in Spc such that
      // iv.contains(rf(x)[i])==true
      // Notice that a box can be formed as a cross section of such
      // bands

   static SubsetOfSpace boxoid(const Rdim& x0, const Rdim& s0,
      const F<Spc,Rdim>& f );
      // Consider X=
      // {x \in Rdim | x0[i]-s0[i]<=x[i]<=x0[i]+s0[i], i=1,...dim}
      // The SubsetOfSpace returned by the present function is
      // f^-1(X). If f is isometric, the radius of this set is simply
      // given by s0.abs(). The problem is, that f -as used here-
      // does not carry a tag saying 'I am isometric'. We thus
      // have to make tests, and set r accordingly.

   static SubsetOfSpace sphere(Spc const& x0, R r0, R accRel=1e-3);
      // r0*accRel will be the typical discretization length

   static R box(const Rdim& x, const Rdim& x0, const Rdim& s0);
      // auxiliar function quantity
      // indicator function of the the box x0[i]-s0[i]<=x[i]<=x0[i]+s0[i]
      // Thus s0[i] is  h a l f  the interval length in i-direction
};

///////////////////////// class Texture ///////////////////////////////
// tool for 2,3-dimensional visualization

class Texture: public CpmFields::ScalarField<Spc,R,Group>{// textures
   // method to assign a periodic pattern to space. This is essential for
   // enabling anaglyphic visualization of arbitrary shapes.
   // By deriving from ScalarField we inherit linear arithmetics and
   // operator *=(Group)
public:
   typedef CpmFields::ScalarField<Spc,R,Group> Base;
   typedef Texture Type;
   virtual F<Spc,R>* clone()const{ return new Texture(*this);}

explicit Texture(const V<R>& omega, const V<R>& delta,
   R p=1, Z method=1);
   // The return value of operator()(const Spc& sp) is a function of
   // yy[i]=0.5*(1+cos(delta[i]+sp[i]*omega[i])), i=1,...Spc.dimension().
   // From this list of numbers one forms the final value of
   // operator()(const Spc& sp) by expressions which are controlled
   // by method and p. p matters only for method 2.
   // method=1: 'additive method' (see code in cpmdimxcpp.h)
   // method=2: 'extended additive method'
   // method= not 1 or 2: 'multiplicative method'

explicit Texture(R width=0.001);
   // Special case of previous constructor. All components of delta are
   // 0, and those of omega are 2*pi/width.

Texture(const VV<R>& mat, Iv xI, Iv yI, Z method=1);
   // Defines a texture on the 1-2-plane of RefSys() by repeating
   // the rectangle xI x YI and filling each rectangle with the
   // 'image matrix' mat. The argument method controls the extension
   // of mat from Z x Z to R x R. method=0 means pixel-wise
   // representation
   // method=1 is a fast simplex interpolation. Other values give error.
   // For this application, the values of mat[i][j] should belong to
   // the interval [0,1]

explicit Texture(const F<Spc,R>& fIn):Base(fIn){}
   // flexible definition from any function of the right type

explicit Texture(const F<C,R>& fIn);
   // Uses the function fIn in order to define the pattern in the
   // 1-2-plane. If Spc::dimension()==3, the pattern extends into
   // the 3-direction in the trivial manner that it does not depend
   // on the 3-coordinate

R eval(const RefSys& rf,const Spc& sp)const
   // Here the value is computed under the assumtion that the
   // pattern is fixed with respect to a system of reference
   // given as the first argument. It would be nice to have the
   // name operator()(arg1,arg2) for this function. Unfortunately the
   // name operator() declared in this class would hide the
   // most important operator() of the base class. This would be
   // more inconvenient than introducing a verbal name here.
{
   Group g0=rf.toMasSys(); Spc x=sp;
   return operator()(x*=g0);
}

};

///////////////////////// class Triangle /////////////////////////////////

class Triangle{ // triangle, few methods
   Spc p1_, p2_, p3_;
   typedef Triangle Type;
public:
   CPM_IO
   Triangle():p1_(),p2_(),p3_(){}
   Triangle(Spc const& p1, Spc const& p2, Spc const& p3):
      p1_(p1),p2_(p2),p3_(p3){}
   Word nameOf()const{ return "Triangle";}
   static Triangle xArrowHead(R l, Spc p=Spc(), R slimness=3)
      // slimness is here the ratio hc/c, see hc() for notation.
   {
      R h = l/(slimness*2);
      Vec v1(0,h);
      Vec v2(0,-h);
      Vec v3(l,0);
      return Triangle(p+v1,p+v2,p+v3);
   }

   static Triangle yArrowHead(R l, Spc p=Spc(), R slimness=3 )
      // slimness is here the ratio hc/c, see hc() for notation.
   {
      R h = l/(slimness*2);
      Vec v1(h,0);
      Vec v2(-h,0);
      Vec v3(0,l);
      return Triangle(p+v1,p+v2,p+v3);
   }

   Spc p1()const{ return p1_;}
   Spc p2()const{ return p2_;}
   Spc p3()const{ return p3_;}
   R hc()const{ return (p1_.cen(p2_)-p3_).abs();}
      // Assume a notation as in school geometry:
      // p1_ = A, p2_ = B, p3_ = C
      // and corresponding sides a,b,c. Then hc is the 'height' from C
      // onto c.
   R c()const{ return (p1_-p2_).abs();}
      // side c, see hc
   Triangle& operator *=(Group const& g);
      // translation and rotation combined
   Triangle& operator +=(Vec const& a);
      // translation
   Triangle& operator *=(AxVec const& A)
      { return operator*=(Group(A));}
      // rotation
   Triangle& scl_(Spc const& fix, R fac)
      //: scale
   {
      p1_.scl_(fix,fac);
      p2_.scl_(fix,fac);
      p3_.scl_(fix,fac);
      return *this;
   }
};

//////////////////////// class Path //////////////////////////////////////

class Path{ // array of points as a discrete representation of a curve
   // array of points considered as following one after the other
   // just according to their indexing. Lean structure that allows
   // the natural definition of operations that make sense for
   // curves but not for general graphs.
   // Correctly formed path satisfy (x_[i+1]-x_[i]).abs()>acc_
   // for all i \in {1,...n_-1}.
   // 2003-11-11: adding the dependent quantities y_,e_,l_
   // that allow a very efficient implementation of the relevant
   // algorithms.
   // Visalization works through class Skeleton. See constructor
   // Skeleton(Path const&).

   V<Spc> x_;
      // list of points
   Vo<R> y_;
      // y_[i] = arc length (i.e. length along the path *this
      // from x_[1] to y_[i]
   V<Vec> e_;
      // e_[i]= unit vector from x_[i] to x_[i+1]
   Vo<R> l_;
      // l_[i]=length of vector x_[i+1]-x_[i]
   R1 length_;
      // total length
   Z1 n_;
      // number of points
   R1 acc_;
      // accuracy: points that are closer to each other than
      // this are considered not as different and one of them will
      // be purged

   Z adjust(R acc);
      // eliminates repetition points (acc determines what these are
      // precisely) and sets n_ and length_ accordingly.
      // Returns the number of points that had to be eliminated.
      // if the argument is 0, then the real accuracy is set
      // according to a specified fraction of the shortest inter-
      // point distance

   void autoAcc(){ if (n_>1) acc_=R1((length_.x1/n_.x1)*1e-3);}
      // sets the accuracy reasonably (not according
      // to all existing points in x_, since x_ may already be spoilt
      // by adding points that differ from intended ones merely by
      // numerical noise

   R update();
      // adjusts the values of n_,length_,y_,l_,e_ according to x_
      // returns the minimum of the l_[i]

   typedef Path Type;

public:
   CPM_IO
   explicit Path():x_(0){}
   Path(V<Spc> const& vsp, R acc=0):x_(vsp)
      // polygon path from list of points
   {adjust(acc);cpmassert(isSmooth(),"Path(V<Spc>)");}

   Path(Spc p1, Spc p2, R acc=0):x_(V<Spc>{p1,p2})
      // line as a Path
   {adjust(acc);}

   Path(Spc p1, Spc p2, Spc p3, R acc=0):x_(V<Spc>{p1,p2,p3})
      // polygon path from three points
   {adjust(acc);cpmassert(isSmooth(),"Path(V<Spc>)");}

   Path(Spc p1, Spc p2, Spc p3, Spc p4, R acc=0):
      x_(V<Spc>{p1,p2,p3,p4})
      // polygon path from four points
      {adjust(acc);cpmassert(isSmooth(),"Path(V<Spc>)");}

   Path(Spc p1, Spc p2, Spc p3, Spc p4, Spc p5, R acc=0):
      x_(V<Spc>{p1,p2,p3,p4,p5})
      // polygon path from five points
      {adjust(acc);cpmassert(isSmooth(),"Path(V<Spc>)");}
   // Path(p1,p2,p3,p4,p1) is a closed box for suitably chosen
   // pi.

   Path(Spc p1, Spc p2, Spc p3, Spc p4, Spc p5, Spc p6, R acc=0):
      x_(V<Spc>{p1,p2,p3,p4,p5,p6})
      // polygon path from six points
      {adjust(acc);cpmassert(isSmooth(),"Path(V<Spc>)");}

   Path(Spc const& p1, Vec const& v1, Spc const& p2,
      Angle const& phi, Z n=10);
      // Constructs a polygon Path of n points, starting at point p1
      // in direction v1 and ends at p2 with direction v1*av
      // where av is the AxVec has a rotation angle and a
      // rotation axis perpendicular to v and p2-p1.
      // Based on a three point Bezier curve.

   Path(Spc const& p1, Vec const& v1, Spc const& p2,
      Vec const& v2, Z n=10);
      // Constructs a polygon Path of n points, starting at point p1
      // in direction v1 and ending at p2 with direction v2.
      // Actually it is the three point Bezier curve defined by
      // p1,pc,p2 where the control point pc is
      // taken as the point of closest approach of the lines
      // (p1,v1) and (p2,v2). If there is no plane in which
      // p1,p2,v1,v2 lie (and thus also pc) the actual initial and final
      // directions, pc-p1, p2-pc will deviate from the intended
      // directions v1 and v2.

   Path(Spc const& p1, Spc const& p2,
      Vec const& v2, R s=0.25, Z n=10);
      // Constructs a polygon Path of n points , starting at point p1
      // and ending at p2 with direction v2. The parameter s should
      // be selected in [0,1]. For small s, the initial direction
      // points nearly to p2 and the requested final direction will be
      // achieved by a narrow curve. For s close to 1, the initial
      // direction and the final direction are close to perpendicular
      // to each other and the path is as constanly curved as
      // possible.

   Z dim()const{ return n_;}
// non-mutating access to the points
   R length()const{ return length_;}
   // old name
   R len()const{ return length_;}
   //: length
   Spc firstPoint()const{ return x_[1];}
   Spc lastPoint()const{ return x_[n_];}
   Spc operator[](Z i)const{ return x_[i];}
      // returning the points
// non-mutating access to directions
   Vec firstDir()const;
      // first direction,
      // does not return unit vectors
      // returns the zero-vector if no reasonable
      // direction can be given
   Vec lastDir()const;
      // last direction

   Vec dir(Z i)const;
      //: direction
      // direction at point i where this has a meaning and Vec() else

   R acc()const{ return acc_;}
      //: accuracy

   void setAcc(R acc){ acc_=CpmRootX::sup<R>(0.,acc);adjust(acc_);}
      //: set accuracy

   Path& operator *=(Group const& g);
      // translation and rotation combined
   Path& operator +=(Vec const& a);
      // translation
   Path& operator *=(AxVec const& A);
      // rotation arround the first point
   Path& operator *=(R fac);
      // scaling relative to the first point

   Path& rot_(Spc const& fix, AxVec const& A);

   Path scl(Spc const& fix, R fac)const;
      //: scale
      // Returns a path resulting from *this by
      // multiplying all position vectors relative to
      // the origin 'fix' by factor fac.
      // Notice that having the path's first point as fixed
      // point (as in operator*=(R)) is not always what is needed.
   Path& scl_(Spc const& fix, R fac){ return *this=scl(fix,fac);}

   V<Path> thc(R r, Z n)const;
      //: thick
      // This function returns n shifted
      // copies, where the n shift vectors have length r and
      // result from equispaced points on a circle (not disk!).
      // This is a brut force method to simulate a line thickness
      // of 2*r. Works fine. But the suitable n has to be found by
      // experimentation for given r. If n is too small, there remain
      // voids, if it is too large, the function takes too much time.

   V<Path> brkLine(R d, R frac=0.5)const;
      //: broken line
      // Returned is a list of Paths which result from *this by leaving
      // out a fraction of relative length frac of the periodicity
      // element d. The actual periodicity element may differ a bit
      // from d since a minimum adjustment is done which lets the
      // end points of *this always are the end of a visible line piece.

   Path rev()const{ return Path(x_.rev(),acc_);}

   Path operator()(const F<Spc,Spc>& g)const
   { return Path(x_(g),0);}
      // transforming the components by a function
      // Since the function may span orders of magnitude
      // the accuracy has to be set from x_(g) which is
      // triggered by the 0-argument

// generation functions as static functions (constructors
// can't have a descriptive name, so named functions
// are more convenient to use).

   static Path arc(const Spc& p1, const Spc& center,
      const AxVec& rot, Z n=10);
      // circular arc obtained by applying a rotation rot arround
      // center in n steps on point p1.

   static Path arc2(const Spc& p1, const Spc& aimPoint,
      const AxVec& rot, Z n=10);
      // circular arc obtained by deviating from initial direction
      // p1 --> aimPoint, having arc length |aimPoint-p1| and
      // lies in the plane containing p1 and aimPoint and being
      // perpendicular to rot.

   static Path line(const Spc& pi, const Spc& pf, Z n=10);
      // straight line starting at pi and ending at pf
      // represented as a chain of n equidistant connected
      // points. (n-1 edges). cpmassert for n>1.

   static V<Path> brkLine(const Spc& pi, const Spc& pf, R d, R frac=0.5)
      //: broken line
      // straight line starting at pi and ending at pf.
      // Line is broken according to the control mechanism explained
      // for the other overload of brkLine
   {
      Path temp=Path::line(pi,pf,2);
      return temp.brkLine(d,frac);
   }

   static Path parabola(const Spc& pi, const Spc& pc, const Spc& pf,
      Z n=10);
      // parabola with initial point pi, final point pf, and
      // Bezier control point pc. The intermediate points
      // correspont to equidistant positions in time, where the time span
      // needed for the whole path is 1. The meaning of the control
      // point is to express the initial velocity as (pc-pi)/(1/2)
      // and the final velocity as (pf-pc)/(1/2).
      // cpmassert for n>2.

   static Path ellipse(const Spc& pc, R a1, R a2, Z n=16);
      // generates an ellipse in the 1-2-plane around point pc
      // approximated by a closed flattened n-polygon.
      // We start from a circle
      // of radius a1 and multiply the 2-coordinates by a2/a1
      // (everything relativ to the master system of reference).
      // half-axis a1 in 1-direction
      // half-axis a2 in 2-direction
      // For a1==a2 (circle) more efficient code is generated by
      // avoiding the flattening loop.

   static Path quarterEllipse(Spc const& p1, Vec const& v1,
      Spc const& p2, Z n=16);
      // generates an elliptic quadrant line, approximated as a
      // polygon of n+1 points.
      // The eliptic arc starts in p1 at direction v1 and ends in p2 with
      // a direction perpendicular to v1. The direction from the
      // beginning to the end of the arc varies monotoneously from
      // the initial direction di to the final direction
      // di +/- 90 degrees. If p2 does not lie in the half-plane
      // determined by ((p2-p1)|v1)>=0 an assertion (cpmassert) fails,
      // and the consequences of this range from ignoration to program
      // stop depending on the status of macro cpmdebug.

   static Path circle(const Spc& pc, R r, Z n=16); // 4 is experiment
      // see function ellipse of which the present function is a special
      // case

   static Path bezier(V<Spc> const& vs, Z n);
      // Bezier curve connecting vs.first() with vs.last() represented as
      // a polygon of n points

   static Path locInt(V<Spc> const& vs, Z n);
      //: local interpolation
      // smooth local interpolated curve connecting vs.first() with
      // vs.last() represented as a polygon of n points

   static Path cubCur(Spc const& p1, Vec const& v1,
      Spc const& p2, Vec const& v2, Z n=64);
      //: cubic curve
      // Represents the unique cubic curve x: [0,1] --> Spc
      // satisfying x(0)=p1, x(1)=p2, x'(0)=v1, x'(1)=v2 as
      // a polygon of n points (n>1 asserted).
      // See e.g. the nice book G.Aumann, K.Spitzmueller:
      // Computerorientierte Geometrie, BI 1993 p. 330, (11.7).
      // Although this function has the same argument structure
      // as the constructor
      // Path(Spc const& p1, Vec const& v1, Spc const& p2,
      // Vec const& v2, Z n=10);
      // it depends on the directions in a different way: In the
      // present function also the length of v1 and v2 matter (as it
      // obviously does for tangent vectors of a curve). For assessing
      // what length should be used in a particular curve design
      // situation one has to take into account that the parameter
      // interval for the whole curve is [0,1] which may be very
      // different from the geometrical length of the path, so that
      // the parametrization may be far away from the natural one
      // by arc length.

   static Path cubCur(Spc const& p1, Vec const& v1, R k1,
      Spc const& p2, Vec const& v2, R k2, Z n=64);
      // This function allows representing exactly the same curves
      // as the previous one. However, the parameters are are
      // better suitable for curve design. Independent of the
      // length of v1 and v2, they enter only by the unit vectors which
      // they determine. The factors k1 and k2 influence the relative
      // influence of the direction (the larger the corresponding factor
      // the longer piece of path that follows the tangent line closely.
      // Changing k1 and k2 by a common factor makes the curve longer
      // until it finally may be forced to develop a crossing point.
      // See p. 335, Fig. 11.3 in the book by Aumann and Spizmueller.
      // For k1=k2=1, one gets a curve which does not deviate much from
      // the connecting line from p1 to p2. For larger values, say 2
      // or 3, the increased arc length of the curve gives freedom to
      // follow the directions given at the ends to a more visible
      // degree and make the change from the initial direction into
      // the final one more smoothly.
      // With this function it was easy to create a greek lower case theta
      // (the variant which is completely different from the upper case
      // theta) by only three of such curves.

    static Path waveLine(Spc const& p1, Spc const& p2, R ampl,
       R omega, Vec const& tr=Vec(1,0), Z n=100);
      //: wave line
      // A wavy line (as photon lines in Feynman diagrams) from
      // p1 to p2. Actually this is a line in space in the form of a
      // linearly polarized wave. The actual spatial angular frequency
      // differs from the input omega as little as possible and
      // nevertheless lets the distance between p1 and p2 be an integer
      // number of half-waves. Notice lambda=2*pi/omega.
      // The actual transversal amplitude vector
      // is a unit vector which lies in the plane detemined by the
      // vectors p2-p1 and tr and is (of course) perpendicular to
      // the direction p2-p1.
      // The amplitude of the wave is given exactly by the argument ampl
      // whereas tr enters only as direction.

   Word nameOf()const{ return "Path";}
      //: name of

   VecGraph flatten()const;
      // returns the VecGraph object resulting from projecting
      // *this to the (1,2) plane

   const V<Spc>& operator()(void)const{ return x_;}
      // accessing the basic data via () as in
      // Path p=...;
      // V<Spc> vs=p();

   X2<Path,Path> divideByArc(R t)const;
      // let res be the return value.Then the intention is
      // *this == res.c1() & res.c2()
      // res.c1().length()==t

   X2<Path,Path> divideByPoint(const Spc sep)const
      // direct derivative from the previous function
   {
      return divideByArc(project(sep).c3());
   }

   Path operator &(const Path& p)const;
      // append p at the end of a copy of *this
      // and returns the result. If *this and p
      // fit perfectly together, the connection point
      // does not get doublicated. If they don't fit together
      // they simply get connected by a segment (i.e.
      // the list of points gets simply concatenated)

   Path moveAndMerge(const Path& p)const;
      // moves p such that it can be smoothly (without a kink)
      // appended at the end of a copy of *this. The result of
      // this appending procedure is returned

   Path replace(const Spc& s1, const Spc& s2, const Path& p)const;
      // The operation has a clear meaning only if s1 and s2
      // ly on the path or near to it so that their projected
      // points s1_, s2_ are conceptually clear.
      // Denote u1 the first of points {s1_,s2_} that we encounter
      // when going from firstPoint() along the path. Let u2
      // be the other one in {s1_,s2_}.
      // The resut is: (*this) from firstPoint() to u1
      // then p (connected by move and merge)
      // then (*this) from u2 to lastPoint() (connected by move and
      // merge)

   X3<Spc,Vec,R> project(const Spc& s)const;
      // Returning geometrical data, which allow to move s onto the
      // path *this in cases in which s is close to this path.
      // Precisely:
      // For n_==0 the function returns X2<Spc,R>(Spc(),0).
      // For n_==1 the function returns X2<Spc,R>(x_[1],0).
      // For n>=2 (the only case of interest) the function returns
      // (s0,e0,l0) \in X3<Spc,Vec,R> with s0,e0, and l0 defined as
      // follows:
      // We consider the n_-1 segments ('Strecken') x_[i]-->x_[i+1]
      // and treat them as point sets. Then s has a well-defined distance
      // to each of the segments and we select the segment for which
      // this distance is minimal. (If there are more with the same
      // distance, which is extremly unprobable, we select the one with
      // the smallest index). s0 is by definition the point on the line
      // which is determined by the selected segment which is closest to
      // s. The unit vector corresponding to the selected segment is e0.
      // Appart from points that are not near to the path or close to its
      // s0 will belong to its line-determining segment and will thus
      // belong to the path (which is made of these segments).
      // l0 is the distance of s0 from the initial point x_[1], measured
      // along the path.

   R toArc(const Spc& s)const{ return project(s).c3();}
      //: to arc
      // returns a number that characterizes the position of s along the
      // path. The function does not assume that s belongs to *this, but
      // if it does the result is easier to explain: it is the distance
      // from x_[1] measured along the path. Actually clear from code.

   X3<Spc,Vec,Z> fromArc(R t)const;
      //: from arc
      // Consider:
/*
         Path p=...;
         R t=...;
         Spc s=p.fromArc(t).c1();
         R t0=p.toArc(s);
         bool test=(t==t0);
*/
      // Then we get always the value true for the variable test.
      // Also returns the unit vector of tangent direction
      // as ret.c2() and i0=ret.c3() such that l_[i0]<=t<l_[i0+1]

   R turn()const;
      // returns the total winding angle from beginning to end,
      // e.g 2*pi for a closed circle, or 2*n*pi for a path obtained from
      // walking around a circle n times before meeding the initial
      // point again. Notice that it would waist information
      // if we would give the result as an Angle (which is unable
      // to hold the 'number of full turns').
      // The accurate mode should use the value of
      // e_[1].angTo(e_[n_-1]) for correcting the rounding errors
      // occuring during many steps in possibly many turns of the path.
      // Tests with very large number of points showed that there is
      // hardly a need for such an accuratification.

   Group con()const;
      //: connection
      // connection along the whole path. Consider
      // Path p=...;
      // Group g=p.con();
      // Then the following expressions evaluate to true:
      // (p.firstPoint()*=g) == p.lastPoint(),
      // (p.firstDir()*=g) == p.lastDir().
      // Very interesting concept. I knew connections only for
      // linear groups; the present form for an affine (inhomgeneous)
      // transformation group is in my head since ,say, two years
      // and was implemented here in order to describe the release of
      // a bended wire, whithout knowing the chain of states through
      // which bending was done. (program CO-Former).

   R turn2(R t1, R t2)const;
      // Going along the parth from fromArc(t1).c1() to fromArc(t2).c1()
      // (this might be a walk opposite to the inherent path direction)
      // we add together the rotations from path element to path element
      // (as numbers not as angles, i.e no mod-2*pi-reduction is
      // involved) and the sum is the result. One has
      //    turn2(t1, t2) == - turn2(t2, t1)
      // and for t1 <= t2 and t2-t1 sufficiently small that no completed
      // turn is made:
      //  turn2(t1,t2)==
      // (Vec e1=fromArc(t1).c2(), Vec e2=fromArc(t2).c2(), e1.angTo(e2) )

   Group con2(R t1, R t2)const;
      // generalizes con in just the same way in which turn2 generalizes
      // turn

   void static test(Z i);
      // Test tool

   bool isSmooth()const;
      // returns true if there are no reversions in direction
      // from point to point
};

///////////////////////// class Skeleton ///////////////////////////////

class Skeleton:public DiGraph<Spc>{// directed graph of points
   // Basic tool for creating spatial
   // structures to be visualized with objects of type Camera
   typedef DiGraph<Spc> Base;
   typedef DiGraph<Spc> CloneBase;
   typedef Skeleton Type;
public:
   virtual DiGraph<Spc>* clone()const{ return new Skeleton(*this);}
      // I/O from Base
   explicit Skeleton(Z n=0):Base(n){}
   explicit Skeleton(const V<Spc>& vsp, bool connect=false):
      Base(vsp){ if (connect) connectAsPath();}
      // notice that there are no edges yet, so that the
      // object is invisible when invocing
      // Camera ca=...; Skeleton sk=...; Color col=...;
      // ca.mark(sk,col);

   explicit Skeleton(Path const& p):Base(p()){connectAsPath();}
      // Typical visalization:
      // Path p=...; Camera ca=...; Color col=...;
      // ca.mark(Skeleton(p),col);

   Skeleton(Spc const& p1, Spc const& p2, Spc const& p3,
      bool connect=false): Base(V<Spc>{p1,p2,p3})
      { if (connect) connectAsPath();}
      // constructing a Skeleton from 3 points

   explicit Skeleton(const VV<Spc>& vsp, bool connect=false):
      Base(vsp.lin()){ if (connect) connectAsPath();}
      // using a matrix of points as input. If connections are
      // generated these duo not (yet ?) reflect the matrix mesh
      // geometry

   const Spc& lastPoint()const
      { return operator[](dim());}
   const Spc& firstPoint()const
      { return operator[](1);}
   Skeleton& operator *=(Group const& g);
      // translation and rotation combined
   Skeleton& operator +=(Vec const& a);
      // translation
   Skeleton& operator *=(AxVec const& A);
      // rotation

   Skeleton& scl_(Spc const& fix, R fac);
      //: scale

   Skeleton con(R w=0.5)const;
      // creates the dual graph, where each edge is transformed into the
      // vertex which is the center of the edge in the normal case w=0.5.
      // Since the edges are directed, it makes sense to put the new
      // vertex closer to the old source vertex (w<0.5) or closer to the
      // old destination vertex (w>0.5). The edges of the new graph are
      // lines that join just those vertices which in the old picture
      // where adjacent edges.

   void connectAsPath();
      // adds edges 1->2,2->3,4-->5,... n-1->n
      // and makes the edges vertexes visible

   void clsPath_(){ addEdge(x_.dim(),1);}
      //: close path
      // adds an edge from the last vertex to the first one

// generation functions as static functions (constructors
// can't have a descriptive name, so named functions
// are more convenient to use).

   static Skeleton arc(const Spc& p1, const Spc& center,
      const AxVec& rot, Z n=10);
      // circular arc obtained by applying a rotation rot arround
      // center in n steps on point p1.

   static Skeleton line(const Spc& pi, const Spc& pf, Z n=10);
      // straight line starting at pi and ending at pf
      // represented as a chain of n equidistant connected
      // points. (n-1 edges). cpmassert for n>1.

   static Skeleton parabola(const Spc& pi, const Spc& pc, const Spc& pf,
      Z n=10);
      // parabola with initial point pi, final point pf, and
      // Bezier control point pc. The intermediate points
      // correspond to equidistant positions in time, where the time span
      // needed for the whole path is 1. The meaning of the control
      // point is to express the initial velocity as (pc-pi)/(1/2)
      // and the final velocity as (pf-pc)/(1/2).
      // cpmassert for n>2.

   static Skeleton ellipse(const Spc& pc, R a1, R a2, Z n=16,
      bool connect=true);
      // generates an ellipse in the 1-2-plane arround
      // the origin approximated by a flattened
      // n-polygon. We start from a circle of radius a1
      // and multiply the 2-coordinates by a2/a1
      // (everything relativ to the master system of reference).
      // half-axis a1 in 1-direction
      // half-axis a2 in 2-direction
      // if connect==true the points are connected as a polygon,
      // otherwise the graph has no edges.
      // For a1==a2 (circle) more efficient code is generated by
      // avoiding the flattening loop.

   static Skeleton circle(const Spc& pc, R r, Z n=16, bool connect=true);
      // generates a circle in the 1-2-plane arround point pc
      // the origin approximated by a  n-polygon.
      // if connect==true the points are connected as a polygon,
      // otherwise the graph has no edges.

   static Skeleton bezier(V<Spc> const& vs, Z n);
      // Bezier curve connecting vs.first() with vs.last() represented as
      // a polygon of n points

   static Skeleton cone(R r, R h, Z n=16);
      // Pyramide over a regular n-polygon of higth h

   static Skeleton cylinder(R r, R h, Z n=16);
      // Cylinder over a regular n-polygon of higth h

   static Skeleton dblPyr(R length, R d, Z n=8);
      // method for generating a double pyramide ....

   static Skeleton rec(R a, R b);
      //: rectangle
      // Returns a rectangle with corners at (0,0),(a,0),(a,b),(0,b)

   Word nameOf()const{ return "Skeleton";}

   VecGraph flatten(Z connect)const;
      // returns the VecGraph object resulting from projecting
      // *this to the (1,2) plane
      // Argument connect will be given to the VecGraph constructor
};

///////////////////////// class WireFrame //////////////////////////////

class WireFrame{ // array of Skeleton objects
   // An array of skeletons is considered a new
   // entity; notice that it would be probably quite inefficient to
   // convert such an array into a single Skeleton by forming the obvious
   // union of sets.

   V<Skeleton> rep;

   typedef WireFrame Type;
public:
   CPM_IO
   virtual WireFrame* clone()const{ return new WireFrame(*this);}

explicit WireFrame(const V<Skeleton>& vsk):rep(vsk){}
explicit WireFrame(const Skeleton& sk):rep(V<Skeleton>(1,sk)){}
   WireFrame():rep(){}

   WireFrame con(R w=0.5)const;
      // creates the dual graph. See Skeleton::con(R)

   WireFrame& operator *=(Group const& g);
   WireFrame& operator +=(Vec const& a);
   WireFrame& operator *=(AxVec const& A);

// transforming the components of the WireFrame separately
// no restriction on dimensions since only those indexes are
// used for which both arrays have components
   WireFrame& operator *=(V<Group> const& g);
   WireFrame& operator +=(V<Vec> const& a);
   WireFrame& operator *=(V<AxVec> const& A);

// combining WireFrames
   WireFrame& operator &=(WireFrame const& wf)
      { rep&=wf.rep; return *this;}
   WireFrame& operator &=(Skeleton const& sk)
      { rep&=sk; return *this;}

// access
   const Skeleton& operator[](Z i)const{ return rep[i];}
   Z dim()const{ return rep.dim();}
   Word nameOf()const{ return "WireFrame";}

   static WireFrame tripod(R length, R relWidth=0.025, R relGap=0.05,
      Z n=3);
      // generates a system of orthogonal axes such that
      // axis n of the master system is made out of n pieces.

   static WireFrame globe(R r, R f=0, Z nLong=8, Z nLat=4, Z nCirc=16,
      bool axes=false);
      // method for generating a rotation ellipsoid like
      // the earth body.
      // r: equatorial radius
      // f: flattening factor: polar radius=r*(1-f)
      // nLong: we show longitude circles of angular distance
      //    180 degrees/nLong
      // nLat: we show latitude circles of angular distance
      //    90 degrees/nLat
      // nCirc: each 'circle' actually is a regular nCirc-polygon.
      // For f!=0 the latitude circles neither refer to topocentric
      // latitude nor to bodycentric latitude. Rather, they simply
      // result from flattening of a spherical globe and its
      // uniquely defined latitude circles. Flattening occurs in the
      // 2-direction (?? or 3 ?)

   static WireFrame spline(const V<Spc>& vs, const Spc& pc1, Z n=10);
      // 'smooth' merge of parabolas where the first one is
      // Skeleton::parabola(vs[1],pc1,vs[2]),
      // the second one:
      // Skeleton::parabola(vs[2],pc2,vs[3]),
      // where pc2 is given by pc2-vs[2]=vs[2]-pc1
      // the third one:
      // Skeleton::parabola(vs[3],pc3,vs[4]),
      // where pc3 is given by pc3-vs[3]=vs[3]-pc2
      // ....
      // So, when considered a path in space (parametrized by time)
      // the position and velocity are continuous functions,
      // only the acceleration changes discontinuously from one
      // parabola to the next
};
