//? cpmhomunitvecxcpp.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


// common implementation for Vec::homUnitVec(Z,Z,Z,R,R)

namespace{
// Tools for implementing function Vec::homUnitVec(Z,Z,Z,R,R).
// Motion of particles on the surface of a sphere.
// This is a nice mechanical system, with a compact configuration space
// without boundaries.
// Taking into account the serving role of the concept, we do not
// build in the flexibility (e.g. different masses for different
// particles) that a treatment as a mechanical system would suggest.

class Pb{ // class for defining single partices
   Vec x_;
      // position relative to the center of the sphere
   AxVec omg_;
      // omega, angular velocity with respect
      // to the center of the sphere
public:
   typedef Pb Type;
   static R eps_; // regularizing the Coulomb force
   CPM_NAM(Pb)
   CPM_IO
   Pb(Vec const& x=Vec(1,0,0)): x_(x), omg_(){}
     // argument needs to be a unit-vector (no adjustment for efficiency)
   Vec x()const{ return x_;}
   Pb& rec_(AxVec const& torque, R dt){omg_+=(torque*(dt)); return *this;}
      //: reaction
      // to a 'Drehstoß'
   AxVec tor(Pb const& p)const;
     // torque felt by *this due to the presence of particle p
   void step_(R dt);
      // free motion for a time span dt
   void stop_(){ omg_=AxVec();}
      // stopping motion
};

bool Pb::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("Pb",str);
   cpmp(x_);
   cpmp(omg_);
   return true;
}

bool Pb::scanFrom(istream& str)
{
   cpms(x_);
   cpms(omg_);
   return true;
}

R Pb::eps_=1e-2;

void Pb::step_(R dt)
{
   AxVec rot=omg_.exp(dt);
   x_*=rot;
}

AxVec Pb::tor(Pb const& p)const
{
   Vec to_p=p.x_-x_;
   R r=to_p.nor_();
   R fac=-R(1)/(r*r+eps_); // not singular
   Vec f=to_p*fac;
   return x_.cross(f);
 }

class PbSys{ // particle system
   V<Pb> vp_;
public:
   Vec operator[](Z i){ return vp_[i].x();}
   PbSys(V<Pb> const& vp):vp_(vp){}
   void stop_(){ for (Z i=vp_.b();i<=vp_.e();++i) vp_[i].stop_(); }
   void step_(R dt);
};

void PbSys::step_(R dt)
// direct midpoint integrator
{
   Z i,j,n=vp_.dim();
   R tau=dt*0.5;
   for (i=1;i<=n;++i) vp_[i].step_(tau);
   V<AxVec> vtor(n); // vtor[i] is the torque felt by
   // particle i due to the presence of the n-1 other particles
   for (i=1;i<=n;++i){
      for (j=1;j<i;++j){ // short loop
         AxVec vpij=vp_[i].tor(vp_[j]);
         vtor[i]+=vpij;
         vtor[j]-=vpij;// making use of actio = reactio
      }
   }
   for (i=1;i<=n;++i) vp_[i].rec_(vtor[i],dt).step_(tau);
}

} // namespace

V<Vec> Vec::homUnitVec(Z n, Z nSteps, Z stopPeriod,
   R dtFac, R eps)
{
   R dt=dtFac/cpmsqrt(R(n));
   Z i,ic=0;
   V<Pb> p(n);
   for (i=1;i<=n;++i){
      Vec xi=Vec::ranUnitVec();
      p[i]=Pb(xi);
   }
   Pb::eps_=eps;
   PbSys ps(p);
   i=0;
   while (ic<nSteps){
      i++;
      ps.step_(dt);
      if (i%stopPeriod==0){
         ic++;
         ps.stop_();
      }
   }
   V<Vec> res(n);
   for (i=1;i<=n;++i) res[i]=ps[i];
   return res;
}
