//? cpmshapesxcpp.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   Description: dimension-independent part of implementation
   of geometrical objects
*/

using namespace CpmStd;

///////////////////////// class Cylinder //////////////////////////////

bool CylinderDat::prnOn(ostream& out)const
// return preliminary
{
   CpmRoot::writeTitle("CylinderDat",out);
   x0.prnOn(out);
   rho.prnOn(out);
   hh.prnOn(out);
   return true;
}

bool CylinderDat::scanFrom(istream& in)
// return preliminary
{
   x0.scanFrom(in);
   rho.scanFrom(in);
   hh.scanFrom(in);
   return true;
}

R CylinderDat::indFnc(const Rdim& x)const
   {
      using CpmArrays::R_Vector;
      R val1=Iv(x0[1]-hh,x0[1]+hh)(x[1]);
      Z d=Spc::dimension();
      R_Vector y(d-1);
      for (Z i=2;i<=d;i++){
         y[i-1]=x[i]-x0[i];
      }
      R val2=y.abs()-rho;
      return CpmRootX::sup<R>(val1,val2);
   }

///////////////////// class Boxoid //////////////////////////////////////

bool BoxoidDat::prnOn(ostream& out)const
// return preliminary
{
   CpmRoot::writeTitle("BoxoidDat",out);
   x0.prnOn(out);
   s.prnOn(out);
   p.prnOn(out);
   return true;
}

bool BoxoidDat::scanFrom(istream& in)
// return preliminary
{
   x0.scanFrom(in);
   s.scanFrom(in);
   p.scanFrom(in);
   return true;
}

R BoxoidDat::indFnc(const Rdim& x)const
{
   Z d=x0.dim();
   R s1=0,s2=0;
   for (Z i=1;i<=d;i++){
      R ha=s[i]; // half axis
      R xi=cpmabs(x[i]-x0[i]);
      ha=cpmpow(ha,p);
      xi=cpmpow(xi,p);
      s1+=ha;
      s2+=xi/ha;
   }
   R pInv=R(1)/p;
   s1=cpmpow(s1,pInv);
   s2=cpmpow(s2,pInv);
   return s1*(s2-1.);
}

R BoxoidDat::getRadius()const
{
   if (p>R1(2)) return s.abs();
   else{
      Z d=Spc::dimension();
      CpmArrays::R_Vector rv(d);
      for (Z i=1;i<=d;i++){
         R hs=cpmabs(s[i]);
         rv[i]=hs;
      }
      return rv.sup();
   }
}

//////////////////////////// class Ellipsoid ///////////////////////////

bool EllipsoidDat::prnOn(ostream& out)const
// return preliminary
{
   CpmRoot::writeTitle("EllipsoidDat",out);
   x0.prnOn(out);
   s.prnOn(out);
   return true;
}

bool EllipsoidDat::scanFrom(istream& in)
// return preliminary
{
   x0.scanFrom(in);
   s.scanFrom(in);
   return true;
}

R EllipsoidDat::indFnc(const Rdim& x)const
{
   Z d=x0.dim();
   R s1=0,s2=0;
   for (Z i=1;i<=d;i++){
      R ha=s[i]; // half axis
      R xi=x[i]-x0[i];
      ha*=ha;
      xi*=xi;
      s1+=ha;
      s2+=xi/ha;
   }
   s1=cpmsqrt(s1);
   s2=cpmsqrt(s2);
   return s1*(s2-1.);
}

R EllipsoidDat::getRadius()const
{
   Z d=Spc::dimension();
   CpmArrays::R_Vector rv(d);
   for (Z i=1;i<=d;i++){
      R hs=cpmabs(s[i]);
      rv[i]=hs;
   }
   return rv.sup();
}

R EllipsoidDat::vol()const
{
   R res= cpmpi;
   Z d=Spc::dimension();
   if (d==3) res*=(4./3);
   for (Z i=1;i<=d;i++) res*=s[i];
   return res;
}

////////////////////////// class Box ///////////////////////////////////

bool BoxDat::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("BoxDat",str);
   cpmp(x0);
   cpmp(s);
   return true;
}

bool BoxDat::scanFrom(istream& str)
{
   cpms(x0);
   cpms(s);
   return true;
}

R BoxDat::vol()const
{
   R res=1;
   Z d=Spc::dimension();
   for (Z i=1;i<=d;i++) res*=(s[i]*2);
   return res;
}

R BoxDat::getRadius()const
{
   R res=0;
   Z d=Spc::dimension();
   for (Z i=1;i<=d;i++) res+=(s[i]*s[i]);
   res=cpmsqrt(res);
   return res;
}

R BoxDat::indFnc(const Rdim& x)const
{
   return SubsetOfSpace::box(x,x0,s);
}

Box::Box(const V<Spc>& vs, const RefSys& rs)
{
   Z i,j,n=vs.dim(),m=dimVec;

   Group g=rs.toMasSys();
   V<Rdim> vsp(n);
   for (i=1;i<=n;i++){
      Spc pi=vs[i];
      pi*=g;
      vsp[i]=pi.toRdim();
   }
   VV<R> dir(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         dir[i][j]=vsp[j][i];
      }
   }
   V<Iv> ivDir(m);
   for (i=1;i<=m;i++) ivDir[i]=Iv(dir[i]);
   Rdim x0;
   Rdim s;
   for (i=1;i<=m; i++){
      x0[i]=ivDir[i].mean();
      s[i]=ivDir[i].radius();
   }
   Box box(x0,s);
   box*=!g;
   *this=box;
}


V<Spc> Box::ran(Z n)const
// random sequence of n points in the Box
{
   V<Spc> res(n);
   Z i,j,ir=0;
   Rdim xs;
   using CpmRoot::ran;
   Z d=Spc::dimension();
   for (i=1;i<=n;i++){
      for (j=1;j<=d;j++){
         R rr=ran(R(1.),ir++);
            // values between -1 and +1
         xs[j]=(dat.x0)[j]+rr*(dat.s)[j];
      }
      res[i]=refSys(xs);
   }
   return res;
}

V<Spc> Box::lattice(Z n, R site)
{
   cpmmessage("Box::lattice(...) started");
   if (site<=0 || n<1) cpmerror("Box::lattice(...): bad arguments");
   Z d=Spc::dimension();
   Z i,j,k,l;
   Rdim s=dat.s;
   V<Z> ns(d);
   V<R> ss(d);
   Z nCells=1;

   for (i=1;i<=d;i++){
      R site_i=s[i]*2;
      Z nt=cpmtoz(site_i/site);
      ns[i]=(nt<=1 ? 1 : nt);
      ss[i]=(site_i/ns[i]);
      nCells*=ns[i];
   }

   Z nRes=cpminf(n,nCells);
      // number of points to be generated

   R dx=ss[1];
   R dy=ss[2];
   R dz=(d==3) ? ss[3] : R(0);

   Rdim corner;
   corner[1]=          (dat.x0)[1]-s[1]-dx*0.5;
   corner[2]=          (dat.x0)[2]-s[2]-dy*0.5;
   if (d==3) corner[3]=(dat.x0)[3]-s[3]-dz*0.5;

   V<Spc> res(nRes);
   Rdim xs;
   R x,y,z;

   l=0;
   if (d==2){
      y=corner[2];
      for (i=1;i<=ns[2];i++){
         y+=dy;
         xs[2]=y;
         x=corner[1];
         for (j=1;j<=ns[1];j++){
            if (++l>nRes) break;
            x+=dx;
            xs[1]=x;
            res[l]=refSys(xs);
         }
      }
   }
   else{
      z=corner[3];
      for (i=1;i<=ns[3];i++){
         z+=dz;
         xs[3]=z;
         y=corner[2];
         for (j=1;j<=ns[2];j++){
            y+=dy;
            xs[2]=y;
            x=corner[1];
            for (k=1;k<=ns[1];k++){
               if (++l>nRes) break;
               x+=dx;
               xs[1]=x;
               res[l]=refSys(xs);
            }
         }
      }
   }
   cpmmessage("Box::lattice(...) done");
   return res;
}

//////////////////////////// class Sphere ///////////////////////////

bool SphereDat::prnOn(ostream& out)const
// return preliminary
{
   CpmRoot::writeTitle("SphereDat",out);
   x0.prnOn(out);
   r.prnOn(out);
   return true;
}

bool SphereDat::scanFrom(istream& in)
// return preliminary
{
   x0.scanFrom(in);
   r.scanFrom(in);
   return true;
}
