//? cpmvectorgraphics.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_VECTORGRAPHICS_H
#define CPM_VECTORGRAPHICS_H

/*
Purpose: Declaration of graphic classes VecGraph and
   VecGraphs representing vector graphics.
   Class Figure for generating graphical LaTex code.
   (No recent experience with this class).
*/

#include <cpmgraph.h>
#include <cpmdigraph.h>
#include <cpmrfunc.h>

namespace CpmGraphics{
   
   using namespace CpmStd;

   using CpmRoot::Word;
   using CpmRoot::Z;
   using CpmRootX::Z1;
   using CpmRoot::R;
   using CpmRoot::C;
   using CpmArrays::V;
   using CpmImaging::Color;
   using CpmFunctions::R_Func;

   using CpmGeo::DiGraph;
   using CpmGeo::Angle;

///////////////////// class Figure ///////////////////////////////////////
// defines the analogon of a screen window as a 'paper window'
// on which graphical LATEX commands work.

class Figure{ 
// the analogon of a screen window as a 'paper window', related to LaTex

// independent data (all belong to a class with default constructor)

   Iv px_;  // picture range in x
   Iv py_;  // picture range in y

   Iv rx_;  // range in x
   Iv ry_;  // range in y

   R_Func fx_;
   R_Func fy_;
   R_Func gx_;
   R_Func gy_;

   Word nx_; // name of the quantity to which the x-axis refers
   Word ny_; // name of the quantity to which the y-axis refers

   Word tit_; // title of the diagram

   Word cap_; // caption of the figure

   Z1 withOri_;
      // withOrigin!=0 controls autoscaling such that the origin
      // is always on the diagram
   Z1 newGrid_;
      // only =0 or !=0 matters. If it is 0 (should be an acception) one
      // let the grid remain unchanged and draw curves on the frame with
      // a meaning of object space wich is no longer in agreement with
      // the grid and the grid text
   Z1 autoScl_;
      // only =0 or !=0 matters. Influences show(..) such that CL CU are
      // adjusted to the data to be represented prior to drawing them.
      // 'autoscaling'
      // The change of CL and CU remains till new values are assigned
      // by new action
   Word gridText_;
   R2 gridTextPos_;
      // text position is understood in the Frame coordinate system
      // (xF,yF)
      // i.e. (0,0) is left lower corner, (1,1) is right upper corner
public:

   Figure(Iv px, Iv py, Iv rx=Iv(0,1), Iv ry=Iv(0,1));
   void draw(R2 a, R2 b, ostream& str, Z np);
   void drawLine(R2 a, R2 b, ostream& str, Z np);

   R getOx()const{ return px_[1];}
   R getOy()const{ return py_[1];}
   R getLx()const{ return px_[2]-px_[1];}
   R getLy()const{ return py_[2]-py_[1];}
   const Word& cap()const{ return cap_;}
   Word& cap(){ return cap_;}
};

///////////////////////// class VecGraph ///////////////////////////////

class VecGraph:public DiGraph<C>{ 
// vector graphics, directed graph with vertices of type C 
// An instance of this class is a list of points in C with some
// of them indicated as connected. The function mark
// creates a graphical representation of the object by drawing
// a line of the color given by attribute 'color' between all
// connected points (here the translation from C to Screen points
// is the one provided by the argument Graph gr of function mark.
// As constructed by the constructors of this class, the points
// have no connections at all. One may create such connections
// using the functions set_neg and addEdge of the base class.
// The most powerful way of generating VecGraphs is by
// using a object camera of type CpmCamera2::Camera or
// CpmCamera3::Camera and to use  camera.capt(obj), where obj
// is of type CpmGeo::Skeleton or CpmGeo::WireFrame. By moving
// object camera, one gets all projective images of the original
// so that there is no need for defining geometrical transforms
// of VecGraphs (it is not without meaning that we are here
// in namespace CpmGraphics and not in CpmGeo !)

   typedef DiGraph<C> Base;
   typedef DiGraph<C> CloneBase;
   typedef VecGraph Type;
   Color color;
   static R dx,dy;
      // x-increment corresponding to 1 graphical pixel and
      // y-increment corresponding to 1 graphical pixel
public:
   virtual DiGraph<C>* clone()const{ return new VecGraph(*this);} 

   CPM_IO
      // I/O from Base
   static void set_dx_dy(Graph const& gr){dx=gr.dx();dy=gr.dy();}
      //: set dx, dy
      // Some graphical elements, such as small circles or disks to be
      // used as representation of points have to be composed of many
      // straight lines and the suitable number of these lines depends
      // on the pixel size in the final image on screen or as image file.
      // This function makes class VecGraph aware of the resolution that
      // will be used. Of course, it is required that gr was given 
      // previously the size that will be used later in calling function
      // mark(Graph& gr).
      
   static R get_dx(){ return dx;}
   static R get_dy(){ return dy;}
   VecGraph(Z n=0):Base(n){}
   explicit VecGraph(const V<C>& vsp, Z connect=1):Base(vsp,connect){}

   Word nameOf()const{ return "VecGraph";}
   void mark(Graph& gr)const;
   void markTriAng(Graph& gr)const;
      // mark a filled triangle, determined by the first 3 components
      // of array x_. No action if there are less than 3 points.
   Color getColor()const{ return color;}
   void setColor(Color c){ color=c;}
   void setColor(Z i){ color=Color(i);}

// 3 functions not yet implemented
   VecGraph shift(C s)const;
   VecGraph rot(C p, Angle rot)const;
   VecGraph scale(R fac)const;

   static VecGraph arc(C p1, C center, Angle rot, Z np=10);
      //: arc
      // circular arc obtained by applying a rotation rot arround
      // center in n steps on point p1.

   static VecGraph line(C pi, C pf, Z np=2);
      //: line
      // Straight line starting at pi and ending at pf 
      // represented as a chain of n equidistant connected
      // points. (n-1 edges); cpmassert for n>1.

   static VecGraph circle(C pc, R r, Z np=64);
      // Circle arround pc with radius r,
      // Polygon with np points. If np<=0 a reasonable value will be
      // determined and used from dx and dy.

   static VecGraph disk(C pc, R r, Z np=64);
      //: disk
      // Filled circle arround pc with radius r
      // actually a circle(pc,r,np) as outer boundary 
      // and a spiral in the interior. The design of this 
      // spiral aims at filling all pixels in the interior.
      // np<=0 has an effect analogous to that in function circle.
      // If one finds unfilled pixels one can always increase np.
      // Computation time may increase considerably, though.

   static VecGraph ellipse(C pc, R a1, R a2, Z np=64);
      // Ellipse with center pc, semi-axis a1 in 1-direction
      // and a2 in 2-direction 

   static VecGraph parabola(C pi, C pc, C pf, Z np=16);
      // Parabola with initial point pi, final point pf, and 
      // Bezier control point pc. The intermediate points 
      // correspont to equidistant positions in time, where the time span
      // needed for the whole path is 1. The meaning of the control 
      // point is to express the initial velocity as (pc-pi)/(1/2)
      // and the final velocity as (pf-pc)/(1/2).
      // cpmassert for n>2.

   static VecGraph bezier(V<C> const& p, Z n=256);
      // Bezier curve connecting p.first() with p.last() represented as a
      // polygon of n points.
};

///////////////////////// class VecGraphs //////////////////////////////

class VecGraphs{ // array of VecGraph objects
   // An array of VecGraph objects is considered a new
   // unity; notice that it would be probably quite inefficient to
   // convert such an array into a single VecGraph by
   // forming the obvious set union.
   CpmArrays::V<VecGraph> rep;

   typedef VecGraphs CloneBase;
   typedef VecGraphs Type;
public:
   CPM_IO
   virtual VecGraphs* clone()const{ return new VecGraphs(*this);}

explicit VecGraphs(const V<VecGraph>& vsk):rep(vsk){}
   VecGraphs():rep(){}

// combining VecGraphss
   VecGraphs& operator &=(const VecGraphs& wf)
      { rep&=wf.rep; return *this;}

   VecGraphs& operator &=(const VecGraph& sk)
      { rep&=sk; return *this;}

// access
   const VecGraph& operator[](Z i)const{ return rep[i];}
   VecGraph& operator[](Z i){ return rep[i];}

   Z dim()const{ return rep.dim();}
      //: dimension

   Word nameOf()const{ return "VecGraphs";}
      //: name of

   void mark(Graph& gr)const
      //: mark
   { for (Z i=1;i<=dim();i++) rep[i].mark(gr);}

   void epic(ostream& str, Figure& fg);

   static VecGraphs polygon(V< X2<C,R> > const& p, Z np=0);
      // Polygon connecting the points p[i].c1(). The points get a mark
      // in the form of a 'circle of radius r=p[i].c2()'. This term is
      // understood here as to mean no marking for r==0 and marking with
      // a filled disk of radius |r| if r<0. For r > 0 the circle has a
      // blank interior. Since a polygon with many points allows to 
      // simulate a smooth curve, one can create by this function smooth 
      // curves on which some places are marked by circles so that the
      // curve is invisible in the interior of the circles. In this case
      // the majority of points in the list p have radius 0. 2009-08-28
      // If the argument np is larger than 0, it lets the circles
      // represent as regular polygons with np vertices. Otherwise an
      // value for np is determined from the resolution dx, dy of 
      // VecGraph.
};

}  // namespace

#endif
