//? cpmvectorgraphics.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmvectorgraphics.h>
#define CPM_DIMDEF2
#include <cpmdim.h>

using namespace CpmGraphics;
using namespace CpmRoot;
using namespace CpmDim2;

Figure::Figure(Iv px, Iv py, Iv rx, Iv ry):px_(px),py_(py),rx_(rx),
   ry_(ry)
{
   fx_=R_Func(rx_,px_);
   fy_=R_Func(ry_,py_);
}

void Figure::drawLine(R2 a, R2 b, ostream& str, Z np)
{
   R2 a2(fx_(a[1]),fy_(a[2]));
   R2 b2(fx_(b[1]),fy_(b[2]));
   str<<endl<<"   "<<"\\dottedline{"<<np<<"}("<<a2[1]<<","<<a2[2]<<")\
   ("<<b2[1]<<","<<b2[2]<<")";
   // actually we need only to draw a line; the epic package available to
   // me produces no actaptable quality
}

void Figure::draw(R2 a, R2 b, ostream& str, Z np)
{
   R2 a2(fx_(a[1]),fy_(a[2]));
   R2 b2(fx_(b[1]),fy_(b[2]));
   R2 c=(a2+b2)*0.5;
   str<<endl<<"   "<<"\\qbezier["<<np<<"]("<<a2[1]<<","<<a2[2]<<")("
      <<c[1]<<","<<c[2]<<")("<<b2[1]<<","<<b2[2]<<")";
   // actually we need only to draw a line; the epic package available to
   // me produces no actaptable quality
}

/////////////////////////// class VecGraph ///////////////////////////////

R VecGraph::dx=0.002;
R VecGraph::dy=0.002;

void VecGraph::mark(Graph& gr)const
{
   using CpmArrays::S;
   using CpmLinAlg::Z2;
   S<Z2> edg=edges();
   for (Z i=1;i<=edg.car();i++){
      Z2 zi=edg[i];
      R2 a(x_[zi[1]]); // x is an protected array of C's
         // in base class DiGraph<C>
      R2 b(x_[zi[2]]);
      gr.draw(a,b,color);
         // thus Graph::draw is the basic 'machine'
         // for drawing polygons (simulating curved
         // lines) in my graphics system
   }
}

void VecGraph::markTriAng(Graph& gr)const
{
   if (dim()<3) return;
   gr.fillTriangle(R2(x_[1]),R2(x_[2]),R2(x_[3]),color);
}

bool VecGraph::prnOn(ostream& str)const
{
   cpmwt("VecGraph");
   if (!Base::prnOn(str)) return false;
   cpmp(color);
   return true;
}

bool VecGraph::scanFrom(istream& str)
{
   if (!Base::scanFrom(str)) return false;
   cpms(color);
   return true;
}

VecGraph VecGraph::arc(C p1, C center, Angle rot, Z np)
{
   return Skeleton::arc(Spc(p1),Spc(center),AxVec(rot),np).flatten(1);
}

VecGraph VecGraph::line(C pi, C pf, Z np)
{
   return Skeleton::line(Spc(pi),Spc(pf),np).flatten(1);
}

Z adjust(R r, R dr)
{
   R phih=cpmacos(R(1)-dr*0.5/r);
   R rnp=cpmpi/phih;
   Z res = cpmrnd(rnp);
   if (res<32) res=32;
   return res;
}

R dPhi(R r, R dr)
{
   return cpmacos(R(1)-dr*0.5/r);
}

VecGraph VecGraph::circle(C pc, R r, Z np)
{
   if (np<=0){
      R dr=C(dx,dy).abs();
      np=adjust(r,dr);
   }
   return Skeleton::circle(Spc(pc),r,np).flatten(2);
}

/********************
VecGraph VecGraph::disk(C pc, R r, Z np)
{
   V<C> vCir=circle(pc,r).get_x();
   V<C> vSpiral(np);
   R phi=0, dphi=0.1;
   R ri=r, dr=r/np;

   for (Z i=1; i<=np;++i){
      C zi(ri,phi,"polar");
      phi+=dphi;
      ri-=dr;
      vSpiral[i]=pc+zi;
   }
   return VecGraph(vCir&vSpiral);
}
***********************/

VecGraph VecGraph::disk(C pc, R r, Z np)
{
   R dr;
   V<C> vCir;
   if (np<=0){
      dr=C(dx,dy).abs();
      vCir=circle(pc,r,0).get_x();
   }
   else{
      dr=r*(R(1)-cpmcos(cpmpi/np));
      vCir=circle(pc,r,np).get_x();
   }
   V<C> vSpiral(0);
   R phi=0;
   R ri=r;
   while (ri>0){
      C zi(ri,phi,"polar");
      vSpiral&=(pc+zi);
      R dphi=dPhi(ri,dr)*0.2;
      phi+=dphi;
      ri-=dr*(dphi/cpmpi)*0.2;
   }
   return VecGraph(vCir&vSpiral);
}

VecGraph VecGraph::ellipse(C pc, R a, R b, Z np)
{
   return Skeleton::ellipse(Spc(pc),a,b,np).flatten(2);
}

VecGraph VecGraph::parabola(C pi, C pc, C pf, Z np)
{
   return Skeleton::parabola(Spc(pi),Spc(pc),Spc(pf),np).flatten(1);
}

VecGraph VecGraph::bezier(V<C> const& p, Z n)
{
   Z pd=p.dim();
   V<Spc> ps(pd);
   for (Z i=1;i<=pd;++i) ps[i]=Spc(p[i]);
   return Skeleton::bezier(ps,n).flatten(1);
}

//////////////////////// class VecGraphs /////////////////////////////////

void VecGraphs::epic(ostream& str, Figure& fg)
{
   using CpmArrays::S;
   using CpmLinAlg::Z2;
   R ox=fg.getOx();
   R oy=fg.getOy();
   R lx=fg.getLx();
   R ly=fg.getLy();
   str.precision(6);
   str.setf(std::ios_base::fixed,std::ios_base::floatfield);
   str<<endl<<"% Output of function VecGraphs::epic(...)";
   str<<endl<<"\\setlength{\\unitlength}{1mm}";
   str<<endl<<"\\begin{figure}";
   str<<endl<<"\\begin{picture}("<<lx<<","<<ly<<")("<<ox<<","<<oy<<")";
   for (Z i=1;i<=dim();i++){
      VecGraph vgi=rep[i];
      S<Z2> edg=vgi.edges();
      for (Z j=1;j<=edg.car();j++){
         Z2 zj=edg[j];
         Z zj1=zj[1];
         Z zj2=zj[2];
         R2 a(vgi[zj1]);
         R2 b(vgi[zj2]);
         fg.drawLine(a,b,str,1);
      }
   }
   str<<endl<<"\\end{picture}";
   str<<endl<<"\\caption{"<<fg.cap()<<"}";
   str<<endl<<"\\end{figure}";
}

VecGraphs VecGraphs::polygon(V< X2<C,R> > const& p, Z np)
// UM 2009-08-28
      // Polygon connecting the points p[i].c1(). The points get a mark
      // in the form of a 'circle of radius r=p[i].c2()'. This term is
      // understood here as to mean no marking for r==0 and marking with
      // a filled disk of radius |r| if r<0. For r > 0 the circle has a
      // blank interior. Since a polygon with many points allows to
      // simulate a smooth curve, one can create by this function smooth
      // curves on which some places are marked by circles so that the
      // curve is invisible in the interior of the circles. In this case
      // the majority of points in the list p have radius 0. 2009-08-28
{
   Z i,pd=p.dim();
   if (pd==0) return VecGraphs();
// First we create a list of the points which have to be marked
   using std::vector;
   vector<Z> ml; // mark list
   for (i=p.b();i<=p.e();++i){
      R ri=p[i].c2();
      if (ri!=0.0) ml.push_back(i);
   }
   Z md=(Z)ml.size();
   if (md==0){ // the we have a pure polygon (no circles)
      V<C> pl(pd);
      for (i=1;i<=pd;++i) pl[i]=p[i].c1();
      VecGraph vg(pl);
      return VecGraphs(V<VecGraph>{vg});
   }
// Now md > 0
   V<VecGraph> vvg(md);
   for (i=1;i<=md;++i){
      X2<C,R> xcr=p[ml[i-1]];
      C pi=xcr.c1();
      R ri=xcr.c2();
      if (ri>0) vvg[i]=VecGraph::circle(pi,ri,np);
      else vvg[i]=VecGraph::disk(pi,-ri,np);
   }
   VecGraphs res(vvg);
   if (pd==1 && md==1) return res;
// Now pd>1 since pd >= md and case md==0 already treated.
// for each pair p[i],p[i+1] we add a line which just connects p[i] and
// p[i+1] if none of those is maked, and a suitably shorted line else.
   for (i=1;i<pd;++i){
      X2<C,R> xc1=p[i], xc2=p[i+1];
      C z1=xc1.c1(), z2=xc2.c1();
      R r1=xc1.c2(), r2=xc2.c2();
      bool m1=r1!=0.0, m2=r2!=0.0; // mark 1, mark 2
      C a,b;
      if (m1==false){
         a=z1;
      }
      else{
         C shift=z2-z1;
         shift.nor_(r1);
         a=z1+shift;
      }
      if (m2==false){
         b=z2;
      }
      else{
         C shift=z1-z2;
         shift.nor_(r2);
         b=z2+shift;
      }
      VecGraph li=VecGraph::line(a,b);
      res&=li;
   }
   return res;
}

bool VecGraphs::prnOn(ostream& str)const
// UM 2009-08-28
{
   cpmwt("VecGraphs");
   cpmp(rep);
   return true;
}

bool VecGraphs::scanFrom(istream& str)
// UM 2009-08-28
{
   cpms(rep);
   return true;
}

#undef CPM_DIMDEF2
