//? cpmfieldstatedirac1.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_FIELDSTATEDIRAC1_H_
#define CPM_FIELDSTATEDIRAC1_H_
/*
   Description: Development and test of quantized fields
      in one spatial dimension
*/
//#define CPM_M
#undef CPM_M
// if CPM_M is active, CpmArrays::M<> is used instead of std::map<>.
// Much slower, unfortunately: presently the factor is ~ 16
// More resently: factor ~ 100. The factor dependson the space dimension
// and is not analysed satisfactorily so far. Anyways, it makes no sense to
// use CpmArrays::M instead of std::map

#if !defined(CPM_M)
   #include <map>
#endif

#include <cpmx.h>
#include <cpmstatedirac1.h>
#include <cpmoperator.h>

using CpmRoot::Z;
using CpmLinAlg::Z2;
using CpmRoot::R;
using CpmRoot::C;
//using CpmLinAlg::C4;
using CpmRootX::Z1;
using CpmRootX::R1;
using CpmGraphics::Frame;
using CpmGraphics::Graph;
using namespace std;
using namespace CpmArrays;
using namespace CpmFunctions;
using CpmArrays::X3;

namespace CpmQuantumMechanics{

class OpFD1;

/////////////////////////////// class KyD1 ///////////////////////////////

class KyD1{ // index set for an orthonormal base of FieldSpaceD1
   // auxiliary class needed as an index set for an orthonormal
   // base of FieldSpaceD1.
   // Actually the instances of KyD1 are multi-indexes i.e.
   // lists of integers. List length is 2*s_ and list values are
   // not negative in our present case, due to their interpretation
   // as occupation numbers. s_ is the number of lattice points in space.
// independent;
   V<Z> v_;
      // Z-valued array, the components play the role of occupation
      // numbers at lattice points.
// dependent
   Z d_;
      // v_.dim();
   Z s_;
      // Number of lattice points in space, s_==v_.dim()/2.
// private function
   R sigma(Z i)const;
      // function needed to deal with fermions;
   void ini_(){ d_=v_.dim(); cpmassert(d_%2==0,"KyD1::ini_()"); s_=d_/2;}
// static members
   static Z lim_;
      // limit to the occupation numbers, 1 for fermions,
      // large for bosons
   static Z eps_;
      // +1 for fermions , -1 for bosons, set depending on
      // lim_=1 (for fermions) or larger (for bosons)
// static member functions
   static void setEps(void){ if (lim_==1) eps_=1; else eps_=-1;}
      // setting eps_ consistently

   typedef KyD1 Type;

public:
   CPM_IO
   CPM_ORDER
   CPM_NAM(KyD1)
   explicit KyD1(V<Z> const& v):v_(v){ini_();}
   explicit KyD1(Z s=0):v_(s*2),d_(s*2),s_(s){}
      // Basic constructor. Safe since gives error for s<0
      // Notice that all components (i.e. all occupation numbers are 0)
      // by initialization of V<Z>.

   Z s()const{ return s_;}
     //. dimension (number of space points)

   Z d()const{ return d_;}
      //: dimension (of v_ )

   Z sup()const;
      //: supremum
      // returns the value of the largest component for s_>0
      // and -1 else

   Z inf()const;
      //: infimum
      // returns the value of the smallest component for s_>0
      // and -1 else

   void sup_(KyD1 const& k);
      //: supremum
      // changes *this into the component-wise maximum
      // of *this and k.
      // Error if both have not the the same dimension

// basic tools for defining creation operators
// (see code defining crt(Z), crt(State)
   bool up_(Z i);
   R up(Z i)const;

// basic tools for defining destruction operators
// (see code defining des(Z), des(State)
   bool down_(Z i);
   R down(Z i)const;
   Z operator[](Z i)const{ return v_.cui(i);} // needed for crt and des
      // in up_,up,down_,down, and operator[] it is now assumed that the
      // index is in proper range. The destruction and creation operators
      // use this functions in a manner which takes care.

// bringing about the space-index, spinor-index separation
 //  Z2 toZ2(Z i)const{ return Z2(i/4 + 1,i%4);}
   Z2 toZ2(Z i)const{ Z j=i-1; return Z2(1+j/2,1+j%2);}
 //  Z toZ(Z2 const& ij)const{ return (ij.x1-1)*4+ij.x2;}

// setting the boson-fermion alternative
   static void setLim(Z i){ lim_=CpmRoot::posVal(i); setEps();}
   static void setBosons(Z lim=1024){ lim_=lim; setEps();}
   static void setFermions(){ lim_=1;setEps();}

// getting the descriptors referring to the boson-fermion alternative
   static Z getLim(){ return lim_;}
   static Z getEps(){ return eps_;}
};

///////////////////////// class FieldStateD1 /////////////////////////////

class FieldStateD1:public QMBase{
   // state space in occupation number representation
   // General considerations on finite-dimensional Hilbert spaces.
   // Let an orthonormal basis of Hilbert space H be represented as a
   // class, say B. This means that there is an interpretation according
   // to which each instance of B is an element of an orthonormal basis of
   // H and vice versa. Then each element psi \in H determines a
   // function B --> C, b |--> <b|psi> ( where <|> denotes the scalar
   // product in H and C are the complex numbers) and this mapping
   // H --> (B --> C) is linear and an isomorphism of Hilbert spaces if
   // the class B --> C is made a Hilbert space in a natural manner.
   // The class B-->C is naturally represented as CpmArrays::M<B,C> in C+-
   // or as std::map<B,C> in the C++ standard library (former STL).
   // If the cardinality |B|=:d is large, this is much more economic
   // than the class of all psi \in V<C> with psi.dim()=d.
   // In the present case we deal with a Hilbert space of states for which
   // the dimension d is written as a product of two natural numbers
   // which are the data members s_ and lim_ of the present class. For
   // convenience we refer to these numbers as s and l in the present
   // discussion. There are two natural interpretations of this space:
   //
   // 1. Tensor product of s single-particle spaces, each of dimension l.
   // Here we  consider the single-particle spaces as attached to s space
   // points, which may (or may not) form a linear lattice, the biotope of
   // the system. As mentioned already, these 'localized' state spaces all
   // have the same dimension l, and we consider them as identical to
   // some l-dimensional Hilbert space H_0 which has an orthonormal
   // base B_0.
   // The physical interpretation is that at each of the s space points
   // (sites) x_i, i=1,...,s,  there is a particle p_i (or other kind of
   // quantum system), which is a fermion for l=1 and a boson else. This
   // is not important for the state space as a Hilbert space but only for
   // the definition of operators.
   // A basis of H is given by the set of lists (b_1,...b_s) where b_i
   // \in B_0. Put differently, a natural basis of H has the form (B_0)^s
   // so that the dimension of H is d = l^s and thus grows exponentially
   // with s. If, however, l=1 we have d=1 for all s so that this case is
   // not interesting for fermions, so lets consider bosons only.
   // For bosons it is natural to consider for each site i the 'local'
   // creation and destruction operators c_i and d_i satisfying
   // [d_i,c_i]=1. Since operators referring to different sites commute
   // we have the general canonical relations [d_i,c_j]=\delta_{i,j},
   // [d_i,d_j]=0, [c_i,c_j]=0 . Since operators referring directly to
   // different sites never anti-commute, a correspondingly trivial
   // realization of canonical anti-commutation relations is not possible.
   // In the bosonic case, it is a consitent physical picture that at each
   // site there is placed a boson (oscillator). The state of the whole
   // system is, however, not given by assigning to each site a local
   // state. This would only lead to pure product states (nuclear states)
   // in the tensor product state. The generic state is a linear
   // combination of nuclear states. Notice that this is exactly what
   // holds for the state space of s distinguishable bosons. If these
   // bosons have a two-dimensional state space (qu-bits), e. g.
   // neutral atoms in electronic ground-state with respect to the
   // normal fine structure but showing a two-level hyperfine splitting.
   // For the states of this bosonic system there is also a representation
   // in terms of fermions via the Jordan-Wigner transformation, which we
   // do not consider here.

   // 2. Fock space over a s-dimensional single-particle space.
   // This works for bosons and fermions in nice parallelity and is the
   // case to which the creation and destruction operators defined here
   // directly refer. It is instructive to look at the single particle
   // space as the space of complex functions on B_0 = {x_1,x_2,...,x_s}.
   // For each i \in B_0 there is a creation operator c_i and its
   // Hermitean conjugate d_i (destruction operator) and a
   // orthonormal base of H is of the form
   // (n_1! n_2!...n_s!)^{-1/2} {c_1}^{n_1} {c_2}^{n_2}...{c_s}^{n_s} |0>
   // =: |n_1,n_2,...,n_s> where n_i \in {0,...s-1}, notice 0!=1
   // see e.g. Messiah I, p. 453, (XII.68) For fermions n_i \in {0,1}
   // and all pre-factors are 1.
   // On the other hand, simply declaring the set B :=
   // { (n_1,n_2,...,n_s)| n_i \in {0,...s-1}}
   // as an orthonormal base can be implemented by considering the
   // Hilberspace M<B,C> as mentioned above. In this 'Hilbertspace in
   // occupation number representation' creation and destruction operators
   // can be defined by the formulas
   // d_i |n_1,...n_i,...,n_s> = {n_i}^{1/2} |n_1,...,n_i - 1,...,n_s>
   // c_i |n_1,...n_i,...,n_s> = {n_i + 1}^{1/2} |n_1,...,n_i + 1,...,n_s>
   // (see Schweber, p. 130 (47), p. 131 (51))
   // for bosons and
   // d_i |n_1,...n_i,...,n_s> = (-1)^{q_i} n_i |n_1,...,n_i - 1,...,n_s>
   // c_i |n_1,...n_i,...,n_s> =(-1)^{q_i}(1-n_i)|n_1,...,n_i + 1,...,n_s>
   // where q_i := n_1 + n_2 + ... + n_(i-1)
   // (see Schweber, p. 138 (87), p. 139 (89),(90))
   // for fermions
   // As an effect of this definition, we have
   // |n_1,n_2,...,n_s> =
   // (n_1!n_2!...n_s!)^{-1/2} {c_1}^n_1 {c_2}^n_2...{c_s}^n_s |0,0,...,0>

   Bio b_;
      // combines number of space points and an interval in space
      // hosting these points

   Z lim_;
      // limit for occupation number, 1 for fermions.
      // Should be large enough not to have an influence
      // on results if >1 (then we have bosons). So the type
      // of statistics is a state attribute here.
//   bool pbc_;
      // only periodic boundary condition
#if defined(CPM_M)
   typedef CpmArrays::M<KyD1,C> Mkc;
#else
   typedef std::map<KyD1,C> Mkc;
#endif
   // associative array either from C+- or from the standard library
   Mkc m_;
   FieldStateD1(Bio b, Mkc const& m, Z lim):
      b_(b), lim_(lim),m_(m){}

   FieldStateD1 mod(Mkc const& m)const{return FieldStateD1(b_,m,lim_);}
      // returns a modified version of *this

   typedef FieldStateD1 Type;

   Type aux(CC2 const& mat1, CC2 const& mat2, R const& potj, Z j)const;

public:
   CPM_IO
   CPM_SCALAR_PRODUCT_INTERFACE(Type,C)
   CPM_PURE_VECTOR_SPACE_INTERFACE(Type,C)
   CPM_PURE_SCALAR_INTERFACE(Type,R)
   CPM_NAM(FieldStateD1)

   FieldStateD1():b_(),lim_(1){}

   FieldStateD1(Bio const& b, Z kind, Z lim);
      // Creating the normalized ground state on a lattice of points
      // as given by b (Fock vacuum state) for kind==1 and the zero-vector
      // on the same lattice else.

   FieldStateD1 zero()const{ return FieldStateD1(b_,0,lim_);}
      //: zero
      // zero-vector with the same general data as *this

   FieldStateD1 vac()const{ return FieldStateD1(b_,1,lim_);}
      //: vacuum
      // Fock vacuum with the same general data as *this

   FieldStateD1(KyD1 const& k, Iv const& iv, C const& z, Z lim):
      b_(k.s(),iv), lim_(lim){ m_[k]=z; }

   FieldStateD1(StateD1 const& st,Z lim);
      // takes the lattice size from n=st.dim(). Creates a
      // FieldStateD1 which is obtained by letting the
      // operator sum from i=1 to i=n of crt(i)*st[i] act on
      // the ground state FieldStateD1(n,1). The actual implementation
      // is more direct and more efficient.

   FieldStateD1(V<StateD1> const& vs,Z lim);
      // Assumes that the components of vs are States on the same
      // lattice.
      // Let m=vs.dim(). Then the generated space is a m-particle
      // state.

   void set_(KyD1 const& k, C const& z){ m_[k]=z;}
      // notice that mutating functions end in _ by convention

   Z size()const{ return b_.d();}
      // Cardinality of the underlying point lattice. Since StateD1
      // consists of C2-valued functions on this point lattice, the
      // one-particle space
      // over which FieldStateD1 happens to be the Fock space

   Z dim()const{ return b_.d();}
      // Not actually the dimension of FieldStateD1 (which is unlimited).
      // but the dimension of the derived State given by function
      // toState.
      // More generally it is the dimension of the one-particle space
      // over which FieldStateD1 happens to be the Fock space

   Bio bio()const{ return b_;}

   C get(KyD1 const& k)const;
      // gets the value of the wave function at 'occupation index'
      // i.e. at a list of discrete positions given by k.

   C operator[](KyD1 const& k)const{ return get(k);}
      // notice that return m_[k] would not define a constant operator
      // since it sets m_[k]=0 in cases (i.e. for values of k) in which
      // m_[k] was not given a value earlier.

  // C operator[](Z i)const;
      // returns a sum of all those values m_[k] for which k[i]>0.
      // This is OK for rough visualization of the state. If one would
      // use the probability density instead of the amplitude density
      // ('the wave function') this would under the name marginal
      // distribution of the probability density of all particles.
      // Proper index range is 0,1,...s_-1

 //  StateD toState(bool ignorePhases)const;
      // If ignorePhases==0, this returns a State s such that
      // s.dim()=size() and s[i]=(*this)[i] for all i \in {0,...s_-1}
      // with operator[] just introduced. If ignorePhases!=0 the
      // resulting State res describes the 'marginal probability density'
      // i.e. res[i] is the probability to find a particle at place i
      // irrespective of other particles being arround elsewhere.

 //  StateD den()const;
      //: density
      // probability density as a distribution over the s_ discrete
      // points.
      // The square root could be interpreted as a real state that
      // is conveniently to display.
      // Meaning has to be checked, clarify connection to Bethe-Salpeter
      // amplitudes.

 //  Observable toObservable(bool ignorePhases)const;

 //  Type crt(Z i)const;
      // crt =creator = creation operator
      // 'creates an oscillator quantum at lattice position i.
      // Notice that for i out of range (with respect to the state
      // to which the operator is being applied) effectively a lattice
      // point within the proper range is used (considering the
      // index range a torus)

//   Type des(Z i)const;
      // destructor (not in the sense of C++) = destruction operator.
      // The basic property of crt(i) and des(j) is:
      //    psi.crt(i).des(j)-psi.des(j).crt(i)==psi*krn(i,j)
      // which corresponds to
      //    [a(i),a*(j)]=[des(i),crt(j)]=krn(i,j)*1
      // taking into account the order of application given by
      // postfix notation.
      // Notice that for i out of range (with respect to the state
      // to which the operator is being applied) effectively a lattice
      // point within the proper range is used (considering the
      // index range a torus).
      // Old explanation which holds for bosons only,
      // see des(State const& st) for the inclusion of fermions.

   Type crt(StateD1 const& st)const;
      // The natural situation is st.dim()==size().
      // equals sum of crt(i)*st(i)

   Type crt(Z i, Z a)const{ return crt(StateD1(i,a,b_));}

   Type des(StateD1 const& st)const;
      // The natural situation is st.dim()==size().
      // equals sum of des(i)*st(i).con()

   Type des(Z i, Z a)const{ return des(StateD1(i,a,b_));}

      // For R eps=Ky::getEps();
      // and State s1,s2; FieldStateD1 fs; s1.dim()==s2.dim()==fs.size();
      // fs.crt(s2).des(s1)+fs.des(s1).crt(s2)*eps-fs*(s1|s2)==0

 //   Type phi(Z i)const{  return (crt(i)+des(i))*facPhi;}
       // The couple (phi,psi) is the natural self-adjoint
       // derivate of the mutually adjoint couple (crt,des)
       // of operators.
       // [phi(i),psi(j)]=C(0,krn(i,j))*1

 //  Type psi(Z i)const{ return (crt(i)-des(i))*facPsi;}
      // see comment to previous function phi

   Type phi(StateD1 const& st)const{  return (crt(st)+des(st))*facPhi;}
      // Going from from crt(i),des(j) to crt(s1), des(s2) is
      // analogous to going from psi(i),phi(j) to psi(s1),phi(s2)

   Type psi(StateD1 const& st)const{ return (crt(st)-des(st))*facPsi;}
      // For R eps=Ky::getEps();
      // and State s1,s2; FieldStateD1 fs; s1.dim()==s2.dim()==fs.size();
      // C sp12=((s1|s2)-(s1|s2).con()*eps)*C(0,0.5);
      // fs.psi(s2).phi(s1)+fs.phi(s1).psi(s2)*eps-fs*sp12==0;

//   Type cook1(Observable const& ob)const;
      // Lets the observable ob (assuming ob.dim()==size()) act
      // on *this by the Fock-Cook construction. Since the matrix
      // elements of ob get automatically periodically continued,
      // it is not necessary that ob.dim()==size() holds.

//   Type cook2(Observable const& ob)const;
      // Lets the observable ob (assuming ob.dim()==size()) act
      // on *this by the Fock-Cook construction. Since the matrix
      // elements of ob get automatically periodically continued, it is
      // not necessary that ob.dim()==size() holds.

 //  Type manyParHam(R2_Func const& pot1, R2_Func const& pot2, R t)const;
      //: 'many particle Hamiltonian'
      // returns a FieldStateD1 which results from application of
      // a many-body hamiltonian with position dependent single particle
      // potential pot1 and pair potential pot2. Both potentials are
      // functions of two variables (instances of R2_Func), where the
      // second argument is time. The first argument is an absolute
      // particle position in case of pot1 and a relative position of
      // a particle pair in the case of pot2.
      // Positions are always given by QMBase::posNor(d) and thus are in
      // the range [-1/2,+1/2]

   Type ham(R const& m, R_Vector const& pot=R_Vector())const;
      //: Hamiltonian

//   Type hub(R a, R b, R t)const;
        // Hamiltonian of the Hubbard model

   R nor_(R r=1);
      // returns the norm of *this and changes *this into
      // a FieldStateD1 of norm r by multiplying with a real factor.
      // If *this was the zero FieldStateD1, it will be turned into
      // a 'standard State' of norm r

   R sup()const;
      // max of absolute value of components

   Type reduce(Z d=32)const;
      // returns a state res that no occupation number higher
      // than d occurs

//   void show(Frame& f, Z mult_i=8, Z mult_j=4,
//      Z iShift=0, Z jShift=0, bool ignorePhases=false)const;
      //: show
      // This is the comment to the obsolete version. A valid
      // implementation of graphical representation is still missing.
      // notice that the state space has a very high dimension even
      // if s_ and Ky::lim_ have moderate values.
      // one component of rep will be represented as a rectangle of
      // mult_i*mult_j pixels. The whole image will be shifted by
      // (iShift,jShift) with respect to the left upper corner
      // of the graphics window f.
      // Actually the components of rep are not converted to a screen
      // color directly but before get renormalized by a real factor
      // chosen such that the dominant component of rep gets an absolute
      // value of 1. If the true values of the components of a state s
      // are to be shown use
      // Graph g;
      // g.show(s.toRealForm())
      // If ignorePhases differs from 0, what is represented is the
      // probability density instead of the complex-valued wave function.

  //    void mark(Graph& f, Z ipolMeth=1, R gamma=0.5, R secFac=1.1)const;
         //: mark
         // See State2::mark in cpmstate2.h for the meaning of the
         // arguments. Shows the two-particle component of Fock-state
         // *this by finally calling
         // State2::mark(f,ipolMeth,gamma,secFac).

 //  X3<Ky,Z1,R1> dgn()const;
      // diagnosis for state *this. The first component of the result
      // gives the maximum occupation number found at each lattice point,
      // the second component gives the number of non-zero components,
      // and the third component gives the norm of the state.

   Type makeReal()const;
      // returns a FieldStateD1 with all components of *this
      // replaced by their absolute value

 //  static R test(Z com);
      //: test
      // test of the most important relations.
      // returns the sum of errors (also the unavoidable
      // small numeric ones, so that only large values
      // indicate malfunction.

//   static R graphicsTest(Z com);
      //:: graphics test
      // test of graphic representation modalities.
inline FieldStateD1 dot(OpFD1 const& op)const;
   };

////////////////// class OpFD1 //////////////////////////////////////////

class OpFD1{ // operators in Dirac space as functions
   typedef OpFD1 Type;
   Bio b_;
   Fa<FieldStateD1,FieldStateD1> f_;
public:
   CPM_NAM(OpFD)
   OpFD1(){}
   explicit OpFD1(Bio const& b);
      // The zero-operator
   OpFD1(Bio const& b, F<FieldStateD1,FieldStateD1> const& f):b_(b),f_(f){}

   bool mat(OpFD1 const& op)const{ return b_==op.b_;}
      //: match
   bool mat(FieldStateD1 const& psi)const{ return b_==psi.bio();}
      //: match
   F<FieldStateD1,FieldStateD1> f()const{ return f_;}
   Bio bio()const{ return b_;}
      // biotope

   static OpFD1 crt(StateD1 const& st);
      //: creator
   static OpFD1 des(StateD1 const& st);
      //: destructor
   static OpFD1 crt(Z i, Z a, Bio const& b);
   static OpFD1 des(Z i, Z a, Bio const& b);
   static OpFD1 ham(Bio const& b, R const& m, R_Vector const& pot);
   static OpFD1 cook(OpD1 const& op);

   OpFD1 operator+(OpFD1 const& op)const
   { cpmassert(mat(op),"OpFD1 +"); return OpFD1(b_,f_+op.f_);}

   OpFD1 operator-(OpFD1 const& op)const
   { cpmassert(mat(op),"OpFD1 -"); return OpFD1(b_,f_-op.f_);}

   OpFD1 operator-()const
   { return OpFD1(b_,-f_);}

   OpFD1 operator*(C const& z)const
   { return OpFD1(b_,multR<FieldStateD1,FieldStateD1,C>(z,f_));}

   OpFD1 operator*(OpFD1 const& op)const
   { cpmassert(mat(op),"OpFD1 *"); return OpFD1(b_,f_.circ(op.f_));}

   OpFD1& operator*=(OpFD1 const& op)
   { cpmassert(mat(op),"OpFD1 *="); return *this=OpFD1(b_,f_.circ(op.f_));}

   FieldStateD1 operator*(FieldStateD1 const& psi)const
   { cpmassert(b_==psi.bio(),"OpFD1 * FieldStadeD1");return f_(psi);}

   FieldStateD1 operator()(FieldStateD1 const& psi)const
   { cpmassert(b_==psi.bio(),"OpFD1(StadeD1)");return f_(psi);}

   R norm()const; // operator norm
      //: norm
   R dtLim()const{return R(2)/norm();}
};

inline FieldStateD1 FieldStateD1::dot(OpFD1 const& op)const
{ return op(*this)*C::Ibar;}
// d/dt psi = H psi/i => i d/dt psi = H psi => psi(t) = exp(-i t H) psi(0)

} // namespace

#endif
