//? cpmlogic1.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_LOGIC1_H_
#define CPM_LOGIC1_H_
/*
   Description: Class Version of Propositional Logic.
      First example of a class based on expression trees.
      Coding method for expressions taken from Chapter 8 of
      A. Koenig, B. Moo: Ruminations on C++, Addison-Wesley 1996
*/
#include <cpms.h>

namespace CpmMathFound{ // foundation of mathematics
   
   using namespace CpmStd;

   using CpmRoot::Word;
   using CpmRoot::Z;
   using CpmArrays::S;
   using CpmArrays::V;
   using CpmArrays::VV;
   using CpmRootX::B;
   using CpmRootX::Z1;

class Prop{ // atomic propositions 
   // 'atomic' propositions characterized by a number
   // and a truth value. Suited as template argument for
   // PropLog<>
   B b_;
   Z1 i_;
   static Z count_;

public:
   typedef Prop Type;
   CPM_ORDER;
   CPM_IO;

   Prop():i_(1){} // no count activities
   explicit Prop(Z i):i_(i){if (i_>count_) count_=i_; if (i_<1) i_=1;}
      // creates a Prop x such that
      // x.toWord()="pk" where k is the character string denoting
      // the number i_, and x.eval()=false
   Prop(Z i, bool b):i_(i),b_(b)
      {if (i_>count_) count_=i_; if (i_<1) i_=1;}
      // as before but x.eval()=b_
   explicit Prop(Word w): i_(++count_){}
      // creates (independent of the value of w) a Prop which is
      // different from any Prop created before
   explicit Prop(Word w, bool b):i_(++count_),b_(b){}
      // as before but the truth value

   void setValue(bool b){ b_=b;}

   bool eval()const{ return b_;}
   Z index()const{ return i_;}
   Word toWord()const{ return "p"&i_.toWord();}
   Word nameOf()const{ return "Prop";}
   static Z count(){ return count_;}
};

//////////// class Expr_node<> ///////////////////////////////////////

template <class X>
class PropLog;

template <class X>
class Expr_node{ // expression nodes
   friend class PropLog<X>;
   Z use; // direct implementation of reference counting
protected:
   Expr_node():use(1){}
   virtual ~Expr_node(){}
   virtual void setValue(const V<B>&)=0;
   virtual Word toWord()const =0;
   virtual Word nameOf()const =0;
   virtual bool eval()const=0;
};

//////////// class PropLog ///////////////////////////////////////////
/* The mathematical structure of PropLog<X> is beautiful. I never found
it spelled out in full elegance.
We consider the class B as enriched by the operations 'and', 'or' and
'not'. We have the diagram:

             f1
     X ------------> B
  g1 |               | g2
     v       f2      v
  PropLog<X> ------> B

where f1 is arbitrary, g1 is given by the constructor
PropLog<X>(const X&) which inserts X into PropLog<X>, and f2,g2 is a
closure to a commutative diagram such that f2 and g2 are morphisms with
respect to the 'and','or', and 'not' operations. 
f1 and f2 are named eval in our application.
Since PropLog<X> consists of expressions in X, eval as a function on X
gets recursively extended to whole PropLog<X>. This logic seems to
work always if only we consider the same operations on B and on
PropLog<X>.
Actually, it is the raison d'etre of expressions that their evaluation
is a well defined operation if the evaluation of the nodes is defined.

It looks to me that the interpretation of this structure as Logic
and, as such, as a foundation of mathematics is misleading. Here
the structure is coded in a formal language (which C++ is).

See cpmtests.h, Test_logic<> for a verification of this structure
for selected points.

*/

template <class X> class Leaf;
template <class X> class Unary_node; 
template <class X> class Binary_node;

template <class X>
class PropLog{ // propositional logic
   // creates propositional logic for the case that the
   // atomic formulae are of type X
   Expr_node<X> *p;

   PropLog(Expr_node<X>* p_):p(p_){}
      // not safe, thus private

   S<X> atoms;
      // set of different atomic formulas on which
      // the expression depends. This object is to be known
      // for generating all truth-value assignments
public:
// Koenigs stuff arround reference counting
   ~PropLog(){if (--p->use==0) delete p;}
      // destructor
   PropLog<X>& operator=(const PropLog<X>& rhs)
   { atoms=rhs.atoms; rhs.p->use++; if (--p->use==0) delete p; p=rhs.p;
    return *this;}
      // assignement
   PropLog(const PropLog<X>& t){ atoms=t.atoms;p=t.p; ++p->use;}
      // copy constructor

   PropLog(){ p=new Leaf<X>(X());}
      // default constructor, together with the previous 3 functions, the
      // value interface is implemented

   explicit PropLog(const X& x){ p=new Leaf<X>(x); atoms.add(x);}
      // constructor for a atomic proposition

   static V< PropLog<X> > list( Z n);
      // creates a list of instances each of which is independent
      // af all instances created before

   static VV< PropLog<X> > matrix(Z m, Z n);
      // creates a matrix of instances each of which is independent
      // af all instances created before
   
   PropLog<X> non()const;
      // returns the negation of *this
   PropLog<X> et(const PropLog<X>& rhs )const;
      // returns *this 'and' rhs
   PropLog<X> vel(const PropLog<X>& rhs )const;
      // returns *this 'or' rhs
   PropLog<X> aut(const PropLog<X>& rhs )const
   { return (non().et(rhs)).vel((*this).et(rhs.non()));}
      // returns *this 'exclusive or' rhs 
   PropLog<X> implies(const PropLog<X>& rhs )const{ return non().vel(rhs)\
   ;}
      // returns *this --> rhs
   PropLog<X> iff(const PropLog<X>& rhs )const
   { return implies(rhs).et(rhs.implies(*this)); }

   static PropLog<X> xorV(const V<PropLog<X> >& vp);
      // returns a proposition which 
      // true iff exactly one of the vp[i] is true. This is not
      // simply vp[1].xor(vp[2]).xor(vp[3]),...... See 
      // implementation code for the formal definition

   static PropLog<X> orV(const V<PropLog<X> >& vp);
      // returns a proposition which is vp[1].vel(vp[2]).vel(vp[3]),...

   static PropLog<X> andV(const V<PropLog<X> >& vp);
      // returns a proposition which is vp[1].et(vp[2]).et(vp[3]),...
   
   typedef PropLog<X> Type;
   CPM_IO
   CPM_ORDER
      // IO and ORDER not yet implemented; not trivial but the lines seem
      // clear
   Word toWord()const{ return p->toWord();}
      // textual representation of the instance *this
   Word nameOf()const;
      // textual representation of the type or class
   bool eval()const{ return p->eval();}
      // returns the truth value which results from the imported
      // truth-value assignment to the atoms
   void setValue(const V<B>& vb){ p->setValue(vb);}
      // the atom at with at.index()=i on which *this depends
      // will get the truth value vb[i]. It was hard work to discover
      // how to access the leafs of expression *this if it is
      // finished. The solution is given by virtual functions
      // in the node classes.
   bool taut();
      // returns true iff *this is tautological

   bool implied(const V< PropLog<X> >& vp);
      // returns true if vp implies *this, or, expressed
      // differently, *this is a consequence (is implied,
      // thus the name) by vp.

   S<X> getAtoms()const{ return atoms;}
      // returns the set of atoms of *this
};

template <class X>
V< PropLog<X> > PropLog<X>::list( Z n)
{
   V< PropLog<X> > res(n);
   res[1]=PropLog<X>(X(""));
   for (Z i=2;i<=n;i++) res[i]=PropLog<X>(X(""));
   return res;
}

template <class X>
VV< PropLog<X> > PropLog<X>::matrix(Z m, Z n)
{
   static Z mL=4;
   static Word loc("PropLog<X>::matrix(Z m, Z n)");
   cpmmessage(mL,loc&" started");
   cpmassert(m>0,loc);
   cpmassert(n>0,loc);
   Z i,j,mn=m*n;
   Z k=0;
   V< PropLog<X> > resLin=list(mn);
   VV< PropLog<X> > res(m,n);
   for (i=1;i<=m;i++) for (j=1;j<=n;j++) res[i][j]=resLin[++k];
   cpmmessage(mL,loc&" done");
   return res;
}

template <class X>
Word PropLog<X>::nameOf()const
{
   X x;
   Word w1="Expression of Propositional Logic with propositions of type ";
   Word w2=x.nameOf();
   return w1&w2;
}

template <class X>
PropLog<X> PropLog<X>::non()const
{
   PropLog<X> res(new Unary_node<X>("non",*this));
   res.atoms=atoms;
   return res;
}

template <class X>
PropLog<X> PropLog<X>::et(const PropLog<X>& rhs)const
{
   PropLog<X> res(new Binary_node<X>("et",*this,rhs));
   res.atoms=atoms|rhs.atoms;
      // was & till 2006-06-25
   return res;
}

template <class X>
PropLog<X> PropLog<X>::vel(const PropLog<X>& rhs)const
{
   PropLog<X> res(new Binary_node<X>("vel",*this,rhs));
   res.atoms=atoms|rhs.atoms;
   return res;
}

class V_bool: public V<B>{ // array of bool-values, helper class
public:
   V_bool(const V<B>& vb):V<B>(vb){}
   V_bool(Z n):V<B>(n){}
   void next(); 
      // changes *this into the next instance
      // (same dim, according to lexicographic
      // order derived from false < true
   bool end()const; // returns true for the last vb
      // which is (true,true,...,true)
};

template <class X>
bool PropLog<X>::taut()
{
   Z mL=2;
   Word loc("PropLog<X>::taut()");
   cpmmessage(mL,loc&" started");
   Z counter=0;
   Z cStep=10000;
   Z count=0;
   Z dat=atoms.dim();
   Z d=atoms[dat].index();
   V_bool vb(d);
   Z cnt=X::count(); // static function
   cpmmessage(mL,"Number of Bits="&cpm(d)&", count="&cpm(cnt));
   for (;!vb.end();vb.next()){
      counter++;
      if (counter==cStep){
         count++;
         cpmmessage(mL+1,"count="&cpmwrite(count));
         counter-=cStep;
      }
      setValue(vb);
      if (!eval()) return false;
   }
   cpmmessage(mL,loc&" done");
   return true;
}

/*
template <class X>
bool PropLog<X>::implied(const V< PropLog<X> >& vpx)
{
   PropLog<X> known=andV(vpx);
   PropLog<X> toShow=known.implies(*this);
   return toShow.taut();
}
*/

template <class X>
bool PropLog<X>::implied(const V< PropLog<X> >& vpx)
// considerably faster than the previous trivial implementation
{
   Z mL=2;
   Word loc("PropLog<X>::implied(V< PropLog<X> >)");
   cpmmessage(mL,loc&" started");
   Z counter=0;
   Z cStep=10;
   Z count=0;
   S<X> s=atoms;
   V< PropLog<X> > vp=vpx;
   Z i,np=vp.dim();
   for (i=1;i<=np;i++){
     // s=(s|(vp[i].getAtoms())); // successive union
     s.add(vp[i].getAtoms()); // successive union
   } 
   Z n=s.dim();
   Z ns=s[n].index();
   V_bool vb(ns);
   cpmmessage(mL,"Number of atoms="&cpmwrite(n)
   &", nBits="&cpmwrite(ns));
   bool b,bVal;
   bool first=true;
   bool last=false;
Label:
   counter++;
   if (counter==cStep){
      count++;
      cpmmessage(mL+1,"count="&cpmwrite(count));
      counter-=cStep;
   }
   if (first) first=false; else { vb.next(); last=vb.end();}
      // the initial truth-value assignment vb=(false,false,....false)
      // has to be taken as it is and not just replaced by vb.next()
   for (i=1;i<=np;i++){
      vp[i].setValue(vb);
      bVal=vp[i].eval();
      if (!bVal) goto Label;
         // then the truth-value assignment gives not true for all
         // premisses; so we can't learn from this truth-value assignment
         // and turn to the next one
   }
   // now the premisses are all true for the present values of vb
   setValue(vb);
   b=eval();
   if (!b){
      return false;
      // for present vb the premisses are true and *this is false
   }
   else if (!last){
      goto Label;
   }
   else{
      return true;
   }
}

template <class X>
PropLog<X> PropLog<X>::xorV(const V< PropLog<X> >& vp)
{
   Z i,j,k,d=vp.dim();
   if (d==0){
      return PropLog<X>();
   }
   else if (d==1){
      return vp[1];
   }
   else if (d==2) {
      PropLog<X> res=vp[1].aut(vp[2]);
      return res;
   }
   else if (d==3) {
      V<PropLog<X> > resP(d);
      resP[1]=vp[1].et(vp[2].non()).et(vp[3].non());
      resP[2]=vp[2].et(vp[3].non()).et(vp[1].non());
      resP[3]=vp[3].et(vp[1].non()).et(vp[2].non());
      return orV(resP);
   }
   else{ // the previous case in automated form
      V<PropLog<X> > resP(d);
      for (i=1;i<=d;i++){
         resP[i]=vp[i];
         k=i+1;
         for (j=1;j<d;j++){
            if (k>d) k=1;
            resP[i]=resP[i].et(vp[k].non());
            k++;
         }
      }
      return orV(resP);
   }
}

template <class X>
PropLog<X> PropLog<X>::orV(const V< PropLog<X> >& vp)
{
   Z i,d=vp.dim();
   if (d==0){
      return PropLog<X>();
   }
   else if (d==1){
      return vp[1];
   }
   else{
      PropLog<X> res=vp[1];
      for (i=2;i<=d;i++) res=res.vel(vp[i]);
      return res;
   }
}

template <class X>
PropLog<X> PropLog<X>::andV(const V< PropLog<X> >& vp)
{
   Z i,d=vp.dim();
   if (d==0){
      return PropLog<X>();
   }
   else if (d==1){
      return vp[1];
   }
   else{
      PropLog<X> res=vp[1];
      for (i=2;i<=d;i++) res=res.et(vp[i]);
      return res;
   }
}

/////////////////////// class Leaf<> /////////////////////////

template <class X>
class Leaf: public Expr_node<X>{ // leafs in expression trees
   friend class PropLog<X>;
   X atom;
   Leaf(const X& x):atom(x){}
   void setValue(bool b_){ atom.setValue(b_);}
   Word toWord()const{ return "("&atom.toWord()&")";}
   Word nameOf()const{ return "node0";}
   bool eval()const{ return atom.eval();}
   void setValue(V<B> const&);
};

template <class X>
void Leaf<X>::setValue(const V<B>& vb)
{
   Z i=atom.index();
   Z vbd=vb.dim();
   if (i<1) cpmerror("Leaf<X>::setValue(const V<B>& vb): i<1, i="
      &cpm(i));
   if (i>vbd){
      cpmerror("Leaf<X>::setValue(const V<B>& vb): i>vb.dim(), i="
      &cpm(i)&" vb.dim()="&cpm(vbd) );
   }
   atom.setValue(vb[i]);
}

/////////////////////// class Unary_node<> ///////////////////////////////

template <class X>
class Unary_node: public Expr_node<X>{//unary nodes in expression trees
   friend class PropLog<X>;
   Word op;
   PropLog<X> opnd;
   Unary_node(const Word& op_,const PropLog<X>& opnd_):
      op(op_),opnd(opnd_){}
   Word toWord()const;
   Word nameOf()const{ return "node1";}
   bool eval()const;
   void setValue(const V<B>&);
};

template <class X>
Word Unary_node<X>::toWord()const
{
   return "("&op&opnd.toWord()&")";
}

template <class X>
bool Unary_node<X>::eval()const
{
   if (op=="non"){
      return !opnd.eval();
   }
   else{
      cpmerror("Unary_node<X>::eval() un-valid node type");
      return false;
   }
}

template <class X>
void Unary_node<X>::setValue(const V<B>& vb)
{
   opnd.setValue(vb);
}

/////////////////////// class Binary_node<> //////////////////////////////

template <class X>
class Binary_node: public Expr_node<X>{// binary nodes in expression trees
   friend class PropLog<X>;
   Word op;
   PropLog<X> left;
   PropLog<X> right;
   Binary_node(const Word& op_, const PropLog<X>& le, 
      const PropLog<X>& ri): op(op_),left(le),right(ri){}
   Word toWord()const;
   Word nameOf()const{ return "node2";}
   bool eval()const;
   void setValue(const V<B>&);
};

template <class X>
Word Binary_node<X>::toWord()const
{
   return "("&left.toWord()&op&right.toWord()&")";
}

template <class X>
bool Binary_node<X>::eval()const
{
   if (op=="et") return left.eval()&&right.eval();
   else if (op=="vel") return left.eval()||right.eval();
   else {
      cpmerror("Binary_node<X>::eval() un-valid node type");
      return false;
   }
}

template <class X>
void Binary_node<X>::setValue(const V<B>& vb)
{
    left.setValue(vb);
   right.setValue(vb);
}

//////////////////// class PropLogic /////////////////////////////

typedef PropLog<Prop> PropLogic;

} // namespace

#endif
