//? cpmoperator.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_OPERATOR_H_
#define CPM_OPERATOR_H_ 
/*
   Description: objects for describing quantum mechanical operators
      as time-dependent functions from Space to Space
*/

#include <cpmtrj.h>

namespace CpmQuantumMechanics{

   using namespace CpmArrays;
   using namespace CpmFunctions;
   using namespace CpmGraphics;
   using CpmRootX::B;
   using CpmFunctions::R_Func;

////////////////////// class Operator /////////////////////////

class Operator: public QMBase{ // observables as functions 
   typedef Operator Type;
   typedef X2<State,R> XSR;
   Z d_;
   F< XSR ,State> f_;

   static State conInTime(XSR const& xsr, F<State,State> const& f)
   { return f(xsr.c1());}

   static State mul2(XSR const& xsr, Operator const& op1, 
      Operator const& op2)
   {  return op1.f_(XSR(op2.f_(xsr),xsr.c2()));}

   static State add2(XSR const& xsr, Operator const& op1, 
      Operator const& op2)
   { return op1.f_(xsr)+op2.f_(xsr);}

   static State dif2(XSR const& xsr, Operator const& op1, 
      Operator const& op2)
   { return op1.f_(xsr)-op2.f_(xsr);}

   static State mulC(XSR const& xsr, Operator const& op, C const& z)
   { return op.f_(xsr)*z;}

   static State mulR(XSR const& xsr, Operator const& op, R const& r)
   {  return op.f_(xsr)*r;}

   Operator( Z d, F<XSR,State> const& f):d_(d),f_(f){}
   Operator(Z d, F<State,State> const& f);
    
public:

   CPM_PURE_VECTOR_SPACE_INTERFACE(Type,C)
   CPM_PURE_SCALAR_INTERFACE(Type,R)
   CPM_NAM(Operator)

   Operator operator*(Operator const& op)const;

   Operator(){}

   explicit Operator(Observable const& ob);

   State actOn(State const& st, R t)const{ return f_(XSR(st,t));}
      // Operator acts on state at time t

   State operator()(State const& st, R t=0)const
      // Operator acts on state at time t
      { return f_(XSR(st,t));}
      
   State operator*(XSR const& xsr)const{ return f_(xsr);}
      // another form of actOn;

   Observable toObservable(R t)const;

   static Operator sinParHam(R2_Func const& pot,Z d, R lambda=1.);

   static Operator freePar(Z d);
};

/////////////////////// class TrjDMI_Opr /////////////////////////////////
// trajectory for time-dependent Hamiltonians.
// similar to TrjDMI but replaces Observables by
// Operators (in the definition of which time-dependence
// is allowed) and which are much more efficient in higher-
// dimensional spaces. Also, the dimension of the space
// needs not to be a power of two 

class TrjDMI_Opr: public TrjDMI<State>{ // trajectories det. by operators 
   Operator opr_;
   virtual State op(State const& x, R t)const{ return opr_.actOn(x,t);}
public:
   typedef TrjDMI_Opr Type;
   virtual TrjBase* clone()const{ return new TrjDMI_Opr(*this);}
   TrjDMI_Opr(){}
   TrjDMI_Opr(Operator const& opr, State const& x, TrjCon const& tc):
   TrjDMI<State>(x,tc),opr_(opr){}
   virtual Word text()const{ return Word("TrjDMI_Opr");}
};

} // namespace

#endif
