//? cpmqmx20180827.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_QMX_H_
#define CPM_QMX_H_
/*
   Description: An attempt to reconstruct the treatment of spin chains
      as given as an example in my article on separated quantum mechanics
      from 2008.
*/

#include <cpmv.h>
#include <cpmc.h>
#include <cpmgraph.h>
#include <cpmrecordhandler.h>
#include <cpmclinalg.h>
#include <map>

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmRoot::Word;
using CpmRoot::randomR;

using CpmRootX::B;
using CpmRootX::RecordHandler;

using CpmGeo::Iv;

using CpmArrays::V;
using CpmArrays::X2;
using CpmArrays::X3;
using CpmArrays::R_Vector;
using CpmArrays::Outside;

using CpmFunctions::F;

using CpmGraphics::Graph;

using CpmLinAlg::R2;
using CpmLinAlg::R3;
using CpmLinAlg::C2;
using CpmLinAlg::C4;
using CpmLinAlg::CC2;
using CpmLinAlg::CC4;
using CpmLinAlg::CCCC2;

using std::map;
using std::vector;

// interfaces as declaration macros

#define CPM_QM_PURE_VECTOR_SPACE_INTERFACE(X,S)\
public:\
   X operator - (void)const;\
   X& operator *=(S);\
   X& operator +=(X const&);\
   X& operator -=(X const&);\
   X operator *(S)const;\
   X operator +(X const&)const;\
   X operator -(X const&)const;

#define CPM_QM_PURE_SCALAR_INTERFACE(X,S)\
public:\
    X& operator *=(S);\
    X operator *(S)const;

#define CPM_QM_SCALAR_PRODUCT_INTERFACE(X,S)\
public:\
   X conj(void)const;\
   S operator |(X const&)const;\
   R normSqr(void)const{ return (operator|(*this)).real();}\
   R norm(void)const{ return ::sqrt(normSqr());}\
   R absVal(void)const{ return norm();}\
   R absSqr(void)const{ return normSqr();}

namespace CpmQM{

// Within this namespace we alway use units in which the numerical values
// of hDirac and c are 1. hPlanck thus has value 2*pi (=CpmRoot::Pi2).
// Three structs for compact input from an ini-file

////////////////////////////// class SysConx //////////////////////////////

   struct SysConx{ // run control
      Z np;
      R h;
      R qXY;
      R qZ;
      R fracRim;
      R phi0;
      R theta0;
      R flipProbability;
   public:
      SysConx(RecordHandler const& rch, Word const& sec)
      {
         cpmrhf(np);
         cpmrhf(h);
         cpmrhf(qXY);
         cpmrhf(qZ);
         cpmrhf(fracRim);
         cpmrhf(phi0);
         cpmrhf(theta0);
         cpmrhf(flipProbability);
      }
   };

////////////////////////////// class RunConx //////////////////////////////
   struct RunConx{ // run control
      R dt;
      Z dtAdjMethod;
      Z nIter;
      R eps;
      B adalf;
      B enfNor; // enforce normalization
      R tWaitIter;
      Z nSteps;
      Z viewPeriod;
      Z docuPeriod;
      R tWaitIni;
      R tWaitView;
      R tWaitDocuNormError;
      R tWaitDocuEnergyError;
      R tWaitRes;
    public:
      RunConx(RecordHandler const& rch, Word const& sec)
      {
         cpmrhf(dt);
         cpmrhf(dtAdjMethod);
         cpmrhf(nIter);
         cpmrhf(eps);
         cpmrhf(adalf);
         cpmrhf(enfNor);
         cpmrhf(tWaitIter);
         cpmrhf(nSteps);
         cpmrhf(viewPeriod);
         cpmrhf(docuPeriod);
         cpmrhf(tWaitIni);
         cpmrhf(tWaitView);
         cpmrhf(tWaitDocuNormError);
         cpmrhf(tWaitDocuEnergyError);
         cpmrhf(tWaitRes);
      }
   };

////////////////////////////// class DisConx //////////////////////////////
   struct DisConx{ // display control
      Z ipolMet;
      Z colMet;
      R gamma;
      R secFac;
      B autoScl;
      B showCor;
         //: show correlations
      B showCls;
         //: show classical
         // show also corresponding data for classical particles
      B areaSty;
         //: area style (prefer representation as a color
         // image ove line graphics
      B wrtOnShow;
         //: write on show
   public:
      //DisConx():ipolMet(-1),gamma(1),secFac(-1),autoScl(true){}
      DisConx(RecordHandler const& rch, Word const& sec)
      {
         cpmrhf(ipolMet);
         cpmrhf(colMet); // colorizing Method
         cpmrhf(gamma);
         cpmrhf(secFac);
         cpmrhf(autoScl);
         cpmrhf(showCor);
         cpmrhf(showCls);
         cpmrhf(areaSty);
         cpmrhf(wrtOnShow);
      }
   };

/////////////////////////// class Iax /////////////////////////////////////

class Iax{ // interaction descriptor for spin one half
   // Interaction descriptor for spin 1/2
      CC2 h1_;
         // one-particle interaction
      CCCC2 h12_;
         // pair interaction
   public:
      Iax(R h, R qXY, R qZ): h1_(CC2::sig3*h),
      h12_(CCCC2::spinSpinXYZ(qXY, qZ)){}

      Z dim()const{ return 2;}
         //: dimension
      C2 operator()(C2 const& x)const{ return h1_*x;}
      C2 operator()(C2 const& x, C2 const& y)const
      { return h12_(x,y).trc1(x);}

      C avr(C2 const& x, C2 const& y)const
         //: average
      { return (y|operator()(x,y));}

   };

//////////////////////////// class SpinNx ////////////////////////////////

   class SpinNx{ // spin chain quantum system
      Z np_;
      Iax ia_;
      V<C2> vx_;
      V<C2> vv_;
      R t_{0.};
      Z sc_{0}; // step counter
      RunConx rc_;
      typedef SpinNx Type;

   public:
      CPM_NAM(SpinNx)

      static V<C2> dot(V<C2> const& vx, Iax const& ia);
      // returns the velocities determined by the arguments

      static R nvx(V<C2> const& vx){
         Word loc("SpinNx::nvx(V<C2>)");
         Z mL=3;
         CPM_MA
         R r=1.;
         for (Z i=vx.b();i<=vx.e();++i){ //r*=vx.cui(i).abs();
            R ri=vx.cui(i).abs();
            if (ri==0.) cpmwarning("ri=0, i="&cpm(i));
            r*=ri;
         }
         CPM_MZ
         return r;
      }

      static V<C2> iniState(SysConx const& sc){
         V<C2> res(sc.np,C2(sc.theta0,sc.phi0));
         Z dRim=cpmrnd(sc.np * sc.fracRim);
         R flip=sc.flipProbability;
         for (Z i=1;i<=sc.np;++i){
            if (randomR()<flip && i>dRim && i+dRim<=sc.np) res[i].swap_();
         }
         return res;
      }

      SpinNx(V<C2> const& vx, R h, R qXY, R qZ, RunConx const& rc):
         np_(vx.dim()), ia_(h,qXY,qZ), vx_(vx), rc_(rc)
      {
         for (Z i=1;i<=np_;++i){
            vx_.cui(i).nor_(); // ensuring that the vx_.cui(i) are
               // unit vectors
         }
         vv_=dot(vx_,ia_);
         Z dtAdjMethod=rc_.dtAdjMethod;
         if ( dtAdjMethod==0) ;
         else if (dtAdjMethod==1) rc_.dt*=cpmsqrt(np_)*cpminv(eTot().abs()\
         );
         else adj_dt_();
      }

      R getEps()const{ return rc_.eps;}
      void setEps_(R eps){ rc_.eps=eps;}

      R get_dt()const{ return rc_.dt;}
      void set_dt_(R dt){ rc_.dt=dt;}

      R t()const{ return t_;}
         //: t

      Iax ia()const{ return ia_;}

      Z parDim()const{ return 2;}
         //: particle dimension

      R n()const
         //: norm
         // Notice that vx_ is the tensor product of the states
         // vx_.cui(i). This implies the product structure of the norm.
      {
         //R r=1;
         //for (Z i=1;i<=np_;++i) r*=vx_.cui(i).abs();
         //return r;
         return nvx(vx_);
      }

      C sclPrd(V<C2> const& vx)const
         //: scalar product
      {
         C z(1.);
         for (Z i=1;i<=np_;++i) z*=(vx_.cui(i)|vx.cui(i));
         return z;
      }

      C hInt()const;
         //: h interaction (h for hamiltonian)

      R hInt_i(Z i)const;
         //
      R eSep_i(Z i)const;
      R eTot_i(Z i)const;
      R theta_i(Z i)const{ return vx_[i].blcAng()[1];}
      R phi_i(Z i)const{ return vx_[i].blcAng()[2];}
      R2 blc_i(Z i)const{ return vx_[i].blcAng();}

      R sig3()const
      {
         R res=0;
         for (Z i=1;i<=np_;++i){
            C2 ci=vx_.cui(i);
            res+=(ci|CC2::sig3*ci)[1];
         }
         return res/np_;
      }

      R3 sig(Z i)const;
         //: sigma
         // expectation values of the three Pauli matrices
         // at chain position i

      V<C2> col()const;
         //: color

      R2 blcAng()const
         //: Bloch angles
         // returns mean values of Bloch angles over the chain
      {
         R r1=0, r2=0;
         for (Z i=1;i<=np_;++i){
            R2 ri=vx_.cui(i).blcAng();
            r1+=ri[1];
            r2+=ri[2];
         }
         R npInv=1./np_;
         return R2(r1*npInv,r2*npInv);
      }

      C eSep()const;

      C eTot()const{ return eSep()+hInt();}

      void step_();
         //: step
         // implemented as by means of stepLean

      void stepLean_();
         //: lean step

      void mark(Graph& g, DisConx const& dc)const;
         //: mark

      void run_(Graph& g, DisConx const& dc);
         //: run

      void adj_dt_();
         //: adjust dt
         // Assumes that dt_ as set by the constructor. Successively halves
         // the steps till the rc_.nIter steps run fine. For security the
         // step is then halved again to give it the final value.

      void lineMark(Graph& g, DisConx const& dc)const{
         using CpmImaging::Color;
         using CpmLinAlg::Z2;
         static Z jShift=1;
         V<C2> wx=col();
         Z gw=g.w();
         Z gh=g.h();
         Z pix=gh/np_;
         Z mult_i=pix;
         Z mult_j=pix;
         Z js=pix,is=pix;
         Z i,j;
         Z d=np_*mult_i;
         jShift+=js;

         Z iState=1;
         Z iCounter=0;
         Z iEff=is;
         Color c;

         for (i=1;i<=d;i++){
            if (iCounter==0){
               C val;
               if (dc.colMet==1){ // Bloch angles based coloring
                  val=wx.cui(iState).blc();
               }
               else if (dc.colMet==2){
                  val=wx.cui(iState)[1];
               }
               else if (dc.colMet==3){
                  val=wx.cui(iState)[2];
               }
               else if (dc.colMet==4){
                  val=wx.cui(iState)[1]*wx.cui(iState)[2];
               }
               else val=C(1.);
               c=Color(val);
            }
            iCounter++;
            if (iCounter==mult_i){
               iCounter=0; // triggers new evaluation of c in the next loop
               iState++;
            }
            for (j=0;j<mult_j;j++){
               g.put(Z2(iEff,(j+jShift)%gw),c);
            }
               iEff++;
         }
      }
  };

} // namespace

#endif
