//? cpmset.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_SET_H_
#define CPM_SET_H_
/*
   Description: Defines a class of finite sets, free of
      any further typing. Of course, it is implied that also
      the elements are finite sets, and so on. Such sets are called
      'hereditarily finite'. This is a version without 'ur-elements' so
      that any set can be written as a uniquely determined finite list of
      symbols '{' and '}'.

   History: Started 2002-11-13 after preliminary implementation in RUBY.
   2005-07-16 a lot of functions added:
      V<Set> getValues(Set const& x)const;
      Set getFunctionValue(Set const& x)const;
      static Set makeNumericFunction(Z n, const F<Z,Z>& f);
      static Set makeCyclicGroup(Z n);
      Z numberOfBraces()const{ return val.size();}
      static Set makeNumber(Z n);
      bool isGroup()const;
      static bool test1(Z n);
      static bool test2(Z n);
      That test2 runs successfully for all arguments <=10
      (for n=10 the Set encoding of S(10) is by 43190 braces
      which translates into 43190 bits) creates confidence that the
      main functions (one needs all of the work horses in the group
      test) are correctly implemented.
      To do: Graphical representation as a mountain chain given by the
      brace level.
   2005-08-02 attribute isValid eliminated, attribute n changed to n_
   2005-11-03 adding to the comments to n_ and add
   2008-04-27 extending 'General considerations'
   2013-01-29 During the last week I got makeCyclicGroup(n)
      and isGroup working and a representation of Set instances
      as a color image (for very large Sets, as they result from
      makeCyclicGroup(30) this is more effective than a linear
      representation. Many of the functions mentioned in the
      2005-07-entry turned out not to be implemented, especially
      the functions test1 and test2. Presumably I unintentionally
      somehow changed the latest version of cpmset.cpp against an
      earlier, incomplete, version.
*/

#include <cpmnumbers.h>
#include <cpmword.h>
#include <cpmv.h>
#include <cpmf.h>
#include <vector>
#include <cpmrvector.h>
#include <cpmimg24.h>

////////////////////////// class Set ///////////////////////////////////
/*
General considerations:
2005-08-02
The way how more complex mathemathical entities emerge from simpler ones
shows striking similarities with the way in which more complex biological
entities emerge from simpler ones. At least in the object oriented
approach to mathematical entities, each such entity has a characteristic
finite system of symbols that determines its 'state' and that can be
compared with the DNA molecule in the nucleus of a living cell or -
sometimes more appropriately - with a gene
(which is a subsection of such a DNA molecule).
Boolean values are the most simple 'genes' in mathematics. Finite lists
of those are the natural numbers in binary coding (see cpmnr.(h,cpp)).
'Stone-age coding' by strings of a single character
(see cpmmathfound.(h,cpp)) is a very vasteful encoding of natural
numbers.
In order to build a hierarchy of mathematical notions which is easy to
use, one should consider boolean values and natural numbers already
defined before considering sets. Having B and N (natural symbols for the
boolean type and the natural number type) it is very important to define
for arbitrary types T the type V<T> of finite sequences of quantities of
type T. If this all is given, we have a king's way to sets, which then
provide a universal framework for mathematics. Actually, already
N:={0,1,2,...} can be obtained as V<B> e.g. with the correspondence
0 <--> false, 1 <---> true and by the agreement that a V<B> string b
indexed as (b_p,b_{p-1},...,b_1,b_0) corresponds to the number
n = 2^{p} * b_p + 2^{p-1} * b_{p-1} + ... + 2^{1} * b_1 + 2^{0} * b_0.
Here, for each string b with more than one component (i.e. p>0)
we have b_p=1.
*/

namespace CpmArrays{

   using std::string;
   using std::vector;
   using CpmRoot::Z;
   using CpmRoot::Word;
   using CpmFunctions::F;
   using CpmGraphics::Img24;

// auxiliar torus class that implements as operator *
// the addition of integers modulo n

class Set{  // represents general finite sets
   // The public interface of the class uses the classes bool, Z,
   // V<bool>, V<Set>, and F<Z,Z>. In mathematics one would call a set
   // finite even if each of its finitely many elements would be infinite.
   // So {N,Z,Q,R,C,H} is a finite set. This is obviously not ment here.
   // Set represents what in Mathematics today is called
   // 'Grothendieck universe of hereditarily finite sets'.

   static char a;
      // character which represents '{' in a set theoretical expression

   static char z;
      // character which represents '}' in a set theoretical expression

// independent data
   string val;
      // representation of *this as a string of characters a and z
      // It is the complete reresentation, thus {} for the void set
// dependent data
   Z n_;
      // number of elements;
      // too dangerous to name this number simply n as I did till
      // 2005-08-01. Actually I had some dubious code using n as a local
      // variable and as a function argument
      // not independent, equals elm.size()

   vector<string> elm;
      // to get to elm one has to remove the outer
      // pair of braces to get to the list of elements. There are no
      // separators in the list and none are needed since proper pairing
      // of the braces uniqely determines the subunits. These subunits
      // than are exactly the val data members of the elements of *this.
      // The process which creates elm from val ensures that the
      // order in which these elements appear in elm agrees with the
      // lexicographic order of a,z-valued strings (a<z, of course).

// end of data

   typedef Set Type;

// data all well typed; therefore the copy constructor and assignment
// operator will be defined correctly by the compiler

   void ini();
      // initializes elm from val

   void iniFromElm();
      // initializes val from elm

   static void unwrap(string&);
      // taking appart the enclosing braces

   explicit Set(const string& str):val(str){ ini();}
      // Making a set from its val data member.
      // Use only if str is well formed. Thus private.

   explicit Set(const vector<string>& vs):elm(vs){ iniFromElm();}
      // Making a set from its elm data member.
      // Use only if vs is well formed. Thus private.

   string rep()const{ return val;}

public:

   CPM_ORDER_PLAIN
   virtual Set* clone()const{ return new Set(*this);}
   CPM_IO
   Word nameOf()const{ return "Set";}

// constructors.

   Set(void);
      // creates the void set

   explicit Set(V<Set> const& vs);
      // creates a set the elements of which are just the components of
      // vs (clearly, multiply appearing components get used only once,
      // otherwise we would not get a set)

   explicit Set(Z n, bool start1=true);
      // The natural n-element set: void for n<=0
      // For n>=1 it has as elements the sets
      // Set::makeNumber(1),....,Set::makeNumber(n) for start1=true
      // Set::makeNumber(0),....,Set::makeNumber(n-1) for start1=false
      // Notice that for all sets of numbers indexing is in falling order:
      // For instance
      // Set s(6);
      // for (Z i=1;i<=s.dim();++i) cout<<s[i].number()<<endl;
      // returns (up to newline characters) 6 5 4 3 2 1. If we like to get
      // the numbers in natural order, i.e. 1 2 3 4 5 6,  we may write
      // for (Z i=s.dim();i>=1;--i) cout<<s[i].number()<<endl;

// cardinality access
   Z card()const{ return n_;}
   Z dim()const{ return n_;} // for uniformity with other array types
   Z size()const{ return n_;} // for uniformity with STL code

// element read access

   Set operator[](Z i)const;
      // Returns the i-th element if  i \in {1,...card()}
      // else error

   Set first()const{ return (*this)[1];}
      // returns the first element of *this

   Set last()const{ return (*this)[n_];}
      // returns the last element of *this

   Set operator()(Z i)const;
      // continuation as constant below and above the limits
      // Thus no error for any value of i

   V<Set> operator()(void)const;
      // returns an array of all elements in natural order

// read access to internals

   string getVal()const{return val;}
   string getElm(Z i)const{ return elm[i-1];}

// set formation operations

   Set push()const;
   // push is the basic creative method. It returns {uv...w}
   // for *this=uv...w
   // note that the returned set has just one element

   V<Set> pop()const;
    // Let the result be denoted res. If *this has just one element,
    // res.dim()==1 and res[1] is just this element. Otherwise
    // res.dim()==0.
    // For all sets s thus (s.push().pop()).dim()==1 and
    // (s.push().pop())[1]==s
    // Notice thar returning a V<Set> instead of a Set is here
    // only a way of exception handling.

   Set operator |(Set const& s)const;
      // forming the union of *this and s (corresponds to the
      // 'or' operation for indicator functions)

   static Set uni(V<Set> const& vs);
      // union (unfortunately 'union' is a key-word in C++ so it can't be
      // used as a name) of the sets in the family vs

   static Set sec(V<Set> const& vs);
      // section of the sets in the family vs

   Set unionOfElements()const{ return uni(operator()());}
      // since the elements of *this are sets, we are allowed to
      // form the set union of them all. This set then is returned.

   Set sectionOfElements()const{ return sec(operator()());}
      // since the elements of *this are sets, we are allowed to
      // form the set section of them all. This set then is returned.

   Set subSet(const V<bool>& vb)const;
      // Let Set s;
      // s[i] will be an element of s.subSet(vb) iff vb[i]==true

   Set powerSet()const;
      // set of all subsets of *this

   Set add(Set const& s)const{ return (*this)|(s.push());}
      // adds s as an element (if not yet there)
      // The function f: Set x Set --> Set, (x,y) |--> x.add(y) plays
      // a decisive role in genertating Set:
      // Assume Set s=....
      // Then with
      // V<Set> vs=s();
      // Z n=vs.dim();
      // we have s=f(...f(f(vs[1],vs[2]),vs[3])...,vs[n])
      // Notice also that {s}=f({},s). So it is quite clear that starting
      // from {} as an element of Set and forming f(x,y) for yet formed
      // sets x,y one obtains any set. Thus in Backus Naur Notation:
      // <Set> ::= {}|f(<Set>,<Set>)

   Set operator &(Set const& s)const;
      // forming the section of *this and s (corresponds to the
      // 'and' operation for indicator functions)

   Set operator -(Set const& s)const;
      // forming *this\s i.e. the set of all those elements of
      // *this which don't belong to s (relative complement)

   Set operator ^(Set const& s)const;
      // forming the symmetric difference of s1 and s2
      // (corresponds to 'exclusive or' for indicator functions)

// boolean operations

   bool isVoid(void)const{ return n_==0;}
      // returns true if *this is void

   bool hasElem(Set const& s)const;
      // returns true iff s is an element of *this

   bool ni(Set const& s)const{ return hasElem(s);}
      // returns true iff s is an element of *this (same as previous
      // function).
      // from LATEX symbol \ni for the mirror image of the epsilon-like
      // 'is element'-symbol

   bool in(Set const& s)const{ return s.hasElem(*this);}
      // returns true iff *this is an element of s
      // from LATEX symbol \in for the epsilon-like 'is element'-symbol

   Z locate(Set const& s)const;
      // returns 0 if s is not an element of *this
      // and its index if it is an element of *this

   bool isSubSetOf(Set const& s)const;
      // returns true iff *this is a subset of s

   bool isSuperSetOf(Set const& s)const;
      // returns true iff s is a subset of *this

   Set wkp(Set const& y)const;
   // generates the Wiener Kuratowski pair
   // {{*this},{*this,y}}=:WKP(*this,y)
   // notice WKP(x,y):={{x},{x,y}} ==> WKP(x,x)={{x}}
   // see e.g. Paul Taylor: Practical Foundations of Mathematics,
   // Cambridge 1999, p. 121

   Set orderedPair(Set const& y)const{ return wkp(y);}
      // more explicit name for previous function

   Set operator *(Set const& y)const;
      // Cartesian product *this times y
      // x*y={WKP(a,b) | a \in x and b \in b}

   V<Set> getTwoFactors()const;
      // Let res be the return value.
      // Then res.dim()==0 if *this is not a WKP.
      // Else *this=WKP(res[1],res[2])
      // Consider Set p=s1.wkp(s2).wkp(s3).wkp(s4);
      // V<Set> v=p.twoFactors(). Then
      // v[1]=s1.wkp(s2).wkp(s3), v[2]=s4
      // The factorization is unique and splits off always the last
      // factor of a multiple product; we would never get factors like
      // s1.wkp(s2), s3.wkp(s4)

   static Set tupel(V<Set> const& vs);
      // creates a n-tupel (n=vs.dim()) by iterating the WKP
      // construction:
      // Lets write [x,y]:=WKP(x,y)
      // tupel(x,y,z):=[[x,y],z]
      // tupel(x,y,z,z'):=[tupel(x,y,z),z'].
      // This is a well defined procedure in which no definition problems
      // can arrise.

   static Set car(V<Set> const& vs);
      // cartesian product of all these sets
      // vs[1]*vs[2]*...*vs[n] consists of all
      // tupel[e1,...en] with e1 \in vs[1],...en \in vs[n]
      // Further: Set::car(vs)=(...((vs[1]*vs[2])*vs[3])...*vs[n])

   V<Set> getFactors()const;
      // decomposing a tupel into its composing components (factors since
      // tupel formation is analogous to multiplication in some way)
      // tupel(V<Set> const& vs).getFactors()==vs
      // For the result res we have res.dim()==0 if *this is not a tuple
      // Implementation contained a severe error which was eliminated
      // 2013-01-27.

// topic: relations and their algebra

   bool isRelation()const;
      // returns true iff all elements of *this are wkp's

   V<Set> decompose()const;
      // returns a list of subsets
      // where res[i] is the subset of *this of those elements which
      // are i-tuples

   V<Set> extractRelation()const;
      // If *this contains wkps as elements, the subset of those
      // will be returned as res[1]; res[2] will be the domain of
      // relation res[1], and res[3] will be the range of it.
      // If no element of *this is a wkp, res will be returned as
      // V<Set>(0). Notice that some of the elements of *this
      // which end up in res[1] may be decomposable further, but this
      // is not the task of the present function.

   bool isFunction()const;
      // returns true iff all elements of *this are wkp's
      // and if there apear no two different wkp's which agree
      // in their first component

   V<Set> getValues(Set const& x)const;
      // returns a list of y's such that (x,y) is an element of *this.
      // If no such y exists (e.g. if *this has no ordered pair as
      // element at all) then the returned list has length 0.
      // Notice that returning the void set would not be safe in this
      // case since the void set could be a legal return value of a
      // function.

   Set getFunctionValue(Set const& x)const;
      // returs the uniquely defined function value if it exists
      // and fails if it does not exist

 //  static V<Set> makeFunction(V<Set> const& rel);
      // let rel be a result from extract Relation.
      // If rel.dim()==0 we return V<Set>(0). Else we
      // return the mapping from res[2].powerSet() to res[3].powerSet()
      // which is induced by relation rel[1].

   static Set makeNumericFunction(Z n, const F<Z,Z>& f);
      // returns the the set
      // {(1,f(1)),...,(n,f(n)}
      // Of course, here 1 is a shorthand for Set::makeNumber(1),
      // and f(n) for Set::makeNumber(f(n)).

   Word toWord()const{ return Word(val);}

   Z numberOfBraces()const{ return (Z)val.size();}

// coding natural numbers (zero included) as sets

   static Set makeNumber(Z n);
      // Set::makeNumber(0)={}, Set::makeNumber(1)={{}},
      // Set::makeNumber(2)={{{}}}, ...
      // If n is negative, we set makeNumber(n):=makeNumber(-n) to define
      // the function for all values of the argument
      // Notice: the universal void set is here being identified
      // with the number 0.

   Z number()const;
      // returns the representation of *this as a standard number if
      // *this is of the form {{{{{... }}}}}....; else it returns -1.
      // Notice that Z allows to hold 'exceptional values' since in this
      // case only values >=0 can appear as regular values. So we can
      // deviate from the establishe method which in this case would be
      // to return a V<Z> of dim 0 in the irregular case and of dim 1 in
      // the regular one which then would hold the regular value as the
      // first component. We will have to return to this method for
      // coding boolean values since there are only two such values
      // which leaves no room for coding 'nil' as a boolean value

   bool isNumber()const{ return number()>=0;}

// coding boolean values as sets

   static Set makeBoolean(bool b)
      { return (b ? makeNumber(1) : makeNumber(0));}

   V<bool> boolean()const
   {
      if (*this==makeNumber(0)) return V<bool>(1,false);
      else if (*this==makeNumber(1)) return V<bool>(1,true);
      else return V<bool>(0);
   }

   bool isBoolean()const
   { return !boolean().isVoid();}

   R_Vector grpRep()const;
   //: graphical representation
   // So far not very instructive linear representation
   // realizing my first idea for a graphical representation.
   // For larger sets, this asks for too large a screen for
   // faithful rendition.

   Img24 imgRep(R aspRat, R prob=0., Z mult=1)const;
   //: image representation
   // The linear two-values array val is visulized
   // line by line in a square color image where the
   // two values 'a' and 'z' are represented as red
   // and green color. The format is governed by aspRat
   // ( = width/height). Since the number val.car() of
   // data to be represented needs not fit exactly an
   // rectangle of this aspect ratio, the last row of
   // pixels may not be populated. They are shown in blue
   // color.
   // Valid values of the second argument are in [0,1].
   // If prob>0 the pixel color is under partial control
   // of a random generator. For prob = 1 this control
   // is total. The reason for introducing this second argument
   // is that it allows an interesting phenomenon to be studied:
   // Wheras the sequence of pixels along a line is a faithful
   // representation of the set (as written out in terms of braces)
   // the sequence of pixels in columns is artificially set by
   // the line length (which again is determined by aspRat).
   // Given that, one may be surprised to see in the image
   // representation of some sets, e.g. those generated as
   // Set::makeCyclicGroup(n) (e.g. n = 10, 16, 30), contiguous
   // pixel regions which extend in vertical direction even
   // stronger than in horizontal direction. Also the size of some
   // of these contiguous is surprisingly large. By giving prob
   // the value 1 one gets a binary purely stochastical picture
   // with much smaller contiguous monochrome areas. Giving prob
   // intermediate values on gets images with partial stochastic
   // influence which may assist understanding.
   // Giving mult a value > 1 represents each pixel of the original
   // image as a mult x mult square which is a easy way to get
   // well printable output even after converting to JPG.

// coding groups as sets

   static Set makeCyclicGroup(Z n);
      // returns the cyclic group of order n, coded as the set
      // ({0,...n-1},product,inversion,unitelement)
      // Tests whether the functions provided so far are expressive
      // enough for defining this conveniently. The answer is yes!

   bool isGroup()const;
      // returns true if *this is a group, else false
      // We agree that a group is a 4-tupel (X,mult,inv,e), where
      // mult: XxX-->X, inv: X-->X, e \in X and the following
      // properties hold:
      // mult(x,mult(x',x")) = mult(mult(x,x'),x") for all x,x',x" \in X
      // mult(x,inv(x)) = e for all x \in X
      // mult(x,e) = x for all x \in X

// test functions

   static bool test1(Z n);
      // Concerned with making a group and testing whether it is one.
      // Returns true if the test was done and everything turned out to
      // be correct, otherwise it returns false.

   static bool test2(Z n);
      // Concerned with graphical representation, here of a cyclic
      // group.

   //static bool test3(Z n);
      // concerned with making a number and testing whether it is one
};

} // namespace

/*********************** discussion of Set ********************************
I discuss the relation of Set to ZFC. My source is the excellent
presentation Ebbinghaus,Flum,Thomas: Einführung in die mathematische\
 Logik.
Spektrum 1998, pp. 117 ff.

Extensionality: Although not by definition, s1 == s2 implies
a) s1.card()==s2.card() and s1[i]==s2[i] for 1<=i<=s1.card().
On the other hand, if a) is valid for s1, s2 we certainly have s1==s2.
So, extensionality holds for Set.

pair formation: The constructor explicit Set(V<Set> const& vs);
for vs.dim()==2 creates the pair set {vs[1],vs[2]}

union of elements: The function
   Set unionOfElements()const;
creates it

power set: made by function Set powerSet()const;

infinity: Of course, there are no instances of Set that have infinitely
many elements. So the axiom of infinity does not hold in Set.
On the other hand, for any natural number n there are many instances of
Set that have just n elements.

choice: Trivially valid. Let V<Set> vs a family of pair-wise disjoint
non-void sets.
Define: Set f(Set const& s){ return s[1];}
Then Set(vs(f)) is the set which contains from each vs[i] just one element.

'Aussonderung': Trivial for finite sets. Since the set-defining boolean
expression can applied to the finitely many elements of the given set in
turn. Set subSet(const V<bool>& vb)const; does just this.

replacement: Trivial for finite sets.

************************************************************************/

#endif
