//? cpmstate4.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_STATE4_H_
#define CPM_STATE4_H_
/*
   Description: Development and test of a class
      that represent quantum mechanical pure states of spin 1/2 particles.
*/

#include <cpmstate.h>
#include <cpmclinalg.h>
#include <cpmgraph.h>
#include <cpmrecordhandler.h>

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmRootX::B;
using CpmRootX::RecordHandler;
using CpmGraphics::Frame;
using CpmGraphics::Graph;
using CpmLinAlg::Z2;
using CpmLinAlg::C4;
using CpmLinAlg::CC4;
using namespace std;
using namespace CpmArrays;
using namespace CpmFunctions;


namespace CpmQuantumMechanics{

   class Y{
      // auxiliar class
      // Task is defining function State4::dot(...)const,
      // or put in different words, the Hamiltonian.

      R m_; // mass
      V<R> x_; // list of 1D positions
      F<R,R> pot_; // scalar potential
      V<R> phi_; // dependent, made from x_ and pot_
         // list of potential values
      R h_; // spatial discretization length, made from x_
      B dmi_; // direct midpoint integrator, initialized as false
      B nr_; // non-relativistic dynamics, initialized as false
   public:
      Y(R m, V<R> const& x, F<R,R> const& pot):
        m_(m),x_(x),pot_(pot),phi_(x_(pot_)),
        h_((x_.last()-x_.fir())/(x_.dim()-1)){}
      Z dim()const{ return x_.dim();}
      R get_h()const{ return h_;}
      R get_m()const{ return m_;}
      void set_dmi(bool b){ dmi_=B(b);}
      bool get_nr()const{ return nr_;}
      void set_nr(bool b){ nr_=B(b);}
      bool get_dmi()const{ return dmi_;}
      V<R> get_phi()const{ return phi_;}
      void showPhi(Graph g)const;
         //: show phi
         // showing the potential phi_

   };

struct ShowCon
// show control
{
   bool autoScale;
      // if true, the data xr,yr have no influence
   Iv xr;
      // x-interval selected for display
   Iv yr;
      // y-interval selected for display
   S<Z> ir;
      // a subset of {1,2,3,4,5,6,7,8} for
      // selecting among the 8 real components (= 4 complex ones)
      // of the Dirac wave function
   R fac;
      // magnification factor for visualization of deviations
      // from norm conservation
   R secFac;
      //: security factor
      // Factor between -1 and +1 for setting dt from the
      // stability limit
   R tWait,tWait2,tWait3;
   Z imgPeriod;
   Word appendix;
   ShowCon():autoScale(true),fac(1.),secFac(0.85),tWait(0.),
      tWait2(0.),tWait3(0.){}
      // default constructor
      // all non-class data explicitely initialized
   ShowCon(RecordHandler const& rch, Word sec);
      // work horse constructor from ini-file entries
};

class State4{
// Dirac electron in 1D

   V<C4> rep_;
      // Array with indexes starting with 1
      // List of Dirac amplitudes
   Iv iv_;
      // interval along 1-axis in space as biotope

   V<R> x_;
      // list of positions

public:
   typedef State4 Type;
   CPM_IO
   Word nameOf()const{ return "State4";}

   State4():rep_(),iv_(){}

   State4(V<C> const& v1, V<C> const& v2, V<C> const& v3,
      V<C> const& v4, Iv const& iv);

   State4(V<C4> const& rep, Iv const& iv):
      rep_(rep),iv_(iv),x_(iv_.cen(rep_.dim())){}

   State4(Z n, Iv const& iv);
      // generates random wave functions, not normalized,
      // on an n-point lattice

   State4(Z n, Iv const& iv, T4< F<R,C> > const& f);

   State4 operator+(State4 const&)const;
   State4& operator+=(State4 const& s){ return *this = *this+s;}
   State4 operator-(State4 const&)const;
   State4& operator-=(State4 const& s){ return *this = *this-s;}
   State4 operator-()const;
   State4 con()const;
      //: (complex) conjugate
   State4 d()const;
      // Differentiation
   C operator |(State4 const&)const;
      // scalar product
   State4 operator*(C const&)const;
   State4 operator*(R const&)const;
   R absSqr()const{ return (*this|*this)[1];}
      //: absolute (value) squared
   R abs()const{ return cpmsqrt(absSqr());}
   R nor_()
      //: normalize
      // changes *this into a unit-vector and returns the
   {
      R r=abs();
      R rInv=cpminv(r);
      *this=(*this)*rInv;
      return r;
   }

   State4 nor()const
      //: normalize
   {
      R r=abs();
      R rInv=cpminv(r);
      return (*this)*rInv;
   }

   State4 rev()const;
   State4 beta()const;

   State4 setPhi(V<R> const& phi)const;

   void boost_(R frac);

   State4 dot(Y const& y)const;
      //: dot (symbol for time derivation)
      // time derivative according to Dirac's equation
   C enr(Y const& y)const{ return (*this|(*this).dot(y)*C::I);}
   Z dim()const{ return rep_.dim();}
   V<R> get_x()const{ return x_;}
   V<C4> get_rep()const{ return rep_;}
   Iv get_iv()const{ return iv_;}
   void showAsCurves(Graph& g, ShowCon const& sc)const;

   R enrNorm(Y const& y, R eps=1e-6, Z iterMax=1000,
      bool randomize=true)const;
   //: energy norm
   // Operatornorm of the Hamilton operator. It gets calculated by
   // iterated application of dot(y) on either *this or a random state
   // which shares all data with *this except the components of rep_.

};


class Trj4{
   R t_;
   State4 x_;
   Y y_;
   State4 v_;
   ShowCon sc_;
   R dt_;
   C e_;
   R hNorm_;
public:
   Trj4(State4 const& x, R m, F<R,R> const& pot,RecordHandler const& rch,
      Word const& w="control"):t_(0.),x_(x.nor()),
      y_(m,x_.get_x(),pot),sc_(rch,w){adj_();}

   Trj4(R t0, State4 const& x, Y const& y, RecordHandler const& rch,
      Word const& w="control"):t_(t0),x_(x.nor()),
      y_(y),sc_(rch,w){adj_();}

   Y get_y()const{ return y_;}
   State4 get_x()const{ return x_;}
   State4 get_v()const{ return v_;}
   R get_t()const{ return t_;}
   R get_dt()const{ return dt_;}
   C get_e()const{ return e_;}
   R get_hNorm()const{ return hNorm_;}

   void set_dt(){ hNorm_=x_.enrNorm(y_); dt_=sc_.secFac*2.0/hNorm_;}
   // for sec<1 the first component of the return value should provide
   // stable
   // behavior of function run_. The second component of the return value
   // is the norm of the energy operator (Hamiltonian).

   void set_dmi(bool b){y_.set_dmi(b);}
   bool get_dmi()const{ return y_.get_dmi();}

   void rev_(){dt_=-dt_;}

   void adj_(){ v_=x_.dot(y_); set_dt();e_=(x_|v_)*C(0,-1);}
      //: adjust


   void step_();
      // most economic form of the densified step of the
      // asynchronous leap-frog method

   void runLean_(R tFinal, Graph g);
      //: run lean
      // no recording of control curves concerning energy conservation
      // and norm conservation

   void run_(R tFinal, Graph g);
      //: run
      // tFinal: till that time the run lasts.
      // dt: time step, g visualization tool

   void concentrate_();
      //: eliminates the 'sister wave packet'.

};

} // namespace

#endif
