//? cpmsttd1 (Kopie).h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_STTD1_H_
#define CPM_STTD1_H_
/*
   Description: Development and test of a class SttD1.
      Its scope is the same as StateD1. The implementation systematically
      makes use of the  Eigen libraries data types for matrices and
      vectors. In my boost series application it was by a factor 2.5
      faster than the version built on StateD1. The adverse aspect of this
      dependency is that the class does not work for multiple precision
      arithmetics.
*/

#include <cpmv.h>
#include <cpmc.h>
#include <cpmtypes.h>
#include <cpminterfaces.h>
#include <cpmrfunc.h>
#include <cpmconstphys.h>
#include <cpmgraph.h>
#include <cpmobservable.h> // contains cpmstate.h
#include <cpmclinalg.h>
#include <cpmfile.h>
#include <cpmbiotope.h> // defines class Bio
#include <cpmsystemdependencies.h>

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmGraphics::Frame;
using CpmGraphics::Graph;
using CpmGeo::Iv;
using CpmLinAlg::Z2;
using CpmLinAlg::R2;
using CpmLinAlg::C2;
using CpmLinAlg::CC2;
using CpmSystem::OFileStream;

using namespace std;
using namespace CpmArrays;
using namespace CpmFunctions;
using namespace Eigen;
using std::complex;

namespace CpmQuantumMechanics{

typedef std::complex<R> Cs;
typedef Matrix<Cs,Dynamic,1> Vc;
typedef Matrix<R,Dynamic,1> Vre;
typedef Matrix<Cs,Dynamic,Dynamic> Mc;

class SttD1; // State Dirac 1-dimensional

SttD1 matOpr(SttD1 const& st, CC2 const& mat);
   //: matrix operator
SttD1 pOpr(SttD1 const& st);
   //: p operator
SttD1 xOpr(SttD1 const& st);
   //: x operator

//////////////////// class SttD1 /////////////////////////////////////////
// D for Dirac

class SttD1 { // spinor-valued wave functions on a finite 1-D lattice
   // Finite-dimensional Quantum mechanical state space with Hilbert
   // space structure and 'periodic boundary conditions'. Notice however,
   // that a fixed value of the dimension of the Hilbert space is not part
   // of the present concept.

   typedef SttD1 Type;
   typedef C ScalarType;

   Vc val_ ; // values of wave function
   Iv s_; // interval in space functioning as biotope
      // independent quantities

   Z n_;  // dimension of complex linear space, = val_size()
   Z d_;  // cardinality of the biotope lattice
   Bio b_; // number of discretization points together with s_
      // dependent quantity

   void adjust_(){ // setting the dependent member + check of consistency
      n_ = val_.size();
      cpmassert(n_%2==0,"SttD1::adjust_()");
      d_=n_/2;
      b_=Bio(d_,s_);
   }

public:

   CPM_IO
   CPM_PURE_VECTOR_SPACE_INTERFACE(Type,C)
   CPM_SCALAR_PRODUCT_INTERFACE(Type,C)
   CPM_PURE_SCALAR_INTERFACE(Type,R)
   CPM_NORMALIZE
      // Hilbert space with anti-unitary operator con()
   bool mat(SttD1 const& st)const{ return s_==st.s_ && n_==st.n_;}
      //: match
   CPM_NAM(SttD1)

   SttD1(){}
      // default constructor

   SttD1(Vc const& v, Iv const& s):val_(v),s_(s){ adjust_();}

   SttD1(V<C2> const& v, Iv const& s);

   explicit SttD1(Bio const& b);
      // Constructing the zero-vector. All components are zero spinors.

   SttD1(C const& z, Bio const& b);
      // Constructing the wave function for which all spinor components
      // have the value z.

   SttD1(F<R,C2> const& f, Bio const& b);
      // Here the values of the wavefunction are given by a function f.

   SttD1(Bio const& b, R const& xPeak, R const& sigma , R const& p,
   R const& a1, R const& a2, R const& phase1, R const& phase2,
   bool compact);

   static CC2 unit_;
   static CC2 alpha1_;
   static CC2 beta_;

   Vc const& c()const{ return val_;}
   Vc & c(){ return val_;}

   R_Vector pos()const{ return R_Vector(b_.points());}
      //: positions
      // returns the positions of the lattice points

   V<C2> val()const;
      //: values
      // returns the values or rep_ with the positions removed

   C2 val(Z i)const{
      cpmassert(i>0&&i<=d_,"");
      i--;
      return C2(C(val_(i)),C(val_(i+d_)));
   }
      //: values
      // returns value of the wave function at the

   R_Vector valVec()const{ return pos();}
      //: values (as) vector
      // bio().points() as a R_Vector

   R_Matrix valMat()const;
      //: values (as)  matrix
      // the 2 components in real representation by dividing into real and
      // imaginary part. Together with absolute value as the first row of
      // the resulting matrix.
      // Actually all these enumerated real quantities are not given by
      // their value but by the value of their squares. These are
      // probabilities and have a more direct meaning than the amplitudes.

   X2<R_Vector,R_Matrix> valVecMat()const
      {
         return X2<R_Vector,R_Matrix>(valVec(),valMat());
      }
      //: values (as) vector (and) matrix
      // the 4 components in real representation by dividing into real and
      // imaginary part. Together with absolute value as the first row of
      // the resulting matrix.

   SttD1 applyCC2(CC2 const&)const;
   //:: apply CC4
   // multiplies all spinor values with the same CC4-matrix

   SttD1 p()const;
      //. p (for momentum)
      // returns the state which results from *this by application of the
      // momentum operator.

   SttD1 operator *(R_Func const& f)const;
   // multiplies all spinor values with the values of a real function on
   // s_

   Z d()const{return b_.d();}
      // returns the cardinality of the biotope

   Z dim()const{return n_;}
      //: dimension
      // returns dimension of the complex linear state space

   Iv s()const{ return s_;}
      //. space

   Bio bio()const{ return b_;}
      //: biotope

   SttD1& nul_()
      //: nullify, non-constant function
      // sets all the spinor components of *this to zero
   {
      val_=Vc::Constant(n_,Cs(0.,0.));
      return *this;
   }

   Z b()const{ return 1;}
      //. begin (of valid indexes)

   Z e()const{ return n_;}
      //. end (of valid indexes)

   R xc(Z i)const{ return s_.cen(i,b_.d());}
      //. x coordinate
      // read access to the position coordinate by indexes

    R gap()const{ return s_.abs()/b_.d();}
      //: gap
      // distance between neighbor-points. 1/gap() is a mass.
      // It is the largest mass which can be used in the Dirac Hamiltonian.

   R lambdaMin()const{ return gap()*R(2);}

   R pMax()const{ return R(b_.d())/s_.abs();}
      // pMax := 1/gap and -pMax are the largest and the smallest
      // eigenvalue
      // of the momentum operator mom(). It can best be seen from the
      // unitary transform momFrq() which is a multiplication operator.
      // Notice that this differs from the self-suggesting ansatz
      // pMax = h/lambdaMin (since hDirac=1, we have hPlanck=2*pi)
      // = 2*pi/(2*gap) = pi/gap.

   R mMax()const{ return pMax();}
      // mMax = pMax = 1./gap
      // introduced for ease of use

   R normHa(R const& m)const{ R p=pMax(); return cpmsqrt(m*m+p*p);}
      // Exact value of normH(m).

   SttD1 ran(Z j=0)const;
      //: random
      // returns a unit vector. The vector *this enters by b_
      // and  n o t  by val_. So, *this may be a zero vector.

   R dis(Type const& psi)const;
      //: distance

   void mark(Graph& g)const;
      //: mark
      // auto-scaled polygon representation of the four complex valued
      // as real and complex part together with the absolute value of the
      // wave function

   void mark2(Graph& g, SttD1 const& sc, R const& scale=1)const;
      // Shows *this together with a comparison state
      // allows to boost the difference etween the two by a factor.

   static void markArr(V<SttD1> const& vs, Graph& g);
      // mark an array of states, where the autoscale function
      // takes all array components into account

   void gp(ostream& out, Iv const& iv=Iv())const;
      // writes data to out from which gnuplot can construct the
      // showAsCurves representation.
      // iv is an interval that will be indicated by drawing its
      // indicator function if iv is not void.

   bool isVal()const;
      //: is valid
      // is true if none of the spinor components has exceptional values
      // such as infinity

   friend
   SttD1 pOpr(SttD1 const& st){ return st.p();}

   friend
   SttD1 xOpr(SttD1 const& st){ return st*R_Func::x1;}

   friend
   SttD1 matOpr(SttD1 const& st, CC2 const& mat)
      { return st.applyCC2(mat);}

   V<SttD1> ortBase()const;
      //: orthonormal base
      // canonical base of states which live on the same biotope as *this
   static R ortTest(V<SttD1> const& vs);
      //: orthonormality test
      // returns the deviation of system vs from orthonormality
 };

 //////////////////// class ObsD1 /////////////////////////////////////////

 class SpcObjD1;

 class ObsD1{ // ObsD1s as matrices in Dirac space
   // no reference to the spatial aspects of the biotope enter here.
   // Ditect implementation of 'Matrix Mechanics'.

   typedef ObsD1 Type;
   friend class SpcObjD1;
   Mc val_;
   Z n_;
   Z d_;

   void adjust_()
   {
      n_=val_.rows();
      Z n2=val_.cols();
      cpmassert(n_==n2,"ObsD1::adjust_()");
      cpmassert(n_%2==0,"ObsD1::adjust_()");
      d_=n_/2;
   }

public:
   bool mat(ObsD1 const& obs)const{ return val_.rows()==obs.val_.rows();}
      //: match
   bool mat(SttD1 const& st)const{ return val_.rows()==st.dim();}
      //:match
   CPM_NAM(ObsD1)
   ObsD1(){}
   ObsD1(CC2 const& cc, Bio const& b);
   explicit ObsD1(Mc const& val):val_(val){adjust_();}
   ObsD1(V<SttD1> const& can, F<SttD1,SttD1> const& f);

   Z d()const{ return d_;}
   Z n()const{ return n_;}
   Mc const& val()const{ return val_;}
   Mc & val(){ return val_;}

   ObsD1 adj()const{ return ObsD1(val_.adjoint());}
   ObsD1 trn()const{ return ObsD1(val_.transpose());}

   ObsD1 operator+(ObsD1 const& op)const
   { cpmassert(mat(op),"ObsD1 +"); return ObsD1(val_+op.val_);}

   ObsD1 operator-(ObsD1 const& op)const
   { cpmassert(mat(op),"ObsD1 -"); return ObsD1(val_-op.val_);}

   ObsD1 operator-()const
   { return ObsD1(-val_);}

   ObsD1 operator*(C const& z)const
   { return ObsD1(val_*z.std());}

   ObsD1 operator*(R const& r)const
   { return ObsD1(val_*Cs(r));}

   ObsD1 sym()const{ return ((*this) + adj())*0.5;}

   ObsD1 operator*(ObsD1 const& op)const
   {
      Z mL=1;
      Word loc("ObsD1::operator*(ObsD1)");
      CPM_MA
      cpmassert(mat(op),"ObsD1 *");
      ObsD1 res(val_*op.val_);
      CPM_MZ
      return res;
   }

   SttD1 operator*(SttD1 const& psi)const
   {
      cpmassert(mat(psi),"ObsD1 * StadeD1");
      return SttD1(val_*psi.c(),psi.s());
   }

   SttD1 operator()(SttD1 const& psi)const{ return operator*(psi);}

// creation of particular (named) observables by static functions.
   static ObsD1 p(Bio const& b);
   static ObsD1 tp(Bio const& b);
   static ObsD1 tm(Bio const& b);
   static ObsD1 x(Bio const& b);
   static ObsD1 applyCC2(Bio const& b, CC2 const& mat);
   static ObsD1 h(Bio const& b, R const& m);
   static ObsD1 booster(Bio const& b, R const& m);

   R abs2()const{ return val_.squaredNorm();} // operator norm squared
   R abs()const{ return cpmsqrt(abs2());} // operator norm

};

class SpcObjD1{ //eigenvectors/values in combination
      // All eigen vectors match SttD1(b_)
      Bio b_;
      Vre sp_; // spectrum
      Mc ev_;   // eigen vectors
      Mc eva_;   // adjoint version of ev_
      Z d_;
      Z n_;
      V<SttD1> eigenvectors_;
      V<C> spectrum_;
   public:
      typedef SpcObjD1 Type;
      CPM_IO
      CPM_NAM(SpcObjD1)
      friend class ObsD1;
      Bio bio()const{ return b_;}
      Z d()const{ return d_;}
      Z n()const{ return n_;}
      V<C> eval()const{ return spectrum_;}
      V<SttD1> evec()const{ return eigenvectors_;}
/**********************
      V<C> c1()const{ return spectrum;}
      V<SttD1> c2()const{ return eigenVectors;}

      const C& operator()(Z i)const{ return spectrum[i];}
         // eigenvalues by (), notice that i has to be in
         // {1,...,l_}.
      const SttD1& operator[](Z i)const{ return eigenVectors[i];}
         // eigenvectors by [], notice that i has to be in
         // {1,...,l_}, the order is such that the ground-state of
         // the harmonic oscillator is obtained for i=l_.

      SpcObjD1():b_(),l_(0){}

      SpcObjD1(Bio const& b);
         // has the 'canonical vectors' as eigenvectors
         // and the canonical count index 1,2,....4*b.d() for spectrum
         // and eigenvectors.

************************************************/
      SpcObjD1(ObsD1 const& obs , Bio const& bio, bool test=false);
         // spectral representation of the 'operator' op. Only if op
         // is a linear function and gets applied to states which
         // match with op the result makes sense. The observable to
         // which op belongs needs to be Hermitean.

      bool mat(SttD1 const& a)const{ return a.mat(SttD1(b_));}
         //: match
         // returns true if a has the same biotope as the eigenVectors

      ObsD1 operator()(F<R,C> const& f)const;
      ObsD1 map(R_Func const& f)const;
      SttD1 apply(F<R,C> const& f, SttD1 const& st);//{ return SttD1();}
      SttD1 apply(R_Func const& f, SttD1 const& st);//{ return SttD1();}
      R_Vector sp()const;
         // spectral values from sp_ but with indexing beginning at 1

      R spectralRadius()const
      {
         R_Vector rs(n_);
         for (Z i=1;i<=n_;i++) rs[i]=spectrum_[i].abs();
         return rs.sup();
      }

      void setSpectrum_(C z){ for (Z i=1;i<=n_;i++) spectrum_[i]=z;}

      void mark(Graph& gr, Z ipolMethod=0, R gamma=0.5)const;
         // modern implementation, writes on image memory.
         // Visible only upon gr.vis().

      void markSpectrum(Graph& gr)const;
         // real part of eigenvalues, as autoscaled graph

      bool writeSpectrum(ostream& str)const;
         // writes the eigenvalues to str, returns true
         // if writing was successful

      void map_(R_Func const& f)
         // real function f acts on the real part of the spectrum
      {
         for (Z i=1;i<=n_;i++){
            C ci=spectrum_[i];
            R ri=ci[1];
            R ii=ci[2];
            spectrum_[i]=C(f(ri),ii);
         }
      }

      SttD1 operator *(SttD1 const& state)const;
      // SpcObjD acts on state just as the underlying observable.

      SttD1 operator()(SttD1 const& state)const{ return (*this)*state;}

      V<C> rep(SttD1 const& state)const;
         //: representation
         // Returns the list of development coefficients of state
         // in the basis given by the list eigenVectors.

      void map_(const F<C,C>& f)
         // f acts on the spectrum
      {for (Z i=1;i<=n_;i++) spectrum_[i]=f(spectrum_[i]);}

      void map_(const F<R,C>& f)
         // f acts on the spectrum
      {for (Z i=1;i<=n_;i++) spectrum_[i]=f(spectrum_[i].real());}

      void expi_(R t)
         // changes *this into exp(i*t*(*this))
      {for (Z i=1;i<=n_;i++) spectrum_[i]=(spectrum_[i]*C(0.,t)).exp();}

   };

} // namespace

#endif
