//? cpmsttd1.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_STTD1_H_
#define CPM_STTD1_H_
/*
   Description: Development and test of a class SttD1.
      Its scope is the same as StateD1. The implementation systematically
      makes use of the  Eigen libraries data types for matrices and
      vectors. In my boost series application it was by a factor 2.5
      faster than the version built on StateD1. The adverse aspect of this
      dependency is that the class does not work for multiple precision
      arithmetics.
      The earlier version in which I tried to have a parallel
      implementation using either Eigen or C+- is now considered obsolete.
      One has indexes starting with 0 in Eigen and starting with 1 in C+-
      and this makes a parallel treatment tedious.
*/

#include <cpmv.h>
#include <cpmc.h>
#include <cpmtypes.h>
#include <cpminterfaces.h>
#include <cpmrfunc.h>
#include <cpmconstphys.h>
#include <cpmgraph.h>
#include <cpmobservable.h> // contains cpmstate.h
#include <cpmclinalg.h>
#include <cpmfile.h>
#include <cpmbiotope.h> // defines class Bio
#include <cpmsystemdependencies.h>

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmGraphics::Frame;
using CpmGraphics::Graph;
using CpmGeo::Iv;
using CpmLinAlg::Z2;
using CpmLinAlg::R2;
using CpmLinAlg::C2;
using CpmLinAlg::CC2;
using CpmSystem::OFileStream;

using namespace std;
using namespace CpmArrays;
using namespace CpmFunctions;
//#if defined(CPM_USE_EIGEN)
using namespace Eigen;
//#endif
   // The attempt to make project dirac work without Eigen is not yet
   // finished. Probably not useful at all.
//using std::complex;

namespace CpmQuantumMechanics{

//#if defined(CPM_USE_EIGEN)
typedef std::complex<R> Cs;
typedef Matrix<Cs,Dynamic,1> Vc;
typedef Matrix<R,Dynamic,1> Vre;
typedef Matrix<Cs,Dynamic,Dynamic> Mc;
//#else
//typedef CpmRoot::C Cs;
//typedef C_Vector Vc;
//typedef R_Vector Vre;
//typedef C_Matrix Mc;
//#endif

class SttD1; // State Dirac 1-dimensional

SttD1 matOpr(SttD1 const& st, CC2 const& mat);
   //: matrix operator
SttD1 pOpr(SttD1 const& st);
   //: p operator
SttD1 xOpr(SttD1 const& st);
   //: x operator
SttD1 xcOpr(SttD1 const& st);
   //. central x operator
   // position operator relative to the center of the biotope
SttD1 ccOpr(SttD1 const& st);
   //. charge conjugation operator
SttD1 trOpr(SttD1 const& st);
   //. time reversal operator
SttD1 srOpr(SttD1 const& st);
   //. space reversal operator
   // reversal is relative to the center of the biotope

//////////////////// class SttD1 /////////////////////////////////////////
// D for Dirac

class SttD1 { // spinor-valued wave functions on a finite 1-D lattice
   // Finite-dimensional Quantum mechanical state space with Hilbert
   // space structure and 'periodic boundary conditions'. Notice however,
   // that a fixed value of the dimension of the Hilbert space is not part
   // of the present concept.

   typedef SttD1 Type;
   typedef C ScalarType;

   Vc val_ ; // values of wave function
      // val_(0),....val_(d_-1) are the first components of the
      // d_ spinors and
      // val_(d_),....val_(n_-1)
      // are the second components of the d_ spinors. To be sure: we
      // have one two-component spinor for each point of the biotope.
   Iv s_; // interval in space functioning as biotope
      // independent quantities
   Iv sup_;
      // support (i.e. the maximum domain outside of which the wave
      // function vanishes)
   Z n_;  // dimension of complex linear space, = val_size()
   Z d_;  // cardinality of the biotope lattice
   Bio b_; // number of discretization points together with s_
      // dependent quantity

   void adjust_(){ // setting the dependent member + check of consistency
      n_ = val_.size();
      cpmassert(n_%2==0,"SttD1::adjust_()");
      d_=n_/2;
      b_=Bio(d_,s_);
      sup_=s_; // always initialized this way; but can be altered
         //using function setSup_(Iv)
   }

   void spaceReversal_(){ // reordering the spinorvalues by reading them
      // in opposite direction. For the full space reversal operator see
      // method SttD1 SttD1::sr()const.
      Vc valMem = val_;
      for (Z i=1;i<=d_;++i){
         val_(i-1)=valMem(d_-i);
         val_(i-1+d_)=valMem(n_-i);
      }
   }

public:

   CPM_IO
   CPM_PURE_VECTOR_SPACE_INTERFACE(Type,C)
   CPM_SCALAR_PRODUCT_INTERFACE(Type,C)
   CPM_PURE_SCALAR_INTERFACE(Type,R)
   CPM_NORMALIZE
      // Hilbert space with anti-unitary operator con(). The 3 macros
      // with parameters are defined in cpmstate.h
   bool mat(SttD1 const& st)const{ return s_==st.s_ && n_==st.n_;}
      //: match
   CPM_NAM(SttD1)

   SttD1():n_(0),d_(0){}
      // default constructor

   SttD1(Vc const& v, Iv const& s):val_(v),s_(s){ adjust_();}

   SttD1(V<C2> const& v, Iv const& s);

   SttD1(Z i, Z a, Bio const& b);
      // Set d=b.d(). Then i \in {1,...,d} and a \in {1,2} are the
      // normal values of the indexes. For i not in this normal range
      // the function nevertheless is defined by shifting i into the
      // normal range by adding or subtracting integer multiples of
      // d. This corresponds to the toroidal topology of the biotope.
      // A corresponding normalization for a is not implemented since
      // there toroidal structure plays no role.

   explicit SttD1(Bio const& b);
      // Constructing the zero-vector. All components are zero spinors.

   SttD1(C const& z, Bio const& b);
      // Constructing the wave function for which all spinor components
      // have the value z.

   SttD1(F<R,C2> const& f, Bio const& b);
      // Here the values of the wavefunction are given by a function f.

   SttD1(Bio const& b, R const& xPeak, R const& sigma , R const& p,
   R const& a1, R const& a2, R const& phase1, R const& phase2,
   bool compact);
      // phase1 and phase2 in degrees

   static CC2 unit_;
   static CC2 alpha1_;
   static CC2 beta_;

   Vc const& c()const{ return val_;}
   Vc & c(){ return val_;}
   C operator()(Z i)const{ return C(val_(i));}
      // notice 0<=i<n_

   Iv getSup()const{ return sup_;}
      //: get support
   void setSup_(Iv const& iv){ sup_=iv;}
      //: set support

   R_Vector pos()const{ return R_Vector(b_.points());}
      //: positions
      // returns the positions of the lattice points

   V<C2> val()const;
      //: values
      // returns the values or rep_ with the positions removed

   void setVal_(Z i, Z a, C const& z){
      cpmassert(i>0&&i<=d_,"");
      cpmassert(a>0&&a<=2,"");
      i--;
      if (a==1) val_(i)=z.std();
      else val_(i+d_)=z.std();
   }

   R valSup2()const;
      // maximum of val(i).abs2()

   R valInf2()const;
      // minimum of val(i).abs2()

   C2 val(Z i)const{
      cpmassert(i>0&&i<=d_,"");
      i--;
      return C2(C(val_(i)),C(val_(i+d_)));
   }
      //: values
      // returns value of the wave function at the

   R_Vector valVec()const{ return pos();}
      //: values (as) vector
      // bio().points() as a R_Vector

   R_Matrix valMat(bool logScale=false)const;
      //: values (as)  matrix
      // the 2 components in real representation by dividing into real and
      // imaginary part. Together with absolute value as the first row of
      // the resulting matrix.
      // Actually all these enumerated real quantities are not given by
      // their value but by the value of their squares. These are
      // probabilities and have a more direct meaning than the amplitudes.

   X2<R_Vector,R_Matrix> valVecMat(bool logScale=false)const
      {
         return X2<R_Vector,R_Matrix>(valVec(),valMat(logScale));
      }
      //: values (as) vector (and) matrix
      // the 4 components in real representation by dividing into real and
      // imaginary part. Together with absolute value as the first row of
      // the resulting matrix.

   SttD1 applyCC2(CC2 const&)const;
   //:: apply CC2
   // multiplies all spinor values with the same CC2-matrix

   SttD1 p()const;
      //. p (for momentum)
      // returns the state which results from *this by application of the
      // momentum operator.

   SttD1 operator *(R_Func const& f)const;
   // multiplies all spinor values with the values of a real function on
   // s_

   SttD1 loc(Iv const& iv)const
      { return (*this)*R_Func::charFunc(iv);}
   //: localization operator
   // multiplies all spinor values with the indicator function of iv

   SttD1 cc()const;
      //. charge conjugation
      // *this |--> cc() is an anti-unitary mapping
      // Here the charge conjugation 2x2-matrix is alpha1_ (see my
      // D1-implementation in Mathematica). This acts on the complex-
      // conjugate wave function.

   SttD1 tr()const;
      //. time reversal

   SttD1 sr()const;
      //. space reversal

   Z d()const{return b_.d();}
      // returns the cardinality of the biotope

   Z dim()const{return n_;}
      //: dimension
      // returns dimension of the complex linear state space

   Iv s()const{ return s_;}
      //. space

   R cen()const{ return s_.cen();}
      //: center

   Bio bio()const{ return b_;}
      //: biotope

   SttD1& nul_()
      //: nullify, non-constant function
      // sets all the spinor components of *this to zero
   {
      val_=Vc::Constant(n_,Cs(0.,0.));
      return *this;
   }

   Z b()const{ return 1;}
      //. begin (of valid indexes)

   Z e()const{ return n_;}
      //. end (of valid indexes)

   R xc(Z i)const{ return s_.cen(i,b_.d());}
      //. x coordinate
      // read access to the position coordinate by indexes

    R gap()const{ return s_.abs()/b_.d();}
      //: gap
      // distance between neighbor-points. 1/gap() is a mass.
      // It is the largest mass which can be used in the Dirac Hamiltonian.

   R lambdaMin()const{ return gap()*R(2);}

   R pMax()const{ return R(b_.d())/s_.abs();}
      // pMax := 1/gap and -pMax are the largest and the smallest
      // eigenvalue
      // of the momentum operator mom(). It can best be seen from the
      // unitary transform momFrq() which is a multiplication operator.
      // Notice that this differs from the self-suggesting ansatz
      // pMax = h/lambdaMin (since hDirac=1, we have hPlanck=2*pi)
      // = 2*pi/(2*gap) = pi/gap.

   R mMax()const{ return pMax();}
      // mMax = pMax = 1./gap
      // introduced for ease of use

   R normHa(R const& m)const{ R p=pMax(); return cpmsqrt(m*m+p*p);}
      // Exact value of normH(m).

   SttD1 ran(Z j=0)const;
      //: random
      // returns a unit vector. The vector *this enters by b_
      // and  n o t  by val_. So, *this may be a zero vector.

   R dis(Type const& psi)const;
      //: distance

   void mark(Graph& g, bool logScale=false)const;
      //: mark
      // auto-scaled polygon representation of the four complex valued
      // as real and complex part together with the absolute value of the
      // wave function.
      // The second argument sets shows the y-values as log10(|y|).

   void mark2(Graph& g, SttD1 const& sc, R const& scale=1)const;
      // Shows *this together with a comparison state
      // allows to boost the difference etween the two by a factor.

   static void markArr(V<SttD1> const& vs, Graph& g);
      // mark an array of states, where the autoscale function
      // takes all array components into account

   void gp(ostream& out, bool logScale=false)const;
      // writes data to out from which gnuplot can construct the
      // showAsCurves representation.
      // The second argument sets shows the y-values as log10(|y|).

   bool isVal()const;
      //: is valid
      // is true if none of the spinor components has exceptional values
      // such as infinity

   friend
   SttD1 pOpr(SttD1 const& st){ return st.p();}

   friend
   SttD1 xOpr(SttD1 const& st){ return st*R_Func::x1;}

   friend
   SttD1 xcOpr(SttD1 const& st){ return xOpr(st)-st*st.cen();}

   friend
   SttD1 ccOpr(SttD1 const& st){ return st.cc();}

   friend
   SttD1 trOpr(SttD1 const& st){ return st.tr();}

   friend
   SttD1 srOpr(SttD1 const& st){ return st.sr();}

   friend
   SttD1 matOpr(SttD1 const& st, CC2 const& mat)
      { return st.applyCC2(mat);}

   V<SttD1> ortBase()const;
      //: orthonormal base
      // canonical base of states which live on the same biotope as *this
   static R ortTest(V<SttD1> const& vs);
      //: orthonormality test
      // returns the deviation of system vs from orthonormality

  };

 //////////////////// class ObsD1 /////////////////////////////////////////

 class SpcObjD1;

 class ObsD1{ // ObsD1s as matrices in Dirac space
   // no reference to the spatial aspects of the biotope enter here.
   // Direct implementation of 'Matrix Mechanics'.
   // Notice that ObsD1 does not contain anti-linear operators.
   // Charge conjugation, for instance, can only be represented as a
   // function SttD1 --> SttD1. See function SttD1::cc()const.

   typedef ObsD1 Type;
   friend class SpcObjD1;
   Mc val_;
   Z n_;
   Z d_;

   void adjust_()
   {
      n_=val_.rows();
      Z n2=val_.cols();
      cpmassert(n_==n2,"ObsD1::adjust_()");
      cpmassert(n_%2==0,"ObsD1::adjust_()");
      d_=n_/2;
   }

public:
   bool mat(ObsD1 const& obs)const{ return val_.rows()==obs.val_.rows();}
      //: match
   bool mat(SttD1 const& st)const{ return val_.rows()==st.dim();}
      //:match
   CPM_NAM(ObsD1)
   ObsD1(){}
   ObsD1(CC2 const& cc, Bio const& b);
   explicit ObsD1(Mc const& val):val_(val){adjust_();}
   ObsD1(V<SttD1> const& can, F<SttD1,SttD1> const& f);

   Z d()const{ return d_;}
   Z n()const{ return n_;}
   Mc const& val()const{ return val_;}
   Mc & val(){ return val_;}

   ObsD1 adj()const{ return ObsD1(val_.adjoint());}
   ObsD1 trn()const{ return ObsD1(val_.transpose());}

   ObsD1 operator+(ObsD1 const& op)const
   { cpmassert(mat(op),"ObsD1 +"); return ObsD1(val_+op.val_);}

   ObsD1 operator-(ObsD1 const& op)const
   { cpmassert(mat(op),"ObsD1 -"); return ObsD1(val_-op.val_);}

   ObsD1 operator-()const
   { return ObsD1(-val_);}

   ObsD1 operator*(C const& z)const
   { return ObsD1(val_*z.std());}

   ObsD1 operator*(R const& r)const
   { return ObsD1(val_*Cs(r));}

   ObsD1 sym()const{ return ((*this) + adj())*0.5;}

   ObsD1 operator*(ObsD1 const& op)const
   {
      Z mL=1;
      Word loc("ObsD1::operator*(ObsD1)");
      CPM_MA
      cpmassert(mat(op),"ObsD1 *");
      ObsD1 res(val_*op.val_);
      CPM_MZ
      return res;
   }

   SttD1 operator*(SttD1 const& psi)const
   {
      cpmassert(mat(psi),"ObsD1 * StadeD1");
      return SttD1(val_*psi.c(),psi.s());
   }

   SttD1 operator()(SttD1 const& psi)const{ return operator*(psi);}
      // An ObsD1 object obs acts on a state psi as obs(psi)

// creation of particular (named) observables by static functions.
   static ObsD1 p(Bio const& b);
      //. momentum
      // p as usual in physics

   static ObsD1 tp(Bio const& b);
      //. translation (in) positive (direction)

   static ObsD1 tn(Bio const& b);
      //. translation (in) negative (direction)

   static ObsD1 x(Bio const& b);
      //. position
      // x as usual in physics

   static ObsD1 x(Bio const& b, R const& x0);
      //. position
      // x as usual in physics
      // position relative to point x0

   static ObsD1 x(Bio const& b, R_Func const& f);
      //. function of position
      // Function f applied to the position operator in the sense
      // of function calculus of operators.

   static ObsD1 x(Bio const& b, R_Vector const& f);

   static ObsD1 applyCC2(Bio const& b, CC2 const& mat);
      //:: apply CC2
      // Function SttD1 SttD1::applyCC2(CC2 const&)const as an observable.

   static ObsD1 h(Bio const& b, R const& m);
      //. hamiltonian
      // Free Dirac Hamiltonian as an observable.

   static ObsD1 h(Bio const& b, R const& m, R_Func const& pot);
      //. hamiltonian
      // Dirac Hamiltonian with potential as an observable.

   static ObsD1 h(Bio const& b, R const& m, R_Vector const& pot);
      //. hamiltonian
      // Dirac Hamiltonian with potential as an observable.

   static ObsD1 booster(Bio const& b, R const& m);
      //:: booster
      // Domenicos's nomenclature for the operator which brings about an
      // infintesimal Lorentz tansformation (also called a 'boost')

   R abs2()const{ return val_.squaredNorm();}
      // operator norm squared

   R abs()const{ return cpmsqrt(abs2());}
      // operator norm

};

class SpcObjD1{ //eigenvectors/values in combination
      // All eigen vectors match SttD1(b_)
      Bio b_;
      Vre sp_; // spectrum
      Mc ev_;   // eigen vectors
      Mc eva_;   // adjoint version of ev_
      Z d_;
      Z n_;
      V<SttD1> eigenvectors_;
      V<C> spectrum_;
      void adjustSp_(){ // after change in specrtum_ one has to
         // change sp_ accordingly
         for (Z i=1,j=0;i<=n_;i++,j++) sp_(j)=spectrum_.cui(i).real();
      }
   public:
      typedef SpcObjD1 Type;
      CPM_IO
      CPM_NAM(SpcObjD1)
      friend class ObsD1;
      Bio bio()const{ return b_;}
      Z d()const{ return d_;}
      Z n()const{ return n_;}
      V<C> eval()const{ return spectrum_;}
      V<SttD1> evec()const{ return eigenvectors_;}
/**********************
      V<C> c1()const{ return spectrum;}
      V<SttD1> c2()const{ return eigenVectors;}

      const C& operator()(Z i)const{ return spectrum[i];}
         // eigenvalues by (), notice that i has to be in
         // {1,...,l_}.
      const SttD1& operator[](Z i)const{ return eigenVectors[i];}
         // eigenvectors by [], notice that i has to be in
         // {1,...,l_}, the order is such that the ground-state of
         // the harmonic oscillator is obtained for i=l_.

      SpcObjD1():b_(),l_(0){}

      SpcObjD1(Bio const& b);
         // has the 'canonical vectors' as eigenvectors
         // and the canonical count index 1,2,....4*b.d() for spectrum
         // and eigenvectors.

************************************************/
      SpcObjD1(ObsD1 const& obs , Bio const& bio, bool test=false);
         // spectral representation of the 'operator' op. Only if op
         // is a linear function and gets applied to states which
         // match with op the result makes sense. The observable to
         // which op belongs needs to be Hermitean.

      bool mat(SttD1 const& a)const{ return a.mat(SttD1(b_));}
         //: match
         // returns true if a has the same biotope as the eigenVectors

      ObsD1 operator()(F<R,C> const& f)const;
      ObsD1 map(R_Func const& f)const;
      SttD1 apply(F<R,C> const& f, SttD1 const& st);//{ return SttD1();}
      SttD1 apply(R_Func const& f, SttD1 const& st);//{ return SttD1();}
      R_Vector sp()const;
         // spectral values from sp_ but with indexing beginning at 1

      R spectralRadius()const
      {
         R_Vector rs(n_);
         for (Z i=1;i<=n_;i++) rs[i]=spectrum_[i].abs();
         return rs.sup();
      }

      void setSpectrum_(C z)
      {
         for (Z i=1;i<=n_;i++) spectrum_[i]=z;
         adjustSp_();
      }

      void mark(Graph& gr, Z ipolMethod=0, R gamma=0.5)const;
         // modern implementation, writes on image memory.
         // Visible only upon gr.vis().

      void markSpectrum(Graph& gr)const;
         // real part of eigenvalues, as autoscaled graph

      bool writeSpectrum(ostream& str)const;
         // writes the eigenvalues to str, returns true
         // if writing was successful

      void map_(R_Func const& f)
         // real function f acts on the real part of the spectrum
      {
         for (Z i=1;i<=n_;i++){
            C ci=spectrum_[i];
            R ri=ci[1];
            R ii=ci[2];
            spectrum_[i]=C(f(ri),ii);
         }
         adjustSp_();
      }

      SttD1 operator *(SttD1 const& state)const;
      // SpcObjD acts on state just as the underlying observable.

      SttD1 operator()(SttD1 const& state)const{ return (*this)*state;}

      V<C> rep(SttD1 const& state)const;
         //: representation
         // Returns the list of development coefficients of state
         // in the basis given by the list eigenVectors.

      void map_(const F<C,C>& f)
         // f acts on the spectrum
      {
         for (Z i=1;i<=n_;i++) spectrum_[i]=f(spectrum_[i]);
         adjustSp_();
      }

      void map_(const F<R,C>& f)
         // f acts on the spectrum
      {
         for (Z i=1;i<=n_;i++) spectrum_[i]=f(spectrum_[i].real());
         adjustSp_();
      }

      void expi_(R t)
         // changes *this into exp(i*t*(*this))
      {
         for (Z i=1;i<=n_;i++) spectrum_[i]=(spectrum_[i]*C(0.,t)).exp();
         adjustSp_();
      }

   };

} // namespace

#endif
