//? cpmh.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   Description: cpmh.cpp see cpmh.h
*/

#include <cpmh.h>

using namespace CpmRoot;
using namespace CpmLinAlg;

/////////////////////////// H //////////////////////////////////////////
R normArc(R arc)
// converts arbitrarily large arcs (positive
// or negative) back to an arc in the range (-pi,pi]
{
   const R period=cpm2pi;
   const R i_period=R(1)/period;
   R rev=arc*i_period;
   rev-=cpmfloor(rev);
    if (rev>0.5) rev-=1.;
   return (rev*period);
}

R H::normalize()
{
   R n=abs();
   if (n>0){
      R nInv=R(1)/n;
      a0*=nInv;
      a*=nInv;
   }
   return n;
}

R& H::operator[](Z i)
{
   if (i>=3) return a.x3;
   else if (i==2) return a.x2;
   else if (i==1) return a.x1 ;
   else return a0;
}

const R& H::operator[](Z i)const
{
   if (i>=3) return a.x3;
   else if (i==2) return a.x2;
   else if (i==1) return a.x1 ;
   else return a0;
}

H H::fromRot(R phi, const R3& axis)
{
   R phiHalf=normArc(phi)*0.5;
   R b0=cpmcos(phiHalf);
   R3 b=axis;
   b.normalize();
   R s=cpmsin(phiHalf);
   b*=s;
   return H(b0,b);
}

bool H::prnOn(ostream& str)const
{
   if (!CpmRoot::writeTitle("H",str)) return false;
   cpmp(a0);
   cpmp(a.x1);
   cpmp(a.x2);
   cpmp(a.x3);
   return true;
}

bool H::scanFrom(istream& str)
{
   cpms(a0);
   cpms(a.x1);
   cpms(a.x2);
   cpms(a.x3);
   return true;
}

Z H::com(const H& h)const
{
   return toR4().com(h.toR4());
}

H H::operator *(const H& h)const
   // return  H(a0*h.a0-(a|h.a), a*h.a0+h.a*a0+crsPrd(a,h.a));
   // would be more transparent, but probably less efficient
{
   return
   H(a0*h.a0-a.x1*h.a.x1-a.x2*h.a.x2-a.x3*h.a.x3,
     a.x1*h.a0 + h.a.x1*a0 + a.x2*h.a.x3-a.x3*h.a.x2,
     a.x2*h.a0 + h.a.x2*a0 + a.x3*h.a.x1-a.x1*h.a.x3,
     a.x3*h.a0 + h.a.x3*a0 + a.x1*h.a.x2-a.x2*h.a.x1);
}

H H::inv(void)const
// same logic as inv(R) for error handling
{
   static const Z maxMes=100;
   static Z mes=1;
   R n=norm();
   if (n==0.){
      goto DIVBYZERO;
   }
   else{
      R nInv=R(1)/n;
      return H(a0*nInv,a*(-nInv));
   }

DIVBYZERO:

   if (mes==maxMes){
      mes++; // needed !
      cpmmessage("H::inv(): argument is 0 ... messages discontinued");
   }
   if (mes<maxMes){
      mes++;
      cpmwarning("C::inv(): argument is 0;  H() returned");
   }
   return H();
}

X2<R,R3> H::rot()const
{
   R2 x(a0,a.abs());
      // by this method we don't need the normalized form of
      // *this! This is rather elegant and efficient
   R2 rphiHalf=x.carToPol();
   R phiHalf=rphiHalf.x1;
   R3 an=a;
   an.normalize();
   return X2<R,R3>(phiHalf*2,an);
}

R3 H::operator *(const R3& x)const
{
   return (operator*( H(x)*inv() )).a;
}

