//? cpmnr.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmnr.h>
#include <cpmtypes.h>

using CpmMathFound::Nr;
using namespace std;
using namespace CpmRoot;
using namespace CpmSystem;
using namespace CpmRootX;

Nr::Nr(string const& s):rep(s)
{
   Z mL=3;
   cpmdbg=2; // makes the program stop if cpmassert fails
   static Word loc="Nr::Nr(string)";
   CPM_MA
   int l=rep.length();
   cpmassert(l>=1,loc);
   char x=rep[0];
   cpmassert(x=='1',loc);
   for (int i=1;i<l;++i){
      x=rep[i];
      cpmassert(x=='0' || x=='1',loc);
   }
   CPM_MZ
}

Nr::Nr(Z i)
{
   Nr aux; // =1
   if (i>1){
      for (Z k=1; k<i;++k) aux=aux.suc();
   }
   *this=aux;
}

Nr Nr::app(char x)const
{
   cpmassert(x=='0' || x=='1',"Nr::app(char)");
   string res=rep+x;
   return Nr(res);
}

Nr Nr::cut()const
{
   if (sin()) return *this;
   int n=rep.length();
   n--;
   string res=rep;
   res.erase(n,1);
   return Nr(res);
}

Nr Nr::suc()const
{
   if (sin()){
      char x=last();
      cpmassert(x=='1',"Nr::suc()");
      return Nr("10");
   }
   else{
      char x=last();
      Nr c=cut();
      if (x=='0') return c.app('1');
      else return c.suc().app('0'); // here recursion takes place
   }
}

Nr Nr::pre()const
{
   if (sin()) return *this;
   else if (rep=="10") return Nr("1");
   else if (rep=="11") return Nr("10");
   else{
      char x=last();
      Nr c=cut();
      if (x=='1') return c.app('0');
      else return c.pre().app('1'); // here recursion takes place
   }
}

bool Nr::operator==(Nr const& n)const
{
   bool b=n.sin();
   if (sin()){
      if (b) return last()==n.last();
      else return false;
   }
   else{
      if (b) return false;
      else return (last()==n.last() && cut()==n.cut());
         // here recursion takes place
   }
}

bool Nr::operator<(Nr const& n)const
{
   bool b=n.sin();
   if (sin()){
      if (b) return false;
      else return true;
   }
   else{
      if (b) return false;
      else{
         Nr c1=cut();
         Nr c2=n.cut();
         char x1=last();
         char x2=n.last();
         if (c1<c2) return true;
         else if (c1==c2 && x1=='0' && x2=='1') return true;
         else return false;
      }
   }
}

/*
Nr Nr::operator+(Nr const& n)const
// non-recursive implementation based on 'counting down'
// very slow!! as could be expected
{
   Nr res=n.suc();
   if (sin()) return res;
   else{
      Nr x1=*this;
      while (!x1.sin()){
         x1=x1.pre();
         res=res.suc();
      }
      return res;
   }
}
*/

Nr Nr::operator+(Nr const& n)const
// much faster recursive implementation
{
   if (sin()) return n.suc();
   else if (n.sin()) return suc();
   else{
      Nr c1=cut();
      char x1=last();
      Nr c2=n.cut();
      char x2=n.last();
      char x;
      if (x1=='0') x=x2;
      else if (x2=='0') x=x1;
      else{
         x='0';
         c1=c1.suc();
      }
      return (c1+c2).app(x);
   }
}

/*
Nr Nr::operator*(Nr const& n)const
// non-recursive implementation based on successive addition
// slow, as could be expected
{
   Nr res=n;
   if (sin()) return res;
   else{
      Nr x1=*this;
      while (!x1.sin()){
         x1=x1.pre();
         res=res+n;
      }
      return res;
   }
}
*/

Nr Nr::operator*(Nr const& n)const
{
   if (sin()) return n;
   else if (n.sin()) return *this;
   else{
      Nr c1=cut();
      char x1=last();
      Nr c2=n.cut();
      char x2=n.last();
      char x;
      Nr res=(c1*c2).app('0').app('0'); // here recursion takes place
      if (x1=='1') res=res+c2.app('0');
      if (x2=='1') res=res+c1.app('0');
      if (x1=='1' && x2=='1') res=res+Nr();
      return res;
   }
}

bool Nr::prnOn(ostream& str)const
{
   str<<rep<<endl;
   return str.good();
}

Nr Nr::random(Z z)
{
   static unsigned long int l=137;
      // never overflows on ++
   cpmassert(z>=1,"Nr::random(Z)");
   Z zAct=randomZ(z,l++);
   bool b;
   string res="1";
   for (Z k=1;k<zAct;++k){
      b=(randomR(l++)>0.5);
      if (b) res+='1';
      else res+='0';
   }
   return Nr(res);
}

Z Nr::test(Z z)
{
// we test statements that are true and others that are wrong
// so that a corrupted function == which gives 'true'
// always can't fake success.

   Z err=0;
   Nr n1=Nr::random(z);
   cout<<endl<<"n1="<<n1;

   bool b=(n1==n1); // should hold
   if (!b) err++;
   cpmmessage("err1="&cpm(err));

   b=(n1<n1); // should nothold
   if (b) err++;
   cpmmessage("err2="&cpm(err));

   b=(n1==n1.suc()); // should not hold
   if (b) err++;
   cpmmessage("err3="&cpm(err));

   b=(n1<n1.suc()); // should hold
   if (!b) err++;
   cpmmessage("err4="&cpm(err));

   Nr n2=Nr::random(z);
   cout<<endl<<"n2="<<n2;
   b=n1+n2==n2+n1;
   if (!b) err++;
   cpmmessage("err5="&cpm(err));

   int sum=0;
   if (n1<n2) sum++;
   if (n1==n2) sum++;
   if (n2<n1)sum++;
   if (sum!=1) err++;
   cpmmessage("err6="&cpm(err));

   Nr n3=Nr::random(z);
   cout<<endl<<"n3="<<n3;
   b=(n1*(n2+n3)==n2*n1+n1*n3); // should hold
   if (!b) err++;
   cpmmessage("err7="&cpm(err));

   b=(n1*(n2.suc()+n3)==n2*n1+n1*n3); // should not hold
   if (b) err++;
   cpmmessage("err8="&cpm(err));

   n1=Nr::random(z);
   cout<<endl<<"n1="<<n1;
   n2=n1.suc();
   b=(n1==n2);
   if (b) err++;
   cpmmessage("err9="&cpm(err));
   n2=n1;
   b=(n1==n2);
   if (!b) err++;
   cpmmessage("err10="&cpm(err));

   n2=n1.suc().suc().suc().suc();
   n3=n1+Nr(4);
   b=(n2==n3);
   if (!b) err++;
   cpmmessage("err11="&cpm(err));

   n3=n1+Nr(3);
   b=(n2==n3);
   if (b) err++;
   cpmmessage("err12="&cpm(err));

   if (err==0){
      cpmmessage("test OK");
   }
   else{
      cpmmessage("test failed");
   }
   return err;
}
