//? cpmq.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmq.h>
#include <cpmsystem.h>

using namespace CpmRoot;

void Q::normalize()
{
   if (qN==0){   // should not happen
      cpmerror("Q::normalize(): denominator = 0, not expected to happen");
   }

   if (qN<0){
      qN=-qN;
      qZ=-qZ;
   } // now qN is a natural number

   if (qZ==0){  // result is 0
      qZ=0;
      qN=1;
      return;
   }

   if (qZ==qN){  // thus result is 1
      qZ=1;
      qN=1;
      return;
   }

   Z pZ, sig;
   if (qZ>0){
      pZ=qZ;
      sig=1;
   }
   else{
      pZ=-qZ;
      sig=-1;
   } // pZ>0 always

// Now pZ,qN are positive and different

   Z a1,a2,a3;
   if (pZ>qN) {
      a1=pZ; a2=qN;
   }
   else {
      a1=qN; a2=pZ;
   }
   do {
      a3=a1%a2;
      a1=a2;
      a2=a3;
   } while (a3!=0);
   pZ /= a1;   // dividing by the greatest common divisor a1
   qN /=a1;
   qZ=pZ*sig;
}

Q::Q(R x, R eps, Z nOK)
{
   const Z nMax=10000;
   const R epsMin=1./nMax;
   const R epsExact=1e-10;
   cpmassert(eps>=epsMin,"Q::Q(R x, R eps, Z nOK)");
   R nr=R(1)/eps;
   Z sig=1;
   if (x<0) {sig=-1; x=-x;}
   Z n_, nRes=cpmtoz(cpmceil(nr)), n=nRes;
   Z m, mRes=cpmrnd(x*n); // mRes/nRes meets
      // accuracy requirements

   while (nRes>nOK){ // trying to reduce magnitude of denominator
      m=cpmrnd(x*n);
      Q q(m,n);
      n_=q.denominator();
      if (n_<nRes){ // reduction successful
         nRes=n_;
         mRes=q.numerator();
         if ( cpmabs(x-(R)mRes/nRes)<=epsExact) goto LABLE;
      }
      n++; // enlarge denominator in order to be able to get a smaller
      // one after cencellation
      if (n>nMax){  // should stop
         cpmwarning("Q(R,R,Z)::numerator larger than required");
         goto LABLE;
      }
   }
LABLE:
   qZ=mRes*sig;
   qN=nRes;
}

Word Q::toWord()const
      // for a nice printable representation
{
   Word sep_i=""; // initial
   Word sep_c="/"; // central
   Word sep_f=""; // final separator
   Word res;
   if (qN==1) res=cpmwrite(qZ);
   else res=cpmwrite(qZ)&sep_c&cpmwrite(qN);
   return sep_i&res&sep_f;
}

Z Q::com(const Q& x) const
// natural order
{
   Z za=qZ * x.qN;
   Z zb=x.qZ * qN;
   if (za<zb) return 1;
   else if (za>zb) return -1;
   else return 0;
}

bool Q::prnOn(ostream& out )const
{ return write(qZ,out)&&write(qN,out);}

bool Q::scanFrom(istream& in ){ return read(qZ,in)&&read(qN,in);}

Word Q::nameOf(void) const{ return Word("Q");}

Q Q::inv(void)const
{
   if (qZ==0)
      cpmerror("Q::inv(): numerator = 0, not expected to happen");
   Word w("intentionally");
   if (qZ>0) return Q(w,qN,qZ); else return Q(w,-qN,-qZ);
      // fast constructor since qZ,qN are already free of common factors
}

Q Q::neg(void)const{ return Q("x",-qZ,qN);}

Q Q::net(Z i)const
{ if (i==1) return Q(Z(1),Z(1)); else return Q(Z(0),Z(1));}

Q Q::abs(void)const{ return ( qZ>=0 ? *this : Q("x",-qZ,qN) );}

Q Q::absSqr(void)const{  return Q(qZ*qZ,qN*qN);}

Q Q::con(void)const{ return *this;}

Z Q::hash()const
{
   R a=qZ;
   a*=qN;
   return CpmRoot::hash(a);
}

namespace{

Z prod(Z a, Z b)
// Product of integers with overflow warning
{
   typedef long double Rld;
   Rld ar=(Rld)a, br=(Rld)b, cr;
   cr=ar*br;
   Z res=a*b;
   Rld indicator=cr-(Rld)res;
   if (indicator!=0.) cpmwarning("Q: Overflow in multiplying integers");
   return res;
}

} // namespace

Q& Q::operator +=(const Q& x)
{
   qZ=prod(qZ,x.qN) + prod(x.qZ,qN);
   qN=prod(qN,x.qN);
   normalize();
   return *this;
}

Q& Q::operator +=(const Z& s){ *this+=Q(s); return *this;}

Q& Q::operator *=(const Q& x)
{
   qZ=prod(qZ,x.qZ);
   qN=prod(qN,x.qN);
   normalize();
   return *this;
}

Q& Q::operator *=(const Z& z)
{
   qZ=prod(qZ,z);
   normalize();
   return *this;
}

Q Q::ran(Z j)const
{
   using CpmRoot::ran;
   using CpmRoot::ranT;
   using CpmRoot::absT;
   using CpmRoot::posVal;
   const Z fZ=11;
   const Z fN=13;
   Z q1Z=qZ*fZ;
   Z q1N=qN*fN;
   Z j0=1000;
   Z j_=(j==0 ? 0 : ranT<Z>(j0,j));
   Z a=ranT<Z>(q1Z,j);
   R a1=absT<Z>(a); a1=cpmsqrt(a1); a=cpmtoz(a1);
   Z b=1+posVal(ranT<Z>(q1N,j+j_));
   R b1=R(b); b1=cpmsqrt(b1); b=cpmtoz(b1);
   return Q(a,b);
}

Q Q::test(Z cpl)const
{
   Z a,b;
   a=3*7*CpmRoot::testT<Z>(a,cpl);
   b=5*11;
   return Q(a,b);
}

R Q::dis(const Q& z)const
{
   R r1=(R)(*this);
   R r2=(R)z;
   return CpmRoot::dis(r1,r2);
}
