//? cpmsi.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmsi.h>
#include <cpmconstphys.h>
#include <cpmrmatrix.h>

using namespace CpmPhysics;
using namespace CpmRoot;
using namespace CpmArrays;

Z Si::nDim=7;

Si Si::m(1.,1,0,0,0,0,0,0);
Si Si::kg(1.,0,1,0,0,0,0,0);
Si Si::s(1.,0,0,1,0,0,0,0);
Si Si::A(1.,0,0,0,1,0,0,0);
Si Si::K(1.,0,0,0,0,1,0,0);
Si Si::mol(1.,0,0,0,0,0,1,0);
Si Si::cd(1.,0,0,0,0,0,0,1);

Si Si::c(c_PHYS,1,0,-1,0,0,0,0);
Si Si::e(e_PHYS,0,0,1,1,0,0,0);
Si Si::h(h_PHYS,2,1,-1,0,0,0,0);
Si Si::h_(h__PHYS,2,1,-1,0,0,0,0);
Si Si::G(G_PHYS,4,-1,-2,0,0,0,0);
Si Si::u(mu_PHYS,0,1,0,0,0,0,0);
Si Si::eps0(eps0_PHYS,-3,-1,4,2,0,0,0);
Si Si::mu0(my0_PHYS,1,1,-2,-2,0,0,0);

void Si::adjust(void)
{
   Z d=qv.dim();
   hasDim=false;
   for (Z i=1;i<=d;i++){
      if (qv[i]!=Q()) { hasDim=true; return;}
   }
   if (type==SI){
      if (d!=nDim)
         cpmerror("Si::adjust(): Si quantity dimension mismatch");
   }
}

Si::Si( R in_val, Q in_m, Q in_kg,
   Q in_s, Q in_A, Q in_K, Q in_mol, Q in_cd ):
      val(in_val),qv(nDim),type(SI)
{
   val=in_val;
   qv[1]=in_m;
   qv[2]=in_kg;
   qv[3]=in_s;
   qv[4]=in_A;
   qv[5]=in_K;
   qv[6]=in_mol;
   qv[7]=in_cd;
   adjust();
}

Si::Si(R in_val,Z in_m,Z in_kg,Z in_s,Z in_A,Z in_K,Z in_mol,Z in_cd ):
   val(in_val),qv(nDim),type(SI)
{
   qv[1]=Q(in_m);
   qv[2]=Q(in_kg);
   qv[3]=Q(in_s);
   qv[4]=Q(in_A);
   qv[5]=Q(in_K);
   qv[6]=Q(in_mol);
   qv[7]=Q(in_cd);
   adjust();
}

Si::Si(R valIn, Q_Vector qvIn, TYPE t):val(valIn), qv(nDim),type(t)
   // qvIn is cut to length nDim if too long and filled by
   // zero if too short
{
   Z dMin=CpmRootX::inf(nDim,qvIn.dim());
   for (Z i=1;i<=dMin;i++) qv[i]=qvIn[i];
   adjust();
}

Si Si::convert(const V<Si>& newBase, TYPE newType)const
// See UM/SI 95-4-21/1-2 Conversion of Units
{
   const Z nOK=10;
   const R eps=0.001;
   R_Matrix a(nDim,nDim);
   Z i,j;
   for (i=1;i<=nDim;i++) for (j=1;j<=nDim;j++)
      a[j][i]=(R)(newBase[i][j]);    // transposition implied
   a=!a;
   R_Vector x(nDim);
   for (i=1;i<=nDim;i++)
      x[i]=(R)(qv[i]);
   R_Vector y=a*x;  // no transposition here
   Si res;
   for (i=1;i<=nDim;i++){
      Q q(y[i],eps,nOK);
      res[i]=q;
   }
   R prod=val;
   for (i=1;i<=nDim;i++){
      R ri=(newBase[i]).val;
      R ei=-res[i];
      prod*=cpmpow(ri,ei);
   }
   res.val=prod;
   res.type=newType;
   res.adjust();
   return res;
}

Si Si::fromSiToNat()const
{
   V<Si> base(nDim);
   base[1]=c;
   base[2]=h;
   base[3]=G;
   base[4]=e;
// natural units for these last three types of quantities
// are not so convincingly determined; let them be!
   base[5]=TEMPERATURE();
   base[6]=SUBSTANCE();
   base[7]=INTENSITY();
   return convert(base,NATCONST);
}

Si& Si::operator+=(const Si& x)
{
   if (type!=x.type)
      cpmerror("Si: += : type mismatch");
   if (qv != x.qv)
      cpmerror("Si: += : dimension mismatch");
   val+=x.val;
   return *this;
}

Si& Si::operator*=(const Si& x)
{
   if (type!=x.type) cpmerror("Si: *= : type mismatch");
   val*=x.val;
   qv+=x.qv;
   adjust();
   return *this;
}

Si& Si::operator*=(const R& s)
{
   val*=s;
   return *this;
}

Si Si::neg(void)const
{
   return Si(-val,qv);
}

Si Si::inv(void)const
{
   if (val==0.) cpmerror("Si::inv(): val=0");
   return Si(R(1)/val,-qv);
}

Si Si::net(Z i)const
{
   if (i==0) return Si(0,qv);
      // there is a additive neutral for each value of qv !
   else return Si(1);
}

R Si::absSqr()const{ R r=cpmabs(val); return r*r;}

R Si::abs()const{ return cpmabs(val);}

Si Si::con(void)const{ return *this;}

R Si::dis(const Si& z)const
{
   if (qv!=z.qv) return 1; // quantities of different dimension
      // are maximally different
   else return CpmRoot::dis(val,z.val);
}

namespace{

R pow_(R x, Z k)
// k-th power of x realized as successive multiplication
{
   if (x==0.){
      cpmwarning("Si::pow_(...) negative power of 0 set to 0");
      return 0.;
   }
   Z n;
   if (k<0){ n=-k; x=R(1)/x;} else n=k;
   R res=x;    // one factor
   for (int i=1;i<n;i++) res*=x; // n-1 multiplications
       // thus in toto n factors
   return res;
}

} // namespace

Si Si::pow(Q q)const
{
   R resVal;
   if (q.denominator()==1){
      Z qn=q.numerator();
      resVal=pow_(val, qn);
      return Si(resVal,qv*q);
   }
   else if (val<0){
      cpmerror("Si::pow(Q): broken power of a negative value");
      return *this; // never come here
   }
   else{
      resVal=cpmpow(val,(R)q);
      return Si(resVal,qv*q);
   }
}

// overloading pow() with SI base and int exponent
Si Si::pow(Z y)const
{
   R resVal;
   resVal=pow_(val, y);
   return Si(resVal,qv*Q(y));
}

Word Si::nameOf(void)const{ return Word("Si");}

Word Si::toWord()const
{
   Word res=cpmwrite(val);
   if (!hasDim) return res;

   V<Word> u(nDim);

   if (type==SI){
      u[1]="m";
      u[2]="kg";
      u[3]="s";
      u[4]="A";
      u[5]="K";
      u[6]="mol";
      u[7]="cd";
   }
   else if (type==NATCONST){
      u[1]="c";
      u[2]="h";
      u[3]="G";
      u[4]="e";
      u[5]="K";
      u[6]="mol";
      u[7]="cd";
   }
   else {
      u[1]="e1";
      u[2]="e2";
      u[3]="e3";
      u[4]="e4";
      u[5]="e5";
      u[6]="e6";
      u[7]="e7";
   }
   for (Z i=1;i<=nDim;i++){
      Q qi=qv[i];
      if (qi!=Q()) res&=" * "&u[i]&"^"&qi.toWord();
   }
   return res;
}

bool Si::prnOn( ostream& str ) const
{
   CpmRoot::writeTitle("Si",str);
   cpmp(val);
   cpmp(qv);
   cpmp((Z)type);
   return true;
}

bool Si::scanFrom (istream& str)
{
   cpms(val);
   cpms(qv);
   Z zt;
   cpms(zt);
   type=(TYPE)zt;
   adjust();
   return true;
}

Z Si::hash() const
{
   R a=val;
   for (Z i=1;i<=nDim; i++) a+=cpmabs((R)qv[i]);
   return CpmRoot::hash(a);
}

Si Si::ran(Z j)const
// ran should not change the qv, since one
// would get incompatible test values for addition
   {
      R valR=cpmran(val,j);
      return Si(valR,qv,type);
   }

Si Si::test(Z tsz)const
{
   R r=1;
   Q q(Z(1));
   r=CpmRoot::test(r,tsz);
   Q q0=q.test(tsz);
   Si res(*this); // res has the same type as *this
   res.val=r;
   Z i,j=tsz;
   for (i=1;i<=nDim;i++) (res.qv)[i]=q0.ran(j++);
   res.adjust();
   return res;
}


Z Si::com(const Si& z) const
{
   if (val<z.val) return 1;
   if (val>z.val) return -1;

   if (qv<z.qv) return 1;
   if (qv>z.qv) return -1;

   return 0; // *this==v
}

void Si::test(Z stop)
 {
   const Z tvs=16;
   Si q(1);
   q.test(tvs);
   cout<<endl<<"test = "<<q;
   //if (stop) getchar();
   Z i=137;
   Si q1=q.ran(i++);
   cout<<endl<<" First randomly generated test object q1: "<<endl;
   cout<<q1;
   //if (stop) getchar();

   Si q2=q.ran(i++);
   cout<<endl<<" Second randomly generated test object q2: "<<endl;
   cout<<q2;
   //if (stop) getchar();

   Si q3=q.ran(i++);
   cout<<endl<<" Third randomly generated test object q3: "<<endl;
   cout<<q3;
   //if (stop) getchar();

   using CpmRoot::ran;
   using CpmRoot::abs;

   R sum=0;
   R r;
   R r1=ran(r,i++);
   R r2=ran(r,i++);

   cout<<endl<<"Test of *,+,- :";
   Si q4=(r1+r2)*q3;
   Si q5=-q3*r1+r2*(-q3);
   q5+=q4;

   cout<<endl<<"q5="<<q5<<" should have value = 0 ";

   sum+=q5.abs();

   cout<<endl<<" errorSum= "<<sum<<" 0 is OK";
  // if (stop) getchar();

   cout<<endl<<"Test of ! * /:";
   q4=!(q1*q2*q3);
   q5=!q3*!q1*!q2;
   q5/=q4;
   cout<<endl<<"q5="<<q5<<" should be number 1 ";
   Si q6;
   q6=q6.net(1);
   q5-=q6;
   cout<<endl<<"q5="<<q5<<" should be number 0 ";
   sum+=q5.abs();
   cout<<endl<<" errorSum= "<<sum<<" 0 is OK";
   //if (stop) getchar();

   cout<<endl<<"Demo of calculations with natural constants";
    LENGTH m(1);
   TIME s(1);
   Si cVal=c*s/m; // SI::c is static
   cout<<endl<<"value of c = "<<cVal;

   ACCELERATION g(9.81);
   TIME t(2);
   Si s1=(g/2)*t*t;
   cout<<endl<<" s1= "<<s1<< " should be the length 19.62 m";
   cout<<endl<<"Test of file I/O: ";

   Si q7,q8,q9;

   ofstream out("testdata.txt");
   if (!out) cpmerror("Si::test(): output file cannot be opened");
   out<<q1;
    out<<q2;
     out<<q3;
   out.close();

   ifstream in("testdata.txt");
   if (!in) cpmerror("Si::test(): input file cannot be opened");
   in>>q7;
     in>>q8;
     in>>q9;

   cout<<endl<<" This was read from file: "<<endl
   <<q7<<endl<<q8<<endl<<q9<<endl<<" looks normal ?";
   in.close();
   sum+=(q7-q1).abs();
   sum+=(q8-q2).abs();
   sum+=(q9-q3).abs();
   cout<<endl<<" errorSum= "<<sum<<", should be close to 0";
   cout<<endl<<" End of test of class "<<"Si"<<endl;
 }

void Si::PlanckUnits(const Word& fileName)
{
   const Z nx=8;
   ofstream out(-fileName);
   V<Si> x(nx);
   x[1]=LENGTH();
   x[2]=MASS();
   x[3]=TIME();
   x[4]=CURRENT();
   x[5]=VOLTAGE();
   x[6]=RESISTANCE();
   x[7]=Si::eps0;
   x[8]=Si::u;
   V<Word> w(nx);
   w[1]="LENGTH";
   w[2]="MASS";
   w[3]="TIME";
   w[4]="CURRENT";
   w[5]="VOLTAGE";
   w[6]="RESISTANCE";
   w[7]="eps0";
   w[8]="u";
   for (Z i=1;i<=nx;i++){
      Word wi=w[i]&": ";
      wi&=x[i].toWord();
      wi&=" = ";
      wi&=x[i].fromSiToNat().toWord();
      cout<<endl<<wi;
      out<<endl<<wi;
   }
}

