//? cpmtrj.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmtrj.h>
#include <cpmframes.h>
#include <cpmrmatrix.h>

   using namespace CpmArrays;
   using CpmRootX::Record;
   using CpmGraphics::Frames;
   using CpmGraphics::Graph;
   using CpmQuantumMechanics::State;
   using CpmQuantumMechanics::Observable;
   using CpmQuantumMechanics::SpectralObject;
   using CpmQuantumMechanics::TrjCon;
   using CpmQuantumMechanics::TrjBase;
   using CpmQuantumMechanics::TrjSR;
   using CpmQuantumMechanics::TrjDMI;
   using CpmQuantumMechanics::TrjDMI_Obs;
   using CpmQuantumMechanics::Trj2;
   using std::vector;

////////////////// struct TrjCon ///////////////////////////////////////
TrjCon::TrjCon(Record const& rc)
{
   Word sec("TrjCon");
   Z pedRed=0; // pedantry reduced
   rc.get(sec,"autoAdj",autoAdj_);
   rc.get(sec,"autoStep",autoStep_);
   rc.get(sec,"dt",dt_);
   rc.get(sec,"enfNor",enfNor_);
   rc.get(sec,"ignorePhases",ignPhs_);
   rc.get(sec,"nSteps",nSteps_);
   rc.get(sec,"displayMode",disMode_,pedRed);
   rc.get(sec,"secFac",secFac_);
   rc.get(sec,"tWait",tWait_);
   rc.get(sec,"viewingPeriod",vwnPer_);
   rc.get(sec,"filingPeriod",filPer_,pedRed);
   rc.get(sec,"imagingPeriod",imgPer_,pedRed);
   rc.get(sec,"dph1",dph1_,pedRed);
   rc.get(sec,"dph2",dph2_,pedRed);
   rc.get(sec,"dphMethod",dphMethod_,pedRed);
   rc.get(sec,"corGrnMet",corGrnMet_,pedRed);
   rc.get(sec,"nPhi",nPhi_,pedRed);
   rc.get(sec,"nBins",nBins_,pedRed);
   rc.get(sec,"gam",gam_,pedRed);
   rc.get(sec,"intMet",intMet_,pedRed);
      // t_Tot will be initialized in function TrjBase::adjustBase_()
      // also for the case dt_<0 an adjustment will be made there.
}

bool TrjCon::prnOn(ostream& str)const
{
   cpmwt("TrjCon");
   cpmp(autoAdj_);
   cpmp(autoStep_);
   cpmp(dt_);
   cpmp(enfNor_);
   cpmp(ignPhs_);
   cpmp(nSteps_);
   cpmp(disMode_);
   cpmp(secFac_);
   cpmp(tWait_);
   cpmp(vwnPer_);
   cpmp(filPer_);
   cpmp(imgPer_);
   cpmp(dph1_);
   cpmp(dph2_);
   cpmp(dphMethod_);
   cpmp(corGrnMet_);
   cpmp(nPhi_);
   cpmp(nBins_);
   cpmp(gam_);
   cpmp(intMet_);
   return true;
}

bool TrjCon::scanFrom(istream& str)
{
   cpms(autoAdj_);
   cpms(autoStep_);
   cpms(dt_);
   cpms(enfNor_);
   cpms(ignPhs_);
   cpms(nSteps_);
   cpms(disMode_);
   cpms(secFac_);
   cpms(tWait_);
   cpms(vwnPer_);
   cpms(filPer_);
   cpms(imgPer_);
   cpms(dph1_);
   cpms(dph2_);
   cpms(dphMethod_);
   cpms(corGrnMet_);
   cpms(nPhi_);
   cpms(nBins_);
   cpms(gam_);
   cpms(intMet_);
   return true;
}

///////////////////// class TrjBase /////////////////////////////////////
// code fragments: "shared code"

#define CPM_SC1 R waitTime=tc_.tWait_;\
   Z nCol=g.getCol();\
   Z width=4;\
   Z nImgLin=nCol/width;\
   Z nRow=g.getLin();\
   Z iFr=2;\
   Z height=(nRow-iFr)/d_;\
   Z ii=iFr,jj=-width;\
   bool fullFrame=false;\
   Z countImg=0;\
   Z showCount=0;

#define CPM_SC2 if(vwnAlarm_()){\
         jj+=width;\
         showCount++;\
         if (showCount==nImgLin) fullFrame=true;\
         Z4 z4(height,width,ii,jj);\
         showAsLine(g,z4);\
         if (fullFrame){\
            fullFrame=false;\
            showCount=0;\
            jj=-width;\
            if (imgAlarm_()){\
               countImg++;\
               Word imf="TrjBase_runAsLine";\
               g.write(imf&cpm(countImg,3));\
               }\
         }\
         cpmwait(waitTime);\
      }

Z TrjBase::runAsLine(Graph& g, ostream& str)
{
   Z mL=1;
   Word loc("TrjBase::runAsLine(Graph)");
   CPM_MA
   Z n=tc_.nSteps_;
   if (n<1) return 0;
   R nInv=1./n;
   R tTotInv=R(1.)/tc_.tTot_;
   R dt=tc_.dt_;
   CPM_SC1
   Z i=0;
   while(keepGoing()){ // shows n states
      cpmprogress("Loop",R(t_)*tTotInv);
      i++;
      CPM_SC2
      if (filAlarm_()) fileState_(str);
      step_(dt);
   }
   cpmprogress("Loop",1);
   CPM_MZ
   return i;
}

void TrjBase::playAsLine(Graph& g, istream& str)
{
   Z mL=1;
   Word loc("TrjBase::playAsLine(Graph)");
   CPM_MA
   bool goOn=readState_(str);
   // notice that CPM_SC1 needs to know d_; we should not
   // assume the calling state to give this information
   // in a way that is consistent with the data from str
   CPM_SC1
   Z frame=1;
   while (goOn && keepGoing()){
      cpmvalue("frame#",++frame,2);
      goOn=readState_(str);
      CPM_SC2
   }
   CPM_MZ
}
#undef CPM_SC1
#undef CPM_SC2

Z TrjBase::runAsArea(Graph& g, ostream& str )
{
   using namespace CpmImaging;
   using namespace CpmArrays;
   using namespace CpmSystem;

   Z mL=1;
   Word loc("TrjBase::runAsArea(Graph)");
   CPM_MA

   R waitTime=tc_.tWait_;
   R secFac=tc_.secFac_;
   bool autoAdj=tc_.autoAdj_;
   bool ignPhs=tc_.ignPhs_;
   R dt=tc_.dt_;
   R tTot=tc_.tTot_;
   Z nSteps=tc_.nSteps_;
      // even in the case of automatic timestepping nSteps determines
      // the number of state records
   Z nStepsMod=nSteps+2;

   R_Vector times(nStepsMod);
   R_Vector norms(nStepsMod);
   V<C> energies(nStepsMod);
   R aspRatMem=g.getAspRat();
   g.clr();
   g.setAspRatHoldHeight(1.);

   Z i,d=dim();
   Z countAct=1;
   Z countImg=0;

// storing initial values
   times[countAct]=t();
   norms[countAct]=n_1();
   energies[countAct]=enr();

   Z ncl=2;
   V<Color> cl(ncl);
   cl[1]=RED;  // color for real part
   cl[2]=GREEN;   // color for imaginary part
// 'dynamics' loop (showing a one-parameter group)
   R nStepsInv=1./nSteps;
   Z count=0;
   showAsArea(g);
   cpmmessage("showing the initial state");
   R tIniWait=2.5;
   Z mesPane=2; // message pane
   cpmwait(tIniWait,mesPane);
   while (keepGoing()){
      cpmprogress("Loop",R(count)*nStepsInv);
      step_(dt);
      count++;
      if (filAlarm_()) fileState_(str);
// graphics part
      if (vwnAlarm_()){
         countAct++;
         times[countAct]=t();
         norms[countAct]=n_1();
         energies[countAct]=enr();
         showAsArea(g);
         cpmwait(waitTime);
      }
      if(imgAlarm_()){
         countImg++;
         Word imf="TrjBase_runAsArea";
         g.write(imf&cpm(countImg,3));
      }
   }
   cpmmessage("keep going loop finished");
   times=times.resize(countAct);
   norms=norms.resize(countAct);
   energies=energies.resize(countAct);
   g.clr();
   g.setAspRatHoldHeight(aspRatMem);
   Frames frs(g,Z(1),Z(2));
   Graph g1(frs.at(1,1));
   Graph g2(frs.at(1,2));
   g1.setWithOrigin(0);
   g2.setWithOrigin(0);
   g1.addText("(norm of state) - 1");
   g1.show(times,norms);
   C e0=energies[1];
   R e0abs=e0.abs();
   bool b0=(e0abs==0.);
   R fac0= b0 ? R(1) : R(1)/e0abs;
   R_Matrix en(ncl,countAct);
   cpmmessage("next loop comes soon");
   for (i=1;i<=countAct;++i){
      C val=energies[i];
      val-=e0;
      en[1][i]=val[1]*fac0;
      en[2][i]=val[2]*fac0;
   }
   if (b0)
       g2.addText("deviation of total energy from initial value");
   else
       g2.addText(
         "relative deviation of total energy from initial value");
   g2.show(times,en,cl);
   cpmmessage("addText done");
   CPM_MZ
   return count;
}

void TrjBase::playAsArea(Graph& g, istream& str )
{
   Z mL=1;
   Word loc("TrjBase::playAsArea(Graph)");
   CPM_MA
   R waitTime=tc_.tWait_;
   R secFac=tc_.secFac_;
   bool ignPhs=tc_.ignPhs_;
   Z ncl=ignPhs ? 1 : 3;
   V<Color> cl(ncl);
   cl[1]=BLACK; // color for absolute value
   if (ncl==3){
      cl[2]=RED;  // color for real part
      cl[3]=GREEN;   // color for imaginary part
   }

   vector<R> tim;
   vector<R> nor;
   vector<C> en;
   bool goOn=true;
   Z countImg=0;
   Z frame=0;
   while (goOn && keepGoing()){
      cpmvalue("frame#",++frame,2);
      goOn=readState_(str);
      if(vwnAlarm_()){
         tim.push_back(t());
         nor.push_back(n_1());
         en.push_back(enr());
         showAsArea(g);
         cpmwait(waitTime);
      }
      if(imgAlarm_()){
         countImg++;
         Word imf="TrjBase_runAsArea";
         g.write(imf&cpm(countImg,3));
      }
   }

   Z countAct=(Z)tim.size();
   R_Vector times(tim);
   R_Vector norms(nor);
   g.clr();
   Frames frs(g,Z(1),Z(2));
   Graph g1(frs.at(1,1));
   Graph g2(frs.at(1,2));
   g1.setWithOrigin(0);
   g2.setWithOrigin(0);
   g1.show(times,norms);
   g1.addText("norm of state - 1");
   C e0=en[0];
   R fac0=R(1)/e0.abs();
   R_Matrix em(ncl,countAct);
   for (Z i=1;i<=countAct;++i){
      C val=en[i-1];
      val-=e0;
      em[1][i]=val.abs()*fac0;
      if (ncl==3){
         em[2][i]=val[1]*fac0;
         em[3][i]=val[2]*fac0;
      }
   }
   g2.show(times,em,cl);
   g2.addText("relative deviation of total energy from initial value");
   CPM_MZ
}

Z TrjBase::runAsCur(Graph& g, ostream& str )
{
   using namespace CpmImaging;
   using namespace CpmArrays;
   using namespace CpmSystem;

   Z mL=1;
   Word loc("TrjBase::runAsCur(Graph)");
   CPM_MA

   R waitTime=tc_.tWait_;
   R secFac=tc_.secFac_;
   bool autoAdj=tc_.autoAdj_;
   bool ignPhs=tc_.ignPhs_;
   R dt=tc_.dt_;
   R tTot=tc_.tTot_;
   Z nSteps=tc_.nSteps_;
   Z nStepsMod=nSteps+2;

   R_Vector times(nStepsMod);
   R_Vector norms(nStepsMod);
   V<C> energies(nStepsMod);

// showing the initial condition
   Z i,d=dim();
   R secFacInv=R(1)/secFac;
   R yMax=sup();
   R yScale=yMax*secFac;
   Iv band;
   if (autoAdj) band=Iv(yMax*secFacInv,yMax*secFac);
   g.setAutoScale(0);
   g.setX(Iv(0,d));
   if (autoAdj){
      g.setY(Iv(-band.sup(),band.sup()));
   }
   else{
      g.setY(Iv(-yScale,yScale));
   }

   Z ncl=ignPhs ? 1 : 3;
   V<Color> cl(ncl);
   cl[1]=BLACK; // color for absolute value
   if (ncl==3){
      cl[2]=RED;  // color for real part
      cl[3]=GREEN;   // color for imaginary part
   }

   Z countAct=1;
   Z countImg=0;

   times[countAct]=t();
   norms[countAct]=n_1();
   energies[countAct]=enr();

   R_Matrix res(ncl+1,d);
   for (i=1;i<=d;i++){
     // res[1][i]=i-1;
      res[1][i]=i;
     // C val=getRep()[i-1];
      C val=getRep()[i];
      res[2][i]=val.abs();
      if (ncl==3){
         res[3][i]=val[1];
         res[4][i]=val[2];
      }
   }
   g.clr();
   g.addText(
      "Initial state for evolution. Real part red, Imaginary part green");
   g.show(res,cl);
   cpmwait(waitTime*3+2);
// 'dynamics' loop (showing a one-parameter group)
   R nStepsInv=R(1)/R(nSteps);
   Z count=0;
   while (keepGoing()){  // one state already shown,
      cpmprogress("Loop",R(count)*nStepsInv);
      step_(dt);
      count++;
      if (count>nSteps){ // no longer happens, but also does not hurt
         ostringstream ost;
         ost<<"t_="<<t_<<" tTot="<<tTot<<" count="<<count<<endl;
         cpmwarning(ost);
         cpmerror("bad time range");
      }
      if (filAlarm_()) fileState_(str);
// graphics part
      if (vwnAlarm_()){
         countAct++;
         times[countAct]=t();
         norms[countAct]=n_1();
         energies[countAct]=enr();

         for (i=1;i<=d;i++){
            //C val=getRep()[i-1];
            C val=getRep()[i];
            res[2][i]=val.abs();
            if (ncl==3){
               res[3][i]=val[1];
               res[4][i]=val[2];
            }
         }
         if (autoAdj){
            yMax=sup();
            bool bc=band.contains(yMax);
            if (!bc){
               band=Iv(yMax*secFacInv,yMax*secFac);
               g.setY(Iv(-band.sup(),band.sup()));
             //  g.writeGridText();
            }
         }
         g.clear();
         g.addText(text());
         g.show(res,cl);

         if(imgAlarm_()){
            countImg++;
            Word imf="TrjBase_runAsCurves";
            g.write(imf&cpm(countImg,3));
         }
         cpmwait(waitTime);
      } // vwnAlarm_
   }
   cpmmessage("keep going loop finished");
   times=times.resize(countAct);
   norms=norms.resize(countAct);
   energies=energies.resize(countAct);
   g.clr();
   Frames frs(g,Z(1),Z(2));
   Graph g1(frs.at(1,1));
   Graph g2(frs.at(1,2));
   g1.setWithOrigin(0);
   g2.setWithOrigin(0);
   g1.addText("(norm of state) - 1");
   g1.show(times,norms);
   C e0=energies[1];
   R e0abs=e0.abs();
   bool b0=(e0abs==0.);
   R fac0= b0 ? R(1) : R(1)/e0abs;
   R_Matrix en(ncl,countAct);
   cpmmessage("next loop comes soon");
   for (i=1;i<=countAct;++i){
      C val=energies[i];
      val-=e0;
      en[1][i]=val.abs()*fac0;
      if (ncl==3){
         en[2][i]=val[1]*fac0;
         en[3][i]=val[2]*fac0;
      }
   }

   if (b0)
       g2.addText("deviation of total energy from initial value");
   else
       g2.addText(
         "relative deviation of total energy from initial value");
   g2.show(times,en,cl);
   cpmmessage("addText done");

   CPM_MZ
   return count;
}

void TrjBase::playAsCur(Graph& g, istream& str )
{
   using namespace CpmImaging;
   using namespace CpmArrays;
   using namespace CpmSystem;

   Z mL=1;
   Word loc("TrjBase::playAsCur(Graph)");
   CPM_MA

   R waitTime=tc_.tWait_;
   R secFac=tc_.secFac_;
   bool autoAdj=tc_.autoAdj_;
   bool ignPhs=tc_.ignPhs_;

   vector<R> tim;
   vector<R> nor;
   vector<C> en;

   bool goOn=readState_(str);
   cpmassert(goOn,loc);

   Z i,d=dim();
   R secFacInv=R(1)/secFac;
   R yMax=sup();
   R yScale=yMax*secFac;
   Iv band;
   if (autoAdj) band=Iv(yMax*secFacInv,yMax*secFac);
   g.setAutoScale(0);
   g.setX(Iv(0,d));
   if (autoAdj){
      g.setY(Iv(-band.sup(),band.sup()));
   }
   else{
      g.setY(Iv(-yScale,yScale));
   }
   Z ncl=ignPhs ? 1 : 3;
   V<Color> cl(ncl);
   cl[1]=BLACK; // color for absolute value
   if (ncl==3){
      cl[2]=RED;  // color for real part
      cl[3]=GREEN;   // color for imaginary part
   }

 //  Z countAct=1;
   Z countImg=0;

   tim.push_back(t());
   nor.push_back(n_1());
   en.push_back(enr());

   R_Matrix res(ncl+1,d);
   for (i=1;i<=d;i++){
      res[1][i]=i-1;
      C val=getRep()[i-1];
      res[2][i]=val.abs();
      if (ncl==3){
         res[3][i]=val[1];
         res[4][i]=val[2];
      }
   }
   g.clr();
   g.addText("Initial state for evolution.\
 Real part red, Imaginary part green)");
   g.show(res,cl);
   cpmwait(waitTime*3+2);
   Z frame=0;
   while (goOn && keepGoing()){
      goOn=readState_(str);
      cpmvalue("frame#",++frame,2);
// graphics part
      if (vwnAlarm_()){
         tim.push_back(t());
         nor.push_back(n_1());
         en.push_back(enr());

         for (i=1;i<=d;i++){
           // C val=getRep()[i-1];
            C val=getRep()[i];
            res[2][i]=val.abs();
            if (ncl==3){
               res[3][i]=val[1];
               res[4][i]=val[2];
            }
         }
         if (autoAdj){
            yMax=sup();
            bool bc=band.contains(yMax);
            if (!bc){
               band=Iv(yMax*secFacInv,yMax*secFac);
               g.setY(Iv(-band.sup(),band.sup()));
             //  g.writeGridText();
            }
         }
         g.clear();
         g.addText(text());
         g.show(res,cl);
         if(imgAlarm_()){
            countImg++;
            Word imf="TrjBase_runAsCurves";
            g.write(imf&cpm(countImg,3));
         }
         cpmwait(waitTime);
      } // vwnAlarm_
   }
   R_Vector times(tim);
   R_Vector norms(nor);
   Z countAct=times.dim();

   g.clr();
   Frames frs(g,Z(1),Z(2));
   Graph g1(frs.at(1,1));
   Graph g2(frs.at(1,2));
   g1.setWithOrigin(0);
   g2.setWithOrigin(0);
   g1.addText("norm of state - 1");
   g1.show(times,norms);

   C e0=en[0];
   R e0abs=e0.abs();
   bool b0=(e0abs==0.);
   R fac0= b0 ? R(1) : R(1)/e0abs;
   R_Matrix em(ncl,countAct);
   for (i=1;i<=countAct;++i){
      C val=en[i-1];
      val-=e0;
      em[1][i]=val.abs()*fac0;
      if (ncl==3){
         em[2][i]=val[1]*fac0;
         em[3][i]=val[2]*fac0;
      }
   }
   if (b0)
      g2.addText("deviation of total energy from initial value");
   else
      g2.addText("relative deviation of total energy from initial value"\
      )\
      ;
   g2.show(times,em,cl);
   CPM_MZ
}

bool TrjBase::prnOn(ostream& str)const
{
   cpmwt("TrjBase");
   cpmp(tc_);
   cpmp(t_);
   cpmp(dt_);
   cpmp(n_1_);
   cpmp(enr_);
   return true;
}

bool TrjBase::scanFrom(istream& str)
{
   cpms(tc_);
   cpms(t_);
   cpms(dt_);
   cpms(n_1_);
   cpms(enr_);
   adjust_(); // makes cyc_, d_, n1_
   return true;
}

//////////////////////// class TrjSR ///////////////////////////////

void TrjSR::adjust_()
      // sets dt_ in accordance with so, x_, and QMBase::phiStep
{
   Z mL=1;
   Word loc("TrjSR::adjust_()");
   CPM_MA
   TrjB::adjust_();
   static R fac=QMBase::get_hDiracInv();
   Z d=so_.dim();
   cpmassert(d==d_,"TrjSR::adjust_()");
   R n1=R(1.)+n_1_;
   cpmmessage("mult ahead");
   State sox=so_*x_;
   cpmmessage("mult done");
   enr_=eTot();
   R n2=sox.norm()*fac;
   dt_=QMBase::step(n1,n2);
   cof_=V<C>(d);
   for (Z i=1;i<=d;i++) cof_[i]=(so_[i]|x_);
   CPM_MZ
}

TrjSR::TrjSR(Observable const& a, State const& x, TrjCon const& tc):
TrjB(x,tc), so_(a.spectralRepresentation()){}

TrjSR::TrjSR(const SpectralObject& so, const State& x, TrjCon const& tc):
TrjB(x,tc),so_(so){}

void TrjSR::step_(R h)
//experiment: dephasing in addition to dynamics
{
   static R fac=-QMBase::get_hDiracInv();
   if (!autoStep()) dt_=h;
   if (dt_!=R1(0)){
      SpectralObject so(so_);
      so.expi(fac*dt_);
      x_=so*x_;
   }
   Z im=trjCon().dphMethod_;
   Z corGrnMet=trjCon().corGrnMet_;
   if (im>0){
      R p1=trjCon().dph1_;
      R p2=trjCon().dph2_;
      x_.dePhase(p1,p2,im);
   }
   if (corGrnMet){
      Z nPhi=trjCon().nPhi_;
      Z nBins=trjCon().nBins_;
      x_=x_.corGrn(nPhi,nBins);
   }
   t_+=dt_;
   stepCon_+=1;
   norIfEnf_();
   enr_=eTot();
}

bool TrjSR::prnOn(ostream& str)const
{
   cpmwt("TrjSR");
   if (! TrjB::prnOn(str)) return false;
   cpmp(so_);
   return true;
}

bool TrjSR::scanFrom(istream& str)
{
   if (! TrjB::scanFrom(str)) return false;
   cpms(so_);
   adjust_();
   return true;
}

/////////////////// class TrjDMI_Obs ///////////////////////

bool TrjDMI_Obs::prnOn(ostream& str)const
{
   cpmwt("TrjDMI_Obs");
   if (! TrjDMI<State>::prnOn(str)) return false;
   cpmp(obs_);
   return true;
}

bool TrjDMI_Obs::scanFrom(istream& str)
{
   if (! TrjDMI<State>::scanFrom(str)) return false;
   cpms(obs_);
   adjust_();
   return true;
}

////////////////////class Trj2 ///////////////////////////////

void Trj2::step_(R h)
{
// For the not very probable situation that the two
// trajectories should ask for rather different auto
// steps, here is a reasonable method to handle the
// case. Essential for reliable comparison is that
// both trajectories do the same time step so that we have
// a common time parameter for both trajectories.
   static Z toggle=1;
   static R dtLast=0;
   toggle=-toggle;
   bool as1=rep_(1).autoStep();
   bool as2=rep_(2).autoStep();
   if (toggle==1){
      rep_(2).step_(h);
      R dt=rep_(2).dt();
      if (dtLast>0 && dt>dtLast) dt=(dt+dtLast)*0.5;
      rep_(1).setAutoStep_(false);
      rep_(1).step_(dt);
      rep_(1).setAutoStep_(as1);
      dtLast=dt;
   }
   else{
      rep_(1).step_(h);
      R dt=rep_(1).dt();
      if (dtLast>0 && dt>dtLast) dt=(dt+dtLast)*0.5;
      rep_(2).setAutoStep_(false);
      rep_(2).step_(dt);
      rep_(2).setAutoStep_(as2);
      dtLast=dt;
   }
}

/*
void Trj2::step_(R h)
{
   rep_(1).step_(h);
   rep_(2).step_(h);
}
*/

void Trj2::run(Graph& g, ostream& str1, ostream& str2)
{
   using namespace CpmImaging;
   using namespace CpmArrays;
   using namespace CpmSystem;
   rep_(1).adjust_();
   rep_(2).adjust_();

   TrjCon tc=rep_(2).trjCon();

   R waitTime=tc.tWait_;
   R secFac=tc.secFac_;
   bool autoAdj=tc.autoAdj_;
   Z nSteps=tc.nSteps_;
   R dt=tc.dt_;
   bool ignPhs=tc.ignPhs_;

   g.clr();
   Frames frs(g,Z(1),Z(2));
   Graph g1(frs.at(1,1));
   Graph g2(frs.at(1,2));
   Frames frs2(g2,Z(2),Z(1));
   Graph g12(frs2.at(1,1));
   Graph g22(frs2.at(2,1));

 // showing the initial condition
   Z i,d=dim();
   R secFacInv=R(1)/secFac;
   R y1=rep_(1).sup();
   R y2=rep_(2).sup();
   R yMax=CpmRootX::sup<R>(y1,y2);
   R yScale=yMax*secFac;
   Iv band;
   if (autoAdj) band=Iv(yMax*secFacInv,yMax*secFac);

   g1.setAutoScale(0);
   g1.setX(Iv(0,d));
   g12.setAutoScale(0);
   g12.setX(Iv(0,d));
   g22.setAutoScale(0);
   g22.setX(Iv(0,d));
   if (autoAdj){
      g1.setY(Iv(-band.sup(),band.sup()));
      g12.setY(Iv(-band.sup(),band.sup()));
      g22.setY(Iv(-band.sup(),band.sup()));
   }
   else{
      g1.setY(Iv(-yScale,yScale));
      g12.setY(Iv(-yScale,yScale));
      g22.setY(Iv(-yScale,yScale));
   }
   Z ncl=ignPhs ? 1 : 3;
   V<Color> cl(ncl);
   cl[1]=BLACK; // color for absolute value
   if (ncl==3){
      cl[2]=RED;  // color for real part
      cl[3]=GREEN;   // color for imaginary part
   }

   Z ncl1=ncl+1;
   Z nSteps1=nSteps+2;
   R_Vector times(nSteps1);
   R_Vector n_1(nSteps1);
   R_Vector diff(nSteps1);

   R_Matrix res1(ncl1,d);
   R_Matrix res2(ncl1,d);
   R_Matrix res3(ncl1,d);
   for (i=1;i<=d;i++){
      res1[1][i]=i-1;
      res2[1][i]=i-1;
      res3[1][i]=i-1;
      C val1=rep_(1).getRep()[i-1];
      C val2=rep_(2).getRep()[i-1];
      C val3=val1-val2;
      res1[2][i]=val1.abs();
      res2[2][i]=val2.abs();
      res3[2][i]=val1.abs()-val2.abs();
      if (ncl1==4){
         res1[3][i]=val1[1];
         res1[4][i]=val1[2];
         res2[3][i]=val2[1];
         res2[4][i]=val2[2];
         res3[3][i]=val3[1];
         res3[4][i]=val3[2];
      }
   }
   g1.clr();
   g1.addText("Initial state for evolution. Real part red, Imaginary\
    part\
    green");
   g1.show(res1,cl);
   g12.clr();
   g12.addText("Initial state for evolution. Real part red, Imaginary\
    part green");
   g12.show(res2,cl);
   cpmwait(waitTime*3+2);

   Z countAct=0;
   R valn=rep_(2).n_1();
   R valDiff=(getRep2()-getRep1()).abs();
    countAct++;
    times[countAct]=t();
    n_1[countAct]=valn;
    diff[countAct]=valDiff;
// 'dynamics' loop (showing a one-parameter group)
   R nStepsInv=1./nSteps;
   for (Z count=0;count<nSteps;count++){  // one state already shown,
        // so we show nSteps more, making up nSteps+1 states and nSteps
        // integration steps
      cpmprogress("Loop",R(count)*nStepsInv);
      step_(dt);
      if (filAlarm_()) fileState_(str1,str2);
// graphics part
      if (vwnAlarm_()){
         valn=rep_(2).n_1();
         valDiff=(getRep2()-getRep1()).abs();
         countAct++;
         times[countAct]=t();
         n_1[countAct]=valn;
         diff[countAct]=valDiff;
         for (i=1;i<=d;i++){
            res1[1][i]=i-1;
            res2[1][i]=i-1;
            res3[1][i]=i-1;
            C val1=getRep1()[i-1];
            C val2=getRep2()[i-1];
            C val3=val1-val2;
            res1[2][i]=val1.abs();
            res2[2][i]=val2.abs();
            res3[2][i]=val1.abs()-val2.abs();
            if (ncl1==4){
               res1[3][i]=val1[1];
               res1[4][i]=val1[2];
               res2[3][i]=val2[1];
               res2[4][i]=val2[2];
               res3[3][i]=val3[1];
               res3[4][i]=val3[2];
            }
         }
         if (autoAdj){
            y1=getRep1().sup();
            y2=getRep2().sup();
            yMax=CpmRootX::sup<R>(y1,y2);
            bool bc=band.contains(yMax);
            if (!bc){
               band=Iv(yMax*secFacInv,yMax*secFac);
               g1.setY(Iv(-band.sup(),band.sup()));
//               g1.writeGridText();
               g12.setY(Iv(-band.sup(),band.sup()));
               g22.setY(Iv(-band.sup(),band.sup()));
            //   g12.writeGridText();
            //   g22.writeGridText();
            }
         }
      }
      g.clear();
      g1.addText(text1());
      g12.addText(text2());
      g22.addText("Difference");
      g1.show(res1,cl);
      g12.show(res2,cl);
      g22.show(res3,cl);
      cpmwait(waitTime);
   }
   times=times.resize(countAct);
   n_1=n_1.resize(countAct);
   diff=diff.resize(countAct);

   g12.clr();
   g22.clr();
   g12.setAutoScale(1);
   g22.setAutoScale(1);
   g22.setWithOrigin(0);
   g12.addText("deviation of norm from 1");
   g12.show(times,n_1);
   g22.addText("distance between trajectories");
   g22.show(times,diff);
}

void Trj2::play(Graph& g, istream& str1, istream& str2)
{
   Z mL=2;
   Word loc="Trj2::play(Graph,istream,istream)";
   CPM_MA
   cpmwarning(loc&": not yet implemented");
   CPM_MZ
}//? end of cpmtrj.cpp, version 2004-11-15

