//? cpmwaves.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


// cpmwaves.cpp
// see cpmwaves.h

#include <cpmwaves.h>

using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmArrays;
using namespace CpmFunctions;

using CpmWaves::Wave;
using CpmWaves::Waves;

/////////////////////// class Wave ///////////////////////////////

Wave::Wave(Z d0, R c0, R dx0):d(d0),x(d),v(d),c(c0),dx(dx0)
{ finish();}

R_Vector Wave::acc(const R_Vector& x)
{
   Z d=x.dim();
   R_Vector res(d);
   for (Z i=2;i<d;i++){ // i.e. we let the boundary values i=1 and i=n
      // equal to 0. Thus the positions of the boundary points will
      // not be computed correctly. This is no problem since the
      // boundary points will be correctly computed by the
      // neighbor-systems
      res[i]=(x[i-1]-x[i]*2+x[i+1])*fac;
   }
   return res;
}

void Wave::step(R dt)
// terse code (following a proposal by Thomas Dera)
{
   R dth=dt*0.5; // dt half
   if (method==0){ // Euler method; direct, simple, but unstable
      // also 4 loops over the chain (2 calls to add, 1 call to +,
      // 1 copy
      R_Vector vOld=v;
      v.add(acc(x),dt); // v now changed, vNew=vOld+acc*dt
      x=x.add(v+vOld,dth);
         // efficient way of saying xNew=xOld+vOlt*dt+acc*(dt^2)/2
   }
   else if(method==1){ // explicit midpoint method; well known for
      // ordinary differential equations. Mostly interpreted in a tricky
      // way as a 2-step method which then goes under the name Leap Frog.
      // Since discretized partial differential equations are time
      // discrete
      // ordinary differential equations in an obvious way, there is
      // nothing
      // fundamentally new involved here. Nevertheless, it seems not to
      // be
      // usual to treat partial differential equations this way. Why ?
      // I tested the method also for the Schroedinger equation and found
      // norm and energy conserved to a very impressive degree (also
      // long term !)
      //
      // Three loops over the chain (3 calles to add)
      x.add(v,dth);
         // midpoint configuration
      x.add(v.add(acc(x),dt),dth);
   }
   else{ // ment as terse form of Euler, obsolete!
      // Not working; when + is being evaluated v already was changed and
      // the
      // original value is no longer available (both arguments of
      // + are references!) so the result is the same as
      // x.add(2*v.add(acc(x),dt),dth); Although the behavior for my
      // present
      // test case is not bad, this method has no value since it gives an
      // error proportional to t^2 for constant acceleration. This was a
      // hard lesson;
      // thanks to Thomas Dera for helping to understand the situation.
      x.add(v+v.add(acc(x),dt),dth);
   }
}

/*********
void Wave::prnOn(ostream& str)const
{
   d.prnOn(str);
   c.prnOn(str);
   dx.prnOn(str);
   fac.prnOn(str);
   method.prnOn(str);
   x.prnOn(str);
   v.prnOn(str);
}
*****************/

bool Wave::prnOn(ostream& str)const
{
   cpmp(d);
   cpmp(c);
   cpmp(dx);
   cpmp(fac);
   cpmp(method);
   cpmp(x);
   cpmp(v);
   return true;
}

bool Wave::scanFrom(istream& str)
{
   cpms(d);
   cpms(c);
   cpms(dx);
   cpms(fac);
   cpms(method);
   cpms(x);
   cpms(v);
   return true;
}

R_Matrix Wave::rep()const
{
   Z i;
   R_Matrix res(2,d);
   R pos=0;
   for (i=1;i<=d;i++){
      res[1][i]=pos;
      res[2][i]=x[i];
      pos+=dx;
   }
   return res;
}

/////////////////////// class Waves ///////////////////////////////

Waves::Waves(const V<Z>& vz, R c0, R dx0):d(vz.dim()),vw(d),c(c0),dx(dx0)
{
  for (Z i=1;i<=d;i++) vw[i]=Wave(vz[i],c0,dx0);
}

Waves::Waves(const V<Wave>& vw0):d(vw0.dim()),vw(vw0),
   c(vw0[1].get_c()),dx(vw0[1].get_dx())
{}

void Waves::adjust()
{
/********
   Z mesLim=3;
   if (cpmverbose>=mesLim)
      cpmmessage("Waves::adjust() started");
   for (Z i=1;i<=d;i++){
      Z iL=vw.cyclic(i-1);
      Z iR=vw.cyclic(i+1);
      R2 left=vw[iL].getRight();
      R2 right=vw[iR].getLeft();
      vw[i].setRight(right);
      vw[i].setLeft(left);
   }
   if (cpmverbose>=mesLim)
      cpmmessage("Waves::adjust() done");
***********/
   // old function cyclic does no longer exist, i
   // if needed it could easily be reconstructed
}

void Waves::step(R dt)
{
   for (Z i=1;i<=d;i++) vw[i].step(dt);
   adjust();
}

void Waves::setState(const R_Func& ampl, const R_Func& vel)
{
   Z mesLim=3;
   if (cpmverbose>=mesLim)
      cpmmessage("Waves::setState(...) started");
   R x=0;
   for (Z i=1;i<=d;i++){
      Z di_=vw[i].dim();
      for (Z j=2; j<di_; j++){ // set only the
         // propriatary points of each component
         x+=dx;
         R ax=ampl(x);
         R vx=vel(x);
         vw[i].setLocalState(j,ax,vx);

      }
   }
   adjust();
   if (cpmverbose>=mesLim)
      cpmmessage("Waves::setState(...) done");
}

Iv Waves::getSpace()const
{
   Z i,ic=0;
   for (i=1;i<=d;i++){
      ic+=vw[i].dim()-2;
   }
   return Iv(0,R(ic)*dx);
}

R_Matrix Waves::rep()const
      // graphical object representing the amplitude distribution
{
   Z i,j,ic=0;
   for (i=1;i<=d;i++){
      ic+=vw[i].dim()-2;
   }
   Z dRep=ic;
   R_Matrix res(2,dRep);
   ic=0;
   R x=0;
   for (i=1;i<=d;i++){
      Z di_=vw[i].dim();
      for (j=2; j<di_; j++){ // set only the
         // propriatary points of each component
         R2 ls=vw[i].getLocalState(j);
         ic++;
         x+=dx;
         res[1][ic]=x;
         res[2][ic]=ls[1];
      }
   }
   return res;
}

bool Waves::prnOn(ostream& str)const
{
   cpmp(d);
   cpmp(c);
   cpmp(dx);
   cpmp(method);
   cpmp(vw);
   return true;
}

bool Waves::scanFrom(istream& str)
{
   cpms(d);
   cpms(c);
   cpms(dx);
   cpms(method);
   cpms(vw);
   return true;
}
