//? cpmbodyx.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
Description: External Systems, an important building block for
   Heterogeneous systems

  Design consideration: The aim is to organize the interaction of
  subsystems to form a larger system

*/

   using CpmRoot::R;
   using CpmRoot::Z;

   using CpmRootX::Z1;
   using CpmRootX::R1;
   using CpmRootX::B;
   using CpmRootX::Record;
   using CpmRootX::RecordHandler;

   using CpmGeo::Angle;
   using CpmGeo::Iv;

   using CpmFunctions::F;
   using CpmFunctions::F1;

   using CpmArrays::V;
   using CpmArrays::X4;
   using CpmArrays::R_Vector;
   
   using CpmLinAlg::R2;

   using CpmGraphics::Graph;
   using CpmGraphics::Frame;
   using CpmGraphics::Frames;
   using CpmGraphics::SitRec;
   using CpmGraphics::UPP;
   using CpmGraphics::LOW;
   using CpmGraphics::LEFT;
   using CpmGraphics::RIG;

   using CpmImaging::Color;

   using CpmProperties::BodyProperties;
   using CpmProperties::ParticleProperties;

   using CPM_DIM::Group;
   using CPM_DIM::Spc;
   using CPM_DIM::Vec;
   using CPM_DIM::AxVec;
  
   using CPM_DIM::Plane;
   using CPM_DIM::SubsetOfSpace;
   using CPM_DIM::Sphere;
   using CPM_DIM::Rdim;
   using CPM_DIM::Skeleton;

   using CPM_RIGID_BODY::RigidObject;
   using CPM_RIGID_BODY::PointPar;
   using CPM_RIGID_BODY::ComPar;

   using CPM_CAMERA::Camera;
   
//doc namespace CpmPhysicsx{

/////////////////////////// class ExtSys ////////////////////////////////

class ExtSys: public RigidObject{ // external systems
   // uses Spc, Vec, AxVec, Force, ForTor,
   // VectorField; depends on SphPar, Graph, Camera
   // RigidObject defined in cpmcomparx.h

// External systems for confining, shaping, transporting systems of
// carrier and toner particles should be instances of classes which are
// derived from this class. This will allow the template class
// ParDynSys<> to execute the proper dynamics of the combined system.

protected:
   // The class is made to define derived classes; so it looks
   // acceptable to have protected data.

   Z1 active;
      // If active==0, the system does not interact with particles
      // and is invisible on screen.

   Z1 field_only;
      // If fieldOnly!=0, the system interacts with particles only via
      // field and not via interaction() i.e. the function interaction
      // is defined such that it can be called without creating
      // computational burden.
      // Type was Z instead of Z1 and initialisation of field_only was
      // missing in the ExtSys default constructor till 2000-4-7.
      // Terrible consequences!!!

   Vec momentum;
      // holds the time-integral of the forces applied to the system:
      // not necessarily mass times velocity (in most cases mass is
      // treated as infinitely large)
      
   AxVec angularMomentum;
      // holds the time-integral of the torques relative to x applied
      // to the system

   Vec force;
      // accumulates all forces acting on *this as a result of
      // calling interaction(...). In order to sum up only those
      // interactions that correspond to a physical state, one has
      // to call clear() at the right time. See the implementation
      // of function steplet(...) in file cpmpardynsys.h for proper
      // usage
   AxVec torque;
      // see function force: accumulates all torques ....

   ScalarField fieldU;
    
   VectorField field;
      // This is a field which the system creates and which may influence
      // other systems

    void transformStreamableData(const Group& g)
      // transforms the streamable data from ExtSys which are not
      // invariant under group action (so we don't transform the
      // parameter active)
   {
      RigidObject::transformStreamableData(g);
      momentum*=g;
      angularMomentum*=g;
   }

// group action to transform the function objects (non-streamable data)
// This operation may not be useful in derived classes since there the
// transformation of field has to be defined from transformed field
// sources which can be described by a streamable data type

    virtual void transformNonStreamableData(const Group& g)
      {field*=g; fieldU*=g;}
      // transforms the streamable data from ExtSys which are not
      // invariant under group action (so we don't transform the 
      // parameter active)

// no redefinition of update()

public:
   typedef ExtSys Type;
   CPM_IO
      // I/O for the streamable part of data
   CPM_NAM(ExtSys)

   virtual RigidObject* clone()const{ return new ExtSys(*this);}

   ExtSys(void){}
   explicit ExtSys(const RigidObject& ro,
         const VectorField& vf=VectorField()):
            RigidObject(ro),active(1),field(vf){}

   Vec get_momentum()const{ return momentum;}

   AxVec get_angularMomentum()const{ return angularMomentum;}

   Vec getMeanForce()const{ return momentum*cpminv(t_);}

// new functions to registrate the weight on bodies (2005-06-13)
   Vec getForce()const{ return force;}
      // see explanation for 'force'
   AxVec getTorque()const{ return torque;}
      // see explanation for 'torque'
   void clear(){ force=Vec(); torque=AxVec();}
      // see explanation for 'force'

   void setActivity(Z act){active=act;}

   bool fieldOnly()const{ return field_only!=0;}

   void makeFieldOnly(){ field_only=Z1(1);}

   virtual R dist(const Spc& x)const{ x;return 0;}
      // will be defined in a useful manner in derived classes

   virtual V<PointCharge> influence(const PointCharge& q)const
      // mirror charge created by q when interacting with *this
      // has to be redefined if body is metallic or dielectric
      { q; return V<PointCharge>(0);}

   virtual  V< X4<Force,Spc,Force,Spc> > forceSpots(
      const ComPar& part)const
      // List of forces which a component particles feel from the Body
      // (and which are not mediated by data member field) together with
      // the 'contact points' on the side of the component particles and
      // the system *this.
      // This function defines class behavior. In order to define this
      // function, one has to define the shape of the system and the
      // mechanisms that generate the forces.
      // alse see implementation of Body::forceSpots(...)
   {  part; return V< X4<Force,Spc,Force,Spc> >(0);}

   virtual  X2<Force,Spc> forceSpot(const PointPar& part )const
      // since the two contact points (this on part and the one 
      // on *this) have to be united anyway, and also the force
      // on part and *this determine each other (their sum is 0)
      // we give here not X4<Force,Spc,Force,Spc> but
      // simply X2<Force,Spc>
      // also see implementation of Body::forceSpot(...)
   { part; return X2<Force,Spc>();}

   ForTor interaction(const ComPar& cl, const R& dt);
      // returns the force and torque excerted on a particle cl
      // and adds to the data members momentum and angularMomentum of
      // *this the contribution corresponding to having the reaction
      // forces acting for a time span dt on the system.

   Force interaction(const PointPar& p, const R& dt);
      // returns the force (torque=0) excerted on a particle p
      // and adds to the data members momentum and angularMomentum of
      // *this the contribution corresponding to having the reaction
      // forces acting for a time span dt on the system.

   virtual R potElastic(PointPar const& p)const{ p; return 0;}
      // potential energy of p due to interaction with the body
   
   virtual R potElastic(ComPar const& p)const{p; return 0;}
      // this functions potElastic will be defined meaningfully in Body,
      // where a shape is defined

// accessing the field values, from the general to the specific

   VectorField getField(void)const{ return field;}
   ScalarField getU(void)const{ return fieldU;}
  
   FieldSpace getField(Spc const& x)const{ return field(x);}
   R getU(Spc const& x)const{ return fieldU(x);}

   VecE getE(Spc const& x)const{ return field(x).c1();}
   VecBH getBH(Spc const& x)const{ return field(x).c2();}
   VecG getG(Spc const& x)const{ return field(x).c3();}

   void setField(VectorField const& f){field=f;}
   void setU(ScalarField const& f){fieldU=f;}
 
// potential energy of a particle due to the field member

   virtual R potField(PointPar const& p)const
   { return p.pot(getField(p.get_x()));}
      // potential energy of p due to interaction with the 
      // electro-magnetic-gravitational field generated by object *this.
      // The formula is exact only if the field is homogeneous
      // Since the particle has no magnetic moment, there is no
      // contribution of the magnetic field
   virtual R potField(ComPar const& p)const
   { return p.pot(getField(p.get_x()));}
      // potential energy of p due to interaction with the 
      // electro-magnetic-gravitational field generated by object *this.
      // The formula is exact only if the field is homogeneous
      // Since the particle has a magnetic moment, there is a contribution
      // of the magnetic field

  // miscellaneous
   virtual R dis(const ExtSys& )const;
      // this function can be meaningfully defined yet on this level

   virtual void mark(const Camera& sc)const{sc;}
      // visualizing the field of an idividual body is not useful
      // since the field of all external bodies has to summed up an shown
      // as one entity. The shape, that should be shown, is not yet
      // defined in this class

   bool isActive()const{ return active == Z1(0) ? false : true;}

   void setAct(bool act){active = (act==true  ? Z1(1) : Z1(0));}

   virtual SubsetOfSpace toSubOfSpc()const{ return SubsetOfSpace();}
      //: to subset of space
      // A subset of space whch is associated with *this. Will be
      // implemented in a meaningful way in the derived class Body.


// list of new virtual functions that are needed to let 
// derivations of this class act as external systems in
// class Station

virtual bool boxSel(const ComPar& cl, Iv iv, Z i=2)
{ cl; iv; i; return false;}
      // box selector, first valid implementation in class Pot
 
virtual bool boxSel(const PointPar& cl, Iv iv, Z i=2)
{ cl; iv; i; return false;}
      // box selector
};

///////////////////////////// class Body /////////////////////////////

class Body: public ExtSys{ // external bodies 
   // adds the shape, and physical properties
   // as variables, from which then the interaction quantities (such as
   // forces) can be calculated.

protected:

// independent quantities

   R1 pinhole_;
      // if a particle has a radius smaller than this it will
      // not feel a force from the shape and thus may penetrate the body

   BodyProperties bodyProp;
      // The Body has physical properties

   SubsetOfSpace shape;
      // needed for defining repulsive forces which prevent particles
      // from penetrating the body

   Color color;
      // color for visualization of the contour of shape

   void transformNonStreamableData(const Group& g)
   {
      ExtSys::transformNonStreamableData(g);
      shape*=g;
   }

public:
   void setColor() // initializing color
   {
      color=Color(Vec::dimension==2 ? 
         CpmImaging::LIGHTGRAY: 
         CpmImaging::INVISIBLE);
   }
      // Note that in the additive coloring mode (which is the most
      // information preserving one) WHITE is not usefull since it makes
      // white independent of the color to which it gets added (just as
      // it would be the case in the color setting mode). So, we use a
      // gray instead of white.

// no redefinition of update(), transformStreamableData(const Group&)

//public:
   typedef Body Type;
   CPM_IO
   CPM_NAM(Body)
   virtual RigidObject* clone()const{ return new Body(*this);} 

   Body(){}

   explicit Body(const ExtSys& es,
      const BodyProperties& bp=BodyProperties(),
      const SubsetOfSpace& sp=SubsetOfSpace()):
         ExtSys(es), bodyProp(bp),shape(sp)
   {
      shape.set_x(x_);
      shape.set_r(r_);
      setColor();
   }

   void set_bodyProp(const BodyProperties& bp){ bodyProp=bp;}

   void set_shape(const SubsetOfSpace& sh)
   {
      shape=sh;
      shape.set_x(x_);
      shape.set_r(r_);
   }

   void setPinhole(R r){pinhole_=R1(r);}
   R getPinhole()const{ return pinhole_;}

   R getEeff()const{ return bodyProp.E_;}

   R get_dr()const{ return shape.get_dr();}
   void set_dr(R dr){ shape.set_dr(dr);}

   R get_ds()const{ return shape.get_ds();}
   void set_ds(R ds){ shape.set_ds(ds);}

   void set_E(R E){  bodyProp.E_=E; }
   void setCofOfNorRes(R cnr){ bodyProp.scnr_=cnr*cnr;} 
      // setting the 'coefficient of normal restitution

   V< X4<Force,Spc,Force,Spc> > forceSpots(const ComPar& cl)const;
      // virtual in ExtSys

   X2<Force,Spc> forceSpot(const PointPar& p)const;
      // virtual in ExtSys

   R gap(PointPar const& p)const;
      // if this is negative, it is an overlap

   R gap(ComPar const& p)const;
      // if this is negative, it is an overlap
 
// potentials of elastic interaction with (the shape of) *this
   R potElastic(PointPar const& p)const;
      // potential energy of p due to interaction with the body
   R potElastic(ComPar const& p)const;

   virtual SubsetOfSpace toSubOfSpc()const{ return shape;}
         //: to subset of space
 
   virtual void mark(const Camera& sc)const
   {     
      if (active && !SysControl::ignoreExternalSystems) 
         sc.mark(shape,LOW,color); 
   }
      // notice that the field should only be showed if added together
      // with all other external systems (argument 'color' was missing
      // till 03-01-07
      // In 3-dimensional space the representation of shapes by 
      // a 'curtain of light rays' is not trivial to interprete but
      // useful. See explanations to function 
      // void mark(const SubsetOfSpace&, SitRec i, const Color& c )const;
      // in cpmcamerax.h for details.

// see SubsetOfSpace for an explanation of the following three functions
   virtual Plane tangentialPlane(const Spc& x)const
      { return shape.tangentialPlane(x);}

   X2<R,Vec> distAndDirection(const Spc& x)const
      { return shape.distAndDirection(x);}
      // the step for numerical differentiation is shape.get_dr().

   X4<Spc,Spc,Vec,R> clsApp(const PointPar& pa)const;
      // closest approach
      // returns the points of closest approach. Let res be the return
      // value, then res.first is the point p on the surface of *this
      // for which the vector pa.get_x()-p is orthogonal to the surface
      // of *this. If *this and pa don't overlap, this is the pair of
      // points of minimum distance.
};

///////////////////// strct HomFilData /////////////////////////////////
// Data for a homogeneous field

struct HomFilData{ // data for a homogeneous field
   VecE fE_;
      // components of a homogeneous E-field
   VecBH fBH_;
      // components of a homogeneous BH-field
   VecG fG_;
      // components of a homogeneous G-field

   HomFilData(){}

   HomFilData(VecE const& fE, VecBH const& fBH, VecG const& fG):
      fE_(fE),fBH_(fBH),fG_(fG){}

   bool prnOn(ostream& str)const
   {
      cpmwt("HomFilData");
      cpmp(fE_);
      cpmp(fBH_);
      cpmp(fG_);
      return true;
   }

   bool scanFrom(istream& str)
   {
      cpms(fE_);
      cpms(fBH_);
      cpms(fG_);
      return true;
   }
   typedef HomFilData Type;
   CPM_IO_IMPL
};

///////////////////// class HomFil /////////////////////////////////////
// Homogeneous field

class HomFil: public ExtSys{ // homogeneous field
   // not derived from Body, since no shape
   // is associated with it

   HomFilData hfd_;

protected:
   void update();
   void transformNonStreamableData(const Group& g){g;}
public:
   typedef HomFil Type;
   CPM_NAM(HomFil)
   virtual ExtSys* clone()const{ return new HomFil(*this);} 
   CPM_IO
   HomFil(){}
   HomFil(ExtSys const& es, HomFilData const& hfd):ExtSys(es),hfd_(hfd)
   { field_only=1; HomFil::update();}

   void freeStep(R dt)
   {
      RigidObject::freeStep(dt);
      update();
   }

   void setOmega(AxVec const& om){ RigidObject::setOmega(om);update();}

   HomFilData getHomFilData()const{ return hfd_;}
   void mark(const Camera& sc)const{sc;}

};

////////////////////////// class VibData ////////////////////////////////

class VibData{ // vibration data
// vibration data. Representative for present (2006-02-28)
// coding style
   Rdim amp_;
      // amplitude
   Rdim omg_;
      // omega, angular velocity of vibration
   Rdim del_;
      // delta, phase of vibration 
   typedef VibData Type;
public:
   CPM_IO
   CPM_NAM(VibData)
   VibData(){}
   VibData(Record const& rec, Word const& sec);
   // A typical input for vibration taking place in the 1-2-plane only:
/*
   pot vibration data  
      // section title
   Rs vibAmp=0.05 0.05 0
      // three amplitudes
   Rs vibFrq=1 1 0
      // three vibration frequencies (in rotations per second)
   Rs vibPhs=0 0.25 0
      // three relative phases (ranging from 0 to 1 for a full rotation)
*/
   // The vibration is a generalized Lissayou motion t |--> shift, where
   // shift[i]:=vibAmp[i]*sin(2*PI*(vibFrq[i]*t+vibPhs[i]));
   // i=1,2(,3)
   // For the data from the example Record above we have a circular motion
   // since the frequencies of x-motion and y-motion are equal and there
   // is a quarter-wave shift between them. 

   Group vib(R t)const;
      // vibration shift at time t, in the data format that fits ito the
      // function X::upd() where X is a class derived from Body.
      // See explanation of the previous function for the type of motion
      // and the meaning of the parameters defining it.
};

/////////////////////// class Pot /////////////////////////////////////

struct PotData{ // data for a container vessel
   // auxiliary structure to collect geometric
   // data of a container vessel

   Rdim s;
      // length of the axis-parallel rectangular parallel-epiped
      // in 1,2,(3)-direction
   B sideWalls_;
      // if this is false, there are no side walls which
      // limit motion in 1-direction. This helps to make
      // use of toroidal geometry
   
   PotData(){}
      // notice that all data members belong to a class

   PotData(R s1, R s2, R s3):s(s1,s2,s3),sideWalls_(true){}

   explicit PotData(Rdim const& s_):s(s_),sideWalls_(true){}

  void setSideWalls(bool b){ sideWalls_=B(b);}

  bool getSideWalls()const{ return sideWalls_;} 

   bool prnOn(ostream& str)const
   {
      cpmwt("PotData");
      cpmp(s);
      cpmp(sideWalls_);
      return true;
   }

   bool scanFrom(istream& str)
   {
      cpms(s);
      cpms(sideWalls_);
      return true;
   }
   typedef PotData Type;
   CPM_IO_IMPL
};

class Pot: public Body{ // box-shaped vessel
   // a vessel having the shape of a rectangular parallel-epiped
   // (Quader in German)
   // Not closed from above; for closing we introduce class Lid. 

// independent quantities

   PotData dat;

   VibData vd;

// dependent quantities

   V<Spc> corners;

protected:

   void upd();
      // updates the values of the dependent quantities according to the
      // values of the independent ones
   void transformNonStreamableData(const Group& g){g;upd();}
public:
   typedef Pot Type;
   virtual RigidObject* clone()const{ return new Pot(*this);} 
   CPM_IO
   CPM_NAM(Pot)

   Pot(){}
      // does not give a consistent object since Pot::upd() is not called

   Pot(const Body& bo, const PotData& dat_, VibData const& vd_=VibData()):
      Body(bo),dat(dat_), vd(vd_){ upd();}

   void setPotDat(Rdim s_)
   {
      dat.s=s_;
      upd();
   }

   Rdim getPotDat()const
   {
      return dat.s;
   }

   R get_h()const{ return dat.s[2];}

   PotData getPotData()const{ return dat;}

   void freeStep(R dt)
   {
      RigidObject::freeStep(dt);
      upd();
   }

   bool boxSel(const ComPar& cl, Iv iv, Z i=2);
      // box selector
      // returns true if cl.get_x()[i] belongs to the interval
      //    Iv(iv[1]*get_h(),iv[2]*get_h()) 
      // Thus a small zone below the Lid gets described as follows:
      //    cl in zone <==> boxSel(cl,Iv(0.9,1),2)==true
      // Notice that this refers to the master system (and not to the
      // Pot system of reference. From the Pot we only use the
      // value of get_h() as a scale factor.

   bool boxSel(const PointPar& cl, Iv iv, Z i=2);
      // box selector
      // returns true if cl.get_x()[i] belongs to the interval
      //    Iv(iv[1]*get_h(),iv[2]*get_h()) 
      // Thus a small zone below the Lid gets described as follows:
      //    cl in zone <==> boxSel(cl,Iv(0.9,1),2)==true
      // Notice that this refers to the master system (and not to the
      // Pot system of reference. From the Pot we only use the
      // value of get_h() as a scale factor.

   void mark(const Camera& sc)const;
};

/////////////////////// class Lid /////////////////////////////////////

struct LidData{ // data of a lid for the container vessel
   // auxiliary structure to collect geometric
   // data of a lid for the container vessel
   Iv iv1 ;
// extension in 1-direction
   Iv iv3 ;
// extension in 3-direction
   R1 y0;
// 2-coordinate of the lid's plane; there is no extension in 
// 2-direction

   B isBottom;
   // if this is true, the repulsive body is below the plane
   // y=y0

   LidData(){}
      // notice that all data members belong to a class

   LidData(Iv ivX, Iv ivZ, R y):
      iv1(ivX),iv3(ivZ),y0(y){}

   bool prnOn(ostream& str)const
   {
      CpmRoot::writeTitle("LidData",str);
      cpmp(iv1);
      cpmp(iv3);
      cpmp(y0);
      cpmp(isBottom);
      return true;
   }

   bool scanFrom(istream& str)
   {
      cpms(iv1);
      cpms(iv3);
      cpms(y0);
      cpms(isBottom);
      return true;
   }

   typedef LidData Type;
   CPM_IO_IMPL
};


class Lid: public Body{ // a cover of the pot

   LidData dat;

   VibData vd;

// dependent quantities

   V<Spc> corners;

protected:

   void upd();
      // updates the values of the dependent quantities according to the
      // values of the independent ones
   void transformNonStreamableData(const Group& g){g;upd();}
public:
   typedef Lid Type;
   
   virtual RigidObject* clone()const{ return new Lid(*this);} 
   CPM_IO
   CPM_NAM(Lid)

   Lid(){}
      // does not give a consistent object since Lid::upd() is not called

   Lid(const Body& bo, const LidData& dat_,
      VibData const& vd_=VibData()):
      Body(bo),dat(dat_), vd(vd_){ upd();}

   LidData getData()const{ return dat;}

   void freeStep(R dt)
   {
      RigidObject::freeStep(dt);
      upd();
   }

   void makeBottom(){dat.isBottom=true;}
      // per default any Lid is of ceiling type, i.e. with the
      // free space below its plane (and the repulsive body above
      // above its plane). This function changes the Lid *this
      // such that free space is above the plane (and ...).

   void mark(const Camera& sc)const;
};

/////////////////////// class Fen /////////////////////////////////////

struct FenData{ // data for a grid-like separator (fence) of the pot
   // auxiliary structure to collect geometric
   // data for a grid-like separator that divides the container
   // vessel in two parts.

   Iv iv1 ;
// extension in 1-direction
   Iv iv3 ;
// extension in 3-direction
   R1 r;
// radius of the cylinders, the pickets of the fence
   R1 g;
// free gap between the pickets 

   FenData(){}
      // notice that all data members belong to a class

   FenData(Iv ivX, Iv ivZ, R radius, R gap):
      iv1(ivX),iv3(ivZ),r(radius),g(gap){}

   bool prnOn(ostream& str)const
   {
      CpmRoot::writeTitle("FenData",str);
      cpmp(iv1);
      cpmp(iv3);
      cpmp(r);
      cpmp(g);
      return true;
   }

   bool scanFrom(istream& str)
   {
      cpms(iv1);
      cpms(iv3);
      cpms(r);
      cpms(g);
      return true;
   }

   typedef FenData Type;
   CPM_IO_IMPL
};


class Fen: public Body{ // separator (fence) that divides the pot 
   // Fen stands for Fence
   // a way to seperate pot into two parts for the study of osmosis.
   // An array of cylindrical obstacles ('pickets') is a more simple
   // semi-permeable divice than a membrane with holes in it.

// independent data
   FenData dat;
      // determines the geometry

// dependent data
   V<R> cenX;
      // x-values of the centers of the pickets

protected:

   void upd();
      // updates the values of the dependent quantities according to the
      // values of the independent ones
   void transformNonStreamableData(const Group& g){g;upd();}
public:
   typedef Fen Type;
   virtual RigidObject* clone()const{ return new Fen(*this);} 
   CPM_IO
   CPM_NAM(Fen)

   Fen(){}
      // does not give a consistent object since Fen::upd() is not
      // called

   Fen(const Body& bo, const FenData& dat_):Body(bo),dat(dat_)
   { upd();}

   FenData getData()const{ return dat;}

   void freeStep(R dt)
   {
      RigidObject::freeStep(dt);
      upd();
   }

   void mark(const Camera& sc)const;
};

////////////////////////// class SphTank ///////////////////////////////
// Spherical tank
// The container with the most simple and regular mathematical structure
// that is expected to be useful in testing integrators. While the Pot
// container may give rise to exceptional dynamical states when particles
// hit the container walls near to an edge or even near to a corner, 
// hitting the wall of the spherical tank should always be a completely
// regular event. 

struct SphTankData{ // data for a spherical tank
   // auxiliary structure to collect geometric
   // data of a container vessel

   R1 r_;
      // length of the axis-parallel rectangular parallel-epiped
      // in 1,2,(3)-direction

   SphTankData(){}
      // notice that all data members belong to a class

   explicit SphTankData(R r):r_(r){}

   bool prnOn(ostream& str)const
   {
      CpmRoot::writeTitle("SphTankData",str);
      cpmp(r_);
      return true;
   }

   bool scanFrom(istream& str)
   {
      cpms(r_);
      return true;
   }
   typedef SphTankData Type;
   CPM_IO_IMPL
};


class SphTank: public Body{ // spherical tank
// independent quantities

   SphTankData dat_;
// no dependent quantities

protected:

   void upd();
      // updates the values of the dependent quantities according to the
      // values of the independent ones
   void transformNonStreamableData(const Group& g){g;upd();}
public:
   typedef SphTank Type;
   virtual RigidObject* clone()const{ return new SphTank(*this);} 
   CPM_IO
   CPM_NAM(SphTank)

   SphTank(){}
      // does not give a consistent object since SphTank::upd() is not
      // called

   SphTank(const Body& bo, const SphTankData& dat):Body(bo),dat_(dat)
   { upd();}

   void setSphTankDat(R r)
   {
      dat_.r_=r;
      upd();
   }

   R get_h()const{ return R(2)*dat_.r_;}

   SphTankData getSphTankData()const{ return dat_;}

   void freeStep(R dt)
   {
      RigidObject::freeStep(dt);
      upd();
   }

   void mark(const Camera& sc)const;
};

/////////////////////// class Drum /////////////////////////////////////

struct DrumData{ // data of a rotating drum
   // auxiliary structure to collect geometric
   // and kinematic data of a rotating drum

   R1 amp;
      // amplitude: amplitude of drum surface roughness

   R1 pointsPerHalfwave;
      // points per halfwave (for visualization of the contour
      // 8 is OK for most cases. Needs not to be integer

   R1 k1;
      // spatial frequency of drum surface roughness in 1-direction

   R1 k3;
      // spatial frequency of drum surface roughness in 3-direction

   R1 del1;
      // delta 1: phase constant of drum surface roughness in 1-direction

   R1 del3;
      // delta 3: phase constant of drum surface roughness in 3-direction

   R1 r0;
      // radius of the drum

   R1 omg; 
      // angular velocity of rotation arround the 3-axis (also in 2D!)

   B hollow;

   DrumData(){}
      // notice that all data members belong to a class

   DrumData(RecordHandler const& rch, Word const& sec, Z ped=1);

   bool prnOn(ostream& str)const
   {
      cpmwt("DrumData");
      cpmp(amp);
      cpmp(pointsPerHalfwave);
      cpmp(k1);
      cpmp(k3);
      cpmp(del1);
      cpmp(del3);
      cpmp(r0);
      cpmp(omg);
      cpmp(hollow);
      return true;
   }

   bool scanFrom(istream& str)
   {
      cpms(amp);
      cpms(pointsPerHalfwave);
      cpms(k1);
      cpms(k3);
      cpms(del1);
      cpms(del3);
      cpms(r0);
      cpms(omg);
      cpms(hollow);
      return true;
   }

   typedef DrumData Type;
   CPM_IO_IMPL
};

class Drum: public Body{ // a rotating rough cylindrical cavity

// independent quantities

   DrumData drd;

   VibData vd;

// dependent quantities

   Z nP;
   // number of points for visulaization, dependent - needs not
   // to be written to file

protected:
   void upd();
      // updates the values of the dependent quantities according to the
      // values of the independent ones
   void transformNonStreamableData(Group const& g){g;upd();}

public:
   typedef Drum Type;
   virtual RigidObject* clone()const{ return new Drum(*this);} 
   CPM_IO
   CPM_NAM(Drum)

   Drum(){}
      // does not give a consistent object since updateDrum is not
      // called

   Drum(Body const& bo, DrumData const& drd_, 
      VibData const& vd_=VibData()):Body(bo),drd(drd_),vd(vd_)
   { AxVec om; om[3]=drd.omg; setOmega(om),upd();}

   void setSurRog(R amp, R k1, R k3, R del1, R del3)
      // set surface roughness
   {
      drd.amp=amp;
      drd.k1=k1;
      drd.k3=k3;
      drd.del1=del1;  
      drd.del3=del3;
      upd();
   }

   R_Vector getSurRog()const
   {
      R_Vector res(5);
      res[1]=drd.amp;
      res[2]=drd.k1;
      res[3]=drd.k3;
      res[4]=drd.del1;
      res[5]=drd.del3;
      return res;
   }

   DrumData getDrumData()const{ return drd;}

   R get_r0()const{ return drd.r0;}

   void freeStep(R dt)
   {
      RigidObject::freeStep(dt);
      upd();
   }

   void mark(const Camera& sc)const;
      // Shows the contour only.
};
