//? cpmcomparcpp.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


using CPM_CAMERA::Camera;

using CPM_DIM::Spc;
using CPM_DIM::Vec;
using CPM_DIM::AxVec;
using CPM_DIM::Group;
using CPM_DIM::RefSys;
using CPM_DIM::dimVec;
using CPM_DIM::SubsetOfSpace;
using CPM_DIM::Sphere;
using CPM_DIM::WireFrame;

using CPM_PHYSICS::Force;
using CPM_PHYSICS::ForTor;
using CPM_PHYSICS::VectorField;
using CPM_PHYSICS::ExtSys;
using CPM_PHYSICS::VecE;
using CPM_PHYSICS::VecBH;
using CPM_PHYSICS::VecG;

using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmLinAlg;
using namespace CpmGraphics;
using namespace CpmAlgorithms;
using namespace CpmArrays;
using namespace CpmFunctions;
using namespace CpmImaging;
using namespace CpmProperties;
using namespace CpmPhysics;
using CpmImaging::Histogram;
   // this is the dimension-independent part (mainly physical constants,
   // so far)

///////////////////// variables of the namespace /////////////////////

bool accExp=false;
// false is OK
// accurate exponential function (from the Lie algebra of the rotation
// group to this group). If this is false a fast and generally
// sufficient approximation is used.

bool markComParMec=false;
   // if this is true, the function ComPar::mark(Camera...) shows
   // the axes of inertia and the vectors of angular velocity and
   // angular momentum

using  CpmProperties::SysControl;
   // new rigid body integrator is defined there, true is OK

bool balance=false;

namespace{
const R facEl=cpmsqrt(2.)/3.;
   // factor appearing in Hertz' formula

} // anon namespace

///////////////// class RigidObject /////////////////////////////////////

void RigidObject::make(R r, const Spc& x, const Vec& v,
         const AxVec& ori, const AxVec& omega)
{
   if (r<0){
      r=0;
      cpmwarning(
         "RigidObject::make(R r,...) r was negative, set to 0");
   }
   r_=r;
   x_=x;
   v_=v;
   ori_=ori;
   omega_=omega;
   RigidObject::update();
}

void RigidObject::transformStreamableData(const Group& g)
{
   x_*=g;
   AxVec rt=g.rot();
   v_*=rt;
   ori_*=rt;
   omega_=omega_.adjointAction(rt);
      // 'omega_=' was missing till 2005-02-14
}

void RigidObject::freeStep(R dt)
{
   if (dt==0.) return;
   t_+=dt;
   if (locked_) return;
// adjusting position from linear velocity
   Vec sh=v_*dt;
   x_+=sh;
// adjusting angular position from constant angular velocity
   AxVec rt=omega_.exp(dt);
   ori_*=rt;
   RigidObject::update();
}

void RigidObject::adjustOmega(const Spc& xp, const Vec& vp)
// omega = (xRel x vRel)/(xRel|xRel) , 'x'= cross product
{
   Vec vRel=vp-v_;
   Vec xRel=xp-x_;
   omega_=xRel.cross(vRel)*CpmRoot::inv(xRel|xRel);
}

bool RigidObject::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("RigidObject",str);
   cpmp(r_);
   cpmp(x_);
   cpmp(v_);
   cpmp(ori_);
   cpmp(omega_);
   cpmp(t_);
   return true;
}

bool RigidObject::scanFrom(istream& str)
{
   cpms(r_);
   cpms(x_);
   cpms(v_);
   cpms(ori_);
   cpms(omega_);
   cpms(t_);
   RigidObject::update(); // we did print only the independent
      // quantities
      // notice that update is virtual and we have to prevent that the
      // versions of derivates get called here
      // scope resolution was missing till 2000-12-21
   return true;
}

R RigidObject::dis(const RigidObject& ro)const
{
   using CpmRootX::sup;
   using CpmRootX::inf;
   R dTrans=(x_-ro.x_).abs();
   R phi=(ori_*(-ro.ori_)).phi();
   R dRot=phi*r_.sup(ro.r_);
      // we transform a rotation into a peripheral distance in order to
      // make spatial distances and rotational ones directly comparable.
      // In actual applications we always compare the same cluster at
      // different points in time, so the two radii r and ro.r are
      // typically the same so that it makes no difference whether we
      // combine those by sup or inf or mean
   R dPos=dTrans+dRot; // adding is 'more continuous' than sup()
   dPos=dPos/(R(2)*r_.inf(ro.r_));   // divided by diameter is more
   // natural
   // than by radius
   return dPos;
}

//////////////////////// struct PhyParPrp //////////////////////////////

bool PhyParPrp::prnOn(ostream& str)const
{
   if (!CpmRoot::writeTitle("PhyParPrp",str) ) return false;
   cpmp(m_);
   cpmp(r_);
   cpmp(q_);
   cpmp(E_);
   cpmp(U0_);
   cpmp(r0_);
   cpmp(a_);
   cpmp(scnr_);
   cpmp(forceType_);
   return true;
}

bool PhyParPrp::scanFrom(istream& str)
{
   cpms(m_);
   cpms(r_);
   cpms(q_);
   cpms(E_);
   cpms(U0_);
   cpms(r0_);
   cpms(a_);
   cpms(scnr_);
   cpms(forceType_);
   return true;
}

///////////////////// struct PointParC /////////////////////////////////

bool PointParC::prnOn(ostream& str)const
{
   if (!CpmRoot::writeTitle("PointParC",str) ) return false;
   cpmp(p_);
   cpmp(id_);
   cpmp(idValid_);
   cpmp(color_);
   return true;
}

bool PointParC::scanFrom(istream& str)
{
   cpms(p_);
   cpms(id_);
   cpms(idValid_);
   cpms(color_);
   return true;
}

///////////////////// struct PointParV /////////////////////////////////

bool PointParV::prnOn(ostream& str)const
{
   if (!CpmRoot::writeTitle("PointParV",str) ) return false;
   cpmp(owner_);
   cpmp(x_);
   cpmp(v_);
   cpmp(vDot_);
   cpmp(t_);
   cpmp(fixed_);
   return true;
}

bool PointParV::scanFrom(istream& str)
{
   cpms(owner_);
   cpms(x_);
   cpms(v_);
   cpms(vDot_);
   cpms(t_);
   cpms(fixed_);
   return true;
}

///////////////////// class PointPar ///////////////////////////////////

PointPar::PointPar(RecordHandler const& rch, Word const& sec, Z j)
{
   Word loc("PointPar(RecordHandler,Word,Z)");
   PhyParPrp ppp;
   // PhyParPrp being a state of the art struct, all its components are
   // initialized
   ContactProperties cp(rch,sec);
   ppp.set_cp(cp);
   Z ped=2; // ped = 'pedantic'
   R r;
   rch.read(sec,"r",r,ped);
      // we need r
   ped=1;
   R dr=0;
   rch.read(sec,"dr",dr,ped); // we don't need dr
   Iv ivr(r-dr,r+dr); // works also for dr<0
   r=ivr.ran(j);
   cpmassert(r>0,loc);
   ppp.r_=r;
// getting the mass
   R m;
   ped=1;
   bool b=rch.read(sec,"m",m,ped);
   if (b && m>0) ppp.m_=m;
   else{
      ped=2;
      // if we don't get rho, we are lost, since m<=0 is not OK
      R rho;
      b=rch.read(sec,"rho",rho,ped);
      R vol=CpmDim3::Spc::vol(r);
      m=vol*rho;
      ppp.m_=m;
   }
// getting the charge. Unlike for m, the value 0 is OK
// so there is no problem if all input fails
   R q=0;
   ped=1;
   b=rch.read(sec,"q",q,ped);
   if (b && q!=0) ppp.q_=q;
   else{
      R sigma=0;
      b=rch.read(sec,"sigma",sigma,ped);
      R sur=CpmDim3::Spc::sur(r);
      q=sur*sigma;
      ppp.q_=q;
   }
   ped=1;
   // read all relevant data that are there;
   // this then overrides the initialization.
   Z forceType=4; // if forceType is not there, the behavior
      // should be the same as for ComPar
   rch.read(sec,"forceType",forceType,ped);
   ppp.forceType_=forceType;
   if (forceType!=4){
      R U0=0,r0=0,a=0;
      b=rch.read(sec,"U0",U0,ped);
      if (b) ppp.U0_=U0;
      b=rch.read(sec,"r0",r0,ped);
      if (b) ppp.r0_=r0;
      b=rch.read(sec,"a",a,ped);
      if (b) ppp.a_=a;
   }
   *this=PointPar(ppp);
   Z color;
   b=rch.read(sec,"color",color,ped);
   if (b) cc.color_=color;
 //  cpmdebug(ppp);
}

AxVec PointPar::angMom(Spc const& p)const
{
   Vec xp=get_x()-p;
   AxVec res=xp.cross(mom());
   return res;
}

bool PointPar::prnOn(ostream& str)const
{
   if (!CpmRoot::writeTitle("PointPar",str) ) return false;
   cpmp(cc);
   cpmp(cv);
   return true;
}

bool PointPar::scanFrom(istream& str)
{
   cpms(cc);
   cpms(cv);
   return true;
}

void PointPar::set_id(Z id1)
{
   if (!cc.idValid_){ // a valid id gets never changed
      cc.id_=id1;
      cc.idValid_=B(true);
   }
}

B PointPar::isVal(const V<PointParV>& vv)
   // returns true if there are no reasons to
   // reject data vv
   // Present implementation based on comparing velocities with
   // SysControl::vCrit
{
   if (SysControl::vCrit<=0) return B(true);
   Z mL=2;
   const R lazyFactor=1.5;
      // since the velocity never exceeds 2*SysControl::vCrit
      // setting this factor >2 prevents any influence from it.
   static Word loc("PointPar::isVal(const V<PointParV>& vv)");
   CPM_MA
   R vStop=lazyFactor*SysControl::vCrit;
   Z i,j,d=Spc::dimension(),n=vv.dim();
   R_Matrix vel(d,n);
   Z count=0;
   for (j=1;j<=n;j++){
      Vec vj=vv[j].v_;
      if (vj.abs()>vStop) count++;
      for (i=1;i<=d;i++){
         vel[i][j]=cpmabs(vj[i]);
      }
   }
   V<R> vMax(d);
   for (i=1;i<=d;i++){
      vMax[i]=vel[i].sup();
   }
   ostringstream ost;
   ost<<endl<<"Output of PointPar::isVal(const V<PointParV>& vv)";
   ost<<endl<<"concerning maximum velocity in i-direction, i=1,2,(3)";

   for (i=1;i<=d;i++){
      ost<<endl<<"vMax["<<i<<"]="<<vMax[i];
   }
   B res(true);
   if (count>0) res=B(false);
   if (!res){
      cpmwarning(loc&" "&cpm(count)&" particles move unexpectedly fast");
      cpmmessage("Details: "&Word(ost.str()));
   }
   CPM_MZ
   return res;
}

B PointPar::isVal(const V<PointPar>& vc)
{
   Z vd=vc.dim();
   V<PointParV> vv(vd);
   for (Z i=1;i<=vd;i++) vv[i]=vc[i].cv;
   return PointPar::isVal(vv);
}

namespace{
   const R redFac=0.25;
   // If a particle velocity is found to be larger than
   // vCrit, this is considered an exceptonal situation and
   // the velocity gets reduced (by multiplication by a
   // suitable real factor) to a value redFac*vCrit;
}

void PointPar::reg(void)
{
   if (SysControl::vCrit<=0) return;
   R v1_a=cv.v_.abs();
   if (v1_a>SysControl::vCrit){
      R red1=redFac*SysControl::vCrit/v1_a;
      cv.v_*=red1;
   }
}

void PointPar::operator *=(R fac)
{
   cpmassert(fac>0,"PointPar::operator *=(R fac)");
   R fac3=fac*fac*fac;
   cc.p_.r_*=fac;
   cc.p_.r0_*=fac;
   cc.p_.a_*=fac;
   cc.p_.m_*=fac3;
}

void PointPar::intAct(R dt, const Force& fs, const FieldSpace& fval)
{
   static Word loc("PointPar::intAct(...)");
   fval; // to avoid warning
   R m=cc.p_.m_;
   cpmassert(m>0,loc);
   R mInv=R(1)/m;
   Vec acc=fs.f*mInv;
   R acca=acc.abs();
   if (SysControl::aCrit>0 && acca>SysControl::aCrit)
      acc*=(SysControl::aCrit/acca);
   Vec dv=acc*dt;
   if (SysControl::accMem){
      R dth=dt*0.5;
      dv-=cv.vDot_*dth;
      cv.vDot_=acc;
   }
   if (!isLoc()) cv.v_+=dv;
}

void PointPar::mark(Graph& g, const Color& color)const
{
   Spc x=get_x();
   R r=get_r();
   R2 center(x[1],x[2]);
   g.fillCircle(center,r,color);
}

void PointPar::mark(const Camera& sc, const Color& col)const
{
   if (col.transparent()) return;

   R r=get_r();
   Spc x=get_x();
   if (!SysControl::coloredParticles){
      if (SysControl::howToMarkParticles==1){
         WireFrame globe=WireFrame::globe(r,0,3,2,12);
         Vec shift=x-Spc();
         globe*=Group(shift);
         sc.mark(globe,col);
      }
      else{
         RefSys rs;
         V<Vec> tp=rs.getVectors();
         for (Z k=1;k<=tp.dim();k++) tp[k]*=r;
         sc.mark(x-tp[1],x+tp[1],col);
         sc.mark(x-tp[2],x+tp[2],col);
         if (tp.dim()==3) sc.mark(x-tp[3],x+tp[3],col);
      }
   }
   else{
      sc.mark(x,r,col);
   }
   if (SysControl::dtMult>0){
      Vec v=get_v();
      R dt=SysControl::tStep*SysControl::dtMult;
      // time used for visualizing the velocity as a distance
      sc.mark(x,x+v*dt,col);
   }
}

SubsetOfSpace PointPar::toSubOfSpc()const
{
   SubsetOfSpace res=Sphere(get_x(),get_r());
   return res;
}

Spc PointPar::frcCen(const PointPar& pa)const
{
   X2<R,Vec> prv=(pa.cv.x_-cv.x_).polDec();
   Vec n=prv.second();
      // vector from x to pa.x
   R s=(cc.p_.r_()-pa.cc.p_.r_())*0.5;
   return cv.x_.cen(pa.cv.x_)+n*s;
}

Force PointPar::forceFrictionAdhesionElastic( R h, const Vec& n12,
         R r1, R r2, const Vec& v1, const Vec& v2,
         const ContactProperties& cp1, const ContactProperties& cp2 )
{
   if (h>SysControl::intactRange) return Force();
   R rHarm=harmonicMean(r1,r2);
   if (h>=rHarm) return Force();
   ContactProperties cp=cp1.combine(cp2);
   R eRatio=cp.scnr_;
   R friction=cp.mu_;
      // friction coefficient
   R v0=cp.v0_;
      // tangential reaction force is nearly 0. Values <=0
      // have to be avoided; class SysControl dos not guarantee this
   R v0Min=1e-6;
   if (v0<v0Min) v0=v0Min;
   R fAdhesive=0;
   if (!SysControl::noAdhesion){
      // here fAdhesive is implemented trivially. A proper
      // implementation is essential for the electrophotographic
      // application which was among the origins of the present program.
   }
// elastic repulsion of elastic spheres at loaded contact (formulas of
// Heinrich Hertz) (only contributing if overlapping)
   R fElastic=0;
   Vec fTan;
   if (!SysControl::noElasticRepulsion && h<0.){
      R Eeff=cp.E_;
      fElastic=-facEl*Eeff*cpmsqrt(rHarm)*cpmpow(-h,R(1.5));
         // should repell 1 from 2, thus minus sign.
      if (!SysControl::noFriction){
         Vec v12=v2-v1;
         R vn=(v12|n12); // scalar product
         if (vn>0) fElastic*=eRatio;
         // inelasticity is described by letting the force
         // in release phase be by a factor smaller than in the
         // compression phase. Since energy is force times distance,
         // this should imply for normal collisions that
         // kinetic energy after the collision =
         // kinetic energy before the collision * eRatio
         Vec vTan=v12-n12*vn;
         R avt=vTan.nor_(1);
         R red=-avt/(v0+avt); // 0 for avt==0 and -1 for avt >> v0
         fTan=vTan*(friction*fElastic*red);
      }
   }
   return Force(n12*(fAdhesive+fElastic)+fTan);
}

R PointPar::elasticPotential( R h, const Vec& n12, R r1, R r2,
   ContactProperties const& cp1, ContactProperties const& cp2 )
{
   R res=0;
   n12; // not used here
   if (h>SysControl::intactRange) return res;
   R rHarm=harmonicMean(r1,r2);
   if (h>=rHarm) return res;
   ContactProperties cp=cp1.combine(cp2);
   R eRatio=cp.scnr_;
   if (!SysControl::noElasticRepulsion && h<0.){
      R Eeff=cp.E_;
      res+=facEl*Eeff*cpmsqrt(rHarm)*cpmpow(-h,R(2.5))*0.4;
   }
   return res;
}

R PointPar::collisionTime(R r1, R r2, R m1, R m2, R v,
      const ContactProperties& cp1, const ContactProperties& cp2)
{
   Word loc("PointPar::collisionTime(...)");
   R Eeff=harmonicMean(cp1.E_,cp2.E_);
      // =3/(2D) with LL's D
   R k=Eeff*cpmsqrt(harmonicMean(r1,r2)*0.5)*8/R(15);
         // 0.5*harmonicMean(r1,r2)=r1*r2/(r1+r2)
         // (8/15) comes from (4/5)*(2/3) see LL p.39
   R mu=harmonicMean(m1,m2)*0.5;
   cpmassert(k>0,loc);
   R res=(mu/k);
   res*=res;
   cpmassert(v>0,loc);
   res/=v;
   return cpmpi*30*cpmpow(res,R(0.2))/R(16);
}

Force PointPar::forceContact(const PointPar& p,
   const Vec& vThis, const Vec& vp)const
{
   X2<R,Vec> prv=gapAndDirection(p);
   R h=prv.first();
   Vec n=prv.second();
   ContactProperties cp1=get_cp();
   ContactProperties cp2=p.get_cp();
   R r1=cc.p_.r_;
   R r2=p.cc.p_.r_;
   return forceFrictionAdhesionElastic(h,n,r1,r2,vThis,vp,cp1,cp2);
}

Force PointPar::forces( const PointPar& pp )const
{
   Force f;
   if (cc.id_==pp.cc.id_) return f;
    // i.e. particle does not directly interact with itself
   Vec xRel=pp.cv.x_-cv.x_;
   R dRel=xRel.nor_();
      // now xRel is the unit vector pointing from *this to pp
   if (dRel > SysControl::intactRange) return f;

   PointCharge pc,pc_pp;
   if(!SysControl::noInteractionChargeToCharge ||
      !SysControl::noMagneticInteractionChargeToCharge){
      pc=getPointCharge();
      pc_pp=pp.getPointCharge();
   }
   if(!SysControl::noInteractionChargeToCharge){
      // Coulomb force
      f.addForce(pc.force(pc_pp));
   }
   if(!SysControl::noMagneticInteractionChargeToCharge){
      // this control structure implies that one may enable the magnetic
      // interaction while disabling the electrostatic one. This is
      // expected to be useful since for non-relativistic velocities
      // the magnetic interaction would else be dominated by the
      // electrostatic one
      f.addForce(pc.forceMag(cv.v_,pc_pp,pp.get_v()));
         // this contains the Lorentz force from the magnetic
         // field which pp generates (if it moves)
   }
   if(!SysControl::noInteractionMassToMass){
      // Newton force
      PointMass pm=getPointMass();
      PointMass pm_cl=pp.getPointMass();
      f.addForce(pm.force(pm_cl));
   }
   if(!SysControl::noElasticRepulsion){
      // 'elastic force'
      Z fT1=cc.p_.forceType_;
      Z fT2=pp.cc.p_.forceType_;
      R fM=0; // not matching force types should produce
         // no interaction
      if (fT1==fT2){
         if (fT1==0){
            // no force, fM=0 on initialization
            ;
         }
         else if (fT1==4){
            // Hertz formula for elastic repulsion of spheres
            R r1=get_r();
            R r2=pp.get_r();
            R overlap=r1+r2-dRel;
            if (overlap<=0);
            else{
               R E=harmonicMean(cc.p_.E_,pp.cc.p_.E_);
               R r=harmonicMean(r1,r2);
               fM=-facEl*E*cpmsqrt(r)*cpmpow(overlap,R(1.5));
            }
         }
         else{ // chemical forces
            R U0=harmonicMean(cc.p_.U0_,pp.cc.p_.U0_);
            R r0=(cc.p_.r0_()+pp.cc.p_.r0_())*0.5;
            R a=(cc.p_.a_()+pp.cc.p_.a_())*0.5;
            if (fT1==1) fM=R_Func::dPotLJ(U0,r0)(dRel);
            else if (fT1==2) fM=R_Func::dPotMorse(U0,r0,a)(dRel);
            else if (fT1==3) fM=R_Func::dPotUM(U0,r0,a)(dRel);
            else cpmwarning
            ("Invalid force type in PointPar::forces(PointPar)");
         }
      }
      if (fM!=0){
         if (!SysControl::noFriction){ // Only then hysteresis
            // of the elastic force is taken into account.
            // Condition added 2009-06-03. Was always correct in
            // function forceFrictionAdhesionElastic and thus in
            // class ComPar.
            Vec vRel=pp.cv.v_-cv.v_;
            R scnr=harmonicMean(cc.p_.scnr_,pp.cc.p_.scnr_);
            R dirFac=(vRel|xRel);
            if (dirFac>0){ // action only for particles that move
            // as to increasing their distance.
               if (fM>0){ // attraction
                  if (scnr>0) fM/=scnr; // increase attraction (0< res<=1)
               }
               else{ // repulsion
                  fM*=scnr; // reduce repulsion
               }
            }
         }
         Vec vf=xRel*fM;
            // fM>0 attraction, fM<0 repulsion
         f+=vf;
      }
   }
   return f;
}

R PointPar::pot( const PointPar& pp )const
{
   R res=0;
   if (cc.id_==pp.cc.id_) return res;
   Vec xRel=pp.cv.x_-cv.x_;
   R dRel=xRel.nor_();
   if (dRel > SysControl::intactRange) return res;

   if(!SysControl::noInteractionChargeToCharge){
      PointCharge pc=getPointCharge();
      PointCharge pc_pp=pp.getPointCharge();
      res+=pc.pot(pc_pp);
   }

   if(!SysControl::noInteractionMassToMass){
      PointMass pm=getPointMass();
      PointMass pm_cl=pp.getPointMass();
      res+=pm.pot(pm_cl);
   }
   if(!SysControl::noElasticRepulsion){
      Z fT1=cc.p_.forceType_;
      Z fT2=pp.cc.p_.forceType_;
      if (fT1==fT2){
         if (fT1==0){
            // no force, fM=0 on initialization
            ;
         }
         else if (fT1==4){
            // Hertz formula for elastic repulsion of spheres
            R r1=get_r();
            R r2=pp.get_r();
            R overlap=r1+r2-dRel;
            if (overlap<=0);
            else{
               R E=harmonicMean(cc.p_.E_,pp.cc.p_.E_);
               R r=harmonicMean(r1,r2);
               res+=facEl*E*cpmsqrt(r)*cpmpow(overlap,R(2.5))*0.4;
            }
         }
         else{ // chemical forces
            R U0=harmonicMean(cc.p_.U0_,pp.cc.p_.U0_);
            R r0=(cc.p_.r0_()+pp.cc.p_.r0_())*0.5;
            R a=(cc.p_.a_.x1+pp.cc.p_.a_)*0.5;
            if (fT1==1) res+=R_Func::potLJ(U0,r0)(dRel);
            else if (fT1==2) res+=R_Func::potMorse(U0,r0,a)(dRel);
            else if (fT1==3) res+=R_Func::potUM(U0,r0,a)(dRel);
            else cpmwarning
            ("Invalid force type in PointPar::pot(PointPar)");
         }
      }
   }
   return res;
}

Force PointPar::forces( const FieldSpace& fs)const
{
   Vec  res;
   if (!SysControl::noInteractionMagneticField){
      VecBH fB=fs.c2();
      Vec LorentzForce=fB.cross(cv.v_)*(-cc.p_.q_);
      res+=LorentzForce;
   }
   if (!SysControl::noInteractionElectricField){
      Vec fieldE=fs.c1();
      res+=fieldE*cc.p_.q_;
   }
   if (!SysControl::noInteractionGravitationField){
      Vec fieldG=fs.c3();
      res+=fieldG*cc.p_.m_;
   }
   return Force(res);
}

R PointPar::pot( const FieldSpace& f)const
{
   R res=0;
   if(!SysControl::noInteractionElectricField){
      res+=getPointCharge().potEnr(f.c1());
   }
   if(!SysControl::noInteractionGravitationField){
      res+=getPointMass().potEnr(f.c3());
   }
   return res;
}

R PointPar::potElastic(PointPar const& p)const
   // potential of the elastic forces contributing to forceContact
{
   if (idn(p)) return 0;
      // a particles doesn't interact with itself
   R h=gap(p);
   R res=0;
   if (!SysControl::noElasticRepulsion && h<0.){
      h=-h; // now we have overlap and h>0
      const R fac=0.4*cpmsqrt(2.)/3; // not negative here
      R Eeff=harmonicMean(cc.p_.E_,p.cc.p_.E_);
      R rHarm=harmonicMean(cc.p_.r_,p.cc.p_.r_);
      res=fac*Eeff*cpmsqrt(rHarm)*cpmpow(h,R(2.5));
      // integral of h^1.5= (h^2.5)/2.5 = (h^2.5)*0.4
   }
   return res;
}

Force PointPar::forces( const VectorField& f)const
{
   FieldSpace fs=f(cv.x_);
   return PointPar::forces(fs);
}

Force PointPar::forces( const RigidObject& ro)const
{
   if (SysControl::eta==0. || SysControl::noFriction) return Force();
   R eta=SysControl::eta;
   Vec vRel=cv.v_-ro.get_v();
   R vRelAbs=vRel.abs();
   R pi_eta_r=cpmpi*eta*cc.p_.r_;
   return Force(-vRel*6*pi_eta_r);
}

SourceOfVectorField PointPar::source()const
{
   ElcDipole ed;
   MagDipole md;
   return SourceOfVectorField(getPointCharge(),ed,md,getPointMass());
}

R PointPar::sclVal(const Word& w)const
{
   if (w=="mass") return cc.p_.m_;
   else if (w=="velocity") return cv.v_.abs();
   else if (w=="kinetic_energy") return kinEnr();
   else if (w=="charge") return cc.p_.q_;
   else cpmerror("PointPar::sclVal("&w&") not defined");
   return 0;
}

void PointPar::showProperty(const Word& theme, const V<PointPar>& vp,
   const Camera& ca, const R_Vector& parameters)
{
   Z mL=3;
   Word loc("PointPar::showProperty(...)");
   CPM_MA
   const Z nBinMax=200;
   const R electricMomentCompareFactor=4;
   const bool histogramCumulative=true;
   Z i,j,nPlanes,nBins,n;
   R_Matrix val;
   Iv iv;
   if (theme=="HistogramVelocity"){
      nPlanes=3;
      n=vp.dim();
      nBins=inf<Z>(n,nBinMax);
      val=R_Matrix(nPlanes,n);
      R maxv=parameters[1];
      iv=Iv(-maxv,maxv);
      for (i=1;i<=nPlanes;i++)
         for (j=1;j<=n;j++)
            val[i][j]=vp[j].get_v()[i];
   }
   else if (theme=="Histogram_xPosition"){
      nPlanes=1;
      n=vp.dim();
      nBins=inf<Z>(n,nBinMax);
      val=R_Matrix(nPlanes,n);
      R maxv=parameters[1];
      iv=Iv(-maxv,maxv);
      for (i=1;i<=nPlanes;i++)
         for (j=1;j<=n;j++)
            val[i][j]=vp[j].get_x()[1];
   }
   else if (theme=="Histogram_zPosition"){
      if (Spc::dimension()==2) return;
      nPlanes=1;
      n=vp.dim();
      nBins=inf<Z>(n,nBinMax);
      val=R_Matrix(nPlanes,n);
      R maxv=parameters[1];
      iv=Iv(-maxv,maxv);
      for (i=1;i<=nPlanes;i++)
         for (j=1;j<=n;j++)
            val[i][j]=vp[j].get_x()[3];
   }
   Histogram his(nPlanes,nBins,iv[1],iv[2],theme);
   for (i=1;i<=nPlanes;i++)
      for (j=1;j<=n;j++)
         his.put(i,val[i][j]);
   his.show(ca.getGraph(),histogramCumulative);
   CPM_MZ
}

///////////////////// struct ComParC ////////////////////////////////////
// constant data of ComPar

bool ComParC::prnOn(ostream& str)const
{
   cpmwt("ComParC");
   cpmp(r);
   cpmp(volume);
   cpmp(ri);
   cpmp(xr0);
   cpmp(I0);
   cpmp(ej0);
   cpmp(prop);
   cpmp(id);
//   cpmassert(id!=0,"ComParC::prnOn(istream&)");
   cpmp(idValid);
   return true;
}

bool ComParC::scanFrom(istream& str)
{
   cpms(r);
   cpms(volume);
   cpms(ri);
   cpms(xr0);
   cpms(I0);
   cpms(ej0);
   cpms(prop);
   cpms(id);
//   cpmassert(id!=0,"ComParC::scanFrom(istream&)");
   cpms(idValid);
   return true;
}

///////////////////// struct ComParV ////////////////////////////////////
// variable data of ComPar

bool ComParV::prnOn(ostream& str)const
{
   cpmwt("ComParV");
   cpmp(owner);
   cpmp(x);
   cpmp(v);
   cpmp(vDot);
   cpmp(ori);
   cpmp(omega);
   cpmp(omgDot);
   cpmp(qd);
   cpmp(t);
   cpmp(fixed);
   return true;
}

bool ComParV::scanFrom(istream& str)
{
   cpms(owner);
   cpms(x);
   cpms(v);
   cpms(vDot);
   cpms(ori);
   cpms(omega);
   cpms(omgDot);
   cpms(qd);
   cpms(t);
   cpms(fixed);
   return true;
}

///////////////////// class ComPar //////////////////////////////////////
//
// Since also in the 'two-dimensional' model, the bodies are 3
// dimensional one needs 3-dimensional computation also in 2D case
typedef CpmDim3::Vec Vec3;
typedef CpmDim3::Spc Spc3;

namespace{ // anonymous namespace for hiding names

RR3 inertia(R m, const Vec3& r)
// Huette E 11 (7), in 2-dimensional case only res.x3.x3 is important
{
   RR3 res;
   res.x1.x1=m*(r[2]*r[2]+r[3]*r[3]);
   res.x2.x2=m*(r[1]*r[1]+r[3]*r[3]);
   res.x3.x3=m*(r[1]*r[1]+r[2]*r[2]);

   res.x1.x2=-m*r[1]*r[2];
   res.x1.x3=-m*r[1]*r[3];
   res.x2.x3=-m*r[2]*r[3];

   res.x3.x1=res.x1.x3;
   res.x3.x2=res.x2.x3;
   res.x2.x1=res.x1.x2;
   return res;
}

Spc3 f2Create(const Z&i, const Vec3& vL, const Vec3& vU)
{
   using CpmRoot::ran;
   Z i0=i;
   Z shift=23;
   Z d=vL.dim();
   R x0=1;
   Vec3 vr;
   for (Z j=1;j<=d;j++){
      R xr=ran(x0,i0);
      xr=(xr+1.)*0.5;
      i0+=shift;
      vr[j]=vL[j]+xr*(vU[j]-vL[j]);
   }
   return Spc3()+vr;
}

bool chi0(const Spc3& x, const Spc3& x0, const R& r)
// indicator function of a sphere arround x0 with radius r (true inside
// the sphere)
{
   return (x-x0).abs()<=r;
}

bool f2chi(const Spc3& x, const V<Spc3>& xList, const V<R>& rList)
{
   Z n=xList.dim();
   for (Z i=1;i<=n;i++){
      bool val=chi0(x,xList[i],rList[i]);
      if (val==true) return true; // fast end, needs not to test all
      // spheres
   }
   return false;
}

R boolToR(bool x){ return (x==true ? 1 : 0);}

R c0(const Spc3& x){ return 1;}
R c1(const Spc3& x){ return x[1];}
R c2(const Spc3& x){ return x[2];}
R c3(const Spc3& x){ return x[3];}
R c11(const Spc3& x){ return x[2]*x[2]+x[3]*x[3];}
R c12(const Spc3& x){ return -x[1]*x[2];}
R c13(const Spc3& x){ return -x[1]*x[3];}
R c22(const Spc3& x){ return x[1]*x[1]+x[3]*x[3];}
R c23(const Spc3& x){ return -x[2]*x[3];}
R c33(const Spc3& x){ return x[1]*x[1]+x[2]*x[2];}

RR3 Steiner(const RR3& J, const Vec3& x, R m)
//TT Huette E 31 (8). Argument is J^A and result J^S in the notation
// there
{
   RR3 res;
   R x11=x[1]*x[1]*m;
   R x22=x[2]*x[2]*m;
   R x33=x[3]*x[3]*m;
   R x12=x[1]*x[2]*m;
   R x13=x[1]*x[3]*m;
   R x23=x[2]*x[3]*m;
   res[1][1]=J[1][1]-x22-x33;
   res[2][2]=J[2][2]-x11-x33;
   res[3][3]=J[3][3]-x11-x22;
   res[1][2]=J[1][2]+x12;
   res[1][3]=J[1][3]+x13;
   res[2][3]=J[2][3]+x23;
   res[2][1]=res[1][2];
   res[3][1]=res[1][3];
   res[3][2]=res[2][3];

   return res;
}

} // anonymous namespace

void ComPar::make(R rs, const V<Vec>& xr0_, const V<R>& ri_,
   const ParticleProperties& prop_, const Vec& ej0_, const Vec& v_,
   const AxVec& omega_)
// as this is implemented here, the computation is completely determined
// by the arguments of the function.
{
// method parameters
   const Z nT=20000;
      // so many useful data accquired in stochastical integration
   const Z nMax=1000000;
      // max of trials in stochastic integration
   const Palette colorCarrier=BLUE;
   const Palette colorToner=RED;

// Setting some data directly from the arguments of the function
   Z mL=3;
   Word loc("ComPar::make(...)");
   CPM_MA
   cv.v=v_;
   cv.omega=omega_;
   cc.xr0=xr0_;
   cc.ri=ri_;
   cc.ej0=ej0_;
   cc.prop=prop_;

   cv.ori=AxVec();
   cv.x=Spc();

   n=cc.xr0.dim(); // number of particles, is a dependent data element
      // will be set again at the end, does not cause harm
   cpmassert(n==cc.ri.dim(),loc); // consistency check

   if (n==1){ // here is no need for stochastical integration
      cc.r=rs;
      cc.xr0[1]=Vec();
      cc.ri[1]=rs;
      cc.volume=Spc3::vol(rs);
      R m=R(1)*cc.volume*cc.prop.rho;
      R IxxSphere=m*rs*rs*0.4; // (2/5)*m*r^2
      RR3 matInertia;
      matInertia[1][1]=IxxSphere;
      matInertia[2][2]=IxxSphere;
      matInertia[3][3]=IxxSphere;
      cc.I0=InTens(matInertia);
         // cv.x already correct
   }
   else{

// Threedimensional part: Even for the 2-dimensional model, particles are
// made out of three-dimensional spheres; their motion is restricted to
// a plane, though.

// First we find a rectangular box (German 'Quader') in which the whole
// body is fully contained.
      Vec3 vUpper;
      Vec3 vLower;
      Z d=vUpper.dim(); // d=3

      vUpper.set(-hugeNumber);
      vLower.set(hugeNumber);
      Vec3 vVal;
      R rVal;
      R vU,vL;
      Z i,k;
      for (i=1;i<=n;i++){
         for (k=1;k<=dimVec;k++) vVal[k]=cc.xr0[i][k];
         rVal=cc.ri[i];
         for (k=1;k<=d;k++){
            vU=vVal[k]+rVal;
            vL=vVal[k]-rVal;
            if ( vU>vUpper[k] ) vUpper[k]=vU;
            if ( vL<vLower[k] ) vLower[k]=vL;
         }
      }
      Vec3 sites=vUpper-vLower;
      R volumeBox=1;

      for (k=1;k<=d;k++) volumeBox*=sites[k];
      cpmmessage(mL,"volumeBox="&cpm(volumeBox));

// What we need now is a sequence of Vecs xr0_ which is uniformely
// distributed in vLower[i]<=xr0_[i]<=vUpper[i]

      F<Z,Spc3> create=F2<Z,Vec3,Vec3,Spc3>(vLower,vUpper)(f2Create);
      V<Spc3> xi(n); // initialized by Spc3()
      Vec3 shift; // only Vec-components will become initialized
      for (i=1;i<=n;i++){
         for (k=1;k<=dimVec;k++){ shift[k]=cc.xr0[i][k];}
         xi[i]+=shift;
      }
      F<Spc3,bool> chi=F2<Spc3,V<Spc3>,V<R>,bool>(xi,cc.ri)(f2chi);
      Z norMet=1;
      RandomIntegration<Spc3> ran(create,chi,norMet);
      // third argument introduced 2006-01-05. Prior to this
      // modification, the effect of this call was as it would
      // be now for norMet=0. This old and wrong version made
      // the quantity xx[1] equal to 1 in the following! It has
      // to come out as the volume of the particle divided by the
      // volume of the enclosing box.
      // was wrong till 2006-01-06
      Z nList=1+d+(d*(d+1))/2;
      // the list contains the volume, center-of-mass position, and the
      // non-redundant components of the tensor of inertia
      V<F<Spc3,R> > f(nList);
      Z iList=1;
      f[iList++]=F<Spc3,R>(c0);
      f[iList++]=F<Spc3,R>(c1);
      f[iList++]=F<Spc3,R>(c2);
      f[iList++]=F<Spc3,R>(c3);
      f[iList++]=F<Spc3,R>(c11);
      f[iList++]=F<Spc3,R>(c12);
      f[iList++]=F<Spc3,R>(c13);
      f[iList++]=F<Spc3,R>(c22);
      f[iList++]=F<Spc3,R>(c23);
      f[iList++]=F<Spc3,R>(c33);

      V<R> xx=ran(f,nT,nMax);
      cpmassert(xx[1]<1,loc);
      // any union of spheres is smaller in volume than the enclosing box
      // Was not satisfied till 2006-01-06
      cc.volume=R1(xx[1]*volumeBox);
      // xx[1] should be smaller than 1
      R mPrel=cc.volume.x1*cc.prop.rho;
      cpmassert(mPrel>0,loc);

      Vec3 xCoM; // position vector of center of mass with respect to
      // master
   // origin
      xCoM[1]=xx[2];
      xCoM[2]=xx[3];
      xCoM[3]=xx[4]; // otherwise all centers are in a plane so
      // that this component is zero

      Vec xCoM_; // 'copy' of xCoM with proper dimension
      for (k=1;k<=dimVec;k++){ xCoM_[k]=xCoM[k];}

      for (i=1;i<=n;i++)  cc.xr0[i]-=xCoM_;
      // aha! now the xr0 refer to the center of mass
      R_Vector dis(n);
      for (i=1;i<=n;i++) dis[i]=cc.xr0[i].abs()+cc.ri[i];
      cc.r=dis.sup();
      // This gives rise to the bounding sphere centered at the
      // center of mass. Has to be multiplied by a factor
      // for the final result.
//   cpmdata<<endl<<"First attempt for radius is "<<cc.r<<endl; //
// temporary
      // this is the right radius for the center of mass as middle-point
      RR3 matInertia;
      matInertia[1][1]=mPrel*xx[5];
      matInertia[1][2]=mPrel*xx[6];
      matInertia[1][3]=mPrel*xx[7];
      matInertia[2][2]=mPrel*xx[8];
      matInertia[2][3]=mPrel*xx[9];
      matInertia[3][3]=mPrel*xx[10];
      matInertia[3][2]=matInertia[2][3];
      matInertia[2][1]=matInertia[1][2];
      matInertia[3][1]=matInertia[1][3];
      matInertia=Steiner(matInertia,xCoM,mPrel);

      // now the inertia tensor (matrix) refers to the center of mass
      cc.I0=InTens(matInertia);
      if (balance) cc.I0.balance();
// Now we have to adjust to the volume we want
      R volumeAim=Spc3::vol(rs);
      cpmassert(cc.volume()>0,loc);
      R scaleFactor3=volumeAim/cc.volume;
      R scaleFactor=cpmpow(scaleFactor3,R(1.)/3);
      R scaleFactor5=cpmpow(scaleFactor3,R(5.)/3);
      cc.I0.scale(scaleFactor5);
      cc.volume=volumeAim;
      cpmassert(cc.volume()>0,loc);
      cc.r*=scaleFactor;
      for (i=1;i<=n;i++) cc.ri[i]*=scaleFactor;
      xCoM_*=scaleFactor;
      cv.x+=xCoM_;
      // now x is the center of mass
      for (i=1;i<=n;i++) cc.xr0[i]*=scaleFactor;
   }
   createAllDependentData();
   CPM_MZ
}

void ComPar::spin(const AxVec& omega_)
{
   cv.omega+=omega_;
   updateInternals();
}

// tools for creating unit vectors which rather uniformely fill the
// surface of a sphere

namespace{

Vec fCreate(Z const& i, R const& r, Z const& j)
// create Vec's of length r randomly in i
// For different values of j one should get unrelated functions
// of i
{
   Z shift1=CpmRoot::ranT<Z>(Z(137137),j);
   Z shift2=CpmRoot::ranT<Z>(Z(13713),j);
   Z shift3=CpmRoot::ranT<Z>(Z(1371),j);
   R r0=1;
   Z i1=i+shift1;
   R x=CpmRoot::ran(r0,i1);
   Z i2=13*i+shift2;
   R y=CpmRoot::ran(r0,i2);
   Vec a(x,y);
   if (dimVec==3){
      Z i3=17*i+shift3;
      a[3]=CpmRoot::ran(r0,i3);
   }
   return (a.polDec().second())*r;
}

bool fCond1(Vec const& x){ x; return true;}
   // trivial condition, needed to satisfy the interface of class Place

bool fCond2(X2<Vec,Vec> const& x, R const& r)
   // condition to ensure minimum distance r
{
   Vec rel=x.c1()-x.c2();
   R d=rel.abs();
   return (d>=r ? true : false);
}

} // namespace

ComPar::ComPar(R d, const ParticleProperties& pp,
   const DumbbellData& dumData, Z j)
{
   j; // not used
   R rSphere=d*0.5;
   Vec e1(1,0);
   R rc=dumData.rc;
   R re=dumData.re;
   R phi=dumData.phi;
   Vec eMagDir(cpmcos(phi),cpmsin(phi));
   V<Vec> vp(3);
   vp[1]=e1;
   vp[3]=-e1;
   V<R> rList(3);
   rList[1]=re;
   rList[2]=rc;
   rList[3]=re;
   make(rSphere,vp,rList,pp,eMagDir);
}

ComPar::ComPar(R d, const ParticleProperties& pp,
   const PairData& pairData, Z j)
{
   j; // not used
   R rSphere=d*0.5;
   Vec e1(1,0);
   V<Vec> vp(2);
   vp[2]=e1;

   R r1=pairData.r1;
   R r2=pairData.r2;
   R phi=pairData.phi;

   V<R> rList(2);
   rList[1]=r1;
   rList[2]=r2;

   Vec eMagDir(cpmcos(phi),cpmsin(phi));
   make(rSphere,vp,rList,pp,eMagDir);
}

ComPar::ComPar(R d, const ParticleProperties& pp,
   const NuggetData& nugData, Z j)
{
   Z mL=3;
   Word loc("ComPar(R,ParticleProperties,NuggetData)");
   CPM_MA
   using CpmRoot::ran;
   R rSphere=d*0.5;
   Z np=nugData.np;
   cpmassert(np>0,loc);
   if (np==1){
      V<Vec> vp(1); // one zero-vector
      V<R> rp(1,rSphere); // one radius
         // actually we only need vp and rp to have dimension 1,
         // the value of the components does not matter due to the
         // definition of function make.
      Vec dir_j;
      if (pp.J>0){ // the we change dir_j into a randomly selected unit
         // vector
         dir_j=fCreate(np+1,1.,j);
      }
      make(rSphere,vp,rp,pp,dir_j);
   }
   else{
      R overlapFac=nugData.overlapFac;
      R isotropyFac=nugData.isotropyFac;
      Z trialMax=nugData.trialMax;

// now we want to place np points on the surface of a sphere in a
// stochastical manner but nevertheless in a way that even for np small
// this surface will be filled as completly as possible. Since the
// corresponding logic is available with our Place-class, we will assure
// this by prescribing a certain minimum distance. The proper selection
// of this has to be dimension-dependent. Since (irrespective of the fact
// that all spheres are three-dimensional in our situation) the points
// under present consideration have to fill a disk in the 2-dimensional
// version and a sphere in the 3-dimensional version. We divide the
// available 'space' (area or volume) into np equal parts and convert
// this 'space per point' back into a radius. In this proceture the Pi's
// and numerical factors cancel out and we get the following simple
// prescription for radiusPerPoint.
      Vec v0;
      Z dimSurface=v0.dim()-1;
      // dimension of the surface of a 'sphere' in Spc
      R radiusPerPoint=rSphere*cpmpow((R)np,R(-1.)/dimSurface);
      R safety=(dimSurface==1 ? 3. : 2.);
      // semi-empirical factors
      R dMin=safety*radiusPerPoint*isotropyFac;
      // dMin will serve a s a minimum distance for np points to be
      // placed stochastically within the sphere of radius rSphere.
      // For dimSurface=1, rSphere=1, and np=4, we have radiusPerPoint
      // =1/4 dMin=savety*1/4. Since dMin=sqrt(2) is the upper value
      // compatible with 4 particles (maximum mutual separation), savety
      // has to be smaller than 5.657 presently I use 3. In 3 dimensions
      // one has to reduce this in order to get not too many toners made
      // out of two spheres instead of 3. One could either invest a bit
      // more work in the theory of the process (polyeders for lower
      // n's or to use repelling dynamics for equalizing the distances
      // between random initial points.

      F<Z,Vec> create=F2<Z,R,Z,Vec>(rSphere,j)(fCreate);
      // creates points on the surface of the sphere in a stochastic
      // manner
      F<Vec,bool> cond1(fCond1);
      F<X2<Vec,Vec>,bool> cond2=F1<X2<Vec,Vec>, R, bool>(dMin)(fCond2);
      Place<Vec> pl(create,cond1,cond2); // creating the placing object
      // this function determines whether ComPar::ComPar will always
      // create the same stochastical geometry or always a new one. This,
      // again, depends on function create. Presently it is made to
      // always
      // create new geometries.
      V<Vec> vp=pl(np,trialMax); // invocing the action of the placing

     // object
      np=vp.dim();   // pl may not be able to deliver np positions
      R rPart=safety*radiusPerPoint*overlapFac;
      // its better to define this as a multiple of dMin instead of
      // radiusPerPoint since this allows to correct for different form
      // factors in 2 and 3 dimensions
      V<R> rp(np,rPart);
      Vec dir_j;
      if (pp.J>0){ // the we change dir_j into a randomly selected unit
   // vector
         dir_j=fCreate(np+1,1.,j);
      }
      make(rSphere,vp,rp,pp,dir_j);
   }
   CPM_MZ
}

ComPar::ComPar(RecordHandler const& rch, Word const& sec, Z j)
{
   Z mL = 1;
   Word loc("ComPar(RecordHandler, sec = "&sec&" )");
   CPM_MA
   ContactProperties cp(rch,sec);
   ParticleProperties pp(cp);
   NuggetData nd(rch,"nugget data");
   R r;
   Z ped=2;
   bool b;
   rch.read(sec,"r",r,ped);
   ped=1;
   R dr(0);
   b=rch.read(sec,"dr",dr,ped);
   Iv ivr(r-dr,r+dr); // works also for dr<0
   r=ivr.ran(j);
   cpmassert(r>0,loc);

// getting rho or m trying m first
   R m,rho;

   b=rch.read(sec,"m",m,ped);
   if (b && m>0){
      R vol=CpmDim3::Spc::vol(r);
      cpmassert(vol>0,loc);
      rho=m/vol;
   }
   else{
      ped=2;
      rch.read(sec,"rho",rho,ped);
   }
   pp.rho=rho;
// getting q or sigma trying q first.
   R q(0),sigma(0);
   ped=1;
   b=rch.read(sec,"q",q,ped);
   if (b && q!=0){ // q==0 is considered a way to say that sigma
      // should be tried directly; if also this fails, sigma
      // will be be set to 0, which is OK
      R sur=CpmDim3::Spc::sur(r);
      cpmassert(sur>0,loc);
      sigma=q/sur;
   }
   else{
      ped=1;
      rch.read(sec,"sigma",sigma,ped);
   }
   pp.sigma=sigma;
   ped=1;

   Z color;
   b=rch.read(sec,"color",color,ped);
   if (b) pp.color=Color::toPal(color);

   R J(0),eps(1),fRelax(0);
   pp.J=J; pp.eps=eps,pp.fRelax=fRelax;

   b=rch.read(sec,"J",pp.J,ped);
   b=rch.read(sec,"eps",pp.eps,ped);
   b=rch.read(sec,"fRelax",pp.fRelax,ped);

// we should not simply read the particle properties
// since rho and sigma might not be there and be replaced by
// q and m
   *this=ComPar(r*2,pp,nd,j);
   CPM_MZ
}

void ComPar::set_m_r(R m, R r,const NuggetData& nd)
{
   if (r<=0) r=SysControl::singRange;
   cpmassert(m>0,"ComPar::set_m_r_E(...)");
   R d=r*2;
   R vol=CpmDim3::Spc::vol(r);
   R rho=m/vol;
   ParticleProperties pp;
   pp.rho=rho;
  // pp.E_=effE;
   ComPar prel(d,pp,nd);
      // this constructor together with the initializations
      // of rho and vol assures that get_m() actually returns
      // m (tested!).
   R rAct=prel.get_r();
   R fac=r/rAct;
   prel*=fac;
      // this adjusts the size but changes the mass
      // now we have to adjust the density to account for this
   R massPrel=prel.get_m();
   R fac2=m/prel.get_m();
   prel.multiplyParameter("rho",fac2);
   R massAct=prel.get_m();
   *this=prel;
}

X2<V<ComPar>,R> ComPar::makePair(R dC, R dT, Z npC, Z npT,
   ParticleProperties ppC, ParticleProperties ppT,
   R overlapFacC, R overlapFacT, R isotropyFacC, R isotropyFacT,
   R facSafty, const Vec& dir )
{
// We understand the arguments 'diameter' as diameter of a sphere having
// the same mass as the irregularly shaped particle.
// for TC=0 we create a single carrier particle with a charge given by
// ppC.sigma. The settling down-time dt is 0 in this case.

   const R dtDefault=1e-6;
   const R tinyTC=1e-6;
   const Z trialMax=4000;

   Word loc("ComPar::makePair(...)");
   NuggetData ndC;
   ndC.np=npC;
   ndC.overlapFac=overlapFacC;
   ndC.isotropyFac=isotropyFacC;
   ndC.trialMax=trialMax;
   ComPar car(dC, ppC, ndC);
   NuggetData ndT;
   ndT.np=npT;
   ndT.overlapFac=overlapFacT;
   ndT.isotropyFac=isotropyFacT;
   ndT.trialMax=trialMax;
   ComPar ton(dT, ppT, ndT);
   V<ComPar> res(2);
   res[1]=ton;
   res[2]=car;
   R rtAct=ton.get_r();
   R rcAct=car.get_r();
   R dEff=facSafty*(rtAct+rcAct);
   Vec vv=dir*dEff;
   res[1]*=Group(vv); // only toners get moved
// estimating the time needed for settling down
   R s=dEff-rcAct+rtAct; // gap + (security = rtAct)
   R qt=ton.get_q();
   R qc=car.get_q();
   R mt=ton.get_m();
   R mc=car.get_m();
   R mProd=mc*mt;
   cpmassert(mProd>0,loc);
   R mRedInv=(mc+mt)/mProd;
   R Force=influence_PHYS*qt*qc/(dEff*dEff);
      // the charge qt
   R acc=cpmabs(Force*mRedInv);
   cpmassert(acc>0,loc);
   R dt=cpmsqrt(s*2/acc);

   return X2<V<ComPar>,R>(res,dt);
}

void ComPar::set_id(Z id1)
{
   if (!cc.idValid){ // a valid id gets never changed
      cc.id=id1;
      cc.idValid=B(true);
   }
}

void ComPar::updateInternals()
// x, v, ori, omega, qd are assumed to have valid values.
// All quanties (except those that are constant anyway) get changed to
// their actual value
{
// dependent quantities to be built from body-frame data
// (name ending in '0')

   I=cc.I0;
   I*=cv.ori;

   Iinv=Iinv0;
   Iinv*=cv.ori;

   for (Z i=1;i<=n;i++){
      xr[i]=cc.xr0[i]*cv.ori;
      part[i].set_x(cv.x+xr[i]);
      part[i].set_v(cv.v+cv.omega.cross(xr[i]));
   }

// The following treatment of magnetic moment and charge is a stripped
// down version of that of particles. If one would introduce more point
// charges at positions different from x this would be more apparent

   j.setDir(cc.ej0*cv.ori);
   j.set_x(cv.x);
   q.set_x(cv.x);
   cv.qd.set_x(cv.x);
}

void ComPar::createAllDependentData()
// x, v, ori, omega are assumed to have valid values.
// All quanties (except those that are constant anyway) get changed to
// their actual value. This function will be called with data read from
// file. These may be corrupt. This should not stop the program but
// should'nt go unnoticed either. So we will throw exceptions.
{
// dependent quantities to be buitlt from body-frame data
// (name ending in '0')
   Z mL=3;
   Word loc("ComPar::createAllDependentData()");
   CPM_MA

   I=cc.I0;
   I*=cv.ori;

   Iinv0=!cc.I0;
   Iinv=Iinv0;
   Iinv*=cv.ori;

   m=R1(R(cc.volume)*R(cc.prop.rho));
   cpmassert(m()>0,loc);
   mInv=R(1)/m();
   n=cc.ri.dim();
   R qVal=cc.get_q();
      // total charge
 // building the SphPar's
   xr=V<Vec>(n);
   part=V<PointPar>(n);
   V<R> frac(n);
   Z i;
   R sum=0;
   for (i=1;i<=n;i++){
      R rii=cc.ri[i];
      rii*=(rii*rii);
      frac[i]=rii;
      sum+=rii;
   }
   cpmassert(sum>0,loc);
   R fracInv=R(1)/sum;
   ContactProperties cp=(ContactProperties)cc.prop;
   for (i=1;i<=n;i++){
      Vec xri=cc.xr0[i]*cv.ori;
      xr[i]=xri;
      part[i].set_x(cv.x+xri);
      part[i].set_v(cv.v+cv.omega.cross(xri));
      part[i].set_r(cc.ri[i]);
      R volumeFraction=frac[i]*fracInv;
      part[i].set_q(volumeFraction*qVal);
      part[i].set_m(volumeFraction*m);
      part[i].set_cp(cp); // does this set E ?
   }
   q=PointCharge(qVal,cv.x);
   q.set_r(SysControl::singRange);
      // important finding 2004-11-02: if the regularization radius
      // varies from particle to particle, the electrostatic forces
      // violate action = reaction
   R jVal=R(cc.volume)*cc.prop.J;
   j=MagDipole((cc.ej0*cv.ori)*jVal,cv.x);
   isMag_=(cc.prop.J!=0.);
   R epsr=cc.prop.eps;
   alpha=cc.volume()*3*CpmPhysics::eps0_PHYS*(epsr-1)/(epsr+2);
   cv.qd.set_x(cv.x);
   cv.qd.set_r(SysControl::singRange);
   CPM_MZ
}

void ComPar::operator *=(const Group& g)
{
// analysing argument g, for having in place the most elementary parts
// for efficient usage
   if (g==Group()) return;
   Vec sh=g.shift();
   AxVec rt=g.rot();

// center of mass
   cv.x*=g;
   cv.v*=rt;
   cv.vDot*=rt;
// electric Dipole Moment
   cv.qd*=rt;

// orientation
   cv.ori*=rt;
   cv.omega=cv.omega.adjointAction(rt);
   cv.omgDot=cv.omgDot.adjointAction(rt);
   // 'cv.omega=' was missing till 2005-02-14

// for x,v; ori, omega having their final values, we update all other
// quantities
   updateInternals();
}

ComPar& ComPar::ref_(Vec const& nv, Spc const& s0)
{
   I.ref_(nv);
      // inertia descriptor reflected
   V<Vec> rel(n);
   Z i;
   for (i=rel.b();i<=rel.e();++i)
      rel[i]=(part[i].get_x() - cv.get_x()).ref(nv);
      // relative geometry reflected
   cv.ref_(nv,s0);
      // location in space reflected
   AxVec ob = !cv.ori;
   cc.I0=(I*=ob);
   for (i=rel.b();i<=rel.e();++i) cc.xr0[i]=rel[i]*ob;
   createAllDependentData(); // guaranties consistency.
      // may involve a bit of redundant work. This is not
      // considered a problem since reflections are only
      // useful in preparing initial conditions and are not
      // caled in the implementation of dynamics.
   return *this;
}

void ComPar::ranRot(R angleRange, Z j)
{
   Group g(Angle(angleRange,DEG));
   Group gj=g.ran(get_x(),j);
   operator *=(gj);
}

void ComPar::operator *=(R fac)
{
   Z i;
   cpmassert(fac>0,"ComPar::operator *=(R fac)");
   R fac2=fac*fac;
   R fac3=fac2*fac;
   R fac5=fac3*fac2;
   cc.r*=fac;
   m*=fac3;
   cc.volume*=fac3;
   mInv/=fac3;
   q*=fac2; // surface charge density constant
   j*=fac3; // volume density of magnetic charge constant
   I.scale(fac5);
   Iinv.scale(R(1)/fac5);
   cc.I0.scale(fac5);
   Iinv0.scale(R(1)/fac5);

   for (i=1;i<=n;i++){
      xr[i]*=fac;
      cc.xr0[i]*=fac;
      cc.ri[i]*=fac; // 2006-04-28
      part[i].cc.p_.r_*=fac;
      part[i].cc.p_.q_*=fac2;
      part[i].cc.p_.m_*=fac3;
      part[i].cv.x_=cv.x+xr[i];
      part[i].cv.v_=cv.v+cv.omega.cross(xr[i]);
   }
}

void ComPar::freeStep(R dt)
{
   if (dt==0.) return;
   cv.t+=dt;
   if (!isLoc()) cv.x+=cv.v*dt;
   if (SysControl::newRigBodyInt){
      R tau=dt*0.5;
      AxVec gc=cv.omega.exp(tau,accExp);
      InTens Ic=I.adjointAction(gc);
      AxVec dIw=(Ic*cv.omega).cross(cv.omega);
      // notice for L:=I*omega, we have with Euler: L'= M
      // here M(=torque)=0 hence L'=0 (' = time derivative)
      // L'=I*omega'+omega.cross(I*omega)
      // Hence I*omega' = (I*omega).cross(omega),
      // omega' = Iinv*((I*omega).cross(omega)).
      // This formula gives much poorer conservation of angular momentum
      // and energy than the formula omega' =
      // Iinv*((Ic*omega).cross(omega))
      // adopted here. The correction I->Ic is semi-empirical
      AxVec dw=Iinv*dIw*dt;
      //AxVec dw=(!Ic)*dIw*dt; // experiment, inferior behavior
         // this is the change of  angular velocity over the whole step.
         // I would have set !Ic instead of Iinv. For unknown
         // reasons this gives energy errors of first order in dt,
         // wheras the Iinv version gives second order energy error.
         // Both variants give only first order error for the
         // conserved angular momentum component for rotation of a
         // magnetized particle in a homogeneous magnetic field.
         // They also give first order for all angular momentum
         // components in the case of free rotation.
      AxVec wc=cv.omega+dw*0.5;
         // only half the velocity increase is to be taken for
         // generating the position change as a uniform motion over
         // the whole timestep
      AxVec rt=wc.exp(dt,accExp);
      cv.ori*=rt;
      cv.omega+=dw;
   }
   else{
      AxVec rt=cv.omega.exp(dt,accExp);
      cv.ori*=rt;
      // v and omega remain unchanged (this corresponds to using
      // the 'zero-times iteration variant' of the publication
      // which is of first order only if forces depend on velocity
      // (as our friction forces do, also in the hysteresis treatment
      // of friction). One could change v and omega according to
      // memorized values of vDot and omegaDot which presently are not
      // available and would presumably introduce instability.
      // The electric moment is mainly controlled by the direction of the
      // external electric field; so it is not natural to rotate it
      // with the particle. Therefore, don't include qd*=rt;
   }
   updateInternals();
}

void ComPar::reg(void)
{
   if (SysControl::vCrit<=0) return;
   R v1_a=cv.v.abs();
   R v2_a=cv.omega.abs()*cc.r; // linear velocity at periphery
   if (v1_a>SysControl::vCrit){
      R red1=redFac*SysControl::vCrit/v1_a;
      cv.v*=red1;
   }
   if (v2_a>SysControl::vCrit){
      R red2=redFac*SysControl::vCrit/v2_a;
      cv.omega*=red2;
      for (Z i=1;i<=n;i++){
         part[i].set_v(cv.v+cv.omega.cross(xr[i]));
      }
   }
}

void ComPar::intAct(R dt, ForTor const& ft, FieldSpace const& fv)
{
   if (dt==0.) return;
// computing accelerations
   Vec force=ft.get_force();
   Vec acc1=force*mInv;
   if (SysControl::aCrit>0){ // only then the correction mechanism
      // applies
      R acc1a=acc1.abs();
      if (acc1a>SysControl::aCrit) acc1*=(SysControl::aCrit/acc1a);
   }
   AxVec torque=ft.get_torque();
   AxVec acc2=Iinv*torque;
   if (SysControl::aCrit>0){ // only then the correction mechanism
      // to change acc2 applies
      R acc2a=R(cc.r)*acc2.abs();
      if (acc2a>SysControl::aCrit) acc2*=(SysControl::aCrit/acc2a);
   }
   Vec dv=acc1*dt;
   AxVec domg=acc2*dt;
   if (SysControl::accMem){
      R dth=dt*0.5;
      dv-=cv.vDot*dth;
      domg-=cv.omgDot*dth;
      cv.vDot=acc1;
      cv.omgDot=acc2;
   }
   if (!isLoc()) cv.v+=dv;
   cv.omega+=domg;

   VecE fe=fv.c1();
   if (fe.abs()!=0) setElcDipole(dt,fe);
}

ForTor ComPar::forces( const ComPar& cl )const
{
   if (cc.id==cl.cc.id) return ForTor();
    // i.e. particle does not directly interact with itself
   Vec xRel=cl.cv.x-cv.x;
   R dRel=xRel.abs();
   R rSum=cc.r()+cl.cc.r();
   R gapCoarse=dRel-rSum;
   if (gapCoarse > SysControl::intactRange) return ForTor();
      // dRel-r-cl.r is the distance of the clusters if those are
      // idealized as spheres (see meaning of r)
   bool farAway=(gapCoarse>0.);
      // no contact force if farAway==true
   Z i,k;
   ForTor ft1;
// contact interaction
   if (!SysControl::noAdhesion||!SysControl::noElasticRepulsion){
      if (!farAway){
         for (i=1;i<=n;i++){
            PointPar p1=part[i];
            for (k=1;k<=cl.n;k++){ // notice that i and k refer
            //to different clusters, so that
            // i==k plays no special role
               PointPar p2=cl.part[k];
               Spc sp12=p1.frcCen(p2);
               Vec v1=get_v(sp12);
               Vec v2=cl.get_v(sp12);
               Force f12=p1.forceContact(p2,v1,v2);
               ft1.add(f12,sp12,get_x());
            }
         }
      }
   }
   if(!SysControl::noInteractionChargeToCharge ){
      ft1.addForce(q.force(cl.q));
   }
   if(!SysControl::noMagneticInteractionChargeToCharge){
      // this control structure implies that one may enable the magnetic
      // interaction while disabling the electrostatic one. This is
      // expected to be useful since for non-relativistic velocities
      // the magnetic interaction would else be dominated by the
      // electrostatic one
      ft1.addForce(getPointCharge().forceMag(get_v(),
         cl.getPointCharge(),cl.get_v()));
         // this contains the Lorentz force from the magnetic
         // field which pp generates (if it moves)
   }
   // magnetic interaction
   if (!SysControl::noInteractionMagnetToMagnet){
      ft1+=j.forceAndTorque(cl.j);
   }
   if (!SysControl::noInteractionElectricDipole){
      // electric dipole interaction
      ft1+=cv.qd.forceAndTorque(cl.cv.qd);
   }
   // gravitational interaction
   if(!SysControl::noInteractionMassToMass){
      PointMass pm=getPointMass();
      PointMass pm_cl=cl.getPointMass();
      ft1.addForce(pm.force(pm_cl));
   }
   return ft1;
}

R ComPar::potElastic( const ComPar& cl )const
{
   if (idn(cl)) return 0;
    // a particle does not directly interact with itself
   Vec xRel=cl.cv.x-cv.x;
   R dRel=xRel.abs();
   R rSum=cc.r()+cl.cc.r();
   R gapCoarse=dRel-rSum;
   R res=0;
   if (gapCoarse > 0) return res;
   for (Z i=1;i<=n;i++){
      PointPar pi=part[i];
      for (Z k=1;k<=cl.n;k++){ // notice that i and k refer
            //to different clusters, so that
            // i==k plays no special role
         PointPar pk=cl.part[k];
         res+=pi.potElastic(pk);
      }
   }
   return res;
}

R ComPar::pot( const ComPar& cl )const
{
   // electric dipole term yet missing; will be added after the
   // magnetic term has been tested
   // also elastic and adhesive part missing
   if (cc.id==cl.cc.id) return 0;
    // i.e. particle does not directly interact with itself
   R res=0;
   if(!SysControl::noInteractionChargeToCharge){
      res+=getPointCharge().pot(cl.getPointCharge());
   // electrostatic energy
   }
   if (!SysControl::noInteractionMagnetToMagnet){
      res+=get_j().pot(cl.get_j());
      // magnetostatic energy (included 2004-10-31)
   }
   if(!SysControl::noInteractionMassToMass){
      res+=getPointMass().pot(cl.getPointMass());
      // gravitational energy
   }
   res+=potElastic(cl);
//   cpmdebug(get_v()); // v should be zero for locked particles
   return res;
}

R ComPar::pot( const FieldSpace& f)const
{
   R res=0;
   if(!SysControl::noInteractionElectricField){
      res+=q.potEnr(f.c1());
   }
   if(!SysControl::noInteractionMagneticField){
      res+=j.potEnr(f.c2());
   }
   if(!SysControl::noInteractionGravitationField){
      res+=getPointMass().potEnr(f.c3());
   }
   return res;
}

AxVec ComPar::angMom(Spc const& p)const
{
   Vec xp=get_x()-p;
   AxVec res=xp.cross(mom())+I*cv.omega;
   return res;
}

R ComPar::angMomCom(Spc const& p, Vec const& dir)const
{
   Z dim=Spc::dimension();
   AxVec ang=angMom(p);
   if (dim==2) return ang[3];
   else return dir[1]*ang[1]+dir[2]*ang[2]+dir[3]*ang[3];
}

ForTor ComPar::forces(VectorField const& f)const
{
   Word loc("ComPar::forces(VectorField)");
   Z mL=3;
   CPM_MA
   static const R c=0.2; // controls differentiation step,
      //should be smaller than 0.5
   R cr=c*cc.r;
   FieldSpace fs=f(cv.x);
 //  fs.prnOn(cpmcerr);
   ForTor  res;
   if (!SysControl::noInteractionMagneticField){
      FieldBH BHfield=f.c2();
      res=j.forceAndTorque(BHfield,cr);
      VecBH vbh=fs.c2();
      Vec LorentzForce=vbh.cross(get_v())*(-get_q());
      res.addForce(LorentzForce);
   }
   if (!SysControl::noInteractionElectricField){
      VecE E=fs.c1();
      res.addForce(q.get_q()*E);
      FieldE Efield=f.c1();
      res+=cv.qd.forceAndTorque(Efield,cr);
   }
   if (!SysControl::noInteractionGravitationField){
      VecG G=fs.c3();
      res.addForce(m*G);
   }
 //  res.c1().prnOn(cpmcerr);
 //  res.c2().prnOn(cpmcerr);
   CPM_MZ
   return res;
}

ForTor ComPar::forces( const FieldSpace& fs)const
{
   Word loc("ComPar::forces(FieldSpace)");
   Z mL=3;
   CPM_MA
   ForTor  res;
   if (!SysControl::noInteractionMagneticField){
      // there are good reasons from my electrophotography model
      // that in a 2D model there are magnetic field lines
      // lying in the 2D system plane. So it is not allowed there
      // to model the magnetic fields as AxVec, what would be the
      // natural choice for a common representation of the Lorentz
      // force in 2D and 3D.
      VecBH fsBH=fs.c2();
      res.addTorque(j.torque(fsBH));
      Vec LorentzForce=fsBH.cross(get_v())*(-get_q());
      res.addForce(LorentzForce);
   }
   if (!SysControl::noInteractionElectricField){
      VecE fsE=fs.c1();
      res.addForce(q.force(fsE));
   }
   if (!SysControl::noInteractionGravitationField){
      VecG fsG=fs.c3();
      res.addForce(m*fsG);
   }
   CPM_MZ
   return res;
}

ForTor ComPar::forces( const RigidObject& ro)const
{
   if (SysControl::eta==0. || SysControl::noFriction) return ForTor();
   R eta=SysControl::eta;
   Vec vRel=cv.v-ro.get_v();
   AxVec omegaRel=cv.omega-ro.getOmega();
      // is OK for angular velocities !!!
   R vRelAbs=vRel.abs();
   R vRelRotAbs=R(cc.r)*omegaRel.abs();
   R pi_eta_r=cpmpi*eta*cc.r;
   Force force(-vRel*pi_eta_r*6);
   AxVec torque=-omegaRel*8*pi_eta_r*cc.r*cc.r;
   return ForTor(force,torque);
}

R ComPar::dis(const ComPar& cl)const
   // Not limited by 1 as most dis functions. Roughly spoken, this
   // is actual distance in units of particle diameter.
{
   using CpmLinAlg::R2;
   R dTrans=(cv.x-cl.cv.x).abs();
   R sr=(cc.r()+cl.cc.r());
   R phi=(cv.ori*(-cl.cv.ori)).phi();
   R dRot=sr*phi;
      // we transform a rotation into a peripheral distance in order to
      // make spatial distances and rotational ones directly comparable.
   R2 dPos2(dTrans,dRot);
   R dPos=dPos2.abs();
   return dPos*cpminv(sr);
}

void ComPar::mark(Graph& g)const
{
   static const Color particleColor(GREEN);
   for (Z i=1;i<=n;i++) part[i].mark(g,particleColor);
}

void ComPar::mark(const Camera& sc, const Color& col)const
{
   static bool minimalSkeleton=true;
   static bool markBoundingSphere=false;
   if (!markComParMec){
      const R reduct=0.4; // was 0.2
      if (col.transparent()) return;
      Z i;
      if (!SysControl::coloredParticles
         && SysControl::howToMarkParticles==0 ){
            // Then s system of orthogonal tripods is drawn
         Vec e1(1,0);
         e1*=cv.ori; // now this is a vector with a fixed
            //direction in the body frame
         Z k;
         for (i=1;i<=n;i++){
            R ri=part[i].cc.p_.r_;
            Spc xi=part[i].cv.x_;
            V<Vec> tp=xr[i].base(e1);
               // two vectors xr[i], e1 determine an orthonormal tripod
            for (k=1;k<=tp.dim();k++) tp[k]*=ri;
            sc.mark(cv.x,xi+tp[1],col);
            if (minimalSkeleton){ // most simple representation
               sc.mark(xi,xi+tp[2],col);
               if (tp.dim()==3) sc.mark(xi,xi+tp[3],col);
            }
            else{ // actual space fill well represented
               sc.mark(xi-tp[2],xi+tp[2],col);
               if (tp.dim()==3) sc.mark(xi-tp[3],xi+tp[3],col);
            }
         }
         if (SysControl::dtMult>0){
            R dt=SysControl::tStep*SysControl::dtMult;
            // and a velocity indicator for the center of mass
            Spc xcm=cv.x;
            Vec vs=cv.v;
            Spc xcm2=xcm+vs*dt;
            sc.mark(xcm,xcm2,col);
         }
      }
      else{ // independent representation of the components
         // each of them has now a velocity indicator so that
         // rotation of the particle becomes visible directly
         for (i=1;i<=n;i++) part[i].mark(sc,col);
         if (markBoundingSphere){
            PointPar ce=part[1]; // marking the bounding sphere too
            ce.set_x(cv.x);
            ce.set_r(cc.r);
            ce.mark(sc,col);
         }
      }
      if (isMag_){ // the direction of the magnetic moment is to be
         // indicated
         Color colJ(col);
            // color for marking the magnetic moment
         Vec jj=j.get_j().polDec().c2()*cc.r;
         Spc xSymbol=cv.x+jj;
         PointPar sp1;
         R rSymbol=reduct*cc.r/cpmsqrt(R(n));
         sp1.set_r(rSymbol);
         if (SysControl::coloredParticles && Spc::dimension()==2){
            colJ=Color(WHITE);
            // here a disk looks better as the center of the particle
            // than as the hat of the 'arrow'
            sp1.set_x(cv.x);
         }
         else{
            sp1.set_x(xSymbol);
         }
         sp1.mark(sc,colJ); // show symbol particle
         sc.mark(cv.x,xSymbol,colJ); // show connecting line
      }
   }
   else{ // markComParMec
      static bool firstrun=true;
      static R angMomFac=1;
      static R angVelFac=1;
      Word loc="ComPar::mark(Camera,Color)";
      R r=get_r();
      Vec e1=I.e1()*I.i1();
      Vec e2=I.e2()*I.i2();
      Vec e3=I.e3()*I.i3();
      R a1=e1.abs();
      cpmassert(a1>0,loc);
      R renorm=r/a1;
      Spc x=get_x();
      e1*=renorm;
      e2*=renorm;
      e3*=renorm;
      sc.mark(x,x+e1,col);
      sc.mark(x,x+e2,col);
      sc.mark(x,x+e3,col);
      Vec om(cv.omega[1],cv.omega[2],cv.omega[3]);
      if (firstrun){
         R oma=om.abs();
         if (oma>0) angVelFac=r/oma;
      }
      om*=angVelFac;
      Spc x1=x+om;
      PointPar sym1;
      R r1=r*0.02;
      sym1.set_r(r1);
      sym1.set_x(x1);
      sc.mark(x,x1,col);
      sym1.mark(sc);
      AxVec io=I*cv.omega;
      Vec iom(io[1],io[2],io[3]);
      if (firstrun){
         R ioma=iom.abs();
         if (ioma>0) angMomFac=r/ioma;
         firstrun=false;
      }
      iom*=angMomFac;
      Spc x2=x+iom;
      PointPar sym2;
      R r2=r*0.04;
      sym2.set_r(r2);
      sym2.set_x(x2);
      sc.mark(x,x2,col);
      sym2.mark(sc);
   }
}

// Now I/O for ComPar works through the division of data into a
// ComParC (constant data) and a ComParV (variable data) component.
// Till 02-03-13 writing on file and MPI-messaging used different
// functions, so that it was hard to see whether strange
// dynamical behavior of configurations read from file was based
// on a difference in available data.

bool ComPar::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("ComPar",str);
   cpmp(cc);
   cpmp(cv);
   return true;
}

bool ComPar::scanFrom(istream& str)
{
   cpms(cc);
   cpms(cv);
   createAllDependentData();
   return true;
}

void ComPar::setDirJ(const Vec& dir)
{
   Vec v1=getDirJ();
   AxVec av=v1.rotTo(dir);
   operator*=(Group(cv.x,av));
}

// new version that expects E to be the consistent mean field
// (that is constant during a complete integration step (step
// as opposed to steplet) so that fRelax and dt need not to
// be used
void ComPar::setElcDipole(R dt, const VecE& E)
{
   static R Emax=3e6; // maximum electric field that
   // may exist in air: 3MV/m
   if (SysControl::noInteractionChargeToCharge ||
      SysControl::noInteractionElectricDipole) return;
   dt; // not used
   VecE Eeff=E;
   R ea=Eeff.abs();
   if (ea>Emax) Eeff*=(Emax/ea);
   Vec qdInduced=Eeff*alpha;
   cv.qd.setMoment(qdInduced);
}

bool ComPar::multiplyParameter(Word const& w, R fac)
{
   Word loc("ComPar::multiplyParameter(...)");
   if (w=="charge"||w=="sigma"){
      if (fac==1.) return true;
      cc.prop.sigma*=fac;
      q*= fac;
      for (Z i=1;i<=n;i++) part[i].cc.p_.q_*=fac;
   }
   else if (w=="alphaAdhesion"){
      if (fac==1.) return true;
      cc.prop.alphaAdhesion_*=fac;
     // for (Z i=1;i<=n;i++) part[i].cp.alphaAdhesion*=fac;
      for (Z i=1;i<=n;i++) part[i].cc.p_.alphaAdhesion_*=fac;
   }
   else if (w=="relativeRangeAdhesion"){
      if (fac==1.) return true;
      cc.prop.relativeRangeAdhesion_*=fac;
      for (Z i=1;i<=n;i++) part[i].cc.p_.relativeRangeAdhesion_*=fac;
   }
   else if (w=="effectiveE"){
      if (fac==1.) return true;
      cc.prop.E_*=fac;
      for (Z i=1;i<=n;i++) part[i].cc.p_.E_*=fac;
   }
   else if (w=="scnr"){
      if (fac==1.) return true;
      cc.prop.scnr_*=fac;
      for (Z i=1;i<=n;i++) part[i].cc.p_.scnr_*=fac;
   }
   else if (w=="J"){
      if (fac==1.) return true;
      cc.prop.J*=fac;
      j*=fac;
      if (fac==0) isMag_=false;
   }
   else if (w=="rho"){
      if (fac==1.) return true;
      cpmassert(fac!=0,loc);
      R facInv=R(1)/fac;
      cc.prop.rho*=fac;
      m*=fac;
      mInv*=facInv;
      I.scale(fac); // notice that the geometry is fixed
      Iinv.scale(facInv);
      cc.I0.scale(fac);
      Iinv0.scale(facInv);
      for (Z i=1;i<=n;i++) part[i].cc.p_.m_*=fac;
   }
   else {
        // even if fac==1 we return false if w does not match a defined
        // meaning
      cpmwarning("ComPar::multiplyParameter(...): "&w&
         " is invalid argument");
      return false;
   }
      // should not stop for a parameter which is not in the list
   return true;
}

 void ComPar::set_cp(const ContactProperties& cp)
  // setting the ContactProperties subset of the ParticleProperties
{
   cc.prop.set_cp(cp);
   for (Z i=1;i<=n;i++) part[i].set_cp(cp);
}

void ComPar::set_pp(const ParticleProperties& pp)
 // setting the ContactProperties subset of the ParticleProperties
{
   set_cp(pp);
   R fac,valOld,valNew;
   Word w;

   w="J";
   valOld=cc.prop.J;

   if (valOld!=0){
      valNew=pp.J;
      fac=valNew/valOld;
      multiplyParameter(w,fac);
         // if fac==0, this function takes care of updating isMag
   }
   else{
      cc.prop.J=pp.J;
      R jVal=R(cc.volume)*cc.prop.J;
      j=MagDipole((cc.ej0*cv.ori)*jVal,cv.x);
      isMag_=(jVal!=0);
   }

   w="rho";
   valOld=cc.prop.rho;
   valNew=pp.rho;
   cpmassert(valOld!=0,"ComPar::set_pp(...)");
   fac=valNew/valOld;
   multiplyParameter(w,fac);

   cc.prop.eps=pp.eps;
}

void ComPar::set_E(R E)
{
   ContactProperties cp=cc.prop;
   cp.E_=E;
   set_cp(cp);
}

void ComPar::setCofOfNorRes(R cnr)
{
   ContactProperties cp=cc.prop;
   cp.scnr_=cnr*cnr;
   set_cp(cp);
}

R ComPar::gap(const RigidObject& ro, Z i)const
{
   Z k;
   Vo<R> rox(n);
   Vo<R> roxAbs(n);
   for (k=1;k<=n;k++) rox[k]=ro.relCor(part[k].get_x(),i);
   for (k=1;k<=n;k++) roxAbs[k]=cpmabs(rox[k]);
   for (k=1;k<=n;k++){
      roxAbs[k]-=cc.ri[k];
   }
   T2<Z> pzz=roxAbs.indInfSup();
   Z iMin=pzz.c1();
   R d0=rox[iMin];
   R r0=cc.ri[iMin];
   return d0>0 ? d0-r0 : d0+r0;
}

R ComPar::gap(const ComPar& cl)const
{
   Vec xRel=cl.cv.x-cv.x;
   R dRel=xRel.abs();
   R rSum=cc.r()+cl.cc.r();
   R gapCoarse=dRel-rSum;
   if (gapCoarse>0) return gapCoarse;
   R res=rSum;
   for (Z i=1;i<=n;i++){
      PointPar pi=part[i];
      for (Z k=1;k<=cl.n;k++){ // notice that i and k refer
            //to different clusters, so that
            // i==k plays no special role
         PointPar pk=cl.part[k];
         R gik=pi.gap(pk);
         if (gik<res) res=gik;
      }
   }
   return res;
}

SubsetOfSpace ComPar::toSubOfSpc()const
{
   SubsetOfSpace res=Sphere(part[1].get_x(),part[1].get_r());
   for (Z i=2;i<=n;i++){
      res=res|Sphere(part[i].get_x(),part[i].get_r());
   }
   return res;
}

ComParC ComPar::getC()const
{
   return cc;
}

ComParV ComPar::getV()const
{
   return cv;
}

void ComPar::set(const ComParC& clc, const ComParV& clv)
{
   cc=clc;
   cv=clv;
   createAllDependentData();
}

ComPar::ComPar(const ComParC& clc, const ComParV& clv)
{
   set(clc,clv);
}

V<ComPar> ComPar::neutralize(const V<ComPar>& vc)
{
   Z mL=4;
   Word loc("ComPar::neutralize(V<ComPar>&)");
   cpmmessage(mL,loc&" started");
   Z i,nvc=vc.dim();
   if (nvc==1) return vc;
   R qP=0;
   R qN=0;
   R qi;
   for (i=1;i<=nvc;i++){
      qi=vc[i].get_q();
      if (qi>0) qP+=qi; else qN-=qi; //qN becomes positive too
   }
   if (cpmverbose>mL){
      cpmcerr<<endl<<" prior to neutralizing"<<endl;
      cpmdebug(qP);
      cpmdebug(qN);
      cpmdebug(qP-qN);
   }
   if (qP==0. || qN==0.){
      cpmwarning(loc&
      ": all charges have the same sign, input returned without change"
      );
      return vc;
   }
   R f=qP/qN;
   f=cpmsqrt(f);
   R fInv=R(1)/f;
   V<ComPar> res(vc);
   for (i=1;i<=nvc;i++){
      qi=vc[i].get_q();
      if (qi>=0){
         res[i].set_q(qi*fInv);
         res[i].cc.prop.sigma*=fInv;
            // this is needed since IO is such that
            // charge isn't written to file but reconstructed
            // from r and prop.sigma. This is more legacy than
            // a good idea
      }
      else{
         res[i].set_q(qi*f);
         res[i].cc.prop.sigma*=f;
      }
   }
   qP=0;
   qN=0;
   for (i=1;i<=nvc;i++){
      qi=res[i].get_q();
      if (qi>0) qP+=qi; else qN-=qi;
   }
   if (cpmverbose>mL){
      cpmcerr<<endl<<" after neutralizing"<<endl;
      cpmdebug(qP);
      cpmdebug(qN);
      cpmdebug(qP-qN);
   }
   cpmmessage(mL,loc&" done");
   return res;
}

SourceOfVectorField ComPar::source()const
{
   return SourceOfVectorField(q,cv.qd,j,getPointMass());
}

VectorField ComPar::farField(const V<ComPar>& par,
   const V<Spc>& pointList)
{
   Z i,np=par.dim();
   V<SourceOfVectorField> sl(np);
   for (i=1;i<=np;i++) sl[i]=par[i].source();
   sl=SourceOfVectorField::redistribute(sl,pointList);
   VectorField res;
   for (i=1;i<=sl.dim();i++) res+=sl[i].field();
   return res;
}

VectorField ComPar::meanField(const V<ComPar>& par,
   const V<Spc>& pointList)
{
   Z i,j,np=par.dim(),n=pointList.dim();
   V<SourceOfVectorField> sl(np);
   for (i=1;i<=np;i++) sl[i]=par[i].source();
   sl=SourceOfVectorField::redistribute(sl,pointList);
      // complexity proportional to np
   V<FieldSpace> val(n);
   for (i=1;i<=n;i++){ // complexity proportional to n*n not np*np !
      Spc xi=pointList[i];
      for (j=1;j<=n;j++){
         if (i==j) continue;
         val[i]+=sl[j].eval(xi);
      }
   }
   return VectorField(pointList,val);
}

B ComPar::isVal(const V<ComParV>& vv)
   // returns true if there are no reasons to
   // reject data vv
   // Present implementation based on comparing velocities with
   // SysControl::vCrit
{
   if (SysControl::vCrit<=0) return B(true);
   Z mL=2;
   const R lazyFactor=1.5;
      // since the velocity never exceeds 2*SysControl::vCrit
      // setting this factor >2 prevents any influence from it.
   static Word loc("ComPar::isVal(const V<ComParV>& vv)");
   CPM_MA
   R vStop=lazyFactor*SysControl::vCrit;
   Z i,j,d=Spc::dimension(),n=vv.dim();
   R_Matrix vel(d,n);
   Z count=0;
   for (j=1;j<=n;j++){
      Vec vj=vv[j].v;
      if (vj.abs()>vStop) count++;
      for (i=1;i<=d;i++){
         vel[i][j]=cpmabs(vj[i]);
      }
   }
   V<R> vMax(d);
   for (i=1;i<=d;i++){
      vMax[i]=vel[i].sup();
   }
   ostringstream ost;
   ost<<endl<<"Output of ComPar::isVal(const V<ComParV>& vv)";
   ost<<endl<<"concerning maximum velocity in i-direction, i=1,2,(3)";

   for (i=1;i<=d;i++){
      ost<<endl<<"vMax["<<i<<"]="<<vMax[i];
   }
   ost<<endl<<"SysControl::aCrit="<<SysControl::aCrit;
   ost<<endl<<"SysControl::vCrit="<<SysControl::vCrit;
   ost<<endl<<"vStop="<<vStop;
   B res(true);
   if (count>0) res=B(false);
   if (!res){
      cpmwarning(loc&" "&cpm(count)&" particles move unexpectedly fast");
      cpmmessage("Details: "&Word(ost.str()));
   }
   CPM_MZ
   return res;
}

B ComPar::isVal(const V<ComPar>& vc)
{
   Z vd=vc.dim();
   V<ComParV> vv(vd);
   for (Z i=1;i<=vd;i++) vv[i]=vc[i].cv;
   return ComPar::isVal(vv);
}

R ComPar::sclVal(const Word& w)const
{
   if (w=="mass") return m;
   else if (w=="velocity") return cv.v.abs();
   else if (w=="angular_velocity") return cv.omega.abs();
   else if (w=="kinetic_energy") return kinEnr();
   else if (w=="charge") return get_q();
   else if (w=="electric_dipole") return get_qd().abs();
   else if (w=="electric_polarization")
      return get_qd().abs()/cc.volume;
   else if (w=="magnetic_dipole") return get_j().abs();
   else if (w=="volume") return cc.volume;
   else if (w=="surface") return getSurface();
   else cpmerror("ComPar::sclVal("&w&") not defined");
   return 0;
}

void ComPar::showProperty(Word const& theme, V<ComPar> const& vp,
   Camera const& ca, R_Vector const& parameters)
// here 'Carrier' means particle with non-vanishing magnetic moment
// 'Toner' means particle which is not a carrier particle.
{
   Z mL=3;
   Word loc("ComPar::showProperty(...)");
   CPM_MA
   const Z nBinMax=200;
   const R electricMomentCompareFactor=4;
   const bool histogramCumulative=true;
   Z i,j,nPlanes,nBins,n;
   R_Matrix val;
   Iv iv;
   if (theme=="HistogramMagneticMoment"){
      nPlanes=3;
      n=vp.dim();
      V<bool> sel(n);
      for (i=1;i<=n;i++) sel[i]=vp[i].isMag();
      V<ComPar> vSel=vp.select(sel);
      n=vSel.dim();
      nBins=inf<Z>(n,nBinMax);
      val=R_Matrix(nPlanes,n);
      R_Vector scale(n);
      for (j=1;j<=n;j++)
            scale[j]=vSel[j].get_j().getMoment().abs();
      R maxv=scale.sup();
      iv=Iv(-maxv,maxv);
      for (i=1;i<=nPlanes;i++)
         for (j=1;j<=n;j++)
            val[i][j]=vSel[j].get_j().getMoment()[i];

   }
   else if (theme=="HistogramTonerElectricMoment"){
      nPlanes=3;
      n=vp.dim();
      V<bool> sel(n);
      for (i=1;i<=n;i++) sel[i]=!vp[i].isMag();
      V<ComPar> vSel=vp.select(sel);
      n=vSel.dim();
      nBins=inf<Z>(n,nBinMax);
      val=R_Matrix(nPlanes,n);
      R E_bias=parameters[1];
      R_Vector alpha(n);
      for (j=1;j<=n;j++)
            alpha[j]=vSel[j].getAlpha();
      R maxv=electricMomentCompareFactor*E_bias*alpha.sup();
      iv=Iv(-maxv,maxv);
      for (i=1;i<=nPlanes;i++)
         for (j=1;j<=n;j++)
            val[i][j]=vSel[j].get_qd().getMoment()[i];
   }
   else if (theme=="HistogramCarrierElectricMoment"){
      nPlanes=3;
      n=vp.dim();
      V<bool> sel(n);
      for (i=1;i<=n;i++) sel[i]=vp[i].isMag();
      V<ComPar> vSel=vp.select(sel);
      n=vSel.dim();
      nBins=inf<Z>(n,nBinMax);
      val=R_Matrix(nPlanes,n);
      R E_bias=parameters[1];
      R_Vector alpha(n);
      for (j=1;j<=n;j++)
            alpha[j]=vSel[j].getAlpha();
      R maxv=electricMomentCompareFactor*E_bias*alpha.sup();
      iv=Iv(-maxv,maxv);
      for (i=1;i<=nPlanes;i++)
         for (j=1;j<=n;j++)
            val[i][j]=vSel[j].get_qd().getMoment()[i];
   }
   else if (theme=="HistogramTonerVelocity"){
      nPlanes=3;
      n=vp.dim();
      V<bool> sel(n);
      for (i=1;i<=n;i++) sel[i]=!vp[i].isMag();
      V<ComPar> vSel=vp.select(sel);
      n=vSel.dim();
      nBins=inf<Z>(n,nBinMax);
      val=R_Matrix(nPlanes,n);
      R maxv=parameters[1];
      iv=Iv(-maxv,maxv);
      for (i=1;i<=nPlanes;i++)
         for (j=1;j<=n;j++)
            val[i][j]=vSel[j].get_v()[i];
   }
   else if (theme=="HistogramCarrierVelocity"){
      nPlanes=3;
      n=vp.dim();
      V<bool> sel(n);
      for (i=1;i<=n;i++) sel[i]=vp[i].isMag();
      V<ComPar> vSel=vp.select(sel);
      n=vSel.dim();
      nBins=inf<Z>(n,nBinMax);
      val=R_Matrix(nPlanes,n);
      R maxv=parameters[1];
      iv=Iv(-maxv,maxv);
      for (i=1;i<=nPlanes;i++)
         for (j=1;j<=n;j++)
            val[i][j]=vSel[j].get_v()[i];
   }
   else if (theme=="Histogram_xPosition"){
      nPlanes=1;
      n=vp.dim();
      nBins=inf<Z>(n,nBinMax);
      val=R_Matrix(nPlanes,n);
      R maxv=parameters[1];
      iv=Iv(-maxv,maxv);
      for (i=1;i<=nPlanes;i++)
         for (j=1;j<=n;j++)
            val[i][j]=vp[j].get_x()[1];
   }
   else if (theme=="Histogram_yPosition"){
      nPlanes=1;
      n=vp.dim();
      nBins=inf<Z>(n,nBinMax);
      val=R_Matrix(nPlanes,n);
      R maxv=parameters[1];
      iv=Iv(-maxv,maxv);
      for (i=1;i<=nPlanes;i++)
         for (j=1;j<=n;j++)
            val[i][j]=vp[j].get_x()[2];
   }
   else if (theme=="Histogram_zPosition"){
      cpmassert(Spc::dimension()==3,loc);
      nPlanes=1;
      n=vp.dim();
      nBins=inf<Z>(n,nBinMax);
      val=R_Matrix(nPlanes,n);
      R maxv=parameters[1];
      iv=Iv(-maxv,maxv);
      for (i=1;i<=nPlanes;i++)
         for (j=1;j<=n;j++)
            val[i][j]=vp[j].get_x()[3];
   }
   else{
      cpmerror(loc&": unvalid theme "&theme);
   }
   Histogram his(nPlanes,nBins,iv[1],iv[2],theme);
   for (i=1;i<=nPlanes;i++)
      for (j=1;j<=n;j++)
         his.put(i,val[i][j]);
   his.show(ca.getGraph(),histogramCumulative);
   CPM_MZ
}

void ComPar::showVelocities(const V<ComPar>& vp, const Camera& ca)
{
   Z i,j,d=Spc::dimension(),n=vp.dim();
   Z nBins=64;
   R vMin=0;
   R vMax=SysControl::vCrit*2;
   Histogram his(d,nBins,vMin,vMax);
   for (i=1;i<=d;i++){
      for (j=1;j<=n;j++){
         his.put(i,cpmabs((vp[j].get_v())[i]));
      }
   }
   his.show(ca.getGraph());
}

Z ComPar::com(const ComPar& p)const{ return cv.x.com(p.cv.x);}
// comparison based solely on position
