//? cpmcomparx.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   cpmcomparx.h
   Description: compar stands for composite particles. Declares a class of
   irregularly shaped particles as clusters of overlapping spherical
   particles. In close analogy to this, also simple spherical
   particles are treated.

   The treatment of rigid body motion is in full agreement with Huette p.
   E33 equ. (16). The time step according to the half-step linear advance
   scheme (explicit midpoint method) that arrises from the MPM-method
   (see
   Ulrich Mutze: Predicting Classical Motion Directly from the Action
   Principle II, http://www.ma.utexas.edu/mp_arc/index-99.html, entry
   99-271) after limit of velocity spread to zero is implemented here
   much more elegantly than before.

   Emphasis was put on defining all structures such that no formal
   changes are needed to switch from 2 space dimensions to three ones.
   This askes for a consequent distinction between axial and polar
   vectors, which brings out the physical structure very nicely. I had to
   work hard (more than 50 pages of calculations to see things clear
   enough) to find the present framework although it does not deviate in
   any point from best practicies.

*/

using namespace CpmRoot;

using CPM_DIM::Vec;
using CPM_DIM::AxVec;
using CPM_DIM::Spc;
using CPM_DIM::Group;
using CPM_DIM::RefSys;
using CPM_DIM::Plane;
using CPM_DIM::SubsetOfSpace;
using CPM_DIM::InTens;
using CPM_DIM::Rdim;
using CPM_DIM::Rdim_;
using CPM_DIM::Box;

using CPM_PHYSICS::Force;
using CPM_PHYSICS::ForTor;
using CPM_PHYSICS::PointMass;
using CPM_PHYSICS::PointCharge;
using CPM_PHYSICS::ElcDipole;
using CPM_PHYSICS::MagDipole;
using CPM_PHYSICS::PatchCharge;
using CPM_PHYSICS::FieldSpace;
using CPM_PHYSICS::VecE;
using CPM_PHYSICS::VecBH;
using CPM_PHYSICS::VecG;
using CPM_PHYSICS::ScalarField;
using CPM_PHYSICS::VectorField;
using CPM_PHYSICS::SourceOfVectorField;
using CPM_PHYSICS::FieldE;
using CPM_PHYSICS::FieldBH;
using CPM_PHYSICS::FieldG;

using CPM_CAMERA::Camera;

using CpmRootX::R1;
using CpmRootX::Z1;
using CpmRootX::B;
using CpmRootX::RecordHandler;
using CpmProperties::SysControl;
using CpmProperties::BodyProperties;
using CpmProperties::ContactProperties;
using CpmProperties::ParticleProperties;
using CpmProperties::DumbbellData;
using CpmProperties::NuggetData;
using CpmProperties::PairData;
using CpmLinAlg::R2;
using CpmGraphics::Graph;
using CpmImaging::Color;
using CpmImaging::Palette;
using CpmImaging::INVISIBLE;
using CpmImaging::RED;
using CpmImaging::GREEN;
using CpmArrays::V;
using CpmArrays::X2;
using CpmArrays::X4;
using CpmArrays::R_Vector;
using CpmArrays::R_Matrix;
using CpmFunctions::F;
using CpmGeo::Angle;
using CpmGeo::DEG;
using CpmGeo::Iv;

// notice that we are here in namespace CpmRigidBody2 or CpmRigidBody3
// so we are introducing here two variables for each name
// e.g. CpmRigidBody2::accExp and CpmRigidBody3::accExp

//doc namespace CpmRigidBodyx{

extern bool accExp;
// accurate exponential function (from the Lie algebra of the rotation
// group to this group). If this is false a fast and generally
// sufficient approximation is used. The free rotation of asymmetrical
// ridgid bodies behaves much more stable if accExp=true

extern bool markComParMec;
   // if this is true, the function ComPar::mark(Camera...) shows
   // the axes of inertia and the vectors of angular velocity and
   // angular momentum

extern bool balance;
   // if this is true, all tensors of inertia get changed
   // to the effect that their moments of inertia are equal
   // to the mean value of the 3 original values

////////////////// class RigidObject /////////////////////////////////////

   class RigidObject{ // describes the kinematical state of a rigid body
      // describes the kinematical state of a rigid body. The kinematical
      // state (sometimes also called dynamical state), by definition,
      // does not only carry full information on a configuration ('a
      // point in configuration space') but on any of the following
      // physically equivalent concepts:
      // (i) two configurations for two points in time which are 'close
      // together'.
      // (ii) an 'infinitesimal' path in configuration space
      // (iii) a point in configuration space and a vector from the
      // local vector space (tangent space) in the sense of differential
      // geometry.

      // The configuration of a RigidObject is described by the position
      // of a distinguished point and an element of the rotation group
      // SO(3) which generates the present 'orientation of the object in
      // space' from some reference orientation. Less formally the SO(3)
      // element will be called angular position (in German:'Winkellage';
      // the German word 'Lage' nicely combines partial meanings of lay,
      // location, position, constellation to a general notion of 'the
      // characteristics of placement in space').

      // If we fix an origin in space, we may describe position in space
      // by a position vector and combine the position vector (which
      // describes a translation) and the SO(3) element to an element of
      // the Euclidean group. So that the configuration space of a
      // RigidObject may be identified with the Euclidean group. This is
      // studies in some detail in my Mathematica file
      // euclideangroup2.nb.

   protected:

// independent quantities

      Z locked_{0};
         // if !=0, freeStep(dt) will act idle, except of replacing
         // t by t+dt

      R1 r_;
         // smallest value such that a sphere of radius r arround x
         // contains all matter forming the object

      Spc x_;
         // center-of-mass position or otherwise characteristic position

      Vec v_;
         // velocity of point x

      AxVec ori_;
         // 'orientation' in the meaning of 'angular position'.
         // Global rotation which creates the present
         // angular orientation of the object from a reference
         // configuration.
         // From the definition of AxVec one sees that the components of
         // this vector closely related to what normally is called Euler
         // parameters of the rotation ('rotation matrix') under
         // consideration. The agreement would be exact if in the
         // definition of the Euler parameters one would replace the
         // sin-function by the tan-function.

      AxVec omega_;
         // angular velocity of collective object motion arround x.
         // angular velocity gives the time trend in angular position.
         // The velocity vi at position xi then is given by
         // vi=v+omega.cross(xi-x)

      R t_=R(0);
         // new element, time to which the quantities refer

// dependent quantity

      RefSys rs_;
         // The 'body-fixed' system of reference determined by x and ori

// non-public functions:

      virtual void update(){ rs_=RefSys(x_,ori_);}
         // updating the dependent quantities according to the actual
         // values of the independent ones. Needs not to be called if
         // only velocities (v or omega) were changed

      void make(R r, const Spc& x, const Vec& v,
         const AxVec& ori, const AxVec& omega);
         // constructing a RigidObject out of the independent quantities

// Now we consider transformations under Group. We need only to consider
// independent quantities, since only those have to be transformed. The
// dependent ones will be obtained via update() from the transformed
// independent quantities. This procedure guarantees the constancy of
// internal geometry independent of the number of transformations acting
// on a system. Here we divide the independent quantities into streamable
// ones i.e. those that can be written on file and retreived from files
// with sufficient accuracy and acceptable effort, and non-streamable
// ones like instances of F<X,Y>.
// In the final classes to be used, non-streamable data will always
// be dependent on streamable ones and thus will get values from calling
// update().
// In intermediate classes like Body, one needs quantities like shape and
// field in order to formulate interaction laws at a point in time where
// a definition of field and shape in terms of streamable data cannot yet
// be available if one wants to hold things sufficiently general. In
// order to make these intermediate classes, like Body, selfconsistent
// one has to define the transformation law also for non-streamable data.

      virtual void transformStreamableData(const Group& g);

      virtual void transformNonStreamableData(const Group& g){g;}
         // there are no nonstreamable data yet
    public:
      typedef RigidObject Type;
      CPM_IO_V
      CPM_NAM_V(RigidObject)

      virtual RigidObject* clone()const{ return new RigidObject(*this);}
         // this definition from allows to make
         // RigidObjects and instances of derived classes into the
         // polymorphic array components of the polymorphic array Vp<>.
         // See cpmp.h. Actually we will need Vp<ExtSys>, with ExtSys
         // derived from RigidObject as template arguments.

      RigidObject(R r=0):r_(r){}
         // Object at rest in the master system of reference

      RigidObject(R r, const Spc& x, const Vec& v, const AxVec& ori,
         const AxVec& omega){ make(r,x,v,ori,omega);}
         // constructing a RigidObject out of the independent quantities

      R get_r()const{ return r_;}
      void set_r(R rf){ r_=rf;}

      Spc get_x()const{ return x_;}
      Vec get_v()const{ return v_;}

      R get_t()const{ return t_;}
      void set_t(R tf){ t_=tf;}

      AxVec getOri()const{ return ori_;}
      AxVec getOmega()const{ return omega_;}
      RefSys getRefSys()const{ return rs_;}

      Group frmMasSys()const{ return rs_.frmMasSys();}
         // from master system
         // returns the transformation which creates rs from the master
         // system of reference

      Group toMasSys()const{ return rs_.toMasSys();} // to master system
         // returns the transformation which creates the master system of
         // reference from system rs

      R relCor(const Spc& p, Z i)const{ return rs_(p)[i];}
         // relative coordinate of p = coordinate of x in reference
         // system rs i=1,2,3 (3 gives always 0 in two dimensional case)

      Vec get_v(const Spc& p)const{ return v_+omega_.cross(p-x_);}
         // velocity at an arbitrary point p of the object

      virtual void set_v(const Vec& v){ v_=v;}
         // setting linear velocity.
         // virtual since adding data in derived classes askes for their
         // proper handling

      virtual void setOmega(const AxVec& omega){ omega_=omega;}
         // setting angular velocity.

      virtual void spin(const AxVec& omega){ omega_+=omega;}
         // adding to angular velocity

      virtual void boost(const Vec& v){ v_+=v;}
         // adding to linear velocity

      virtual void freeStep(R dt);
         // state change during time span dt under the assumption that v
         // and omega remain constant. Notice that in the absence of
         // external forces v is constant but that the absence of torques
         // does not imply that omega is constant unless the tensor of
         // inertia is a multiple of the unit tensor. Actually, we use
         // RigidObject never as dynamical objects for which inertia
         // plays a role. So there is no tensor of inertia defined for
         // RigidObject. This will change with classes PointPar and
         // ComPar. Notice that
         //    freeStep(dt)
         // is identical (appart from t+=dt, which belongs to freeStep)
         // to applying operator *=(g) for the following instance of
         // Group: g=(omega.exp(dt),x-x*omega.exp(dt)+v*dt).
         // For the sake of efficiency we do not make use of this.

      virtual void freeMot(R tf){ freeStep(tf-t_);}
         // changes the present state into the state at time tf employing
         // a mode of motion in which velocity and angular velocity are
         // constant.

      void operator*=(const Group& g)
         // has always this form, building blocks are virtual and thus
         // adjust the function so that additional data of derived
         // classes get transformed properly
      {
         transformStreamableData(g);
         transformNonStreamableData(g);
         update();
      }

      void lock_(bool fix){ locked_=fix; if (fix) v_=Vec();}
         //: lock
      void unLock_(){ locked_=0;}
         //: un-lock
      bool isLoc()const{ return locked_!=0;}
         //: is locked
         // answers the question whether *this is locked or not

      virtual void adjustOmega(const Spc& xp, const Vec& vp);
         // Adjusts the angular velocity omega such that after calling
         // the present function we have: get_v(xp)==vp. The quantities
         // x and v remain unchanged.

      R dis(const RigidObject& ro)const;
         // returns the 'the distance' of *this and ro as a fraction to
         // 2r. Thus the value is not limited by 1. The distance is a sum
         // of the distance between the centers, and the distance between
         // body-fixed peripheral points in a configuration in which the
         // centers are shifted back to coincidence.
   };

/////////////////////// struct PhyParPrp /////////////////////////////

struct PhyParPrp{ // physical particle properties
      R1 m_;
         // mass
      R1 r_;
         // an interaction radius, in collision of particles
         // with walls which are implemented by classes derived from
         // CpmPhysicsx::Body it acts as a radius of a rigid sphere
         // that starts to feel reaction forces as soon as this sphere
         // overlaps with the body which represents the wall.
         // In addition. r_ is the regularization radius of the
         // point charge and the point mass associated with the particle.
      R1 q_;
         // charge

      R1 alphaAdhesion_;
      R1 relativeRangeAdhesion_;
      R1 E_;
         // E/(1-sigma*sigma); effective Young's modulus of Hertz spheres
      R1 mu_;
      R1 v0_;
      R1 scnr_;
         // Notice that the 'coefficient of normal restitution (cnr)
         // (1 for elastic collisions, <1 for inelastic) defined as a
         // reduction factor for normal velocity, not for kinetic energy
         // (although the latter, the square of  cnr is needed in the
         // computation of forces (our only usage)

      R1 U0_,r0_,a_;
         // U0_,r0_,a_ are the three constants defining a
         // inter-particle potential (see attribute forceType)
         // for 'chemical' forces

      Z1 forceType_;
         // 0 no chemical interaction
         // 1 Lennard Jones
         // 2 Morse
         // 3 UM
         // 4 Hertz type elastic repulsion

      void set_cp(ContactProperties const& cp)
      {
         alphaAdhesion_=cp.alphaAdhesion_;
         relativeRangeAdhesion_=cp.relativeRangeAdhesion_;
         E_=cp.E_;
         scnr_=cp.scnr_;
         mu_=cp.mu_;
         v0_=cp.v0_;
      }

      ContactProperties get_cp()const
      {
         ContactProperties res;
         res.alphaAdhesion_=alphaAdhesion_;
         res.relativeRangeAdhesion_=relativeRangeAdhesion_;
         res.E_=E_;
         res.mu_=mu_;
         res.v0_=v0_;
         res.scnr_=scnr_;
         return res;
      }

      typedef PhyParPrp Type;
      CPM_IO
      PhyParPrp(){}
      PhyParPrp(R d, ParticleProperties const& pp)
      {
         set_cp(pp);
            // sets E_,mu_,v0_,scnr_ and the adhesion values
         r_=d*0.5;
         m_=pp.rho*CpmDim3::Spc::vol(r_);
         q_=pp.sigma*CpmDim3::Spc::sur(r_);
         forceType_=4;
      }
      PhyParPrp(R m, R r, R q=0):m_(m),r_(r),q_(q){}
   };

////////////////////// class PointParC //////////////////////////////////
// PointPar stands for point particle
// constant quantities associated with an instance of PointPar

struct PointParC{ // constant quantities of point particles

   PhyParPrp p_;
      // a set of material properties which is shared by all particles
      // of family part

   Z1 id_;
      // identification number, allows identification. Usefull
      // especially in parallel programs where a single particle
      // may have to be handled by various processes.

   B idValid_;
      // initialized as false, device to avoid that id can be changed
      // despite the fact that set_id is public
   Z1 color_;
      // artificial attribute for visualization
      // Color(color_) is actually a color according to
      // my Palette. Since Color(0) = Color(INVISIBLE) the default
      // particle should get an anaglyph representation
   typedef PointParC Type;
   CPM_IO
   CPM_NAM(PointParC)
   PointParC(){}
   PointParC(R m, R r, R q=0):p_(m,r,q){}
   explicit PointParC(const PhyParPrp p):p_(p){}
   PointParC(R d, ParticleProperties const& p):p_(d,p){}
   Z get_id()const{ return id_;}
   bool get_idValid()const{ return idValid_;}
   void setCol(Palette  c){color_=Z1(Color::toZ(c));}
   Color getCol()const{ return Color(Z(color_));}
};

////////////////////// class PointParV //////////////////////////////////
// variable quantities associated with an instance of PointPar

struct PointParV{ // variable quantities of point particles
   // there is no angular velocity!
   Z1 owner_;
      // needed for implementing parallelism
   Spc x_;
      // position
   Vec v_;
      // velocity
   Vec vDot_;
      // acceleration, automatically initialized as zero
      // which is what is needed
   R1 t_;
      // time for which data are valid
   B fixed_;
      // If this is true, the center of mass is held fixed in dynamics.
   typedef PointParV Type;
   CPM_IO
   CPM_NAM(PointParV)
   PointParV(){}
   explicit PointParV(Spc x, Vec v=Vec(), R t=0):x_(x),v_(v),t_(t){}
   Spc get_x()const{ return x_;}
   Vec get_v()const{ return v_;}
   bool get_fix()const{ return fixed_;}
   PointParV& ref_(Vec const& n, Spc const& s0)
      //: reflect
   {
      x_.ref_(n,s0);
      v_.ref_(n);
      vDot_.ref_(n);
      return *this;
    }
};

//////////////////////// class PointPar /////////////////////////////////

   class PointPar{ // point particles

      // Describes kinematical states (i.e. positions and velocities) of
      // point particles. Since 2006-02-02 there is the acceleration
      // added. This serves as a memory for easy implementation of my
      // new integrator which is second order even for velocity-dependent
      // forces. Not all comments may already reflect this new situation.
      // As far as force laws are concerned, these point particles have
      // a radius (action radius), but they show no inertia of rotation
      // arround their center; actually, there is no degree of freedom
      // corresponding to such a rotation.

      PointParC cc;
      PointParV cv;
         // notice: there are no dependent data members

   public:
      friend class ComPar;
      typedef PointPar Type;
      CPM_IO
      CPM_NAM(PointPar)

      PointPar(){}

     explicit PointPar(PhyParPrp const& ppp, Spc const& x=Spc(),
        Vec const& v=Vec()): cc(ppp),cv(x,v){}
      // work horse constructor

      PointPar(const PointParC& cc1, const PointParV& cv1):
         cc(cc1),cv(cv1){}
      // obvious

      PointPar(R d, const ParticleProperties& pp):
      cc(d,pp),cv(){}
         // For uniformity with ComPar; the data pp.J
         // pp.eps, pp.fRelax which deal with magnetic and
         // electric dipoles are passive here, since missing
         // rotational inertia one can't handle the torques on
         // PointPar's that would result from such dipoles.

      PointPar(R d, const ParticleProperties& pp,
         const NuggetData& pd, Z j=1): cc(d,pp),cv(){pd;j;}
      PointPar(R d, const ParticleProperties& pp,
         const DumbbellData& pd, Z j=1): cc(d,pp),cv(){pd;j;}
      PointPar(R d, const ParticleProperties& pp,
         const PairData& pd, Z j=1): cc(d,pp),cv(){pd;j;}
         // these three only for uniformity with ComPar, last argument
         // has no influence

      PointPar(RecordHandler const& rch, Word const& sec, Z j=1);
         // Flexible input which allows to use a rho input if
         // no mass input is there; and correspondingly uses
         // a sigma input if no q input is there. Due to
         // the 'universality' of the data type RecordHandler,
         // it is easy to provide a constructor with the same
         // signature also for ComPar

      R get_m()const{ return cc.p_.m_;}
      R get_r()const{ return cc.p_.r_;}
      R get_q()const{ return cc.p_.q_;}

      Z get_id()const{ return cc.id_;}

      void set_id(Z id1);// a valid id gets never changed unless
        // the id is cancelled explicitely via cancel_id

      void setOwner(Z owner){ cv.owner_=owner;}
      Z getOwner()const{ return cv.owner_;}
      bool isVal()const{ return cc.idValid_;}
      void cancel_id(){ cc.idValid_=false;}
      bool idn(PointPar const& pa)const{
         return (isVal()&&pa.isVal()&&(get_id()==pa.get_id())) ?
         true : false;}

      Spc get_x()const{ return cv.x_;}
      Vec get_v()const{ return cv.v_;}
      PhyParPrp get_p()const{ return cc.p_;}

      PointParC getC()const{ return cc;}
      PointParV getV()const{ return cv;}
      void set(const PointParC& cc1, const PointParV& cv1)
      {cc=cc1; cv=cv1;}

      void set_m(R m){cc.p_.m_=m;}
      void set_r(R r){cc.p_.r_=r;}
      void set_m_r(R m, R r, const NuggetData& pd=NuggetData())
         // last argument for uniformity with ComPar, here it is
         // passive
      {pd;cc.p_.m_=m;cc.p_.r_=r;}

      void set_q(R q){cc.p_.q_=q;}
      void rev_q(){cc.p_.q_*=-1;}
      void invertCharge(){cc.p_.q_*=-1; }
         // changing the sign of the charges
      void set_E(R E){cc.p_.E_=E;}
      void setCofOfNorRes(R cnr){cc.p_.scnr_=cnr*cnr;}

      void set_cp(ContactProperties const& cp){ cc.p_.set_cp(cp);}
      ContactProperties get_cp()const{ return cc.p_.get_cp();}

      R getEeff()const{ return cc.p_.E_;}

      void setPot(R U0, R r0, R a, R cnr, Z forceType)
      {
         cc.p_.U0_=U0; cc.p_.r0_=r0; cc.p_.a_=a;
         cc.p_.scnr_=cnr*cnr; cc.p_.forceType_=forceType;
      }

      void setForType(Z forceType){cc.p_.forceType_=forceType;}

      void set_x(const Spc& x){ cv.x_=x;}
      void set_v(const Vec& v){ cv.v_=v;}
      void set_p(const PhyParPrp& p){ cc.p_=p;}
      void setCol(Palette c){cc.setCol(c);}

      R get_t()const{ return cv.t_;}
      Color getCol()const{return cc.getCol();}
      void set_t(R tf){ cv.t_=tf;}

      PointCharge getPointCharge()const
      { return PointCharge(cc.p_.q_,cv.x_,SysControl::singRange);}

      PointMass getPointMass()const
        { return PointMass(cc.p_.m_,cv.x_,SysControl::singRange);}

      void boost(const Vec& v){ cv.v_+=v;}

      void ranBoost(Vec const& v, R dv, R angleRange, Z j)
      {
         Vec vj=v.randomize(dv,Angle(angleRange,DEG),j);
         boost(vj);
      }

      void spin(const AxVec& omega_){omega_;}
         // does nothing; for uniformity with ComPar

      PointPar& ref_(Vec const& nv, Spc const& s0)
         //: reflection
      {
         cv.ref_(nv,s0);
         return *this;
      }

      PointPar& ref_(Plane const& p)
      { return ref_(p.n(),p.o());}

      void operator *=(const Group& a){cv.x_*=a; cv.v_*=a; cv.vDot_*=a;}
         // action of the basic symmetry group on a particle. Notice that
         // x and v belong to different classes (Spc and Vec,
         // respectively) the group element a does different things to
         // them.

      void ranRot(R angleRange, Z j){angleRange;j;}
         // Random rotation arround the center of mass which is trivial
         // here.
         // Definition for uniformity with ComPar

      void operator *=(R fac);
         // scaling a point particle such that all linear dimensions
         // get multiplied by fac and mass by fac^3 (charge remains
         // constant since quantized in nature)

      R dist(const PointPar& pa)const
         { return (cv.x_-pa.cv.x_).abs();}
         // returns the distance of centers

      R dis(const PointPar& pa)const
      { return dist(pa)/(cc.p_.r_.x1+pa.cc.p_.r_.x1);}
         // returns the 'the distance' of *this and pa divided by the sum
         // of the radii of *this and pa. Thus the value is not limited by
         // 1.

      R gap(const PointPar& pa)const
      { return dist(pa)-(cc.p_.r_.x1+pa.cc.p_.r_.x1);}
         // if the spheres occupied by the particles don't overlap, this
         // is their distance in the sense of geometry (distance of
         // closest point pair) If they overlap, the return value is
         // negative and gives (appart from the sign) the overlap
         // distance. Even if a small particle occupies the interior
         // of a large particle it is not too unreasonable a measure
         // of which the repulsive force should be a increasing function.
         // Notice that the minimum of the function is -r-pa.r

      R corGap(const PointPar& pa)const{ return gap(pa);}
         // coarse gap

      X2<R,Vec> gapAndDirection(const PointPar& pa)const
         // returns gap and the unit vector pointing from *this to pa
         // (center to center)
      {
         X2<R,Vec> pp3=(pa.cv.x_-cv.x_).polDec();
         return X2<R,Vec>(pp3.c1()-cc.p_.r_-pa.cc.p_.r_, pp3.c2());
      }

       Spc frcCen(const PointPar& pa)const;
         // returns a point in which contact forces can reasonably
         // assumed to act. This point equals the point p1.cen(p2)
         // mentioned in function clsApp.

      void freeStep(R dt){
         if (!isLoc()) cv.x_+=dt*cv.v_;
         cv.t_+=dt;}
         // free motion

      void freeMot(R tf){ freeStep(tf-cv.t_);}
         // free motion. t=tf after call

      void phase1(R dt){ if (SysControl::accMem) cv.v_+=cv.vDot_*dt;}
         // named according to Section 7 in GranMat publication

      void intAct(R dt, const Force& fs,
         const FieldSpace& fval=FieldSpace()); // interaction
      // Last  argument not used

      void reg(void);
         // regularization of the motion by reducing the velocity
         // if this is larger than SysControl::vCrit
         // no action for SysControl::vCrit

      Force forces( const PointPar& pp )const;
         // returns the force acting on the particle *this
         // resulting from the presence of a particle pp.

      R pot( const PointPar& pp )const;
         // returns the potential of forces above

      Force forceContact(PointPar const& p,
         Vec const& v, Vec const& vp)const;
         // force which particle *this feels as a result of the presence
         // of p. Contact interaction forces are treated as originating
         // at the points of closest approach (see function clsApp) and
         // acting on those points. Friction forces depend on velocities,
         // for which we have to insert the velocities of the points
         // of closest approach. These velocities are given by the last
         // two arguments of the function. If among the data of PointPar
         // there would be an angular velocity, then we would express v
         // and vp in terms of PointPar-data; so we have to provide them
         // as arguments.
         // This  friction scheme is introduced here since with respect to
         // interaction, the polyspherical particles ComPar to be defined
         // later will generate their interaction forces as rigidly
         // connected PointPars. Even if such a ComPar consists of a
         // single component, it will have inertia and the freedom to
         // rotate. Then the friction forces damp relative rotation
         // of spherical particles so that ComPars behave dynamically
         // normal if they are not asymmetrical and thus may be made of
         // a single constituent. This helps to improve speed.
         // Contact force laws all are coded in function
         // forceFrictionAdhesionElastic(...)

      R potElastic(PointPar const& p)const;
         // potential of the elasic forces contributing to forceContact

      Force forces( const VectorField& f)const;
         // returns the force acting on the particle *this
         // resulting from the presence of electric, magnetic, and
         // gravitational fields.

      Force forces( const FieldSpace& f)const;
         // returns the force acting on the particle *this
         // resulting from the presence of homogeneous electric,
         // magnetic,
         // and gravitational fields.

      R pot( const FieldSpace& f)const;
         // potential energy pf particle *this resulting from the
         // presence of homogeneous electric, magnetic, and gravitational
         // fields.

      Force forces( const RigidObject& f)const;

      R kinEnr()const{ return cc.p_.m_.x1*0.5*(cv.v_|cv.v_);}
         // kinetic energy

      ScalarField getU()const
         // potential associated with the types of interaction to be
         // considered
         {
            if (SysControl::potentialToShow==3)
               return getPointMass().fieldU();
            else if (SysControl::potentialToShow==2)
               return ScalarField();
            else
               return getPointCharge().fieldU();
         }

      FieldE getE()const{ return getPointCharge().field();}
      FieldBH getBH()const{ return getPointCharge().fieldBH(cv.v_);}
      FieldG getG()const{ return getPointMass().field();}

      VectorField getField(bool noGravity=true)const
      {
         if (noGravity) return VectorField(getE(),getBH(),FieldG());
         else return VectorField(getE(),getBH(),getG());
      }

      Vec mom()const{ return cv.v_*cc.p_.m_;}
         //: momentum

      AxVec angMom(Spc const& p)const;
         // returns the total (mechanical) angular momentum
         // of *this relative to the point p in space

      void mark(Graph& g, const Color& color)const;
         // Draws a filled disk with particle's radius according to
         // the object space metric of g. Here color acts a an instance
         // of the base class rgb of Color, so that Color(INVISIBLE)
         // acts as rgb(0,0,0) which is Color(BLACK)

      void mark(Graph& g)const{ Color col=getCol(); mark(g,col);}
         // same as previous function but uses the own color for drawing

      void mark(const Camera& sc, const Color& color)const;
         // If SysControl::coloredParticles, the representation is by
         // a filled circular disk, without a velocity indicator.
         // if SysControl::howToMarkParticles==1 a wire frame
         // representation using color is drawn. A velocity indication is
         // also  given as a line startingfrom the center.

      void mark(const Camera& sc)const
         { Color col=getCol(); mark(sc,col) ;}
         // same as previous function but uses the own color for drawing

      SubsetOfSpace toSubOfSpc()const;
         //: to subset of space
         // This helps visualisation since for objects of type
         // SubsetOfSpace a Camera can create a good anaglyphic
         // representation via its function mark.
         // Notice that class Body has a member function of the same name
         // and signature.

     R sclVal(const Word& w)const;
         // scalar value
         // returns a scalar value designated by w, where
         // the following cases for w are implemented
         // "mass", "velocity", "kinetic_energy", "charge".
         // In cases where w indicates a vector quantity,
         // the actual value from this function is the
         // absolute value of the vector quantity.

      SourceOfVectorField source()const;
         // unified treatment of field sources
         // of particle *this

      void stop(R red=0, bool selective=false )
         {
            if (selective){
               cv.v_[3]*=red;
               cv.vDot_[3]*=red;
            }
            else{
               cv.v_*=red;
               cv.vDot_*=red;
            }
         }
         // brings *this to rest (state with all velocities equal to
         // zero) or reduces all velocities of *this by the factor red.

    //  void fix_(){ cv.fixed_=B(1);}
         //: fix
         // make the particle fixed at its present position

      void lock_(bool fix){ cv.fixed_=B(fix); if (fix) cv.v_=Vec();}
         //: lock
      void unLock_(){ cv.fixed_=B(false);}
         //: un-lock
      bool isLoc()const{ return cv.fixed_;}
         //: is locked
         // answers the question whether *this is locked or not

      bool isMag()const{ return false;}
         // ' is magnetic'
         // for uniformity with ComPar

      void deElectrize(){}
         // setting the electric moment to zero
         // for uniformity with ComPar

      void removeMagDipole(){}
         // setting the magnetic moment to zero
         // for uniformity with ComPar

      static B isVal(const V<PointPar>& vp);
         // returns true if there are no reasons to
         // reject data vp
         // Implementation based on calling
         // static B isVal(const V<PointParV>& vv)

      static B isVal(const V<PointParV>& vv);
         // returns true if there are no reasons to
         // reject data vv
         // Present implementation based on the velocity distribution
      static void showProperty(const Word& theme, const V<PointPar>& vp,
         const Camera& ca, const R_Vector& parameters);

      static Force forceFrictionAdhesionElastic( R gap, const Vec& n12,
         R r1, R r2, const Vec& v1, const Vec& v2,
         const ContactProperties& cp1, const ContactProperties& cp2 );
      // a new implementation for which the following description of
      // friction forces is not used. Hysteresis of normal force instead
      // and conventional friction law for tangential force.
      // The idea to use hysteresis of normal forces for modeling
      // friction was told to me by Dr. Bernhard Eberhardt at the
      // University of Tübingen in 1998.
      // n12 is the unit vector directed from 1 to 2

      static R elasticPotential( R gap, const Vec& n12,
         R r1, R r2, ContactProperties const& cp1,
         ContactProperties const& cp2 );
      // returns the potential of the elastic contribution to force
      // given by the previous function

      static R collisionTime(R r1, R r2, R m1, R m2, R v,
         const ContactProperties& cp1, const ContactProperties& cp2);
         // This is the time during which two centrally colliding elastic
         // sheres are in contact as computed first by H. Hertz 1881
         // (Source: Landau, Lifschitz: Lehrbuch der Theoretischen
         // Physik,
         // Band 7, 5. Auflage, Akademie Verlag 1983 § 9 p. 39)
         // Here v is the abs-value of relative velocity of the
         // spheres.

         // We consider two objects, distinguished by index 1 and 2.
         // Each has a mass, a velocity, a radius, and an instance of
         // ContactProperties.
         // The location of the objects is characterized (not
         // completely!) by the geometry of their 'region of closest
         // approach' ('contact region'). This is a numerical value of
         // the 'gap' (positive if there is really a gap and no overlap;
         // so -gap is 'the overlap') and a unit vector n12 from object 1
         // to object 2.
         // These quantities obviously have a well defined meaning if
         // both objects are spherical particles. In this case v1 and v2
         // are not the velocities of the particle centers but the
         // velocities of the points of closest approach (cf. comment to
         // function
         //  Force forceContact(const SphPar& p,
         //     const Vec& v, const Vec& vp)const; )
         // One of the objects is allowed to grow to a
         // half-space (radius tends to infinity). (For two half-spaces
         // the situation can only be applied if their limiting planes
         // are parallel.)
         // The return value of the function is the force that object 1
         // feels from object 2. It is a sum of three contributions:
         //    1. relative friction
         //    2. adhesion
         //    3. elastic repulsion
         // which will be described now for the most important case that
         // the two objects are of type SphPar.
   };

//////////////////////// class ComParC //////////////////////////////////
// ComPar stands for composite particle
// constant quantities associated with an instance of ComPar

struct ComParC{ // constant quantities of composite particles
   typedef ComParC Type;
   ComParC(){}
   CPM_IO
   CPM_NAM(ComParC)
   R1 r;
      // smallest value such that a sphere of radius r arround x
      // contains all particles of the cluster
      // (thus r= sup{ |xr[i]|+ part[i].r : i=1,...n}
   R1 volume;
      // the volume of the body
      // This is assumed to define also the surface of the body
      // as the surface of the corresponding sphere
   V<R> ri;
      // list of the radii of the constituent particles
   V<Vec> xr0;
      // list of the particle centers relative to center of mass in
      // the 'body-fixed frame'. These are needed in order not to
      // allow the bodies geometry vary with accumulating numerical
      // errors
   InTens I0;
      // Tensor of inertia ( with respect to x and the body fixed
      // axes)
   Vec ej0;
      // a unit vector the components of which refer to the body fixed
      // axes, which gives the direction of the magnetic moment. If
      // the magnetic moment is zero, ej0 will neverthelss be
      // initialized in every constructor

   ParticleProperties prop;
      // a set of material properties which is shared by all particles
      // of family part

   Z1 id;
      // identification number, allows identification. Usefull
      // especially in parallel programs where a single particle
      // may have to be handled by various processes.

   B idValid;
      // initialized as false, device to avoid that id can be changed
      // despite the fact that set_id is public

   bool isToner()const{ return !(prop.J>0);}
   R getVolume()const{ return volume;}
   R get_m()const{ return volume.x1*prop.rho;}
   R get_q()const
   { return cpmpow(cpmpi*volume*volume*36,R(1)/R(3))*prop.sigma;}
      // correct even if looks strange
   Z get_id()const{ return id;}
   bool get_idValid()const{ return idValid;}
   ContactProperties get_cp()const{ return (ContactProperties)prop;}

   ComParC& ref_(Vec const& n)
   {
      I0.ref_(n);
      ej0.ref_(n);
      for (Z i=xr0.b();i<=xr0.e();++i) xr0[i].ref_(n);
      return *this;
   }

   ComParC& operator *=(AxVec& av)
   {
      for (Z i=xr0.b();i<=xr0.e();++i) xr0[i]*=av;
      ej0*=av;
      I0*=av;
      return *this;
   }
};

//////////////////////// class ComParV //////////////////////////////////
// variable quantities associated with an instance of ComPar

struct ComParV{// variable quantities of composite particles
   typedef ComParV Type;
   ComParV(){}
   CPM_IO
   CPM_NAM(ComParV)

   Z1 owner;

   Spc x;
      // center-of-mass position

   Vec v;
      // velocity of center-of-mass

   Vec vDot;
      // acceleration of center-of-mass

   AxVec ori;
      // ori stands for orientation
      // global rotation which creates the present orientation of the
      // cluster relative to a reference configuration (at creation or
      // for the initial state of a dynamical process)

   AxVec omega;
      // angular Velocity of collective cluster motion arround center
      // of mass. The velocities v[i] of the individual particles
      // then are given by v[i]=v+omega.cross(xr[i])

   AxVec omgDot;
      // Time derivative of omega, angular acceleration

   ElcDipole qd;
      // The dipole is assumed to be located at x, thus qd.x=x. This
      // has the efficiency advantage, that the torque on qd directly
      // translates into a torque on *this.
      // addition 2001-4-1

   R1 t;
      // time, see RigidObject::t (addition 2000-12-12 after having
      // re-united ComPar and ComPar2

   B fixed;
      // If this is true, the center of mass is held fixed in dynamics

   Spc get_x()const{ return x;}
   Vec get_v()const{ return v;}
   bool get_fix()const{ return fixed;}

   ComParV& ref_(Vec const& n, Spc const& s0)
   {
      x.ref_(n,s0);
      v.ref_(n);
      vDot.ref_(n);
      ori.ref_(n);
      omega.ref_(n);
      omgDot.ref_(n);
         //qd.mir_(i); not yet defined
      return *this;
    }
};

//////////////////////////// class ComPar ///////////////////////////
// ComPar stands for composite particle
/*
Design consideration concerning representation of irregularly shaped
bodies as clusters of (potentially overlapping) spherical particles
which are rigidly connected

It is clear that any shape can be modelled by composing spheres (of
various sizes) if we can use sufficiently many for this purpose. Since we
have to be concerned about computational efficiency, it is important to
design the scheme such that already a few spheres make up shapes that
behave reasonably (for instance which simulate rather smooth surfaces
without deep holes which would restrict relative mobility of the
clusters too much. For this purpose (pointed out by Thomas Dera) it is
reasonable to let the spheres overlap (as we do here!).

The parameters which describe the physical properties of the surface
(e.g. Hamaker constant) or bulk interaction properties like elastic
constants and constants of dielectricity, are chosen here to have the
same values for all particles belonging to one cluster. This is mainly
in order not to have too many data associated with a cluster.
If this concern should turn out to be only a minor one, one should
integrate all physical parameters in the description of PointPar.

Also for reasons of efficiency, I decided to represent the electrical and
magnetic structure of a ComPar by a single point charge and a single
magnetic dipole. It would be not difficult, however, to endow each
constituent particle with its own point charge and magnetic moment. This
would allow to model toner particles with non-uniform charge distribution
as discussed for instance in
   James Q. Feng and Dan A. Hays: Theory of Electric Field Detachment of
   Charged Toner Particles in Electrophotography
   Journal of Imaging Science and Technology, Vol. 44/#1 January/February
   2000 p. 19-25

2000-6-15: Now SysControl::singleCharge controls whether a single charge
for a ComPar is used or whether each constituent particle has its own
charge.
*/

   class ComPar{ // composite particles
      // could be derived from RigidObject. Actually it is not
      // in order to avoid any degradation of efficiency.
      // The class implements rigid body dynamic via a propriatory(?)
      // version of the explicit midpoint method. The new part is the
      // parallelism in treating 'spatial position' and 'angular
      // position'. We associate (x,v) with space position and
      // (ori,omega) with angular position.
      // The unifying point of view is that x and ori belong to Lie
      // groups and v and omega belong to the Lie algebras of the
      // corresponding Lie groups. (Clearly, we may unify the two groups
      // under consideration into a single group, the Euclidean group).
      // Operations, which in mechanics textbooks are done in real-valued
      // 'generalized coordinates' are done here directly in the
      // operations defined for Lie groups and Lie algebras. This adds
      // both to clarity of structure and efficiency of the numerical
      // implementation. Some of the concepts involved are explained in
      // class RigidObject.
      // Since 2006-02-02 there are the rates of change for velocity and
      // angular velocity  added. These serve as a memory for easy
      // implementation of my new integrator which is second order even
      // for velocity-dependent forces. Not all comments may already
      // reflect this new situation.

// independent data (only those have to be written on file,
// dependent ones will be reconstructed by the function updateInternals.

      ComParC cc;
      // constant data during program run

      ComParV cv;
      // data that change with state

// dependent data; notice that we do not use one of the dependent data
// members of RigidObject: RefSys rs. This is for saving memory space
// since we typically will have to handle many instances of type ComPar.

      Z1 n;
         // number of particles in the cluster

      R1 m;
         // total mass of the cluster

      R1 mInv;
         // inverse of m

      V<Vec> xr;
         // list of the particle positions relative to center of mass

      InTens I;
         // Tensor of inertia ( with respect to x and the system of
         // master axes)
      InTens Iinv;
         // inverse of I
      InTens Iinv0;
         // inverse of I0
      V<PointPar> part;
         // array of particles i with x[i], v[i], not referring to the
         // center of mass but to reference space (master space).
         // Data for masses come from here, values for position and
         // velocity are derived from x, v, omega, and xr[i]
      PointCharge q;
         // The charge is assumed to be located at x, thus q.x=x. Having
         // a position associated with q itself is redundant in the
         // present setting. If one wants to incorporate non-trivial
         // charge distributions, one would have to have an array of
         // charge points ( i.e. V<PointCharge> ) in close analogy
         // to array part.

      MagDipole j;
         // The dipole is assumed to be located at x, thus j.x=x. This
         // has the efficiency advantage, that the torque on j directly
         // translates into a torque on *this.

      B isMag_;
         // true if the magnetic moment is different from zero

      R1 alpha;
         // polarizability

// functions

      void make(R r, const V<Vec>& xRel, const V<R>& radii,
         const ParticleProperties& p, const Vec& ej, const Vec& v_=Vec(),
         const AxVec& omega_=AxVec());
         // This function may be called for any instance of ComPar
         // (e.g. the default one) and will initialize explicitely all
         // data members.
         // The array of xRel gives the positions of the centers of solid
         // spheres relative to the origin of the master system of
         // reference.
         // The array radii gives the corresponding radii of these
         // spheres.
         // The cluster to be formed is the union of these spheres.
         // This quantities are understood as referring to an arbitrary
         // common unit of length. The geometry given such up to a
         // scaling transformation will be scaled such that the volume of
         // the set union of these spheres (which, by definition, is the
         // volume of the cluster, is the volume of a sphere of  radius r.
         // This volume will be filled with matter such that the density
         // is p.rho.
         // Volume, center-of-mass, tensor of inertia of this all will be
         // found by stochastic integration.
         // The surface of the cluster comes into play in order to allow
         // the prescription of its total charge in the form
         // surface*p.sigma. Since the surface charge built up by
         // tribo-charging is far from a uniform distribution, it is not
         // clear that the exact geometrical surface (which would not be
         // easy to calculate for the set union of overlapping spheres)
         // is the best quantity to enter the formula given above.
         // To avoid complications of doubtful return we simply define
         // the 'active surface' of the cluster as 4*pi*r^2, the
         // surface of a sphere that has the same volume as the cluster.
         // The (Coulomb) magnetic moment (SI unit is Tesla) of the
         // cluster is prescribed as volume*J as absolute value and the
         // direction of the magnetic moment is given by ej.
         // Finally we shift the whole object such that the center of
         // mass coincides with the origin and set the center-of-mass
         // velocity equal to v and the angular velocity equal to omega.

       void updateInternals();
         // x, v, ori, omega are assumed to have valid values.
         // All quanties (except those that are constant anyway) get
         // changed to their actual value. The non-constant dependent
         // quantities get re-built from body-frame data (name ending in
         // '0') so that the geometry remains unaffected by numerical
         // errors resulting from composing many rotations.
         // This turned out to be very important and was the motivation
         // for introducing the body-frame data in 99-10-2.
         // Since this operation is called extremely often, it should be
         // very efficient. Therefore, in arrays of quantities (which
         // always are fixed in number) we only change the components
         // and don't assign the whole array. This causes some
         // code doublication in the next function.

       void createAllDependentData();
         // This assigns to all dependent data the proper value
         // (assuming that all independent data have proper value) not
         // making any assumption on the present values of these
         // dependent data. In particular, vectors get assigned and not
         // only their components.

public:
   typedef ComPar Type;
   virtual ComPar* clone()const{ return new ComPar(*this);}
   CPM_IO_V
   CPM_ORDER
   CPM_NAM(ComPar)
   ComPar(){}
      // does nothing useful

   ComPar(RecordHandler const& rch, Word const& sec, Z j=1);
      // Flexible input which allows to define a ComPar
      // from the same rch and the same sec for which also
      // PointPar(rch,sec) makes sense.
      // Letting j vary produces different instances based on different
      // random decisions

   ComPar(R d, const ParticleProperties& pp, const NuggetData& pd,
      Z j=1);
   // stochastically shaped cluster of pd.np (or less in exceptional
   // cases) constituents, which are spherical particles of type
   // PointPar.
   // The characteristics of this cluster are as follows:
   // 1. All constituents have the same radius
   // 2. The volume v of the cluster, which by definition is the volume
   //    of the set-union of the constituents, equals (4/3)*Pi*r^3 where
   //    r:=d/2.
   // 3. The centers of all constituents can be represented by instances
   //    of Spc and thus lie on a plane in the two-dimensional version
   //    of the program (i.e. for using namespace CpmDim2).
   //    Moreover, the centers of the constituents lie on a circle in the
   //    2-dimensional version(=:2D) and on the surface of a sphere in
   //    the 3-dimensional version (=:3D).
   //    Let the radius of this sphere/circle be r0. For np points
   //    populating such a manifold 'rather uniformly' a natural distance
   //    to the next neighbor is 3*r0/np in 2D and 2*r0/sqrt(np) in 3D.
   //    Let us denote this value by disMean. The factors 3 and 2 in this
   //    definition are a bit arbitrary and also sub-optimal. I did not
   //    modify them for compatibility with old files.
   // 4. The distance between the centers of any two constituents is
   //    greater or equal to isotropyFac*disMean. If isotropyFac is
   //    sufficiently large, it becomes impossible to place np points on
   //    a sphere/circle in such a manner that this distance requirement
   //    is met. If the trialMax positions from the random generator were
   //    tested, there will no new ones and the cluster will formed out
   //    of the valid points placed so far. These then will be less than
   //    np. If isotropyFac is considerably lower than 1, the constituents
   //    are allowed to populate their available space rather
   //    non-uniformely, resulting in a strongly non-spherical shape of
   //    the cluster.
   // 5. The radius of each constituent is overlapFac*disMean.
   //    overlapFac should be >=1 for getting a cluster that has no
   //    holes it its surface.
   // 6. The mass of the cluster is v*pp.rho, the magnetic moment v*pp.J
   //    (see 2. for v).
   // 7. The charge of the cluster is (4*Pi*r^2)*pp.sigma
   //    (see 2. for r).
   // 8. The center-of-mass position of the cluster coincides with the
   //    origin of the master system of reference.
   // 9. All constituent particles have the same ContactProperties
   //    attribute cp=(ContactProperties)pp
   //
   // The volume and the tensor of inertia are calculated rather
   // efficiently and accurately by stochastic integration ('Monte Carlo
   // Method'). For the true surface of a set-union of spheres no
   // efficient algorithm is at hand, so I use the surface of a sphere
   // as a surrogate. See the comment to function ComPar::make(...) for
   // details.

   ComPar(R d, const ParticleProperties& pp, const DumbbellData& pd,
      Z j=1);
    // Regular elongated object made of three spheres.
    // d as in the previous function.
    // Here we have three spheres the centers of which lie on a straight
    // line in an equi-spaced manner. Let D0 be the distance from the
    // center of the central sphere to any of the two spheres in
    // end-position. Let Rc be the radius of the central sphere. The two
    // end-spheres have the same radius Re.
    // Then the meaning of the arguments rc and re is rc=Rc/D0 and
    // re=Re/D0. In order for the spheres to overlap we need rc+re>1.
    // For re>rc we have a dumbbell and for re<rc we have a kind of a
    // base ball.
    // If the particles are treated as magnetic, the magnetic moment
    // forms an angle phi (in rad) with the axis of rotational symmetry
    // of the dumbbell.
    // Dumbbells were proposed 2000-5-5 by Mike Regan as a model for
    // experimental Kodak carrier material.

   ComPar(R d, const ParticleProperties& pp, const PairData& pd, Z j=1);
    // Describes elongated objects made of two spheres.
    // Rather trivial spin-off from Dumbbell, might be useful as the
    // simplest non-spherical particle.

    static X2<V<ComPar>,R> makePair(R dC, R dT, Z npC, Z npT,
            ParticleProperties ppC, ParticleProperties ppT,
            R overlapFacC=2, R overlapFacT=2,
            R isotropyFacC=1, R isotropyFacT=1,
            R facSafty=1.05, const Vec& dir=Vec(0,1));
         // This function does nearly the same than the previous one.
         // It creates only one carrier c and one toner t, determines the
         // charges from the surface of the particles and ppC.sigma and
         // ppT.sigma, and allows to set the direction of the vector
         // which points from the carrier center to the toner center
         // (center = center-of-mass).

      V<PointPar> getConstituents()const{ return part;}
         // getting the family of constituent particles in order to let
         // external systems have all the data they need for interaction
         // Notice that the constituents carry the same ContactProperties
         // as the whole object.

      ForTor forces( const ComPar& cl )const;
         // returns the force and torque acting on the cluster *this
         // resulting from the presence of a cluster cl. Notice that
         // forces on the constituent particles are the primary mediators
         // of interaction and that the combination to a net force and a
         // torque is completely within the responsibility of class
         // ComPar.
         // If get_id()==cl.get_id() the returned value is ForTor().
         // So in summing over particle pairs it is not necessary to
         // exclude the case i=j in most contexts.
         // In the present version, only the contact forces are mediated
         // by the constituent particles, wheras magnetic and electric
         // forces originate from and are felt by pointcharges,
         // electric moments, and magnetic moments that are located in
         // the center-of-mass of the cluster. The electric moments were
         // added in 2001-4-1 and are set proportional to a polarizabilty
         // and an external electric field, where the polarizability
         // is proportional to the volume of the ComPar.(see
         // void ComPar::createAllDependentData()):
         //    R epsr=prop.eps;
         //    alpha=volume*3*CpmPhysics::eps0_PHYS*(epsr-1)/(epsr+2);
         // In the time stepping algorithm, we separate the formation
         // of an induced electric moment and the generation of forces
         // (see ForTor ComPar::forces( const ComPar& cl )const) in a
         // natural manner. Notice also that
         //
         //    FieldE getE()const{ return q.field()+ej.field();}
         //
         // lets the electric moment contribute to field generation.
         // Since May 2000 there is an option (SysControl::singleCharge=0)
         // to replace the single point charge mentioned above by point
         // charges in the centers of the constituent particles.
         // The contact part of the force (i.e. friction, adhesion,
         // elastic repulsion) is implemented such that contact
         // forces originate from and act on the points of closest
         // approach between constituent particles and not the centers
         // of the constituent particles (as was the case in early
         // versions). This makes clusters consisting of a single
         // particle behave similar to larger clusters.
         // If also the friction forces are treated as acting directly
         // on the particle center, they are not able to prevent
         // densely packed one-particle clusters from spinning with
         // in arbitrary angular velocity (varying from particle to
         // particle). If the forces act on the periphery of the
         // particles only isolateted pairs can practice 'contact
         // spinning' and only if they arrange their angular velocities
         // opposit to each other. This more complex mode of motion is
         // much less likely to become excited by numerical noise.

      R pot( const ComPar& cl )const;
         // returns the potential of the electric and gravitational
         // forces for the simple model that all charge and all mass is
         // collected in the point cv.x (center-of-mass). Since March
         // 2005 there is added also the potential energy of the elastic
         // forces.  Notice that the definition of the dynamics makes no
         // use of pot. The only reason for having pot is to demonstrate
         // conservation of energy.

      R potElastic( const ComPar& cl )const;
         // returns potential energy of the elastic forces.

      ForTor forces( const VectorField& f)const;
         // returns the force and torque acting on the cluster *this
         // resulting from the presence of electric, magnetic, and
         // gravitational fields.

      ForTor forces( const FieldSpace& f)const;
         // returns the force and torque acting on the cluster *this
         // resulting from the presence of homogeneous electric,
         // magnetic, and gravitational fields.

      R pot( const FieldSpace& f)const;
         // potential energy pf particle *this resulting from the
         // presence of homogeneous electric, magnetic, and gravitational
         // fields.

      ForTor forces( const RigidObject& ro)const;
         // returns the force and torque acting on the cluster *this
         // resulting from the presence of a resistive medium such as air
         // which is at rest relative to ro. Such a resistive force may
         // either be used to model damping by motion in air or it may be
         // used as an artificial means to avoid numerical problems in
         // case that too large time steps led to too violent motion.
         // The present implementation is as follows: The damping medium
         // has dynamical viscosity  SysControl::eta and exerts damping
         // force and torque according to Stokes damping force formula
         // for a sphere of radius r: -6*pi*eta*v*r and damping torque
         // -8*pi*eta*r^3*omega. See L.D.Landau, E.M.Lifschitz: Lehrbuch
         // der Theoretischen Physik, Band VI, Hydrodynamik, S. 75
         // (20,17) and S. 77.

// time stepping and components for the explicit midpoint method

      void freeStep(R dt);
         // Integrator for free motion. Does not assume constant
         // angular velocity as RigidObject did

      void phase1(R dt)
      {
         if (SysControl::accMem){
            if (!isLoc()) cv.v+=cv.vDot*dt;
            cv.omega+=cv.omgDot*dt;
         } // 2010-08-24: curly brackets were missing, so the last
         // part was done always, so only the version with
         // SysControl::accMem==true was correct. Only this was used
         // in my published work.
      }

      void intAct(R dt, ForTor const& fs,
         FieldSpace const& fval=FieldSpace());
         // change of v and omega and electric moment during dt.
         // fs is the whole force-torque on the particle, including
         // contributions from electric moments. fval should be
         // chosen as the present value of the outside field at point x.
         // It is needed to update the value of the electric dipole
         // moment qd. Presently only the electric component of fval
         // matters since the magnetic moment is treated as constant.
         // (ideal permanent magnets).
         // Contains what was called Phase 3 in Section 7 of the GranMat
         // publication.

      R get_t()const{ return cv.t;}
      void set_t(R tf){ cv.t=tf;}

      void freeMot(R tf){ freeStep(tf-cv.t);}
      //: free motion
      // t=tf after call

// transformations

      void operator *=(const Group& g);
         // active transformation acting on a ComPar

      ComPar& ref_(Vec const& n, Spc const& s0);
         //: reflection

      ComPar& ref_(Plane const& p){ return ref_(p.n(),p.o());}

      void ranRot(R angleRange, Z j);
         //: random rotation
         // Rotates *this arround the center of mass by a rotation angle
         // in [0,angleRange) (angleRange in degrees) and arbitrary
         // rotation axis. Angle and axis depend randomly on j.

      void operator *=(R fac);
         // scaling a cluster such that rho, sigma, J (see constructor)
         // have constant values and linear dimensions are magnified by
         // fac;

      void boost(const Vec& v)
      { cv.v+=v; for (Z i=1;i<=n;i++) part[i].boost(v);}
         // active transformation which has on data the same influence as
         // changing the sytem of reference into a uniformely moving one

      void ranBoost(Vec const& v, R dv, R angleRange, Z j)
      {
         Vec vj=v.randomize(dv,Angle(angleRange,DEG),j);
         boost(vj);
      }

      void spin(const AxVec& omega_);
         // add opmega_ to the angular velocity omega and adjust particle
         // velocities accordingly

      void stop(R red=0, bool selective=false )
         {
            if (selective){
               cv.v[3]*=red;
               cv.vDot[3]*=red;
            }
            else{ cv.v*=red;cv.omega*=red;cv.vDot*=red;cv.omgDot*=red;}
            updateInternals();
         }
         // brings *this to rest (state with all velocities equal to
         // zero) or reduces all velocities of *this by the factor red.


// Make the center of mass of the particle fixed at its present
// position and set linear velocity to zero.  Rotational degrees of
// freedom will not be obstructed. Notice that in a system of locked
// particles which interact, say, by dipole-dipole interaction, the total
// angular momentum is not conserved (since the fixation system will feel
// a torque to which it is assumed not to react mechanically). The
// three functions control locking (our name for this fixation option).
// Locking was introduced 2010-08-30 - 2010-09-02.

      void lock_(bool fix){ cv.fixed=B(fix); if (fix) cv.v=Vec();}
         //: lock
      void unLock_(){ cv.fixed=B(false);}
         //: un-lock
      bool isLoc()const{ return cv.fixed;}
         //: is locked
         // answers the question whether *this is locked or not

      void reg(void);
         // see PointPar::reg()

      bool isMag()const{ return isMag_;}

      void deElectrize(){ cv.qd.setMoment(Vec());}
         // setting the electric moment to zero

      void removeMagDipole(){ cc.prop.J=0.; isMag_=false; j=MagDipole();}
         // setting the magnetic moment to zero. It is important to make
         // the change in the 'genes' (cc in this case) too, so that
         // the function createAllDependentData() will not set isMag_
         // back to true.

      void reflect(const Plane& pl)
      {cv.x=pl.reflect(cv.x); cv.v=pl.reflect(cv.v);updateInternals();}
         // reflection at plane. Notice that reflection acts trivially on
         // axial vectors ori and omega

      void invertCharge(){ q.invertCharge();}
         // changing the sign of the charges

      void set_cp(const ContactProperties&);
         // setting the ContactProperties subset of the
         // ParticleProperties

      void set_pp(const ParticleProperties&);
         // setting the ContactProperties subset of the
         // ParticleProperties

      void set_m_r(R m, R r, const NuggetData& pd=NuggetData());
         // Changes *this into a ComPar with given mass, geometrical
         // radius (not radius of a sphere of equal volume!), and shape.
         // for uniformity with PointPar, this allows to set also the
         // shape via the fourth argument (in class PointPar this
         // argument is passive).

      void set_E(R E);
         // setting effective E

      void setCofOfNorRes(R cnr);//{ cc.prop.scnr_=cnr*cnr;}
         // setting the coefficient of normal restitution

      void setCol(Palette c){cc.prop.color=c;}
         // setting the visualization color

      void setForType(Z i){i;}
         // presently there are no different force types for class ComPar
         // so this is trivially implemented

      Color getCol()const{return Color(cc.prop.color);}
         // getting the visualization color

      void set_x(const Spc& sp){ operator*=(Group(sp-get_x()));}

      void setDirJ(const Vec& v);
         // rotates the particle *this such that the magnetic
         // dipole moment and v are parallel

      Vec getDirJ()const{ return j.get_j();}
         // returns the magnetic dipole moment as a vector

      ParticleProperties get_pp()const{ return cc.prop;}
         // since ParticleProperties can be get and set, the internal
         // structure of ParticleProperties needs not tho be known
         // to the class ComPar

      bool multiplyParameter(Word const& w, R fac);
         // multiplies a parameter indicated by argument w by factor fac.
         // More precisely:
         // If w does not belong to the list:
         // "charge", "sigma, "alphaAdhesion", "relativeRangeAdhesion",
         // "effectiveE", "scnr", "J", "rho"
         // we do nothing except of returning 'false'.
         // If w belongs to the previous list, we multipliy the parameter
         // denoted w by factor fac and adjust all dependent parameters
         // accordingly.
         // For fac==1, this is done very efficiently by doing nothing.

      R kinEnr()const
         //: kinetic energy
      {
         R res=(cv.omega|I*cv.omega);
         if (!isLoc()) res+=m.x1*(cv.v|cv.v);
         return res*0.5;
      }

      R totEnr(const VecBH& H)const{ return kinEnr()+j.potEnr(H);}
         // total energy of *this in a homogeneous magnetic field of
         // strength H

      Vec elcMom(const Spc& x0=Spc())const{ return (cv.x-x0)*get_q();}
         // returns the moment of the electric charge relative to point
         // x0; Summing up these values of a set of ComPars of zero net
         // charge gives a result which is independent of x0 and is the
         // electric dipol moment of the charges

      MagDipole get_j()const{ return j;}
      ElcDipole get_qd()const{ return cv.qd;}
      void setElcDipole(R dt, const VecE& E);
         // assuming an externally created electric field strength
         // at x, we set the electric dipole moment qd essentially as
         // qdSat=polarizability of the cluster times E.
         // Actually the present value of qd is shifted towards qdSat to
         // a degree which is complete only if dt is large compared
         // to the dielectric relaxation time tRelax=1./(prop.fRelax)
      PointCharge getPointCharge()const{ return q;}
      PointMass getPointMass()const
         // the corresponding mass point, regularization radius taken
         // from point charge
         { return PointMass(m,cv.x,q.get_r());}

      SourceOfVectorField source()const;
         // unified treatment of electromagnetic field sources
         // of particle *this

      R get_q()const{return q.get_q();}
         // getting the value of the charge
      void set_q(R q_){ q.set_q(q_);}
         // setting the value of the charge
      const Spc& get_x()const{ return cv.x;}
      Vec get_v()const{ return cv.v;}

      Vec mom()const{ return isLoc() ? Vec() : cv.v*m;}
         // linear momentum

      AxVec angMom(Spc const& p)const;
         // returns the total (mechanical) angular momentum
         // of *this relative to the point p in space

      R angMomCom(Spc const& p, Vec const& dir)const;
         //: angular momentum component
         // In 3D: Component of angMom(p) in the direction defined by the
         // unit-vector dir.
         // In 2D: Indepenently of dir we return the single component
         // of angMom(p).
         // This function allows to use angular momentum also in contexts
         // that don't know class AxVec (as in cpmpardynsys.h).

      Vec get_v(Spc const& p)const
      {  return isLoc() ? cv.omega.cross(p-cv.x) :
         cv.v+cv.omega.cross(p-cv.x);}
         // velocity at an arbitrary point p of the object

      R get_r()const{ return cc.r;}
      R getVolume()const{ return cc.volume;}
      R getSurface()const
         { return cpmpow(cc.volume.x1*cc.volume*36*cpmpi,R(1)/3);}
         // this is the surface of a sphere that has the same volume
         // as *this
      R get_m()const{ return m;}
      R getAlpha()const{ return alpha;}
      R get_sigma()const{ return get_q()/getSurface(); }
      R getEeff()const{ return cc.prop.E_;}
      ContactProperties get_cp()const{ return cc.get_cp();}
      R sclVal(const Word& w)const;
         // scalar value
         // returns a scalar value designated by w, where
         // the following cases for w are implemented
         // "mass", "velocity", "angular_velocity",
         // "kinetic_energy", "charge", "electric_dipole",
         // "magnetic_dipole", "volume", "surface"
         // In cases where w indicates a vector quantity,
         // the actual value from this function is the
         // absolute value of the vector quantity.

      Z get_id()const{ return cc.id;}

      void set_id(Z id1);// a valid id gets never changed unless
        // the id is cancelled explicitely via cancel_id

      void setOwner(Z owner1){ cv.owner=owner1;}
      Z getOwner()const{ return cv.owner;}
      bool isVal()const{ return cc.idValid;}
      void cancel_id(){ cc.idValid=false;}

      bool idn(ComPar const& cp)const{ // 'identical'
         return (isVal()&&cp.isVal()&&(get_id()==cp.get_id())) ?
         true : false;}

      ScalarField getU()const
         // potential associated with the types of interaction to be
         // considered
         {
            if (SysControl::potentialToShow==3)
               return PointMass(m,cv.x).fieldU();
            else if (SysControl::potentialToShow==2)
               return j.fieldU();
            else
               return q.fieldU()+cv.qd.fieldU();
         }
      FieldE getE()const{ return q.field()+cv.qd.field();}
      FieldBH getBH()const{ return j.field();}
      FieldG getG()const{ return PointMass(m,cv.x).field();}
         // gravitational field generated by the cluster.
      VectorField getField(bool noGravity=true)const
      {
         if (noGravity) return VectorField(getE(),getBH(),FieldG());
         else return VectorField(getE(),getBH(),getG());
      }
         // returns the electric, magnetic, gravitational field as a
         // whole. If the noGravity argument is set to true, the
         // gravitational field is not calculated and thus is left in its
         // default value zero. So no superfluous computational burden
         // results for applications on electromechanics of small
         // particles

      R dis(const ComPar& cl)const;
         // returns the 'the distance' of *this and cl as a fraction of
         // r+cl.r. Thus the value is not limited by 1.

      R gap(const RigidObject& ro, Z i)const;
         // returns the clear gap between *this and the j-k-plane
         // of ro.rs where {j,k}={1,2,3}\{i}

      R gap(ComPar const& cl)const;
         // returns the clear gap between *this and cl.
         // Is negative in the case of overlap. If *this and cl
         // are sufficiently far away that even the bounding spheres
         // don't overlap, we simply return the clear gap between
         // the bounding sheres (very efficient and sufficiently
         // accurate).

      R corGap(ComPar const& cp)const // coarse gap
      { return
         (get_x()-cp.get_x()).abs()-(get_r()+cp.get_r());
      }

      void mark(Graph& )const;
         // marks the position (and orientation) on screen.
         // Invokes SphPar::mark(Graph& ) for all constituent particles.
         // Not useful in 3 dimensions.

      void mark(const Camera& ca, const Color& col)const;
         // Visualization via a Camera. Presently the 2-dimensional
         // version is created from the 3-dimensional one by simple
         // translation without exploiting the obvious simpifications
         // occurring in this case. In the three-dimensional case
         // only lines (as opposed to solid areas) give a good
         // representation in anaglyphic mode, thus the representation is
         // by a system of lines.
         // Lines crossing at right angle mark the center of a
         // constituent spherical particle. The point where several lines
         // meet is the center-of-mass.
         // A line ending in a 'blob' gives the direction of a magnetic
         // moment. All lines not ending in a 'blob' end on the surface
         // of the ComPar to be represented.
         // If col is not invisible, it will be used instead the
         // color of the particle.
         // If the name-space global boolean variable markComParMec is
         // true not the shape is indicated but the axes of inertia and
         // the vectors of angular velocity and angular momentum.
         // If dt is not 0, the rectilinear path that the particle
         // would go during dt based its present velocity will also be
         // indicated by a line starting at the center of mass.

      void mark(const Camera& sc)const{Color col=getCol(); mark(sc,col);}
         // same as previous function but uses the own color for drawing

      SubsetOfSpace toSubOfSpc()const;
         //: to subset of space
         // This helps visualisation since for objects of type
         // SubsetOfSpace a Camera can create a good anaglyphic
         // representation via its function mark.
         // Notice that class Body has a member function of the same name
         // and signature.

// new economic storage method
      ComParC getC()const;
      ComParV getV()const;
      void set(const ComParC&, const ComParV&);
      ComPar(const ComParC&, const ComParV&);

// methods that work on arrays of ComPars
      static V<ComPar> neutralize(const V<ComPar>& vc);
         // creates from an array of particles one that hat total charge
         // 0 by multiplying all positive charges by one factor and all
         // negative charges by the inverse of this factor.

      static VectorField farField(const V<ComPar>& vc, const V<Spc>& vx);
         // returns a vector field that is made by replacing the
         // field created by particles vc by a a sum of simple multipole
         // fields with poles located at vx[i], i=1,...vx.dim().
         // Naturally all particles located in the Voronoy neighborhood
         // of vx[i] are used to built the multipole at vx[i].

      static VectorField meanField(const V<ComPar>& vc, const V<Spc>& vx);
         // The previous field, as the name says, is not intended to be
         // represent something useful when evaluated at points which
         // are close to any of the points vx. The present construction
         // aims at filling this gap and to define the field generated by
         // the particles in the part of space where the particles are
         // located. A verbal description of the method tends to be less
         // readable than the transparent code in
         // cpmextendedparticlescpp.h

      static B isVal(const V<ComParV>& vv);
         // returns true if there are no reasons to reject data vv
         // Present implementation based on the velocity distribution

      static B isVal(const V<ComPar>& vc);
         // returns true if there are no reasons to reject data vc
         // Implementation based on calling
         // static B isVal(const V<ComParV>& vv)

      static void showVelocities(const V<ComPar>& vp, const Camera& ca);
      static void showProperty(const Word& theme, const V<ComPar>& vp,
         const Camera& ca, const R_Vector& parameters);
   };
