//? cpmforcescpp.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


using namespace CpmLinAlg;
using namespace CpmGraphics;
using namespace CpmRoot;
using namespace CpmFunctions;
using namespace CpmArrays;
using namespace CpmPhysics;
   // dimension-independent part
using namespace CPM_DIM;

using CPM_DIM::dimVec;
using CPM_DIM::Rdim_;
using CPM_DIM::Skeleton;

   using CpmProperties::SysControl;
   using CpmImaging::Color;
   using CpmImaging::Image;
   using CpmImaging::Palette;
   using CpmImaging::INVISIBLE;
   using CpmImaging::YELLOW;
   using CPM_CAMERA::Camera;
   using CpmProperties::transitionFunction;

namespace{
   const R mu=CpmPhysics::my0_PHYS;
   const R muInv=R(1)/mu;
   const R kel=CpmPhysics::influence_PHYS;
   const R kmag=CpmPhysics::mag_influence_PHYS;
  // const R kmagX=mu*0.25*cpmpiInv; // introduced 2010-08-30
      // since we work with Coulomb magnetic moments the standard
      // factor is kmagX and not kmag.
   const R kgrav=-CpmPhysics::G_PHYS;
   inline void regularize(R& x, R r){ x=cpmsqrt(x*x+r*r);}
      // regularization device
}

////////////////////// class VecBH /////////////////////////////////////

void VecBH::ini(R3 const& a, Word const& ind)
{
   Z mL=3;
   Word loc("VecBH::ini(R3,Word)");
   CPM_MA
   if (Spc::dimension()==3){
      // both av_ and v_ have 3 components
      if (ind=="H"){ // v_ initialized correctly
         // av_ needs definition
         v_[1]=a[1];
         av_[1]=a[1]*mu;

         v_[2]=a[2];
         av_[2]=a[2]*mu;

         v_[3]=a[3];
         av_[3]=a[3]*mu;
      }
      else if (ind=="B"){
         av_[1]=a[1];
         v_[1]=a[1]*muInv;

         av_[2]=a[2];
         v_[2]=a[2]*muInv;

         av_[3]=a[3];
         v_[3]=a[3]*muInv;
      }
      else{
         cpmerror(loc&": indicator should be B or H not "&ind);
      }
   }
   else{ // then Spc::dimension()==2 and av_ has 1 component
      // and v_ two. av_ wil be initialized by a[3]
      // v_ by a[1], a[2]
      if (ind=="H"){
         v_[1]=a[1];
         v_[2]=a[2];
         av_[3]=a[3]*mu;
      }
      else if (ind=="B"){
         v_[1]=a[1]*muInv;
         v_[2]=a[2]*muInv;
         av_[3]=a[3];
      }
      else{
         cpmerror(loc&": indicator should be B or H not "&ind);
      }
   }
   CPM_MZ
}

VecBH::VecBH(const Vec& v, Word const& ind)
{
   R3 a; // all components 0, in particular a[3]
   a[1]=v[1];
   a[2]=v[2];
   if (Spc::dimension()==3) a[3]=v[3];
   ini(a,ind);
}

bool VecBH::prnOn(ostream& str)const
{
   cpmwt("VecBH");
   cpmp(av_);
   cpmp(v_);
   return true;
}

bool VecBH::scanFrom(istream& str)
{
   cpms(av_);
   cpms(v_);
   return true;
}

///////////////// class ForTor /////////////////////////////////////////

Z ForTor::com(const ForTor& a)const
{
   if (f<a.f) return 1;
   if (f>a.f) return -1;
   if (t<a.t) return 1;
   if (t>a.t) return -1;
   return 0;
}

ForTor::ForTor(const V< X2<Force,Spc> >& vx, Spc const& sp):f(),t()
{
   Z i,n=vx.dim();
   for (i=1;i<=n;i++){
      Vec vi=vx[i].c2()-sp;
      operator +=(ForTor(vx[i].c1(),vi));
   }
}

///////  class PointCharge //////////////////////////////////////////

bool PointCharge::prnOn( ostream& str )const
{
   CpmRoot::writeTitle("PointCharge",str);
   cpmp(q);
   cpmp(r);
   cpmp(x);
   return true;
}

bool PointCharge::scanFrom( istream& str )
{
   cpms(q);
   cpms(r);
   cpms(x);
   return true;
}

PointCharge PointCharge::mirror(const Plane& pl)const
{
   return PointCharge(-q,pl.reflect(x),r);
}

PointCharge PointCharge::mirror(const Plane& pl, R eps)const
{
   R qm=q*(R(1)-eps)/(R(1)+eps); // Jackson: Classical Electrodynamics,
      // Wiley 1962, p. 112, (4.51)
   return PointCharge(qm,pl.reflect(x),r);
}

Force PointCharge::force(VecBH const& BH, Vec const& v)const
{
   return Force(BH.cross(v)*(-q));
}

Force PointCharge::force( const PointCharge& pc )const
// this is the exact gradient of the following potential
// so that energy conservation holds for a system with forces
// and potential energy given by force and pot
{
   if (SysControl::noInteractionChargeToCharge) return Force();
   Vec xr=x-pc.x; // points from pc to *this
   R rr=xr.abs();
   if (rr>SysControl::intactRange) return Vec();
   R Fabs=q*pc.q*kel*cpmpow(r*r+rr*rr,R(-1.5));
   return Force(xr*Fabs);
}

Force PointCharge::forceMag(Vec const& v, PointCharge const& pc,
   Vec const& vpc )const
{
   if (SysControl::noMagneticInteractionChargeToCharge) return Force();
   VecBH vbhx=PointCharge::magneticField(x,pc,vpc);
   Vec temp=vbhx.cross(v);
   return Force(temp*(-q));
}

R PointCharge::pot( const PointCharge& pc )const
{
   if(SysControl::noInteractionChargeToCharge) return 0;
   Vec xr=x-pc.x; // points from pc to *this
   R rr=xr.abs();
   if (rr>SysControl::intactRange) return 0;
   return q*pc.q*kel*cpmpow(r*r+rr*rr,R(-0.5));
}

Force PointCharge::force( const V<PointCharge>& vpc )const
{
   Force f;
   Z n=vpc.dim();
   for (Z i=1;i<=n;i++) f+=force(vpc[i]);
   return f;
}

Force PointCharge::force(const Plane& pl, R eps)const
{
   if(SysControl::noInteractionChargeToWall) return Vec();
   R h=pl.dist(x);
   if (h<0) h=r*0.5;
   Vec n=pl.n();
   Spc x1=x-n*(h*2);
   R qAux=-q/(eps+1);
   R q1=(eps-1)*qAux;
   return force(PointCharge(q1,x1));
}

Force PointCharge::force(const Plane& pl, R eps, R d)const
{
   if(SysControl::noInteractionChargeToWall) return Vec();
   R h=pl.dist(x);
   if (h<0) h=r*0.5;
   Vec n=pl.n();
   Spc x1=x-n*(h*2);
   Spc x2=x1-n*(d*2);
   R qAux=-q/(eps+1);
   R q1=(eps-1)*qAux;
   R q2=qAux*2;
   return force(PointCharge(q1,x1))+force(PointCharge(q2,x2));
}

VecE PointCharge::electrostaticField(Spc const& x, const PointCharge& pc)
{
   if (pc.q==0.) return VecE();
   Vec xr=x-pc.x;
   R rr=xr.abs();
   regularize(rr,pc.r);
   return VecE(xr*(pc.q*kel*cpmpow(rr,-3)));
      // OK for pc.q>0, this has to point away from pc.x
      // (like the force on a positive charge probe.
}

VecBH PointCharge::magneticField(Spc const& x, const PointCharge& pc,
         const Vec& v)
{
   if (pc.q==0.) return VecBH();
   Vec xr=x-pc.x;
   R rr=xr.nor_();
   R nom=R(1)/((rr*rr+pc.r*pc.r)*cpmpi*4);
   Vec dI=v*pc.q;
   AxVec dH=dI.cross(xr)*nom;
   // my source: Weizel I, p. 371, (27) where I*ds  corresponds
   // to my dI
   R3 res; // all components are 0
   if (Spc::dimension()==3){

      res[1]=dH[1];
      res[2]=dH[2];
      res[3]=dH[3];

   }
   else{
      res[3]=dH[3];
   }
   return VecBH(res,"H");
}

VecE PointCharge::electrostaticField(Spc const& x,
   const V<PointCharge>& vpc)
{
   VecE res;
   Z n=vpc.dim();
   for (Z i=1;i<=n;i++) res+=electrostaticField(x,vpc[i]);
   return res;
}

R PointCharge::electrostaticPotential(Spc const& x,
   const PointCharge& pc)
   //
{
   if (pc.q==0.) return 0.;
   Vec xr=x-pc.x;
   R rr=xr.abs();
   regularize(rr,pc.r);
   return pc.q*kel/rr;
}

R PointCharge::electrostaticPotential(Spc const& x,
         const V<PointCharge>& vpc)
{
   R res=0;
   Z n=vpc.dim();
   for (Z i=1;i<=n;i++) res+=electrostaticPotential(x,vpc[i]);
   return res;
}

FieldE PointCharge::field()const
         // returns the electric field generated by the source *this
{
   return FieldE(F1<Spc,PointCharge,VecE>(*this)(electrostaticField));
}

FieldBH PointCharge::fieldBH(const Vec& v)const
{
   return FieldBH(F2<Spc,PointCharge,Vec,VecBH>(*this,v)(magneticField));
}

ScalarField PointCharge::fieldU()const
         // returns the electric field generated by the source *this
{
   return F1<Spc,PointCharge,R>(*this)(electrostaticPotential);
}

///////////////////////// class PatchCharge ///////////////////////////

namespace{//anon

R reg(R y, R r) // stay away from y==0
{
   R sig=(y>=0 ? 1 :-1);
   return sig*cpmsqrt(y*y+r*r);
}

bool indFunc(const Rdim_& x, const R& w, const R& depth)
{
   bool b1=cpmabs(x[1])<=w*0.5;
   bool b2=true;
   if (dimVec==3){
      b2=cpmabs(x[3])<=depth*0.5;
   }
   return b1&&b2;
}

// implementations based on the anti-derivation method
// (unbestimmtes Integral). See maintheorem.nb for methodological
// considerations on the connection of indefinite integrals
// (anti-derivations) and definite integrals over rectangles.
// Unfortunately this implementation was delayed by not
// realizing for 3 days that Mathematicas ArcTan(x,y)
// is not C's atan2(x,y) but atan2(y,x) even in the creation
// of C-code. (Solved 03-04-09)

/******************/
inline void f3(R& fx, R& fy, R& fz, R x, R y, R z)
// name appendix 3 says that the first 3 arguments are of ref type
// The reason for this argument structure is that
// one needs to compute r only once
{
   R r=cpmsqrt(x*x+y*y+z*z);
   fx=cpmlog(y+r);
   fy=cpmlog(x+r);
   fz=-cpmatan2(x*y/z,r);
}
/********************/

/************
inline R3 fa(R x, R y, R z)
{
   R r=cpmsqrt(x*x+y*y+z*z);
   return R3(cpmlog(y+r),cpmlog(x+r),-cpmatan(x*y/(z*r)));
}
***************/

inline R f(R x, R y, R z)
{
   R r=cpmsqrt(x*x+y*y+z*z);
   return x*cpmlog(y+r)+y*cpmlog(x+r)-z*cpmatan((x*y)/(z*r));
}
   // f this is one of the many functions
   // AntiDer(x) AntiDer(y) (x^2+y^2+z^2)^(-1/2)
   // see my Mathematica file potential_and_field_of_rectangle.nb
   // a surprisingly simple function.
   // fx,fy,fz are the corresponding partial derivatives of f.
} //anon

B PatchCharge::makeFieldFromFieldU(false);
   // false is by far faster (and more precise) so
   // true is only acceptable for debugging.

const R facSqr=400;
   // factor 20 in distance
   // controls the decision from what distance from the patch on,
   // the patch should be replaced by a point charge

V< X2<R2,Z> > PatchCharge::corners()const
{
   R a=w*0.5;
   R b=depth*0.5;
   V< X2<R2,Z> > res(4);
   res[1]=X2<R2,Z>(R2(a,b),1);
   res[2]=X2<R2,Z>(R2(a,-b),-1);
   res[3]=X2<R2,Z>(R2(-a,b),-1);
   res[4]=X2<R2,Z>(R2(-a,-b),1);
   return res;
}

R PatchCharge::electrostaticPotential(Spc const& p,
   const PatchCharge& sc)
{
   if (sc.sigma==0.) return 0;
   R a=sc.w*0.5;
   R b=sc.depth*0.5;
   Rdim xr=sc.rs(p);
   R y=0;
   R x=xr[1];
   R z=reg(xr[2],sc.r);
   if (dimVec==3) y=-xr[3];

   R a2=a*a;
   R b2=b*b;
   R disSqr=x*x+y*y+z*z;
   R sizeSqr=a2+b2;
   if (disSqr>facSqr*sizeSqr){ // use point charge approximation
      PointCharge pc(sc.sigma*4*a*b,sc.get_x());
      return PointCharge::electrostaticPotential(p,pc);
   }
   else{
      R xp=x+a, xm=x-a, yp=y+b, ym=y-b;
      R res=f(xp,yp,z)-f(xp,ym,z)-f(xm,yp,z)+f(xm,ym,z);
      res*=influence_PHYS*sc.sigma;
      return res;
   }
}

VecE PatchCharge::electrostaticField(Spc const& p,
   const PatchCharge& sc)
// see potential_and_field_of_rectangle.nb by UM
{
   if (sc.sigma==0.) return VecE();
   R a=sc.w*0.5;
   R b=sc.depth*0.5;
   Rdim xr=sc.rs(p);
   R y=0;
   R x=xr[1];
   R z=reg(xr[2],sc.r);
   if (dimVec==3) y=-xr[3];

   R a2=a*a;
   R b2=b*b;
   R disSqr=x*x+y*y+z*z;
   R sizeSqr=a2+b2;
   if (disSqr>facSqr*sizeSqr){ // use point charge approximation
      PointCharge pc(sc.sigma*4*a*b,sc.get_x());
      return PointCharge::electrostaticField(p,pc);
   }
   if (makeFieldFromFieldU){
      Vec res=CPM_DIM::grad(p, F1<Spc,PatchCharge,R>(sc)
         (electrostaticPotential), sc.r);
         // uses fast and simplistic gradient formula
      return VecE(-res);
   }
   else{
      // direct field formula, which is numerically less complex
      // and significantly faster

      R fx1,fy1,fz1, fx2,fy2,fz2, fx3,fy3,fz3, fx4,fy4,fz4;
      R xp=x+a, xm=x-a, yp=y+b, ym=y-b; // p: plus, m: minus

      f3(fx1,fy1,fz1,xp,yp,z);
      f3(fx2,fy2,fz2,xp,ym,z);
      f3(fx3,fy3,fz3,xm,yp,z);
      f3(fx4,fy4,fz4,xm,ym,z);

      R Ex=fx1-fx2-fx3+fx4;
      R Ey=dimVec==3 ? fy1-fy2-fy3+fy4 : R(0);
      R Ez=fz1-fz2-fz3+fz4;

      Vec ex=sc.rs.getVectors()[1];
      Vec ey=(dimVec==3 ? -sc.rs.getVectors()[3] : Vec() );
      Vec ez=sc.rs.getVectors()[2];

      Vec res= (dimVec==3 ? ex*Ex+ey*Ey+ez*Ez : ex*Ex+ez*Ez);
      res*=(-influence_PHYS*sc.sigma);
      return VecE(res);
   }
}

ScalarField PatchCharge::fieldU()const
{
   return F1<Spc,PatchCharge,R>(*this)(electrostaticPotential);
}

FieldE PatchCharge::field()const
{
   return FieldE(F1<Spc,PatchCharge,VecE>(*this)(electrostaticField));
}

SubsetOfPlane PatchCharge::getPatch()const
{
   return SubsetOfPlane(rs,F2<Rdim_,R,R,bool>(w,depth)(indFunc));
}

V<Spc> PatchCharge::range()const
{
   Rdim c1,c2;
   R wh=w*0.5;
   c1[1]=-wh;
   c2[1]=wh;
   V<Spc> res(2);
   res[1]=rs(c1);
   res[2]=rs(c2);
   return res;
}

void PatchCharge::mark(const Camera& ca, Color c, bool shAx)const
// shAx = show axes
{
   R wh=w*0.5;
   R dh=depth*0.5;

   if (Spc::dimension()==2){
      Rdim p1,p2;
      p1[1]=wh;
      p2[1]=-wh;
      V<Spc> p(2); // was 4 till 2004-06-28
      p[1]=rs(p1);
      p[2]=rs(p2);
      Skeleton sk(p,true);
      ca.mark(sk,c);
   }
   else{

      Rdim p1,p2,p3,p4;

      p1[1]=wh;
      p1[3]=dh;

      p2[1]=wh;
      p2[3]=-dh;

      p3[1]=-wh;
      p3[3]=-dh;

      p4[1]=-wh;
      p4[3]=dh;

      V<Spc> p(4);

      p[1]=rs(p1);
      p[2]=rs(p2);
      p[3]=rs(p3);
      p[4]=rs(p4);

      Skeleton sk(p,true);
      sk.clsPath_();
      ca.mark(sk,c);
   }
   if (shAx) ca.mark(rs,c,wh);
}

V<PatchCharge> PatchCharge::increase(const V<Vec>& vv)const
{
   Z i,vvd=vv.dim();
   V<PatchCharge> res(vvd,*this);
   for (i=1;i<=vvd;i++) res[i]+=vv[i];
   return res;
}

bool PatchCharge::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("PatchCharge",str);
   cpmp(rs);
   cpmp(w);
   cpmp(sigma);
   cpmp(r);
   return true;
}

bool PatchCharge::scanFrom(istream& str)
{
   cpms(rs);
   cpms(w);
   cpms(sigma);
   cpms(r);
   return true;
}


R PatchCharge::test()
{
   if (dimVec==2){
      cpmerror("R PatchCharge::test() not implemented for 2D");
      return 0;
   }
   const Z iStart=3;
   const R range=13.7;
   const Z n=128;

   B makeFieldFromFieldU_mem=makeFieldFromFieldU;

   RefSys rfs;
   Z ir=iStart;
   Iv ivAngle(-180,180); // intervals are very convenient
      // input data for shaping the random generation process.
   R w1=ivAngle.ran(ir++);
   R w2=ivAngle.ran(ir++);
   R w3=ivAngle.ran(ir++);
   Angle alpha1(w1);
   Angle alpha2(w2);
   Angle alpha3(w3);
   AxVec a1(alpha1,1);
   AxVec a2(alpha2,2);
   AxVec a3(alpha3,3);
   Group g1(a1);
   Group g2(a2);
   Group g3(a3);
   rfs*=g1;
   rfs*=g2;
   rfs*=g3;
   Iv ivRange(-range,range);
   Vec sh;
   sh[1]=ivRange.ran(ir++);
   sh[2]=ivRange.ran(ir++);
   sh[3]=ivRange.ran(ir++);
   Group g4(sh);
   rfs*=g4;
   Iv ivSize(0,range*0.5);
   R width=ivSize.ran(ir++);
   R depth=ivSize.ran(ir++);
   R sigma=1.;
   R r=CpmRootX::inf<R>(width,depth)*1e-6;
   PatchCharge pc1(rfs,width,depth,sigma,r);
   V<Spc> tp(n); // tp = test points
   Z i;
   for (i=1;i<=n;i++){
      R x=ivRange.ran(ir++);
      R y=ivRange.ran(ir++);
      R z=ivRange.ran(ir++);
      Rdim xyz;
      xyz[1]=x;
      xyz[2]=y;
      xyz[3]=z;
      tp[i]=rfs(xyz);
   }
   R res=0;
   V<VecE> res1(n);
   makeFieldFromFieldU=false;
   FieldE f1=pc1.field();
   for (i=1;i<=n;i++){
      res1[i]=f1(tp[i]);
   }
   RefSys rs0;
// different style of creating and using the 'same' object
   PatchCharge pc2(rs0,width,depth,sigma,r);
   pc2*=g1;
   pc2*=g2;
   pc2*=g3;
   pc2*=g4;
   V<VecE> res2(n);
   makeFieldFromFieldU=true;
   for (i=1;i<=n;i++){
      res2[i]=pc2.eval(tp[i]);
   }
   for (i=1;i<=n;i++) res+=res1[i].dis(res2[i]);
      // since the res[i] may have very large or small 'values'
      // summing up (res1[i]-res2[i]).abs() would
      // be misleading
   res/=n;
   makeFieldFromFieldU=makeFieldFromFieldU_mem;
   ostringstream ost;
   ost<<"Class test for PatchCharge(): mean error was "<<res<<"."<<endl<<
   "   This should be well below 1e-5"<<endl<<
   "   Some data of the random patch: width="<<width<<" depth="<<depth<<
      " r="<<r<<endl<<
   "   Number of field points where the two methods of computing the\
 field were compared: "<<n<<endl;
   cpmmessage(ost);

   return res;
}

////////////// class MultiPatchCharge ////////////////////////////////////

void MultiPatchCharge::ini_(void)
{
   S<R2> sr;
   M<R2,Z> mrz;
   Z i;
   R2 c1(xr_[1],yr_[1]);
   R2 c2(xr_[1],yr_[2]);
   R2 c3(xr_[2],yr_[1]);
   R2 c4(xr_[2],yr_[2]);
   sr.add(c1); sr.add(c2); sr.add(c3); sr.add(c4);
   R2 cc1=c1, cc2=c2, cc3=c3, cc4=c4;
   Z z1=1, z2=-1, z3=-1, z4=1;
   mrz[cc1]=z1;
   mrz[cc2]=z2;
   mrz[cc3]=z3;
   mrz[cc4]=z4;

   bool bi;
   for (i=v_.b();i<=v_.e();++i){
      R2 vi=v_[i];
      cc1=c1+vi;
      cc2=c2+vi;
      cc3=c3+vi;
      cc4=c4+vi;
      bi=sr.addAct(cc1);
      if (bi) mrz[cc1]=z1; else mrz[cc1]+=z1;
      bi=sr.addAct(cc2);
      if (bi) mrz[cc2]=z2; else mrz[cc2]+=z2;
      bi=sr.addAct(cc3);
      if (bi) mrz[cc3]=z3; else mrz[cc3]+=z3;
      bi=sr.addAct(cc4);
      if (bi) mrz[cc4]=z4; else mrz[cc4]+=z4;
   }
   Z sumzi=0;
   for (i=mrz.b();i<=mrz.e();++i){
      R2 ci=mrz.x(i);
      Z  zi=mrz.y(i);
      if (zi!=0){
         sx_.add(X2<R2,Z>(ci,zi));
         cpmdebug(zi);
         sumzi+=zi;
      }
   }
/*   cpmdebug(sr);
   cpmdebug(mrz); */
   cpmdebug(sx_);
   cpmdebug(sumzi);
   cpmdebug(sr.car());
   R cost_rel=(sx_.car()*1.0)/sr.car();
   cpmdebug(cost_rel);
}

R MultiPatchCharge::ePot(Spc const& x)const
{
   R res=0;
   R sig=sigma_;
   if (sig==0. || dimVec==2) return res;
   Z i;
   Rdim xr=rs_(x);
   R2 p(xr[1],xr[2]);
   R z=reg(xr[3],r_);
   for (i=sx_.b();i<=sx_.e();++i){
      X2<R2,Z> si=sx_[i];
      R2 pi=si.c1()-p;
      Z wi=si.c2();
      res+=R(wi)*f(pi.x1, pi.x2, z);
   }
   res*=influence_PHYS*sigma_;
   return res;
}

ScalarField MultiPatchCharge::fieldU()const
{
   return F1<Spc,MultiPatchCharge,R>(*this)
   (MultiPatchCharge::electrostaticPotential);
}

////////////// class MultiVolCharge ////////////////////////////////////

/***********
// Is a valid anti-derivative accoding to Mathematica, but does not
// work here
inline R ad3(R x, R y, R z)
{
   R r=cpmsqrt(x*x+y*y+z*z);
   return z*y*cpmlog(x+r)+z*x*cpmlog(y+r)+x*y*cpmlog(z+r)
      -z*z*atan(x*y/(z*r))-y*y*atan(x*z/(y*r))-0.5*x*x*atan(y*z/(x*r))
      +0.5*y*y*atan2(x*z,y*r)+0.5*z*z*atan2(x*y,z*r);
 }
 ***************/

// Working form of anti-derivative, also accepted by Mathematica.
// 2010-07-15
R ad3(R x, R y, R z)
{
   R x2=x*x, y2=y*y, z2=z*z;
   R xy=x*y, yz=y*z, zx=z*x;
   R r=cpmsqrt(x2+y2+z2);
   return yz*cpmlog(x+r)+zx*cpmlog(y+r)+xy*cpmlog(z+r)
      -R(0.5)*(x2*cpmatan(yz/(x*r))+y2*cpmatan(zx/(y*r))+z2*cpmatan(xy/\
      (z*r)));
 }

void MultiVolCharge::ini_(void)
{
   S<R3> sr;
   M<R3,Z> mrz;
   Z i;
   R3 c1(xr_[1],yr_[1],zr_[1]);
   R3 c2(xr_[1],yr_[1],zr_[2]);
   R3 c3(xr_[1],yr_[2],zr_[1]);
   R3 c4(xr_[1],yr_[2],zr_[2]);
   R3 c5(xr_[2],yr_[1],zr_[1]);
   R3 c6(xr_[2],yr_[1],zr_[2]);
   R3 c7(xr_[2],yr_[2],zr_[1]);
   R3 c8(xr_[2],yr_[2],zr_[2]);// The upper point has weight 1!
   sr.add(c1); sr.add(c2); sr.add(c3); sr.add(c4);
   sr.add(c5); sr.add(c6); sr.add(c7); sr.add(c8);
   R3 cc1=c1, cc2=c2, cc3=c3, cc4=c4;
   R3 cc5=c5, cc6=c6, cc7=c7, cc8=c8;
   Z z1=-1, z2=1, z3=1, z4=-1;
   Z z5=1, z6=-1, z7=-1, z8=1;
   mrz[cc1]=z1;
   mrz[cc2]=z2;
   mrz[cc3]=z3;
   mrz[cc4]=z4;
   mrz[cc5]=z5;
   mrz[cc6]=z6;
   mrz[cc7]=z7;
   mrz[cc8]=z8;

   bool bi;
   for (i=v_.b();i<=v_.e();++i){
      // loop over the translation vectors
      R3 vi=v_[i];
      cc1=c1+vi;
      cc2=c2+vi;
      cc3=c3+vi;
      cc4=c4+vi;
      cc5=c5+vi;
      cc6=c6+vi;
      cc7=c7+vi;
      cc8=c8+vi;
      bi=sr.addAct(cc1);
      if (bi) mrz[cc1]=z1; else mrz[cc1]+=z1;
      bi=sr.addAct(cc2);
      if (bi) mrz[cc2]=z2; else mrz[cc2]+=z2;
      bi=sr.addAct(cc3);
      if (bi) mrz[cc3]=z3; else mrz[cc3]+=z3;
      bi=sr.addAct(cc4);
      if (bi) mrz[cc4]=z4; else mrz[cc4]+=z4;
      bi=sr.addAct(cc5);
      if (bi) mrz[cc5]=z5; else mrz[cc5]+=z5;
      bi=sr.addAct(cc6);
      if (bi) mrz[cc6]=z6; else mrz[cc6]+=z6;
      bi=sr.addAct(cc7);
      if (bi) mrz[cc7]=z7; else mrz[cc7]+=z7;
      bi=sr.addAct(cc8);
      if (bi) mrz[cc8]=z8; else mrz[cc8]+=z8;
   }
   Z sumzi=0;
   for (i=mrz.b();i<=mrz.e();++i){
      // collecting the non-vanishing contributions
      R3 ci=mrz.x(i);
      Z  zi=mrz.y(i);
      if (zi!=0) {
         sx_.add(X2<R3,Z>(ci,zi));
         cpmdebug(zi);
         sumzi+=zi;
      }
   }
/*   cpmdebug(sr);
   cpmdebug(mrz); */
   cpmdebug(sx_);
   cpmdebug(sumzi);
   cpmdebug(sr.car());
   R cost_rel=(sx_.car()*1.0)/sr.car();
   cpmdebug(cost_rel);
}

R MultiVolCharge::ePot(Spc const& x)const
{
   R res=0;
   R sig=sigma_;
   if (sig==0. || dimVec==2) return res;
   R eps=R(r_)*R(r_);
   Rdim xr=rs_(x);
   R3 p(xr[1],xr[2],xr[3]);
   for (Z i=sx_.b();i<=sx_.e();++i){
      X2<R3,Z> si=sx_[i];
      R3 pi=si.c1()-p;
      Z wi=si.c2();
      res+=ad3(reg(pi.x1), reg(pi.x2), reg(pi.x3))*wi;
      // avoid division by 0
   }
   res*=influence_PHYS*sigma_;
   return res;
}

ScalarField MultiVolCharge::fieldU()const
{
   return F1<Spc,MultiVolCharge,R>(*this)
   (MultiVolCharge::electrostaticPotential);
}

////////////// class ElcDipole ///////////////////////////////////////////

AxVec ElcDipole::torque(const VecE& E)const
{
   return d.cross(E);
}

bool ElcDipole::prnOn( ostream& str )const
{
   CpmRoot::writeTitle("ElcDipole",str);
   cpmp(d);
   cpmp(r);
   cpmp(x);
   return true;
}

bool ElcDipole::scanFrom( istream& str )
{
   cpms(d);
   cpms(r);
   cpms(x);
   return true;
}

ForTor ElcDipole::forceAndTorque( const F<Spc,VecE>& fE, R r)const
// number of fE-evaluations minimized
{
   if(SysControl::noInteractionElectricField ||
      SysControl::noInteractionElectricDipole) return ForTor();
   X2<R,Vec> dd=d.polDec();
   R  dabs=dd.first();
   Vec n=dd.second();
   Vec a=r*n;
   Spc x1=x+a;
   Vec E1=fE(x1);
   Spc x2=x-a;
   Vec E2=fE(x2);
   R Qel=dabs*CpmRoot::inv(r*2);
      // see Weizel I, p. 321, Abb 103, p. 322 first formula
   Vec F=Qel*(E1-E2);
   Vec E0=0.5*(E1+E2); // in order to avoid a third evaluation at the
   // center
   AxVec T=d.cross(E0);
   return ForTor(F,T);
}

ForTor ElcDipole::forceAndTorque( const ElcDipole& ed )const
{
   if(SysControl::noInteractionElectricDipole) return ForTor();
   Vec xr=x-ed.x;
   X2<R,Vec> xrpd=xr.polDec();
   R  rr=xrpd.first();
   if (rr>SysControl::intactRange) return ForTor();
   Vec n=xrpd.second();
   regularize(rr,r);
   R fac1=kel*cpmpow(rr,-3);
   R fac2=fac1*3/rr;
   R dn=d|n;
   R dd=d|ed.d;
   R ddn=ed.d|n;
   Vec E=n*3*ddn-ed.d;
   E*=fac1;
   AxVec T=d.cross(E);
   Vec F=ed.d*dn+d*ddn+n*(dd-dn*ddn*5);
   F*=fac2;
   return ForTor(F,T);
}

VecE ElcDipole::electrostaticDipoleField(Spc const& x,
   const ElcDipole& md)
{
   if(SysControl::noInteractionElectricDipole) return VecE();
   Vec xr=x-md.x;
   X2<R,Vec> xrpd=xr.polDec();
   R  rr=xrpd.c1();
   regularize(rr,md.r);
   Vec n=xrpd.c2();
   R fac=kel*cpmpow(rr,-3);
   R dn=md.d|n;
   Vec E=n*(dn*3)-md.d;
   E*=fac;
   return VecE(E);
}

FieldE ElcDipole::field()const
   // returns the magnetic field generated by the source *this
{
   return FieldE(F1<Spc,ElcDipole,VecE>(*this)(electrostaticDipoleField));
}

R ElcDipole::electrostaticDipoleFieldU(Spc const& x,
   const ElcDipole& md)
{
   if(SysControl::noInteractionElectricDipole) return 0.;
   Vec xr=x-md.x;
   X2<R,Vec> xrpd=xr.polDec();
   R  rr=xrpd.c1();
   regularize(rr,md.r);
   Vec n=xrpd.c2();
   R fac=kel*cpmpow(rr,-2);
   R dn=md.d|n;
   return fac*dn;
}

ScalarField ElcDipole::fieldU()const
         // returns the magnetic field generated by the source *this
{
   return F1<Spc,ElcDipole,R>(*this)(electrostaticDipoleFieldU);
}

///////  class MagPole //////////////////////////////////////////

Force MagPole::force( const MagPole& pc )const
{
   if(SysControl::noInteractionMagnetToMagnet) return Vec();
   Vec xr=x-pc.x; // points from pc to *this
   R rr=xr.abs();
   if (rr>SysControl::intactRange) return Vec();
   regularize(rr,r);
   R fac=kmag*q*pc.q*cpmpow(rr,-3); // positive if charges have equal sign
   return Force(xr*fac); // repulsive if fac is positive, OK
}

VecBH MagPole::magnetostaticField(Spc const& x, const MagPole& mp)
{
   Vec xr=x-mp.x;
   R rr=xr.abs();
   rr=(rr<mp.r ? mp.r : rr);
   return VecBH(xr*(mp.q*kmag*cpmpow(rr,-3)),"H");
}

FieldBH MagPole::field()const
         // returns the electric field generated by the source *this
{
   return F1<Spc,MagPole,VecBH>(*this)(magnetostaticField);
}

bool MagPole::prnOn( ostream& str )const
{
   CpmRoot::writeTitle("MagPole",str);
   cpmp(q);
   cpmp(r);
   cpmp(x);
   return true;
}

bool MagPole::scanFrom( istream& str )
{
   cpms(q);
   cpms(r);
   cpms(x);
   return true;
}

 /// class MagDipole ///////////////////////////////////////////////

AxVec MagDipole::torque(const VecBH& H)const
{
   return j.cross(H.v());
}

bool MagDipole::prnOn( ostream& str )const
{
   CpmRoot::writeTitle("MagDipole",str);
   cpmp(j);
   cpmp(r);
   cpmp(x);
   return true;
}

bool MagDipole::scanFrom( istream& str )
{
   cpms(j);
   cpms(r);
   cpms(x);
   return true;
}

ForTor MagDipole::forceAndTorque( const F<Spc,VecBH>& fH, R r)const
// number of fH-evaluations minimized
{
   if(SysControl::noInteractionMagneticField) return ForTor();
   X2<R,Vec> jd=j.polDec();
   R  jabs=jd.c1();
   Vec n=jd.c2();
   Vec a=r*n;
   Spc x1=x+a;
   Vec H1=fH(x1).v();
   Spc x2=x-a;
   Vec H2=fH(x2).v();
   R Qmag=jabs*CpmRoot::inv(r*2);
      // see Weizel I, p. 321, Abb 103, p. 322 first formula
   Vec F;
   if (!SysControl::noForceMagneticField) F=Qmag*(H1-H2);
      // in some situations it helps if a magnetic field only orients a
      // magnetized particle without carrying it away
   Vec H0=0.5*(H1+H2); // in order to avoid a third evaluation at the
   // center
   AxVec T=j.cross(H0);
   return ForTor(F,T);
}

ForTor MagDipole::forceAndTorque( const MagDipole& d )const
// see UM Ariton 99-4-21/4-6 und Nachtrag
{
   if(SysControl::noInteractionMagnetToMagnet) return ForTor();
      // was missing till 99-11-4
   Vec xr=x-d.x;
   X2<R,Vec> xrpd=xr.polDec();
   R  rr=xrpd.c1();
   if (rr>SysControl::intactRange) return ForTor();
   Vec n=xrpd.c2();
   regularize(rr,r);
   R fac1=kmag*cpmpow(rr,-3);
   R fac2=fac1*3/rr;
   R jn=j|n;
   R jd=j|d.j;
   R djn=d.j|n;
   Vec H=n*(djn*3)-d.j;
   H*=fac1;
   AxVec T=j.cross(H);
   Vec F=d.j*jn+j*djn+n*(jd-jn*djn*5);
   F*=fac2;
   return ForTor(F,T);
}

VecBH MagDipole::magnetostaticDipoleField(Spc const& x,
   const MagDipole& md)
{
   if(SysControl::noInteractionMagneticField) return VecBH();
   Vec xr=x-md.x;
   X2<R,Vec> xrpd=xr.polDec();
   R  rr=xrpd.c1();
   regularize(rr,md.r);
   Vec n=xrpd.c2();
   R fac=kmag*cpmpow(rr,-3);
   R jn=md.j|n;
   Vec H=n*3.*jn-md.j;
   return VecBH(H*=fac,"H");
}

R MagDipole::magnetostaticDipoleFieldU(Spc const& x,
   const MagDipole& md)
{
   if(SysControl::noInteractionMagneticField) return 0.;
   Vec xr=x-md.x;
   X2<R,Vec> xrpd=xr.polDec();
   R  rr=xrpd.c1();
   regularize(rr,md.r);
   Vec n=xrpd.c2();
   R fac=kmag*cpmpow(rr,-2);
   R jn=md.j|n;
   return fac*jn;
}

R MagDipole::pot(MagDipole const& md)const
// Potential energy for two magnetic dipoles is
// treated in a Wikipedia article on dipole-dipole interaction.
// Here the dipole moments are Ampere's moments, wheras I use
// Coulomb's moments.
 {
    if(SysControl::noInteractionMagnetToMagnet) return 0.;
    Vec n=x-md.x;
    R  rr=n.nor_();
    regularize(rr,md.r);
    R fac=kmag*cpmpow(rr,-3);
    Vec jm=md.j;
    return fac*((j|jm)-(n|j)*(n|jm)*3.0);
 }

FieldBH MagDipole::field()const
{
   return F1<Spc,MagDipole,VecBH>(*this)(magnetostaticDipoleField);
}

ScalarField MagDipole::fieldU()const
{
   return F1<Spc,MagDipole,R>(*this)(magnetostaticDipoleFieldU);
}

////////////////////// FieldSpace /////////////////////////////////////

bool FieldSpace::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("FieldSpace",str);
   cpmp(E_);
   cpmp(BH_);
   cpmp(G_);
   return true;
}

bool FieldSpace::scanFrom(istream& str)
{
   cpms(E_);
   cpms(BH_);
   cpms(G_);
   return true;
}

/////////  class VectorField ///////////////////////////////////////////
namespace{ // anonymous namespace

FieldSpace combineEBHG(Spc const& x,
   const FieldE& fE, const FieldBH& fBH, const FieldG& fG)
{
   VecE vE=fE(x);
   VecBH vBH=fBH(x);
   VecG vG=fG(x);
   return FieldSpace(vE,vBH,vG);
}

Spc funcPeriodize(Spc const& x0, Spc const& x1, Spc const& x2)
{
   Word loc("funcPeriodize(...)");
   Vec v12=x2-x1;
   Vec v=v12;
   R l=v12.nor_();
   cpmassert(l>0,loc);
   Vec v10=x0-x1;
   R a=(v12|v10);
   R n=cpmfloor(a/l);
   return x0-v*n;
}

} // namespace

VectorField::VectorField(const FieldE& fE,
   const FieldBH& fBH, const FieldG& fG):
Base(F3<Spc,FieldE,FieldBH,FieldG,FieldSpace>(fE,fBH,fG)(combineEBHG)){}

static VecE f1(Spc const& x, const VectorField& f){return f(x).c1();}
static VecBH f2(Spc const& x, const VectorField& f){return f(x).c2();}
static VecG f3(Spc const& x, const VectorField& f){return f(x).c3();}

void ScalarField::periodize(Spc const& x1, Spc const& x2)
{
   leftCombine(F2<Spc,Spc,Spc,Spc>(x1,x2)(funcPeriodize));
}

void FieldE::periodize(Spc const& x1, Spc const& x2)
{
   leftCombine(F2<Spc,Spc,Spc,Spc>(x1,x2)(funcPeriodize));
}

FieldE VectorField::c1()const
{
   return FieldE(F1<Spc,VectorField,VecE>(*this)(f1));
}

FieldBH VectorField::c2()const
{
   return FieldBH(F1<Spc,VectorField,VecBH>(*this)(f2));
}

FieldG VectorField::c3()const
{
   return FieldG(F1<Spc,VectorField,VecG>(*this)(f3));
}

namespace{ // anonymous namespace

Vec v1(Spc const& x, const VectorField& f){return f(x).c1();}
Vec v2(Spc const& x, const VectorField& f){return f(x).c2().v();}
Vec v3(Spc const& x, const VectorField& f){return f(x).c3();}

FieldSpace fSetU(Spc const& x, const ScalarField& sf,
   const VectorField& vf)
{
   FieldSpace res=vf(x);
   Vec v=grad(sf)(x);
   VecE ve(-v);
   res.c1()=ve;
   return res;
}

FieldSpace fSetE(Spc const& x, const FieldE& fE, const VectorField& vf)
{
   FieldSpace res=vf(x);
   res.c1()=fE(x);
   return res;
}

FieldSpace fSetBH(Spc const& x, FieldBH const& fBH, VectorField const& vf)
{
   FieldSpace res=vf(x);
   res.c2()=fBH(x);
   return res;
}

FieldSpace fSetG(Spc const& x, const FieldG& fG, const VectorField& vf)
{
   FieldSpace res=vf(x);
   res.c3()=fG(x);
   return res;
}

FieldSpace faddE(Spc const& x, const FieldE& f, const VectorField& vf)
{
   FieldSpace res=vf(x);
   res.c1()+=f(x);
   return res;
}

FieldSpace faddBH(Spc const& x, const FieldBH& f, const VectorField& vf)
{
   FieldSpace res=vf(x);
   res.c2()+=f(x);
   return res;
}

FieldSpace faddG(Spc const& x, const FieldG& f, const VectorField& vf)
{
   FieldSpace res=vf(x);
   res.c3()+=f(x);
   return res;
}

} // namespace

VecField VectorField::operator[](Z i)const
{
   cpmassert(i>0 && i<4,"VecField VectorField::operator[](Z i)const");
   if (i==1) return VecField(F1<Spc,VectorField,Vec>(*this)(v1));
   if (i==2) return VecField(F1<Spc,VectorField,Vec>(*this)(v2));
   else return VecField(F1<Spc,VectorField,Vec>(*this)(v3));
}

VectorField::VectorField(const V<Spc>& x, const V<FieldSpace>& y):
Base(new CpmAlgorithms::AveragingInterpolator<Spc,FieldSpace>(x,y)){}


void VectorField::setU(const ScalarField& sf)
{
   *this=F2<Spc,ScalarField,VectorField,FieldSpace>(sf,*this)(fSetU);
}

void VectorField::setE(const FieldE& fE)
{
   *this=F2<Spc,FieldE,VectorField,FieldSpace>(fE,*this)(fSetE);
}

void VectorField::setBH(const FieldBH& fBH)
{
   *this=F2<Spc,FieldBH,VectorField,FieldSpace>(fBH,*this)(fSetBH);
}

void VectorField::setG(const FieldG& fG)
{
   *this=F2<Spc,FieldG,VectorField,FieldSpace>(fG,*this)(fSetG);
}

void VectorField::add(const FieldE& f)
{
   *this=F2<Spc,FieldE,VectorField,FieldSpace>(f,*this)(faddE);
}

void VectorField::add(const FieldBH& f)
{
   *this=F2<Spc,FieldBH,VectorField,FieldSpace>(f,*this)(faddBH);
}

void VectorField::add(const FieldG& f)
{
   *this=F2<Spc,FieldG,VectorField,FieldSpace>(f,*this)(faddG);
}

Vec gradient(Spc const& x , const ScalarField& f,const R& h)
{
   return CPM_DIM::grad(x,f,h);
}

VecField grad(const ScalarField& sf, R h)
{
   return F2<Spc,ScalarField,R,Vec>(sf,h)(gradient);
}

static Vec optTransfer(const Vec& v)
// This function will be used in ScalarField::mark() and
// VectorField::mark() only.
// It turns out that fields the order of magnitude of which varies
// strongly over the display area get visible representaton only at a few
// points if represented linearily. So, in analogy to imaging transfer
// curves I allow for a transformation of the display values too. Notice
// that one has to separate into absolute value and direction if the
// direction of the gradient is to be represented correctly.
{
   R r=v.polDec().c1();
   if (r==0) return v;
   R fac=cpmpow(r,SysControl::fieldGamma)/r;
   return v*fac;
}

void markVecField(const VecField& f, const Camera& sc,
      R& scaleFactor, Z gridPointsHorizontal, Z gridPointsVertical,
      const Color& color)
{
   Z mL=3;
   Word loc("markVecField(...)");
   cpmmessage(mL,loc&" started, scaleFactor="&cpmwrite(scaleFactor));
   cpmassert(gridPointsHorizontal>0 && gridPointsVertical>0,loc);
// getting from gr representative points in object space
   C zLL(sc.getxLeft(),sc.getyLower());
      // lower left point of rectangle associated with gr
   R dx=sc.getdX()/gridPointsHorizontal;
   R dy=sc.getdY()/gridPointsVertical;
   C cell(dx,dy);
   R cellSize=CpmRootX::inf(dx,dy)*0.5;
      // independent of its direction, the arrow should never leave
      // its cell
   C pointLL=zLL+cell*0.5;
   Z nP=gridPointsHorizontal*gridPointsVertical;
   V<Spc> spc(nP);
   Z i,j,iCounter=1;
   R pLx=pointLL[1];
   R pLy=pointLL[2];
   R spx=pLx;
   R spy=pLy;
   for (i=1;i<=gridPointsHorizontal;i++){
      for (j=1;j<=gridPointsVertical;j++){
         spc[iCounter++]=sc.inverse(C(spx,spy),1.);
         spy+=dy;
      }
      spy=pLy;
      spx+=dx;
   }
// recording field strength's at the test points

   V<Vec> fields(nP);

   for (i=1;i<=nP;i++){
      Vec fVal=f(spc[i]);
      fields[i]=optTransfer(fVal);
   }
   // if the scaleFactor is not yet determined we have to make
   // field vectors and position vectors comparable (they typically
   // don't even have the same dimension)
   R tWait=-1;
   if (scaleFactor<=0){
      tWait=2;
      Vo<R> abs(nP);
      for (i=1;i<=nP;i++) abs[i]=fields[i].abs();
      R scale=abs.sup();
      if (scale<=0){
         cpmwarning("denied to draw a trivial vector field");
         return; // nothing worthwhile to be drawn
      }
      scaleFactor=cellSize/scale;
   }
   R rHead=cellSize*0.04;
   for (i=1;i<=nP;i++){
      Spc spcEnd=spc[i]+fields[i]*scaleFactor;
      sc.mark(spc[i],spcEnd,color,rHead);
         // argument color introduced 2001-11-26
   }
   cpmwait(tWait);
   cpmmessage(mL,loc&" done");
}

R VectorField::scaleFactorE=-1;
R VectorField::scaleFactorM=-1;
R VectorField::scaleFactorG=-1;

void VectorField::mark(const Camera& sc)const
{
   if (SysControl::markElectricField>=0){
      // was > 0 till 2003-06-16 thus disabling anaglyphic field
      // representation
      VecField fE=(*this)[1];
      Z colNumber=SysControl::markElectricField;
      Color c(INVISIBLE);
      if (colNumber>0) c=Color(colNumber);
      markVecField(fE,sc,scaleFactorE,SysControl::nxElectricField,
         SysControl::nyElectricField,c);
         // SysControl::markElectricField is the color, 0 for
         // anaglyphic representation, negativ for no representation
   }
   if (SysControl::markMagneticField>=0){
         // remark as in electric case
      VecField fM=(*this)[2];
      Color c(INVISIBLE);
      Z colNumber=SysControl::markMagneticField;
      if (colNumber>0) c=Color(colNumber);
      markVecField(fM,sc,scaleFactorM,SysControl::nxMagneticField,
         SysControl::nyMagneticField,c);
   }
   if (SysControl::markGravityField>=0){
      VecField fG=(*this)[3];
      Color c(INVISIBLE);
      Z colNumber=SysControl::markGravityField;
      if (colNumber>0) c=Color(colNumber);
      markVecField(fG,sc,scaleFactorG,SysControl::nxGravityField,
         SysControl::nyGravityField,c);
   }
}

//////////////////// ScalarField //////////////////////////////
namespace{
   R fConAux(Spc const& x, FieldSpace const& fs,
      Spc const& p)
   {
      Z pts=SysControl::potentialToShow;
      Vec fc;
      if (pts==2) fc=fs.c2().v();
      else if (pts==3) fc=(Vec)fs.c3();
      else fc=(Vec)fs.c1();
      return ((p-x)|fc);
   }
} // namespace

ScalarField::ScalarField(FieldSpace const& fs, Spc const& p):
Base(F2<Spc,FieldSpace,Spc,R>(fs,p)(fConAux)){}

void ScalarField::mark(const Camera& ca)const
{
   Z mL=2;
   Word loc("ScalarField::mark(..)");
   CPM_MA
   Response rsp=ca.getResponse();
   bool adjust=!rsp.normalizeGamma();
   Z ny=SysControl::nyScalarFields;
   cpmassert(ny>=16,loc); // warn or stop
   if (ny<16) ny=16;
   Z nx=cpmtoz(R(ny)*ca.getGraph().getAspRat());
   Iv ivx=ca.getGraph().getIvX();
   Iv ivy=ca.getGraph().getIvY();
   V<R> vx=ivx.centers(nx);
   V<R> vy=ivy.centers(ny);
   cpmassert(vx.dim()==nx,loc);
   cpmassert(vy.dim()==ny,loc);
   R_Matrix val(ny,nx); // indexed like a image matrix
   Z i,j;
   C xij;
   Z count=0;
   R nxInv=R(1)/nx;
   for (i=1;i<=nx;i++){
      xij[1]=vx[i];
      for (j=1;j<=ny;j++){
         xij[2]=vy[j];
         Spc sij=ca.inverse(xij,1.);
         val[ny-j+1][i]=operator()(sij);
      }
      count++;
      if (cpmverbose>1) cpmprogress("mark-loop",R(count)*nxInv);
   }
   Iv riv;
   if (adjust){
      Iv ivRange=rsp.getRangeX();
      X2<R_Matrix,Iv> xri=val.adjustRange(ivRange);
      val=xri.c1();
      riv=xri.c2();
   }
   Image<R> img(val.toVV());
   if (adjust){
      R vL=riv.inf();
      R vU=riv.sup();
      if (cpmverbose>0){
         R xT=0.25, yT=0.1;
         Color cT(Z(255),Z(0),Z(0));
         Word wT=cpm(vL)&" <=val<= "&cpm(vU);
         ca.getGraph().addText(wT,xT,yT,cT);
      }
   }
   Z intPolMet=1; // bilinear
   img.mark(ca.getGraph(),rsp,intPolMet);
   CPM_MZ
}

ScalarField::ScalarField(const V<Spc>& x, const V<R>& y):
Base(new CpmAlgorithms::AveragingInterpolator<Spc,R>(x,y)){}

////////////////// class SourceOfVectorField ///////////////////////////

void SourceOfVectorField::add(const PointCharge& s)
{
   R qs=s.get_q();
   q+=qs;
   d+=(s.get_x()-x)*qs;
   updateQD();
}

void SourceOfVectorField::add(const ElcDipole& s)
{
   d+=s.get_d();
   updateD();
}

void SourceOfVectorField::add(const MagDipole& s)
{
   j+=s.get_j();
   updateJ();
}

void SourceOfVectorField::add(const PointMass& s)
{
   R ms=s.get_m();
   m+=ms;
   updateM();
}

void SourceOfVectorField::add(const SourceOfVectorField& s)
{
   R qs=s.get_q();
   q+=qs;
   d+=(s.d+(s.get_x()-x)*qs);
   j+=s.j;
   m+=s.m;
   update();
}

bool SourceOfVectorField::prnOn( ostream& str )const
{
   cpmwt("SourceOfVectorField");
   cpmp(q);
   cpmp(d);
   cpmp(j);
   cpmp(m);
   cpmp(r);
   cpmp(x);
   return true;
}

bool SourceOfVectorField::scanFrom( istream& str )
{
   cpms(q);
   cpms(d);
   cpms(j);
   cpms(m);
   cpms(r);
   cpms(x);
   update();
   return true;
}

V<SourceOfVectorField> SourceOfVectorField::redistribute(
   const V<SourceOfVectorField>& sl, const V<Spc>& pl)
   // sl: source list, pl: point list
{
   Z mL=2;
   static Word loc("SourceOfVectorField::redistribute(...)");
   CPM_MA
   Z sld=sl.dim(),pld=pl.dim();
   Z i,k; // j is attribute!
   Vo<R> distances(pld);
   V<SourceOfVectorField> res(pld);
   R_Vector vr(sld);
   for (i=1;i<=sld;i++) vr[i]=sl[i].get_r();
   R rMean=vr.mean();
   for (k=1;k<=pld;k++) res[k]=SourceOfVectorField(pl[k],rMean);
   for (i=1;i<=sld;i++){
      Spc xi=sl[i].get_x();
      for (k=1;k<=pld;k++) distances[k]=(pl[k]-xi).absSqr();
         // absSqr() faster than abs()
      Z iMin=distances.indInf();
         // indInf very efficient, no ordering
      res[iMin].add(sl[i]);
   }
   CPM_MZ
   return res;
}

V<SourceOfVectorField> SourceOfVectorField::redistribute2(
   const V<SourceOfVectorField>& sl, const V<Spc>& pl,
   const V<Z>& whereToGo)
{
   Z mL=3;
   static Word loc("SourceOfVectorField::redistribute2(...)");
   CPM_MA
   Z sld=sl.dim(),pld=pl.dim();
   Z i,k; // j is attribute!
   V<SourceOfVectorField> res(pld);
   R rMean=SysControl::singRange; // OK since not of
      // any influence in the application in cpmpardynsys.h
   for (k=1;k<=pld;k++) res[k]=SourceOfVectorField(pl[k],rMean);
      // here no index mistake can happen
   for (i=1;i<=sld;i++){
      Z wi=whereToGo[i];
      cpmassert(wi>0,loc);
      cpmassert(wi<=pld,loc);
         // in case that something went wrong with 'whereToGo'
      res[wi].add(sl[i]);
   }
   CPM_MZ
   return res;
}

///////  class PointMass //////////////////////////////////////////

bool PointMass::prnOn( ostream& str )const
{
   CpmRoot::writeTitle("PointMass",str);
   cpmp(m);
   cpmp(r);
   cpmp(x);
   return true;
}

bool PointMass::scanFrom( istream& str )
{
   cpms(m);
   cpms(r);
   cpms(x);
   return true;
}

Force PointMass::force( const PointMass& pc )const
// exact force to the following regularized potential
{
   if(SysControl::noInteractionMassToMass) return Vec();
   Vec xr=x-pc.x; // points from pc to *this
   R rr=xr.abs();
   if (rr>SysControl::intactRange) return Vec();
   R Fabs=kgrav*m*pc.m*cpmpow(r*r+rr*rr,R(-1.5));
   return Force(xr*Fabs);
}

R PointMass::pot( const PointMass& pc )const
{
   if(SysControl::noInteractionMassToMass) return 0;
   Vec xr=x-pc.x; // points from pc to *this
   R rr=xr.abs();
   if (rr>SysControl::intactRange) return 0;
   return kgrav*m*pc.m*cpmpow(r*r+rr*rr,R(-0.5));
}

VecG PointMass::gravitationalField(Spc const& x, const PointMass& pc)
{
   Vec xr=x-pc.x;
   R rr=xr.abs();
   rr=(rr<pc.r ? pc.r : rr);
   return VecG(xr*(pc.m*kgrav*cpmpow(rr,-3)));
}

R PointMass::gravitationalPotential(Spc const& x, const PointMass& pc)
{
   Vec xr=x-pc.x;
   R rr=xr.abs();
   rr=(rr<pc.r ? pc.r : rr);
   return pc.m*kgrav/rr;
}

FieldG PointMass::field()const
{
   return F1<Spc,PointMass,VecG>(*this)(gravitationalField);
}

ScalarField PointMass::fieldU()const
{
   return F1<Spc,PointMass,R>(*this)(gravitationalPotential);
}
