//? cpmforcesx.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*

   Description: Classes for sources of forces and related physical
   quantities.
   The decision, whether geometry is treated in 2 dimensions
   or three dimensions is made by macro substitutions for
   CPM_DIM and CPM_CAMERA. See cpmforces.h for the mechanism which
   defines these macros and which puts the quantities to be defined in
   the present file into namespaces CpmPhysics2 or CpmPhysics3 according
   to the status of macros CPM_DIMDEF2 and CPM_DIMDEF3.
   This is not a header file for its own but cpmforce.h is.
   One finds the same mechanism for a common definition of 2D and
   3D versions of physical and geometrical concepts throughout the
   C+- class system.
   In the 2D case, sources of fields never are asumed to have an
   infinite extension in the direction perpendicular to the
   representation plane. Instead, only all vectorial quantities
   (such as velocities or directions of dipoles) are assumed to
   be parallel to this representation plane. In particular, our
   point charges are point charges generating fields decreasing as
   r^-2 (not logarithmic !) also in 2D; they thus are no charged thin
   wires.
*/

using namespace CpmRoot;
using CPM_DIM::AxVec;
using CPM_DIM::Vec;
using CPM_DIM::Spc;
using CPM_DIM::Group;
using CPM_DIM::Plane;
using CPM_DIM::SubsetOfPlane;
using CPM_DIM::SubsetOfSpace;
using CPM_DIM::RefSys;
using CPM_DIM::Rdim;
using CPM_CAMERA::Camera;

using CpmProperties::SysControl;
using CpmGraphics::Graph;
using CpmImaging::Color;
using CpmImaging::Palette;
using CpmImaging::GREEN;
using CpmRootX::B;
using CpmRootX::R1;
using CpmLinAlg::R2;
using CpmLinAlg::R3;
using CpmFunctions::F;
using CpmFunctions::F1;
using CpmArrays::V;
using CpmArrays::X2;
using CpmArrays::S;
using CpmArrays::Iv;

//doc namespace CpmPhysicsx{

//////////////////////// class VecE ////////////////////////////////////
// electrical field strength E (V/m in SI) (electric force/charge)
// There is no need to introduce D-Fields as united with E-fields
// in a way analogous to the magnetic case

   class VecE: public Vec{ // electrical field strength E

   public:
      VecE():Vec(){}
      explicit VecE(const Vec& a):Vec(a){}
      explicit VecE(V<R> const& vr):Vec(vr){}
      explicit VecE(Rdim const& vr):Vec(vr){}
      Word nameOf()const{ return Word("VecE");}
   };

//////////////////////// class VecG /////////////////////////////////////
// gravitational field strength (acceleration = gravity force/mass)

   class VecG: public Vec{ // gravitational field strength
   public:
      VecG():Vec(){}
      explicit VecG(const Vec& a):Vec(a){}
      explicit VecG(V<R> const& vr):Vec(vr){}
      explicit VecG(Rdim const& vr):Vec(vr){}
      Word nameOf()const{ return Word("VecG");}
   };

////////////////////// class VecBH /////////////////////////////////////
// Magnetic induction B (V.s.m^-2 in SI) and magnetic field strength H
// (A/m in SI) in a single class.
// This has a particular behavior in 2D since we need here a field in the
//
// system plane to study chain formation of magnetized particles
// and also a perpendicular field for studying the influence of the
// Lorentz force. Thus, also in 2 dimensions, the field has 3 components.
// In 2D thus the dimension of the quantities say that H as a 2-component
// object is typed Vec, wheras the perpendicular field is typed AxVec.
// Also in 3D it is useful to have H and B available at once. But in this
// case these two are proportional to each other.
// The situation that in 2D, both av_ and v_ are different from zero
// is not strictly physical. The present interpretation of it is that
// from av_ one computes the Lorentz force and from v_ one computes
// the force and torque on magnetic dipoles.

class VecBH{//magnetic induction B and magnetic field strength H combined
      AxVec av_; // B (relative to unit Tesla)
      Vec v_; // H (relative to unit A/m)
      VecBH(AxVec const& av, Vec const& v):av_(av),v_(v){}
         // private constructor
      void ini(R3 const& a, Word const& ind);
         // compact initialization tool. All components are
         // understood as referring all to the SI unit T
         // or all to the SI unit A/m. The first case is
         // understood for ind="B", the second one for
         // ind="H".
   public:
      typedef VecBH Type;
      CPM_IO
      VecBH():av_(),v_(){}
         // zero field
      VecBH(Vec const& v, Word const& ind);
         // the components of v will be interpreted as
         // H components if ind=="H" and as B components if
         // ind=="B".
      VecBH(V<R> const& vr, Word const& ind ){ ini(R3(vr),ind);}
         // the numbers vr[i] get interpreted as having the dimension of
         // H if ind=="H" and as B components if ind=="B".
      VecBH(R3 const& r3, Word const& ind ){ ini(r3,ind);}
         // the numbers r3[i] get interpreted as having the dimension of
         // H if ind=="H" and as B components if ind=="B".
      Word nameOf()const{ return Word("VecBH");}
      Vec v()const{ return v_;}
         // H-field in umits A/m ('Ampere windings per meter')
      AxVec av()const{ return av_;}
         // B-field in units T (Tesla)
      Z dim()const{ return 3;}
      VecBH neg()const{ return VecBH(-av_,-v_);}
      VecBH operator*(R r)const{ return VecBH(av_*r,v_*r);}
      VecBH& operator-=(VecBH const& x)
         { av_-=x.av_; v_-=x.v_;  return *this;}
      VecBH& operator+=(VecBH const& x)
         { av_+=x.av_; v_+=x.v_;  return *this;}
      VecBH& operator*=(R r){ av_*=r; v_*=r;  return *this;}
      VecBH& operator*=(Group const& g){ av_*=g; v_*=g;  return *this;}
      Vec cross(Vec const& w)const{ return av_.cross(w);}
         // Defined such that the Lorentz force on a charge q of
         // velocity w is cross(w)*(-q)
   };

///////////////////////// class Force //////////////////////////////////

   struct Force{  // force
      // Example of a compact definition of an auxiliary class.
      Vec f;
      typedef Force Type;
      typedef R ScalarType;
      CPM_LINEAR
      CPM_ORDER

      Force():f(){}

      Force(const Vec& force):f(force){}
      Word nameOf()const{ return Word("Force");}

      Vec get_f()const{ return f;}

      void set_f(const Vec& f_){ f=f_;}
      void addForce(const Force& f_){ f+=f_.f;}

      Force& operator *=(const AxVec& d)
      { f*=d;  return *this;}

      Force& operator *=(const Group& g)
      { AxVec d=g.rot(); f*=d; return *this;}

      bool isVal()const{ return f.isVal();}

   };

   inline Force& Force::operator +=(const Force& a)
      { f+=a.f; return *this;}
   inline Force& Force::operator -=(const Force& a)
      { f-=a.f;  return *this;}
   inline Force& Force::operator *=(const R& a)
      { f*=a; return *this;}
   inline Force& Force::neg()
      { f.neg();  return *this;}
   inline Z Force::dim()const
      {return f.dim();}

///////////////////////// class ForTor //////////////////////////////////

   class ForTor{ // force and torque combined
      // combining force and torque into a linear Space.
      // Example of a compact definition of an auxiliary class.
      // It turns out that this class is also needed for describing
      // the combination of velocity and angular velocity
      // and also for acceleration and angular acceleration.
      // Actually it is a representation of the Lie-Algebra of Group
      Vec f;
      AxVec t;
   public:
      typedef ForTor Type;
      typedef R ScalarType;
      CPM_LINEAR
      CPM_ORDER

      ForTor():f(),t(){}

      ForTor(const Force& f_, const Vec& v):
         f(f_.f),t(v.cross(f_.f)){}
         // Let the force f_ act on point pAct, then let the torque
         // refer to point pRef, then ForTor(f_,pAct-pRef) is the
         // object of type ForTor which corresponds to the situation

      ForTor(const Force& force, const AxVec& torque):f(force.f),
         t(torque){}

      ForTor(const V<X2<Force,Spc> >& vx, const Spc& sp);
         // combines forces, that are given together with their point of
         // action into a resulting total force and a torque with respect
         // to point sp

      Word nameOf()const{ return Word("ForTor");}

      void add(const Force& f_, const Spc& pAct, const Spc& pRef)
         // adds to *this the contribution of force f_ acting on point
         // pAct under the assumption that the torque refers to the
         // reference point pRef. One could think that a better notion of
         // ForTor should include the reference point for the torque
         // among the data members. Will not be carried out, though.
      { f+=f_.f; t+=(pAct-pRef).cross(f_.f);}

      Vec get_force()const{ return f;}
      AxVec get_torque()const{ return t;}

      const Vec& c1()const{ return f;}
      const AxVec& c2()const{ return t;}
         // semantically neutral access to components

      void set_force(const Vec& f_){ f=f_;}
      void addForce(const Force& f_){ f+=f_.f;}
      void addForce(const Vec& v){ f+=v;}
      void addTorque(const AxVec t_){t+=t_;}
      void set_torque(const AxVec& t_){ t=t_;}

      ForTor& operator *=(const AxVec& d)
      { f*=d; t=t.adjointAction(d); return *this;}
         // notice that t*=d would be a composition of rotations
         // which can't be the right thing
            // 't=' was missing till 2005-02-14

      ForTor& operator *=(const Group& g)
      { AxVec d=g.rot(); f*=d; t=t.adjointAction(d); return *this;}
            // 't=' was missing till 2005-02-14

      bool isVal()const{ return f.isVal()&&t.isVal();}

   };

   inline ForTor& ForTor::operator +=(const ForTor& a)
      { f+=a.f; t+=a.t; return *this;}
   inline ForTor& ForTor::operator -=(const ForTor& a)
      { f-=a.f; t-=a.t; return *this;}
   inline ForTor& ForTor::operator *=(const R& a)
      { f*=a; t*=a; return *this;}
   inline ForTor& ForTor::neg()
      { f.neg(); t.neg(); return *this;}
   inline Z ForTor::dim()const
      {return t.dim()+f.dim();}

/////////////////////// class FieldSpace /////////////////////////////////

   class FieldSpace{
//electric, magnetic, gravitational fieldstrengths into one linear space
      VecE E_;
      VecBH BH_;
      VecG G_;
   public:
      typedef FieldSpace Type;
      typedef R ScalarType;
      CPM_LINEAR
      CPM_IO
      FieldSpace():E_(),BH_(),G_(){}
      FieldSpace(VecE const& E, VecBH const& H, VecG const& G=VecG()):
         E_(E),BH_(H),G_(G){}

      Word nameOf()const{ return Word("FieldSpace");}

// my standard notation for heterotypic components. I'll always
// use this correspondence:
// 1: electric, 2: magnetic, 3: gravitational
      VecE& c1(){ return E_;}
      const VecE& c1()const{ return E_;}
      VecBH& c2(){ return BH_;}
      const VecBH& c2()const{ return BH_;}
      VecG& c3(){ return G_;}
      const VecG& c3()const{ return G_;}
      FieldSpace& operator *=(const Group& g)
         { E_*=g; BH_*=g; G_*=g; return *this;}
      FieldSpace& operator+=(VecE const& v){ E_+=v; return *this;}
      FieldSpace& operator+=(VecBH const& v){ BH_+=v; return *this;}
      FieldSpace& operator+=(VecG const& v){ G_+=v; return *this;}
   };

   inline FieldSpace& FieldSpace::operator +=(FieldSpace const& a)
   { E_+=a.E_; BH_+=a.BH_; G_+=a.G_; return *this;}

   inline FieldSpace& FieldSpace::operator -=(FieldSpace const& a)
   { E_-=a.E_; BH_-=a.BH_; G_-=a.G_; return *this;}

   inline FieldSpace& FieldSpace::operator *=(R const& a)
   {  E_*=a; BH_*=a; G_*=a;  return *this;}

   inline FieldSpace& FieldSpace::neg()
      { E_.neg(); BH_.neg(); G_.neg(); return *this;}

   inline Z FieldSpace::dim()const{return E_.dim()+BH_.dim()+G_.dim();}

/////////////////// class ScalarField ////////////////////////////////////

class ScalarField:public CpmFields::ScalarField<Spc,R,Group>{
      // scalar field with group action
      // since we need functions that cannot yet be defined in
      // CpmFields::ScalarField<Spc,R,Group>, a typedef is not
      // sufficient.
      typedef CpmFields::ScalarField<Spc,R,Group> Base;
   public:
      ScalarField(const Base& f):Base(f){}
      ScalarField(const F<Spc,R>& f):Base(f){}

      explicit ScalarField(const SubsetOfSpace& sos):Base(sos.getF()){}
         // new constructor which mainly helps the visalization of
         // SubsetOfSpace objects via the well-working function
         // void ScalarField::mark(const Camera& sc)const;

      ScalarField(const V<Spc>& x, const V<R>& y);
         // General interpolation method for functions defined on a
         // metric
         // space and taking values in a linear space

      explicit ScalarField(FieldSpace const& fs, Spc const& p=Spc());
         // Potential of a homogeneous VectorField of (constant) strength
         // fs. The component selector c (1: electric, 2: magnetic, 3:
         // gravitation) determines the field under consideration.

      ScalarField():Base(){}

      Word nameOf()const{ return Word("ScalarField");}

      friend Base down(const ScalarField& s);//{ return s;}
         // works !!!

      void periodize(const Spc& x1, const Spc& x2);

      void mark(const Camera& sc)const;
         // Displays the scalar field under the control of the
         // data element typed Response of sc. Implemented by
         // means of image class Image<R> which supports much
         // higher dynamic range than Img24. This function
         // does not simply mark but also displays automaticly.
         // In the way I use it presently it does not cooperate well
         // with the frame number indication in movie replay.
         // This is considered a minor aesthetical problem.

      static ScalarField sum(const V<ScalarField>& vf);
    //  { return Base::sum(vf(F<ScalarField,Base>(down)));}
         // a direct implementation as in Base and avoiding .f was not
         // faster (even a bit slower) in a test with summing over the
         // field of several hundred particles.
   };

   inline CpmFields::ScalarField<Spc,R,Group>
      down(const ScalarField& s){ return s;}
         // works !!!

   inline ScalarField ScalarField::sum(const V<ScalarField>& vf)
   {
      return
      CpmFields::ScalarField<Spc,R,Group>::
         sum(vf(F<ScalarField,Base>(down)));
   }

/////////////////// class FieldE ///////////////////////////////////////

   class FieldE: public CpmFields::Field<Spc,VecE,R,Group>{
      // E-field with group action
      // since we need functions that cannot yet be defined in
      // CpmFields::Field<Spc,VecE,R,Group>, a typedef is not
      // sufficient for the electric field. Could well be
      // that a similar need would arise for the magnetic
      // field too
      typedef CpmFields::Field<Spc,VecE,R,Group> Base;
   public:
      FieldE(const Base& f):Base(f){}
      FieldE():Base(){}
      Word nameOf()const{ return Word("FieldE");}
      void periodize(const Spc& x1, const Spc& x2);
   };

/////////////////// class FieldBH //////////////////////////////////////

   typedef CpmFields::Field<Spc,VecBH,R,Group> FieldBH;

///////////////////// class FieldG //////////////////////////////////////

   typedef CpmFields::Field<Spc,VecG,R,Group> FieldG;

/////////////////// class VecField /////////////////////////////////////

   typedef CpmFields::Field<Spc,Vec,R,Group>  VecField;

   void markVecField(const VecField& f, const Camera& sc,
      R& scaleFactor,
      Z gridPointsHorizontal=10,
      Z gridPointsVertival=6,
      const Color& color=GREEN);
   // scaleFactor is a factor which determines how long the field vectors
   // will be represented on screen. A negative value will not be
   // utilized as input but as a signal to determine a suitable value for
   // this parameter. After function call the variable scaleFactor will
   // hold a proper value that may be used in subsequent calls.

   VecField grad(const ScalarField&, R h=SysControl::singRange);
      // gradient as a vector field

   Vec gradient(const Spc& x , const ScalarField& f,
      const R& h=SysControl::singRange);
      // gradient as a vector-valued function of position

//////////// class VectorField ///////////////////////////////////////////

// Multicomponent fields can be implemented as arrays of mono-component
// fields or as fields with multi-component values.
// Here we use the second method which looks more natural to me. The
// first one would presumably be slightly easier to implement.

   class VectorField:public CpmFields::Field<Spc,FieldSpace,R,Group>{
   // FieldSpace-valued field with group action

      static R scaleFactorE; // controls the graphical representation
      static R scaleFactorM;
      static R scaleFactorG;

   public:
      typedef CpmFields::Field<Spc,FieldSpace,R,Group> Base;
      VectorField(){}
      VectorField(const Base& f):Base(f){}
      VectorField(const F<Spc,FieldSpace>& f):Base(f){}
      explicit VectorField(FieldE const& fe,
         FieldBH const& fh=FieldBH(), FieldG const& fg=FieldG());
         // creating a vector field by prescribing the field components
      VectorField(V<Spc> const& x, V<FieldSpace> const& y);
         // General interpolation method for functions defined on a
         // metric
         // space and taking values in a linear space

      VectorField(const FieldSpace& y):Base(y){}
         // constructor for a constant field, with y as the field value
      Word nameOf()const{ return Word("VectorField");}
      FieldE c1()const;
      FieldBH c2()const;
      FieldG c3()const;
         // components
      VecE E(const Spc& x)const{ return (operator()(x)).c1();}
      VecBH BH(const Spc& x)const{ return (operator()(x)).c2();}
      VecG G(const Spc& x)const{ return (operator()(x)).c3();}

      friend Base downV(const VectorField& s);//{ return s;}

      static void newSclE(){ scaleFactorE=-1;}
      static void newSclM(){ scaleFactorM=-1;}
      static void newSclG(){ scaleFactorG=-1;}

      static void newScl()
      { scaleFactorE=-1; scaleFactorM=-1; scaleFactorG=-1;}
         // managing graphical representation newScl = new scale
      VecField operator[](Z i)const;
         // returns the electric, magnetic, and gravitational field part
         // as a quantity of uniform type. Notice that the implementation
         // in cpmforces.h would be syntactically correct also for a
         // return value VecField&. Calling this modified function causes
         // severe runtime errors (would be a reference to a local
         // variable, which my present compiler does not even warn)
      void setU(const ScalarField&);
         // the electrical part of the field is set such that the
         // argument of the function is the potential of it
      void setE(const FieldE&);
         // the electrical part of the field is set such that it equals
         // the argument of the function

      void setBH(const FieldBH&);
         // the magnetic part of the field is set such that it equals the
         // argument of the function
      void setG(const FieldG&);
         // the gravitational part of the field is set such that it
         // equals the argument of the function

      void add(const FieldE&);
      void add(const FieldBH&);
      void add(const FieldG&);
         // adding is mostly more convenient than setting; no
         // different function names needed

      void mark(const Camera& sc)const;
         // The details of the display mode are controlled by
         // SysControl::nxElectricField
         // SysControl::nyElectricField
         // SysControl::markElectricField
         // SysControl::nxMagneticField
         // SysControl::nyMagneticField
         // SysControl::markMagneticField
         // SysControl::nxGravityField
         // SysControl::nyGravityField
         // SysControl::markGravityField

      static VectorField sum(const V<VectorField>& vf);
      //{ return Base::sum(vf(F<VectorField,Base>(downV)));}
    };

    inline CpmFields::Field<Spc,FieldSpace,R,Group>
       downV(const VectorField& s){ return s;}

   inline VectorField VectorField::sum(const V<VectorField>& vf)
   {
      return CpmFields::Field<Spc,FieldSpace,R,Group>::
         sum(vf(F<VectorField,Base>(downV)));
   }

////////////////////// class PointMass ///////////////////////////////////

   class PointMass{ // point masses
      // describes point masses and the forces they feel and they create

      R m;
         // mass (kg)

      R r;
         // a regularization radius
         // Notice that interacting point-masses have to have all
         // the same regularization radius, otherwise momentum
         // conservation does not hold. So it is wise to use
         // SysControl::singRange as this value

      Spc x;
         // position

   public:
      typedef PointMass Type;
      CPM_IO
      Word nameOf()const{ return Word("PointMass");}
      PointMass():m(0),r(SysControl::singRange){}
         // x gets initilized as zero vectors automatically

      PointMass(R m_, const Spc& x_, R r_=SysControl::singRange):
         m(m_),x(x_),r(r_){}

      R get_m()const{ return m;}

      void set_m(R m_){ m=m_;}

      R get_r()const{ return r;}

      void set_r(R r_){ r=r_;}

      Spc get_x()const{ return x;}

      void set_x(const Spc& x_){ x=x_;}

      PointMass& operator *=(R s){m*=s; return *this;}

      PointMass& operator *=(const Group& a){x*=a; return *this;}
         // action of the basic symmetry group on a point mass.

      PointMass& operator *=(const AxVec& a){a; return *this;}
         // action of rotations on a point mass. The quantity
         // is scalar and so the action is trivial. Included here for
         // uniformity of the interface with similar classes

      PointMass& operator +=(const Vec& a){x+=a; return *this;}
         // action of shift on a point mass.

      Force force( const VecG& G )const{ return Force(m*G);}
         // force on *this acted on by an gravitational field strength G

      Force force( const F<Spc,VecG>& fG)const{ return Force(m*fG(x));}
         // returns the force on the point charge as a result of the
         // interaction with the argument field, which is a vector field
         // giving the spatial distribution of the gravitational field
         // strength (G)

      Force force( const PointMass& pm )const;
         // force on *this caused by a point mass pm

      R pot( const PointMass& pm )const;
         // potential energy of a pair due to gravity

      R potEnr(const VecG& G)const{ return (Spc()-x|G)*m;}
         // potential energy of *this with respect to the homogeneous
         // gravitational field strength G

      FieldG field()const;
         // returns the gravitational field generated by the source *this

      VecG eval(const Spc& x)const
      { return gravitationalField(x,*this);}

      R evalU(const Spc& x)const
      { return gravitationalPotential(x,*this);}

      ScalarField fieldU()const;
         // returns the gravitational potential as a field

      static VecG gravitationalField(const Spc& x, const PointMass& pc);
         // the main physical law as a plain function

      static R gravitationalPotential(const Spc& x, const PointMass& pc);
         // the main physical law as a plain function
   };

/////////////////// class PointCharge ///////////////////////////////////

   class PointCharge{ // point charges
      // describes point charges and the forces they feel and they
      // create.
      // Actually a certain type of radially symmetric continuous
      // charge, that only for vanishing regularization radius r
      // becomes a point charge.

      R q;
         // electric charge (A*s)

      R r;
         // a regularization radius. This can be interpreted as
         // the approximate radius of a radially symmetric
         // diffuse charge distribution depending on distance
         // d like (d*d+r*r)^(-5/2)
         // see my Mathematica file coulombmod2.nb

      Spc x;
         // position

   public:
      typedef PointCharge Type;
      CPM_IO
      Word nameOf()const{ return Word("PointCharge");}
      PointCharge():q(0),r(SysControl::singRange){}
         // x gets initilized as zero vectors automatically

      PointCharge(R q_, const Spc& x_, R r_=SysControl::singRange):
         q(q_),x(x_),r(r_){}

      R get_q()const{ return q;}

      void set_q(R q_){ q=q_;}

      R get_r()const{ return r;}

      void set_r(R r_){ r=r_;}

      Spc get_x()const{ return x;}

      void set_x(const Spc& x_){ x=x_;}

      void invertCharge(){ q=-q;}

      PointCharge mirror(const Plane& pl)const;
         // returns a PointCharge which is the mirror charge created by
         // *this in the presence of a perfectly conducting plane

      PointCharge mirror(const Plane& pl, R eps)const;
         // returns a PointCharge which is the mirror charge created by
         // *this in the presence of a dielectric plane of
         // relative dielectric constant eps. The halfspace
         // where *this is located is assumed to be
         // characterized by relative dielectric constant equal
         // to 1. eps>>1 approximates the behavior of the previous
         // function.

      PointCharge& operator *=(R s){q*=s; return *this;}

      PointCharge& operator *=(const Group& a){x*=a; return *this;}
         // action of the basic symmetry group on a electric point
         // charge.

      PointCharge& operator *=(const AxVec& a){a; return *this;}
         // action of rotations on a electric point charge. The quantity
         // is scalar and so the action is trivial. Included here for
         // uniformity of the interface with similar classes

      PointCharge& operator +=(const Vec& a){x+=a; return *this;}
         // action of shift on a electric point charge.

      void add(const PointCharge& pc){ q+=pc.get_q();}

      Force force(VecBH const& BH, Vec const& v)const;
         // Lorentz force force on *this acted on by an magnetic field
         // vector BH. The velocity of the PointCharge is given as v.

      Force force( const VecE& E )const{ return Force(q*E);}
         // force on *this acted on by an electric field strength E

      Force force( const F<Spc,VecE>& fE)const{ return Force(q*fE(x));}
         // returns the force on the point charge as a result of the
         // interaction with the argument field, which is a vector field
         // giving the spatial distribution of the electric field
         // strength (E)

      Force force( PointCharge const& pc )const;
         // force on *this caused by a point charge pc

      Force forceMag(Vec const& v, PointCharge const& pc,
         Vec const& vpc )const;
         // Lorentz force on *this , moving with velocity v,
         // caused by the magnetic field,
         // generated by point charge pc, moving with velocity vpc.
         // No retardation taken into account.

      R pot( const PointCharge& pc )const;
         // potential energy of a pair due to Coulomb interaction

      R potEnr(const VecE& E)const{ return (Spc()-x|E)*q;}
         // potential energy of *this with respect by the homogeneous
         // electric field of strength E

      Force force( const V<PointCharge>& vpc )const;
         // force on *this caused by the family vpc of point charges

// Interaction with dielectric half-space and with a dielectric layer
// which is made perfectly conducting on the back side (like a photo
// conductor film in electro-photography. Explanations:

// On image charges

   // A electrophotographic photoconductor film is modelled here as a
   // dielectric layer of relative dielectric constant eps and thickness
   // d behind which there is a perfectly conducting half-space (a layer
   // of any thickness has the same effect, if conductivity is perfect).
   // A point charge q placed near the surface of the film will feel a
   // force resulting from influenced electric dipoles in the dielectric
   // layer and re-distributed charges in the conductor. It can be shown
   // (UM/PNP 99-11-17/1-4) that this force is approximately (Thomas
   // Dera, 99-12-6, pointed out that this holds not exactly as I thought
   // before. The condition on continuity of normal component of D
   // is not satisfied exactly. See my Mathematica document pnp.nb )
   // to the force exerted by a charge q1=-q*(eps-1)/(eps+1) located at
   // the mirror image position resulting from reflection at the PC
   // surface and a second charge q2=-q*2/(eps+1) located at the mirror
   // image position resulting from reflection at the surface of the
   // conducting half-space. Notice q+q1+q2=0. The combined field of
   // charge q together with the two image charges is the actual electric
   // field generated by q under the environmental condition posed by the
   // presence of the PC film. This implies that, in the presence of a PC
   // film, each particle feels from a partner particle the direct
   // Coulomb force from the partner particle and the Coulomb forces from
   // the mirror particles of the partner particle. It also feels a force
   // from 'itself' i.e. from its own mirror particles. Since in the
   // present application, the PC is treated as elastically deformable,
   // we
   // have to consider the possibility that the point charge q is placed
   // below the undisturbed PC surface. In this case we define the first
   // image charge one charge regularization radius (see PointCharge::r)
   // ahead of q

// Consider a system V<PointCharge> pcs of n point charges near a PC film
// where the latter is described by Plane pl, R eps, R d.
// Then the force felt by charge i0 is given by f from the following
// algorithm
//
//      Force f=pcs[i0].force(pl,eps,d);
//      for (i=1;i<=n;i++){
//         if (i==i0) continue;
//         f+=pcs[i0].force(pcs[i],pl,eps,d);
//      }
// It thus can be considered consistently as a sum over all particles -
// including the particle feeling the force - we thus have non-trivial
// self-interaction. The PC film appears as a parameter in the particle-
// particle force law and does not appear as an interaction partner (as
// in the erroneous Ansatz: force to a particle = force to that particle
// from the PC + force to that particle from all the other particles.

      Force force(const Plane& pl, R eps )const;
         // force on *this caused by the interaction of *this and a
         // half-space determined by the Plane pl and filled by
         // dielectric
         // matter of relative constant of dielectricity eps.

      Force force(const Plane& pl, R eps, R d)const;
         // force on *this caused by a dielectric layer (thickness d) and
         // a perfectly conducting half-space

      FieldE field()const;
         // returns the electric field generated by the source *this

      FieldBH fieldBH(const Vec& v)const;
         // returns the magnetic field generated by the source *this
         // if this moves with velocity v. No retardation is taken into
         // account

      VecE eval(const Spc& x)const
         { return electrostaticField(x,*this);}

      R evalU(const Spc& x)const
         { return electrostaticPotential(x,*this);}

      ScalarField fieldU()const;
         // returns the electric potential as a field

      static VecE electrostaticField(const Spc& x,const PointCharge& pc);
         // the main physical law as a plain function
         // see my Mathematica file coulombmod2.nb for a
         // consistent physical interpretation of the regularization
         // as a non-singular charge distribution.

      static VecBH magneticField(const Spc& x, const PointCharge& pc,
         const Vec& v);
         // the main physical law as a plain function
         // Magnetic field as generated by a moving charge
         // no retardation taken into account.

      static VecE electrostaticField(const Spc& x,
         const V<PointCharge>& vpc);
         // the main physical law as a plain function

      static R electrostaticPotential(const Spc& x,
         const PointCharge& pc);
         // the main physical law as a plain function

      static R electrostaticPotential(const Spc& x,
         const V<PointCharge>& pc);
   };

///////////////// class PatchCharge ///////////////////////////////////

   class PatchCharge{ // charged planar rectangle (patch)
      // describes a charged planar rectangle (patch) one side of which
      // is of length w and the other of length depth.
      // Till ~ 03-02-xx the quantity depth was not present as a data
      // member since it had the 'fixed value infinity'.
      // Some of the commentaries may still refer to this situation.
      // Since 03-04-09 also formula for fieldU is generalized
      // (surprisingly simple) for this general case.

      RefSys rs;
         // this is a reference frame relative to which the charged area
         // is a subset of the 1-3-plane, given by the condition
         // -w/2 <= x[1] <=w/2, -depth/2 <= x[3] <= depth/2
         // (i.e. the charged patch lies in the 1-3-plane, symmetrical
         // arround the line x[1]=0.
         // Notice that in 2D we can evaluate fields and potentials
         // only in the 1-2-plane but the field generating physical
         // structure (the charged patch) nevertheless has an extension
         // in 3-direction. We prefer here to have the patchcharge in the
         // 1-3 plane so that always the 2-axis (which is the vertical
         // axis both in the 3D and the 2D version of the PaDyAn-model)
         // is perpendicular to the charged rectangle. Since the general
         // theory in my Mathematica files is developed for the rectangle
         // lying in the 1-2-plane there is a slight complication in the
         // implementation that uses coordinates (x,y,z) corresponding to
         // axes (1,3,-2) and does so for 2D and 3D simulaneously.

      R w;
         // width of the patch

      R depth;
         // extension in 3-direction

      R sigma;
         // electric charge areal density
         // total electrical charge thus is w*depth*sigma

      R r;
         // a regularization radius

   public:
      static B makeFieldFromFieldU;
      typedef PatchCharge Type;
      CPM_IO
      Word nameOf()const{ return Word("PatchCharge");}
      PatchCharge():w(0),depth(0),sigma(0),r(SysControl::singRange){}
      PatchCharge(const RefSys& rs_, R w_, R depth_,
         R sigma_=0, R r_=SysControl::singRange):
         rs(rs_),w(w_),depth(depth_),sigma(sigma_),r(r_){}

      R get_r()const{ return r;}
      void set_r(R r_){ r=r_;}

      R get_w()const{ return w;}
      R set_w(R w_){w=w_;}

      void setDepth(R d){ depth=d;}
      R getDepth(){ return depth;}

      void invertCharge(){ sigma=-sigma;}
      R get_sigma()const{ return sigma;}
      void set_sigma(R sigma_){ sigma=sigma_;}

      RefSys get_rs()const{return rs;}
         // read only, manipulation of rs should be confined to calling
         // operators *= and +=

      Spc get_x()const{  return rs.o();}
         // origin of the reference system

      V<Spc> range()const;
         // returns two points on the boundary of the patch
         // the connecting line of which is orthogonal to this
         // boundary

      V< X2<R2,Z> > corners()const;
         // Tool for organizing computation of the field in a way that
         // can be used as a building block in class MultiPatchCharge

      SubsetOfPlane getPatch()const;
         // returns the charged area as a object

      PatchCharge& operator *=(const Group& g){rs*=g; return *this;}
         // action of the basic symmetry group on a charged patch.

      PatchCharge& operator *=(const AxVec& av){rs*=av; return *this;}
         // transformation specialized to rotation

      PatchCharge& operator +=(const Vec& v){rs+=v; return *this;}
         // transformation specialized to translation

      V<PatchCharge> increase(const V<Vec>& vv)const;
         // returns a set of copies of *this, each shifted by
         // a vector which is a component of vv

      FieldE field()const;
         // returns the electric field generated by the source *this

      VecE eval(const Spc& x)const
         { return electrostaticField(x,*this);}

      R evalU(const Spc& x)const
         { return electrostaticPotential(x,*this);}

      ScalarField fieldU()const;
         // returns the electric potential as a field

      static VecE electrostaticField(const Spc& x,
         const PatchCharge& pc );
         // the main physical law as a plain function
         // no longer needed since 03-09-13 since now
         // we use FieldE = -grad(fieldU)

      static R electrostaticPotential(const Spc& x,
         const PatchCharge& pc);
         // the main physical law as a plain function, very
         // transparent implementation.

      void mark(const Camera& sc, Color c,  bool shAx=false)const;
         // if shAx==true the axes of RefSys.rs are shown too,
         // also in color c. Axes 1,2,3 can be discerned by the number of
         // structural elements they are made of.

      static R test();
         // Tests the consistence of computing the electrical field
         // directly or as a gradient of the electrical potential.
         // returns a number that had to be zero if no
         // numerical inaccuracies would be present.
         // More informations get written on cpmcerr so that
         // the return value should actually not be needed.
         // The tedious process of finally making sure that
         // electric field and electrical potential are
         // related in the right way triggered the idea to have
         // such a class test, which could be a model for
         // similar tests of other classes with logically involved
         // functionality. The present intention is not to create a
         // universal investigation tool with many parameters
         // but a randomly chosen object tested at random points.
   };

///////////////// class MultiPatchCharge ///////////////////////////////

   class MultiPatchCharge{
       // A collection of several patch charges.
       // I originally had in mind to model the electrostatic image of
       // a simple letter, say of an 'F' as an aggregation of patch
       // charges. It would be computationally wastefull (although trivial
       // to program) to use simply an object of type V<PatchCharge>.
       // A more economical method takes into account that the field of
       // each PatchCharge is the sum of four terms, one for each
       // corner. For adjacent patches these often cancel out and
       // a system of many mutually adjacent patches may result into
       // a number of surviving terms which is much smaller than the
       // number of patches in the system. Here we organize this
       // cancellation in the constructor code, so that computation
       // of the field can always start from the simplified situation
       // that the field is represented as a sum of only a few terms
       // coming from corners on the outer rim of the multi-patch system.
       // New version which is completely independent of PatchCharge.

      Iv xr_;
         // x-range of reference patch
         // x and y correspond to 1-direction and 2-direction
         // in rs_

      Iv yr_;
         // y-range of reference patch

      RefSys rs_;
         // System of reference, the 1-2-plane of which carries the
         // charged rectangular patches.

      R1 sigma_;
         // electric charge areal density
         // total electrical charge thus is
         // sigma_*xr_.abs()*yr_.abs()

      R1 r_;
         // a regularization radius

      V<R2> v_;
         // List of translation vectors, given by the components in rs_
      S< X2<R2,Z> > sx_;
         // set of corners and weights which have to be used in computing
         // the field.
       void ini_();
         // initializes sx_ from v_ and the first five data
       typedef MultiPatchCharge Type;
   public:
      CPM_IO
      Word nameOf()const{ return Word("MultiPatchCharge");}
      MultiPatchCharge(){}
      MultiPatchCharge(Iv const& xr, Iv const& yr, RefSys const& rs,
         R sigma, R r, V<R2> const& v):
         rs_(rs),xr_(xr),yr_(yr),sigma_(sigma),r_(r), v_(v){ini_();}


      R get_r()const{ return r_;}
      R get_sigma()const{ return sigma_;}
      void invChr_(){ sigma_*=-1;}
         //: invert charge
      RefSys get_rs()const{return rs_;}

      MultiPatchCharge& operator *=(const Group& g){rs_*=g;return *this;}
         // action of the basic symmetry group on a charged multi-patch.

      MultiPatchCharge& operator *=(const AxVec& av)
      {rs_*=av; return *this;}
         // transformation specialized to rotation

      MultiPatchCharge& operator +=(const Vec& v){rs_+=v; return *this;}
         // transformation specialized to translation

      R ePot(Spc const& x)const;

      ScalarField fieldU()const;
         // returns the electric potential as a field

       static R electrostaticPotential(Spc const& x,
          MultiPatchCharge const& pc){ return pc.ePot(x);}
         // the main physical law as a plain function

   };

///////////////// class MultiVolCharge ///////////////////////////////////

   class MultiVolCharge{
      // A collection of several rectangular charged cuboids.
      // See class MultiPatchCharge for details.

      Iv xr_;
         // x-range of reference cuboid
         // x, y, z correspond to 1-direction, 2-direction, 3-direction
         // in rs_

      Iv yr_;
         // y-range of reference cuboid

      Iv zr_;
         // z-range of reference cuboid

      RefSys rs_;
         // System of reference, to the axes of which all
         // cuboids under consideration are parallel.

      R1 sigma_;
         // electric charge density
         // total electrical charge thus is
         // sigma_*xr_.abs()*yr_.abs()*zr_.abs()

      R1 r_;
         // a regularization radius

      V<R3> v_;
         // List of translation vectors, given by the components in rs_
      S< X2<R3,Z> > sx_;
         // set of corners and weights which have to be used in computing
         // the field.
       void ini_();
         // initializes sx_ from v_ and the first five data
       typedef MultiVolCharge Type;
   public:
      CPM_IO
      Word nameOf()const{ return Word("MultiVolCharge");}
      MultiVolCharge(){}
      MultiVolCharge(Iv const& xr, Iv const& yr, Iv const& zr,
         RefSys const& rs, R sigma, R r, V<R3> const& v):
         rs_(rs),xr_(xr),yr_(yr),zr_(zr),sigma_(sigma),r_(r),
         v_(v){ini_();}


      R get_r()const{ return r_;}
      R reg(R y)const
      {
         R ay=cpmabs(y);
         if (ay<r_) return y>=0 ? r_ : -r_;
         else return y;
      }

      R get_sigma()const{ return sigma_;}
      RefSys get_rs()const{return rs_;}

      void invChr_(){ sigma_*=-1;}
         //: invert charge

      MultiVolCharge& operator *=(const Group& g){rs_*=g;return *this;}
         // action of the basic symmetry group on a charged multi-patch.

      MultiVolCharge& operator *=(const AxVec& av)
      {rs_*=av; return *this;}
         // transformation specialized to rotation

      MultiVolCharge& operator +=(const Vec& v){rs_+=v; return *this;}
         // transformation specialized to translation

      R ePot(Spc const& x)const;

      ScalarField fieldU()const;
         // returns the electric potential as a field

       static R electrostaticPotential(Spc const& x,
          MultiVolCharge const& pc){ return pc.ePot(x);}
         // the main physical law as a plain function
   };


//////////////////////// ElcDipole /////////////////////////////////////

   class ElcDipole{ // electric dipoles
      // Describes 'real' electric dipoles (i.e. the limit to zero
      // spacing of a finitely spaced pair of charges. This is the
      // charge distribution for which the function
      // electrostaticDipoleField(...) gives the correct field.

      Vec d;
         // electric dipole moment (A*s*m)

      R r;
         // a regularization radius

      Spc x;
         // position

   public:
      typedef ElcDipole Type;
      CPM_IO
      Word nameOf()const{ return Word("ElcDipole");}

      ElcDipole():r(SysControl::singRange){}
         // x and d get initilized as zero vectors automatically

      ElcDipole(const Vec& d_, const Spc& x_,
         R r_=SysControl::singRange):d(d_),r(r_),x(x_){}

      Vec get_d()const{ return d;}

      Spc get_x()const{ return x;}

      R get_r()const{ return r;}

      void add(const ElcDipole& ed){ d+=ed.d;}
         // adds the electric moment of a dipole

      void set_r(R r_){ r=r_;}

      void set_x(const Spc& x_){ x=x_;}

      void setDir(const Vec& v){ d.setDir_(v);}

      void setMoment(const Vec& v){ d=v;}

      Vec getMoment()const{ return d;}

      R abs()const{ return d.abs();}
         // absolute value of the electric moment

      ElcDipole& operator *=(R s){d*=s; return *this;}

      ElcDipole& operator *=(const Group& a){x*=a; d*=a; return *this;}
         // action of the basic symmetry group on a electric dipole.
         // Notice that x and d belong to different classes (Spc and Vec,
         // respectively) the group element a does different things to
         // them.

      ElcDipole& operator +=(const Vec& a){x+=a; return *this;}
         // action of a shift on a electric dipole.

      ElcDipole& operator *=(const AxVec& a){ d*=a; return *this;}
         // action of a rotation on a the electric dipole vector (place
         // held fixed)

      AxVec torque(const VecE& E)const;
         // torque on *this caused by the electric field strength E

      R potEnr(const VecE& E)const{ return (-d|E);}
         // potential energy of *this with respect by the homogeneous
         // electric field of strength E

      ForTor forceAndTorque( const F<Spc,VecE>& fE,
            R r=SysControl::singRange)const;
         // Force and torque on *this caused by the electric field fE.
         // The force depends on the the gradient of fE. Internally the
         // field is evaluated at two points, the distance of which is
         // 2*r

      ForTor forceAndTorque( const ElcDipole& md )const;
         // Force and torque on *this caused by a electric dipole md.

      FieldE field()const;
         // returns the electric field (E) generated by the source *this

      ScalarField fieldU()const;
         // returns the electric potential as a field

      VecE eval(const Spc& x)const
         { return electrostaticDipoleField(x,*this);}

      R evalU(const Spc& x)const
         { return electrostaticDipoleFieldU(x,*this);}

      static VecE electrostaticDipoleField(const Spc& x,
         const ElcDipole& md);
         // the main physical law as a plain function

      static R electrostaticDipoleFieldU(const Spc& x,
         const ElcDipole& md);
         // the main physical law as a plain function
   };

/////////////////// class MagPole ///////////////////////////////////////

   class MagPole{ // magnetic point charges
      // describes magnetic point charges and the forces they feel and
      // they create

      R q;
         // magnetic charge (V*s)

      R r;
         // a regularization radius

      Spc x;
         // position

   public:
      typedef MagPole Type;
      CPM_IO
      Word nameOf()const{ return Word("MagPole");}
      MagPole():q(0),r(SysControl::singRange){}
         // x gets initilized as zero vectors automatically

      MagPole(R q_, const Spc& x_ ,R r_=SysControl::singRange):
         q(q_),r(r_),x(x_){}

      R get_q()const{ return q;}

      void set_q(R q_){ q=q_;}

      R get_r()const{ return r;}
      void set_r(R r_){ r=r_;}

      Spc get_x()const{ return x;}

      void set_x(const Spc& x_){ x=x_;}

      void invertCharge(){ q=-q;}

      MagPole& operator *=(R s){q*=s; return *this;}

      MagPole& operator *=(const Group& a){x*=a; return *this;}
         // action of the basic symmetry group on a magnetic pole.

      MagPole& operator *=(const AxVec& a){a; return *this;}
         // see PointCharge::operator*=(const AxVec& a)

      MagPole& operator +=(const Vec& a){x+=a; return *this;}
         // action of a shift group on a magnetic pole.

      Force force( VecBH const& a )const{ return Force(a.v()*q);}
         // force on *this acted on by an magnetic field strength H

      Force force( const F<Spc,VecBH>& f)const{ return force(f(x));}
         // returns the force on the magnetic point charge as a result of
         // the interaction with the argument field, which is a vector
         // field giving the spatial distribution of the magnetic field
         // strength (H)

      Force force( const MagPole& mp )const;
         // force on *this caused by a magnetic pole mp

      FieldBH field()const;
         // returns the magnetic field generated by the source *this

      VecBH eval(const Spc& x)const
         { return magnetostaticField(x,*this);}


      static VecBH magnetostaticField(const Spc& x, const MagPole& mp);
         // the main physical law as a plain function
   };

   class MagDipole{ // magnetic dipoles
      // Describes 'real' magnetic dipoles (i.e. the limit to zero
      // spacing of a finitely spaced pair of poles

      Vec j;
         // magnetic dipole moment (V*s*m), so called Coulomb's
         // magnetic moment (as opposed to Amperes magnetic moment)
         // A/m

      R r;
         // a regularization radius

      Spc x;
         // position
   public:
      typedef MagDipole Type;
      CPM_IO
      Word nameOf()const{ return Word("MagDipole");}
      MagDipole():r(SysControl::singRange){}
         // x and j get initilized as zero vectors automatically

      MagDipole(const Vec& j_, const Spc& x_,
         R r_=SysControl::singRange):j(j_),r(r_),x(x_){}

      Vec get_j()const{ return j;}

      Spc get_x()const{ return x;}

      R get_r()const{ return r;}

      void set_r(R r_){ r=r_;}

      void set_x(const Spc& x_){ x=x_;}

      void setDir(const Vec& v){ j.setDir_(v);}

      void setMoment(const Vec& v){ j=v;}

      Vec getMoment()const{ return j;}

      R abs()const{ return j.abs();}
         // absolute value of the magnetic moment

      void add(const MagDipole& md){ j+=md.j;}

      MagDipole& operator *=(R s){j*=s; return *this;}

      MagDipole& operator *=(const Group& a){x*=a; j*=a; return *this;}
         // action of the basic symmetry group on a magnetic dipole.
         // Notice that x and j belong to different classes (Spc and Vec,
         // respectively) the group element a does different things to
         // them.

      MagDipole& operator +=(const Vec& a){x+=a; return *this;}
         // action of a shift on a magnetic dipole.

      MagDipole& operator *=(const AxVec& a){ j*=a; return *this;}
         // action of a rotation on a the magnetic dipole vector (place
         // held fixed)

      AxVec torque(const VecBH& H)const;
         // torque on *this caused by the magnetic field strength H

      R potEnr(const VecBH& H)const{ return (H.v()|-j);}
         // potential energy of *this with respect to the homogeneous
         // magnetic field of strength H

      R pot(MagDipole const& md)const;
       //  { return md.potEnr(eval(md.get_x()));}
         // potential energy of the pair *this and md
         // of magnetic dipoles

      ForTor forceAndTorque( const F<Spc,VecBH>& fH,
            R r=SysControl::singRange)const;
         // Force and torque on *this caused by the magnetic field fH.
         // The force depends on the the gradient of fH. Internally the
         // field is evaluated at two points, the distance of which is
         // 2*r

      ForTor forceAndTorque( const MagDipole& md )const;
         // Force and torque on *this caused by a magnetic dipole md.

      FieldBH field()const;
         // returns the magnetic field generated by the source *this

      ScalarField fieldU()const;
         // returns the magnetic potential as a field (this is related
         // to the potential energy of a magnetic pole just as
         // the electric potential is related to the potential
         // energy of a point charge.

      VecBH eval(const Spc& x)const
         { return magnetostaticDipoleField(x,*this);}

      static VecBH magnetostaticDipoleField(const Spc& x,
         const MagDipole& md);
         // the main physical law as a plain function

      static R magnetostaticDipoleFieldU(const Spc& x,
         const MagDipole& md);
         // the main physical law as a plain function
   };

/////// class SourceOfVectorField //////////////////////////////////////

   class SourceOfVectorField{
      // Sources of static electromagnetic and gravitational fields.
      // Describes the far field of a static electromagnetic and
      // gravitational field by giving its lowest order terms of a
      // multipole expansion

// idependent quantities

      R1 m;
         // mass (kg)

      R1 q;
         // electric charge (A*s), sum of qi


      Vec d;
         // electric dipole moment (A*s*m)
         // sum of qi*(xi-x)

      Vec j;
         // magnetic dipole moment (V*s*m)

      R1 r;
         // a regularization radius

      Spc x;
         // position

// dependent quantities

      PointCharge pc;
      ElcDipole ed;
      MagDipole md;
      PointMass pm;

      void update()
      {
         pc=PointCharge(q,x,r);
         ed=ElcDipole(d,x,r);
         md=MagDipole(j,x,r);
         pm=PointMass(m,x,r);
      }

      void updateQD()
      {
         pc=PointCharge(q,x,r);
         ed=ElcDipole(d,x,r);
      }

      void updateDJ()
      {
         ed=ElcDipole(d,x,r);
         md=MagDipole(j,x,r);
      }

      void updateD(){ ed=ElcDipole(d,x,r);}

      void updateJ(){ md=MagDipole(j,x,r);}

      void updateM(){ pm=PointMass(m,x,r);}

   public:
      typedef SourceOfVectorField Type;
      CPM_IO
      Word nameOf()const{ return Word("SourceOfVectorField");}

      SourceOfVectorField(){}

      SourceOfVectorField(const Spc& x0, R r0):x(x0),r(r0){update();}
         // d and j get initilized as zero vectors automatically

      explicit SourceOfVectorField(const PointCharge& pcI,
         const ElcDipole& edI=ElcDipole(),
         const MagDipole& mdI=MagDipole(),
         const PointMass& pmI=PointMass()):
         q(pcI.get_q()),d(edI.get_d()),j(mdI.get_j()),
         m(pmI.get_m()), r(pcI.get_r()),x(pcI.get_x()){update();}

      PointCharge getCharge()const{ return pc;}
         // evident
      ElcDipole getElcDipole()const{ return ed;}
         // evident
      MagDipole getMagDipole()const{ return md;}
         // evident
      PointMass getMass()const{ return pm;}
         // evident

// trivial access functions
      Spc get_x()const{ return x;}
      R get_r()const{ return r;}
      R get_q()const{ return q;}
      R get_m()const{ return m;}
      Vec get_d()const{ return d;}
      Vec get_j()const{ return j;}
      void set_r(R r_){ r=r_;update();}
      void set_x(const Spc& x_){ x=x_;update();}

// adding sources of static electromagnetic fields
      void add(const PointCharge&);
      void add(const ElcDipole&);
      void add(const MagDipole&);
      void add(const PointMass&);
      void add(const SourceOfVectorField&);

// transformations

      SourceOfVectorField& operator *=(R s)
         {q*=s; d*=s; j*=s; m*=s; update();return *this;}
         // multiplying all sources by a common factor s

      SourceOfVectorField& operator *=(const Group& a)
         {x*=a; d*=a; j*=a; update(); return *this;}
         // action of the basic symmetry group on *this

      SourceOfVectorField& operator +=(const Vec& a)
         {x+=a; update();return *this;}
         // action of a shift

      SourceOfVectorField& operator *=(const AxVec& a)
         { d*=a; j*=a; updateDJ();return *this;}
         // action of a rotation on a the dipole vectors (place x
         // held fixed)

// fields

      VectorField field()const
         // returns the field created by the source *this
      {
         return
         VectorField(pc.field()+ed.field(),md.field(),pm.field());
      }

      ScalarField fieldU()const{
         if (!SysControl::noInteractionMassToMass) return pm.fieldU();
         else  return pc.fieldU()+ed.fieldU();
      }

      FieldSpace eval(const Spc& x)const
      {
         return
         FieldSpace(VecE(pc.eval(x)+ed.eval(x)),md.eval(x),pm.eval(x));
      }

      R evalU(const Spc& x)const{
         if (!SysControl::noInteractionMassToMass) return pm.evalU(x);
         else return pc.evalU(x)+ed.evalU(x);
      }

      static V<SourceOfVectorField> redistribute(
         const V<SourceOfVectorField>& sl, const V<Spc>& pl);
      // relacing a typically large list sl of sources, by
      // typically much smaller such list (the result res of the
      // function) where the positions of the result sources are
      // given as res[i].get_x()==pl[i].
      // Going from sl to res is a course graining procedure.
      // The far field of both lists of sources is the same.
      // The source list sl can play a role similar to a multipole
      // expansion of the far field. Since we need the field also
      // quite close to the point cloud pl, a set of a few
      // multipoles up to dipole order is more efficient than
      // a single multipole up to higher order.
      // The method of defining the course grained field is as follows:
      // Each source sl[j] gets transformed into a a source located
      // at this point pl[k] of the list pl which is closest to
      // sl[j].get_x();

      static V<SourceOfVectorField> redistribute2(
         const V<SourceOfVectorField>& sl, const V<Spc>& pl,
         const V<Z>& whereToGo);
      // Faster version of the previous function, applicable if
      // additional information is abvailable (2004-05-09)
      // relacing a typically large list sl of sources, by
      // typically much smaller such list (the result res of the
      // function) where the positions of the result sources are
      // given as res[i].get_x()==pl[i].
      // Going from sl to res is a course graining procedure.
      // The far field of both lists of sources is the same.
      // The source list sl can play a role similar to a multipole
      // expansion of the far field. Since we need the field also
      // quite close to the point cloud pl, a set of a few
      // multipoles up to dipole order is more efficient than
      // a single multipole up to higher order.
      // The method of defining the course grained field is as follows:
      // Each source sl[j] gets transformed into a a source located
      // pl[whereToGo[j]]
   };
