//? cpmparensx.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   Description: Declares a class ParEns for describing
      ensembles of particles.
*/

using namespace CpmRoot;

using CPM_DIM::Vec;
using CPM_DIM::AxVec;
using CPM_DIM::Spc;
using CPM_DIM::Group;
using CPM_DIM::SubsetOfSpace;
using CPM_DIM::Rdim;
using CPM_DIM::Rdim_;
using CPM_DIM::Box;

using CpmRootX::R1;
using CpmRootX::Z1;
using CpmRootX::B;
using CpmRootX::RecordHandler;
using CpmProperties::SysControl;
using CpmProperties::BodyProperties;
using CpmProperties::ContactProperties;
using CpmProperties::ParticleProperties;
using CpmProperties::ParEnsPrp;
using CpmProperties::DumbbellData;
using CpmProperties::NuggetData;
using CpmProperties::PairData;
using CpmLinAlg::R2;
using CpmArrays::V;
using CpmArrays::X2;
using CpmArrays::X4;
using CpmArrays::R_Vector;
using CpmArrays::R_Matrix;
using CpmFunctions::F;
using CpmGeo::Iv;

//doc namespace CpmRigidBodyx{
////////////////////////// class ParEns ////////////////////////////////
   template <class ParT>
   class ParEns { // particle ensemble
      // This is still the version from electrophotography (toner and
      // carrier particles. But since the generating data type ParEnsPrp
      // is now universal (since derived from Recordable) it will be easy
      // to implement constructors for completely different ensembles,
      // such as a pile of sand.
      // An aggregate of at least one carrier particle and several toner
      // particles near to it such that the net charge is zero. The
      // constructors allow to set the constituents according to toner
      // and carrier data in a format which is common for such data. The
      // grain is electrically neutral as a whole and the charges are
      // adjusted to a charge/mass ratio of the toner material.
      static V<ParT> makeParticles(Z np, const ParEnsPrp& gp);
      static V<ParT> fillBox(const Box& box, ParEnsPrp gp,
         R facVol=1, Z np=0);

      V<ParT> rep;

   public:
      typedef ParEns Type;
      ParEns(ParEnsPrp const& pep, Z np):rep(makeParticles(np,pep)){}
         // This creates the particles near the origin and assumes that
         // proper positioning in space will be carried out by other
         // instances.

       ParEns(ParEnsPrp const& pep, Box const& box, R facVol=1, Z np=0):
         rep(fillBox(box,pep,facVol,np)){}
         // Creates a list of particles which are 'according to
         // specifications pep' and are placed randomly inside the box
         // (without provisions against overlap, since now a modification
         // of the dynamical law unbuilds overlap effectively). Adding
         // up the volumes of the particles gives facVol times the volume
         // of the box. Since making particles is expensive due to the
         // stochastic integration involved, it is possible by giving to
         // np a value > 0 to make only np particles and use them
         // repeatedly.

       CPM_IO_V

       V<ParT> operator()(void)const{ return rep;}

   }; // class ParEns

template <class ParT>
V<ParT> ParEns<ParT>::makeParticles(Z np, ParEnsPrp const& gp)
{
   Z mL=1;
   static Z npSmall=16;
   static Word loc("ParEns<>::makeParticles(Z,ParEnsPrp)");
   CPM_MA
   cpmmessage(mL,"np="&cpmwrite(np));
   Word sec("grain properties");
   R r0T=gp.getR(sec,"dT")*0.5;
   R r0C=gp.getR(sec,"dC")*0.5;
   R sT=gp.getR(sec,"sT");
   R sC=gp.getR(sec,"sC");
   R TC=gp.getR(sec,"TC");
   R qmT=gp.getR(sec,"qmT");
   Word typeT=gp.getW(sec,"typeT");
   Word typeC=gp.getW(sec,"typeC");
   sec="tonerproperties";
   R rhoT=gp.getR(sec,"rho");
   sec="carrierproperties";
   R rhoC=gp.getR(sec,"rho");

   NuggetData nugDatT(gp.rc(),"nuggetdata_for_toner");
   NuggetData nugDatC(gp.rc(),"nuggetdata_for_carrier");

   DumbbellData dumDatT(gp.rc(),"dumbbelldata_for_toner");
   DumbbellData dumDatC(gp.rc(),"dumbbelldata_for_carrier");

   PairData pairDatT(gp.rc(),"pairdata_for_toner");
   PairData pairDatC(gp.rc(),"pairdata_for_carrier");

   V<R> radii(np);
   V<R> radii2(np);
   V<R> masses(np);
   V<bool> toner(np);

   R r,q,m;
   Z i;
   R tonerContent=0;
   R tonerMass=0;
   R carrierMass=0;
   R totalMass=0;
   R spc=0;
   for (i=1;i<=np;i++){
      bool ton=(tonerContent<TC);
         // for TC>0, the first particle will be a toner particle
         // for TC==0, we will never get a toner particle
      if (ton){
         r=r0T;
         r*=cpmexp(CpmAlgorithms::gaussRandomValue(0,sT));
         m=CpmDim3::Spc::vol(r)*rhoT; // particles are 3-dimensional !
         tonerMass+=m;
      }
      else{
         r=r0C;
         r*=cpmexp(CpmAlgorithms::gaussRandomValue(0,sC));
         m=CpmDim3::Spc::vol(r)*rhoC;
         carrierMass+=m;
      }
      totalMass+=m;
      if (totalMass<=0)
         cpmerror("ParEns::makeParticles(Z,...): totalMass<=0");
      tonerContent=tonerMass/totalMass;
      radii[i]=r;
      radii2[i]=r*r;
      masses[i]=m;
      toner[i]=ton;
   }
   R toner_r2=0;
   R carrier_r2=0; // the sums over radius square will help to asign
      // charges in proportion to the surfaces
   for (i=1;i<=np;i++){
      if (toner[i]) toner_r2+=radii2[i]; else carrier_r2+=radii2[i];
   }
   R tonerCharge=tonerMass*qmT;
   R carrierCharge=-tonerCharge;
   R tonerFactor=0,carrierFactor=0;
   if (toner_r2>0) tonerFactor=tonerCharge/toner_r2;
   if (carrier_r2>0) carrierFactor=carrierCharge/carrier_r2;

   V<ParT> res(np);
   R npInv=1./np;
   R qTot=0;
   for (i=1;i<=np;i++){
      cpmprogress("Particle Factory",R(i)*npInv);
      bool ton=toner[i];
      r=radii[i];
      q = ton ? radii2[i]*tonerFactor : radii2[i]*carrierFactor;
      R d=r*2;
      R surface=CpmDim3::Spc::sur(r);
      ParticleProperties pp;
      sec= ton ? "tonerproperties" : "carrierproperties";
      pp.sigma=gp.getR(sec,"sigma");
      pp.J=gp.getR(sec,"J");
      pp.rho=gp.getR(sec,"rho");
      pp.E_=gp.getR(sec,"effectiveE");
      pp.alphaAdhesion_=gp.getR(sec,"alphaAdhesion");
      pp.relativeRangeAdhesion_=gp.getR(sec,"relativeRangeAdhesion");

      if (np>1) pp.sigma=q/surface;
         // notice that pp.sigma is already initialized in gp.
         // If there is charge to mass,...  given also
         // it has to be adjusted to these data
      if (ton){
         if (typeT=="pair"){
            res[i]=ParT(d,pp,pairDatT);
         }
         else if (typeT=="dumbbell"){
            res[i]=ParT(d,pp,dumDatT);
         }
         else{
            res[i]=ParT(d,pp,nugDatT);
         }
      }
      else{
         if (typeC=="pair"){
            res[i]=ParT(d,pp,pairDatC);
         }
         else if (typeC=="dumbbell"){
            res[i]=ParT(d,pp,dumDatC);
         }
         else{
            res[i]=ParT(d,pp,nugDatC);
         }
      }
   }
   for (i=1;i<=np;i++){
      res[i].set_id(i); // added 2004-11-01
      qTot+=res[i].get_q();
   }
   if (np>npSmall){
      R atc=cpmabs(tonerCharge);
      if (atc>0){
         R vQuality=cpmabs(qTot)/atc;
         cpmmessage(mL,"vQuality="&cpmwrite(vQuality));
         cpmassert(vQuality<1e-4,loc);
      }
   }
   CPM_MZ
   return res;
}

template <class ParT>
V<ParT> ParEns<ParT>::fillBox(const Box& box, ParEnsPrp gp,
   R facVol, Z np)
{
   Word sec("grain properties");
   R gpdC=gp.getR(sec,"dC");
   R gpdT=gp.getR(sec,"dT");
   R rPart=(gpdC+gpdT)*0.5;
   R volPart=Spc::vol(rPart);
   cpmassert(volPart>0,"V<ParT> ParEns::fillBox(...)");
   R volAim=box.vol()*facVol;
   Z n=cpmtoz(volAim/volPart);
   if (n<1) n=1;
   Z nAct;
   if (np<=0) nAct=n;
   else nAct=CpmRootX::inf<Z>(np,n);
   V<ParT> vc=ParEns::makeParticles(nAct,gp);
   V<Spc> vs=box.ran(n);
   V<ParT> vcRes(n);
   Spc ori;
   Z iAct=0;
   for (Z i=1;i<=n;i++){
      Vec tr=vs[i]-ori;
      if (iAct<nAct) iAct++;
      else iAct=1;
      ParT pi=vc[iAct];
      pi*=Group(tr);
      vcRes[i]=pi;
   }
   return vcRes;
}

template <class ParT>
bool ParEns<ParT>::prnOn(ostream& str)const
{
   cpmwt("ParEns<...> begin");
   cpmp(rep);
   cpmwt("ParEns<...> end");
   return true;
}

template <class ParT>
bool ParEns<ParT>::scanFrom(istream& str)
{
   cpms(rep);
   return true;
}
