//? cpmrigbodydyn.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
  Describing: Rigid Body Dynamics: a simple system of a few particles
  for simple demonstrations without the ballast of the full-blown system.
  Only three dimensions to get simple file structure.
*/
#include <cpmdimdef.h>
#if defined(CPM_DIMDEF3) // file content trivial otherwise

#include <cpminifilebasapp.h>
#include <cpmcompar.h>
#include <cpmp.h>

namespace CpmPala1_3{
   // Topics namespace, now integrated into PaLa as CpmPala1_3
   // last number is dimension index, presently 2D not implemented
   
   using namespace CpmStd;
   
   using CpmDim3::Vec;
   using CpmDim3::AxVec;
   using CpmDim3::Spc;
   using CpmDim3::Group;
   using CpmDim3::RefSys;
   using CpmDim3::SubsetOfSpace;
   using CpmDim3::BndSph;
   using CpmDim3::Plane;
   using CpmDim3::InTens;
   using CpmDim3::Rdim;
   using CpmDim3::Texture;

   using CpmCamera3::Camera;

   using CpmRigidBody3::RigidObject;
   
   using CpmRigidBody3::PointPar;
   using CpmRigidBody3::PointParC;
   using CpmRigidBody3::PointParV;

   using CpmRigidBody3::ComPar;
   using CpmRigidBody3::ComParC;
   using CpmRigidBody3::ComParV;

   using CpmPhysics3::Force;
   using CpmPhysics3::ForTor;
   using CpmPhysics3::FieldSpace;

   using CpmProperties::ParticleProperties;
   using CpmProperties::NuggetData;   
   using CpmProperties::SysControl; 

   using CpmGraphics::Graph;
   using CpmGraphics::Frame;

   using CpmImaging::Color;
   using CpmImaging::Palette;
   using CpmImaging::INVISIBLE;

   using CpmAlgorithms::CyclicAlarm;
   using CpmApplication::IniFileBasedApplication;

   using CpmRoot::Word;
   using CpmRoot::R;
   using CpmRoot::Z;

   using CpmRootX::Z1;
   using CpmRootX::R1;
   using CpmRootX::B;
   using CpmRootX::RecordHandler;
   using CpmRootX::Record;

   using CpmArrays::R_Vector;
   using CpmArrays::V;
   using CpmArrays::Vp;
   using CpmArrays::Vo;
 
   using CpmFunctions::F;

   using CpmSystem::OFileStream;
   using CpmSystem::IFileStream;

////////////////////////////// class RigBodyDyn ///////////////////////

#if defined(CPM_POINT_PARTICLES)

typedef PointPar ParT;
typedef Force ForT;
typedef PointParC ParCT;
typedef PointParV ParVT;

#else

typedef ComPar ParT;
typedef ForTor ForT;
typedef ComParC ParCT;
typedef ComParV ParVT;

#endif

struct DynCon{ // dynamics control

   Z stepsForTypTime;
   R relTotTime;
   R fitFactor;
   R tWait;
      // wait time for showing frames in function run(Graph)
      // So better rather short such as 0.02
   Z viwPer;
      // viewing period. Only each viwPer_'th computed step will
      // be shown on screen
   Z filPer;
      // filing period. Only each filPer_'th computed step will
      // be written to file
   Z imgPer;
      // filing period. Only each imgPer_'th computed step will
      // be written to image (bitmap)

   R trapPow;
      // power of trap potential

   R lambda;
      // coupling constant of trap potential

   FieldSpace fs;
      // combined homogeneous electric magnetic and
      // gravitational field

   R aCritRel;
      // SysControl::aCrit will be set to aCritRel times 
      // the unit acceleration in the natural units of the system 

   R vCritRel;
      // SysControl::vCrit will be set to vCritRel times 
      // the unit velocity in the natural units of the system 

   Z nRunsForOrd;
      // If this is >1 it asks for this number of runs, where always
      // the next run doubles the number of steps but covers the
      // same period of time. The maximum deviation values that get
      // normally reported for any run get listed for these more 
      // runs here and a polynomial growth law extracted 

   typedef DynCon Type;
   CPM_IO
   DynCon(){}
   DynCon(RecordHandler const& rc, Word const& sec);
      // constructor from objects that can conveniently be initialized 
      // from text
};

struct TypQun{ // 'typical quantities'
   R1 enr,len,mass,time,vel;
      // energy, length, mass, time, velocity
   Spc cenOfMass;
      // center of mass
   SubsetOfSpace bonSph;
      // bounding sphere
};

class RigBodyDyn{ // rigid body dynamics

   DynCon dc_;
      // dynamics control

   V<ParT> vp_;

   V<ForT> accPar_;
      // only used if SysControl::accMem==true

   Z1 stepCon_;
      // step count, initialized as 0
   R1 t_,dt_;
      // time and time step
   R1 enr_;
      // total energy 
   Vec mom_;
      // total momentum
   AxVec angMom_;
      // total angular momentum
   CyclicAlarm viws_;
   CyclicAlarm fils_;
   CyclicAlarm imgs_;

   F<Spc,Vec> trapAcc_;
      // trapping acceleration. This is a straightforward mechanism
      // for enforcing confined motion (i.e. motion that lets all
      // particles remain within a certain volume of space. The force
      // on particle vp_[i] is trapAcc_(xi)*mi, where
      // xi=vp_[i].get_x(); mi=vp_[i].get_m(). We will only consider 
      // radially symmetric fields which don't invalidate the constancy
      // of the angular momentum which we are interested in as a good
      // test for the accuracy of the dynamics

   F<Spc,R> trapPot_;
      // contribution to the potential energy of the trapping field
      // introduced above. For particle vp_[i] this contribution is
      // trapPot_(xi)*mi with xi and mi defined as above.
// Dependent quantities
   
   Z n_;
      // number of particles

   void adjust_();

public:
   typedef RigBodyDyn Type;
   virtual RigBodyDyn* clone()const{ return new RigBodyDyn(*this);}
   CPM_IO_V
   RigBodyDyn(){}
   RigBodyDyn(DynCon const& dc, V<ParT> const& vp):dc_(dc),vp_(vp)
   {adjust_();}
   virtual void step_(R h);
   void run(Camera& cam, ostream& str, Word const& name);
   void play(Camera& cam, ostream& str);
//   bool autoStep()const{ return dc_.autoStep_;}
   R t()const{ return t_;}
   R dt()const{ return dt_;}
   R enr()const{ return enr_;}
   Vec mom()const{ return mom_;}
   AxVec angMom()const{ return angMom_;}
   void adjustStep_(){}
      // should set dt_ to a value for which a step_ is OK
      // not yet implemented

   DynCon dynCon()const{ return dc_;}

   bool viwAlarm_()
   {  return viws_();}
      // the logic is: if dc_.dt_ has a useful value it will
      // be used (together with nSteps_) to determine a time span, 
      // that controls function run,  otherwise all control is
      // based on nSteps_ alone.

   bool filAlarm_()
   {  return fils_();}

   bool imgAlarm_()
   {  return imgs_();}

   Z stepCon()const{ return stepCon_;}

   void updConQun_();
      // update conserved quantities

   TypQun typQun()const;
      // typical energy,mass,length,time, velocity determined by the
      // system

   BndSph bndSph()const;
      // 
};

class App1RigBodyDyn:public CpmApplication::IniFileBasedApplication{//appl
 
public:
   App1RigBodyDyn(const Word& w):
      IniFileBasedApplication(w){}
   void doTheWork();
};

} // namespace
#endif
