//? cpmsys2x.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
  Describing a particle system enclosed by structures
  of type Pot, Lid, Fen suitable for simulating osmosis.
  Without using these confining structures also free
  particle systems such as star clusters can be studied.

  Implementation of fast and parallel dynamics as defined in
  cpmpardynsys.h
*/
namespace CPM_TOP{
   // Topics namespace
   // Can be CpmPala2_2 or CpmPala2_3 as seen from cpmsys2.h.

   using CPM_DIM::Vec;
   using CPM_DIM::AxVec;
   using CPM_DIM::Spc;
   using CPM_DIM::Group;
   using CPM_DIM::RefSys;
   using CPM_DIM::Plane;
   using CPM_DIM::InTens;
   using CPM_DIM::Rdim;
   using CPM_DIM::ParOfSpc;
      // partition of space, see cpmpartition.h

   using CPM_CAMERA::Camera;

   using CPM_RIGID_BODY::RigidObject;

   using CPM_RIGID_BODY::PointPar;
   using CPM_RIGID_BODY::PointParC;
   using CPM_RIGID_BODY::PointParV;

   using CPM_RIGID_BODY::ComPar;
   using CPM_RIGID_BODY::ComParC;
   using CPM_RIGID_BODY::ComParV;

   using CPM_PHYSICS::ExtSys;
   using CPM_PHYSICS::Body;
   using CPM_PHYSICS::Pot;
   using CPM_PHYSICS::Lid;
   using CPM_PHYSICS::Fen;
   using CPM_PHYSICS::SphTank;

   using CPM_PHYSICS::Force;
   using CPM_PHYSICS::ForTor;
   using CPM_PHYSICS::FieldSpace;
   using CPM_PHYSICS::ScalarField;
   using CPM_PHYSICS::VectorField;
   using CPM_PHYSICS::PointCharge;
   using CPM_PHYSICS::SourceOfVectorField;
   using CPM_PHYSICS::MagDipole;

   using CpmDynSys::ParDynSys;

   using CpmGraphics::Graph;
   using CpmGraphics::Frame;
   using CpmGraphics::Frames;

   using CpmImaging::Color;
   using CpmImaging::Palette;
   using CpmImaging::INVISIBLE;

   using CpmProperties::BodyProperties;
   using CpmProperties::ParticleProperties;

   using CpmAlgorithms::Switch;

   using CpmRoot::Word;
   using CpmRoot::R;
   using CpmRoot::Z;

   using CpmRootX::Z1;
   using CpmRootX::R1;
   using CpmRootX::B;
   using CpmRootX::RecordHandler;
   using CpmRootX::Record;

   using CpmArrays::R_Vector;
   using CpmArrays::I_Vector;
   using CpmArrays::R_Matrix;
   using CpmArrays::V;
   using CpmArrays::Vp;
   using CpmArrays::Vo;
   using CpmArrays::X2;
   using CpmArrays::T2;
   using CpmArrays::X4;

   using CpmFunctions::F;

   using CpmSystem::Error;
   using CpmSystem::Timer;
   using CpmSystem::OFileStream;

////////////////////////////// class Sys2 ///////////////////////
// Parallelism now extended to all public functions.
// The name Sys2 reminds us that the objects of this type are made
// of 2 main components: a particle system and a system of
// enclosing  structures which form an environment for the
// particles.

#if defined(CPM_POINT_PARTICLES)

typedef PointPar ParT;
typedef Force ForT;
typedef PointParC ParCT;
typedef PointParV ParVT;

#else

typedef ComPar ParT;
typedef ForTor ForT;
typedef ComParC ParCT;
typedef ComParV ParVT;

#endif

// defining Sys2Base
typedef ParDynSys<ParT,ExtSys,Vec,Spc,Group,Partition,ForT,
   FieldSpace,ScalarField,VectorField,PointCharge,
   SourceOfVectorField,MagDipole,PointPar,Camera> Sys2Base;
   // specialization of the 'universal' template class in
   // cpmpardynsys.h

// defining ExtrSys2
typedef CPM_TOP::ParExt<ParCT,ParVT,ParT,ExtSys,Sys2Base> ExtrSys2;
   // file oriented type needed in
   // explicit Sys2(const ExtrSys2& extr);

class Sys2:public Sys2Base{
// Sys2, the final physical system for confined granular matter dynamics

   mutable B running_;
      // introduced 2005-06-14
      // initialized as false, which is OK
      // A device for allowing remembering physical properties
      // of the initial state in function docuStep(...).
      // Presently only used for documode=="total_energy".

   Z1 nTot_;
      // total number of particles (in all processes)

   V<Z> per_;
      // A permutation of the particles which enters the definition
      // of function asym. If no symmetry is under consideration,
      // one has per_.dim()==0.

   Record rc_;
      // flexible container for data, instead of naming them all.
      // Not named rec_ since this would suggest also to use rec and this
      // would clash with function rec for parallel send and receive.
      // The basic constructor
      //    Sys2(const RecordHandler& rch)
      // brings with it the Record rch(), but rc_ needs not simply (or
      // not only) to hold the data of rch(). It rather may hold
      // additional data which are computed from data in rch() and
      // which may only be needed for some of the control paths
      // in the function block of a constructor, particularly in the one
      // mentioned above. In the present case these are mostly
      // the input quantities 'topic' and 'documode' which determine
      // which additional quantities should be known to the
      // Sys2-object under consideration and construction.


   Sys2(V<ParT> const& vc, Vp<ExtSys> const& ext,
      Partition const& pr, Record const& rec, bool careful=true);
      // auxiliary constructor which lets rank==1 build the
      // whole system and leaves distribution to other
      // processes as a task for additional code.
      // If the last argument has value true, only those
      // particles from vc will be used that are reasonably
      // within the process volume SubsetOfSpace pr.getX().

   explicit Sys2(const ExtrSys2& extr);
      // construction from a file extractor; not for
      // parallel computing. Needed for reading movie files.

   explicit Sys2(V<Word> const& typeList);
      // Setting the exact types for the external system
      // needed in function distributeSys2 as a prerequisite
      // to properly receiving the data describing the
      // external systems. In a sense this is a trick to make
      // my polymorphic array Vp<> work also under the harder
      // conditions of parallelism (2006-02-08).

   void distributeSys2(const Partition&);
      // assumes that rank 1 has all particles available
      // The second argument was introduced in 02-04-23 to
      // avoid long-standing malfunction when reading states from file.
      // The distribution of particles with additional overlap
      // deactivated seems to be appropriate only for states as coming
      // out directly from a clean constructor which did not yet
      // perform dynamical time steps.
      // Does not only set the particles for the other processes but
      // makes the whole station known to ranks different from 1.

   void file1(const Word& fn)const;
      // filing for size==1

   typedef Sys2 Type;

public:

   static X4<V<ParT>,Vp<ExtSys>,Partition,Z>
      readFromMovie(const Word& movieFile, R frac=1);
      // returns the quantities specified in the return value
      // from the ith frame of a moviefile, where i is characterized
      // by a fraction frac of the whole movie.
      // i is returned as the 6th component of the return value.
      // Thus frac=1 corresponds to the last frame of the movie.

   explicit Sys2(const RecordHandler&);
      // parallel version!
      // construction from design data

   explicit Sys2(const Word& movieFile, const Record& rc,
      R frac=1);
      // parallel version!
      // construction from the ith frame of a moviefile
      // where i is characterized by a fraction frac of the whole movie
      // Thus frac=1 corresponds to the last frame of the movie.
      // The argument rc contributes only through its section
      // 'partition_data':
      // It adds the data that define the partition attribute of the
      // base class. This over-writes the value written to file. This
      // has the advantage, that one can use a state from
      // a movie file as initial state of a run for which the number of
      // parallel processors differs from the number that was active when
      // creating the movie file.
      // The profile wave numbers k1Sh, k1PC, k3Sh, k3PC get adjusted to
      // the new partition in order not to disturb the particles by
      // moving them from the end of the partition domain to the beginning
      // (was not the case till 02-07-16).
      // If the new partition corresponds to a smaller region of space
      // than the original one, the number of particles get reduced
      // accordingly.
      // If, however, rc is trivial (i.e. rc.isVoid()==true) the
      // partition attribute will be taken from movieFile and it is
      // assumed that the data of particles and external systems
      // in movieFile are mutually consistent so that the adjustments
      // described earlier will not be made. This is the desired behavior
      // for continuation of moviefiles, i.e for taking the final frame
      // of a movie file as the initial state of a new run.

   Sys2(void);
      // Works in parallel programs since no structures get created
      // which depend on the partition of space (in particular,
      // no particles!). A correct idiom for creating and using a
      // Sys2 in a parallel program is (see padyanapscpp.h)
/*
       Sys2 ton;
       if (...){
            ton=Sys2(aRecordHandler);
               // makes ton the right object in each process
       }
       else{
            ton=Sys2(aWord,aRecord);
               // makes ton the right object in each process
       }
       ton.run(...);
*/
      // notice that the default constructor together with '='
      // allows more flexible program control than non-trivial
      // constructors alone.

// functions that are virtual in Sys2Base

   CPM_IO
      // input and output in C+- style

   CPM_NAM(Sys2)
      // 'modern' form of providing the nameOf function

   virtual void mark(const Camera& ca, bool display,
      const Word& text=Word(), bool fileOnShow=false)const;
      // display system on screen;
      // The parts of the systems for which the various processes are
      // responsible get messaged to process 1 and graphically
      // represented by process 1 in succession.

   virtual X2<Z,Word> docuStep(OFileStream& ofs)const;
      // if active==false, nothing is written to ofs and only the
      // return value is given. The return value puts together the
      // number of data arrays and a diagram title.

   virtual void boost(const Vec& v, R vAngleRange);
      // boost the light particles with  velocity v and the heavy ones
      // with a lower velocity that gives them
      // the same kinetical energy (simulating equal temperatures)

   virtual void setVel();
      // virtual and trivial in Sys2Base, see explanation in
      // cpmpardynsys.h
      // Here we implement the setting of suitable velocity distributions
      // controlled by the state of rc_. The velocities are understood as
      // addition to the existing velocities (which a state may have if
      // it was created as the final state of a pre-run).

   Z parType(Z i)const;
      // returns a discrete type number of par[i]. For the osmosis
      // application 1 for the small particles and 2 for the large ones.
      // In general, the value has to be determined by the state of par[i]
      // and by the state of pc_ (which is highly flexible).

   Z parLoc(Z i)const;
      // returns a discrete type number of par[i]. For the osmosis
      // application 1 for the small particles and 2 for the large ones.

   virtual bool file(const Word& fn, const Timer& tm)const;
      // virtual in base
      // writes the state into a suitable file, which should
      // be made visible by a suitable viewer program on a
      // single computer. Since the number of particles to be
      // considered is huge, it is rather evident that one should
      // file not too much redundant data. This suggests to divide
      // the data of an object in constant ones and those that change
      // with time. This is implemented in the present function.

   virtual Z get_nTot()const{ return nTot_;}
      // overwrites virtual function of Base

// access functions concerning rec_
   Record getRc()const{ return rc_;}

   void setRc(Record const& rc){ rc_=rc;}

// constant functions returning properties of the system

   Spc carCenOfMas()const;
      // center-of-mass of the set of particles

   X2<R,R> totChr()const;
      // Name means 'total charges'.
      // Return.first is the sum of all negative charges of particles.
      // Return.second is the sum of all positive charges of particles.
      // Notice that increasing values from left to right are
      // considered natural order; so negative charges have to come
      // first.

   R getHeight()const;
      // evident

   static T2<Word> docMode(Record const& rc);
      //: documentation mode

   T2<Word> getDocMode()const{ return docMode(rc_);}
      //: get documentation mode
      // Returns the pair (topic,documode),
      // where 'topic' is read from section 'topic under consideration'
      // of rc_ and 'documode' is read from section topic&" data"
      // of rc_.

   R asym(bool log=true)const;
      // asymmetry
      // Measure of asymmetry of particle configuration. Is 0 (or in
      // logarithmic mode, which is the default one, the log10 of a very
      // small number) if particles i and i' (i' := per_[i]) are
      // mirror-images of each other.
      // We don't here parametrize the kind of mirror-symmetry but only
      // consider the symmetry which is defined by point s0 and vector
      // mirDir defined as
      // RefSys master;
      // Spc s0=master.o();
      // Vec mirDir = master.e1();
      // If the newly introduced argoment is true, the log10 of the
      // preliminary result is given (log10(0):=0 here by definition).
      // If per_.dim()==0, the function returns always 1.

   void setPer_(V<Z> const& per){ per_=per;}
};
} // namespace
