//? cpmproperties.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmproperties.h>
#include <cpmconstphys.h>
#include <cpmrfunc.h>

using namespace CpmRoot;
using namespace CpmProperties;
using namespace CpmPhysics;
using CpmRootX::RecordHandler;

R CpmProperties::nonHarmonicMean(R x1, R x2){ return (x1+x2)*0.5;}

R CpmProperties::harmonicMean(R x1, R x2)
// a factor 2 was missing till 2000-3-10, without this factor we have
// harmonicMean(x,x) != x
{
   R nom=x1+x2;
   if (nom==0.){
      if (x1==0) return 0;
         // normal
      cpmwarning("CpmProperties::harmonicMean(...): x1+x2==0");
      cpmmessage("x1="&cpmwrite(x1));
      cpmmessage("x2="&cpmwrite(x2));
      return 0;
   }
   else{
      return (x1*x2*2)/nom;
   }
}

R CpmProperties::mixedMean(R x1, R x2)
{
   return (x1==0 || x2==0) ? nonHarmonicMean(x1,x2) : harmonicMean(x1,x2)\
   ;
}

R CpmProperties::transitionFunction(const R& x, const R& xNear,
   const R& xFar )
// this is a differentiable transit function
{
   if (x<=xNear) return 1;
   if (x>=xFar) return 0;
   cpmassert(xNear<xFar,"CpmProperties::transitionFunction(R,R,R)");
   R p=(x-xNear)/(xFar-xNear);
   return CpmFunctions::R_Func::hill3(p);
}

/////////////////////// struct SysControl //////////////////////////////
Z SysControl::nyScalarFields=32;
Z SysControl::nxElectricField=10;
Z SysControl::nyElectricField=6;
Z SysControl::markElectricField=-1;
Z SysControl::nxMagneticField=11;
Z SysControl::nyMagneticField=8;
Z SysControl::markMagneticField=-1;
Z SysControl::nxGravityField=12;
Z SysControl::nyGravityField=9;
Z SysControl::markGravityField=-1;
R SysControl::fieldGamma=0.25;
Z SysControl::verbose=0;
Z SysControl::fixedStep=0;
R SysControl::intactRange=1e100; // should never become active
R SysControl::singRange=1e-6;
Z SysControl::noAdhesion=1;
   // adhesion forces=0 in the present program anyway
Z SysControl::noElasticRepulsion=0;
Z SysControl::noFriction=0;
Z SysControl::noInteractionChargeToWall=1;
Z SysControl::noInteractionChargeToCharge=0;
Z SysControl::noMagneticInteractionChargeToCharge=1;
Z SysControl::noInteractionMagnetToMagnet=0;
Z SysControl::noInteractionMassToMass=1;
Z SysControl::noInteractionElectricField=0;
Z SysControl::noInteractionElectricDipole=0;
Z SysControl::noInteractionMagneticField=0;
Z SysControl::noForceMagneticField=0;
Z SysControl::noInteractionGravitationField=0;
Z SysControl::noMeanFieldPar=1; // clean 'classical' treatment
   // for small particle numbers, was 0 till 2005-03-26
Z SysControl::ignoreExternalSystems=0;
Z SysControl::coloredParticles=0;
Z SysControl::singleCharge=1;
bool SysControl::accMem=true;
   // true works extremely better for Lorentz forces,
   // for non-velocity dependent forces no obvious difference
bool SysControl::newRigBodyInt=true;
R SysControl::eta=0; // if needed, it should be set more transparently
//R SysControl::eta=1.72e-6; // air
R SysControl::vCrit=0; // no influence if <=0
R SysControl::aCrit=0; // no influence if <=0
   // critical acceleration smallest meaningful length
   // divided by smallest meaningful time squared
bool SysControl::veryFar=true;
Z SysControl::bitmapToShow=0;
Z SysControl::potentialToShow=1;
Z SysControl::howToMarkParticles=0;
bool SysControl::showMeanFil=false;
bool SysControl::showParticles=true;
bool SysControl::showPartFields=false;
bool SysControl::showPartition=false;
bool SysControl::showPrimaryChargesOnly=true;
R SysControl::nearFactor=1;
R SysControl::tSmall=1e-10;
R SysControl::tStep=0;
R SysControl::tWait=0;
R SysControl::dtMult=0;

void SysControl::reset()
{
   nyScalarFields=32;
   nxElectricField=10;
   nyElectricField=6;
   markElectricField=-1;
   nxMagneticField=11;
   nyMagneticField=8;
   markMagneticField=-1;
   nxGravityField=12;
   nyGravityField=9;
   markGravityField=-1;

   fieldGamma=0.25;
   verbose=0;
   fixedStep=0;
   intactRange=1e100; // should never become active
   singRange=1e-6;
   noAdhesion=1;
   noElasticRepulsion=0;
   noFriction=0;
   noInteractionChargeToWall=1;
   noInteractionChargeToCharge=0;
   noMagneticInteractionChargeToCharge=1;
   noInteractionMagnetToMagnet=0;
   noInteractionMassToMass=1;
   noInteractionElectricField=0;
   noInteractionElectricDipole=0;
   noInteractionMagneticField=0;
   noForceMagneticField=0;
   noInteractionGravitationField=0;
   noMeanFieldPar=1;
   ignoreExternalSystems=0;
   coloredParticles=0;
   singleCharge=1;
   accMem=true;
   newRigBodyInt=true;
   eta=0;
   vCrit=0;
   aCrit=0;
   veryFar=true;
   bitmapToShow=0;
   potentialToShow=1;
   howToMarkParticles=0;
   showMeanFil=false;
   showParticles=true;
   showPartFields=false;
   showPartition=false;
   showPrimaryChargesOnly=true;
   nearFactor=1;
   tSmall=1e-10;
   tStep=0;
   tWait=0;
   dtMult=0;
}

void SysControl::read(const RecordHandler& rch, const Word& sec, Z ped)
{
    rch.read(sec, "nyScalarFields", nyScalarFields,ped);
    rch.read(sec, "nxElectricField", nxElectricField,ped);
    rch.read(sec, "nyElectricField", nyElectricField,ped);
    rch.read(sec, "markElectricField", markElectricField,ped);
    rch.read(sec, "nxMagneticField", nxMagneticField,ped);
    rch.read(sec, "nyMagneticField", nyMagneticField,ped);
    rch.read(sec, "markMagneticField", markMagneticField,ped);
    rch.read(sec, "nxGravityField", nxGravityField,ped);
    rch.read(sec, "nyGravityField", nyGravityField,ped);
    rch.read(sec, "markGravityField", markGravityField,ped);
    rch.read(sec, "fieldGamma", fieldGamma,ped);
    rch.read(sec, "verbose", verbose,ped);
    rch.read(sec, "fixedStep", fixedStep,ped);
    rch.read(sec, "intactRange", intactRange,ped);
    rch.read(sec, "singRange", singRange,ped);
    rch.read(sec, "noAdhesion", noAdhesion,ped);
    rch.read(sec, "noElasticRepulsion", noElasticRepulsion,ped);
    rch.read(sec, "noFriction", noFriction,ped);
    rch.read(sec, "noInteractionChargeToWall",
       noInteractionChargeToWall,ped);
    rch.read(sec, "noInteractionChargeToCharge",
       noInteractionChargeToCharge,ped);
    rch.read(sec, "noMagneticInteractionChargeToCharge",
       noMagneticInteractionChargeToCharge,ped);
    rch.read(sec, "noInteractionMagnetToMagnet",
       noInteractionMagnetToMagnet,ped);
    rch.read(sec, "noInteractionMassToMass",noInteractionMassToMass,ped);
    rch.read(sec, "noInteractionElectricField",
      noInteractionElectricField,ped);
    rch.read(sec, "noInteractionElectricDipole",
      noInteractionElectricDipole,ped);
    rch.read(sec, "noInteractionMagneticField",
      noInteractionMagneticField,ped);
    rch.read(sec, "noForceMagneticField",noForceMagneticField,ped);
    rch.read(sec, "noInteractionGravitationField",
      noInteractionGravitationField,ped);
    rch.read(sec, "noMeanFieldPar",noMeanFieldPar,ped);
    rch.read(sec, "ignoreExternalSystems", ignoreExternalSystems,ped);
    rch.read(sec, "coloredParticles", coloredParticles,ped);
    rch.read(sec, "singleCharge", singleCharge,ped);
    rch.read(sec, "accMem", accMem,ped);
    rch.read(sec, "newRigBodyInt", newRigBodyInt,ped);
    rch.read(sec, "eta",eta,ped);
    rch.read(sec, "vCrit",vCrit,ped);
    rch.read(sec, "aCrit",aCrit,ped);
    rch.read(sec, "veryFar",veryFar,ped);
    rch.read(sec, "bitmapToShow",bitmapToShow,ped);
    rch.read(sec, "potentialToShow",potentialToShow,ped);
    rch.read(sec, "howToMarkParticles",howToMarkParticles,ped);
    rch.read(sec, "showMeanFil",showMeanFil,ped);
    rch.read(sec, "showParticles",showParticles,ped);
    rch.read(sec, "showPartFields",showPartFields,ped);
    rch.read(sec, "showPartition",showPartition,ped);
    rch.read(sec, "showPrimaryChargesOnly",showPrimaryChargesOnly,ped);
    rch.read(sec, "nearFactor",nearFactor,ped);
    rch.read(sec, "tSmall",tSmall,ped);
    rch.read(sec, "tStep",tStep,ped);
    rch.read(sec, "tWait",tWait,ped);
    rch.read(sec, "dtMult",dtMult,ped);
}

///////////////// class SysControlState ////////////////////////////////

bool SysControlState::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("SysControlState",str);
   cpmp(nyScalarFields);
   cpmp(nxElectricField);
   cpmp(nyElectricField);
   cpmp(markElectricField);
   cpmp(nxMagneticField);
   cpmp(nyMagneticField);
   cpmp(markGravityField);
   cpmp(nxGravityField);
   cpmp(nyGravityField);
   cpmp(markMagneticField);
   cpmp(fieldGamma);
   cpmp(verbose);
   cpmp(fixedStep);
   cpmp(intactRange);
   cpmp(singRange);
   cpmp(noAdhesion);
   cpmp(noElasticRepulsion);
   cpmp(noFriction);
   cpmp(noInteractionChargeToWall);
   cpmp(noInteractionChargeToCharge);
   cpmp(noMagneticInteractionChargeToCharge);
   cpmp(noInteractionMagnetToMagnet);
   cpmp(noInteractionMassToMass);
   cpmp(noInteractionElectricField);
   cpmp(noInteractionElectricDipole);
   cpmp(noInteractionMagneticField);
   cpmp(noForceMagneticField);
   cpmp(noInteractionGravitationField);
   cpmp(noMeanFieldPar);
   cpmp(ignoreExternalSystems);
   cpmp(coloredParticles);
   cpmp(singleCharge);
   cpmp(accMem);
   cpmp(newRigBodyInt);
   cpmp(eta);
   cpmp(vCrit);
   cpmp(aCrit);
   cpmp(veryFar);
   cpmp(bitmapToShow);
   cpmp(howToMarkParticles);
   cpmp(potentialToShow);
   cpmp(showMeanFil);
   cpmp(showParticles);
   cpmp(showPartFields);
   cpmp(showPartition);
   cpmp(showPrimaryChargesOnly);
   cpmp(nearFactor);
   cpmp(tSmall);
   cpmp(tStep);
   cpmp(tWait);
   cpmp(dtMult);
   return true;
}

bool SysControlState::scanFrom(istream& str)
{
   cpms(nyScalarFields);
   cpms(nxElectricField);
   cpms(nyElectricField);
   cpms(markElectricField);
   cpms(nxMagneticField);
   cpms(nyMagneticField);
   cpms(markMagneticField);
   cpms(nxGravityField);
   cpms(nyGravityField);
   cpms(markGravityField);
   cpms(fieldGamma);
   cpms(verbose);
   cpms(fixedStep);
   cpms(intactRange);
   cpms(singRange);
   cpms(noAdhesion);
   cpms(noElasticRepulsion);
   cpms(noFriction);
   cpms(noInteractionChargeToWall);
   cpms(noInteractionChargeToCharge);
   cpms(noMagneticInteractionChargeToCharge);
   cpms(noInteractionMagnetToMagnet);
   cpms(noInteractionMassToMass);
   cpms(noInteractionElectricField);
   cpms(noInteractionElectricDipole);
   cpms(noInteractionMagneticField);
   cpms(noForceMagneticField);
   cpms(noInteractionGravitationField);
   cpms(noMeanFieldPar);
   cpms(ignoreExternalSystems);
   cpms(coloredParticles);
   cpms(singleCharge);
   cpms(accMem);
   cpms(newRigBodyInt);
   cpms(eta);
   cpms(vCrit);
   cpms(aCrit);
   cpms(veryFar);
   cpms(bitmapToShow);
   cpms(potentialToShow);
   cpms(howToMarkParticles);
   cpms(showMeanFil);
   cpms(showParticles);
   cpms(showPartFields);
   cpms(showPartition);
   cpms(showPrimaryChargesOnly);
   cpms(nearFactor);
   cpms(tSmall);
   cpms(tStep);
   cpms(tWait);
   cpms(dtMult);
   return true;
}

void SysControlState::mem()
{
   nyScalarFields=SysControl::nyScalarFields;
   nxElectricField=SysControl::nxElectricField;
   nyElectricField=SysControl::nyElectricField;
   markElectricField=SysControl::markElectricField;
   nxMagneticField=SysControl::nxMagneticField;
   nyMagneticField=SysControl::nyMagneticField;
   markMagneticField=SysControl::markMagneticField;
   nxGravityField=SysControl::nxGravityField;
   nyGravityField=SysControl::nyGravityField;
   markGravityField=SysControl::markGravityField;
   fieldGamma=SysControl::fieldGamma;
   verbose=SysControl::verbose;
   fixedStep=SysControl::fixedStep;
   intactRange=SysControl::intactRange;
   singRange=SysControl::singRange;
   noAdhesion=SysControl::noAdhesion;
   noElasticRepulsion=SysControl::noElasticRepulsion;
   noFriction=SysControl::noFriction;
   noInteractionChargeToWall=SysControl::noInteractionChargeToWall;
   noInteractionChargeToCharge=SysControl::noInteractionChargeToCharge;
   noMagneticInteractionChargeToCharge=
      SysControl::noMagneticInteractionChargeToCharge;
   noInteractionMagnetToMagnet=SysControl::noInteractionMagnetToMagnet;
   noInteractionMassToMass=SysControl::noInteractionMassToMass;
   noInteractionElectricField=SysControl::noInteractionElectricField;
   noInteractionElectricDipole=SysControl::noInteractionElectricDipole;
   noInteractionMagneticField=SysControl::noInteractionMagneticField;
   noForceMagneticField=SysControl::noForceMagneticField;
   noInteractionGravitationField=
      SysControl::noInteractionGravitationField;
   noMeanFieldPar=SysControl::noMeanFieldPar;
   ignoreExternalSystems=SysControl::ignoreExternalSystems;
   coloredParticles=SysControl::coloredParticles;
   singleCharge=SysControl::singleCharge;
   accMem=SysControl::accMem;
   newRigBodyInt=SysControl::newRigBodyInt;
   eta=SysControl::eta;
   vCrit=SysControl::vCrit;
   aCrit=SysControl::aCrit;
   veryFar=SysControl::veryFar;
   bitmapToShow=SysControl::bitmapToShow;
   potentialToShow=SysControl::potentialToShow;
   howToMarkParticles=SysControl::howToMarkParticles;
   showMeanFil=SysControl::showMeanFil;
   showParticles=SysControl::showParticles;
   showPartFields=SysControl::showPartFields;
   showPartition=SysControl::showPartition;
   showPrimaryChargesOnly=SysControl::showPrimaryChargesOnly;
   nearFactor=SysControl::nearFactor;
   tSmall=SysControl::tSmall;
   tStep=SysControl::tStep;
   tWait=SysControl::tWait;
   dtMult=SysControl::dtMult;
}

void SysControlState::act()const
{
   SysControl::nyScalarFields=nyScalarFields;
   SysControl::nxElectricField=nxElectricField;
   SysControl::nyElectricField=nyElectricField;
   SysControl::markElectricField=markElectricField;
   SysControl::markElectricField=markElectricField;
   SysControl::nxMagneticField=nxMagneticField;
   SysControl::nyMagneticField=nyMagneticField;
   SysControl::markMagneticField=markMagneticField;
   SysControl::nxGravityField=nxGravityField;
   SysControl::nyGravityField=nyGravityField;
   SysControl::markGravityField=markGravityField;
   SysControl::fieldGamma=fieldGamma;
   SysControl::verbose=verbose;
   SysControl::fixedStep=fixedStep;
   SysControl::intactRange=intactRange;
   SysControl::singRange=singRange;
   SysControl::noAdhesion=noAdhesion;
   SysControl::noElasticRepulsion=noElasticRepulsion;
   SysControl::noFriction=noFriction;
   SysControl::noInteractionChargeToWall=noInteractionChargeToWall;
   SysControl::noInteractionChargeToCharge=noInteractionChargeToCharge;
   SysControl::noMagneticInteractionChargeToCharge=
      noMagneticInteractionChargeToCharge;
   SysControl::noInteractionMagnetToMagnet=noInteractionMagnetToMagnet;
   SysControl::noInteractionMassToMass=noInteractionMassToMass;
   SysControl::noInteractionElectricField=noInteractionElectricField;
   SysControl::noInteractionElectricDipole=noInteractionElectricDipole;
   SysControl::noInteractionMagneticField=noInteractionMagneticField;
   SysControl::noForceMagneticField=noForceMagneticField;
   SysControl::noInteractionGravitationField=
      noInteractionGravitationField;
   SysControl::noMeanFieldPar=noMeanFieldPar;
   SysControl::ignoreExternalSystems=ignoreExternalSystems;
   SysControl::coloredParticles=coloredParticles;
   SysControl::singleCharge=singleCharge;
   SysControl::accMem=accMem;
   SysControl::newRigBodyInt=newRigBodyInt;
   SysControl::eta=eta;
   SysControl::vCrit=vCrit;
   SysControl::aCrit=aCrit;
   SysControl::veryFar=veryFar;
   SysControl::bitmapToShow=bitmapToShow;
   SysControl::potentialToShow =potentialToShow;
   SysControl::howToMarkParticles=howToMarkParticles;
   SysControl::showMeanFil=showMeanFil;
   SysControl::showParticles=showParticles;
   SysControl::showPartFields=showPartFields;
   SysControl::showPartition=showPartition;
   SysControl::showPrimaryChargesOnly=showPrimaryChargesOnly;
   SysControl::nearFactor=nearFactor;
   SysControl::tSmall=tSmall;
   SysControl::tStep=tStep;
   SysControl::tWait=tWait;
   SysControl::dtMult=dtMult;
}

//////////////////// class ContactProperties ////////////////////////////
// If the third argument is ped=0 the program will not stop
// if the quantity can't be read. This helps allows to stay compatible
// with data structures that are partly obsolete in newer versions of the
//
// program

ContactProperties::ContactProperties(RecordHandler const& rch,
   Word const& sec)
{
// trivial initialization in case that the file gave nothing
   alphaAdhesion_=0;
   relativeRangeAdhesion_=0;
   E_=0;
   mu_=0;
   v0_=0;
   scnr_=1;
   Z ped=1; // pedantic, value 1 warns on read failure
   // default value 2 stops, value 0 does not even warn
   rch.read(sec,"alphaAdhesion",alphaAdhesion_,ped);
   rch.read(sec,"relativeRangeAdhesion",relativeRangeAdhesion_,ped);
   rch.read(sec,"E",E_,ped);
   rch.read(sec,"mu",mu_,ped);
   rch.read(sec,"v0",v0_,ped);
   R cnr=1;
   bool b=rch.read(sec,"cnr",cnr,ped);
   if (b) scnr_=cnr*cnr;
}

ContactProperties ContactProperties::combine(
   const ContactProperties& cp)const
// combination by harmonic mean lacks a physical basis in most cases
{
   ContactProperties res;
   res.alphaAdhesion_=harmonicMean(alphaAdhesion_,cp.alphaAdhesion_);
   res.relativeRangeAdhesion_=harmonicMean(
   relativeRangeAdhesion_,cp.relativeRangeAdhesion_);
   res.E_=mixedMean(E_,cp.E_);
   res.mu_=nonHarmonicMean(mu_,cp.mu_);
   res.v0_=nonHarmonicMean(v0_,cp.v0_);
   res.scnr_=nonHarmonicMean(scnr_,cp.scnr_);
   return res;
}

bool ContactProperties::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("ContactProperties",str);
   cpmp(alphaAdhesion_);
   cpmp(relativeRangeAdhesion_);
   cpmp(E_);
   cpmp(mu_);
   cpmp(v0_);
   cpmp(scnr_);
   return true;
}

bool ContactProperties::scanFrom(istream& str)
{
   cpms(alphaAdhesion_);
   cpms(relativeRangeAdhesion_);
   cpms(E_);
   cpms(mu_);
   cpms(v0_);
   cpms(scnr_);
   return true;
}

/////////////////////// class ParticleProperties ///////////////////////

ParticleProperties::ParticleProperties(const RecordHandler& rch,
   const Word& sec):ContactProperties(rch,sec),color(CpmImaging::STEREO)
{
   Z ped=1; // pedantic
   Z col;
   bool b=rch.read(sec,"color",col,ped);
   if (b) color=Color::toPal(col);
// to get harmless initialization if nothing can be
// found on file
   rho=1;
   sigma=0;
   J=0;
   eps=1;
   fRelax=1;
   rch.read(sec,"rho",rho,ped);
   rch.read(sec,"sigma",sigma,ped);
   rch.read(sec,"J",J,ped);
   rch.read(sec,"eps",eps,ped);
   rch.read(sec,"fRelax",fRelax,ped);
}

bool ParticleProperties::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("ParticleProperties",str);
   if (!ContactProperties::prnOn(str)) return false;
   cpmp(Color::toZ(color)); // color treated as Z for I/O
   cpmp(rho);
   cpmp(sigma);
   cpmp(J);
   cpmp(eps);
   cpmp(fRelax);
   return true;
}

bool ParticleProperties::scanFrom(istream& str)
{
   if (!ContactProperties::scanFrom(str)) return false;
   Z temp;
   cpms(temp);
   color=Color::toPal(temp);
   cpms(rho);
   cpms(sigma);
   cpms(J);
   cpms(eps);
   cpms(fRelax);
   return true;
}

////////////////////////////// class BodyProperties ////////////////////

BodyProperties::BodyProperties(const RecordHandler& rch,
   const Word& sec): ContactProperties(rch,sec)
{
   Z ped=1;
   eps=1;
   d=0;
   rch.read(sec,"eps",eps,ped);
   rch.read(sec,"d",d,ped);
}

bool BodyProperties::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("BodyProperties",str);
   if (!ContactProperties::prnOn(str)) return false;
   cpmp(eps);
   cpmp(d);
   return true;
}

bool BodyProperties::scanFrom(istream& str)
{
   if (!ContactProperties::scanFrom(str)) return false;
   cpms(eps);
   cpms(d);
   return true;
}

///////////////////////// struct PairData //////////////////////////////

bool PairData::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("PairData",str);
   cpmp(r1);
   cpmp(r2);
   return true;
}

bool PairData::scanFrom(istream& str)
{
   cpms(r1);
   cpms(r2);
   return true;
}

PairData::PairData(const RecordHandler& rch, const Word& sec)
{
   r1=0;
   r2=0;
   rch.read(sec,"r1",r1);
   rch.read(sec,"r2",r2);
}

PairData::PairData(const Record& rc, const Word& sec)
{
   RecordHandler rch(rc);
   PairData res(rch,sec);
   *this=res;
}

///////////////////// struct DumbbellData //////////////////////////////

bool DumbbellData::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("DumbbellData",str);
   cpmp(rc);
   cpmp(re);
   cpmp(phi);
   return true;
}

bool DumbbellData::scanFrom(istream& str)
{
   cpms(rc);
   cpms(re);
   cpms(phi);
   return true;
}

DumbbellData::DumbbellData(const RecordHandler& rch, const Word& sec)
{
   rc=0;
   re=0;
   phi=0;
   rch.read(sec,"rc",rc);
   rch.read(sec,"re",re);
   rch.read(sec,"phi",phi);
}

DumbbellData::DumbbellData(const Record& rc, const Word& sec)
{
   RecordHandler rch(rc);
   DumbbellData res(rch,sec);
   *this=res;
}

/////////////////////// struct NuggetData //////////////////////////////

bool NuggetData::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("NuggetData",str);
   cpmp(overlapFac);
   cpmp(isotropyFac);
   cpmp(np);
   cpmp(trialMax);
   return true;
}

bool NuggetData::scanFrom(istream& str)
{
   cpms(overlapFac);
   cpms(isotropyFac);
   cpms(np);
   cpms(trialMax);
   return true;
}

NuggetData::NuggetData(const RecordHandler& rch, const Word& sec)
{
// a working initialization even if no data are on file
   overlapFac=1.25;
   isotropyFac=0.9;
   np=3;
   trialMax=5000;
   Z ped=1;
   rch.read(sec,"overlapFac",overlapFac,ped);
   rch.read(sec,"isotropyFac",isotropyFac,ped);
   rch.read(sec,"np",np,ped);
   rch.read(sec,"trialMax",trialMax,ped);
}

NuggetData::NuggetData(const Record& rc, const Word& sec)
{
   RecordHandler rch(rc);
   NuggetData res(rch,sec);
   *this=res;
}
