//? tut2oldandgood.cpp
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*************************************************************************
   Studying the Kepler oscillator
*********************************************************************/   
#include <cpmbas1.h> 
   // Actually, this is a collecion of header files.

using namespace CpmRoot; 
using namespace CpmRootX; 
using namespace CpmSystem; 
using namespace CpmArrays; 
using namespace CpmGraphics;

using CpmLinAlg::R2;
using CpmGeo::Iv;
using CpmAlgorithms::solKepEqu;
using CpmAlgorithms::CyclicAlarm;
using CpmImaging::YELLOW;
using CpmImaging::GREEN;
using CpmImaging::darkSky;
using CpmImaging::twiSky;
using CpmImaging::morSky;

class KepOsc{ // 'Kepler oscillator'
// My name for the system defined as the radial part of the Kepler
// motion arround a fixed central mass. Physical units are used in which
// the moving mass, the gravitational potential at unit distance, and the
// angular momentum have value 1.
// see www.ulrichmutze.de/articles/99-271.pdf, Section 4

// Methodologic remarks:
// The primary aim here is to compare various integrators for the
// equation of motion for the physical system mentioned above.
// There are two strategies for implementing this aim:
// 1. Define for each integration method a separate class.
//    This is the normal way to be followed according to the pure doctrine
//    of object oiented programming. It is also very natural since the 
//    various integration methods to be considered need different state
//    data for implementing an integration step as a mapping of the
//    state space into itself.
// 2. Define each integration method as a separate member function
//    within a single class. This requires the state space to contain
//    data which are needed for implementing e.g. the leap-frog integrator
//    and are not needed e.g. for the asynchronous leap-frog integrator.
//    If data belong to a type for which the default constructor is 
//    computationally inexpensive such partially relevant data seem to
//    be acceptable and the present class implementation follows this 
//    path.

// data
// 1. common data   
   R t_,x_,v_; // time, position, velocity
   Z met_; // method selector for functon step_
// 2. additional data for implementing stepALF_(R dt) 
   R1 w_,a_; 
// 3. additional data for implementing stepLF_(R dt) 
   R1 tp_,xp_,vp_; // p for previous
   B dynSttCon_; // true means that the (dynamical) state is (fully) 
      // constructed
   Z1 nIter_; // needed in iniLF
   B stepRed_; // step reduction
// static functions
   static R f(R const& x){R xi=cpminv(x); return (xi-1)*xi*xi;}
      // force law
   static R2 FLF(R2 const& a){return R2(a[2],f(a[1]));}
      // implementation tool for function iniLF
// member functions
   void init_(Z met, R eps, R E); // assumes 0<=eps<1, E arbitrary
      // initializes the common data
   R2 iniLF(R2 const& a, R dt);
      // initialization for LF. The LF method is special in that it
      // the dynamical state depends on dt
   void stepDMI_(R dt); // direct midpoint integrator (2nd order)
      // needs only the common data
   void stepRK_(R dt); // Runge-Kutta (2nd order)
      // needs only the common data
   void stepALF_(R dt); // asynchronous leap-frog (2nd order)
      // needs data w_, a_ for implementation
      // See 'Fortsetzung von UM 10.9.08' in my notebook UM 2008a 
   void stepLF_(R dt); // leap-frog (2nd order)
      // needs tp_,xp_,vp_
public:
   Word nameOf()const{ return "KepOsc, mehod="&cpm(met_);}
   KepOsc(){init_(0,0,0);}
   explicit KepOsc(Z met, R eps, R E=0){init_(met,eps,E);}
   R t()const{ return t_;}
   R x()const{ return x_;}
   R v()const{ return v_;}
   R T()const{ return v_*v_*0.5;}
      //: T (kinetic energy)
   R V()const{ R xi=cpminv(x_); return xi*(xi*0.5-1);}
      //: V potential energy
   R eTot()const{ return T()+V();}
      //: e total
   R force()const{ return f(x_);}
      //: force
      // The superficially correct name 'for' is not correct since it is
      // a C++ key word.
   R_Vector orbDes()const;
      //: orbit descriptors
      // Returns a list of the usual orbit elements
      // a: major semi-axis, eps: numerical excentricity,
      // n: mean motion, E: excentric anomaly, M: mean anomaly,
      // tP: orbital period (revolution time)
   void rev_(){ v_=-v_;}
      //: reversed
      // Notice name ending in '_' since it is a non-constant function
   void extEvl_(R t, R acc=1e-8);
      //: exact evolution
   void step_(R dt) 
      //: step
      // The actual step algorithm depends on method parameters
   {
      if (met_==0) stepRK_(dt);
      else if (met_==1) stepDMI_(dt);
      else if (met_==2){
         if (stepRed_){
            stepALF_(dt*0.5);
            stepALF_(dt*0.5);
         }
         else stepALF_(dt);
      }
      else if (met_==3){
         if (stepRed_){
            stepLF_(dt*0.5);
            stepLF_(dt*0.5);
         }
         else stepLF_(dt);
      }
      else cpmerror("bad value of KepOsc::met_");
   }

   void shf_(R dx, R dv)
      //: shift
   {  
      x_+=dx; v_+=dv;
      if (met_==2){
         w_=v_;
         a_=f(x_);
      }
      else if (met_==3){
         xp_+=dx; 
         vp_+=dv;
        // dynSttCon_=B(false);
      }
      else ;
   }
      
   X3<R,Iv,Iv> orbSize()const;
      //: orbit size
      // intervals which contain the t_-range, x_-range , the y_range
      // for a full revolution

   Color col()const
      //: color for visualization
   { 
      if (met_==0) return WHITE;
      else if (met_==1) return YELLOW;
      else if (met_==2) return GREEN;
      else if (met_==3) return RED;
      else cpmerror("bad value of KepOsc::met_ in KepOsc::col()");
   }
      
   void mark(Graph& g)const{ g.mark(R2(x_,v_),col());}
      //: mark
 
   R2 phsPos()const{ return R2(x_,v_);}
      //: phase postion

   R per()const{ return orbDes()[6];} 
      //:period

   void set_nIter_(Z n){ nIter_=n;}
      //: set nIter

   void setStepRed_(bool b){ stepRed_=B(b);}
 };

void KepOsc::init_(Z met, R eps, R E)
// initializing x_ anv v_ from numerical excentricity and excentric
// anomaly
{
   Word loc("KepOsc::init_(R,R)");
   cpmassert(eps>=0,loc);
   cpmassert(eps<1,loc);
   met_=met;
   t_=0; 
   R ai=(1-eps)*(1+eps);
   R a=1./ai;
   x_=a*(1-eps*cpmcos(E));
   v_=eps*cpmsqrt(a)*cpmsin(E)/x_;
   if (met_==2){ // then we need w_ and a_
      w_=v_;
      a_=f(x_);
   }
}

R2 KepOsc::iniLF(R2 const& a, R dt)
{
   R2 a0=a, a1=a0;
   for (Z i=1;i<=nIter_;++i){
      a1=a0+(FLF(a1)+FLF(a0))*dt*0.5;
   }
   return a1;
}

R_Vector KepOsc::orbDes()const
{
   Word loc("KepOsc::orbDes()");
   Z mL=2;
   CPM_MA
   R e=eTot();
   
   if (e>=0){
      cpmcerr<<"We have e>=0, un-bound state. x_="<<x_<<" v_="<<v_<<endl;
      cpmdebug(V());
      cpmdebug(T());
   }
   cpmassert(e<0,loc); // we need a bound state
      // then also x>0.5
   R a=-0.5*cpminv(e);
   R eps=cpmsqrt(1.-1./a);
   R n=cpmpow(a,-1.5);
   R tP=cpm2pi/n;
   R z1=1-x_/a;
   R z2=x_*v_/cpmsqrt(a);
   C z(z1,z2);
   R E=z.arg();
   R M=E-eps*cpmsin(E);
   CPM_MZ
   return R_Vector("",a,eps,n,E,M,tP);
}

void KepOsc::extEvl_(R t, R acc)
// exact step with t arbitrarily large, based on the
// famous transcendental Kepler equation
{
   Word loc("KepOsc::extEvl_(R,R)");
   Z mL=2;
   CPM_MA
   R_Vector y=orbDes();
   R a=y[1],eps=y[2],n=y[3],M=y[5];
   M+=n*t;
   R E=solKepEqu(M,eps,acc);
   x_=a*(1-eps*cpmcos(E));
   cpmassert(x_>=0.5,loc); // for safty, not needed
   v_=eps*a*a*n*cpmsin(E)/x_;
   t_+=t;
   CPM_MZ
}

void KepOsc::stepDMI_(R dt)
// nice symmetry, f evaluated only once
{
   R tau=dt*0.5;
   t_+=tau;
   x_+=(v_*tau);
   v_+=f(x_)*dt;
   x_+=(v_*tau);
   t_+=tau;
}

void KepOsc::stepRK_(R dt)
// not so nice, involves two evaluations of f
{
   R tau=dt*0.5;
   R a=f(x_);
   R am=f(x_+v_*tau);
   R dv=am*dt;
   R dx=(v_+a*tau)*dt;
   x_+=dx;
   v_+=dv;
   t_+=dt;
}

void KepOsc::stepALF_(R dt)
{
   R tau=dt*0.5;
   R xm=x_+tau*w_;
   R am=f(xm);
   x_+=dt*(v_+tau*a_);
   w_=2*v_-w_+dt*a_;
   v_+=dt*am;
   a_=2*am-a_;
   t_+=dt;
}

void KepOsc::stepLF_(R dt)
{
   if (!dynSttCon_){
      tp_=t_;
      xp_=x_; // here xp_ is previous x_
      vp_=v_; // here vp_ is previous v_
      t_=tp_+dt;
      R2 rp(xp_,vp_);
      R2 r=iniLF(rp,dt);
      x_=r[1];
      v_=r[2];
      dynSttCon_=B(true);
   }
   else{
      R dt2=dt*2;
      R tn=tp_+dt2;
      R vn=vp_+f(x_)*dt2;
      R xn=xp_+v_*dt2;
      tp_=t_;
      vp_=v_;
      xp_=x_;
      t_=tn;
      v_=vn;
      x_=xn;
   }
}

X3<R,Iv,Iv> KepOsc::orbSize()const
{
   Z mL=2;
   Word loc("KepOsc::orbSize()");
   CPM_MA
   R_Vector y=orbDes();
   R a=y[1],eps=y[2],n=y[3],tP=y[6];
   R xMin=a*(1-eps);
   R xMax=a*(1+eps);
   R vMax=eps*a*a*n;
   CPM_MZ
   return X3<R,Iv,Iv>(tP,Iv(xMin,xMax),Iv(-vMax,vMax));
}

//////////////////////////////////////////////////////////////////////////

void cpmplot(V<V<R2> > const& dat, Word const& fileName)
{
   Word loc("cpmplot(V<V<R2>>,Word)");
   Z mL=2;
   CPM_MA
   if (fileName.isVoid()){
      CPM_MZ
      return;
   }
   Z i,j,md=dat.dim();
   V<Z> vd(md);
   for (i=1;i<=md;++i){
      vd[i]=dat[i].dim();
   }
   S<Z> sd(vd);
   cpmassert(sd.card()==1,loc);
   Z nd=sd[1];
   OFileStream ofs(fileName);
   
   for (i=1;i<=nd;++i){
      ofs()<<endl;
      for (j=1;j<=md;++j){
         R2 y=dat[j][i];
         ofs()<<y[1]<<" "<<y[2]<<"    "; 
      }
   }
   CPM_MZ
}

void app1(void) // orbit in phase space (v = momentum since m=1)
// simultaneously for a selected list of integration methods
{
   Word loc("app1");
   Color cb=darkSky; // night sky blue. Each task appi
   // has here its characteristic background color
   RecordHandler rch("app1.ini");
      // This says that we expect data input to be available in file
      // app1.ini in a diretory which is determined by entries in
      // cpmsystemdependencies.h and cpmconfig.ini.
   Word sec="data"; // names 'rch' and 'sec' are to be employed
      // for the macro cpmrh to work
   R eps=0.25, fac=1.2, nPerRev=16, tWaitFinal=1;
   Z nTot=1000,displayPeriod=1,nIter=4;
   bool stepRed=false;
   Word imageName,plotDataName;
   V<Z> methods;
// getting quantities from section 'data' of file app1.ini
   cpmrh(eps); // rh: read handler
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrh(nIter);
   cpmrhf(stepRed);
   cpmrh(fac);
   cpmrh(tWaitFinal);
   cpmrh(imageName);
   cpmrhf(plotDataName);
   cpmrh(displayPeriod);
      // now the program has all its data
   cpmrh(methods);
   S<Z> met(methods); // eliminates doublicates
   Z nm=met.dim();
   V<KepOsc> vk(nm); 
   for (Z i=1;i<=nm;++i){
      vk[i]=KepOsc(met[i],eps);
      vk[i].set_nIter_(nIter);
      vk[i].setStepRed_(stepRed);
   } 
   X3<R,Iv,Iv> os=vk[1].orbSize(); // cartesian product templates
      // are defined from X2 to X8
   R tP=os.c1(); // time for a full period
   cpmdebug(tP);
   R dt=tP/nPerRev; // time step to be used for integration
   Iv ivx=os.c2(); // Iv = interval is a valuable type
   cpmdebug(ivx);
   Iv ivy=os.c3();
   cpmdebug(ivy);
   ivx*=fac; // changing the size of an interval, while holding the center
      // fixed
   ivy*=fac; // same factor for the other direction
   V<V<R2> > vvr(nm,V<R2>(nTot)); // memory for all phase space positions
      // during run
   Graph g(cb); // basic graphical 'window to the world'
   g.setXY(ivx,ivy);
      // this associates physical sizes (in space) with the 
      // 'pixeled' rectangle g
   R nInv=1./nTot; // for use in cpmprogress
   CyclicAlarm cy(displayPeriod); // device for triggering periodic 
      // actions
   for (Z i=1;i<=nTot;++i){
      for (Z j=vk.b();j<=vk.e();++j){
         //vk[j].step_(dt);
         vk[j].mark(g); // 'mark' acts on memory only
         R2 pj=vk[j].phsPos();
         if (i==1) cpmdebug(pj);
         vvr[j][i]=pj;
         vk[j].step_(dt);
      }
      if (cy()||i==nTot){ // cy() becomes 'true' periodically 
         g.dis(); // transfers the whole memory frame
            // to screen. This expensive action should probably not be
            // orderd for each integration step. Object cy allows
            // to achieve this.
         cpmprogress("loop",i*nInv); // progress indicator on
            // pane 1 of he status bar
      }
   }
   g.wrt(imageName); // writes the final frame to a ppm image file
      // if imageName is not the void word.
   cpmplot(vvr,plotDataName);
   cpmwait(tWaitFinal,2); // allows us to inspect the last frame
      // for the time (in seconds) put in here. The second argument
      // says that on pane 2 of the status bar there will be a
      // 'countdown' indication of the time left for inspection.
}

void app2(void) // method error of the two integration methods
// represented by means of interaction picture dynamics. For 
// applying this terminology, we interpret the dynamics as defined by
// the numerical methods as resulting from the exact Kepler dynamics
// by 'interaction'. This interaction picture motion is much slower for 
// the more accurate direct midpoint method. The Runge Kutta motion
// finally explodes and reaches non-bound states for which the exact
// evolution back is no longer defined (by the method KepOsc::extEvl_).
// Therefore the option is implemented to deal only with the stable
// direct midpoint method.
{
   Word loc("app2");
   Color cb=twiSky; // background
   RecordHandler rch("app2.ini");
   Word sec="data";
   R eps=0.25, nPerRev=16, acc=1e-8, tWaitFinal=1;
   Z i,j,nTot=1000,nIter=4;
   Word imageName,plotDataName;
   V<Z> methods;
   bool backEvl=true;
   bool stepRed=true;
   cpmrh(eps);
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrhf(nIter);
   cpmrh(acc);
   cpmrh(tWaitFinal);
   cpmrh(imageName);
   cpmrhf(plotDataName);
   cpmrh(methods);
   cpmrhf(backEvl);
   cpmrhf(stepRed);

   S<Z> met(methods); // eliminates doublicates
   Z nm=met.dim();
   cpmassert(nm>0,"app2()");
   V<KepOsc> vk(nm); 
   for (Z i=1;i<=nm;++i){
      vk[i]=KepOsc(met[i],eps);
      vk[i].set_nIter_(nIter);
      vk[i].setStepRed_(stepRed);
   } 
   KepOsc k0=vk[1];
   X3<R,Iv,Iv> os=k0.orbSize();
   R2 p0=k0.phsPos();
   R xInv=cpminv(os.c2().abs());
   R vInv=cpminv(os.c3().abs());

   V<V<R2> > vvr(nm,V<R2>(nTot));
 
   R tP=k0.per(); // period
   R dt=tP/nPerRev; // time step

   R nInv=1./nTot;
   for (i=1;i<=nTot;++i){ // no graphical actions
      // within the loop. Terribly fast!!!
      KepOsc k1Act=vk[1];
      R tAct=k1Act.t(); // the states are aware of the time to which they
         // belong: Schroedinger picture 
      if (!backEvl){
         KepOsc kExt(k0);
         kExt.extEvl_(tAct,acc); // exact evolution
         R2 posExt=kExt.phsPos(); // storing the exact phase space
         // position
         for (j=vk.b();j<=vk.e();++j){
            R2 diff=vk[j].phsPos()-posExt;
            diff[1]*=xInv;
            diff[2]*=vInv;
            vvr[j][i]=diff;
         }
      }
      else{
         for (j=vk.b();j<=vk.e();++j){
            KepOsc vkj(vk[j]);
            vkj.extEvl_(-tAct,acc);
            R2 diff=vkj.phsPos()-p0;
            diff[1]*=xInv;
            diff[2]*=vInv;
            vvr[j][i]=diff;
         }
      }
      for (j=vk.b();j<=vk.e();++j){
         vk[j].step_(dt);
      }
      cpmprogress("loop",i*nInv,1,2); // only 2 digits for 
         // progress indication to avoid 'much adoo about nothing'
   }
   Graph g(cb);
   g.clr_(cb);
   g.setAutoScale(true);
   g.setNoGridText(false);
   g.setNoGrid(true);
   g.setGridTextColor(LIGHTBLUE);
   V<Color> vc(nm);
   for (j=1;j<=nm;++j) vc[j]=vk[j].col();
   g.mark(vvr,vc); // powerful method for graphical
      // representation of polygons; here one or two 
      // of those
   g.dis();
   g.wrt(imageName); // does nothing if file name void
   cpmplot(vvr,plotDataName);
   cpmwait(tWaitFinal,2);
}

void app3(void) // interaction picture dynamics for not a single
// point as an initial condition but for an ensemple of 
// phase space points which form a closed curve. The deformation
// of this closed curve by interaction picture dynamics is monitored.
// Here the field of view in phase space follows the moving states.
// Otherwise they would slowly drift out. Since the motion is much
// more violent for RK, the two methods are not shown together.
// One has to select the method that one wants to study. 
{
   Word loc("app3");
   Color cb=morSky; // background
   RecordHandler rch("app3.ini");
   Word sec="data";
   R eps=0.25, fac=1.2, nPerRev=16, acc=1e-8, E=0;
   R facX=1, facV=1, tWaitFinal=1, tWaitFrame=0.1;
   Z nTot=1000, nK=12, displayPeriod=1, method=1;
   Word imageName;

   cpmrh(eps);
   cpmrh(E);
   cpmrh(fac);
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrh(acc);
   cpmrh(tWaitFinal);
   cpmrh(tWaitFrame);
   cpmrh(nK);
   cpmrh(facX);
   cpmrh(facV);
   cpmrh(displayPeriod);
   cpmrh(method);
   cpmrhf(imageName);
   
   KepOsc ko(method,eps,E);
   ko.setStepRed_(true); // experiment
   R x0=ko.x();
   R v0=ko.v();
   X3<R,Iv,Iv> os=ko.orbSize();
   R tP=os.c1();
   R dt=tP/nPerRev;

   KepOsc kt(ko);
   kt.step_(dt);
   R xt=kt.x();
   R vt=kt.v();
   R dx=xt-x0;
   R dv=vt-v0;
   R r=R2(dx,dv).abs();
   R rx=r*facX;
   R rv=r*facV;
   V<KepOsc> vk(nK,ko);
   
   R phi=0;
   R dphi=cpm2pi/(nK-1);
   R_Vector xx(nK);
   R_Vector vv(nK);
   
   Z i,j;
   for (i=vk.b();i<=vk.e();++i){
      vk[i].shf_(rx*cpmcos(phi),rv*cpmsin(phi));
      xx[i]=vk[i].x();
      vv[i]=vk[i].v();
      phi+=dphi;
   }
   R_Vector xx0=xx;
   R_Vector vv0=vv;

   R rf=r*fac;
   Iv ivx(-rf,rf);
   Iv ivy=ivx;
   Graph g(cb);
   g.setXY(ivx,ivy);
   R nInv=1./nTot;
   Z nPrev=nTot-1;
   CyclicAlarm cy(displayPeriod);
   for (i=1;i<=nTot;++i){
      for (j=vk.b();j<=vk.e();++j){
         vk[j].step_(dt);
         R tAct=vk[j].t();
         KepOsc koAct=vk[j];
         koAct.rev_();
         koAct.extEvl_(tAct,acc);
         R xxj=koAct.x();
         R vvj=koAct.v();
         xx[j]=xxj;
         vv[j]=vvj;
      }
      g.draw(xx.subMean(),vv.subMean(),ko.col());
      if (i<nPrev){
         if (cy()){
            g.dis();
            cpmprogress("loop",i*nInv);
            cpmwait(tWaitFrame);
            g.clr_(cb);
         }
      }
      else if (i==nPrev) g.clr_(cb); // that only a single 
         // drawing is on the final display
      else{
         R redFac=0.5; // color of initial configuration should 
            // be a bit less bright
         Color ci=ko.col()*redFac;
         g.draw(xx0.subMean(),vv0.subMean(),ci); // the final
         // image should also show the initial configuration
         g.dis(); // display 
         cpmprogress("loop",i*nInv);
      }
   }
   g.wrt(imageName); // does nothing if file name is void
   cpmwait(tWaitFinal,2);
}

void info()
{
   cout<<"Studying the Kepler oscillator"<<endl
   <<" This function takes 0 to 1 integer arguments; values 1 2 3"<<endl
   <<" trigger one of the three basic modes, each other number"<<endl
   <<" triggers execution of these modes in succession."<<endl
   <<" No argument is equivalent to an argument of value 1.";
}

Z CpmApplication::main_(void)
// Selects the task from the command line, if there is a numerical 
// argument following the function name.
// If there is no such argument, the program tries to find tut2.ini
// (actually title_&".ini") Places to search for this are determined by 
// cpmsystemdependencies.h and cpmconfig.ini, where the latter enjoys
// priority.
// If no tut2.ini can be found, the coded default selection takes
// place.
// The functionality of the tasks app1(), app2(), app3() is controlled
// through files ap1.ini, app2.ini, app3.ini, for which search places 
// are determined by the same logic as those of tut2.ini.
{
   title_="tut2";
      // Important! This lets the program try to get input from
      // a file tut2.ini although the name of the executable as 
      // coming out of the building process may be, say, tut2MS.exe.
   Z mL=1;
   Z sel=1; // default selection
   cpmverbose=1;
   if (args_.valInd(2)){ // arrays allow testing the validity
         // of indexes without running into exceptions
      Word w2=args_[2];
      if (w2=="?"){
         info();
         return 0;
      }
      else sel=w2.toZ(); // conversion to Z (integer)
   }
   Word loc("CpmApplication::main_(), sel="&cpm(sel));
   CPM_MA
   
   if (sel==1) app1();
   else if (sel==2) app2();
   else if (sel==3) app3();
   else { // do everything sequentially
      app1();
      app2();
      app3();
   }
   CPM_MZ
   return sel;
}

// end of tut2.cpp
