//? cpmvo.h
//? C+- by Ulrich Mutze. Status of work 2018-12-01.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_VO_H_
#define CPM_VO_H_
/*
   Purpose: Defining a array class Vo<T> which exploits the assumption\
    that
   T supports ordering. See cpmv.h

*/
#include <cpmv.h>
#include <algorithm>

namespace CpmArrays{

///////////////////////// class Vo ///////////////////////////////////

template <class T>
   // We assume that T provides (explicitly or implicitly)
   // copy constructor, and assignment or that T is a built-in type.
   // T < T and T > T assumed.

class Vo : public V<T>{ // version of V with order-related operations
   // derivation which implements the order interface

   typedef Vo<T> Type;
   typedef V<T> Base;
public:

explicit Vo(Z n=0):V<T>(n){}

explicit Vo(IvZ ivz):V<T>(ivz){}

   Vo(Z n, T const& t):V<T>(n,t){}

   Vo(const V<T>& h):V<T>(h){}
         // 'down-cast'-constructor

   virtual V<T>* clone(void)const{ return new Vo(*this);}

   virtual Word nameOf()const
      //: name of
   {
      Word wi="Vo<";
      return wi&CpmRoot::Name<T>()(T())&">";
   }

   V<Z> permutationForIncreasingOrder(void)const;
      // Returned is an array per such that i<j enforces
      //  (*this)[per[i]]<(*this)[per[j]]
      //       or
      // (*this)[per[i]]==(*this)[per[j]]
      // Assumes that valid indexing starts with 1 (i.e. b()==1)

   V<Z> permutationForDecreasingOrder(void)const
      // analogous to previous function
   { return permutationForIncreasingOrder().rev();}

   void order(void);
      // Permutes the indexes such that
      // (*this)[i]<(*this)[j] || (*this)[i]==(*this)[j]
      // whenever i<j. The components (instances of T) are allowed to
      // be large objects since excessive copying is avoided.

   Vo<T>& sort_(void){ order(); return *this;}
      //: sort

   void sortInc_();

   V<Z> perForDecOrd(void)const
      //: permutation for decreasing order
   { return permutationForDecreasingOrder();}

   V<Z> perForIncOrd(void)const
      //: permutation for increasing order
   { return permutationForIncreasingOrder();}

   Vo<T> increasingCopy(void)const;
      // returns a increasingly ordered version of (*this) without
      // changing *this

   Vo<T> incCopy(void)const{ return increasingCopy();}
      //: increasing copy

   Vo<T> strictlyIncreasingVersion(void)const;
      // returns a stricly increasingly ordered version of (*this)
      // which leaves out multiply appearing components
      // (again, without changing *this)

   Vo<T> strIncVer(void)const{ return strictlyIncreasingVersion();}
      //: strictly increasing version

   Vo<T> decreasingCopy(void)const;
      // returns a decreasingly ordered version of (*this) without
      // changing *this

   Vo<T> decCopy(void)const{ return decreasingCopy();}
      //: decreasing copy

   Z find(T const&, bool ordered=true)const;
      //: find
      // if the return value is 0, there is no i=1,...n
      // such that (*this)[i]==t. Otherwise, the return value i is
      // such that (*this)[i]==t.
      // If ordered==true, this assumes that (*this) is ordered in the
      // way achieved by function order(), else no assumptions on
      // ordering. The components are compared against t starting from 1
      // (to dim()). The first occurrence of t gets reported.

   Vo<T> purge(const V<T>& s)const;
      // returned is a list which is obtained from *this by eliminating
      // all components which equal to one of the components of s. Beside
      // of this removal, the order of the components in *this is
      // retained.

   Vo<T> pur(const V<T>& s)const{ return purge(s);}
      //: purge

   Z locate(T const&)const;
      // returned is a value j such that t satisfies
      //    (*this)[j] <= x < (*this)[j+1].
      // (*this) must be monotonic increasing
      // j=b()-1 or j=e()+1 is returned to indicate that t is out of
      // range.

   Z loc(T const& t)const{ return locate(t);}
      //: locate

   Vo<T> enqueue(T const&, bool asc=true)const;
      // assumes that (*this) is monotonic. Then t will be inserted in
      // the orderly correct position into of a copy of (*this). This copy
      // remains monotonic and gets returned.
      // The second argument comes only in action for dim()==1, since
      // then no direction is provided by *this and a direction is needed
      // for the result (for dim()==0, the result has dimm()==1 and no
      // direction).

   Vo<T> enq(T const& t, bool asc=true)const{ return enqueue(t,asc);}
      //: enqueue

   T sup()const;
      //: supremum
      // returns the largest component of *this; faster than getting the
      // largest element by ordering

   T inf()const;
      //: infimum
      // returns the smallest component of *this; faster than getting the
      // largest element by ordering

   Z indSup()const;
      //: index (of) supremum
      // returns the smallest i for which (*this)[i]=(*this).sup()

   Z indInf()const;
      //: index (of) infimum
      // returns the smallest i for which (*this)[i]=(*this).inf()

   T2<Z> indInfSup()const;
      //: returns indInf(), indSup() as a pair in a single pass

   T2<T> infSup()const;
      //: infimum supremum
      // Let res be the return value, then res.first==inf(),
      // res.second==sup(). Fast single-pass operation
};


// Function indexx from Press et al. changed into a function template
// p. 339  (role of M: see p. 333; algorithm is based on Quicksort, see
// sort(), anyway the choice of M may influence performance but never
// puts a limitation on the size of the vector to be ordered.

namespace{
   inline void swapZ(Z& i, Z& j){ Z ii=i; i=j; j=ii;}
}

template <class T>
V<Z> Vo<T>::permutationForIncreasingOrder(void)const
// Indexes an array v[1..n], i.e. outputs the array indx[1..n] such that
// v[indx[j]] is in ascending order for j=1,2,...,n. The input
// quantities n and v are not changed. Especially, v-elements are
// not copied back and forth which is essential if T is a large class
{
   const Z mL=4;
   cpmmessage(mL,"Vo<T>::permutation...(): started");
   const Z M=7;
   const Z NSTACK=50;

   Z nVec=Base::dim();
   Vo<Z> indx(nVec);
   Z i,indxt,ir=nVec,j,k,l=1,jstack=0;
   T a;

   Vo<Z> istack(NSTACK);

   for (j=1;j<=nVec;j++) indx[j]=j;
   for (;;) {
      if (ir-l < M) {
         for (j=l+1;j<=ir;j++) {
            indxt=indx[j];
            a=(*this)[indxt];
            for (i=j-1;i>=1;i--) {
               if ((*this)[indx[i]] < a) break;
               if ((*this)[indx[i]] == a) break;
               indx[i+1]=indx[i];
            }
            indx[i+1]=indxt;
         }
         if (jstack == 0) break;
         ir=istack[jstack--];
         l=istack[jstack--];
      } else {
         k=(l+ir)/2; // was >> 1 instead of /2 before
         swapZ(indx[k],indx[l+1]);
         if ((*this)[indx[l+1]] > (*this)[indx[ir]]) {
            swapZ(indx[l+1],indx[ir]);
         }
         if ((*this)[indx[l]] > (*this)[indx[ir]]) {
            swapZ(indx[l],indx[ir]);
         }
         if ((*this)[indx[l+1]] > (*this)[indx[l]]) {
            swapZ(indx[l+1],indx[l]);
         }
         i=l+1;
         j=ir;
         indxt=indx[l];
         a=(*this)[indxt];
         for (;;) {  // the i<nVec and j>1 test added by UM, otherwise in
               // patological cases of < and > relations we get a range
               // error
            do i++; while (i<nVec && (*this)[indx[i]] < a);
            do j--; while (j>1 && (*this)[indx[j]] > a);
            if (j < i) break;
            swapZ(indx[i],indx[j]);
         }
         indx[l]=indx[j];
         indx[j]=indxt;
         jstack += 2;
         if (jstack > NSTACK) cpmerror("NSTACK too small in ordering()");
         if (ir-i+1 >= j-l) {
            istack[jstack]=ir;
            istack[jstack-1]=i;
            ir=j-1;
         } else {
            istack[jstack]=j-1;
            istack[jstack-1]=l;
            l=i;
         }
      }
   }
   cpmmessage(mL,"Vo<T>::permutation...(): done");
   return indx;
}

// Function locate from Press et al. changed into a function template

//template <class T>
//Z Vo<T>::locate(const T& x)const
//{
//   Z mL=3;
//   Word loc("Z Vo<T>::locate(const T& x)const");
//   CPM_MA
//   Z jl=Base::b()-1;
//   Z ju=Base::e()+1;
//   bool ascnd=(Base::last() > Base::fir());
//   while (ju-jl > 1) {
//      Z jm=(ju+jl)/2;// Press et al. have >>1 istead of /2
//      if ((x > (*this)[jm]) == ascnd)
//         jl=jm;
//      else
//         ju=jm;
//   }
//   CPM_MZ
//   if (jl<Base::e()&& x==(*this)[jl+1]) return jl+1;
//      // Press et al. only assert that "x is between (*this)[j] and
//      // (*this)[j+1]" if j is the return value of the function
//      // So we have to specially treat the case that it is at the
//      // upper end
//   else return jl;
//}

template <class T>
Z Vo<T>::locate(T const& x)const
{
   Z mL=3;
   Word loc("Z Vo<T>::locate(const T& x)const");
   CPM_MA
   Z nc=Base::sz_;
   cpmassert(nc>0,loc);
   Z jl=Base::b()-1;
   Z ju=Base::e()+1;

   T xf=Base::fir();
   T xl=Base::last();
   if (x<xf){
      CPM_MZ
      return jl;
   }
   if (x>xl){
      CPM_MZ
      return ju;
   }
   T* q=std::upper_bound(&Base::p_[0],&Base::p_[nc],x);
   Z ir=Base::b()+(q-Base::p_)-1;
   CPM_MZ
   return ir;
}

// functions based on order relations in T

template <class T>
Vo<T> Vo<T>::increasingCopy(void)const
{
   Vo<Z> per=permutationForIncreasingOrder();
   Z i, n=Base::dim();
   Vo<T> y(n);
   for (i=1;i<=n;i++) y[i]=(*this)[per[i]];
   return y;
}

namespace{
   template <class T>
   bool fEqual(T const& t1, T const& t2){ return t1==t2;}
}

template <class T>
Vo<T> Vo<T>::strictlyIncreasingVersion(void)const
{
   Vo<T> res=increasingCopy();
   return res.condense(fEqual).c1();
}

template <class T>
Vo<T> Vo<T>::decreasingCopy(void)const
{
   Vo<Z> per=permutationForIncreasingOrder();
   Z i, n=Base::dim();
   Vo<T> y(n);
   for (i=1;i<=n;i++) y[1+n-i]=(*this)[per[i]];
   return y;
}

template <class T>
void Vo<T>::sortInc_()
{
   Base::cow_();
   std::sort(&Base::p_[0], &Base::p_[Base::sz_]);
}

//template <class X>
//void V<X>::sortInc_(B b)
//{
//   class cx{ // compare X
//      Z z_;
//   public:
//      cx(B bb):z_(bb ? -1 : 1){}
//      bool operator()(X const& x1, X const& x2 )
//      { return x1.c(x2)==z_;}
//   };
//   cx cp(b);
//   cow_();
//   std::sort(&p_[0], &p_[nc()],cp);
//}

template <class T>
void Vo<T>::order(void){ sortInc_();}
//{
//   V<Z> per=permutationForIncreasingOrder();
//   Z i, n=Base::dim();
//   Vo<T> y(n);
//   for (i=1;i<=n;i++) y[i]=(*this)[per[i]];
//   for (i=1;i<=n;i++) (*this)[i]=y[i];
//}

//template <class T >
//Z Vo<T>::find(T const& t, bool ordered)const
//{
//   Z n=Base::dim();
//   if (n==0) return 0;
//   if (ordered){
//      // fast method
//      Z jl=0, jm, ju=n+1;
//      while (ju-jl > 1) {
//         jm=(ju+jl)/2; // was >> 1 instead of /2 before
//         if (t < (*this)[jm])
//            ju=jm;
//         else
//            jl=jm;
//      }
//      if (jl<=n){
//         if (jl<n && t==(*this)[jl+1]) return jl+1;
//         if (jl>0 && t==(*this)[jl]) return jl;
//      }
//      return 0;
//   }
//   else{
//      Z iRes=0; // correct result if the next loop doesn't find t
//      for (Z i=1;i<=n;i++){
//         if ((*this)[i]==t){ iRes=i; break;}
//      }
//      return iRes;
//   }
//}

template <class T >
Z Vo<T>::find(T const& t, bool ordered)const
{
   cpmassert(ordered==true,"Z Vo<T>::find(T,bool)");
   Z n=Base::dim();
   if (n==0) return 0;
   Z nc=Base::sz_;
   T* q=std::find(&Base::p_[0],&Base::p_[nc],t);
   if (q==&Base::p_[nc]) return 0;
   return Base::b()+(q-Base::p_);
}

template <class T >
Vo<T> Vo<T>::purge(const V<T>& s)const
{
   Z i,j,n=Base::dim(),ns=s.dim();
   V<bool> vs(n,true);
   for (i=1;i<=n;i++){
      T t=(*this)[i];
      for (j=1;j<=ns;j++){
         if( t==s[j] ){ vs[i]=false; break; }
      }
   }
   return Base::select(vs);
}

template <class T >
T Vo<T>::sup()const
{
   Z n=Base::dim();
   cpmassert(n>0,"T Vo<T>::sup()const");
   T t,res=Base::fir();
   for (Z i=Base::b()+1;i<=Base::e();i++){
      t=Base::cui(i);
      if (t>res) res=t;
   }
   return res;
}

template <class T >
Z Vo<T>::indSup()const
{
   Z n=Base::dim();
   cpmassert(n>0,"Vo<T>::indSup()const");
   T t,valMax=(*this)[1];
   Z res=1;
   for (Z i=2;i<=n;i++){
      t=(*this)[i];
      if (t>valMax){ valMax=t; res=i;}
   }
   return res;
}

template <class T >
T Vo<T>::inf()const
{
   Z n=Base::dim();
   cpmassert(n>0,"T Vo<T>::inf()const");
   T t,res=Base::fir();
   for (Z i=Base::b()+1;i<=Base::e();i++){
      t=Base::cui(i);
      if (t<res) res=t;
   }
   return res;
}

template <class T >
Z Vo<T>::indInf()const
// needed as fast as possible
{
   Z n=Base::dim();
   cpmassert(n>0,"Z Vo<T>::indInf()const");
   Z res=0;
   T t,valMin=Base::cui(res);

   for (Z i=1;i<n;i++){
      t=Base::cui(i);
      if (t<valMin){ valMin=t; res=i;}
   }
   return ++res;
}

template <class T >
T2<Z> Vo<T>::indInfSup()const
{
   Z n=Base::dim();
   if (n<=0) cpmerror("Vo<T>::indInfSup(): dim()<=0");
   T t,valMin=(*this)[1],valMax=valMin;
   Z resMin=1;
   Z resMax=1;
   for (Z i=2;i<=n;i++){
      t=(*this)[i];
      if (t<valMin){ valMin=t; resMin=i;}
      if (t>valMax){ valMax=t; resMax=i;}
   }
   return T2<Z>(resMin,resMax);
}

/********************************
template <class T >
T2<T> Vo<T>::infSup()const
{
   Z n=Base::dim();
   if (n<=0) cpmerror("Vo<T>::infSup(): dim()<=0");
   T t,valMin=(*this)[1],valMax=valMin;
   for (Z i=2;i<=n;i++){
      t=(*this)[i];
      if (t<valMin){ valMin=t; }
      if (t>valMax){ valMax=t; }
   }
   return T2<T>(valMin,valMax);
}
*************************************/

template <class T >
T2<T> Vo<T>::infSup()const
{
   Z n=Base::dim();
   if (n<=0) cpmerror("Vo<T>::infSup(): dim()<=0");
   T t,valMin=Base::fir(),valMax=valMin;
   for (Z i=Base::b()+1;i<=Base::e();i++){
      t=Base::cui(i);
      if (t<valMin){ valMin=t; }
      if (t>valMax){ valMax=t; }
   }
   return T2<T>(valMin,valMax);
}

template <class T >
Vo<T> Vo<T>::enqueue(T const& t, bool asc)const
{
   Z mL=3;
   Word loc("Vo<T> Vo<T>::enqueue(T const& t, bool asc)const");
   CPM_MA
   Z d=Base::dim();
   if (d==0) return Vo<T>(1,t);
   if (d==1){
      Vo<T> res(2);
      T t1=(*this)[1];
      if (asc){
         if (t1<t){
            res[1]=t1;
            res[2]=t;
         }
         else{
            res[1]=t;
            res[2]=t1;
         }
      }
      else{
         if (t1<t){
            res[1]=t;
            res[2]=t1;
         }
         else{
            res[1]=t1;
            res[2]=t;
         }
      }
      return res;
   }
   Z i=locate(t);
   CPM_MZ
   return Base::ins(i+1,t);
}
///////////////////////// iterated templates ///////////////////////////
// describing multi-indexed quantities
//////////////////////////// class VVo /////////////////////////////////
// matrices

template <class T>
class VVo: public VV<T>{// version of VV with order-related operations

   typedef VV<T> Base;
   typedef VVo<T> Type;

public:
   CPM_ORDER
// constructors
   VVo(Z d1, Z d2):Base(d1,d2){}
   VVo( Z d1, Z d2, T const& t):Base(d1,d2,t){}
      // all components are equal to t
   VVo(void):Base(){}
   VVo(VV<T> const& x):Base(x){}
   VVo( V< V<T> > const& x):Base(x){}
};

template <class T >
Z VVo<T>::com(VVo<T> const& s)const
{
   Z d1=Base::dim(), d2=s.dim();
   if (d1<d2) return 1;
   else if (d1>d2) return -1;
   else{
      for (Z i=1;i<=d1;i++){
         Vo<T> ti=(*this)[i];
         Vo<T> si=s[i];
         Z ci=CpmRoot::comT<T>(ti,si);
         if (ci!=0) return ci;
      }
      return 0;
   }
}

///////////////////////// class VVVo ///////////////////////////////////
// tensors of rank 3

template <class T>
class VVVo: public VVV<T>{// version of VVV with order-related operations

   typedef VVV<T> Base;
   typedef VVVo<T> Type;

public:
   CPM_ORDER
// constructors
   VVVo(Z d1,Z d2,Z d3):Base(d1,d2,d3){}
   VVVo(Z d1,Z d2,Z d3,T const& t):Base(d1,d2,d3,t){}
      // all components are equal to t
   VVVo(void):Base(){}
   VVVo(VVV<T> const& x):Base(x){}
   VVVo( V< V< V<T> > > const& x):Base(x){}
};

template <class T >
Z VVVo<T>::com(VVVo<T> const& s)const
{
   Z d1=Base::dim(), d2=s.dim();
   if (d1<d2) return 1;
   else if (d1>d2) return -1;
   else{
      for (Z i=1;i<=d1;i++){
         VVo<T> ti=(*this)[i];
         VVo<T> si=s[i];
         Z ci=CpmRoot::comT<T>(ti,si);
         if (ci!=0) return ci;
      }
      return 0;
   }
}

///////////////////////// class VVVVo //////////////////////////////////
// tensors of rank 4

template <class T>
class VVVVo:public VVVV<T>{//version of VVVV with order-related operations

   typedef VVVV<T> Base;
   typedef VVVVo<T> Type;

public:
   CPM_ORDER
// constructors
   VVVVo(Z d1,Z d2,Z d3,Z d4):Base(d1,d2,d3,d4){}
   VVVVo(Z d1,Z d2,Z d3,Z d4,T const& t):Base(d1,d2,d3,d4,t){}
      // all components are equal to t
   VVVVo(void):Base(){}
   VVVVo(VVVV<T> const& x):Base(x){}
   VVVVo( V< V< V< V<T> > > > const& x):Base(x){}
};

template <class T >
Z VVVVo<T>::com(VVVVo<T> const& s)const
{
   Z d1=Base::dim(), d2=s.dim();
   if (d1<d2) return 1;
   else if (d1>d2) return -1;
   else{
      for (Z i=1;i<=d1;i++){
         VVVo<T> ti=(*this)[i];
         VVVo<T> si=s[i];
         Z ci=CpmRoot::comT<T>(ti,si);
         if (ci!=0) return ci;
      }
      return 0;
   }
}

} // namespace

#endif
